local ir = require "lua_iragentlink"
local apiModel = require "api.model"
local models = apiModel:GetModels()
local utils = require "shared.apps.utils"
local api_utils = require "api.utils"
require "uuid"

module(..., package.seeall, inconcert.orbit.application.Decorate)

local function makeBatchId (account, page, batchId)
	return account .. (not api_utils.empty(page) and (":" .. page) or "") .. ":" .. batchId
end

local function GetBatch (web, request)
	local batchFullId = makeBatchId(request.account, request.page, request.batchId)
	local batch = models.BatchManagement.GetBatch(batchFullId)
	if batch then
		return api_utils.replyOk(web, { batch = batch })
	else
		return api_utils.replyNotFound(web, "batch not found")
	end
end

local function Search (web, request)
	local vcc = inconcert.authUser.virtualContactCenter
	request.sort = { { [request.sort] = { order = request.desc and "desc" or "asc" } } }
	request.desc = nil
	request.query.account_name = nil
	request.query.campaign = not api_utils.empty(request.query.campaign) and (utils.AppendVcc(request.query.campaign, vcc)) or nil

	local batchesController = require("apps.batch_management.controllers.batches")
	local skipViewPermission = true
	local allCampaigns = true
	local datsource = models.BatchManagement.GetDatasource()
	local elasticSearchConnectionConfig = api_utils.GetApiConnectionConfig("elasticsearch")
	local status, result = batchesController.SearchBatchesHelper(request, skipViewPermission, allCampaigns, datsource, elasticSearchConnectionConfig)

	return status == 200 and api_utils.replyOk(web, result) or api_utils.replyError(web, status, result.reason)
end

local function CreateBatch (web, request)
	local addresses = {}
	if addresses then
		for _, address in pairs(request.addresses) do
			if not api_utils.empty(address.address) then
				address.data = {
					name			= address.name,
					data			= address.data,
					agentData		= address.agentData,
					reportingGroup	= request.reportingGroup
				}
				address.agentData = nil
				table.insert(addresses, address)
			end
		end
	end

	local batch = {}
	if request.message and request.message.msgtype and request.message.msgtype == "template" and request.message.template then
		batch.msgtype = request.message.msgtype
		batch.template = request.message.template
		batch.text = request.message.template and request.message.template.text or ""
	else
		batch.body = {
			{content = request.message and request.message.text or "", type ="text/plain"}
		}
		batch.subject = request.message and request.message.subject or ""
		batch.text = request.message and request.message.text or ""
	end
	batch.id = request.batchId
	batch.addresses = addresses
	
	--send to ir
	local vcc = inconcert.authUser.virtualContactCenter
	local campaignAtVCC = utils.AppendVcc(request.campaign, vcc)
	local requestId = uuid.new()
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:NewOutgoingBatch({
		proxyType		= request.type,
		campaignAtVCC	= campaignAtVCC,
		account			= request.account,
		page			= request.page,
		batch			= batch,
		apiRequest		= requestId
	})
	result = api_utils.processAsyncResult({
		id = requestId, 
		result = result,
		async = request.async,
		timeout = request.timeout,
		pollingSeconds = request.pollingSeconds
	})
	local status, reason = api_utils.getHttpStatus(result)
	if status ~= 200 then
		LogError("[API:batch_management:CreateBatch] %s - %s", json.encode(request), reason)
		return api_utils.replyError(web, status, reason)
	end
	LogInfo("[API:batch_management:CreateBatch] %s - created", json.encode(request))
	return api_utils.replyOk(web)
end

local function GetBatchAddress (web, request)
	local batchFullId = makeBatchId(request.account, request.page, request.batchId)
	local address = models.BatchManagement.GetAddressStatus(batchFullId, request.address)
	if address then
		return api_utils.replyOk(web, { address = address })
	else
		return api_utils.replyNotFound(web, "address not found")
	end
end

local function GetBatchAddresses (web, request)
	local batchFullId = makeBatchId(request.account, request.page, request.batchId)
	local addresses = models.BatchManagement.GetAddressesStatus(batchFullId, nil, request.addresses)
	if addresses then
		return api_utils.replyOk(web, { addresses = addresses })
	else
		return api_utils.replyNotFound(web, "addresses not found")
	end
end

local function matchAddressFilter (filter, data)
	return  (not filter.status or data.status == filter.status) and
			(not filter.from_status_timestamp or data.status_timestamp >= filter.from_status_timestamp) and
			(not filter.to_status_timestamp or data.status_timestamp <= filter.to_status_timestamp)
end

local function SearchBatchAddresses (web, request)
	LogDebug("[batch_management:SearchBatchAddresses] Query Filters request: %s",json.encode(request))
	local startTime = os.time()

	local batchFullId = makeBatchId(request.account, request.page, request.batchId)
	local addresses = models.BatchManagement.GetAddressesStatus(batchFullId, nil, request.addresses)
	if not addresses then
		return api_utils.replyNotFound(web, "addresses not found")
	end

	local filteredAddresses = {}
	local totalUnfilteredRows = 0
	local totalAdded = 0
	for address, data in pairs(addresses) do
		data.address = address
		data.status_timestamp = data["timestamp_" .. data.status] or data.timestamp
		if matchAddressFilter(request.query, data) then
			table.insert(filteredAddresses, data)
			totalAdded = totalAdded + 1
		end
		totalUnfilteredRows = totalUnfilteredRows + 1
	end
	
	local response = {
		rows = filteredAddresses,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			return tostring(row[request.sort]):upper()
		end,
		desc = request.desc
	}
	request.__attr_id = "address"
	local resultData = utils.MakeSearchResponse(response, request, responseParams)
	
	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogDebug("[API:batch_management:Search] Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", 
		#resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)
	return resultData
end

local function AddAddressToBatch (web, request)
	--send to ir
	local vcc = inconcert.authUser.virtualContactCenter
	local campaignAtVCC = utils.AppendVcc(request.campaign, vcc)
	local requestId = uuid.new()
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:AddAddressToOutgoingBatch({
		proxyType		= request.type,
		campaignAtVCC	= campaignAtVCC,
		account			= request.account,
		page			= request.page,
		batchId			= request.batchId,
		address			= request.address,
		message			= request.message,
		data			= {
			name			= request.name,
			data			= request.data,
			agentData		= request.agentData,
			reportingGroup	= request.reportingGroup
		},
		apiRequest		= requestId
	})
	result = api_utils.processAsyncResult({
		id = requestId, 
		result = result, 
		async = request.async,
		timeout = request.timeout,
		pollingSeconds = request.pollingSeconds
	})
	local status, reason = api_utils.getHttpStatus(result)
	if status ~= 200 then
		LogError("[API:batch_management:AddAddressToBatch] %s - %s", json.encode(request), reason)
		return api_utils.replyError(web, status, reason)
	end
	LogInfo("[API:batch_management:AddAddressToBatch] %s - added with status %s", json.encode(request), result.status)
	return api_utils.replyOk(web, { messageStatus = result.status })
end

--[[----------------------------------------------------------------------
	Mapeo de funciones que quedan disponibles en módulo para luego ser utilizadas en el mapeo de funciones con las urls (/definition/ARCHIVO.lua función "urlLinkedFunction")
--------------------------------------------------------------------------]]
_M.GetBatch             = GetBatch
_M.Search               = Search
_M.GetBatchAddress      = GetBatchAddress
_M.GetBatchAddresses    = GetBatchAddresses
_M.SearchBatchAddresses = SearchBatchAddresses
_M.CreateBatch          = CreateBatch
_M.AddAddressToBatch    = AddAddressToBatch
-------------------------------------------------------------------------

function MapUris( m, authFilter, prefix )
	local f = authFilter
	local p = prefix
	local Log = require("apps.agent.requestFilters").LogRequest
	local jsonValidator = utils.JsonValidator
	local vr = utils.JsonValidatorReply
	local r = utils.JsonReply
	-- mapeo de URLs dinamicas (definidas en \api\definition\full.lua)
	local replyPostFn = Log .. f .. vr
	--local replyGetFn = Log .. f .. r
	local apiType = "batch_management"
	api_utils:MapDynamicURIs(apiType, m, _M, replyPostFn) --, replyGetFn) --, Log)
	--
	m_htmlRender = htmlRender
end
