local utils = require "shared.apps.utils"
local api_utils = require "api.utils"
local apiModel = require "api.model"
local models = apiModel:GetModels()

module(..., package.seeall, inconcert.orbit.application.Decorate)

local m_reportAPI = nil


function ApiExecuteComposite (report,queryParams )
	local cassandra = inconcert.environment.api.cassandra
	m_reportAPI = require("apps.reports").GetApi(cassandra)
	return m_reportAPI:Execute(report, queryParams)
end

--[[
	Función encargada de ejecutar un reporte contra la APP "apps.reports" -> rocka "reporter".

	Si en el POST llegan filtros de CCI, la ejecución del Filtro NO se realiza en la rocka "reporter", sino que se realiza en esta misma función.
	Para ello se valida contra la información de la interacción que queda en "data.reportingData".

	Los datos a retornar son:
		{
			ids             = true, -- retorna solo los IDs dentro de un objeto
			fullData        = true, -- retorna toda la interaccion (completa)
			reportStructure = true, -- retorna la información tal cual la retornan los reportes
			reportOnlyData  = true  -- retorna la información como los reportes, pero solo los datos (no las columnas)
		}
]]--
function API_Execute (reportId, reportData)
	-- el tipo de retorno de datos por defecto son interacciones completas
	local defaultReturnType = "ids"
	local returnInfoType = reportData.returnInfo and reportData.returnInfo.type or defaultReturnType
	local allowedReturnInfoType = {
		ids             = true, -- retorna solo los IDs dentro de un objeto
		fullData        = true, -- retorna toda la interaccion (completa)
		reportStructure = true, -- retorna la información tal cual la retornan los reportes
		reportOnlyData  = true  -- retorna la información como los reportes, pero solo los datos (no las columnas)
	}
	--LogDebug("+ ===========================")
	LogInfo("[API.reports:API_Execute] 1) Invoked API for ........ reportId: '%s', returnInfoType: '%s'.", reportId, returnInfoType)

	if not allowedReturnInfoType[returnInfoType] then
		LogWarning("[API.reports:API_Execute] Unrecognized return info type: %s.", returnInfoType)
		return false, {}
	end

	local cassandra = inconcert.environment.api.cassandra
	m_reportAPI = require("apps.reports").GetApi(cassandra)

	--if reportData.rowsPerPage then
		--reportData.resultSize = reportData.rowsPerPage
	--end
	reportData.rowsPerPage = reportData.rowsPerPage or 20
	reportData.resultSize = reportData.rowsPerPage
	reportData.last_handle = reportData.lastRowOfPage or nil
	--reportData.rawData = true

	--LogDebug("[API.reports:API_Execute] 2) Invoked API for ........ reportData: '%s'.", json.encode(reportData))
	-- tengo que enviar una copia hacia quien invoca a los reportes, por las dudas que modifique los datos de la tabla original
	local copyOfReportDta = utils.DeepCopy(reportData)
	local success, content = m_reportAPI:Execute(reportId, copyOfReportDta)
	if not success and content == "other_node" then
		return success, content
	end
	--LogDebug("[API.reports:API_Execute] 3) Invoked API for ........ success: '%s', content: %s.", success, json.encode(content))

	-- El tipo devuelto por defecto es (returnInfoType="interactions")
	--
	if returnInfoType == "reportStructure" then
		return success, content
	end

	--[[
		Funcion que valida si la interacción puede ser retornada al cliente.
	]]--
	local function CanIncludeInteraction (interaction, interactionId)
		if not interaction then
			-- NO Incluye la interacción, la misma NO existe en REDIS o CASSANDRA
			LogWarning("[API.reports:API_Execute.CanIncludeInteraction] Interaction does not exist: %s.", interactionId)
			return false
		end
		-- SI agrega la interacción
		return true
	end

	local ksInteractions = utils.GetDatasource():CassandraConnection("Interactions")
	local lasHhandle = content.last_handle
	local lastRowsInPage = 0

	--[[
		...
	]]--
	local function ProcessReportRows (reportRows, rowsToReturn, page)
		reportRows = reportRows or {}
		local totalRowsInPage = #reportRows
		page = page + 1

		--LogDebug("- - - - - - - - - - -")
		--LogDebug("[API.reports:API_Execute.ProcessReportRows] (1) Page %d starting process rows %d.", page, totalRowsInPage)
	
		if totalRowsInPage == 0 then
			-- si no existen más paginas
			--LogDebug("[API.reports:API_Execute.ProcessReportRows]   (1.1) Page %d no more rows.", page)
			return
		end

		-- recorro los rows del reporte
		for i, row in ipairs(reportRows) do
			local rowData
			-- recorro las columnas del reporte
			for j, columnValue in ipairs(row) do
				local columnInfo = content.result.columnsMetadata[j]
				local columnName = columnInfo.name
				if "reportOnlyData" == returnInfoType then
					rowData = rowData or {}
					rowData[columnName] = columnValue
				else -- ids o fullData
					-- "exported_content" en reportes de detalle live.
					-- "ID" en reportes de detalle history
					if columnName == "exported_content" or columnName == "ID" then
						local interactionId = columnValue
						-- obtengo la interacción de REDIS o CASSANDRA
						local interaction = models.Interactions.Get(interactionId, ksInteractions)
						-- ahora se valida si la interacción puede ser incluida.
						if CanIncludeInteraction(interaction, interactionId) then
							if "fullData" == returnInfoType then
								-- se devuelve toda la interacción
								rowData = interaction
							else
								-- se devuelve únicamente el id de la interacción
								rowData = {id = interaction.id}
							end
						end
						break
					end
				end
			end

			if rowData then
				-- incluyo la interacción en el array a retornar al cliente.
				table.insert(rowsToReturn, rowData)
			--else
				--LogWarning("[API.reports:API_Execute] Unable to include row: %s.", json.encode(row))
			end
		end
	end

	-- Comienzo la búsqueda de inforación contra Reportes (la primer página de datos)
	local page = 0
	local reportRows = content.result.rows or {}
	local rowsToReturn = {}
	ProcessReportRows(reportRows, rowsToReturn, page)

	ksInteractions:Close()
	--LogDebug("- ===========================")

	-- retorno los datos al cliente.
	local newContent = {
		result = {
			rows = rowsToReturn
		},
		last_handle = lasHhandle
	}
	return success, newContent
end

local function Execute (web, request)
    local encodedParams = web.input["post_data"]
	LogInfo("[API.reports:Execute] Invoked API for ........ post_data: '%s'.", encodedParams)
	LogInfo("[API.reports:Execute] Invoked API for ........ request: '%s'.", json.encode(request or {}))

	local cassandra = inconcert.environment.api.cassandra
	LogDebug("[API.reports:Execute] Invoked API for 0.1)...... cassandra: '%s'.", cassandra)
	m_reportAPI = require("apps.reports").GetApi(cassandra)
	LogDebug("[API.reports:Execute] Invoked API for 1)........ m_reportAPI: '%s'.", m_reportAPI)
	LogDebug("[API.reports:Execute] Invoked API for 2)........ m_reportAPI.Execute: '%s'.", m_reportAPI.Execute)

	--local encodedParams = [[{"utc_offset":180,"executed_by_user":true,"resultSize":20,"firstKey":"","queryFields":{"startDate":"2021-08-01 00:00:00","endDate":"2021-12-09 23:59:59","Type":"CHAT"},"last_handle":null,"pending_fields":["VCC"]}]]
	local data = json.decode(encodedParams)
	-- local reportId = "ReportDetailInboundChatsByCampaign"
	-- "ReportDetailInboundChatsByCampaign"
	local reportId = request.apiDefinition.config.reportId
	local success, content = m_reportAPI:Execute(reportId, data)
	if not success and content == "other_node" then
		return api_utils.replyOtherNode(web, "cache not found in this node")
	end

	return api_utils.replyOk(web, {content = content})
end

--[[----------------------------------------------------------------------
	Mapeo de funciones que quedan disponibles en módulo para luego ser utilizadas en el mapeo de funciones con las urls (/definition/ARCHIVO.lua función "urlLinkedFunction")
--------------------------------------------------------------------------]]
 _M.Execute = Execute
-------------------------------------------------------------------------

function MapUris (m, authFilter, prefix)
	local f = authFilter
	local p = prefix
	local Log = require("apps.agent.requestFilters").LogRequest
	local jsonValidator = utils.JsonValidator
	local vr = utils.JsonValidatorReply
	local r = utils.JsonReply

	-- mapeo de URLs dinamicas (definidas en \api\definition\full.lua)
	local replyPostFn = Log .. f .. vr
	local replyGetFn = Log .. f .. r
	local apiType = "reports"
	api_utils:MapDynamicURIs(apiType, m, _M, replyPostFn, replyGetFn) --, Log)
	--
	m_htmlRender = htmlRender
end
