local md5 = require "md5"
local lua_globaltimer = require "lua_globaltimer"
local apiModel = require "api.model"
local models = apiModel:GetModels()
local utils = require "shared.apps.utils"
local api_utils = require "api.utils"
local admUserController = require("apps.administration.controllers.users")
local allegro_proxy = require "apps.allegro.proxy"
local m_reportAPI = require "api.controllers.reports"
local m_adminModel
local m_helpers = require "helpers"
local validation_def = require("apps.administration.controllers.validations.user").GetDefinitions()
local m_apiType = "users"

module(..., package.seeall, inconcert.orbit.application.Decorate)

local function Search (web, request)
	--Desde Api, permitir validar si los usuarios tienen permiso sobre una o varias APP
	--(Esto pq se agrego en algun momento una validacion para no validar permisos cuando el user es de api)
	if request and request.query and request.query.Application then
		if type(request.query.Application) == "table" and #request.query.Application > 0
		then
			request.skipCanValidationPermission = true
		else
			request.query.Application = nil
		end
	end
	--LogInfo("[API:users:Search]  request %s", json.encode(request))
	local result = admUserController.SearchUsers(request)
	return api_utils.replyOk(web, result)
end

local function Get (web, userId)
	--LogDebug("userId: %s", userId)
	local _, vcc = utils.SeparateVCC(userId)
	local hasPermission, reason = api_utils.CheckOccPermissions(m_apiType, "GetUser", vcc)
	if not hasPermission then
		return api_utils.replyForbidden(web, reason)
	end
	local result = admUserController.Get(web, userId)
	if result.result then
		LogError("[API:users:Get] %s - %s", userId, result.result)
		return api_utils.replyError(web, web.status, result.result)
	end
	return api_utils.replyOk(web, { user = result })
end

local function GetUsersExportInfo (web)
	local vcc = inconcert.authUser.virtualContactCenter
	m_adminModel = m_adminModel or require "apps.administration.model"
	if not m_adminModel.IsCreatedModel() then
		m_adminModel:BuildModels()
	end
	local result = admUserController.GetExportInfo(web)
	LogInfo("[API:users:GetUsersExportInfo] VCC %s - result %s", vcc, json.encode(result))
	if result.result then
		return api_utils.replyError(web, web.status, result.result)
	end
	return api_utils.replyOk(web, { users = result })
end

local function Save (web, request)
	--encode password
	if request.Password then
		request.PlainPasswordFromApi =  request.Password
		request.Password = md5.sumhexa(request.Id:lower() .. request.Password)
	end
	-- LogDebug("[API.users:Save] Saving User: %s", json.encode(request))
	local isValidFields, errorVal = m_helpers.validation:StartValidateJsonValues(request, validation_def.basic)
	if not isValidFields then
		errorVal.result = "ERROR_FIELDS_VALIDATION"
		return api_utils.replyError(web, 400, errorVal)
	end
	local keyspace = utils.GetDatasource():CassandraConnection("Agent")
	local keyspaceHistorical = utils.GetDatasource():CassandraConnection( "Historical" )
	local success, result = admUserController.SaveOneUser(request, keyspace, nil, nil, keyspaceHistorical)
	keyspaceHistorical:Close()
	keyspace:Close()
	if not success then
		local rawErr = result
		--LogError("[API:users:Save] %s\n%s", json.encode(request), json.encode(rawErr))
		return api_utils.replyError(web, 400, (rawErr.msg or rawErr.reason), "SaveUser", rawErr)
	end
	return api_utils.replyOk(web)
end

local function Activate (web)
	local result = admUserController.Activate(web)
	if result.result ~= "OK" then
		return api_utils.replyError(web, web.status, result.result)
	end
	return api_utils.replyOk(web)
end

local function Deactivate (web)
	local result = admUserController.Deactivate(web)
	if result.result ~= "OK" then
		return api_utils.replyError(web, web.status, result.result)
	end
	return api_utils.replyOk(web)
end

local function GetAvailableTelephonyPeers(web)
	local result, err = allegro_proxy.administration.LoadAvailablesPeers()
	if not result then
		local reason = type(err) == "table" and err.reason or err or "ERROR_LOADING_AVAILABLES_PEERS"
		LogError("[API:users:GetAvailablePeers] - %s", reason)
		return api_utils.replyError(web, 500, reason)
	end
	return api_utils.replyOk(web, { peers = result.rows or {} })
end

--[[
	ExecuteReport

		Función que ejecuta la App "Reports" para obtener la información necesaria.
		Si fuese necesario los datos resultantes se pueden procesar.
]]--
local function ExecuteReport (web, request)
	local encodedParams = web.input["post_data"]
	LogInfo("[API.interactions:ExecuteReport] Invoked API Users Detail ........ post_data: '%s'.", encodedParams)
	local data = json.decode(encodedParams)
	if data.queryFields and request.apiDefinition.config.type then
		data.queryFields.Type = request.apiDefinition.config.type:lower()
	end
	local reportId = request.apiDefinition.config.reportId
	-- el tipo de retorno de datos por defecto usuarios es "reportOnlyData"
	local defaultReturnType = "reportOnlyData"
	if not data.returnInfo then
		data.returnInfo = {
			type = defaultReturnType
		}
	else
		-- solo soporta estos 2 tipos de datos a retornar "reportOnlyData" y "reportStructure"
		if data.returnInfo.type ~= "reportOnlyData" or data.returnInfo.type ~= "reportStructure" then
			data.returnInfo.type = defaultReturnType
		end
	end

	local success, content = m_reportAPI.API_Execute(reportId, data)
	if not success and content == "other_node" then
		return api_utils.replyOtherNode(web, "cache not found in this node")
	end
	-- TODO: aca llegan los datos retornados desde la app "reports" rocka "reporter". Si fuese necesario se puede procesar la información:
	--	realizar algo contra ES, REDIS, Cassandra, etc.
	return api_utils.replyOk(web, {content = content})
end

--[[----------------------------------------------------------------------
	Mapeo de funciones que quedan disponibles en módulo para luego ser utilizadas en el mapeo de funciones con las urls (/definition/ARCHIVO.lua función "urlLinkedFunction")
--------------------------------------------------------------------------]]
_M.Get                        = Get
_M.Save                       = Save
_M.Search                     = Search
_M.Activate                   = Activate
_M.Deactivate                 = Deactivate
_M.GetAvailableTelephonyPeers = GetAvailableTelephonyPeers
_M.GetUsersExportInfo         = GetUsersExportInfo
-- Reportes: ejecuta contra la app reports
_M.ExecuteReport              = ExecuteReport
-------------------------------------------------------------------------

function MapUris (m, authFilter, prefix)
	local f = authFilter
	local p = prefix
	local Log = require("apps.agent.requestFilters").LogRequest
	local jsonValidator = utils.JsonValidator
	local vr = utils.JsonValidatorReply
	local r = utils.JsonReply
	-- mapeo de URLs dinamicas (definidas en \api\definition\full.lua)
	local replyPostFn = Log .. f .. vr
	local replyGetFn = Log .. f .. r
	api_utils:MapDynamicURIs(m_apiType, m, _M, replyPostFn, replyGetFn)
	--
	m_htmlRender = htmlRender
end
