local api_utils = require "api.utils"
local oe_utils = require "api.outbound_engine_utils"

module((...), package.seeall)

local m_apiDefaultConfig = api_utils.GetApiURLsDefaultConfig()
local entityType = "outbound_engine_contact"
local baseUri = "/outbound_engine/contacts/"

local AVAILABLE_CHANNELS = {
	CALL = true, 
}

local AVAILABLE_ENGINE_TYPES = {
	PROGRESSIVE = true,
}

local m_apiDefinition = {
	info = {
		name = entityType
	},
	items = {

		--[[
			Get Contact
				url: /outbound_engine/contacts/get/contactid@vcc
				method: GET
		]]--
		{
			entityType = entityType,
			id = "a8a92b10-7da6-4c98-9a7a-1035e5ce854c",
			name = "GetContact",
			description = "Get Contact Information",
			url = baseUri .. "get/(.+)",
			method = "GET",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.read,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "Get"
			}
		},

		--[[
			List Contacts
				url: /outbound_engine/contacts/list/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "5bb9d336-aaed-4592-9ba1-fe807c7497fc",
			name = "ListContacts",
			description = "List Contacts",
			url = baseUri .. "list/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "contactIds", 	required = true,	type = "table", array = true },
				{id = "fullData",  		required = false, 	type = "boolean" },
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "List"
			}
		},

		--[[
			Search Contacts
				url: /outbound_engine/contacts/search/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "a541f4d3-629d-43b8-a2f4-ce47e7d02f7e",
			name = "SearchContacts",
			description = "Search Contacts",
			url = baseUri .. "search/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "processId", 		required = false,	type = "string" },
				{id = "campaignId", 	required = false,	type = "string" },
				{id = "batchId", 		required = false,	type = "string" },
				{id = "skillId", 		required = false,	type = "string" },
				{id = "size",  			required = false, 	type = "number" },
				{id = "pit", 			required = false,	type = "string" },
				{id = "searchAfter", 	required = false,	type = "string" },
				{id = "fullData",  		required = false, 	type = "boolean" },
				{id = "keepOpen",  		required = false, 	type = "boolean" },
				{id = "sortTerms", 		required = false,	type = "table", array = true, children = {
						{id = "name",     		required = true, type = "string" },
						{id = "order",  		required = false, type = "string", domain = { asc = true, desc = true } },
					}
				},
				{id = "fields", 		required = false,	type = "table", array = true },
				oe_utils.GetConditionsDefinition(false),
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "Search"
			}
		},

		--[[
			Get contact by name
				url: /outbound_engine/contacts/get_by_name/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "1cf9193e-58e5-49b7-8e2a-932ac230f6fec",
			name = "GetContactByName",
			description = "Get contact by name",
			url = baseUri .. "get_by_name/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "name", 		required = true,	type = "string" },
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "GetByName"
			}
		},

		--[[
			Get contact by address
				url: /outbound_engine/contacts/get_by_address/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "a27a6b23-7baa-491c-b6fa-e93f08477ddc",
			name = "GetContactByAddress",
			description = "Get contact by address",
			url = baseUri .. "get_by_address/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "addressType", 	required = true,	type = "string" },
				{id = "address", 		required = true,	type = "string" },
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "GetByAddress"
			}
		},

		--[[
			Summary Contacts
				url: /outbound_engine/contacts/summary/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "aa24e6ef-654b-4bab-9395-b394ef44fe0c",
			name = "SummaryContacts",
			description = "Summary Contacts",
			url = baseUri .. "summary/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "processId", 		required = false,	type = "string" },
				{id = "campaignId", 	required = false,	type = "string" },
				{id = "batchId", 		required = false,	type = "string" },
				{id = "skillId", 		required = false,	type = "string" },
				oe_utils.GetConditionsDefinition(false),
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "Summary"
			}
		},

		--[[
			Store Contacts
				url: /outbound_engine/contacts/store/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "0f815584-1a8a-4f63-94a2-6dffd3154485",
			name = "StoreContacts",
			description = "Store Contacts",
			url = baseUri .. "store/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "mix",     		required = false,   type = "boolean" },
				{id = "contacts", 		required = true,	type = "table", array = true, children = {
						{id = "Id",     			required = true, type = "string" },
						{id = "Name",     			required = false, type = "string" },
						{id = "Campaign",     		required = false, type = "string" },
						{id = "Category",     		required = false, type = "string" },
						{id = "AccountOfficer",     required = false, type = "string" },
						{id = "ExternalId",     	required = false, type = "string" },
						{id = "ImportId",     		required = false, type = "string" },
						{id = "AccountGroup",     	required = false, type = "string" },
						{id = "IsClient",     		required = false, type = "boolean" },
						{id = "Addresses",     		required = false, type = "table", array = true,
							children = oe_utils.GetAddressFieldsDefinition() },
						{id = "ManualScheduleDate", required = false, type = "string" },
						{id = "Agent",     			required = false, type = "string" },
						{id = "Skills", 			required = false, type = "table", array = true, children = {
								{id = "SkillId",     	required = true, type = "string" },
								{id = "Mode",  			required = false, type = "string", default = "SORT",
									domain = { SORT = true, REQUIRED = true, PREFERRED = true } },
								{id = "Value",  	required = true, type = "number" },
							}
						},
						{id = "NameValuesSimple", 	required = false, type = "table", array = true, children = {
								{id = "Name",     		required = true, type = "string" },
								{id = "Value",  		required = true, type = "string" },
							}
						},
						{id = "NameValuesSearchDate", 	required = false, type = "table", array = true, children = {
								{id = "Name",     		required = true, type = "string" },
								{id = "Value",  		required = true, type = "string" },
							}
						},
						{id = "NameValuesSearchNumeric", required = false, type = "table", array = true, children = {
								{id = "Name",     		required = true, type = "string" },
								{id = "Value",  		required = true, type = "number" },
							}
						},
						{id = "NameValuesSearchText", 	required = false, type = "table", array = true, children = {
								{id = "Name",     		required = true, type = "string" },
								{id = "Value",  		required = true, type = "string" },
							}
						},
					}
				},
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "medium_light", -- low, medium_light, medium, high
				urlLinkedFunction = "Store"
			}
		},

		--[[
			Set Contact Addresses
				url: /outbound_engine/contacts/set_addresses/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "d7ac62b4-25a8-4273-b990-f4fe68a4b7e5",
			name = "SetContactAddresses",
			description = "Set Contact Addresses",
			url = baseUri .. "set_addresses/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "mix",     		required = false,   type = "boolean" },
				{id = "contactId", 		required = true,   	type = "string" },
				{id = "addresses",     	required = true, 	type = "table", array = true,
					children = oe_utils.GetAddressFieldsDefinition() },
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "medium_light", -- low, medium_light, medium, high
				urlLinkedFunction = "SetAddresses"
			}
		},

		--[[
			Set Contact Values
				url: /outbound_engine/contacts/set_values/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "ca91a4b6-c66e-4a76-889b-629e847609a7",
			name = "SetContactValues",
			description = "Set Contact Values",
			url = baseUri .. "set_values/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "mix",     		required = false,   type = "boolean" },
				{id = "contactId", 		required = true,   	type = "string" },
				{id = "simple", 		required = false, 	type = "table", array = true, children = {
						{id = "Name",     		required = true, type = "string" },
						{id = "Value",  		required = true, type = "string" },
					}
				},
				{id = "date", 			required = false, 	type = "table", array = true, children = {
						{id = "Name",     		required = true, type = "string" },
						{id = "Value",  		required = true, type = "string" },
					}
				},
				{id = "numeric", 		required = false, 	type = "table", array = true, children = {
						{id = "Name",     		required = true, type = "string" },
						{id = "Value",  		required = true, type = "number" },
					}
				},
				{id = "text", 			required = false, 	type = "table", array = true, children = {
						{id = "Name",     		required = true, type = "string" },
						{id = "Value",  		required = true, type = "string" },
					}
				},
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "medium_light", -- low, medium_light, medium, high
				urlLinkedFunction = "SetValues"
			}
		},

		--[[
			Set Contact Skills
				url: /outbound_engine/contacts/set_skills/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "82973a50-7f73-48f3-8dd5-f4d160733148",
			name = "SetContactSkills",
			description = "Set Contact Skills",
			url = baseUri .. "set_skills/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "mix",     		required = false,   type = "boolean" },
				{id = "contactId", 		required = true,   	type = "string" },
				{id = "skills", 		required = true, 	type = "table", array = true, children = {
						{id = "SkillId",     	required = true, type = "string" },
						{id = "Mode",  			required = false, type = "string", default = "SORT",
							domain = { SORT = true, REQUIRED = true, PREFERRED = true } },
						{id = "Value",  	required = true, type = "number" },
					}
				},
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "medium_light", -- low, medium_light, medium, high
				urlLinkedFunction = "SetSkills"
			}
		},

		--[[
			Remove Contact Address
				url: /outbound_engine/contacts/remove_address/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "a69c753d-0f8a-4225-97a1-1e4e606cd319",
			name = "RemoveContactAddress",
			description = "Remove Contact Address",
			url = baseUri .. "remove_address/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "contactId", 		required = true,   	type = "string" },
				{id = "addressType", 	required = true,	type = "string" },
				{id = "address", 		required = true,	type = "string" },
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "medium_light", -- low, medium_light, medium, high
				urlLinkedFunction = "RemoveAddress"
			}
		},

		--[[
			Enable Contact Address
				url: /outbound_engine/contacts/enable_address/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "39c15135-4283-49b9-974f-0957debfbfd4",
			name = "EnableContactAddress",
			description = "Enable Contact Address",
			url = baseUri .. "enable_address/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "contactId", 		required = true,   	type = "string" },
				{id = "addressType", 	required = true,	type = "string" },
				{id = "address", 		required = true,	type = "string" },
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "medium_light", -- low, medium_light, medium, high
				urlLinkedFunction = "EnableAddress"
			}
		},

		--[[
			Disable Contact Address
				url: /outbound_engine/contacts/disable_address/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "0de76309-20f5-4c8d-b941-d3adf168a8db",
			name = "DisableContactAddress",
			description = "Disable Contact Address",
			url = baseUri .. "disable_address/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "contactId", 		required = true,   	type = "string" },
				{id = "addressType", 	required = true,	type = "string" },
				{id = "address", 		required = true,	type = "string" },
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "medium_light", -- low, medium_light, medium, high
				urlLinkedFunction = "DisableAddress"
			}
		},

		--[[
			Reschedule Contact
				url: /outbound_engine/contacts/reschedule/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "a82540a5-8ccc-46c4-a9e1-7456f1623829",
			name = "RescheduleContact",
			description = "Reschedule Contact",
			url = baseUri .. "reschedule/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "processId", 		required = false,	type = "string" },
				{id = "campaignId", 	required = false,	type = "string" },
				{id = "batchId", 		required = true,	type = "string" },
				{id = "contactId", 		required = true,	type = "string" },
				{id = "configuration", 	required = true,	type = "table", children = {
						{id = "restartRetries",  	required = false, type = "boolean" },
						{id = "scheduleDate",  	required = false, type = "string" },
						{id = "agent",  		required = false, type = "string" },
						{id = "channel",  		required = false, type = "string" },
						{id = "kind",  			required = false, type = "string" },
						{id = "address",  		required = false, type = "table", children = oe_utils.GetAddressFieldsDefinition() },
					}
				},
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "Reschedule"
			}
		},

		--[[
			Validate Address
				url: /outbound_engine/contacts/validate_address/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "131acaa6-4fe4-4030-8e3a-53e7c2172d27",
			name = "ValidateAddress",
			description = "Validate Address",
			url = baseUri .. "validate_address/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{ id = "type",     		required = true, type = "string" },
				{ id = "number",     	required = true, type = "string" },
				{ id = "country",     	required = false, type = "string" },
				{ id = "area",     		required = false, type = "string" },
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "ValidateAddress"
			}
		},

		--[[
			Get Contact Results
				url: /outbound_engine/contacts/get_results/
				method: POST
		]]--
		{
			entityType = entityType,
			id = "a82540a5-8ccc-46c4-a9e1-7456f16238291",
			name = "GetContactResults",
			description = "Get Contact Results",
			url = baseUri .. "get_results/",
			method = "POST",
			header = {
				Authorization = "....",
				["Content-Type"] = "application/json"
			},
			body = {
				{id = "processId", 		required = false,	type = "string" },
				{id = "campaignId", 	required = false,	type = "string" },
				{id = "batchId", 		required = false,	type = "string" },
				{id = "contactId", 		required = false,	type = "string" },
				{id = "size",  			required = false, 	type = "number" },
				{id = "offset", 		required = false,	type = "number" },
				{id = "sortTerms", 		required = false,	type = "table", array = true, children = {
						{id = "name",     		required = true, type = "string" },
						{id = "order",  		required = false, type = "string", domain = { asc = true, desc = true } },
					}
				},
				oe_utils.GetConditionsDefinition(false),
			},
			response = {},
			responseErrors = {},
			config = {
				permission = m_apiDefaultConfig.permissions.write,
				rateLimits = api_utils.GetApiURLRateLimit,
				resourceType = "low", -- low, medium_light, medium, high
				urlLinkedFunction = "GetResults"
			}
		},
	}
}

-- ---------------------------------------------

function Get ()
	local m_apiDefinition = m_apiDefinition
	return m_apiDefinition
end