local Synchronize = require "Synchronize"
local https = Synchronize("luasec_mutex", function()
	return require ("ssl.https")
end)
local ltn12 = require "ltn12"
local m_models = require("apps.administration.model"):GetModels()

module(..., package.seeall, inconcert.orbit.application.Decorate)
local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local m_validations = require("apps.administration.controllers.validations.app_notifications").GetDefinitions()

m_htmlRender = nil
m_appLocation = nil

function MapUris(module, htmlRender, prefix, appLocation)

	local LogRequest = require("apps.agent.requestFilters").LogRequest

	module:dispatch_post(utils.JsonValidatorReply..LogRequest..Search, prefix.."/search/")
	module:dispatch_get(utils.JsonReply..LogRequest..List, prefix.."/list/")
	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("administration", "view_apps_notifications", Get), prefix.."/get/(.+)")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("administration", "save_apps_notifications",
		utils.WebJsonValidatorData(m_validations.basic, SaveAppsNotifications)),
		prefix.."/save/"
	)
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("administration", "delete_apps_notifications", RemoveAppsNotifications), prefix.."/delete/")

	module:dispatch_get(Render, prefix .. "/frm/render/")

	module:dispatch_post(utils.JsonValidatorReply..LogRequest..GetHelpDeskApiInfo, prefix.."/helpdesk_api_info/")
	
	m_htmlRender = htmlRender
	m_appLocation = appLocation
end

function Render(web)
	_M.response.DontCache(web)
	local env = { _T = m_htmlRender.extension.makeTextTranslator() }
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("Render form apps_notifications for user %s, group %s", userId, inconcert.authUser.userGroup)

	local html_file = m_appLocation .. "/pages/frm_apps_notifications_content.html"
	local content = m_htmlRender:RenderPageLP(html_file, web, env)
	if not content then
		LogError("HTML apps_notifications does not exist: %s", html_file)
	end
	return content
end

--Returns all apps_notifications in the VCC
function Search(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local fullUserId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local startTime = os.time()

	LogInfo("[ADMINISTRATION.APPS_NOTIFICATIONS:SEARCH] 1 - (START) Search rows for user '%s', group '%s'. Posted data: '%s'", fullUserId, inconcert.authUser.userGroup, web.input["post_data"])

	local postData = web.request
	--Get apps integrations
	local integrations = m_models.VCCs.GetAppsIntegrations(vcc)
	--Get all apps_notifications from model
	local rows = m_models.AppsNotifications.GetAppsNotifications(vcc, nil, nil, integrations)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[ADMINISTRATION.APPS_NOTIFICATIONS:SEARCH] 2 - Filtering rows")

	local _T = m_htmlRender.extension.makeTextTranslator()
	for appId, app in pairs(rows) do
		app.Id = appId
		if utils.MatchFilter(app, postData) then
			totalAdded = totalAdded + 1
			if app.System and app.Description then
				app.Description = _T(app.Description)
			end
			table.insert(filteredRows, app)
		end
	end

	LogInfo("[ADMINISTRATION.APPS_NOTIFICATIONS:SEARCH] 3 - Sorting rows")

	--include only rows that belong to logged user vcc
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = 0
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[ADMINISTRATION.APPS_NOTIFICATIONS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--Returns all apps_notifications in the VCC
function List(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local fullUserId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local startTime = os.time()

	--Get apps integrations
	local integrations = m_models.VCCs.GetAppsIntegrations(vcc)
	--Get all apps_notifications from model
	local rows = m_models.AppsNotifications.GetAppsNotifications(vcc, nil, nil, integrations)

	local _T = m_htmlRender.extension.makeTextTranslator()
	for appId, app in pairs(rows) do
		app.Id = appId
		if app.System and app.Description then
			app.Description = _T(app.Description)
		end
	end
	
	return { apps = rows }
end

function Get(web, appId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc
	LogInfo("Retrieving app notifications appId %s, user %s, vcc %s", appId, userId, vcc)
	
	local integrations = m_models.VCCs.GetAppsIntegrations(vcc)
	local appNotifications = m_models.AppsNotifications.GetAppsNotifications(vcc, appId, nil, integrations)
	LogInfo("App notifications %s: %s", appId, json.encode(appNotifications))

	if appNotifications then
		appNotifications.Id = appId
		if appNotifications.System and appNotifications.Description then
			local _T = m_htmlRender.extension.makeTextTranslator()
			appNotifications.Description = _T(appNotifications.Description)
		end
		return appNotifications
	else
		web.status = 404
		return {result = "NOT_FOUND"}
	end
end

--[[
	Add/Modifies appNotifications in a vcc
]]
function SaveAppsNotifications (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("[ADMINISTRATION.SaveAppsNotifications], user %s, vcc %s", userId, vcc)
	--LogDebug("[ADMINISTRATION.SaveAppsNotifications] --> posted Notifications Apps: %s", web.input["post_data"])

	local postData = web.request
	--Get all apps from model
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	local apps = {}

	for appId, app in pairs(postData) do
		local existingApp = m_models.AppsNotifications.GetAppsNotifications(vcc, appId, keyspace)

		if type(app.Url) ~= "string" or app.Url == "" then
			keyspace:Close()
			web.status = 400
			LogError("App %s doesnt have an Url", appId)
			return {result = "MUST_HAVE_A_URL", reason = "App must have an url"}
		end

		if type(app.Description) ~= "string" or app.Description == "" then
			keyspace:Close()
			web.status = 400
			LogError("App %s doesnt have a Description", appId)
			return {result = "MUST_HAVE_A_DESCRIPTION", reason = "App must have a description"}
		end

		if type(app.Events) ~= "table" then
			keyspace:Close()
			web.status = 400
			LogError("App %s doesnt have an events list", appId)
			return {result = "MUST_HAVE_A_EVENTS_LIST", reason = "App must have an events list"}
		end

		--If this is a new entity, try to find an unused Id for it
		if app.isNew then
			if existingApp then
				keyspace:Close()
				web.status = 400
				LogError("App %s already exists", appId)
				return {result = "ID_ALREADY_EXISTS", reason = "Already exists another app with same id"}
			end
			app.isNew = nil
		else
			if not existingApp then
				keyspace:Close()
				web.status = 400
				LogError("App %s does not exist", appId)
				return {result = "ID_DOES_NOT_EXIST", reason = "Id does not exist"}
			elseif existingApp.System then
				keyspace:Close()
				web.status = 400
				LogError("App %s already exists as a system app", appId)
				return {result = "ID_ALREADY_EXISTS", reason = "Already exists a system app with same id"}
			end
		end

		apps[appId] = app
	end

	local result
	local success, err = m_models.AppsNotifications.SaveAppsNotifications(vcc, apps, keyspace)
	if success then
		result = {result = "OK", idMapping = idMapping}
	else
		web.status = 400
		LogError("[ADMINISTRATION.SaveAppsNotifications] Saving apps notifications error: %s.", json.encode(err))
		result = err
	end

	keyspace:Close()
	return result
end


--[[
Remove apps from a vcc
]]
function RemoveAppsNotifications(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("[ADMINISTRATION.RemoveAppsNotifications], user %s, vcc %s, apps: '%s'", userId, vcc, web.input["post_data"])

	local apps = web.request
	local keyspace = utils.GetDatasource():CassandraConnection("Agent")

	local success, err = m_models.AppsNotifications.RemoveAppsNotifications( vcc, apps, keyspace )
	if not success then
		keyspace:Close()
		web.status = 400
		LogError("[ADMINISTRATION.RemoveAppsNotifications] Removing apps error: %s.", json.encode(err))
		return err
	end

	keyspace:Close()
	return {result = "OK"}
end

--Returns all HelpDesk API info in the VCC
function GetHelpDeskApiInfo(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local enabled, vccHelpDeskConfig = utils.HasHelpDeskIntegration(vcc)

	local response = {}
	if enabled and vccHelpDeskConfig and vccHelpDeskConfig.Server and vccHelpDeskConfig.Server.HOST then
		local helpDeskApiJsonDocUrl = vccHelpDeskConfig.Server.HOST .. "/apidoc/openapi.json"
		local responseBody = {}

		local result, code, headers, statusline = https.request{
			url = helpDeskApiJsonDocUrl,
			method = "GET",
			sink = ltn12.sink.table(responseBody)
		}

		if code == 200 then
			response.success = true
			response.data = table.concat(responseBody, "")
		else
			response.success = false
			response.data = statusline
			LogError("[apps_notifications:GetHelpDeskApiInfo] Error. RESULT: '%s', CODE: '%s', STATUS: '%s'.", result, code, statusline)
		end
	else			   
		response.success = false
		response.data = "HelpDesk not enabled or misconfigured"				   
		LogDebug("[apps_notifications:GetHelpDeskApiInfo] response '%s'", json.encode(response))
	end

	return response
end