module(..., package.seeall, inconcert.orbit.application.Decorate)
local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local PM = require("provisioningmanager")
local m_provisioningVccsModel = require "apps.provisioning.models.vccs"
local jwt = require "shared.jwt"
local audits = require "audits"
m_htmlRender = nil
m_appLocation = nil

local function getSecurity(vcc)
	local success, vccInfo = m_provisioningVccsModel.Get(vcc)

	if not success then
		return false, {result = vccInfo}
	end

	return true, vccInfo.Security
end

local function getSecurityAuditData()
	--Campos auditables de Seguridad
	local auditFields = {
		captcha = true,
		mfa = true,
		ldap = true,
		authentication = true
	}
		
	local vcc = inconcert.authUser.virtualContactCenter
	local success, vccInfo = m_provisioningVccsModel.Get(vcc)
	local securityData = success and m_provisioningVccsModel.LoadVCCSecurity(vccInfo) or {}
	local result = {}
	for key, value in pairs(securityData) do
		local auditField = auditFields[key]
		if auditField and type(auditField) == "table" then
			result[auditField.alias or key] = value
		elseif auditField then
			result[key] = value
		end
	end
	
	if result and result.captcha then
		result.captcha["secretKey"] = nil
		if result.captcha.type or result.captcha.enabled == "0" then
			result.captcha["type"] = nil
			result.captcha["publicKey"] = nil
		end
	end
	return result
end

function GetSecurity(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc
	LogInfo("[ADMINISTRATION.SECURITY:Get] Retrieving Security configuration for user %s, vcc %s", userId, vcc)

	local success, vccSecurity = getSecurity(vcc)
	--LogDebug("[ADMINISTRATION.SECURITY:Get] success: '%s' vccInfo: '%s'.", success, vccInfo)
	if not success then
		web.status = 400
		LogError("[ADMINISTRATION.SECURITY:Get] Error Retrieving Security configuration for user %s, vcc %s: %s", userId, vcc, vccInfo)
	end
	return vccSecurity
end

function UpdateSecurity (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("[ADMINISTRATION.SECURITY:UpdateSecurity] Updating security - user %s, vcc %s", userId, vcc )
	local ok, security = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[ADMINISTRATION.SECURITY:UpdateSecurity] Error Updating security for vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end
	
	--#11602 - LicenseSystem - PM.UpdateVCC and PM.GetVCC now require a previous call to InitHelpersAgentServiceOptions
	local helpers_agentservice_options = utils.Get_m_helpers_agentservice_options()
	local success_agentservice_options, err_agentservice_options = pcall(PM.InitHelpersAgentServiceOptions, helpers_agentservice_options)

	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local success, vccInfo = pcall(PM.GetVCC, ksAgent, vcc, false)	
	if not success then
		ksAgent:Close()
		LogError("[ADMINISTRATION.SECURITY:UpdateSecurity] Error retrieving vcc info for vcc: '%s':'%s", vcc, vccInfo)
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	local authToken = web.vars.auth.authToken

	vccInfo.Security = security
	--load captcha config
	vccInfo.Security = m_provisioningVccsModel.LoadVCCSecurity(vccInfo)

	--El password para conectar con el server externo de autenticacion viene con jwtoken
	local authType = vccInfo.Security.authentication.type
	if authType ~= "BASIC" and authType ~= "MFA" then
		authType = authType:lower()
		if vccInfo.Security[authType].adminPassword then
			local jwtData, jwtErr = jwt.decode(vccInfo.Security[authType].adminPassword, md5.sumhexa(authToken), "HS256")
			if jwtErr then
				LogError("[ADMINISTRATION.SECURITY:UpdateSecurity] Error decoding jwtToken: '%s'", jwtErr)
				web.status = 400
				return {result = "INVALID_REQUEST", code = 400}
			end
			local fakePassword = m_provisioningVccsModel.GetExternalAuthFakePassword() 
			if jwtData.password == fakePassword then 
				--El password no ha cambiado, recuperamos el que tenemos en BD
				vccInfo.Security[authType].adminPassword = m_provisioningVccsModel.GetCurrentExternalAuthPassword(vcc)
			else
				--El password ha cambiado
				vccInfo.Security[authType].adminPassword = jwtData.password
			end
		else
			ksAgent:Close()
			LogError("[ADMINISTRATION.SECURITY:UpdateSecurity] adminPassword for external authentication '%s' is required: '%s'", vccInfo.Security.authentication.type)
			return false, "EXTERNAL AUTHENTICATION PASSWORD INVALID"
		end
	end	

	if authType:lower() == "openid" then
		local openid_configuration_url = vccInfo.Security.openid.openid_configuration_url
		local openIdConfigSuccess, openIdConfig = pcall(PM.GetOpenIdConfiguration, openid_configuration_url)
		if not openIdConfigSuccess or not openIdConfig.success then
			LogError("[ADMINISTRATION.SECURITY:UpdateSecurity] Get openid_configuration_url error: '%s'", openIdConfig.error)
			return false, "OPENID CONFIGURATION URL ERROR"
		end
		vccInfo.Security.openid.openid_configuration_values = openIdConfig.result
	end		

	local old_sec_audit_data = getSecurityAuditData()
	local ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, errMsg = pcall(PM.UpdateVCC, ksAgent, vcc, vccInfo, ksProv)

	ksAgent:Close()
	ksProv:Close()

	if not success then
		LogWarning("[ADMINISTRATION.SECURITY:UpdateSecurity] Error saving vccs '%s' -> '%s'", vcc, errMsg)
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end	

	local new_sec_audit_data = getSecurityAuditData()
	local ksHistorical = utils.GetDatasource():CassandraConnection("Historical")
	audits.Create({
		vcc = vcc,
		entity = "Security",
		user = inconcert.authUser.userName,
		fieldKey = authType:lower(),
		oldFields = old_sec_audit_data,
		newFields = new_sec_audit_data,
		connection = ksHistorical
	})
	ksHistorical:Close()
	--ok
	return { result = "OK" }
end

local function getExternalEntities(data)
	assert(data and type(data) == "table")
	assert(data.source)
	if data.source == "ldap" then
		assert(data.server)
		assert(data.baseDN)
		assert(data.adminUserDN)
		assert(data.password)
		data.useTLS = data.useTLS or false

		return pcall(PM.GetLdapGroups, data.server, data.baseDN, data.adminUserDN, data.password, data.useTLS)
	elseif data.source == "azuread" then
		assert(data.openid_configuration_url)
		assert(data.client_id)
		assert(data.adminPassword or data.password)

		return pcall(PM.GetAzureADGroups, data.openid_configuration_url, data.client_id, data.adminPassword or data.password)
	end
end

function LoadExternalEntities(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local data = json.decode(web.input["post_data"])

	local authToken = web.vars.auth.authToken
	local jwtData, jwtErr = jwt.decode(data.password, md5.sumhexa(authToken), "HS256")
	if jwtErr then
		LogError("[ADMINISTRATION.SECURITY:LoadExternalEntities] Error decoding jwtToken: '%s'", jwtErr)
		web.status = 400
		return { result = "FAIL"}
	end
	data.password = jwtData.password

	local externalAuthFakePassword = m_provisioningVccsModel.GetExternalAuthFakePassword()
	if data.password == externalAuthFakePassword then
		data.password = m_provisioningVccsModel.GetCurrentExternalAuthPassword(vcc)
	end

	LogInfo("[ADMINISTRATION.SECURITY:LoadExternalEntities] Loading external entities for User %s and VCC %s.", vcc, userId)

	local success, entities = getExternalEntities(data)
	if not success then
		LogError("[ADMINISTRATION.SECURITY:LoadExternalEntities] Error Retrieving ")
		web.status = 400
		return { result = "FAIL"}
	end

	if type(entities) == "table" and entities.success == false then
		local _T = m_htmlRender.extension.makeTextTranslator()
		entities.error = entities.error and _T(entities.error) or ""
	end

	return entities
end

function GetExternalEntities(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[ADMINISTRATION.SECURITY:GetExternalEntities] Loading external entities for VCC %s. User %s.", vcc, userId)
	--LogDebug("[ADMINISTRATION.GROUPS:LoadExternalAuthoritationEntities] --> Posted data '%s'.", postedData)

	local success, vccSecurity = getSecurity(vcc)
	if not success then
		web.status = 404
		LogError("[ADMINISTRATION.SECURITY:GetExternalEntities] Error Retrieving Security configuration for user %s, vcc %s: %s", userId, vcc, vccSecurity)
		return {result = "NOT_FOUND"}
	end
	local authType = vccSecurity.authentication.type
	if authType == "BASIC" or authType == "MFA" then
		web.status = 200
		LogInfo("[ADMINISTRATION.SECURITY:GetExternalEntities] vcc %s has not external entities", vcc)
		return {result = ""}
	end

	authType = authType:lower()
	local data = {}
	if authType == "ldap" then
		data.source = "ldap"
		data.server = vccSecurity[authType].server
		data.baseDN = vccSecurity[authType].baseDN
		data.adminUserDN = vccSecurity[authType].adminUserDN
		data.password = m_provisioningVccsModel.GetCurrentExternalAuthPassword(vcc)
		data.useTLS = vccSecurity[authType].useTLS
	elseif authType == "openid" and vccSecurity[authType].sync_method == "AZUREAD" then
		data.source = "azuread"
		data.openid_configuration_url = vccSecurity[authType].openid_configuration_url
		data.client_id = vccSecurity[authType].client_id
		data.adminPassword = m_provisioningVccsModel.GetCurrentExternalAuthPassword(vcc)
	end

	local success, result = getExternalEntities(data)	
	if not success then
		local err = result
		local stringErr = type(err) == "table" and json.encode(err) or err
		LogError("[ADMINISTRATION.SECURITY:GetExternalEntities] Unable to Load external entities. Err: %s", stringErr)
		web.status = 500
		return result
	end

	if type(result) == "table" and result.success == false then
		local _T = m_htmlRender.extension.makeTextTranslator()
		result.error = result.error and _T(result.error) or ""
	end

	web.status = 200
	return result
end

function Render (web)
	_M.response.DontCache(web)
	local env = { _T = m_htmlRender.extension.makeTextTranslator() }
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("[ADMINISTRATION.SECURITY:Render] Render form Security for user %s, group %s", userId, inconcert.authUser.userGroup)

	local html_file = m_appLocation .. "/pages/frm_security_content.html"
	local content = m_htmlRender:RenderPageLP(html_file, web, env)
	if not content then
		LogError("[ADMINISTRATION.SECURITY:Render] HTML Security does not exist: %s", html_file)
	end
	return content
end

--[[
****************************************************************
	MAPEO DE URLS
****************************************************************
]]--

function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest

	local permissionView  = "view_security"
	local permissionSave  = "save_security"

	local replyFn = utils.JsonReply
	local replyValFn = utils.JsonValidatorReply

	local function HasPermission (fn, actionId, responseFn)
		actionId = actionId or permissionSave
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. permissions:CheckPermission("administration", actionId, fn)
	end

	local p = prefix
	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- SECURITY
	module:dispatch_get   	(HasPermission(GetSecurity, permissionView)                                               , p .. "/get/(.*)"                             )
	module:dispatch_put   	(HasPermission(UpdateSecurity, permissionSave)                                            , p .. "/update/(.*)"                          )
	module:dispatch_get   	(HasPermission(GetExternalEntities, permissionView)                                       , p .. "/get_vcc_externalentities/(.*)"        )
	module:dispatch_post	(utils.JsonValidatorReply..LogRequest..HasPermission(LoadExternalEntities, permissionSave), p .."/loadexternalentities/"                 )

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end