local m_models = require("apps.administration.model"):GetModels()
module(..., package.seeall, inconcert.orbit.application.Decorate)
local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local agentservice = require "apps.agent.agentservice"
local agentNotifications = require("apps.agent.agentNotifications")
local m_validations = require("apps.administration.controllers.validations.signatures").GetDefinitions()

local m_moduleLocation = select(2, ...)
m_htmlRender = nil
m_appLocation = nil

function MapUris(module, htmlRender, prefix, appLocation)

	local LogRequest = require("apps.agent.requestFilters").LogRequest

	module:dispatch_post(utils.JsonValidatorReply..LogRequest..Search, prefix.."/search/")
	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("administration", "view_signatures", Get), prefix.."/get/(.+)")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("administration", "save_signatures",
			utils.WebJsonValidatorData(
				m_validations.basic,
				SaveSignatures
			)
		),
		prefix.."/save/"
	)
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("administration", "delete_signatures", RemoveSignatures), prefix.."/delete/")
	module:dispatch_get(utils.JsonReply..LogRequest.. RemoveValidation, prefix.."/delete_validation/(.+)")

	module:dispatch_get(Render, prefix .. "/frm/render/")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end

function Render(web)
	_M.response.DontCache(web)

	local env = { _T = m_htmlRender.extension.makeTextTranslator() }
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("Render form Signatures for user %s, group %s", userId, inconcert.authUser.userGroup)

	local html_file = m_appLocation .. "/pages/frm_signature_content.html"
	local content = m_htmlRender:RenderPageLP(html_file, web, env)
	if not content then
		LogError("HTML Signatures does not exist: %s", html_file)
	end
	return content
end

--Returns all Signatures in the VCC
function Search(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local fullUserId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local startTime = os.time()

	LogInfo("[ADMINISTRATION.SIGNATURES:SEARCH] 1 - (START) Search rows for user '%s', group '%s'. Posted data: '%s'", fullUserId, inconcert.authUser.userGroup, web.input["post_data"])

	local postData = web.request
	--Get all signature from model
	local rows = m_models.Signatures.GetSignatures(vcc)
	local totalUnfilteredRows = utils.GetRowCount(rows)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[ADMINISTRATION.SIGNATURES:SEARCH] 2 - Filtering rows")

	for signatureId, signature in pairs(rows) do
		signatureId = utils.GetId(signatureId)
		signature.Id = signatureId
		signature.Name = signature.Name or signatureId
		if utils.MatchFilter(signature, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, signature)
		end
	end

	LogInfo("[ADMINISTRATION.SIGNATURES:SEARCH] 3 - Sorting rows")

	--include only rows that belong to logged user vcc
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[ADMINISTRATION.SIGNATURES:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

function Get(web, signatureId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc
	LogInfo("Retrieving signatureId %s, user %s, vcc %s", signatureId, userId, vcc)
	local signatures = m_models.Signatures.GetSignatures(vcc, signatureId)
	LogInfo("Signature %s: %s", signatureId, json.encode(signatures))

	if next(signatures) then
		signatures.Name = signatures.Name or signatureId
		return signatures
	else
		web.status = 404
		return {result = "NOT_FOUND"}
	end
end

--[[
	Add/Modifies signatures in a vcc
	posted data must have the form:
{
	signatureId1 = {
		Description = "signature description 1",
	},
	signatureId2 = {
		Description = "signature description 2",
	},
	...
}

]]
function SaveSignatures(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("Save Signatures, userId %s, post data:  %s", userId, web.input["post_data"])


	LogInfo("[SaveSignatures] loading existing signatures")

	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	local existingSignatures = m_models.Signatures.GetSignatures(vcc, nil, keyspace)
	local postData = web.request
	local idMapping = {}

	LogInfo("[SaveSignatures] validating posted signatures")

	local signatures = {}
	for signatureId, signature in pairs(postData) do
		if type(signature.Description) ~= "string" then
			keyspace:Close()
			web.status = 400
			LogError("Signature %s doesnt have a description", signatureId)
			return {result = "MUST_HAVE_A_DESCRIPTION", reason = "Signature must have a description"}
		end
		if type(signature.Name) ~= "string" then
			keyspace:Close()
			web.status = 400
			LogError("Signature %s doesnt have a Name", signatureId)
			return {result = "MUST_HAVE_A_NAME", reason = "Signature must have a name"}
		end

		--If this is a new entity, try to find an unused Id for it
		if signature.isNew then
			local n = 0
			repeat
				n = n + 1
				local repeated = utils.FilterEntities(existingSignatures, function(id) return id == utils.Id_n(signatureId, n) end)
			until not next(repeated)
			idMapping[signatureId] = utils.Id_n(signatureId, n)
			signatureId = idMapping[signatureId]
			signature.Id = signatureId
			signature.isNew = nil
		else
			--modification of an existing entity, must ensure that the entity already exists
			local exists = false
			utils.FilterEntities(existingSignatures, function(id) exists = exists or id == signatureId end)
			if not exists then
				keyspace:Close()
				web.status = 400
				LogError("Signature %s does not exist", signatureId)
				return {result = "ID_DOES_NOT_EXIST", reason = "Id does not exist"}
			end
		end

		--Check repeated signatures (those with same Id or Name)
		local repeated = utils.FilterEntities(existingSignatures, {Name = signature.Name})
		for id in pairs(repeated) do
			if id ~= signatureId then
				keyspace:Close()
				web.status = 400
				LogError("Signature %s has repeated Name or Id with %s", signatureId, id)
				return {result = "MUST_HAVE_UNIQUE_NAME", reason = "Signature must have an unique name"}
			end
		end

		existingSignatures[signatureId] = signature
		signatures[signatureId] = signature
	end

	LogInfo("[SaveSignatures] saving signatures")
	m_models.Signatures.SaveSignatures(vcc, signatures, keyspace)
	
	agentNotifications.NotifyAgents({
		eventType = "administrator_signatures_changed",
		agentsLevel = "VCC_USERS",
		pollerMessage = {
			action = "NEW_UPDATE",
			signatures = signatures
		}
	}, keyspace)

	keyspace:Close()
	return {result = "OK", idMapping = idMapping}
end

--[[
Remove signatures from a vcc
postedData must be of the form: ["signatureId1", "signatureId2", ...]
]]
function RemoveSignatures(web)
	local vcc = inconcert.authUser.virtualContactCenter
	LogInfo("RemoveSignatures -> user '%s', vcc '%s', signatures: '%s'.", inconcert.authUser.userName, vcc, web.input["post_data"])

	local signatures_to_remove = web.request
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	m_models.Signatures.RemoveSignatures( vcc, signatures_to_remove, keyspace )
	
	agentNotifications.NotifyAgents({
		eventType = "administrator_signatures_changed",
		agentsLevel = "VCC_USERS",
		pollerMessage = {
			action = "DELETE",
			signatures = utils.ArrayToHash(signatures_to_remove, "Id")
		}
	}, keyspace)

	keyspace:Close()
	return {result = "OK"}
end

function RemoveValidation(web, signature_id)

	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("Validate removing signature '%s'. User '%s'.", signature_id, userId)

	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	local signatures = m_models.Signatures.GetSignatures(vcc, signature_id, keyspace)

	if next(signatures) then
		local result = {result = "OK"}
		-- 1 - valido si se encuentra en uso en alguna account de alguna campana
		local accounts_signature = m_models.Signatures.GetAccountsBySignature( vcc, signature_id, keyspace )

		local function count(list)
			local count = 0
			for _ in pairs(list) do count = count + 1 end
			return count
		end

		if accounts_signature then
			local count_accounts = count(accounts_signature)
			local msg = LogWarning("Signature '%s'. Is being used for '%d' Accounts: '%s'.", signature_id, count_accounts, json.encode(accounts_signature) )
			result.warning = {
				info = "SIGNATURE_BEING_USED_ON_ACCOUNTS",
				count_accounts = count_accounts,
				reason = msg
			}
		end
		keyspace:Close()
		return result
	else
		keyspace:Close()
		LogError("Validation Error: '%s' signature not found.", signature_id)
		web.status = 404
		return {result = "NOT_FOUND"}
	end
end
