module(..., package.seeall)

local utils = require "shared.apps.utils"
local m_datasource
local allegroGroups = require "apps.allegro.controllers.groups"
local helpDeskGroups = require "apps.helpdesk.controllers.groups"
local cjson = require "cjson"
local PM = require("provisioningmanager")

function GetName()
	return "Groups"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

--[[
	Función que filtra los roles a retornar.
		INCLUDE_ALL           = retorna todos los roles.
		NO_INCLUDE_DEPLOYER   = retorna todos los roles menos el rol deployer.
		ONLY_INCLUDE_DEPLOYER = retorna unicamente el rol deployer.
]]--
local function FilterRoles (roles, includeType)
	local filteredRoles = {}
	includeType = includeType or "NO_INCLUDE_DEPLOYER"
	for name, role in pairs(roles or {}) do
		if includeType == "INCLUDE_ALL" then
			filteredRoles[name] = role

		elseif includeType == "ONLY_INCLUDE_DEPLOYER"
			and tostring(role.IsDeployer) == "1"
		then
			filteredRoles[name] = role

		elseif includeType == "NO_INCLUDE_DEPLOYER"
			and (
				tostring(role.IsDeployer) == "0"
				--and tostring(role.IsFullSystemAccess or "0") == "0"
			)
		then
			filteredRoles[name] = role

		end
	end
	return filteredRoles
end

function GetGroups (vcc, groupId, keyspace, includeType)
	local groups = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "Group",
		superColumns = true,
		key = vcc,
		columns = groupId and {groupId} or {}
	}

	local filteredRoles = FilterRoles(groups, includeType)
	--LogDebug("[ADMINISTRATION.GROUPS:GetGroups] filteredRoles: %s,", json.encode(filteredRoles or {}))

	local roles
	if groupId then
		roles = filteredRoles and filteredRoles[groupId] or {}
	else
		roles = filteredRoles or {}
	end
	--LogDebug("[ADMINISTRATION.GROUPS:GetGroups] roles: %s,", json.encode(roles or {}))

	return roles
end

--[[
	-- Guarda en HelpDesk el Groups
]]--
function UpdateHelpDeskGroups (groups, vcc, keyspace)
	if not utils.HasHelpDeskIntegration(vcc, keyspace) then
		LogInfo("[ADMINISTRATION.GROUPS:UpdateHelpDeskGroups] HELPDESK integration. Updating groups. Vcc '%s' has not configured integration with HelpDesk.", vcc)
		return false, {type = "WITHOUT_HELPDESK_INTEGRATION"}
	end

	local success, err = helpDeskGroups.Save(groups, vcc)
	if not success then
		LogError("[ADMINISTRATION.GROUPS:UpdateHelpDeskGroups] HELPDESK integration error. Updating groups '%s'. Error: '%s'.", json.encode(groups), cjson.encode(err))
		return false, {type = "HELPDESK_INTEGRATION_ERROR", sub_type = err}
	end

	return true
end

--[[
	-- Guarda en Allegro los Groups
]]--
local function UpdateAllegroGroupsIn55 (groups, vcc, keyspace)
	if not utils.HasAllegroIntegration(vcc, keyspace) then
		LogInfo("[ADMINISTRATION.GROUPS:UpdateAllegroGroupsIn55] - Allegro integration Updating groups. Vcc '%s' has not configured integration with allegro.", vcc)
		return false, { type = "WITHOUT_ALLEGRO_INTEGRATION" }
	end

	local success, err = allegroGroups.Save(groups, vcc)
	if not success then
		LogError("[ADMINISTRATION.GROUPS:UpdateAllegroGroupsIn55] - Allegro integration error. Updating groups '%s'. Error: '%s'.", json.encode(groups), json.encode(err))
		return false, { type = "ALLEGRO_INTEGRATION_ERROR", err = err }
	end

	return true
end

function SaveGroups (vcc, groups, onlyUpdateIntegration, keyspace)
	LogInfo("[ADMINISRATION.GROUPS:SaveGroups] Saving groups in VCC '%s' onlyUpdateIntegration: '%s'. Groups: '%s'.", vcc, onlyUpdateIntegration, groups)
	-- Guarda en HelpDesk los roles
	-- Debo mandar a actualizar aca, en este lugar, antes de guardar algo en cassandra,
	--   asi me evito el Rollback y toda esas cosas!!!
	--
	local success, helpDeskInfo = UpdateHelpDeskGroups(groups, vcc, keyspace)
	if not success then
		-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "HELPDESK_INTEGRATION_ERROR",
		--  ya que otro tipo de error, puede ser que no tenga integracion con HelpDesk y por lo tanto en si, no es un error y debe de continuar
		--  la actualizacion en i6.
		if helpDeskInfo and helpDeskInfo.type == "HELPDESK_INTEGRATION_ERROR" then
			if close_ks then keyspace:Close() end
			return false, {result = helpDeskInfo.type, message_error = helpDeskInfo.sub_type}
		end
	end

	-- Guarda en Allegro los groups
	-- Debo mandar a actualizar aca, en este lugar, antes de guardar algo en cassandra, asi me evito el Rollback y toda esas cosas!!!
	--
	local success, err = UpdateAllegroGroupsIn55( groups, vcc, keyspace )
	if not success then
		-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
		--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar
		--  la actualizacion en i6.
		if err and err.type == "ALLEGRO_INTEGRATION_ERROR" then
			if close_ks then keyspace:Close() end
			return false, { result = err.type }
		end
	end

	--#11602 - LicenseSystem - PM.UpdateVCC and PM.GetVCC now require a previous call to InitHelpersAgentServiceOptions
	local helpers_agentservice_options = utils.Get_m_helpers_agentservice_options()
	local success_agentservice_options, err_agentservice_options = pcall(PM.InitHelpersAgentServiceOptions, helpers_agentservice_options)

	local success, vccInfo = pcall(PM.GetVCC, keyspace, vcc, false)
	if not success then
		local stringErr = type(vccInfo) == "table" and json.encode(vccInfo) or vccInfo
		return false, {result = stringErr}
	end

	local vccInfoSecurity = json.decode(vccInfo.Security)
	local vccAuthType = vccInfoSecurity.authentication.type
	local ksProv

	for groupId, group in pairs(groups) do
		if vccAuthType ~= "BASIC" and vccAuthType ~= "MFA" then
			vccAuthType = vccAuthType:lower()
			local externalAuthConfig = vccInfoSecurity[vccAuthType]
			local groupFilters = externalAuthConfig.groupFilters
			for groupName, groupFilter in pairs(groupFilters) do
				if groupName ~= groupId and group.ExternalAuthFilter == groupFilter then
					return false, { result = "DUPLICATED_EXTERNAL_FILTER"}
				end
			end

			--Guardamos el filtro del grupo
			if group.ExternalAuthFilter and group.ExternalAuthFilter ~= "" then
				vccInfoSecurity[vccAuthType].groupFilters[groupId] = group.ExternalAuthFilter
				vccInfo.Security = vccInfoSecurity
				-- Update Vcc
				ksProv = ksProv or utils.GetDatasource():CassandraConnection("Provisioning")
				local success, err = pcall(PM.UpdateVCC, keyspace, vcc, vccInfo, ksProv)
				if not success then
					ksProv:Close()
					LogError("ADMINISRATION.GROUPS:SaveGroups - Error saving external filter:'%s'", err)
					return false, { result = "ERROR_SAVING_EXTERNAL_FILTER"}
				end
			end
		end

		-- Clean up de los datos que no se deben almacenar en cassandra en la CF "Group"
		group.grouppermissions = nil
		group.ExternalAuthFilter = nil
	end

	if ksProv then
		ksProv:Close()
	end
	if onlyUpdateIntegration then
		return true
	end

	utils.SaveRows {
		keyspace = keyspace or "Agent",
		columnFamily = "Group",
		superColumns = true,
		rows = {
			[vcc] = groups
		}
	}

	return true
end

--[[
	-- Delete en Allegro los groups
]]--
local function DeleteAllegroGroupsIn55 ( groups, keyspace )
	local vcc = inconcert.authUser.virtualContactCenter

	if not utils.HasAllegroIntegration( vcc, keyspace ) then
		LogInfo("[ADMINISTRATION.GROUPS:DeleteAllegroGroupsIn55] - Allegro integration. Deleting groups '%s'. Vcc '%s' has not configured integration with allegro.", json.encode(groups), vcc)
		return false, { type = "WITHOUT_ALLEGRO_INTEGRATION" }
	end

	local success, err = allegroGroups.Delete(groups, vcc)
	if not success then
		if err.type == "ALLEGRO_INTEGRATION_ERROR" then
			LogError("[ADMINISTRATION.GROUPS:DeleteAllegroGroupsIn55] - Allegro integration error. Deleting groups '%s'. Response: '%s'.", json.encode(groups), json.encode(err))
		else
			LogInfo("[ADMINISTRATION.GROUPS:DeleteAllegroGroupsIn55] - Allegro integration. Deleting groups '%s'. Response: '%s'.", json.encode(groups), json.encode(err))
		end
		return false, err
	end

	return true
end

function RemoveGroups (vcc, groups, keyspace)
	--[[
		ATENCION:
			En HELPDESK NO se implementa el Delete de Roles, por este motivo NO implementamos la baja contra el API.
	]]--

	local success, err = DeleteAllegroGroupsIn55(groups, keyspace)
	if not success then
		-- Solo tiene que retornar sin eliminar el group en i6, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
		--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar la eliminacion en i6.
		if err and err.type == "ALLEGRO_INTEGRATION_ERROR" then
			keyspace:Close()
			return false, { result = err.type, err = (err.err or err.sub_type) }
		end
	end

	--#11602 - LicenseSystem - PM.UpdateVCC and PM.GetVCC now require a previous call to InitHelpersAgentServiceOptions
	local helpers_agentservice_options = utils.Get_m_helpers_agentservice_options()
	local success_agentservice_options, err_agentservice_options = pcall(PM.InitHelpersAgentServiceOptions, helpers_agentservice_options)
	
	local success, vccInfo = pcall(PM.GetVCC, keyspace, vcc, false)
	if not success then
		local stringErr = type(vccInfo) == "table" and json.encode(vccInfo) or vccInfo
		return false, {result = stringErr}
	end

	local vccInfoSecurity = json.decode(vccInfo.Security)
	local vccAuthType = vccInfoSecurity.authentication.type
	local ksProv
	for groupId, group in pairs(groups) do
		if vccAuthType ~= "BASIC" and vccAuthType ~= "MFA" then
			vccAuthType = vccAuthType:lower()
			local externalAuthConfig = vccInfoSecurity[vccAuthType]
			local groupFilters = externalAuthConfig.groupFilters
			--Eliminamos el grupo y su filtro
			vccInfoSecurity[vccAuthType].groupFilters[group] = nil
			vccInfo.Security = vccInfoSecurity
			-- Update Vcc
			ksProv = ksProv or utils.GetDatasource():CassandraConnection("Provisioning")
			local success, err = pcall(PM.UpdateVCC, keyspace, vcc, vccInfo, ksProv)
			if not success then
				ksProv:Close()
				LogError("ADMINISRATION.GROUPS:RemoveGroups - Error saving external filter:'%s'", err)
				return false, { result = "ERROR_SAVING_EXTERNAL_FILTER"}
			end
		end
	end

	if ksProv then
		ksProv:Close()
	end

	utils.RemoveColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "Group",
		superColumns = true,
		rows = {
			[vcc] = groups
		}
	}

	return true
end
