module(..., package.seeall)

local utils = require "shared.apps.utils"
local m_datasource
local m_helpers = require "helpers"

function GetName ()
	return "VCCs"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

function Get (vcc, keyspace, columns, skipLicenseInfo)
	assert(vcc)
	local vccInfo = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCs",
		key = vcc,
		columns = columns
	} or {}
	if skipLicenseInfo then
		return vccInfo
	end
	
	local options = utils.Get_m_helpers_agentservice_options()
	return m_helpers.utils.HasLicenseSystem() and m_helpers.agentservice.GetVccWithLicenseInfo(options, vccInfo) or vccInfo
end

function LoadClientsLayouts (vcc, keyspace)
	assert(vcc)
	local vccInfo = Get(vcc, keyspace)
	if not vccInfo then
		return false
	end
	local clients = vccInfo.Clients
	if not clients then
		return false
	end
	return clients.Layouts or false
end

function GetClientLayout (vcc, layoutId, keyspace)
	assert(vcc)
	assert(layoutId)
	local layouts = LoadClientsLayouts(vcc, keyspace)
	if not layouts then
		return false
	end
	local layout = layouts[layoutId:upper()]
	if not layout then
		return false
	end
	return layout
end

local function UpdateClientsLayouts (vcc, layouts, keyspace)
	local rows = {}
	rows[vcc] = {
		Clients = {
			Layouts = layouts
		}
	}
	utils.SaveRows {
		keyspace     = keyspace or "Agent",
		columnFamily = "VCCs",
		rows         = rows
	}
end

function SaveClientsLayouts (vcc, newLayouts, keyspace)
	local currentLayouts = LoadClientsLayouts(vcc, keyspace) or {}
	for layoutId, layout in pairs(newLayouts) do
		currentLayouts[layoutId] = layout
	end
	UpdateClientsLayouts(vcc, currentLayouts, keyspace)
	return true
end

function RemoveClientsLayout (vcc, layoutId, keyspace)
	local layouts = LoadClientsLayouts(vcc, keyspace)
	if not layouts then
		return false, "NO_LAYOUTS"
	end

	if not layouts[layoutId] then
		return false, "LAYOUT_DOES_NOT_EXIST"
	end

	layouts[layoutId] = nil
	UpdateClientsLayouts(vcc, layouts, keyspace)

	return true
end

function GetAppsIntegrations (vcc, keyspace)
	assert(vcc)
	local vccInfo = Get(vcc, keyspace)
	if not vccInfo then
		return false
	end
	return vccInfo.AppsIntegrations
end

local function LoadSpecificsAccounts ( accounts, accountType, accountId )
	if not accounts then
		return
	end

	if type( accounts ) == "string" then
		accounts = json.decode( accounts )
	end

	if accountType then
		accountType = accountType:upper()
		local accountsByType = accounts[ accountType ]
		if accountId then
			accountId = accountId:upper()
			return accountsByType and accountsByType[ accountId ]
		else
			return accountsByType
		end
	else
		return accounts
	end
end

function GetAccounts ( vcc, accountType, accountId, keyspace )
	local accounts = utils.GetColumnValue {
		keyspace     = keyspace or "Agent",
		columnFamily = "VCCs",
		key          = vcc,
		column       = "Accounts"
	}

	return LoadSpecificsAccounts( accounts, accountType, accountId )
end

function LoadAllVCCs (keyspace)
	local vccs = utils.GetAllColumns({
		keyspace     = keyspace or "Agent",
		columnFamily = "VCCs",
		startKey     = "",
		endKey       = ""
	}) or {}
	if m_helpers.utils.HasLicenseSystem() then
		local options = utils.Get_m_helpers_agentservice_options()
		for vcc, vccInfo in pairs(vccs) do
			vccInfo = m_helpers.agentservice.GetVccWithLicenseInfo(options, vccInfo)
		end
	end
	return vccs
end

function GetAccountsOfAllVCCs (accountType, keyspace)
	local vccs = LoadAllVCCs(keyspace)
	if not vccs then
		return
	end

	local accounts_by_vcc
	for _, vcc in pairs( vccs ) do
		accounts_by_vcc = accounts_by_vcc or {}
		accounts_by_vcc[ vcc.Id or _ ] = LoadSpecificsAccounts( vcc.Accounts, accountType )
	end

	return accounts_by_vcc
end

local function UpdateAccounts ( vcc, accounts, keyspace )
	local rows = {}
	rows[ vcc ] = {
		Accounts = accounts
	}
	utils.SaveRows {
		keyspace     = keyspace or "Agent",
		columnFamily = "VCCs",
		rows         = rows
	}
end

function SaveAccounts ( vcc, new_accounts, keyspace )
	local vcc_accounts = GetAccounts( vcc, nil, nil, keyspace )
	if not vcc_accounts then
		vcc_accounts = {}
	end

	for _, account in pairs( new_accounts ) do
		local accountType = ( account.proxyType or "" ):upper()
		vcc_accounts[ accountType ] = vcc_accounts[ accountType ] or {}
		local uid = ( account.uid or "" ):upper()
		vcc_accounts[ accountType ][ uid ] = account
	end

	UpdateAccounts( vcc, vcc_accounts, keyspace )
end

function RemoveAccounts ( vcc, accountType, accountId, keyspace )
	local accounts = GetAccounts( vcc, nil, nil, keyspace )
	if not accounts then
		return false, "NO_ACCOUNTS"
	end

	local accountType = ( accountType or "" ):upper()
	local accountsByType = accounts[ accountType ]
	if not accountsByType then
		return false, "ACCOUNT_TYPE_DOES_NOT_EXIST"
	end

	local accountId = ( accountId or "" ):upper()
	local account = accountsByType[ accountId ]
	if not account then
		return false, "ACCOUNT_DOES_NOT_EXIST"
	end

	accountsByType[ accountId ] = nil
	UpdateAccounts( vcc, accounts, keyspace )

	return true
end

