local oo = require "loop.base"
local dateClass = require "LuaOrm.Date"
require "inconcert.translators"
local Avatar = require "apps.agent.avatar"
local interactions = require "apps.agent.controllers.interactions"
local models = require("apps.agent.model"):GetModels()
local ir = require "lua_iragentlink"
local http_mapping = require "apps.agent.result_codes_to_http_codes"
local helpers = require "apps.agent.helpers"
local Fullcontact = require "fullcontact"
local model_perm = require "shared.permissions.models.permissions"

local utils = require "shared.apps.utils"

local function ExecuteFullcontactMethod (callback)
	local agentApp = require "apps.agent"
	agentApp.InitializeFullcontact(callback)
end

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil

local function ReplyJSON(web, data)
	web.headers["Content-Type"] = "application/json"
	if type(data) == "string" then
		return data
	end
	return json.encode(data)
end

--[[ campaign_id and app_id are optionals ]]--
local function HasPermission( action_id, campaign_id, app_id )
	app_id = app_id or "interaction" -- default

	local actions = {
		SAVE      = { interaction = "int_save_contact"     , supervisor = "sup_save_contact"     } ,
		DELETE    = { interaction = "int_delete_contact"   , supervisor = "sup_delete_contact"   } ,
		FAVORITE  = { interaction = "int_favorite_contact" , supervisor = "sup_favorite_contact" } ,
		-- MERGE, utiliza el mismo id que SAVE
		MERGE     = { interaction = "int_save_contact"     , supervisor = "sup_save_contact"     }
	}

	local has_permission = false
	local apps_actions = actions[ action_id ]

	if not apps_actions then
		LogError( "Contact permissions: Unrecognized action id '%s'.", action_id )
		return has_permission
	end

	local user_name = inconcert.authUser.userName
	local vcc = inconcert.authUser.virtualContactCenter
	local action = apps_actions[ app_id ]

	if campaign_id then
		has_permission = model_perm.IsPermissionEnabledForUser( app_id, action, campaign_id )
	else
		local user_campaigns = models.Campaigns.GetUserCampaigns( user_name, vcc )
		for campaign_id in pairs( user_campaigns ) do
			has_permission = model_perm.IsPermissionEnabledForUser( app_id, action, campaign_id )
			if has_permission then
				break
			end
		end
	end

	local user_id = user_name .. "@" .. vcc
	if has_permission then
		--LogDebug("User '%s', has permission ENABLED for action '%s' in campaign '%s'.", user_id, action_id, campaign_id )
	else
		LogWarning("User '%s', has permission DISABLED for action '%s' in campaign '%s'.", user_id, action_id, campaign_id )
	end

	return has_permission
end


local function UpdateInteractionsContact (vcc, actualContact, contact, obsoleteId)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)

	local function store_in_elasticsearch (vcc, interaction_id, contact)
		LogDebug("[store_in_elasticsearch]: Update interaction in elasticsearch: (vcc: %s) %s-%s", vcc, interaction_id, contact.Id)

		if not inconcert.environment.apps.agent.elasticsearch then
			LogError("[store_in_elasticsearch]: Elasticsearch is not configured")
			return false
		end

		local http = require "socket.http"
		local ltn12 = require "ltn12"
		local response_body = {}

		local data = {
			doc = {
				contact = contact.Id,
				contact_name = contact.Name
			}
		}

		local requestInfo = utils.GetElasticSearchInstance():MakeRequestInfo({
			query = data
		})

		local index_name = ("vcc-%s"):format( assert(vcc) )

		local ok, response_code, response_headers, response_status_line = http.request({
			host = inconcert.environment.apps.agent.elasticsearch.host,
			port = inconcert.environment.apps.agent.elasticsearch.port,
			url = ("/%s/interaction/%s/_update"):format(index_name, interaction_id),
			headers = requestInfo.headers,
			method = "POST",
			source = requestInfo.source,
			sink = ltn12.sink.table(response_body),
			redirect = false
		})
		if not ok then
			LogError("%s\n%s\n%s\n%s", response_code, json.encode(response_headers), response_status_line,
				json.encode(response_body))
			return false
		end

		response_body = table.concat(response_body)

		if response_headers["content-type"]:match("^application/json") then
			LogDebug("[store_in_elasticsearch]: response: %s", json.encode(response_body))
			return true, json.decode(response_body)
		end
		LogError("[store_in_elasticsearch]: Unknown response from elasticsearch. headers: %s\nresponse: %s", json.encode(response_headers), response_body)
		return true
	end
	
	local function setInteractionContact (vcc, interaction, contact)
		local result = irLink:SetInteractionContact(interaction.id, contact.Id, contact.Name)
		if not result then
			-- TODO: how we handle partial errors? For example, merge went OK but set interaction
			-- contact fail?
			LogError("Cannot merge contacts '%s' with obsolete contact: '%s', set interaction contact failed", contact.id, request.obsoleteId)
			return false, { result = error.result, reason = "Set interaction contact failed" }
		end
		--store_in_elasticsearch(vcc, interaction.id, contact)
	end

	-- Update obsolete contact interactions if contactId not equal obsoleteId
	local needUpdateObsoleteContactInteractions = contact.Id ~= obsoleteId

	-- Update owner contact interactions because change name contact (ES, redis, cassandra)
	local needUpdateContactInteractions = actualContact and actualContact.Name ~= contact.Name
	
	if 	needUpdateObsoleteContactInteractions then
		-- TODO: this should be done async cause it may take long to accomplish
		local interactions = models.Interactions.GetInteractionsFromContact(obsoleteId, vcc)
		for _, interaction in pairs(interactions) do
			setInteractionContact(vcc, interaction, contact)
		end
	end

	if needUpdateContactInteractions then
		local interactions = models.Interactions.GetInteractionsFromContact(contact.Id, vcc)
		for _, interaction in pairs(interactions) do
			setInteractionContact(vcc, interaction, contact)
		end
	end
end

---
--
function SaveContact (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local contact = web.request

	LogInfo("[AGENT.CONTACTS:SaveContact] User '%s@%s' requested to save contact '%s'", name, vcc, web.input.post_data)

	-- optionals, used by permissions
	local campaign_id = contact.__campaign_id
	local app_id = contact.__app_id -- optional

	if not HasPermission( "SAVE", campaign_id, app_id ) then
		web.status = 403
		return ReplyJSON(web, {result = "FORBIDDEN"})
	end
	-- delete attrs used by permissions
	contact.__campaign_id = nil
	contact.__app_id = nil

	-- we don't allow to save empty external id.
	if contact.externalId and #contact.externalId == 0 then
		web.status = 400
		return ReplyJSON(web, { result = "E_EMPTY_EXT_ID", reason = "Cannot save contact with an empty external id" })
	end

	-- Get the user campaigns
	local userCampaigns = models["Campaigns"].GetUserCampaigns(name, vcc)
	-- Get the current contact
	local currentContact
	ExecuteFullcontactMethod(function()
		currentContact = Fullcontact.Get(contact.Id, vcc)
	end)

	local validCampaigns = {}
	-- if there is no campaigns on posted data
	if not contact.campaigns or #contact.campaigns == 0 then
		contact.campaigns = {}
		--assume the contact is valid on all the user campaigns
		for k, role in pairs(userCampaigns) do
			if role ~= "invited" then
				table.insert(contact.campaigns, k)
			end
		end
	end

	-- de las campañas que envia, quedarse solo con aquellas a las cuales el usuario pertenece y no es un invitado
	for index, id in pairs(contact.campaigns) do
		if userCampaigns[id] and userCampaigns[index] ~= "invited" then
			validCampaigns[id] = id
		end
	end

	-- If we have already a contact on the database
	if currentContact then
		local remainingCampaigns = {}
		-- Keep the campaigns that user can't see
		for index, id in pairs(currentContact.campaigns) do
			if not userCampaigns[id] or userCampaigns[id] == "invited" then
				remainingCampaigns[id] = id
			end
		end
		-- And join them with the ones the user submitted
		for campaign in pairs(remainingCampaigns) do
			validCampaigns[campaign] = campaign
		end
	end

	-- TODO: Move this to helpers?
	local function to_array(t)
		local temp = {}
		for k,v in pairs(t) do
			table.insert(temp, v)
		end
		return temp
	end

	-- Now validCampaign holds the proper set of campaigns:
	-- (already existing campaigns and the ones the user added. The ones the user removed aren't here anymore)

	-- Convert to a proper array.
	contact.campaigns = to_array(validCampaigns)

	if #contact.campaigns > 0 then
		local success, err, partialError
		ExecuteFullcontactMethod(function()
			--[[
				Se invoca al metodo 'AddContactOnlyValidAddresses' para que valide y solo agregue addresses validas.
			]]--
			success, err, partialError = Fullcontact.AddContactOnlyValidAddresses(contact.Id, vcc, contact)
		end)
		local function ReturnErr (savedError)
			if not savedError then
				savedError = {code = "UNKNOWN", msg = "unknown"}
			end
			savedError.result = savedError.code
			savedError.reason = "Unable to save contact: ".. (savedError.msg or "unknown")
			web.status = 500
			LogError("[AGENT.CONTACTS:SaveContact] There is a problem saving contact: '%s': %s", contact.Id, json.encode(savedError))
			return ReplyJSON(web, savedError)
		end

		if not success then
			return ReturnErr(err)
		else
			if partialError then
				return ReturnErr(partialError)
			end
		end

		-- Update owner contact interactions if change name contact (ES, redis, cassandra)
		UpdateInteractionsContact(vcc, currentContact, contact, contact.Id)
	else
		LogError("[AGENT.CONTACTS:SaveContact] Cannot save contact '%s' with no related campaigns", contact.Id)
		web.status = 400
		return "Cannot save contact before beloging to a group, contacts do belong to groups"
	end

	LogInfo("[AGENT.CONTACTS:SaveContact] Success Updated Contact. Requested to save contact '%s'", web.input.post_data)
	return ReplyJSON(web, contact)
end

function MergeContacts (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	LogInfo("[AGENT.CONTACTS:MergeContacts] User '%s@%s' requested to merge contacts '%s' with: '%s' the last one becomes obsolete. Posted data: %s", name, vcc, request.id, request.obsoleteId, web.input.post_data)

	-- app_id is optional
	if not HasPermission( "SAVE", request.campaign_id, request.app_id ) then
		web.status = 403
		return ReplyJSON(web, {result = "FORBIDDEN"})
	end

	-- if the client does not specify contact campaigns
	-- assume all user campaigns
	if not request.contact.campaigns or #request.contact.campaigns == 0 then
		request.contact.campaigns = {}
		local userCampaigns = models["Campaigns"].GetUserCampaigns(name, vcc)
		--assume the contact is valid on all the user campaigns
		for k, role in pairs(userCampaigns) do
			if role ~= "invited" then
				table.insert(request.contact.campaigns, k)
			end
		end
	end

	local actualContact
	ExecuteFullcontactMethod(function()
		actualContact = Fullcontact.Get(request.id, vcc)
	end)

	local success, err
	ExecuteFullcontactMethod(function()
		success, err = Fullcontact.Merge(request.id, vcc, request.contact, request.obsoleteId)
	end)

	if not success then
		LogError("[AGENT.CONTACTS:MergeContacts] Cannot merge new contact '%s' with obsolete contact: '%s', err: %s", request.id, request.obsoleteId, err.msg)
		web.status = 400
		return ReplyJSON(web, {result = err.result, reason = err.msg, newContactId = request.id, obsoleteContactId = request.obsoleteId})
	end

	-- OK --
	web.headers["Content-Type"] = "application/json"

	-- TODO: Maybe the FullContact can return the merged contact?
	local mergedContact
	ExecuteFullcontactMethod(function()
		mergedContact = Fullcontact.Get(request.id, vcc)
	end)
	mergedContact.Avatar = nil

	UpdateInteractionsContact(vcc, actualContact, mergedContact, request.obsoleteId)

	return json.encode(mergedContact or {})
end

---
--
function DeleteContact (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	LogInfo("[AGENT.CONTACTS:DeleteContact] User '%s@%s' requested to delete contact '%s'", name, vcc, web.input.post_data)

	-- campaign_id and app_id is optionals, used by permissions
	if not HasPermission( "DELETE", request.campaign_id, request.app_id ) then
		web.status = 403
		return ReplyJSON(web, {result = "FORBIDDEN"})
	end

	local result, reason
	ExecuteFullcontactMethod(function()
		result, reason = Fullcontact.Delete(request.id, vcc)
	end)
	if not result then
		web.status = 500
		return reason
	end

	web.headers["Content-Type"] = "application/json"
	return "{}"
end

---
--
function GetContact (web, contactID)
	_M.response.DontCache(web)

	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local contact_id_at_vcc = helpers.AddVCC(contactID, vcc)

	--LogInfo("[AGENT.CONTACTS:GetContact] User '%s@%s' requested to retrieve contact '%s'.", name, vcc, contact_id_at_vcc)

	local contact
	ExecuteFullcontactMethod(function()
		contact = Fullcontact.Get(contact_id_at_vcc, vcc)
	end)
	if not contact then
		LogError("[AGENT.CONTACTS:GetContact] contact '%s' not found.", contact_id_at_vcc)
		return { result = "CONTACT_NOT_FOUND", message = "Contact not found" }
	end

	-- TODO los avatars deberian ir a una CF separada para no joder la piola
	contact.Avatar = nil
	interactions.SolveSubscribeIMContactsStatus({contact})

	--LogInfo("[AGENT.CONTACTS:GetContact] Successfully retrieved contact: '%s'.", contact_id_at_vcc)
	return contact
end

---
--
function GetContactByAddress (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	if not request.address then
		web.status = 400
		return "Invalid request"
	end

	local address = request.address
	local campaignId = request.campaignId

	--LogInfo("User '%s@%s' requested to get contact by address: '%s'", name, vcc, json.encode(address) )

	local contact
	local type = address.type or address.Type --TODO ver pq anda llegando como Type o type la key
	ExecuteFullcontactMethod(function()
		contact = Fullcontact.GetByAddress(type, address, vcc)
	end)
	if not contact then
		web.status = 404
		return { result = "CONTACT_NOT_FOUND", message = "Contact not found" }
	end

	-- This is done because the avatar comes with blob from cassandra
	-- and json fails to encode
	contact.Avatar = nil

	--LogInfo("Found contact '%s' by address", contact.Id)

	return contact
end

function GetContactByExternalId (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local userAtVcc = name .."@"..vcc
	local request = web.request

	if not request.externalId or #request.externalId == 0 then
		web.status = 400
		return "Invalid request"
	end

	--LogInfo("[AGENT.GetContactByExternalId] User '%s' requested to get contact by external id: '%s'", userAtVcc, request.externalId)

	local contact
	ExecuteFullcontactMethod(function()
		contact = Fullcontact.GetByExternalId(request.externalId, vcc)
	end)
	if not contact then
		LogWarning("[AGENT.GetContactByExternalId] Contact not found, by external id: '%s'", request.externalId)
		return { result = "CONTACT_NOT_FOUND", message = "Contact not found", externalId = request.externalId }
	end

	-- This is done because the avatar comes with blob from cassandra
	-- and json fails to encode
	contact.Avatar = nil

	--LogInfo("[AGENT.GetContactByExternalId] User: '%s' found contact '%s' by external id: '%s'", userAtVcc, contact.Id, request.externalId)

	return contact
end

---
-- Build and return a normalized address for a contact
--
-- @tparam table  web
-- @tparam table  web.request
-- @tparam string web.request.addressType  The address type (i.e. if mail: "Mail")
-- @tparam string web.request.address      The address (i.e. if mail: "smoreno@inconcertcc.com")
--
function BuildAddress (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local request = web.request

	--LogInfo("[AGENT.CONTACTS:BuildAddress] User '%s' requested to build address: '%s' and address type:'%s'", userId, request.address, request.addressType)

	if not request.address or not request.addressType then
		LogError("[AGENT.CONTACTS:BuildAddress] Invalid request, missing address or addressType parameters")
		web.status = 400
		return { result = "FAIL", message = "Invalid request, missing address or addressType parameters" }
	end

	local address
	ExecuteFullcontactMethod(function()
		address = Fullcontact.BuildAddress(request.addressType, request.address, vcc, request.options)
	end)
	if not address then
		LogError("[AGENT.CONTACTS:BuildAddress] Error building address")
		web.status = 500
		-- TO-DO: Proper error handling
		return { result = "FAIL", message = "Error building address" }
	end

	return address
end

--[[
this function receives a posted filter to query contacts by and returns a page with the matching results
{
	campaigns: {c1, c2, c3}  //optional
	accounts: {twitter, facebook, phone}
	text //free text to search against all contact data
	start //start row
	rows //total rows requested
}
]]--
function SearchContacts (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	LogInfo("[AGENT.CONTACTS:SearchContacts] User '%s@%s' is searching contacts with request '%s'", name, vcc, web.input.post_data)
	if not request.query or not request.last or not request.rows then
		web.status = 400
		return "Incomplete request"
	end

	-- until the contacts have a views view campaign
	-- we just use every campaign to make this query
	local vccCampaigns = {}
	for k, role in pairs(models.Campaigns.GetVCCCampaigns(vcc)) do
		table.insert(vccCampaigns, k.."@"..vcc)
	end

	if #vccCampaigns == 0 then
		LogInfo("[AGENT.CONTACTS:SearchContacts] User %s does not belong to a campaign, no contacts are returned", name)
		web.headers["Content-Type"] = "application/json"
		return json.encode({})
	end

	local filter = {}
	if request.query:len() > 0 then
		filter["Name"] = {exact = false, values = {}}
		table.insert(filter["Name"].values, request.query)

		filter["Address"] = {
			type = "address",
			addressType = request.addressType or nil,
			addressId = request.query:gsub("+", "")
		}
	end
	local sort = {{["Name.keyword"] = {order = "asc"}}}
	local params = {
		filter = filter,
		campaigns = vccCampaigns,
		startRow = request.last,
		rowsPerPage = request.rows,
		vcc = vcc,
		sort = sort
	}
	local success, response
	ExecuteFullcontactMethod(function()
		success, response = Fullcontact.Search(params)
	end)
	if not success then
		web.status = 500
		return response -- contains the error message
	end
	
	interactions.SolveSubscribeIMContactsStatus(response.contacts)

	local encodedResult = json.encode(response)
	--LogDebug("[AGENT.CONTACTS:SearchContacts] -----> Result: '%s'.", encodedResult)

	web.headers["Content-Type"] = "application/json"
	return encodedResult
end

--[[
this function receives a posted filter to query contacts by and returns a page with the matching results
{
	campaigns: {c1, c2, c3}  //optional
	accounts: {twitter, facebook, phone}
	text //free text to search against all contact data
	start //start row
	rows //total rows requested
}
]]--
function ExtendedSearch (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	--LogInfo("[AGENT.CONTACTS:ExtendedSearch] User '%s@%s' is searching contacts with request '%s'", name, vcc, web.input.post_data)
	if not request.query or not request.last or not request.rows then
		web.status = 400
		return "Incomplete request"
	end

	-- until the contacts have a views view campaign
	-- we just use every campaign to make this query
	local vccCampaigns = {}
	for k, role in pairs(models.Campaigns.GetVCCCampaigns(vcc)) do
		table.insert(vccCampaigns, k.."@"..vcc)
	end

	if #vccCampaigns == 0 then
		--LogInfo("[AGENT.CONTACTS:ExtendedSearch] User %s does not belong to a campaign, no contacts are returned", name)
		web.headers["Content-Type"] = "application/json"
		return json.encode({})
	end

	local sort = {{["Name"] = {order = "asc"}}}
	local params = {
		filter = request.query,
		campaigns = vccCampaigns,
		startRow = request.last,
		rowsPerPage = request.rows,
		vcc = vcc,
		sort = sort
	}
	local success, response
	ExecuteFullcontactMethod(function()
		success, response = Fullcontact.Search(params)
	end)
	if not success then
		web.status = 500
		return response -- contains the error message
	end
	--interactions.SolveSubscribeIMContactsStatus(response.contacts)

	web.headers["Content-Type"] = "application/json"
	return json.encode(response)
end

---
-- Adds the selected contact to the current user favorite list
--
function AddToFavorites (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	LogInfo("User '%s@%s' has added to favorites contact info '%s'", name, vcc, web.input.post_data)

	-- campaign_id and app_id is optionals, used by permissions
	if not HasPermission( "FAVORITE", request.campaign_id, request.app_id ) then
		web.status = 403
		return ReplyJSON(web, {result = "FORBIDDEN"})
	end

	local added
	ExecuteFullcontactMethod(function()
		added = Fullcontact.AddToFavorites(name, vcc, request.contactID, request.contactName)
	end)
	return {result = added and "OK" or "Already in Favorites"}
end

---
--	Removes the selected contact from the current user favorite list
--
function RemoveFromFavorites (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	LogInfo("User '%s@%s' has removed from favorites contact info %s", name, vcc, web.input.post_data)

	-- campaign_id and app_id is optionals, used by permissions
	if not HasPermission( "FAVORITE", request.campaign_id, request.app_id ) then
		web.status = 403
		return ReplyJSON(web, {result = "FORBIDDEN"})
	end

	local removed
	ExecuteFullcontactMethod(function()
		removed = Fullcontact.RemoveFromFavorites(name, vcc, request.contactID)
	end)
	return {result = removed and "OK" or "Not in Favorites"}
end

---
-- Sets the external ID of a given contact.
-- The request is a json with contactID and externalId fields.
--
function SetExternalId (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	if not request.contactID or not request.externalId then
		web.status = 400
		return { result = "FAIL", reason = "Invalid request" }
	end

	LogInfo("User '%s@%s' is setting the external id of contact '%s' to '%s'", name, vcc, request.contactID, request.externalId)

	local contact, err
	ExecuteFullcontactMethod(function()
		contact, err = Fullcontact.Get(request.contactID, vcc)
	end)
	if not contact then
		return { result = "Contact not found" }
	end

	contact.externalId = request.externalId
	local ok = contact:Update()

	return { result = ok and "OK" or "Failed to update contact" }
end

---
--	Retrieves the avatar for a specific contact in case it's available
--
function GetContactAvatar (web, id)
	local vcc = inconcert.authUser.virtualContactCenter

	--LogDebug("[CONTACTS:GetContactAvatar] Retrieving contact avatar. id: %s", id)

	web.headers["Content-Type"] = "image/png"
	web.headers["Cache-Control"] = "max-age=172800, public, must-revalidate"

	local avatar, err
	ExecuteFullcontactMethod(function()
		avatar, err = Fullcontact.GetAvatar(id, vcc)
	end)

	if not avatar then
		if err == "Avatar not found" then
			local contact, err
			ExecuteFullcontactMethod(function()
				contact, err = Fullcontact.Get(id, vcc)
			end)
			avatar = Avatar.FindBestAvatar(contact)
			if avatar then
				ExecuteFullcontactMethod(function()
					Fullcontact.SaveAvatar(id, vcc, avatar)
				end)
				return avatar.image
			end
		end
	else
		return avatar
	end

	local genericAvatar = Avatar.GetGenericAvatar()
	if not genericAvatar then
		return
	end
	return genericAvatar.image
end

function GetGenericAvatar (web)
	web.headers["Content-Type"] = "image/png"
	web.headers["Cache-Control"] = "max-age=172800, public, must-revalidate"

	--LogDebug("[CONTACTS:GetGenericAvatar] Retrieving Generic Avatar")

	local genericAvatar = Avatar.GetGenericAvatar()
	if not genericAvatar then
		return
	end
	return genericAvatar.image
end

function GetAddressAvatar (web, addressType, address)
	local vcc = inconcert.authUser.virtualContactCenter
	local addressAvatar

	addressType = addressType:upper()

	--LogDebug("[CONTACTS:GetAddressAvatar] Retrieving address avatar type %s address %s", addressType, address)
	web.headers["Content-Type"] = "image/png"
	web.headers["Cache-Control"] = "max-age=172800, public, must-revalidate"

	if addressType == "IM" then
		local protocol, buddy = string.match(address, "(.*)|(.*)")
		addressAvatar = Avatar.GetIMAvatar(buddy, protocol)
	elseif addressType == "MAIL" then
		addressAvatar = Avatar.GetMailAvatar(address)
	elseif addressType == "TWITTER" then
		-- Se comenta esto hasta que se solucino el incidente #6355
		--local screen_name, campaignAtVCC, account = string.match(address, "(.*)|(.*)|(.*)")
		--addressAvatar = Avatar.GetTwitterAvatar(screen_name, campaignAtVCC, account)
	end

	return (addressAvatar and addressAvatar.image) or GetGenericAvatar(web)
end

---
--	Retrieves favorite list for the user, TODO, should be paginated
--
function GetFavorites (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName

	--LogDebug("Retrieving favorites for user '%s@%s'", name, vcc)

	local result
	ExecuteFullcontactMethod(function()
		result = Fullcontact.GetUserFavorites({
			user = name,
			vcc = vcc
		})
	end)

	return result.contacts
end


---
-- Given a contact, will refresh social information (twitter for now) and save it to the contact
-- @return a json with the contact
--
function RefreshSocialInfo (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = web.request

	--LogDebug("Refreshing social info for user '%s@%s'", name, vcc)

	local contact
	ExecuteFullcontactMethod(function()
		contact = Fullcontact.Get(request.contactID)
	end)
	if not contact then
		web.status = 404
		return "USER_NOT_FOUND"
	end

	contact.Avatar = nil
	local accountsRefreshed = false

	for index, address in ipairs(contact.addresses) do
		if address.Type:lower() == "twitter" then
			local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
			local result = irLink:TwitterGetUser(address.User, request.campaignId, request.twitterAccount)
			if result["result"] == "OK" then
				local user = result["user_info"]
				-- insert new information
				for k, v in pairs(user) do contact.addresses[index][k] = v end
				accountsRefreshed = true
			
			elseif result["reason"] == "USER_NOT_FOUND" then
				web.status = 404
				return "TWITTER_USER_NOT_FOUND"
			
			else
				local http_code = http_mapping.getHTTPCode(result["result"])
				web.status = http_code or 500 --unknown result code
				return "GET_TWITTER_USER_ERROR"
			end
		end
	end

	if accountsRefreshed then
		contact.lastSocialUpdateTime = LuaOrm.Date():toString()
		local res
		ExecuteFullcontactMethod(function()
			res = Fullcontact.Update(request.contactID, vcc, contact)
		end)
		if not res then
			web.status = 500
			return "UNABLE_TO_SAVE_CONTACT"
		end
	end
	return ReplyJSON(web, contact)
end

function GetTwitterUser (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = json.decode(web.input["post_data"])

	--LogDebug("User %s on vcc %s it's trying to get twitter user: @%s data", name, vcc, request.userId)

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:TwitterGetUser(request.userId, request.campaignId, request.account)

	local http_code = http_mapping.getHTTPCode(res["result"])
	web.status = http_code or 500 --unknown result code

	if res.result ~= "OK" then
		LogError("There is an error trying to get twitter user: @%s on account: %s with error: '%s' ", request.userId, request.account, tostring(res.result))
	end

	web.headers["Content-Type"] = "application/json"
	return json.encode(res.user_info)
end

function FollowTwitterUser(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = json.decode(web.input["post_data"])

	--LogDebug("Start following user: '%s', from agent: '%s', on campaign: '%s', on account: '%s' ", request.userId, name, request.campaignId, request.account)

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)

	--this is done in order to avoid api manglers creating interactions from another vcc from the one they're logged in
	local campaignIdAtVCC = helpers.AddVCC(request.campaignId, vcc)
	local res = irLink:TwitterFollow(campaignIdAtVCC,
					request.account,
					request.userId)

	--Convert result code to HTTP code
	local http_code = http_mapping.getHTTPCode(res["result"])
	web.status = http_code or 500 --unknown result code

	return ReplyJSON(web, res)
end

function UpdateContactFollowers(web)
	
	local request = json.decode(web.input["post_data"])
	local vcc = inconcert.authUser.virtualContactCenter

	LogInfo("[UpdateContactFollowers]: contactID: '%s' followers: '%s' vcc: '%s'",request.contactID, request.followers, vcc)
	local contact
	ExecuteFullcontactMethod(function()
		contact = Fullcontact.Get(request.contactID, vcc)
	end)
	if not contact then
		web.status = 404
		return {result = "CONTACT_NOT_FOUND"}
	end

	contact.totalFollowers = request.followers
	contact.followersDate = LuaOrm.Date():toString()
	contact:Update()
	
	return {result = "OK"}
end

function UnFollowTwitterUser(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = json.decode(web.input["post_data"])

	LogInfo("Stop following user: '%s', from agent: '%s', on campaign: '%s', on account: '%s' ", request.userId, name, request.campaignId, request.account)

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)

	--this is done in order to avoid api manglers creating interactions from another vcc from the one they're logged in
	local campaignIdAtVCC = helpers.AddVCC(request.campaignId, vcc)
	local res = irLink:TwitterUnFollow(campaignIdAtVCC,
					request.account,
					request.userId)

	--Convert result code to HTTP code
	local http_code = http_mapping.getHTTPCode(res["result"])
	web.status = http_code or 500 --unknown result code

	return ReplyJSON(web, res)
end

function TwitterShowRelationship(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName .. "@" .. vcc
	local request = json.decode(web.input["post_data"])
	local rid = web.vars.HTTP_X_ICC_RID or "n/a"

	LogInfo("Agent: '%s' asked for Twitter relationship on campaign: '%s', account: '%s' to: '%s'", name, request.campaignId, request.account, request.to_account)

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)

	--this is done in order to avoid api manglers creating interactions from another vcc from the one they're logged in
	local campaignIdAtVCC = helpers.AddVCC(request.campaignId, vcc)
	local res = irLink:TwitterShowRelationship(
					campaignIdAtVCC,
					request.account,
					request.to_account,
					rid)

	if not res then
		web.status = 500
		local errorMessage = res.reason and res.reason.message or "Unknown proxy error"
		LogError("Error getting Twitter relationship from account: '%s' to account: '%s', error: '%s'", request.account, request.to_account, errorMessage)
		return ReplyJSON(web, { result = res.result, reason = errorMessage })
	end

	LogDebug("Success getting Twitter relationship from user: '%s' to user: '%s'", request.account, request.to_account)
	-- if ok, irlink returns relationship directly
	return ReplyJSON(web, res)
end

function TwitterBlockExists(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = json.decode(web.input["post_data"])

	--LogDebug("User %s on vcc %s it's trying to check if block exists on twitter user: @%s on account: %s", name, vcc, request.userId, request.account)

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:TwitterBlockExists(request.campaignId, request.account, request.userId)

	local http_code = http_mapping.getHTTPCode(res["result"])
	web.status = http_code or 500 --unknown result code

	if res.result ~= "OK" then
		LogError("There is an error trying to check if block exists on twitter user: @%s on account: %s with error: '%s' ", request.userId, request.account, tostring(res.result))
	end
	return ReplyJSON(web, res)
end

function TwitterBlockUser(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = json.decode(web.input["post_data"])

	--LogDebug("User %s on vcc %s it's trying to block twitter user: @%s on account: %s", name, vcc, request.userId, request.account)

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:TwitterBlockUser(request.campaignId, request.account, request.userId)

	local http_code = http_mapping.getHTTPCode(res["result"])
	web.status = http_code or 500 --unknown result code

	if res.result ~= "OK" then
		LogError("There is an error trying block twitter user: @%s on account: %s with error: '%s' ", request.userId, request.account, tostring(res.result))
	end
	return ReplyJSON(web, res.response)
end

function TwitterUnBlockUser(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local request = json.decode(web.input["post_data"])

	--LogDebug("User %s on vcc %s it's trying to unblock twitter user: @%s on account: %s", name, vcc, request.userId, request.account)

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:TwitterUnBlockUser(request.campaignId, request.account, request.userId)

	local http_code = http_mapping.getHTTPCode(res["result"])
	web.status = http_code or 500 --unknown result code

	if res.result ~= "OK" then
		LogError("There is an error trying unblock twitter user: @%s on account: %s with error: '%s' ", request.userId, request.account, tostring(res.result))
	end
	return ReplyJSON(web, res.response)
end

------------------------------------------
-- Addresses Book
--
local AddressesBook = {}
---
--
function AddressesBook.LoadByChannel (web, channel)
	_M.response.DontCache(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local loggedUserId = utils.GetCurrentLoggedUser()

	--LogDebug("[AGENT.CONTACTS:AddressesBook.LoadByChannel] User '%s' Loading addresses book by channel '%s' in vcc '%s'", loggedUserId, channel, vcc)

	local params = {vcc = vcc, channel = channel, withContactsInfo = true}
	local success, addressesBook
	ExecuteFullcontactMethod(function()
		success, addressesBook = Fullcontact.AddressesBook.ListByChannel(params)
	end)

	if not success then
		local err = addressesBook
		web.status = 400
		return {result = "LOADING ERROR", err = err}
	end

	-- OK
	--LogDebug("[AGENT.CONTACTS:AddressesBook.LoadByChannel] Successfully retrieved addresses book by channel: '%s' vcc: '%s'.", channel, vcc)
	--LogDebug("[AGENT.CONTACTS:AddressesBook.LoadByChannel] addressesBook: '%s'.", json.encode(addressesBook))

	return addressesBook
end

function AddressesBook.GetAll (web)
	_M.response.DontCache(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local loggedUserId = utils.GetCurrentLoggedUser()

	--LogDebug("[AGENT.CONTACTS:AddressesBook.GetAll] User '%s' Loading addresses book '%s' in vcc '%s'", loggedUserId, channel, vcc)

	local params = {vcc = vcc, withContactsInfo = true}
	local success, addressesBook
	ExecuteFullcontactMethod(function()
		success, addressesBook = Fullcontact.AddressesBook.List(params)
	end)

	if not success then
		local err = addressesBook
		web.status = 400
		return {result = "LOADING ERROR", err = err}
	end

	-- OK
	--LogDebug("[AGENT.CONTACTS:AddressesBook.GetAll] Successfully retrieved addresses book in vcc: '%s'.", vcc)
	--LogDebug("[AGENT.CONTACTS:AddressesBook.GetAll] addressesBook: '%s'.", json.encode(addressesBook))

	return addressesBook
end
-------------------------------------------------------------------

function MapUris (m, htmlRender, prefix)
	local p = prefix
	local Log = require("apps.agent.requestFilters").LogRequest
	local JsonValidator = utils.JsonValidator
	local JsonValidatorReply = utils.JsonValidatorReply
	local JsonReply = utils.JsonReply

	m:dispatch_post(Log .. JsonValidatorReply        .. SaveContact                 , p .. "/save"                               )
	m:dispatch_post(Log .. JsonValidatorReply        .. MergeContacts               , p .. "/merge"                              )
	m:dispatch_post(Log .. JsonValidatorReply        .. DeleteContact               , p .. "/delete"                             )

	m:dispatch_get (Log .. JsonReply                 .. GetContact                  , p .. "/get/(.*)"                           )
	m:dispatch_post(Log .. JsonValidatorReply        .. GetContactByAddress         , p .. "/get_by_address"                     )
	m:dispatch_post(Log .. JsonValidatorReply        .. GetContactByExternalId      , p .. "/get_by_external_id"                 )
	m:dispatch_post(Log .. JsonValidatorReply        .. BuildAddress                , p .. "/build_address"                      )
	m:dispatch_post(Log .. JsonValidatorReply        .. AddToFavorites              , p .. "/add_to_favorites"                   )
	m:dispatch_post(Log .. JsonValidatorReply        .. RemoveFromFavorites         , p .. "/remove_from_favorites"              )
	m:dispatch_get (Log .. JsonReply                 .. GetFavorites                , p .. "/get_favorites"                      )
	m:dispatch_post(Log .. JsonValidatorReply        .. SetExternalId               , p .. "/set_external_id"                    )
	m:dispatch_get (Log .. GetContactAvatar                                         , p .. "/get_contact_avatar/(.*)"            )
	m:dispatch_get (Log .. GetAddressAvatar                                         , p .. "/get_address_avatar/(.*)/(.*)"       )
	m:dispatch_get (Log .. GetGenericAvatar                                         , p .. "/get_generic_avatar"                 )
	m:dispatch_post(Log .. JsonValidator             .. SearchContacts              , p .. "/search"                             )
	m:dispatch_post(Log .. JsonValidator             .. ExtendedSearch              , p .. "/extended_search"                    )

	-- twitter
	m:dispatch_post(Log .. JsonValidator             .. RefreshSocialInfo           , p .. "/twitter/refresh_social_info"        )
	m:dispatch_post(Log .. FollowTwitterUser                                        , p .. "/twitter/follow_user"                )
	m:dispatch_post(Log .. UnFollowTwitterUser                                      , p .. "/twitter/unfollow_user"              )
	m:dispatch_post(Log .. TwitterBlockExists                                       , p .. "/twitter/block_exists"               )
	m:dispatch_post(Log .. TwitterBlockUser                                         , p .. "/twitter/block_user"                 )
	m:dispatch_post(Log .. TwitterUnBlockUser                                       , p .. "/twitter/unblock_user"               )
	m:dispatch_post(Log .. TwitterShowRelationship                                  , p .. "/twitter/show_relationship"          )
	m:dispatch_post(Log .. GetTwitterUser                                           , p .. "/twitter/get_user/"                  )
	m:dispatch_post(Log .. JsonValidatorReply         .. UpdateContactFollowers     , p .. "/update_contact_followers"          )

	-- addresses book
	m:dispatch_get (Log .. JsonReply                 .. AddressesBook.LoadByChannel , p .. "/addresses_book/load_by_channel/(.*)")
	m:dispatch_get (Log .. JsonReply                 .. AddressesBook.GetAll        , p .. "/addresses_book/get_all")
	m_htmlRender = htmlRender
end
