local helpers = require "apps.agent.helpers"

module(..., package.seeall)
local m_datasource
local agentservice = require "apps.agent.agentservice"
local agent_module = require "apps.agent.agent"
local agentInteractions = require "apps.agent.controllers.interactions"
local utils = require "shared.apps.utils"
local cjson = require "cjson"
local allegroUsers = require "apps.allegro.controllers.users"
local helpDeskUsers = require "apps.helpdesk.controllers.users"
local models = require("apps.agent.model"):GetModels()
local PM = require("provisioningmanager")

function GetName ()
	return "Users"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

function GetCassandra ()
	return m_datasource.cassandra
end

local function GetCompleteNameLoggedUser ()
	return inconcert.authUser.userName .. "@" .. inconcert.authUser.virtualContactCenter
end

-- INICIO de funciones del CACHE de Usuarios --------------------------------------------------------------------------------
--
--[[
	EL CACHE de usuarios se ha utilizado para la obtención de Usuarios del VCC.
	Específicamente su uso lo realiza la obtención de Usuarios de la Campana a los cuales se le podrá transferir la interacción (popup de transferencias)
	El mismo se almacena según la VM de LUa en la cual ha caído el request.
	Se cachean datos practicamente estáticos, como lo es el nombre de usuario (algo que nadie o casi nadie en el mundo cambia de nombre, y menos a cada rato)
	El CACHE tiene un tiempo de expiración al cual se le ha ingresado un valor en segundos de (m_expiredCacheUsersSec = 300 segundos '5 minutos')
	Esta validación de expiración se realiza cuando se va a pedir al cache un usuario determinado.
]]--
local m_expiredCacheUsersSec = 300
local m_cachedUsers

local function InitCacheUsers (userVCC, forceCleanUp)
	if not m_cachedUsers then
		m_cachedUsers = {}
	end
	local function Init()
		LogInfo("[InitCacheUsers] Initializing cache for VCC %s. Expires in %s seconds.", userVCC, m_expiredCacheUsersSec)
		m_cachedUsers[userVCC] = {
			cacheStartTime = os.time(),
			list = {}
		}
	end
	if forceCleanUp or not m_cachedUsers[userVCC] then
		Init()
	end
end

local function SetCacheUser (fullUserId, user)
	local _, userVCC = utils.SeparateVCC(fullUserId)
	InitCacheUsers(userVCC)
	m_cachedUsers[userVCC].list[fullUserId] = user
end

local function GetCachedUser (fullUserId)
	if not m_cachedUsers then
		--LogDebug("1 GetCachedUser")
		return
	end
	local _, userVCC = utils.SeparateVCC(fullUserId)
	local cachedUsersInfo = m_cachedUsers[userVCC]
	if not cachedUsersInfo then
		--LogDebug("2 GetCachedUser")
		return
	end
	local now = os.time()
	local elapsedTimeSec = now - cachedUsersInfo.cacheStartTime
	if elapsedTimeSec >= m_expiredCacheUsersSec then
		-- Cleanup del cache para el VCC
		LogInfo("[GetCachedUser] Cache time of users has expired in the VCC %s. Expires in %s seconds and %s seconds have elapsed time.", userVCC, m_expiredCacheUsersSec, elapsedTimeSec)
		local forceCleanUp = true
		InitCacheUsers(userVCC, forceCleanUp)
		return
	end
	--LogDebug("cachedUsersInfo: %s", cjson.encode(cachedUsersInfo))
	local cachedUser = cachedUsersInfo.list[fullUserId]
	--LogDebug("3 GetCachedUser cachedUser: %s fullUserId: %s", cachedUser, fullUserId)
	return cachedUser
end

function ExistUsersCache (vcc)
	if not m_cachedUsers then
		return false
	end
	if not m_cachedUsers[vcc] then
		return false
	end
	local now = os.time()
	local elapsedTimeSec = now - m_cachedUsers[vcc].cacheStartTime
	local expireInSec = m_expiredCacheUsersSec - elapsedTimeSec
	return true, expireInSec
end

--
-- FIN de funciones del CACHE de Usuarios --------------------------------------------------------------------------------

function GetVCCUsers (vcc, keyspace, count)
	count = count or 50000
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCUsers",
		key = vcc,
		columnCount = count
	} or {}
end

local function Lock (key, name, value)
	value = value or GetCompleteNameLoggedUser()
	assert( type(key) == "string" and type(name) == "string")
	--LogInfo("Locking key: '%s', name: '%s' by user: '%s'.", key, name, value)
	local locked = m_datasource.redis:hsetnx(key, name, value)
	if locked then
		--LogInfo("Lock operation is successful on key: '%s', name: '%s' by user: '%s'.", key, name, value)
	else
		LogError("Failed to lock key: '%s', name: '%s' by user: '%s'.", key, name, value)
	end
	return locked
end

local function Unlock (key, name)
	assert( type(key) == "string" and type(name) == "string")
	--LogInfo("Unlocking key: '%s', name: '%s'.", key, name)
	local unlocked = m_datasource.redis:del(key, name)
	if unlocked then
		--LogInfo("Unlock operation is successful on key: '%s', name: '%s'.", key, name)
	else
		LogError("Failed to unlock key: '%s', name: '%s'.", key, name)
	end
	return unlocked
end

local RESERVE_WORKGROUP_KEY = "Agent::Workgroups"
function LockWorkgroup (user)
	--workgroups are locked by vcc
	return Lock(RESERVE_WORKGROUP_KEY, user.VCC)
end

function UnlockWorkgroup (user)
	return Unlock(RESERVE_WORKGROUP_KEY, user.VCC)
end

local RESERVE_USER_KEY = "Agent::User"
function LockUser (user)
	return Lock(RESERVE_USER_KEY,  user.Id .. "@" .. user.VCC)
end

function UnlockUser (user)
	return Unlock(RESERVE_USER_KEY,  user.Id .. "@" .. user.VCC)
end

local RESERVE_USER_BY_MAIL_KEY = "Agent::UserByMail"
function LockMail (user)
	local mailToLock = user.Mail
	if not mailToLock then
		return true
	end
	return Lock(RESERVE_USER_BY_MAIL_KEY, mailToLock)
end

function UnlockMail (user)
	local mailToLock = user.Mail
	if not mailToLock then
		return true
	end
	return Unlock(RESERVE_USER_BY_MAIL_KEY, mailToLock)
end

local RESERVE_USER_BY_PEER_KEY = "Agent::UserByPeer"
function LockPeer (peerId)
	if not peerId then
		return true
	end
	return Lock(RESERVE_USER_BY_PEER_KEY, peerId)
end

function UnlockPeer (peerId)
	if not peerId then
		return true
	end
	return Unlock(RESERVE_USER_BY_PEER_KEY, peerId)
end

function UnlockUserData (user)
	UnlockUser(user)
	UnlockMail(user)
	if user and user.addresses and user.addresses.CALL then
		UnlockPeer(user.addresses.CALL.id)
	end
end

function IsEnabled (full_user_id, keyspace)
	local is_enabled = utils.GetColumnValue({
		keyspace = keyspace or "Agent",
		columnFamily = "Users",
		key = full_user_id,
		column = "Status"
	})
	return ( is_enabled == nil or tonumber(is_enabled) == 1 )
end

function IsSystemUser (full_user_id, keyspace)
	local is_system_user = utils.GetColumnValue({
		keyspace = keyspace or "Agent",
		columnFamily = "Users",
		key = full_user_id,
		column = "IsSystemUser"
	})
	return is_system_user == 1
end

function GetSpecificColumn ( full_user_id, specific_column, keyspace )
	if not specific_column then
		return
	end
	return utils.GetColumnValue({
		keyspace = keyspace or "Agent",
		columnFamily = "Users",
		key = full_user_id,
		column = specific_column
	})
end

--[[
	Returns the first user matching for a specific email address
]]--
function FindUserByEmail (emailAddress, vcc, keyspace)
	local close_keyspace = false
	if not keyspace then
		keyspace = GetCassandra():CassandraConnection("Agent")
		close_keyspace = true
	end

	local userList = keyspace:GetSliceRange(emailAddress,
		"UsersByEmail",
		"", 
		{
			column_names = {},
			slice_range = {
					start = "",
					finish = "",
					reversed = true,
					count = 2000
				}
		}
	)

	if close_keyspace then
		keyspace:Close()
	end

	--LogDebug("Mail a buscar: '%s', userList: '%s'.", emailAddress, userList)

	if userList then
		for k, v in ipairs(userList) do
			local invitedUserName, invitedUserVCC = helpers.SeparateVCC(v.name)
			--LogDebug("v.name: '%s', invitedUserName: '%s', invitedUserVCC: '%s'", v.name, invitedUserName, tostring(invitedUserVCC))
			if invitedUserVCC == vcc then
				return v.name
			end
		end
	end
end

function FindUserByPeer (peer, keyspace)
	local close_keyspace = false
	if not keyspace then
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
		close_keyspace = true
	end

	local userList = keyspace:GetSliceRange(peer,
		"UsersByPeer",
		"",
		{
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = true,
				count = 20000
			}
		}
	)

	if close_keyspace then
		keyspace:Close()
	end

	--LogDebug("Peer a buscar: '%s', userList: '%s'.", peer, userList)

	return userList
end

function FindUsersByState (stateId, keyspace)
	local release = false
	if not keyspace then
		release = true
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
	end

	local userList = keyspace:GetSliceRange(
		stateId,
		"UsersByState",
		"",
		{
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = true,
				count = 5000
			}
		}
	) or {}

	if release then
		keyspace:Close()
	end

	--LogDebug("State to find: '%s', userList: '%s'.", stateId, userList)

	return userList
end

function UserExists (userAtVCC, keyspace)
	local closeKeyspace
	if not keyspace then
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
		closeKeyspace = true
	end
	local res, column = pcall(keyspace.GetColumnValue, keyspace, userAtVCC, "Users", "Id")
	if closeKeyspace then
		keyspace:Close()
	end
	return res and column
end

--[[
	Returns the campaign with account information included

	userAtVCC         (requerido) el id del usuario
	can_add_column_fn (opcional)  funcion que valida si la columna correspondiente al usuario se debe incluir en la respuesta.
	keyspace          (opcional)  el keyspace de quien me invoca, es utilizado para utilizar la misma conexion a cassandra de quien me ha invocado.
]]--
function Get (fullUserId, can_add_column_fn, keyspace, mustLoadFromCache)
	if mustLoadFromCache then
		local cachedUser = GetCachedUser(fullUserId)
		if cachedUser then
			--LogDebug("1 - LOADED USER FROM CACHE %s", fullUserId)
			return cachedUser
		--else
		--	LogDebug("2 - LOADED USER FROM BD %s", fullUserId)
		end
	end

	can_add_column_fn = can_add_column_fn or function() return true end

	local release = false
	if not keyspace then
		release = true
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
	end

	local userColumns = keyspace:GetSliceRange(
		fullUserId,
		"Users",
		"",
		{
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = true,
				count = 2000
			}
		}
	)

	local user = {}
	if userColumns then
		for k, v in ipairs(userColumns) do
			local column_name = v.name
			if can_add_column_fn(column_name) then
			-- En la funcion "JSON_Decode" se realiza "decode" si el valor es un cjson, sino retorna el valor sin modificar.
				user[column_name] = utils.JSON_Decode(v.value)
			end
		end

		local simple_user_id, user_vcc = utils.SeparateVCC(fullUserId)
		if can_add_column_fn("groups") then
			-- agrego las campanas a la cuales pertenece el usuario

			local userCampaigns = models.Campaigns.GetUserCampaigns(simple_user_id, user_vcc)
			user.groups = {}
			for camp_id, role_id in pairs(userCampaigns or {}) do
				table.insert(user.groups, { id = camp_id, role = role_id, invite = true })
			end
		end

		if can_add_column_fn("addresses") then
			--now go fetch the user addresses
			local rows = keyspace:GetSuperRangeSlices(
				"AgentAddresses",  --column family
				"",  --supercolumn
				fullUserId, --startkey
				fullUserId, --endkey
				{
					column_names = {},
					slice_range = {
						start = "",
						finish = "",
						reversed = false,
						count = 2000
					}
				},
				200 --total keys
			) or {}

			user.addresses = {}
			local addressTypeMapper = {
				CHAT = "Chat",
				MAIL = "Mail"
			}
			local hasAllegro = utils.HasAllegroIntegration(user_vcc)
			if hasAllegro then
				addressTypeMapper["CALL"] = "Phone"
			else
				addressTypeMapper["CALL"] = "builtin"
			end
			for key, agentAddresses in rows:Iterator() do
				if key == fullUserId then
					for k, v in pairs(agentAddresses) do
						local address = {}
						for colid, column in pairs(v.columns) do
							address[column.name] = column.value
						end
						--the supercolumn key is the address type, only one allowed by agent
						address.Type = addressTypeMapper[v.name]
						address.ClientType = v.name
						user.addresses[v.name] = address
					end
					break
				end
			end
			if not hasAllegro then
				local removeCallAddress = true -- Remove unless the parameters from peers were retrieved successfully
				local peerId = user.addresses.CALL and user.addresses.CALL.id

				if peerId then
					local tech, peerName = PM.SeparateText(peerId, "_")
					local ksProv = m_datasource.cassandra:CassandraConnection("Provisioning")
					local success, currentPeer = pcall(PM.GetPeerWithHostInfo, ksProv, tech, peerName)
					ksProv:Close()

					if not success or utils.TableIsEmpty(currentPeer) then
						LogWarning("[AGENT:Users.Get] Unable to Assign Peer to User '%s'. Peer %s does not exist.", fullUserId, peerId)
					else
						if not currentPeer.VCC then
							LogWarning("[AGENT:Users.Get] CALL peer %s for agent %s is NOT assigned to any VCC", peerId, fullUserId)
						elseif currentPeer.VCC ~= user_vcc then
							LogWarning("[AGENT:Users.Get] CALL peer %s for agent %s is assigned to a different VCC %s", peerId, fullUserId, currentPeer.VCC)
						else
							for name, value in pairs(currentPeer) do
								--LogDebug(" -> %s = %s", name, value)
								user.addresses.CALL[name] = utils.JSON_Decode(value)
							end
							removeCallAddress = false
						end
					end
				end
				if removeCallAddress then
					user.addresses.CALL = nil
				end
				--LogDebug("[AGENT:Users.Get] fullUserId %s without allegro, user.addresses.CALL %s", fullUserId, json.encode(user.addresses.CALL or {}))
			end
		end
	end

	if release then
		keyspace:Close()
	end

	if mustLoadFromCache then
		--LogDebug("3 - CACHE USER %s", fullUserId)
		SetCacheUser(fullUserId, user)
	end

	return user
end

function UpdateUserStates (states, keyspace)
	if not states then
		return
	end

	local release = false
	if not keyspace then
		release = true
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
	end

	local vcc = inconcert.authUser.virtualContactCenter
	local initial_state = agent_module.GetUserInitialState(vcc)

	local function Update(stateId, userId)
		LogInfo("User: '%s' changed initial state from '%s' to '%s'.", userId, stateId, BREAK)
		-- update column family Users
		keyspace:InsertColumn( userId, "Users", "initialState", BREAK )
		-- update column family UsersByState
		keyspace:RemoveColumn( stateId      , "UsersByState", userId )
		keyspace:InsertColumn( initial_state, "UsersByState", userId , "")
	end

	for i, stateId in ipairs(states) do
		local usersByState = FindUsersByState(stateId, keyspace)
		for k, v in ipairs(usersByState) do
			local userId = v.name
			Update(stateId, userId)
		end
	end

	if release then
		keyspace:Close()
	end
end

function UpdateUserSkills (keyspace, user_id, user_skills)

	user_skills = user_skills or {}
	LogInfo("Updating user skills: User '%s', skills :'%s'.", user_id, cjson.encode(user_skills))

	local close_keyspace
	if not keyspace then
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
		close_keyspace = true
	end

	keyspace:RemoveColumn(user_id, "AgentSkills")
	for _, user_skill in ipairs(user_skills) do
		keyspace:InsertColumn(user_id, "AgentSkills", user_skill.skill, tostring(user_skill.weight))
	end

	if close_keyspace then
		keyspace:Close()
	end

	-- actualizo la info de redis que utiliza el agentservice
	-- arreglar nombres de propiedades
	local skills = {}
	for _, skill in ipairs(user_skills) do
		table.insert(skills, { name = skill.skill, score = skill.weight })
	end
	agentservice:GetAgentClient():UpdateAgentSkills(user_id, skills)
end

--[[
	-- Guarda en Allegro el usuario
]]--
function UpdateAllegroUserIn55 (user, vcc, keyspace, allegroOtherInfo)
	if not utils.HasAllegroIntegration( vcc, keyspace ) then
		LogInfo("[AGENT.USERS:UpdateAllegroUserIn55] Allegro integration. Updating user '%s'. Vcc '%s' has not configured integration with allegro.", user.Id, vcc)
		return false, { type = "WITHOUT_ALLEGRO_INTEGRATION" }
	end

	if user.IsSystemUser == "1" then
		-- Los SystemUser, no se envian a actualizar al server de 5.5. Por aca nunca deberia pasar ya que estamos actualizando en administration.
		LogInfo("[AGENT.USERS:UpdateAllegroUserIn55] Unable to update system user. Updating user '%s'.", user.Id)
		return false, { type = "ALLEGRO_INTEGRATION_UNABLE_UPDATE_SYSTEM_USERS" }
	end

	if user.CreatedBy and user.CreatedBy.version == "allegro" then
		local fromApp = user.CreatedBy.fromApp
		user.CreatedBy.update_number = (user.CreatedBy.update_number or 0) + 1
		if fromApp ~= "administration" and user.CreatedBy.update_number == 1 then
			-- Desde la app PROVISIONING, cuando es la primera vez que se crea un usuario desde allegro (cuando se ha creado el VCC, no se debe mandar a actualizar hacia 5.5)
			LogInfo("[AGENT.USERS:UpdateAllegroUserIn55] Unable to create user '%s'.", user.Id)
			return false, { type = "ALLEGRO_INTEGRATION_UNABLE_UPDATE_FIRST_USER_CREATED_BY_ALLEGRO" }
		end
	end

	local success, err = allegroUsers.Save(user, vcc, allegroOtherInfo)
	if not success then
		LogError("[AGENT.USERS:UpdateAllegroUserIn55] Allegro integration error. Updating user '%s'. Error: '%s'.", user.Id, cjson.encode(err or {}))
		return false, { type = "ALLEGRO_INTEGRATION_ERROR", err = err }
	end

	return true
end

--[[
	-- Guarda en HelpDesk el usuario
]]--
function UpdateHelpDeskUser (user, vcc, keyspace)
	if not utils.HasHelpDeskIntegration(vcc, keyspace) then
		--LogInfo("[AGENT.USERS:UpdateHelpDeskUser] HELPDESK integration. Updating user '%s'. Vcc '%s' has not configured integration with HELPDESK.", user.Id, vcc)
		return false, {type = "WITHOUT_HELPDESK_INTEGRATION"}
	end

	if user.IsSystemUser == "1" then
		-- Los SystemUser, no se envian a actualizar al server de HELPDESK. Por aca nunca deberia pasar ya que estamos actualizando en administration.
		LogInfo("[AGENT.USERS:UpdateHelpDeskUser] Unable to update system user. Updating user '%s'.", user.Id)
		return false, {type = "HELPDESK_INTEGRATION_UNABLE_UPDATE_SYSTEM_USERS"}
	end

	local success, err = helpDeskUsers.Save(user, vcc)
	if not success then
		LogError("[AGENT.USERS:UpdateHelpDeskUser] HELPDESK integration error. Updating user '%s'. Error: '%s'.", user.Id, err)
		-- Sample --> err = {type = "HELPDESK_INTEGRATION_ERROR", sub_type = helpdeskResponse}
		return false, err
	end

	return true
end

--[[
Actualiza en Cassandra informacion relevante de 5.5 para el Usuario.
]]
local function UpdateAllegroInfoInI6 (full_user_id, user, vcc, keyspace )
	LogInfo("[AGENT.USERS:UpdateAllegroInfoInI6] Updating Allegro integration on User '%s'.", full_user_id)

	if not user.AllegroIntegration then
		LogWarning("[AGENT.USERS:UpdateAllegroInfoInI6] Incomplete information: Allegro integration does not exist in User '%s' data.", full_user_id)
	else
		local success, encoded_allegro_information = pcall(cjson.encode, user.AllegroIntegration)
		if success then
			keyspace:InsertColumn(full_user_id, "Users", "AllegroIntegration", encoded_allegro_information)
		else
			LogWarning("[AGENT.USERS:UpdateAllegroInfoInI6] Error encoding allegro information for User '%s', data: '%s'.", full_user_id, user.AllegroIntegration)
		end
	end

	return true
end

--[[
	Saves the user with information included

	* Users (column)

	* UsersByEmail (column)
	* VCCUsers (column)

	* UserCampaigns (column)
	* CampaignMembers (column)

	AgentAddresses (supercolumn)
	-- si es call
	UsersByPeer (column)

	AgentSkills (column)
]]--
function Save (id, vcc, user, keyspace, allegroOtherInfo)
	LogInfo("[AGENT.Users.Save] - Saving User: '%s' VCC: '%s'", id, vcc)
	--LogDebug("1-------> user: %s", cjson.encode(user))
	user.Avatar = nil
	--This is to avoid update user data on external applications if the vcc has them (helpdesk, allegro, ..
	local saveFromLogin = user.saveFromLogin or false
	user.saveFromLogin = nil

	local close_ks = false
	if not keyspace then
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
		close_ks = true
	end

	local full_user_id = id .. "@" .. vcc
	-- Get the synchronous interaction limit
	-- * If it's specified on 'user', use that
	-- * If the user already exists, use the current (don't set anything)
	-- * If the user is new, use the default in the VCC
	local userExists = UserExists(full_user_id, keyspace)
	-- Use the same criteria for FreeSecondsAfterInteraction
	if not user.FreeSecondsAfterInteraction and not userExists then
		local res, value = pcall(keyspace.GetColumnValue, keyspace, vcc, "VCCs", "FreeSecondsAfterInteraction")
		user.FreeSecondsAfterInteraction = res and value
	end

	if not saveFromLogin then
		-- Guarda en HelpDesk el usuario
		-- Debo mandar a actualizar aca, en este lugar, antes de guardar algo en cassandra,
		--   asi me evito el Rollback y toda esas cosas!!!
		--
		local success, helpDeskInfo = UpdateHelpDeskUser(user, vcc, keyspace)
		if not success then
			-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "HELPDESK_INTEGRATION_ERROR",
			--  ya que otro tipo de error, puede ser que no tenga integracion con HelpDesk y por lo tanto en si, no es un error y debe de continuar
			--  la actualizacion en i6.
			if helpDeskInfo and helpDeskInfo.type == "HELPDESK_INTEGRATION_ERROR" then
				if close_ks then keyspace:Close() end
				return false, {result = helpDeskInfo.type, message_error = helpDeskInfo.sub_type}
			end
		end

		-- Guarda en Allegro el usuario
		-- Debo mandar a actualizar aca, en este lugar, antes de guardar algo en cassandra,
		--   asi me evito el Rollback y toda esas cosas!!!
		--
		local success, allegroInfo = UpdateAllegroUserIn55(user, vcc, keyspace, allegroOtherInfo)
		if not success then
			-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
			--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar
			--  la actualizacion en i6.
			if allegroInfo and allegroInfo.type == "ALLEGRO_INTEGRATION_ERROR" then
				if close_ks then keyspace:Close() end
				return false, {result = allegroInfo.type}
			end
		end
	end

	-- Agrega la company info (la que llega en el post, o sino la por default)
	local company_information = user.CompanyInformation or { Location = "", Organization = "", Position = "" }
	keyspace:InsertColumn(full_user_id, "Users", "CompanyInformation", cjson.encode(company_information))

	-- Agrega los custom info (la que llega en el post, o sino la por default)
	local custom_information = user.CustomInformation or {
		custom_agent_info_1 = "", custom_agent_info_2 = "", custom_agent_info_3 = "",
		custom_agent_info_4 = "", custom_agent_info_5 = ""
	}
	keyspace:InsertColumn(full_user_id, "Users", "CustomInformation", cjson.encode(custom_information))

	-- Agrega informacion relevante de las ventanas de los diferentes canales (layouts)
	local clients = user.Clients or {Layouts = {}}
	keyspace:InsertColumn(full_user_id, "Users", "Clients", cjson.encode(clients))

	-- Agrega el historial de passwords
	local pwdHistory = user.PasswordHistory or {}
	keyspace:InsertColumn(full_user_id, "Users", "PasswordHistory", cjson.encode(pwdHistory))

	if user.SupervisorWorkgroups then
--[[
	Enforce the following structure
	{
		campaign1: {wkg1: true, wkg2: true, wkg3:true},
		campaign2: {wkg4: true},
		..
	}
]]
		local wkg_error = {result = "SUPERVISORWORKGROUP_HAS_AN_INVALID_STRUCTURE"}
		if type(user.SupervisorWorkgroups) ~= "table" then
			if close_ks then keyspace:Close() end
			wkg_error.description = "SupervisorWorkgroups is not a table"
			return false, wkg_error
		end

		for campaignName, workgroups in pairs(user.SupervisorWorkgroups) do
			if type(campaignName) ~= "string" then
				if close_ks then keyspace:Close() end
				wkg_error.description = "Campaign name is not a string"
				return false, "SupervisorWorkgroups has an invalid structure"
			end
			if type(workgroups) ~= "table" then
				if close_ks then keyspace:Close() end
				wkg_error.description = "Workgroups name is not a table"
				return false, wkg_error
			end
			for workgroupId, enabled in pairs(workgroups) do
				if type(workgroupId) ~= "string" then
					wkg_error.description = "Workgroup id is not a string"
					if close_ks then keyspace:Close() end
					return false, wkg_error
				end
				if type(enabled) ~= "boolean" then
					wkg_error.description = "Enabled attribute is not a boolean"
					if close_ks then keyspace:Close() end
					return false, wkg_error
				end
			end
		end
		--everything is right, save it
		keyspace:InsertColumn(full_user_id, "Users", "SupervisorWorkgroups", cjson.encode(user.SupervisorWorkgroups))
	end

	if user.WorkgroupId == false then
		keyspace:RemoveColumn(full_user_id, "Users", "WorkgroupId")
	end

	local oldUser
	if userExists then
		oldUser = Get(full_user_id, nil, keyspace)
	end

	--remove jwtToken sended by client and not necessary anymore
	user.jwtToken = nil

	--save the user
	--LogDebug("---> User: '%s'", cjson.encode(user))
	for columnName, columnValue in pairs(user) do
		if type(columnValue) == "string" then
			keyspace:InsertColumn(full_user_id, "Users", columnName, tostring(columnValue))
		end
	end

	local clonedUser = utils.DeepCopy(user)
	clonedUser.Status = clonedUser.Status or "1"
	clonedUser.IsSystemUser = clonedUser.IsSystemUser or "0"
	clonedUser.HasRoleOverride = clonedUser.HasRoleOverride or "0"

	keyspace:InsertColumn(vcc, "VCCUsers", id, json.encode(clonedUser))

	if user.Mail then
		keyspace:InsertColumn(user.Mail, "UsersByEmail", full_user_id, "")
	else
		keyspace:RemoveColumn(full_user_id, "Users", "Mail")
	end

	if oldUser and oldUser.Mail then
		utils.RemoveColumns({
			keyspace = keyspace,
			columnFamily = "UsersByEmail",
			rows = {oldUser.Mail}
		})
	end

	--update column family UsersByGroup
	if user.UserGroup then
		-- insert or update
		local newGroupId = user.UserGroup .. "@" .. user.VccGroup
		keyspace:InsertColumn(newGroupId, "UsersByGroup", full_user_id, "")
		if oldUser then
			local oldGroupId = oldUser.UserGroup .. "@" .. oldUser.VccGroup
			if newGroupId:lower() ~= oldGroupId:lower() then
				keyspace:RemoveColumn(oldGroupId, "UsersByGroup", full_user_id)
			end
		end
	end

	--update column family UsersByState
	if utils.HasAllegroIntegration( vcc, keyspace ) then
		user.initialState = "BREAK"
	else
		user.initialState = user.initialState or "BREAK"
	end

	-- insert or update
	local newStateId = user.initialState
	keyspace:InsertColumn(newStateId, "UsersByState", full_user_id, "")
	if oldUser then
		local oldStateId = oldUser.initialState or "BREAK"
		if newStateId:lower() ~= oldStateId:lower() then
			keyspace:RemoveColumn(oldStateId, "UsersByState", full_user_id)
		end
	end

	--save user campaign and campaign members
	if type(user.groups) == "table" then
		local campaignModel = require("apps.agent.models.campaigns")

		--campaigns with user's interactions
		local problematicCampaigns = {}

		local newCampaigns = {}
		for _, camp in ipairs(user.groups) do
			if not camp.id:match("^(.*)@([^@]+)$") then
				camp.id = camp.id.."@"..vcc
			end
			local campIdWithVCC = camp.id:lower()
			newCampaigns[campIdWithVCC] = camp.role
		end

		local userCampaigns = campaignModel.GetUserCampaigns(id, vcc, keyspace)
		for campaign, role in pairs(userCampaigns) do
			campaign = campaign:lower()
			local campaignId = helpers.StripVCC(campaign, vcc)

			if not newCampaigns[campaign] then
				local assignedInteractions = agentInteractions.RetrieveUserInteractions(id, utils.AppendVcc(campaignId, vcc), keyspace, vcc)
				if next(assignedInteractions) then
					LogWarning("User %s has assigned interactions for campaign %s. Can not remove now", id, campaignId)
					table.insert(problematicCampaigns, {id = campaignId, description = "Has assigned interactions"})
				else --campaign has been deleted
					LogInfo("Removing member %s from campaign %s", full_user_id, campaign)
					agentservice.RemoveAgentFromCampaign(id, campaignId, vcc)
				end
			elseif newCampaigns[campaign] ~= role then
				--only his role has changed
				LogInfo("Changing role to %s for user %s in campaign %s", newCampaigns[campaign], full_user_id, campaign)
				--AddMember works for this purpose too
				agentservice.AddAgentToCampaign(id, campaignId, vcc, newCampaigns[campaign])
			end
		end

		for campaign, role in pairs(newCampaigns) do
			if not userCampaigns[campaign] then
				--its a new campaign
				LogInfo("Adding member %s for campaign %s", full_user_id, campaign)
				local campaignId = helpers.StripVCC(campaign, vcc)
				agentservice.AddAgentToCampaign(id, campaignId, vcc, role)
			end
		end

		if next(problematicCampaigns) then
			if close_ks then keyspace:Close() end
			return false, {result = "USER_HAS_INTERACTIONS_ON_CAMPAIGNS", problematicCampaigns = problematicCampaigns}
		end
	end

	--save user addresses
	if user.addresses then
		local new_peer_id
		for proxyType, address in pairs(user.addresses) do
			if type(address) == "table" then
				for column, value in pairs(address) do
					if type(value) == "string" then
						keyspace:InsertSuperColumn(full_user_id, "AgentAddresses", proxyType, column, value)
					end
				end
				if proxyType == "CALL" then
					if address.tech and address.name then
						new_peer_id = string.format("%s_%s", address.tech, address.name)
						keyspace:InsertColumn(new_peer_id, "UsersByPeer", full_user_id, "")
					end
				end
			end
		end

		local old_call_address = oldUser and oldUser.addresses and oldUser.addresses.CALL

		if old_call_address then
			local old_peer_id = string.format("%s_%s", old_call_address.tech, old_call_address.name)
			-- 1 - Borro la direccion del usuario si antes tenia CALL y ahora no
			if old_peer_id and not new_peer_id then
				LogInfo("[AGENT.Users.Save] - Removing '%s' Old CALL address of user'%s'", old_peer_id, full_user_id)
				utils.RemoveColumns {
					keyspace = keyspace,
					columnFamily = "AgentAddresses",
					superColumns = true,
					rows = {
						[full_user_id] = {"CALL"}
					}
				}
			end
			-- 2 - Si el peer ha cambiado tengo que eliminar la relacion usuario / peer anterior.
			if old_peer_id ~= new_peer_id then
				LogInfo("[AGENT.Users.Save] - Updating relation users by peer. User '%s' old peer: '%s', new peer: '%s'.", full_user_id, old_peer_id, new_peer_id)
				utils.RemoveColumns {
					keyspace = keyspace,
					columnFamily = "UsersByPeer",
					rows = {old_peer_id}
				}
			end
		end
	end

	--actualizo el mail de las addresses del usuario
	if oldUser and oldUser.addresses then
		local oldUserHasMailConfigured = false
		for proxyType, address in pairs(oldUser.addresses) do
			if proxyType == "MAIL" and type(address) == "table" then
				oldUserHasMailConfigured = true
				for column, value in pairs(address) do
					if type(value) == "string" then
						if column == "address" or column == "address_id" then
							if user.Mail then
								keyspace:InsertSuperColumn(full_user_id, "AgentAddresses", proxyType, column, user.Mail)
							else
								keyspace:RemoveSuperColumn(full_user_id, "AgentAddresses", proxyType, column)
								keyspace:RemoveSuperColumn(full_user_id, "AgentAddresses", proxyType, "resource_address")
							end
						end
					end
				end
			end
		end

		if not oldUserHasMailConfigured and user.Mail then
			keyspace:InsertSuperColumn(full_user_id, "AgentAddresses", "MAIL", "address", user.Mail)
			keyspace:InsertSuperColumn(full_user_id, "AgentAddresses", "MAIL", "address_id", user.Mail)
			keyspace:InsertSuperColumn(full_user_id, "AgentAddresses", "MAIL", "resource_address", "")
		end
	end

	UpdateUserSkills( keyspace, full_user_id, user.skills )

	--actualiza informacion sobre allegro integration en cassandra
	if user.AllegroIntegration then
		UpdateAllegroInfoInI6( full_user_id, user, vcc, keyspace )
	end

	if close_ks then keyspace:Close() end

	local updateInfo = {
		integrations = {
			helpdesk = helpDeskInfo,
			allegro  = allegroInfo
		}
	}
	return true, updateInfo
end

--[[
	Saves an avatar for the user
]]--
function SaveAvatar (userAtVCC, avatar)
	local keyspace = m_datasource.cassandra:CassandraConnection("Agent")
	keyspace:InsertColumnWithSize(userAtVCC, "Users", "Avatar", avatar.image, avatar.size)
	keyspace:Close()
end

--[[
	Saves user status message
]]--
function SetStatusMessage (userAtVCC, message)
	local keyspace = m_datasource.cassandra:CassandraConnection("Agent")
	keyspace:InsertColumn(userAtVCC, "Users", "StatusMessage", message)
	keyspace:Close()
	return true
end

---
-- Returns an array with the attention levels a user is in.
-- @return an array of attention level ids.
--
function GetUserAttentionLevels (userAtVcc)
	assert(type(userAtVcc) == "string")
	local levels = m_datasource.redis:smembers("AttentionLevels::ByUser:"..userAtVcc)
	return levels or {}
end

function SetMfaCode(userAtVcc, code, expirationSeconds)
	local key = "UserMfaCode:" .. userAtVcc
	m_datasource.redis:set(key, code)
	m_datasource.redis:expire(key, expirationSeconds)
end

function GetMfaCode(userAtVcc)
	return m_datasource.redis:get("UserMfaCode:" .. userAtVcc)
end

function RemoveMfaCode(userAtVcc)
	m_datasource.redis:del("UserMfaCode:" .. userAtVcc)
end

function GetAgentSkills (vcc, user_id, keyspace)
	local release = false
	if not keyspace then
		release = true
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
	end
	user_id = utils.AppendVcc( user_id, vcc )
	local skills = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "AgentSkills",
		key = user_id
	}
	if release then
		keyspace:Close()
	end
	return skills;
end

--
-- BEGIN: INCO64 - AgentMonitorService
--
function GetUserCaptureStatisticsStatus(username, vcc)
	local agent_at_vcc = username .. "@" .. vcc
	local status_key = "AgentMonitorService::CaptureStatisticsStatus::" .. agent_at_vcc

	-- Obtengo los status WebRTC y Logging
	-- LogDebug("[GetUserCaptureStatisticsStatus] Getting WebRTC status (key is %s) for agent '%s@%s'.", status_key, username, vcc)
	local webrtc_status = m_datasource.redis:hget(status_key, "WebRTC") or "0"

	-- LogDebug("[GetUserCaptureStatisticsStatus] Getting Logging status (key is %s) for agent '%s@%s'.", status_key, username, vcc)
	local logging_status = m_datasource.redis:hget(status_key, "Logging") or "0"

	-- LogDebug("[GetUserCaptureStatisticsStatus] Getting BrowserHealth status (key is %s) for agent '%s@%s'.", status_key, username, vcc)
	local browser_health = m_datasource.redis:hget(status_key, "BrowserHealth") or "0"


	local agentStatus = {
		agent_at_vcc = agent_at_vcc,
		webrtc = webrtc_status == '1',
		logging = logging_status == '1',
		browser_health = browser_health == '1'
	}

	return agentStatus
end
--
-- END: INCO64 - AgentMonitorService
--

--[[------------------------------------------------------------------------------------------
	SUPER USERS
]]--

--[[
]]--
function GetVCCSuperUsers (vcc, ksAgent)
	assert(vcc)
	return utils.GetAllColumns {
		keyspace = ksAgent or "Agent",
		columnFamily = "VCCSuperUsers",
		key = vcc,
		columnCount = 50000
	} or {}
end

--[[
]]--
function GetSuperUser (vcc, simpleId, ksAgent)
	assert(vcc)
	assert(simpleId)
	local superUsers = GetVCCSuperUsers(vcc, ksAgent)
	return superUsers[simpleId]
end

--[[
]]--
function SaveSuperUser (vcc, user, ksAgent)
	assert(vcc)
	assert(user)
	local closeKeyspace = false
	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection("Agent")
		closeKeyspace = true
	end
	ksAgent:InsertColumn(vcc, "VCCSuperUsers", user.Id, cjson.encode({}))
	if closeKeyspace then
		ksAgent:Close()
		ksAgent = nil
	end
end
