/* jshint smarttabs:true, quotmark:double, esversion: 11 */
/* global $, inConcert, hermes */


if (!this.inConcert) {
	this.inConcert = {};
}

(function () {

	"use strict";

	function AgentAPIServer(mode = "agent") {
		this.mode = mode;
		this.incomingEventsCallbacks = {};
		this.outgoingEventsNotifications = {};
		this.incomingEventsSuscriptions = {};
		this.incomingEventsSuscriptionsByToken = {};
		this.incomingEventsInteractionIdsByToken = {};
		this.incomingTokensTimestamps = {};
		this.__initialize();
	}

	if (typeof inConcert.AgentAPIServer !== "function") {
		inConcert.AgentAPIServer = AgentAPIServer;
	}

/************************ Internal ************************/

	AgentAPIServer.prototype.__isEmptyObject = function(data) {
		if (!data) {
			return true;
		}
		for (var key in data) {
			if (data[key]) {
				return false;
			}
		}
		return true;
	};

	AgentAPIServer.prototype.__makeResult = function(result, data = {}) {
		data.result = result;
		return data;
	};

	AgentAPIServer.prototype.__makeOk = function(data) {
		return this.__makeResult("OK", data);
	};

	AgentAPIServer.prototype.__makeError = function(result, reason, data) {
		data = data || {};
		data.reason = reason || "unknown error";
		return this.__makeResult(result || "FAIL", data);
	};

	AgentAPIServer.prototype.__initialize = function() {
		this.__listenHermesEvents();
		if (this.mode !== "login") {
			this.__incomingTokensPooling();
		}
	};

	AgentAPIServer.prototype.__hermesIncomingEventName = function() {
		return "inconcert_agent_api_incoming";
	};

	AgentAPIServer.prototype.__hermesOutgoingEventName = function() {
		return "inconcert_agent_api_outgoing";
	};

	AgentAPIServer.prototype.__callbackPrefix = function() {
		return "callback_";
	};

	AgentAPIServer.prototype.__listenHermesEvents = function() {
		const self = this;

		hermes.on(self.__hermesOutgoingEventName(), function(eventData) {
			if (!$.isArray(eventData) || eventData.length < 2) {
				return;
			}
			var token = eventData[0];
			var eventName = eventData[1];
			var data = eventData.length > 2? eventData[2] : null;
			var callbackId = eventData.length > 3? eventData[3] : null;
			self.__setIncomingToken(token);
			self.__outgoingEvent(token, eventName, data, callbackId);
		});

		hermes.on(self.__callbackPrefix() + self.__hermesIncomingEventName(), function(eventData) {
			if (!$.isArray(eventData) || eventData.length < 4) {
				return;
			}
			var token = eventData[0];
			var eventName = eventData[1];
			var data = eventData[2];
			var callbackId = eventData[3];
			self.__setIncomingToken(token);
			self.__fireIncomingMethodCallback(token, eventName, data, callbackId);
		});
	};

	AgentAPIServer.prototype.__sendHermesMessage = function(topic, token, event, data, callbackId) {
		try {
			hermes.send(topic, [token, event, JSON.parse(JSON.stringify(data)), callbackId], true);
			return true;
		}
		catch(e) {
			console.error("Error sending message to hermes", event, e);
			return false;
		}
	};

	AgentAPIServer.prototype.__outgoingEvent = function(token, event, data, callbackId) {
		const self = this;
		//event custom notifications
		self.__fireOutgoingEventNotification(event, data);
		//method internal notification
		if (self["__" + event]) {
			var callback = callbackId? function(result) {
				self.__outgoingEventCallback(token, event, callbackId, result);
			} : null;
			self["__" + event](data, callback, token);
		}
	};

	AgentAPIServer.prototype.__outgoingEventCallback = function(token, event, callbackId, result) {
		return this.__sendHermesMessage(this.__callbackPrefix() + this.__hermesOutgoingEventName(), 
			token, event, result, callbackId);
	};

	AgentAPIServer.prototype.__getIncomingMethodCallbackDefaultTimeout = function() {
		return 30;
	};

	AgentAPIServer.prototype.__callIncomingMethod = function(event, data, tokens, callback, timeout) {
		tokens = tokens || null;
		var callbackId = callback? inConcert.Utils.generateGuid() : null;
		var status = this.__sendHermesMessage(this.__hermesIncomingEventName(), tokens, event, data, callbackId);
		if (callback) {
			if (status) {
				this.__addIncomingMethodCallback(event, callbackId, callback, timeout);
			}
			else {
				callback(this.__makeError("FAIL", "error sending message"));
			}
		}
		return status;
	};

	AgentAPIServer.prototype.__addIncomingMethodCallback = function(event, callbackId, callback, timeout) {
		if (!event || !callbackId || !callback) {
			return;
		}
		this.incomingEventsCallbacks[event] = this.incomingEventsCallbacks[event] || {};
		this.incomingEventsCallbacks[event][callbackId] = {
			callback : callback,
			timeout : this.__makeIncomingMethodCallbackTimeout(event, callbackId, timeout)
		};
	};

	AgentAPIServer.prototype.__makeIncomingMethodCallbackTimeout = function(event, callbackId, timeout) {
		const self = this;
		if (timeout === 0) {
			return null;
		}
		timeout = timeout || self.__getIncomingMethodCallbackDefaultTimeout();
		return window.setTimeout(function() {
			var result = self.__makeError("TIMEOUT", "timeout waiting for event result");
			self.__fireIncomingMethodCallback("", event, result, callbackId);
		}, timeout * 1000);
	};

	AgentAPIServer.prototype.__getIncomingMethodCallback = function(event, callbackId) {
		return event && callbackId && this.incomingEventsCallbacks[event] && 
			this.incomingEventsCallbacks[event][callbackId] || null;
	};

	AgentAPIServer.prototype.__removeIncomingMethodCallback = function(event, callbackId) {
		var callback = this.__getIncomingMethodCallback(event, callbackId);
		if (!callback) {
			return;
		}
		if (callback.timeout) {
			window.clearTimeout(callback.timeout);
		}
		delete this.incomingEventsCallbacks[event][callbackId];
		if (this.__isEmptyObject(this.incomingEventsCallbacks[event])) {
			delete this.incomingEventsCallbacks[event];
		}
	};

	AgentAPIServer.prototype.__fireIncomingMethodCallback = function(token, event, result, callbackId) {
		var callback = this.__getIncomingMethodCallback(event, callbackId);
		if (!callback) {
			return;
		}
		this.__removeIncomingMethodCallback(event, callbackId);
		callback.callback(result);
	};

	AgentAPIServer.prototype.__fireOutgoingEventNotification = function(event, data) {
		var notifications = this.outgoingEventsNotifications[event];
		if (!notifications) {
			return;
		}
		for (var i = 0; i < notifications.length; i++) {
			notifications[i](data);
		}
	};
	
	AgentAPIServer.prototype.__fireIncomingNotificationFirstTime = function(event, token) {
		if (this.__isTokenSuscribedToEvent(event, token) && this["__" + event + "__FirstTime"]) {
			this["__" + event + "__FirstTime"](token);
		}
	};

	AgentAPIServer.prototype.__isTokenSuscribedToEvent = function(event, token) {
		return this.incomingEventsSuscriptions[event] && this.incomingEventsSuscriptions[event][token] || false;
	};

	AgentAPIServer.prototype.__RefreshEventsSuscription = function(request, _, token) {
		const self = this;
		self.__unsuscribeTokenEvents(token);
		request.events.forEach(function(event) {
			self.__suscribeTokenEvent(event, token);
		});
		self.__setTokenIncomingInteractionId(request.interactionId, token);
	};
	
	AgentAPIServer.prototype.__SuscribeEvent = function(request, _, token) {
		this.__suscribeTokenEvent(request.event, token);
		this.__fireIncomingNotificationFirstTime(request.event, token);
	};

	AgentAPIServer.prototype.__suscribeTokenEvent = function(event, token) {
		if (!event) {
			return;
		}
		this.incomingEventsSuscriptions[event] = this.incomingEventsSuscriptions[event] || {};
		this.incomingEventsSuscriptions[event][token] = true;
		this.incomingEventsSuscriptionsByToken[token] = this.incomingEventsSuscriptionsByToken[token] || {};
		this.incomingEventsSuscriptionsByToken[token][event] = true;
	};

	AgentAPIServer.prototype.__UnSuscribeEvent = function(request, _, token) {
		this.__unsuscribeTokenEvent(request.event, token);
	};

	AgentAPIServer.prototype.__unsuscribeTokenEvent = function(event, token) {
		if (!event) {
			return;
		}
		if (this.incomingEventsSuscriptions[event]) {
			delete this.incomingEventsSuscriptions[event][token];
			if (this.__isEmptyObject(this.incomingEventsSuscriptions[event])) {
				delete this.incomingEventsSuscriptions[event];
			}
		}
		if (this.incomingEventsSuscriptionsByToken[token]) {
			delete this.incomingEventsSuscriptionsByToken[token][event];
			if (this.__isEmptyObject(this.incomingEventsSuscriptionsByToken[token])) {
				delete this.incomingEventsSuscriptionsByToken[token];
			}
		}
	};

	AgentAPIServer.prototype.__unsuscribeTokenEvents = function(token) {
		const self = this;
		Object.keys(self.incomingEventsSuscriptionsByToken[token] || {}).forEach(function(event) {
			self.__unsuscribeTokenEvent(event, token);
		});
	};

	AgentAPIServer.prototype.__SetInteractionId = function(request, _, token) {
		this.__setTokenIncomingInteractionId(request.id, token);
	};

	AgentAPIServer.prototype.__setTokenIncomingInteractionId = function(interactionId, token) {
		this.__removeTokenIncomingInteractionId(token);
		if (interactionId) {
			this.incomingEventsInteractionIdsByToken[token] = interactionId;
		}
	};

	AgentAPIServer.prototype.__removeTokenIncomingInteractionId = function(token) {
		if (this.incomingEventsInteractionIdsByToken[token]) {
			delete this.incomingEventsInteractionIdsByToken[token];
		}
	};

	AgentAPIServer.prototype.__getTokenInteractionId = function(token) {
		return this.incomingEventsInteractionIdsByToken[token];
	};

	AgentAPIServer.prototype.__getInteractionIdTokens = function(interactionId) {
		const self = this;
		return Object.keys(self.incomingEventsInteractionIdsByToken).reduce(function(result, token) {
			if (self.incomingEventsInteractionIdsByToken[token] === interactionId) {
				result[token] = true;
			}
			return result;
		}, {});
	};

	AgentAPIServer.prototype.__hasTokenInteractionId = function(interactionId, token, allowsEmpty = true) {
		let tokenInteractionId = this.__getTokenInteractionId(token);
		return tokenInteractionId? tokenInteractionId === interactionId : allowsEmpty;
	};

	AgentAPIServer.prototype.__hasEventSuscription = function(event, condition) {
		if (this.__isEmptyObject(this.incomingEventsSuscriptions[event])) {
			return false;
		}
		if (condition) {
			return !!Object.keys(this.incomingEventsSuscriptions[event]).find(function(token) {
				return condition(token);
			});
		}
		else {
			return true;
		}
	};

	AgentAPIServer.prototype.__fireIncomingNotification = function(event, data, filter) {
		if (inConcert.logger) {
			inConcert.logger.debug({
				action_name : "AgentAPIServer.prototype.__fireIncomingNotification",
				message     : event
			});
		}
		let tokens = null;
		if (filter !== true) {
			tokens = this.incomingEventsSuscriptions[event];
			if (filter) {
				tokens = inConcert.Utils.filterMap(tokens || {}, function(_, token) {
					return filter(token);
				});
			}
			if (this.__isEmptyObject(tokens)) {
				return;
			}
		}
		this.__callIncomingMethod(event, data, tokens);
	};

	AgentAPIServer.prototype.__hasInteractionEventSuscription = function(event, interactionId) {
		const self = this;
		let filter = interactionId?
			function(token) {
				return self.__hasTokenInteractionId(interactionId, token);
			} : null;
		return self.__hasEventSuscription(event, filter);
	};

	AgentAPIServer.prototype.__fireIncomingInteractionNotification = function(event, data) {
		const self = this;
		let filter = (data.interaction && data.interaction.id)?
			function(token) {
				return self.__hasTokenInteractionId(data.interaction.id, token);
			} : null;
		self.__fireIncomingNotification(event, data, filter);
	};

	AgentAPIServer.prototype.__fireIncomingNotificationToToken = function(event, data, token) {
		if (!this.__isTokenSuscribedToEvent(event, token)) {
			return;
		}
		this.__callIncomingMethod(event, data, token);
	};

	AgentAPIServer.prototype.__setIncomingToken = function(token) {
		if (token) {
			this.incomingTokensTimestamps[token] = (new Date()).getTime();
		}
	};
	
	AgentAPIServer.prototype.__getIncomingTokensTimeout = function() {
		return 30;
	};

	AgentAPIServer.prototype.__validateIncomingTokens = function() {
		const self = this;
		let now = (new Date()).getTime();
		let timeoutMs = self.__getIncomingTokensTimeout() * 1000;
		Object.keys(self.incomingTokensTimestamps).forEach(function(token) {
			if (self.incomingTokensTimestamps[token] + timeoutMs < now) {
				self.__unsuscribeTokenEvents(token);
				self.__removeTokenIncomingInteractionId(token);
				delete self.incomingTokensTimestamps[token];
			}
		});
	};

	AgentAPIServer.prototype.__incomingTokensPooling = function() {
		const self = this;
		window.setInterval(function() {
			self.__validateIncomingTokens();
		}, 3000);
	};

	AgentAPIServer.prototype.__getRecordUrl = function(interaction, part) {
		if (inConcert.app.has_allegro_integration()) {
			return;
		}
		return inConcert.app.getBaseDomain() + "/inconcert/apps/repository/download/?id=" + interaction.id + "&section=" + part;
	};

	/*
		https://testubu.i6.inconcertcc.com/inconcert/apps/allegro/repository/download/?id=0853693D112938A4738D3FF28B44FEF9&section=1&part=1
	*/
	AgentAPIServer.prototype.__getAllegroRecordUrl = function(interaction, section, part) {
		if (!inConcert.app.has_allegro_integration()) {
			return;
		}
		return `${inConcert.app.getBaseDomain()}/inconcert/apps/allegro/repository/download/?id=${interaction.id}&section=${section}&part=${part}`;
	};

/************************ Outgoing events ************************/

	AgentAPIServer.prototype.On = function(event, callback) {
		if (!this[event]) {
			return;
		}
		this.outgoingEventsNotifications[event] = this.outgoingEventsNotifications[event] || [];
		this.outgoingEventsNotifications[event].push(callback);
	};

	AgentAPIServer.prototype.Off = function(event, callback) {
		if (!this.outgoingEventsNotifications[event]) {
			return;
		}
		if (callback) {
			var current = this.outgoingEventsNotifications[event];
			this.outgoingEventsNotifications[event] = [];
			for (var i = 0; i < current.length; i++) {
				if (current[i] !== callback) {
					this.outgoingEventsNotifications[event].push(current[i]);
				}
			}
			if (!this.outgoingEventsNotifications[event].length) {
				delete this.outgoingEventsNotifications[event];
			}
		}
		else {
			delete this.outgoingEventsNotifications[event];
		}
	};

/************************ Aux transformations ************************/
	
	AgentAPIServer.prototype.__makeUser = function() {
		let userId = inConcert.app.getLoggedUserId();
		return userId? {
			Id : userId,
			FirstName : inConcert.app.getLoggedUserFirstName(),
			LastName : inConcert.app.getLoggedUserLastName(),
			FirstSurname : inConcert.app.getLoggedUserFirstSurname(),
			LastSurname : inConcert.app.getLoggedUserLastSurname(),
			CompleteName : inConcert.app.getLoggedUserCompleteName(),
			Avatar : inConcert.app.getLoggedUserAvatar(),
			Group : inConcert.app.getUserGroup(),
			Company : inConcert.app.getLoggedUserCompanyInformation(),
			CustomInfo : inConcert.app.getLoggedUserCustomInformation(),
			State : inConcert.app.getCurrentStatus(),
			StateDescription : inConcert.app.getLoggedUserStatus(),
			CountryCode : inConcert.app.getCountryCode(),
			AreaCode : inConcert.app.getAreaCode(),
			Timezone : inConcert.app.getTimezone(),
			Vcc : inConcert.app.get_vcc_id(),
			SessionData : inConcert.app.getSessionData(),
			Email: inConcert.app.getLoggedUserEmail()
		} : null;
	};

	AgentAPIServer.prototype.__makeMember = function(member) {
		return {
			Id : member.Id,
			FirstName : member.Name || "",
			LastName : member.LastName || "",
			FirstSurname : member.FirstSurname || "",
			LastSurname : member.LastSurname || "",
			CompleteName : member.get_full_name(),
			Avatar : member.getAvatarURI(),
			AttentionLevels : member.getAttentionLevels(true)
		};
	};

	AgentAPIServer.prototype.__makeState = function(state) {
		return {
			Id : state.Id,
			Description : state.Description,
			Break : state.isBreak === "1",
			Custom : state.vcc !== "system"
		};
	};

	AgentAPIServer.prototype.__makeInteraction = function(interaction) {
		if (!interaction) {
			return null;
		}
		let raw = interaction.toRaw();
		if (!interaction.isInflated()) {
			raw.steps = null;
			raw.events = null;
		}
		return raw;
	};

	AgentAPIServer.prototype.__makeAttentionLevel = function(accAttLevel, campaignAttLevel) {
		return {
			Id : accAttLevel.attention_level_id,
			Name : accAttLevel.name || campaignAttLevel.name,
			Distribution: accAttLevel.distribution,
			PriorityDate: accAttLevel.priority_date,
			Priority: accAttLevel.priority,
		};
	};

	AgentAPIServer.prototype.__makeAccount = function(campaign, account) {
		const self = this;
		let raw = {
			Id : account.id,
			Name : account.name,
			Type : account.proxyType,
			Language : account.language,
			Enabled : account.state && account.state !== "0" || false,
			Status : account.status && account.status.status || ""
		};
		let als = account.AttentionLevels || {};
		const campaignAls = campaign.AttentionLevels || {};
		raw.AttentionLevels = Object.keys(als)
			.map(function(attLevelId) {
				return self.__makeAttentionLevel(als[attLevelId], campaignAls[attLevelId]);
			})
			.sort(function(a1, a2) {
				if (a1.Name && a2.Name) { 
					return a1.Name.localeCompare(a2.Name);
				}
				return -1;
			});
		return raw;
	};

	AgentAPIServer.prototype.__makeCampaign = function(campaign, includeAccounts) {
		const self = this;
		return {
			Id : inConcert.Utils.getIdWithoutVCC(campaign.Id),
			Name : campaign.Name,
			Description : campaign.Description,
			Timezone : campaign.ValidationTimeZone,
			Accounts : includeAccounts? 
				campaign.accounts
					.filter(function(account) {
						return includeAccounts === true || account.proxyType === includeAccounts;
					})
					.map(function(account) {
						return self.__makeAccount(campaign, account);
					})
				: null,
			AttentionLevels : Object.keys(campaign.AttentionLevels || {}).map(function(al_id) {
				//Se cambia el nombre del campo para que quede en semejanza con los AL de las cuentas
				const al = campaign.AttentionLevels[al_id];
				return {
					Id : al.id,
					Name : al.name,
					Description: al.description
				};
			}).sort(function(a1, a2) {
				if (a1.Name && a2.Name) { 
					return a1.Name.localeCompare(a2.Name);
				}
				return -1;
			})
		};
	};

	AgentAPIServer.prototype.__makeLabel = function(label) {
		return {
			Id : label.id || label.Id,
			Name : label.name,
			System : label.system && label.system !== "0"
		};
	};

	AgentAPIServer.prototype.__appendVcc = function(value) {
		if (!value) return value;

		return inConcert.Utils.getIdWithoutVCC(value) + "@" + inConcert.app.get_vcc_id();
	};

	AgentAPIServer.prototype.__appendProcessId = function(batchId, processId) {
		if (!batchId || !processId || batchId.indexOf(":") > 0) return batchId;

		return `${batchId}:${this.__appendVcc(processId)}`;
	};

	AgentAPIServer.prototype.__appendVccToContactId = function(request) {
		if (!request.contactId || request.isExternal) return;

		request.contactId = this.__appendVcc(request.contactId);
	};

	AgentAPIServer.prototype.__processFilterValue = function(value, properties, processedValues) {
		if (typeof(value) === "string") {
			if (properties.withVcc) {
				value = this.__appendVcc(value);
			}
			if (properties.caseInsensitive) {
				value = value.toUpperCase();
			}
		}
		if (processedValues && properties.id === "account") {
			value = this.__processAccountFilterValue(value, processedValues.campaign, properties);
		}
		return value;
	};

	AgentAPIServer.prototype.__processAccountFilterValue = function(accountValue, campaignsIds, properties) {
		accountValue = this.__getAccountId(accountValue, campaignsIds) || accountValue;
		if (properties.caseInsensitive) {
			accountValue = accountValue.toUpperCase();
		}
		return accountValue;
	};

	AgentAPIServer.prototype.__makeFilterValues = function(request, properties, processedValues) {
		const self = this;
		if (properties.fixedValue) {
			return properties.fixedValue;
		}
		let values = properties.getFilter? properties.getFilter(request) : request[properties.id];
		if (!values) {
			return properties.defaultValue || values;
		}
		values = $.isArray(values)? values : [values];
		properties = properties || {};
		values = values.map(function(value) {
			return self.__processFilterValue(value, properties, processedValues);
		});
		return properties.map? inConcert.Utils.arrayToBoolMap(values) : values;
	};

	AgentAPIServer.prototype.__makeFilters = function(request, filtersProperties) {
		const self = this;
		let values_d = {};
		return (filtersProperties || []).map(function(filterProperties) {
			filterProperties = $.isPlainObject(filterProperties)? filterProperties : { id : filterProperties };
			let values = self.__makeFilterValues(request, filterProperties, values_d);
			values_d[filterProperties.id] = values;
			return {
				values : values,
				properties : filterProperties
			};
		});
	};

	AgentAPIServer.prototype.__getFilterElementValue = function(element, properties) {
		let value = properties.getValue? properties.getValue(element) :
				element[properties.elementId || properties.id];
		return this.__processFilterValue(value, properties);
	};

	AgentAPIServer.prototype.__processFilter = function(element, filter) {
		if (!filter.values) {
			return true;
		}
		let value = this.__getFilterElementValue(element, filter.properties);
		return !!($.isArray(filter.values)?
			filter.values.find(function(filterValue) {
				return value === filterValue;
			}) : filter.values[value]);
	};

	AgentAPIServer.prototype.__processFilters = function(element, filters, isOr) {
		const self = this;
		if (!element) {
			return false;
		}
		let filter = filters.find(function(filter) {
			let result = self.__processFilter(element, filter);
			return isOr? result : !result;
		});
		return isOr? !!filter : !filter;
	};

	AgentAPIServer.prototype.__processInteractionInflate = function(interaction, inflate, callback) {
		if (inflate && !interaction.isBeingComposed()) {
			interaction.inflate(function() {
				callback();
			});
		}
		else {
			callback();
		}
	};

	AgentAPIServer.prototype.__filterInteractions = function(interactions, request, filtersProperties, callback) {
		const self = this;
		let filters = self.__makeFilters(request, filtersProperties);
		inConcert.Utils.asyncReduce(interactions,
			function(result, interaction, nextStep) {
				if (self.__processFilters(interaction, filters, request.filtersOr)) {
					self.__processInteractionInflate(interaction, request.full, function() {
						result.push(self.__makeInteraction(interaction));
						nextStep(result);
					});
				}
				else {
					nextStep(result);
				}
			},
			function(interactions) {
				callback(self.__makeOk({ interactions : interactions }));
			}, 
			[]);
	};

	AgentAPIServer.prototype.__processServerInteractions = function(interactions) {
		const self = this;
		return interactions.map(function(interaction) {
			return self.__makeInteraction(interaction);
		});
	};

	AgentAPIServer.prototype.__replyServerInteractions = function(interactions, total, last, callback) {
		total = total || 0;
		last = last || 0;
		callback(this.__makeOk({
			interactions : this.__processServerInteractions(interactions),
			total : total,
			last : last,
			areMore : interactions.length > 0 && total > last
		}));
	};

	AgentAPIServer.prototype.__makeServerFilter = function(request, filtersProperties) {
		let filters_a = this.__makeFilters(request, filtersProperties);
		let filters = inConcert.Utils.arrayToMap(filters_a, 
			function(filter) {
				return filter.values;
			},
			function(filter) {
				return filter.properties.elementId || filter.properties.id;
			});
		//#14651: Acomodar filter al nuevo formato incorporado con el nuevo interface de busqueda
		let filter_name_mapping = {
			agent: "agent",
			state: "state",
			type: "type",
			campaign: "campaign",
			account: "account",
			address: "contact_address",
			contact: "contact",
			contactExternalId: "contact",
			contactName: "contact_name",
			text: "content",
			ticket: "ticket",
			disposition: "disposition_code",
			attention_level: "attention_level"
		};
		filters = filters || {};
		let query = {};
		query.condition = "AND";
		query.rules = [];
		let start_date_rule;
		for (const filter in filters) {
			if (Object.hasOwn(filters, filter) && filters[filter] && filter_name_mapping[filter]) {
				let rule = {};
				if(filter === "date_from") {
					start_date_rule = start_date_rule || {};
					start_date_rule.id = "start_date";
					if(start_date_rule.value) {
						start_date_rule.value_to = start_date_rule.value;
						start_date_rule.value = Array.isArray(filters[filter]) ? filters[filter][0] : filters[filter];
						start_date_rule.operator = "between";
					}
					else {
						start_date_rule.value = Array.isArray(filters[filter]) ? filters[filter][0] : filters[filter];
						start_date_rule.operator = "greater_or_equal";
					}
				}
				else if(filter === "date_to") {
					start_date_rule = start_date_rule || {};
					start_date_rule.id = "start_date";
					if(start_date_rule.value) {
						start_date_rule.value_to = Array.isArray(filters[filter]) ? filters[filter][0] : filters[filter];
						start_date_rule.operator = "between";
					}
					else {
						start_date_rule.value = Array.isArray(filters[filter]) ? filters[filter][0] : filters[filter];
						start_date_rule.operator = "less_or_equal";
					}
				}
				else if(filter === "type") {
					rule.id = filter_name_mapping[filter];
					rule.value = filters[filter];
					rule.operator = Array.isArray(filters[filter]) ? "in" : "equal";
					query.rules.push(rule);
				}
				else {
					rule.id = filter_name_mapping[filter];
					rule.value = filter === "state" ? filters[filter] : (Array.isArray(filters[filter]) ? filters[filter][0] : filters[filter]);
					rule.operator = filter === "state" ? "in" : "equal";
					query.rules.push(rule);
				}
			}
		}
		if(start_date_rule) {
			query.rules.push(start_date_rule);
		}	
		return {
			query : query,
			sort: request.sort || [{ start_timestamp: { order: "desc" } }],
			inflate : !!request.full,
			exact : !!request.exact,
			rows : request.rows || 20,
			last : request.last || "",
			paginating : true,
			output : "interactions"
		};
	};

	AgentAPIServer.prototype.__getAssignedInteractionObject = function(interactionId, callback, notInflate) {
		let interaction = inConcert.app.getInteractionManager().getAssignedInteraction(interactionId);
		if (interaction) {
			this.__processInteractionInflate(interaction, !notInflate, function() {
				callback(interaction);
			});
		}
		else {
			callback(null);
		}
	};

	AgentAPIServer.prototype.__getInteractionObjectFromServer = function(interactionId, callback, notInflate, notSearchArchive) {
		inConcert.app.getInteractionManager().getInteraction(interactionId, function(interaction) {
			callback(interaction || null);
		}, !notInflate, !notSearchArchive);
	};

	AgentAPIServer.prototype.__getInteractionObject = function(interactionId, callback, notInflate, notSearchArchive) {
		const self = this;
		self.__getAssignedInteractionObject(interactionId, function(interaction) {
			if (interaction) {
				callback(interaction);
			}
			else {
				self.__getInteractionObjectFromServer(interactionId, callback, notInflate, notSearchArchive);
			}
		}, notInflate);
	};

	AgentAPIServer.prototype.__openInteractionClient = function(interaction, callback) {
		inConcert.app.getInteractionManager().openInteractionClient(interaction);
		callback();
	};

	AgentAPIServer.prototype.__getAndOpenInteractionClient = function(interactionId, callback, notInflate, notSearchArchive) {
		const self = this;
		self.__getInteractionObject(interactionId, function(interaction) {
			if (interaction) {
				self.__openInteractionClient(interaction, function() {
					callback(interaction);
				});
			}
			else {
				callback();
			}
		}, notInflate, notSearchArchive);
	};

	AgentAPIServer.prototype.__getCampaign = function(campaignId) {
		return inConcert.app.getCampaign(this.__appendVcc(campaignId));
	};

	AgentAPIServer.prototype.__getCampaignAccount = function(campaign, accountId) {
		if (typeof(campaign) === "string") {
			campaign = this.__getCampaign(campaign);
		}
		return campaign && campaign.getAccount(accountId || "", null, true) || null;
	};

	AgentAPIServer.prototype.__getCampaignAttentionLevel = function(campaign, alId) {
		if (typeof(campaign) === "string") {
			campaign = this.__getCampaign(campaign);
		}
		return campaign && campaign.getAttentionLevel(alId, true) || null;
	};

	AgentAPIServer.prototype.__getAccountId = function(accountValue, campaigns) {
		const self = this;
		if (campaigns && !$.isArray(campaigns)) {
			if ($.isPlainObject(campaigns)) {
				campaigns = Object.keys(campaigns);
			}
			else {
				campaigns = [campaigns];
			}
		}
		campaigns = (campaigns && campaigns.length)? campaigns : inConcert.app.getCampaigns();
		let accountId = null;
		$.each(campaigns, function(_, campaign) {
			let account = self.__getCampaignAccount(campaign, accountValue);
			if (account) {
				accountId = account.id;
				return false;
			}
		});
		return accountId;
	};

	AgentAPIServer.prototype.__getAttentionLevelId = function(alValue, campaigns) {
		const self = this;
		if (campaigns && !$.isArray(campaigns)) {
			if ($.isPlainObject(campaigns)) {
				campaigns = Object.keys(campaigns);
			}
			else {
				campaigns = [campaigns];
			}
		}
		campaigns = (campaigns && campaigns.length)? campaigns : inConcert.app.getCampaigns();
		let alId = null;
		$.each(campaigns, function(_, campaign) {
			let al = self.__getCampaignAttentionLevel(campaign, alValue);
			if (al) {
				alId = al.id;
				return false;
			}
		});
		return alId;
	};

	AgentAPIServer.prototype.__getDispositions = function(campaignId) {
		let campaign = this.__getCampaign(campaignId);
		return campaign && campaign.getDispositionCodes() || [];
	};

	AgentAPIServer.prototype.__hasTicket = function(campaignId, accountId) {
		let campaign = this.__getCampaign(campaignId);
		return campaign && campaign.hasTicketConfiguredInAccount(accountId) || false;
	};

	AgentAPIServer.prototype.__getTemplateMessageOptions = function(campaignId, accountId, page, callback) {
		let campaign = this.__getCampaign(campaignId);
		if (campaign) {
			campaign.loadTemplateMessageOptions(accountId, page, true, function(options) {
				callback(options || {});
			});
		}
		else {
			callback({});
		}
	};

	AgentAPIServer.prototype.__buildAddress = function(addressType, addressRaw, callback) {
		inConcert.app.getContactManager().buildAddress(addressType, addressRaw, callback);
	};

	AgentAPIServer.prototype.__getContact = function(contactId, callback) {
		inConcert.app.getContactManager().getContact(contactId, callback);
	};

	AgentAPIServer.prototype.__getContactByAddress = function(address, callback) {
		inConcert.app.getContactManager().getContactByAddress(address, callback);
	};

	AgentAPIServer.prototype.__getContactByAddressRaw = function(addressType, addressRaw, callback) {
		const self = this;
		self.__buildAddress(addressType, addressRaw, function(address) {
			if (address) {
				self.__getContactByAddress(address, callback);
			}
			else {
				callback();
			}
		});
	};

	AgentAPIServer.prototype.__getContactByExternalId = function(externalId, callback) {
		inConcert.app.getContactManager().getContactByExternalId(externalId, callback);
	};

	AgentAPIServer.prototype.__buildMessage = function(proxyType, message) {
		if ($.isPlainObject(message)) {
			return message;
		}
		switch(proxyType) {
			case "MAIL":
				return {
					msgtype : "mail",
					body : message || "",
					attachments : []
				};
			default:
				return { msgtype : "text", text : message || "" };
		}
	};

/************************ Incoming user events ************************/

	AgentAPIServer.prototype.UserLoggedIn = function() {
		this.__fireIncomingNotification("UserLoggedIn", { user : this.__makeUser() }, true);
	};

	AgentAPIServer.prototype.__UserLoggedIn__FirstTime = function(token) {
		if (!inConcert.app || !inConcert.app.getLoggedUserId()) {
			return;
		}
		this.__fireIncomingNotificationToToken("UserLoggedIn", { user : this.__makeUser() }, token);
	};

	AgentAPIServer.prototype.UserLoginFailed = function(result, reason) {
		this.__fireIncomingNotification("UserLoginFailed", this.__makeError(result, reason), true);
	};

	AgentAPIServer.prototype.UserLoggedOut = function() {
		this.__fireIncomingNotification("UserLoggedOut", { user : this.__makeUser() });
	};

	AgentAPIServer.prototype.UserStateChanged = function(state) {
		this.__fireIncomingNotification("UserStateChanged", { state : this.__makeState(state) });
	};

/************************ Incoming interaction events ************************/

	AgentAPIServer.prototype.SwitchInteractionId = function(oldInteractionId, newInteractionId) {
		const self = this;
		let tokens = self.__getInteractionIdTokens(oldInteractionId);
		if (!self.__isEmptyObject(tokens)) {
			Object.keys(tokens).forEach(function(token) {
				self.__setTokenIncomingInteractionId(newInteractionId, token);
			});
			self.__callIncomingMethod("SwitchInteractionId", { oldId : oldInteractionId, newId : newInteractionId }, tokens);
		}
	};

	AgentAPIServer.prototype.InteractionAssigned = function(interaction) {
		this.__fireIncomingInteractionNotification("InteractionAssigned", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.__InteractionAssigned__FirstTime = function(token) {
		let interactionId = this.__getTokenInteractionId(token);
		if (interactionId) {
			let interaction = inConcert.app.getInteractionManager().getAssignedInteraction(interactionId);
			if (interaction && !interaction.isBeingComposed()) {
				this.__fireIncomingNotificationToToken("InteractionAssigned", { interaction : this.__makeInteraction(interaction) }, token);
			}
		}
	};

	AgentAPIServer.prototype.InteractionRemovedFromAgent = function(interaction) {
		this.__fireIncomingInteractionNotification("InteractionRemovedFromAgent", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.InteractionRequeued = function(interaction) {
		this.__fireIncomingInteractionNotification("InteractionRequeued", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.InteractionTransferred = function(interaction) {
		this.__fireIncomingInteractionNotification("InteractionTransferred", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.InteractionArchived = function(interaction) {
		this.__fireIncomingInteractionNotification("InteractionArchived", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.InteractionCanceled = function(interaction) {
		this.__fireIncomingInteractionNotification("InteractionCanceled", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.InteractionInWrapup = function(interaction) {
		this.__fireIncomingInteractionNotification("InteractionInWrapup", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.NewInteractionMessages = function(interaction, sinceEvent, eventCount) {
		const self = this;
		if (!self.__hasInteractionEventSuscription("NewInteractionMessages", interaction.id)) {
			return;
		}

		function fireEvent() {
			if (!interaction.events || interaction.events.length < eventCount) {
				return;
			}
			let messages = [];
			for (let i = sinceEvent; i < interaction.events.length; i++) {
				messages.push(interaction.events[i]);
			}
			self.__fireIncomingInteractionNotification("NewInteractionMessages", {
				interaction : self.__makeInteraction(interaction), messages : messages });
		}
		
		if (interaction.events && interaction.events.length >= eventCount) {
			fireEvent();
		}
		else {
			self.__processInteractionInflate(interaction, true, function() {
				fireEvent();
			});
		}
	};

	AgentAPIServer.prototype.NewChatMessage = function(interaction, message) {
		this.__fireIncomingInteractionNotification("NewInteractionMessages", {
				interaction : this.__makeInteraction(interaction), messages : [message] });
	};

/************************ Phone events ************************/

	AgentAPIServer.prototype.CallTaken = function(interaction) {
		this.__fireIncomingInteractionNotification("CallTaken", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallOnRecording = function(interaction, recording) {
		this.__fireIncomingInteractionNotification("CallOnRecording", { interaction : this.__makeInteraction(interaction), recording : recording });
	};

	AgentAPIServer.prototype.CallOnHold = function(interaction, hold) {
		this.__fireIncomingInteractionNotification("CallOnHold", { interaction : this.__makeInteraction(interaction), hold : hold });
	};

	AgentAPIServer.prototype.CallOnMute = function(interaction, muted) {
		this.__fireIncomingInteractionNotification("CallOnMute", { interaction : this.__makeInteraction(interaction), muted : muted });
	};

	AgentAPIServer.prototype.CallOnVideoDisabled = function(interaction, disabled) {
		this.__fireIncomingInteractionNotification("CallOnVideoDisabled", { interaction : this.__makeInteraction(interaction), disabled : disabled });
	};

	AgentAPIServer.prototype.CallOnSwitch = function(interaction, switched) {
		this.__fireIncomingInteractionNotification("CallOnSwitch", { interaction : this.__makeInteraction(interaction), switched : switched });
	};

	let m_previewIds = {};
	AgentAPIServer.prototype.CallOnPreview = function(interaction) {
		//sometimes is duplicated
		if (!interaction || m_previewIds[interaction.id]) {
			return;
		}
		m_previewIds[interaction.id] = true;
		this.__fireIncomingInteractionNotification("CallOnPreview", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallRedialing = function(interaction) {
		this.__fireIncomingInteractionNotification("CallRedialing", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallRedialAccepted = function(interaction) {
		let originId = interaction.getRedialProcess();
		if (originId) {
			this.SwitchInteractionId(originId, interaction.id);
		}
		this.__fireIncomingInteractionNotification("CallRedialAccepted", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallRedialCanceled = function(interaction) {
		this.__fireIncomingInteractionNotification("CallRedialCanceled", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallAppointmentAccepted = function(interaction) {
		this.__fireIncomingInteractionNotification("CallAppointmentAccepted", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallAppointmentCanceled = function(interaction) {
		this.__fireIncomingInteractionNotification("CallAppointmentCanceled", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallTransferring = function(interaction) {
		this.__fireIncomingInteractionNotification("CallTransferring", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallTransferConnected = function(interaction) {
		this.__fireIncomingInteractionNotification("CallTransferConnected", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallTransferCanceled = function(interaction) {
		this.__fireIncomingInteractionNotification("CallTransferCanceled", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallOnConference = function(interaction) {
		this.__fireIncomingInteractionNotification("CallOnConference", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallConferenceConsultationTrying = function(interaction) {
		this.__fireIncomingInteractionNotification("CallConferenceConsultationTrying", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallConferenceConsultationConnected = function(interaction) {
		this.__fireIncomingInteractionNotification("CallConferenceConsultationConnected", { interaction : this.__makeInteraction(interaction) });
	};

	AgentAPIServer.prototype.CallConferenceConsultationCanceled = function(interaction) {
		this.__fireIncomingInteractionNotification("CallConferenceConsultationCanceled", { interaction : this.__makeInteraction(interaction) });
	};

/************************ User methods ************************/

	AgentAPIServer.prototype.__GetLoggedUser = function(request, callback) {
		callback(this.__makeOk({ user : this.__makeUser() }));
	};

	AgentAPIServer.prototype.__GetLoggedUserState = function(request, callback) {
		callback(this.__makeOk({ state : inConcert.app.getCurrentStatus() }));
	};

	AgentAPIServer.prototype.__ListUserStates = function(request, callback) {
		const self = this;
		let states = inConcert.app.getAvailableStates().map(function(state) {
			return self.__makeState(state);
		});
		callback(self.__makeOk({ states : states }));
	};

	AgentAPIServer.prototype.__ChangeUserState = function(request, callback) {
		const self = this;
		if (!request.state) {
			return callback(self.__makeError("FAIL", "must enter an user state"));
		}
		if (request.state === inConcert.app.getCurrentStatus()) {
			return callback(self.__makeOk());
		}
		inConcert.app.changeUserState(request.state, function(response) {
			if (response.success) {
				inConcert.app.getMainWindow().setUserStatus(request.state);
				callback(self.__makeOk());
			}
			else {
				callback(self.__makeError(response.errType, response.reason || "error changing state"));
			}
		});
	};

	AgentAPIServer.prototype.__LogoutUser = function(request, callback) {
		const self = this;
		if (!request.force && !inConcert.app.canExit()) {
			return callback(self.__makeError("FAIL", "can not logout"));
		}
		inConcert.app.logout(function() {
			callback(self.__makeOk());
			window.location = "/inconcert/apps/agent/login/";
		}, "MainWindow-CloseButton");
	};

	AgentAPIServer.prototype.__FirePoller = function(request, callback) {
		inConcert.app.firePoller(request, function(result) {
			if (result.data && !request.returnData) {
				result.data = null;
			}
			callback(result);
		});
	};

	AgentAPIServer.prototype.__FireChatPoller = function(request, callback) {
		let client = inConcert.app.getInteractionManager().getChannelClient("CHAT");
		if (!client) {
			return callback(this.__makeError("NO_CHANNEL", "user has not chat channel"));
		}
		client.firePoller(request, function(result) {
			if (result.data && !request.returnData) {
				result.data = null;
				result.status = null;
			}
			callback(result);
		});
	};

	AgentAPIServer.prototype.__EnableAutomaticPoller = function(request, callback) {
		inConcert.app.enableAutomaticPoller();
		let chatClient = inConcert.app.getInteractionManager().getChannelClient("CHAT");
		if (chatClient) {
			chatClient.enableAutomaticPoller();
		}
		callback(this.__makeOk());
	};

	AgentAPIServer.prototype.__DisableAutomaticPoller = function(request, callback) {
		inConcert.app.disableAutomaticPoller();
		let chatClient = inConcert.app.getInteractionManager().getChannelClient("CHAT");
		if (chatClient) {
			chatClient.disableAutomaticPoller();
		}
		callback(this.__makeOk());
	};

	AgentAPIServer.prototype.__GetLoggedUserDefaultCampaign = function(request, callback) {
		let loggedUserInfo = inConcert.app.getLoggedUserInfo();
		let defaultCampaign = loggedUserInfo.DefaultCampaign; 
		defaultCampaign = defaultCampaign.substring(0, defaultCampaign.indexOf("@"));
		callback(this.__makeOk({ DefaultCampaign : defaultCampaign }));
	};
	
/************************ Campaign methods ************************/

	AgentAPIServer.prototype.__GetCampaigns = function(request, callback) {
		const self = this;
		let campaigns = inConcert.app.getCampaigns();
		let rawCampaigns = campaigns
			.filter(function(campaign) {
				return !request.type || !!(campaign.accounts || []).find(function(account) {
					return account.proxyType === request.type;
				});
			})
			.map(function(campaign) {
				return self.__makeCampaign(campaign, request.type || true);
			});
		callback(this.__makeOk({ campaigns : rawCampaigns }));
	};

	AgentAPIServer.prototype.__GetCampaign = function(request, callback) {
		let campaign = this.__getCampaign(request.id);
		callback(campaign? this.__makeOk({ campaign : this.__makeCampaign(campaign, true) }) :
			this.__makeError("FAIL", "campaign not found"));
	};

	AgentAPIServer.prototype.__GetCampaignAccount = function(request, callback) {
		let campaign = this.__getCampaign(request.campaign);
		let account = campaign? this.__getCampaignAccount(campaign, request.account) : null;
		callback(account? this.__makeOk({ account : this.__makeAccount(campaign, account) }) :
			this.__makeError("FAIL", "account not found"));
	};

	AgentAPIServer.prototype.__GetCampaignDispositions = function(request, callback) {
		callback(this.__makeOk({ dispositions : this.__getDispositions(request.campaign) }));
	};

	AgentAPIServer.prototype.__GetCampaignLabels = function(request, callback) {
		const self = this;
		let campaign = self.__getCampaign(request.campaign);
		if (campaign) {
			let labels_d = campaign.Labels || {};
			let labels = Object.keys(labels_d)
				.sort()
				.map(function(labelId) {
					return self.__makeLabel(labels_d[labelId]);
				});
			callback(self.__makeOk({ labels : labels }));
		}
		else {
			callback(self.__makeError("FAIL", "campaign not found"));
		}
	};

	AgentAPIServer.prototype.__GetAccountTemplateMessageOptions = function(request, callback) {
		const self = this;
		if (request.account) {
			request.account = self.__getAccountId(request.account, request.campaign) || request.account;
		}
		self.__getTemplateMessageOptions(request.campaign, request.account, request.page, function(options_d) {
			let options = inConcert.Utils.hash_to_sorted_array(options_d, null, null, { without_id : true, attr : "id" });
			callback(self.__makeOk({ options : options }));
		});
	};

	AgentAPIServer.prototype.__GetCampaignUsers = function(request, callback) {
		inConcert.app.getInteractionManager().getCampaignUsers(
			this.__appendVcc(request.campaign),
			request.interactionPermissions,
			(users) => callback(this.__makeOk({
				users: users || []
			}))
		);
	};

/************************ Interaction methods ************************/
	
	let AGENT_FILTER = {
		id : "agent",
		withVcc : true,
		getFilter : function({ allAgents }) {
			return !allAgents ? inConcert.app.getLoggedUserId() : null;
		}
	};

	let ASSIGNED_INTERACTIONS_FILTERS = [
		{ id : "campaign", withVcc : true, caseInsensitive : true, map : true },
		{ id : "account", caseInsensitive : true, map : true },
		{ id : "type", caseInsensitive : true, map : true }
	];

	let ASSIGNED_INTERACTIONS_BY_ADDRESS_FILTERS = [
		{ id : "campaign", withVcc : true, caseInsensitive : true, map : true },
		{ id : "account", caseInsensitive : true, map : true },
		{ id : "type", caseInsensitive : true, map : true },
		{ id : "address", elementId : "contactAddress", caseInsensitive : true, map : true }
	];

	let ASSIGNED_INTERACTIONS_BY_CONTACT_FILTERS = [
		{ id : "campaign", withVcc : true, caseInsensitive : true, map : true },
		{ id : "account", caseInsensitive : true, map : true },
		{ id : "type", caseInsensitive : true, map : true },
		{ id : "contactId", elementId : "contact", caseInsensitive : true, map : true }
	];

	let SERVER_INTERACTIONS_FILTERS = [
		{ id : "type", caseInsensitive : true },
		{ id : "campaign", withVcc : true },
		{ id : "account" },
		{ id : "address" },
		{ id : "contactId", elementId : "contact" },
		{ id : "contactExternalId", elementId : "contactExternalId" },
		{ id : "contactName" },
		{ id : "text" },
		{ id : "ticket" },
		{ id : "dateFrom", elementId : "date_from" },
		{ id : "dateTo", elementId : "date_to" }
	];
	
	let WAITING_INTERACTIONS_FILTERS = [...SERVER_INTERACTIONS_FILTERS];

	let QUEUED_INTERACTIONS_FILTERS = [...SERVER_INTERACTIONS_FILTERS,
		{ id : "no_agent", fixedValue : true }
	];

	let ARCHIVE_INTERACTIONS_FILTERS = [...SERVER_INTERACTIONS_FILTERS, AGENT_FILTER,
		{ id : "disposition" }
	];

	let ABANDONED_INTERACTIONS_FILTERS = [...SERVER_INTERACTIONS_FILTERS,
		{ id : "no_agent", fixedValue : true },
		{ id : "disposition" }
	];

	let GLOBAL_SEARCH_INTERACTIONS_FILTERS = [...SERVER_INTERACTIONS_FILTERS,
		{ id : "disposition" },
		{ id : "agent", withVcc : true },
		{ id : "attention_level" },
		{ id : "state", caseInsensitive : true }
	];

	AgentAPIServer.prototype.__GetAssignedInteractions = function(request, callback) {
		let interactions = inConcert.app.getInteractionManager().getAssignedInteractions();
		return this.__filterInteractions(interactions, request, ASSIGNED_INTERACTIONS_FILTERS, callback);
	};

	AgentAPIServer.prototype.__GetAssignedInteractionByAddress = function(request, callback) {
		const self = this;
		request.full = request.full !== false;
		let interactions = inConcert.app.getInteractionManager().getAssignedInteractions();
		return self.__filterInteractions(interactions, request, ASSIGNED_INTERACTIONS_BY_ADDRESS_FILTERS, function(result) {
			callback(result.result === "OK"? 
				self.__makeOk({ interaction : result.interactions.length? result.interactions[0] : null }) : result);
		});
	};

	AgentAPIServer.prototype.__GetAssignedInteractionByContact = function(request, callback) {
		const self = this;
		if (!request.contactId) {
			return callback(self.__makeError("FAIL", "must enter a contact"));
		}
		request.full = request.full !== false;
		self.__appendVccToContactId(request);

		let search = function() {
			let interactions = inConcert.app.getInteractionManager().getAssignedInteractions();
			return self.__filterInteractions(interactions, request, ASSIGNED_INTERACTIONS_BY_CONTACT_FILTERS, function(result) {
				callback(result.result === "OK"? 
					self.__makeOk({ interaction : result.interactions.length? result.interactions[0] : null }) : result);
			});
		};

		if (request.isExternal) {
			self.__getContactByExternalId(request.contactId, function(contact) {
				if (contact) {
					request.contactId = contact.Id;
					search();
				}
				else {
					callback(self.__makeError("FAIL", "contact not found"));
				}
			});
		}
		else {
			search();
		}
	};

	AgentAPIServer.prototype.__GetAppointmentInteractions = function(request, callback) {
		let interactions = inConcert.app.getInteractionManager().getAppointmentInteractions();
		return this.__filterInteractions(interactions, request, ASSIGNED_INTERACTIONS_FILTERS, callback);
	};

	AgentAPIServer.prototype.__GetWaitingInteractions = function(request, callback) {
		const self = this;
		//self.__
		let filter = self.__makeServerFilter(request, WAITING_INTERACTIONS_FILTERS);
		let im = inConcert.app.getInteractionManager();
		filter.query.rules.push({id:"state", operator:"equal",value:["WAITING"]});
		im.searchWaitingInteractions(filter, function(status, result) {
			if (status) {
				self.__replyServerInteractions(result.interactions, result.total, result.last, callback);
			}
			else {
				callback(self.__makeError(result && result.result, result && result.reason || "error searching interactions"));
			}
		});
	};

	AgentAPIServer.prototype.__GetTopQueuedInteractions = function(request, callback) {
		const self = this;
		let im = inConcert.app.getInteractionManager();
		im.getTopQueued(function(interactions) {
			callback(self.__makeOk({ interactions : interactions || [] }));
		}, request.full);
	};

	AgentAPIServer.prototype.__GetQueuedInteractions = function(request, callback) {
		const self = this;
		let im = inConcert.app.getInteractionManager();
		if (request.fromMemory) {
				im.getQueuedInteractions(function(interactions) {
					self.__replyServerInteractions(interactions, interactions.length, interactions.length, callback);
				}, request.full);
		} else {
			request.fromMemory = undefined;
			let filter = self.__makeServerFilter(request, QUEUED_INTERACTIONS_FILTERS);	
			filter.query.rules.push({id:"state", operator:"in",value:["QUEUED","TRANSFER QUEUED"]});
			im.searchQueuedInteractions(filter, function(status, result) {
				if (status) {
					self.__replyServerInteractions(result.interactions, result.total, result.last, callback);
				}	
				else {
					callback(self.__makeError(result && result.result, result && result.reason || "error searching interactions"));
				}
			});
		}
	};

	AgentAPIServer.prototype.__GetArchiveInteractions = function(request, callback) {
		const self = this;
		let filter = self.__makeServerFilter(request, ARCHIVE_INTERACTIONS_FILTERS);
		filter.query.rules.push({id:"state", operator:"equal",value:["FINISHED"]});

		let im = inConcert.app.getInteractionManager();
		im.searchArchiveInteractions(filter, function(status, result) {
			if (status) {
				self.__replyServerInteractions(result.interactions, result.total, result.last, callback);
			}
			else {
				callback(self.__makeError(result && result.result, result && result.reason || "error searching interactions"));
			}
		});
	};

	AgentAPIServer.prototype.__GetAbandonedInteractions = function(request, callback) {
		const self = this;
		let filter = self.__makeServerFilter(request, ABANDONED_INTERACTIONS_FILTERS);
		filter.query.rules.push({id:"state", operator:"equal",value:["ABANDONED"]});
		let im = inConcert.app.getInteractionManager();
		im.searchAbandonedInteractions(filter, function(status, result) {
			if (status) {
				self.__replyServerInteractions(result.interactions, result.total, result.last, callback);
			}
			else {
				callback(self.__makeError(result && result.result, result && result.reason || "error searching interactions"));
			}
		});
	};

	AgentAPIServer.prototype.__GlobalSearchInteractions = function(request, callback) {
		const self = this;
		let filter = self.__makeServerFilter(request, GLOBAL_SEARCH_INTERACTIONS_FILTERS);
		let im = inConcert.app.getInteractionManager();
		im.globalSearch(filter, function(status, result) {
			if (status) {
				self.__replyServerInteractions(result.interactions, result.total, result.last, callback);
			}
			else {
				callback(self.__makeError(result && result.result, result && result.reason || "error searching interactions"));
			}
		});
	};

	AgentAPIServer.prototype.__GetInteraction = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction) {
				callback(self.__makeOk({ interaction : self.__makeInteraction(interaction) }));
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found"));
			}
		});
	};

	AgentAPIServer.prototype.__TakeNextQueuedInteraction = function(request, callback) {
		const self = this;
		let campaign = request.campaign? self.__appendVcc(request.campaign) : null;
		let type = request.type? request.type.toUpperCase() : null;
		inConcert.app.getInteractionManager().takeNextBest(campaign, type, function(status, interactionId) {
			if (status) {
				self.__GetInteraction({ id : interactionId }, callback);
			}
			else {
				callback(self.__makeError("FAIL", interactionId || "error taking interaction"));
			}
		});
	};

	AgentAPIServer.prototype.__TakeQueuedInteraction = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (!interaction) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			if (!interaction.canBeTaken() || ( !interaction.isQueued() && !interaction.isOpen())) {
				return callback(self.__makeError("FAIL", "interaction can not be taken"));
			}
			interaction.take(function(status, message) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", message || "error taking interaction"));
			});
		});
	};

	AgentAPIServer.prototype.__RequeueInteraction = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction) {
				return callback(self.__makeError("FAIL", "interaction not found or not taken"));
			}
			interaction.requeue(function(status, message) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", message || "error requeueing interaction"));
			});
		}, true);
	};

	AgentAPIServer.prototype.__HangupInteraction = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || !interaction.canHangup()) {
				return callback(self.__makeError("FAIL", "interaction not found or not taken"));
			}
			interaction.hangup(function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		}, true);
	};

	AgentAPIServer.prototype.__ArchiveInteraction = function(request, callback) {
		const self = this;
		if (request.disposition && (!request.disposition.name || !request.disposition.treePath)) {
			return callback(self.__makeError("FAIL", "invalid disposition object"));
		}
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction) {
				return callback(self.__makeError("FAIL", "interaction not found or not taken"));
			}
			let dispositions = self.__getDispositions(interaction.campaign);
			if (dispositions.length && !request.disposition) {
				return callback(self.__makeError("FAIL", "Must enter a disposition"));
			}
			
			var getDisposition = function (disposition, id) {
				if (disposition) {
					for (var i = 0; i < disposition.length; i++) {
						if (disposition[i].code === id || disposition[i].name === id) {
							return disposition[i];
						}
						var found = getDisposition(disposition[i].childs, id);
						if (found) return found;
					}
				}
			};
			var dispositionObj;
			if (request.disposition) {
				dispositionObj = getDisposition(dispositions, request.disposition.name);
				if(dispositionObj === undefined)
				{

					return callback(self.__makeError("FAIL", "Disposition does not exist"));
				}
				if(interaction.interactionData && interaction.interactionData.dialer && interaction.interactionData.dialer.contactId){
					request.disposition.dialer = interaction.interactionData.dialer;
					if (dispositionObj.contactCode) {
						request.disposition.dialer.contactCode = dispositionObj.contactCode;
					}
				}
			}
			if(dispositionObj && dispositionObj.needReschedule)
			{
				request.disposition.needReschedule = true;	//This make request go to dialerCallBack on allegro_api
				request.disposition.countryId = request.disposition.countryId || (interaction.completeAddress ? interaction.completeAddress.Country : "");
				request.disposition.areaId = request.disposition.areaId || (interaction.completeAddress ? interaction.completeAddress.Area : "");
				request.disposition.number = request.disposition.number || (interaction.completeAddress ? interaction.completeAddress.Number : interaction.contactAddress);
				request.disposition.zip = request.disposition.zip || (interaction.completeAddress ? interaction.completeAddress.zip : "-1");
				request.disposition.phoneType = interaction.interactionData.completeAddress.Kind || "HOME";
				//Si no disponemos de countryId ni de areaId 
				if(!request.disposition.countryId && !request.disposition.areaId)
				{
					//Si el number contiene guiones, tratamos de parsearlo
					if(request.disposition.number.indexOf("-") >= 0)
					{
						let splitedNumber = request.disposition.number.split("-");
						if(splitedNumber.length === 3)
						{
							request.disposition.countryId = splitedNumber[0];
							request.disposition.areaId = splitedNumber[1];
							request.disposition.number = splitedNumber[2];
						}
					}
					else {//Si no hay guiones en el number lo seteamos como numero internacional. Los " " en countryId y areaId son para que el RecognizePhone no falle
						request.disposition.countryId = " ";
						request.disposition.areaId = " ";
						request.disposition.number = "00" + request.disposition.number;
					}
				}
				//si se recive un alternativephone se debe validar y enviar para reagendar a ese tel en particular.
				if(request.disposition.alternativephone)
				{
					request.disposition.alternativephone.countryId = request.disposition.alternativephone.countryId || (interaction.completeAddress ? interaction.completeAddress.Country : "");
					request.disposition.alternativephone.areaId = request.disposition.alternativephone.areaId;
					request.disposition.alternativephone.number = request.disposition.alternativephone.number;
					request.disposition.alternativephone.phoneType = interaction.interactionData.completeAddress.Kind || "HOME";
					request.disposition.alternativephone.zip = request.disposition.alternativephone.zip || "-1";
					
					if(!request.disposition.alternativephone.countryId && !request.disposition.alternativephone.areaId)
					{
						//Si el number contiene guiones, tratamos de parsearlo
						if(request.disposition.alternativephone.number.indexOf("-") >= 0)
						{
							let splitedNumber = request.disposition.alternativephone.number.split("-");
							if(splitedNumber.length === 3)
							{
								request.disposition.alternativephone.countryId = splitedNumber[0];
								request.disposition.alternativephone.areaId = splitedNumber[1];
								request.disposition.alternativephone.number = splitedNumber[2];
							}
						}
						else {//Si no hay guiones en el number lo seteamos como numero internacional. Los " " en countryId y areaId son para que el RecognizePhone no falle
							request.disposition.alternativephone.countryId = " ";
							request.disposition.alternativephone.areaId = " ";
							request.disposition.alternativephone.number = "00" + request.disposition.alternativephone.number;
						}
					}
				}
				request.disposition.vcc = interaction.VCC;
				request.disposition.campaignId = interaction.campaign.substring(0, interaction.campaign.indexOf("@"));
				if (!request.disposition.importId){
					request.disposition.importId = "";
				}
				if (!request.disposition.batchId){ 
					request.disposition.batchId = "";
				}
				if(interaction.interactionData && interaction.interactionData.dialer && interaction.interactionData.dialer.contactId)
				{
					request.disposition.contactId = interaction.interactionData.dialer.contactId || "";
					request.disposition.importId = interaction.interactionData.dialer.importId || "";
					request.disposition.batchId = interaction.interactionData.dialer.batchId || "";
				}
				else if(interaction.interactionData && interaction.interactionData.gatheredContactData && interaction.interactionData.gatheredContactData.external_id)
				{
					request.disposition.contactId = interaction.interactionData.gatheredContactData.external_id;
				}
				else if(interaction.contactObject && interaction.contactObject.Id)
				{
					request.disposition.contactId = interaction.contactObject.Id;
				}
				else if(interaction.contact) {
					request.disposition.contactId = interaction.contact;
				}
				else if(request.disposition && request.disposition.contactId) {
					// esta condicion es para que no se pise el contactId.
				}
				else
				{
					request.disposition.contactId = "";	
				}
				if (!request.disposition.contactName ) {
					if(interaction.interactionData && interaction.interactionData.gatheredContactData && interaction.interactionData.gatheredContactData.name) {
						request.disposition.contactName = interaction.interactionData.gatheredContactData.name;
					}
					else {
						request.disposition.contactName = (interaction.contactObject && interaction.contactObject.Name) || (interaction.interactionData && interaction.interactionData.contactName) || "No Name";
					}
				}
				let loggedUser = self.__makeUser();
				request.disposition.timeZone = loggedUser.Timezone;
				if(request.disposition.sameAgent)
				{
					request.disposition.agentId = loggedUser.Id;
					delete request.disposition.sameAgent;
				}
				else
				{
					request.disposition.agentId = "";
				}
				if(!request.disposition.scheduleDate)
				{
					return callback(self.__makeError("FAIL", "Disposition requires a valid date"));
				}
				let scheduleDate = new Date(request.disposition.scheduleDate);
				if(!(scheduleDate instanceof Date && !isNaN(scheduleDate.valueOf())))
				{
					return callback(self.__makeError("FAIL", "Disposition requires a valid date"));
				}
				interaction.archive(request.disposition, request.comment, function(status, successful_nodes, unsuccessful_nodes, message) {
					callback(status? self.__makeOk() : self.__makeError("FAIL", message || "error archiving interaction"));
				});
			}
			else
			{
				interaction.archive(request.disposition, request.comment, function(status, message) {
					callback(status? self.__makeOk() : self.__makeError("FAIL", message || "error archiving interaction"));
				});
			}
		}, true);
	};

	AgentAPIServer.prototype.__ReopenInteraction = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (!interaction) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			if (!interaction.isArchived()) {
				return callback(self.__makeError("FAIL", "interaction is not archived"));
			}
			interaction.reopen(function(status, message) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", message || "error reopening interaction"));
			});
		}, true);
	};

	AgentAPIServer.prototype.__CancelInteraction = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || !interaction.isBeingComposed()) {
				return callback(self.__makeError("FAIL", "interaction not found or not been composed"));
			}
			inConcert.app.getInteractionManager().cancelInteraction(interaction);
			callback(self.__makeOk());
		}, true);
	};

	AgentAPIServer.prototype.__CommentInteraction = function(request, callback) {
		const self = this;
		if (!request.comment) {
			return callback(self.__makeError("FAIL", "comment not found"));
		}
		self.__getInteractionObject(request.id, function(interaction) {
			if (!interaction) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			if (interaction.isFinished()) {
				return callback(self.__makeError("FAIL", "interaction is finished"));
			}
			interaction.comment(request.comment, function(status, message) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", message || "error commenting interaction"));
			});
		}, true);
	};

	AgentAPIServer.prototype.__EmitInteractionTicket = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || interaction.isBeingComposed() || interaction.isInternal()) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			if (interaction.getAssignedTicket()) {
				return callback(self.__makeError("FAIL", "interaction already has a ticket"));
			}
			if (!self.__hasTicket(interaction.campaign, interaction.account)) {
				return callback(self.__makeError("FAIL", "interaction has not ticket emiter"));
			}
			interaction.emitTicket(function(status, ticket) {
				callback(status? self.__makeOk({ ticket : ticket }) : 
					self.__makeError("FAIL", "error emiting ticket"));
			});
		}, true);
	};

	AgentAPIServer.prototype.__SearchTicket = function(request, callback) {
		const self = this;
		let api = inConcert.app.getTicketsAPI();
		api.search({ tickets : request.ticket }, function(status, result) {
			callback(status && result.status? self.__makeOk({ interactions : result.interactions || [] }) :
				self.__makeError("FAIL", "error searching tickets"));
		});
	};

	AgentAPIServer.prototype.__UpdateInteractionLabels = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || interaction.isBeingComposed() || interaction.isInternal()) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			let campaign = self.__getCampaign(interaction.campaign);
			let campaignLabels = campaign && campaign.Labels || {};
			let interactionLabels = interaction.labels || {};
			let removedLabels = inConcert.Utils.arrayToBoolMap(request.removedLabels);
			
			let error = false;
			
			let addLabelToList = function(labels, labelId) {
				let label = campaignLabels[labelId];
				if (label) {
					labels[labelId] = {
						id : labelId,
						name : label.name
					};
				}
				else {
					error = "label " + labelId + " not configured";
				}
			};
			let currentLabels = Object.keys(interactionLabels)
				.filter(function(labelId) {
					return removedLabels[labelId];
				})
				.reduce(function(labels, labelId) {
					addLabelToList(labels, labelId);
					return labels;
				}, {});
			let addedLabels = (request.addedLabels || [])
				.filter(function(labelId) {
					return !interactionLabels[labelId];
				})
				.reduce(function(labels, labelId) {
					addLabelToList(labels, labelId);
					return labels;
				}, {});

			if (error) {
				callback(self.__makeError("FAIL", error));
			}
			else if (!$.isEmptyObject(addedLabels) || !$.isEmptyObject(currentLabels)) {
				interaction.updateLabels(addedLabels, currentLabels, function(status) {
					callback(status? self.__makeOk() : 
						self.__makeError("FAIL", "error updating labels"));
				}, "interaction");
			}
			else {
				callback(self.__makeOk());
			}
		}, true);
	};

	AgentAPIServer.prototype.__SetInteractionReadEvents = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || interaction.isBeingComposed() || interaction.isInternal()) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			interaction.setRead(request.read, request.events, function(status) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", "error setting interaction read"));
			});
		}, true);
	};

	AgentAPIServer.prototype.__ReadInteractionEvents = function(request, callback) {
		request.read = true;
		this.__SetInteractionReadEvents(request, callback);
	};

	AgentAPIServer.prototype.__UnreadInteractionEvents = function(request, callback) {
		request.read = false;
		this.__SetInteractionReadEvents(request, callback);
	};

	AgentAPIServer.prototype.__GetInteractionDispositions = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction) {
				self.__GetCampaignDispositions({ campaign : interaction.campaign }, callback);
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found"));
			}
		}, true);
	};

	AgentAPIServer.prototype.__GetInteractionMembersToTransfer = function(request, callback) {
		const self = this;
		let im = inConcert.app.getInteractionManager();
		im.getMembersForTransfer([request.id], function(members) {
			callback(self.__makeOk({
				members : (members || []).map(function(member) {
					return self.__makeMember(member);
				})
			}));
		}, "interaction");
	};

	AgentAPIServer.prototype.__GetInteractionCampaignsAndAccountsToTransfer = function(request, callback) {
		const self = this;
		let im = inConcert.app.getInteractionManager();
		im.getCampaignsAccountsForTransfer({ interactions : [request.id], app_id : "interaction", callback : function(campaigns) {
			campaigns = campaigns || {};
			callback(self.__makeOk({
				campaigns : Object.keys(campaigns)
					.sort()
					.map(function(campaignId) {
						let campaign = campaigns[campaignId];
						let cleanId = inConcert.Utils.getIdWithoutVCC(campaignId);
						return {
							Id : cleanId,
							Name : cleanId,
							Accounts : Object.keys(campaign.accounts)
								.map(function(accountId) {
									return self.__makeAccount(campaign, campaign.accounts[accountId]);
								})
								.sort(function(a1, a2) {
									return a1.Name.localeCompare(a2.Name);
								})
						};
					})
			}));
		} });
	};

	AgentAPIServer.prototype.__GetInteractionAttentionLevelsToTransfer = function(request, callback) {
		const self = this;
		let im = inConcert.app.getInteractionManager();
		im.getAccountsAttentionLevelsForTransfer({ interactions : [request.id], app_id : "interaction", callback : function(campaigns) {
			let campaign = inConcert.Utils.firstMapValue(campaigns || {}) || {};
			let account = inConcert.Utils.firstMapValue(campaign.accounts || {});
			let accountRaw = account? self.__makeAccount(campaign, account) : null;
			callback(self.__makeOk({ attentionLevels : accountRaw && accountRaw.AttentionLevels || [] }));
		} });
	};

	AgentAPIServer.prototype.__TransferInteraction = function(request, makeTarget, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || interaction.isBeingComposed() || interaction.isInternal()) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			let target = makeTarget(interaction);
			interaction.transfer(target, function(status, err) {
				callback(status? self.__makeOk({}) : self.__makeError("FAIL", err || "error transferring interaction"));
			}, "interaction");
		}, true);
	};
	
	AgentAPIServer.prototype.__TransferInteractionToMember = function(request, callback) {
		this.__TransferInteraction(request, function(interaction) {
			return {
				to : "member",
				user : request.member,
				att_level_id : request.attentionLevel
			};
		}, callback);
	};

	AgentAPIServer.prototype.__TransferInteractionToNumber = function(request, callback) {
		this.__TransferInteraction(request, function(interaction) {
			return {
				to : "number",
				number : request.number
			};
		}, callback);
	};
	
	AgentAPIServer.prototype.__TransferInteractionToCampaign = function(request, callback) {
		const self = this;
		self.__TransferInteraction(request, function(interaction) {
			return {
				to : "campaign",
				campaign_id : self.__appendVcc(request.campaign),
				account_id : request.account
			};
		}, callback);
	};

	AgentAPIServer.prototype.__TransferInteractionToAttentionLevel = function(request, callback) {
		const self = this;
		self.__TransferInteraction(request, function(interaction) {
			return {
				to : "attention_level",
				campaign_id : self.__appendVcc(interaction.campaign),
				account_id : interaction.account,
				att_level_id : request.attentionLevel
			};
		}, callback);
	};

	AgentAPIServer.prototype.__ConferenceInteraction = function(request, makeTarget, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || interaction.isBeingComposed() || interaction.isInternal()) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			let target = makeTarget(interaction);
			interaction.conference(target, function(status, err) {
				callback(status? self.__makeOk({}) : self.__makeError("FAIL", err || "error conference interaction"));
			}, "interaction");
		}, true);
	};

	AgentAPIServer.prototype.__ConferenceInteractionToMember = function(request, callback) {
		this.__ConferenceInteraction(request, function(interaction) {
			return {
				to : "member",
				user : request.member
			};
		}, callback);
	};

	AgentAPIServer.prototype.__ConferenceInteractionToNumber = function(request, callback) {
		this.__ConferenceInteraction(request, function(interaction) {
			return {
				to : "number",
				number : request.number
			};
		}, callback);
	};

	AgentAPIServer.prototype.__ScheduleInteraction = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (!interaction) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			if (!interaction.isScheduleEnabled()) {
				return callback(self.__makeError("FAIL", "schedule not supported"));
			}
			let data = {
				type : request.date? "DATE" : "NOW",
				date : request.date,
				number : request.number,
				overridedNumber : request.overridedNumber? true : false
			};
			interaction.schedule(data, function(status, err) {
				callback(status? self.__makeOk({}) : self.__makeError("FAIL", err || "error scheduling call"));
			});
		}, true);
	};

	AgentAPIServer.prototype.__GetInteractionTemplateMessageOptions = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction) {
				interaction.loadTemplateMessageOptions(function(options_d) {
					let options = inConcert.Utils.hash_to_sorted_array(options_d || {}, null, null, 
						{ without_id : true, attr : "id" });
					callback(self.__makeOk({ options : options }));
				});
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found"));
			}
		}, true);
	};

	AgentAPIServer.prototype.__GetInteractionReportingDataFields = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction) {
				callback(self.__makeOk({ fields : interaction.getReportFields(request.scope, request.json) || [] }));
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found"));
			}
		});
	};

	AgentAPIServer.prototype.__GetBaragentAttribute = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || !interaction.isAllegro()) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			//from popup url
			let value = interaction.getPopupUrlAttribute(request.field);
			if (value) {
				self.__makeOk({ value : value });
			}
			//from attributes (like notified variables)
			else {
				interaction.getBaragentAttribute(request.field, function(value) {
					self.__makeOk({ value : value });
				});
			}
		}, true);
	};

	AgentAPIServer.prototype.__GetInteractionReportingDataField = function(request, callback) {
		const self = this;
		if (request.scope === "phone_call") {
			return self.__GetBaragentAttribute(request, callback);
		}
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction) {
				callback(self.__makeOk({ value : interaction.getReportValue(request.scope, request.field, true) }));
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found"));
			}
		});
	};

	AgentAPIServer.prototype.__GetInteractionEvents = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction) {
				callback(self.__makeOk({ events : interaction.events }));
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found"));
			}
		});
	};

	AgentAPIServer.prototype.__GetInteractionUnreadEvents = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction) {
				callback(self.__makeOk({ events : interaction.getUnreadEvents() }));
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found"));
			}
		});
	};
	
	AgentAPIServer.prototype.getInteractionFromRepository = function (interaction, callback)
	{
		inConcert.app.getInteractionManager().getFromRepository(interaction, function(session, error_response, error_msg){
			if (!session) {
				callback(null);
			}
			else {
				if(typeof(session) === "string") {
					try {
						session = JSON.parse(session);
					} catch(e) {
						callback(null);
						return;
					}
				}
				callback(session);
			}
		});
	};

	AgentAPIServer.prototype.__GetChatInteractionMessages = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction) {
				var messages = [];
				self.getInteractionFromRepository({ id : request.id }, function(session) {
					if (session) {
						messages.push(session.messages);
						callback(self.__makeOk({ messages : messages || [] }));
						return;
					}
					else {
						return callback(self.__makeError("FAIL", "interaction not found"));
					}
				});
				return;
			}
			let actualSession = interaction.getSession();
			if (!actualSession) {

				return callback(self.__makeError("FAIL", "interaction session not found"));
			}
			let transferrerIds = interaction.getTransferrerIds();
			(function loadTransferId(index, messages = []) {
				if (index === (transferrerIds.length || 0)) {
					messages.push(actualSession.getMessages());
					callback(self.__makeOk({ messages : messages || [] }));
					return;
				}
				self.getInteractionFromRepository({ id : transferrerIds[index] }, function(session) {
					if (session) {
						messages.push(session.messages);
					}
					loadTransferId(index + 1, messages);
				});
			})(0, []);
		}, true);
	};

	AgentAPIServer.prototype.__GetChatBrowsingHistory = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (!interaction) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			interaction.getBrowsingHistory(function(history) {
				callback(history? self.__makeOk({ history : history }) : self.__makeError("FAIL", "error searching history"));
			});
		}, true);
	};

	AgentAPIServer.prototype.__SetInteractionFinalizationMessage = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || interaction.isBeingComposed() || interaction.isInternal()) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			interaction.getFinalizationMessageConfiguration(function(status) {
				if (!status) {
					return callback(self.__makeError("FAIL", "account hast not finalization message"));
				}
				let finalization = interaction.getFinalizationData() || {};
				finalization.send = !!request.send;
				finalization.mail_to = finalization.send? (request.address || finalization.mail_to || null) : null;
				interaction.setFinalizationData(finalization);
				let openClient = inConcert.app.getInteractionManager().getInteractionClient(interaction);
				if (openClient) {
					openClient.updateFinalization();
					callback(self.__makeOk());
				}
				else {
					self.__openInteractionClient(interaction, function() {
						callback(self.__makeOk());
					});
				}
			});
		}, true);
	};

	AgentAPIServer.prototype.__ForwardInteractionMessages = function(request, callback) {
		const self = this;
		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (!interaction || interaction.isBeingComposed() || interaction.isInternal()) {
				return callback(self.__makeError("FAIL", "interaction not found"));
			}
			if (!interaction.canForward()) {
				return callback(self.__makeError("FAIL", "can not forward messages"));
			}
			let messages = request.events.map(function(index) {
				return interaction.getForwardMessage(index);
			});
			interaction.forward(request.address, messages, function(status, message) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", message || "error forwarding messages"));
			});
		}, true);
	};

	AgentAPIServer.prototype.__CreateOutgoingInteraction = function(request, callback) {
		const self = this;
		self.__appendVccToContactId(request);
		if (!request.message && request.type === "CALL")
			request.message = {};
		if (request.account) {
			request.account = self.__getAccountId(request.account, request.campaign) || request.account;
		}
		if (request.attentionLevel) {
			request.attentionLevel = self.__getAttentionLevelId(request.attentionLevel, request.campaign) || 
				request.attentionLevel;
		}

		let address = null;
		let contact = null;
		function buildAddress(onBuilt) {
			if (!request.address) {
				return onBuilt();
			}
			self.__buildAddress(request.addressType, request.address, function(builtAddress) {
				if (builtAddress) {
					address = builtAddress;
					onBuilt();
				}
				else {
					callback(self.__makeError("FAIL", "error building address"));
				}
			});
		}

		const makeFakeContact = () => {
			contact = {
				Name: request.contactName,
				externalId: (request.isExternal && request.contactId) || null
			};
		};

		function buildContact(onBuilt) {
			if (!request.contactId) {
				makeFakeContact();
				return onBuilt();
			}
			
			function onResult(builtContact) {
				if (builtContact) {
					contact = builtContact;
					return onBuilt();
				}

				const account = self.__getCampaignAccount(request.campaign, request.account);
				if (	!account ||
						account.ContactManagement.Enabled !== "1" ||
						account.ContactManagement.AutoAddOutgoingContact !== "1") {
					makeFakeContact();
					return onBuilt();
				}

				contact = inConcert.app.getContactManager().create();	//Create an empty Contact object
				if (request.isExternal) {
					contact.externalId = request.contactId;
				} else {
					contact.Id = request.contactId;							
				}
				contact.Name = request.contactName || contact.Name;
				contact.updateAddress(address);							//Update contact address
				contact.save(function(ok, createdContact, error) {		//Save contact
					if(ok) {
						contact = createdContact;
						onBuilt();
					}
					else {
						callback(self.__makeError("FAIL",  error.reason || "error creating contact"));
					}
				});
			}

			if (request.isExternal) {
				self.__getContactByExternalId(request.contactId, onResult);
			}
			else {
				self.__getContact(request.contactId, onResult);
			}
		}

		function createInteraction() {
			let panel = inConcert.app.getMainWindow().getPanel("AssignedInteractions");
			let message = request.message? self.__buildMessage(request.type, request.message) : null;
			let interaction = null;
			let response = panel.createOutgoingInteraction(request.type, contact, address, 
				request.campaign? self.__getCampaign(request.campaign) : inConcert.app.getCampaigns(), 
				null, null, request.account || null, {
					fromAPI : true,
					draftMessage : message,
					draftAttentionLevel : request.attentionLevel,
					autoSendDraftMessage : request.autosend || null,
					draftMessageCallback : function(status, error) {
						let result = status? self.__makeOk({ interaction : self.__makeInteraction(interaction) }) :
							self.__makeError("FAIL", error || "error sending message");
						if (!status && interaction && request.autosend) {
							inConcert.app.getInteractionManager().cancelInteraction(interaction);
						}
						callback(result);
					}
				});
			if (response.success) {
				interaction = response.interaction;
			}
			else {
				if (response.conflicts) {
					let errorMessage = "Error creating interaction. Error message:" + response.conflicts.message;
					console.error(errorMessage);
					callback(self.__makeError("FAIL", errorMessage));
				} else {
					callback(self.__makeError("FAIL", "Error creating interaction."));
				}
			}
		}

		buildAddress(function() {
			buildContact(function() {
				createInteraction();
			});
		});
	};

	AgentAPIServer.prototype.__OpenInteractionClient = function(request, callback) {
		const self = this;
		self.__getAndOpenInteractionClient(request.id, function(interaction) {
			callback(interaction? self.__makeOk() : self.__makeError("FAIL", "interaction not found"));
		}, true);
	};

	AgentAPIServer.prototype.__SendMessageToInteraction = function(request, callback) {
		const self = this;

		if (!request.message) {
			return callback(self.__makeError("FAIL", "must enter a message"));
		}
		
		if (request.account) {
			request.account = self.__getAccountId(request.account, request.campaign) || request.account;
		}

		function sendMessage(interaction) {
			let message = self.__buildMessage(interaction.type, request.message);
			interaction.setDraftMessage(message, request.autosend, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error || "error sending message"));
			});
			let openClient = inConcert.app.getInteractionManager().getInteractionClient(interaction);
			self.__openInteractionClient(interaction, function() {
				if (openClient) {
					openClient.loadDraftMessage();
				}
			});
		}

		self.__getAssignedInteractionObject(request.id, function(interaction) {
			if (interaction) {	
				sendMessage(interaction);
			}
			else if (request.reopen) {
				self.__ReopenInteraction({ id : request.id }, function(result) {
					if (result.result === "OK") {
						self.__getInteractionObject(request.id, function(interaction) {
							sendMessage(interaction);
						}, true);
					}
					else {
						callback(result);
					}
				});
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found or closed"));
			}
		}, true);
	};

	AgentAPIServer.prototype.__SendMessageToAddress = function(request, callback) {
		const self = this;
		request.full = false;
		self.__appendVccToContactId(request);
		self.__GetAssignedInteractionByAddress(request, function(result) {
			if (result.interaction) {
				request.id = result.interaction.id;
				self.__SendMessageToInteraction(request, callback);
			}
			else if (result.createIfNotFound) {
				self.__CreateOutgoingInteraction(request, callback);
			}
			else {
				callback(self.__makeError("FAIL", "an assigned interaction not found"));
			}
		});
	};

	AgentAPIServer.prototype.__SendMessageToContact = function(request, callback) {
		const self = this;
		request.full = false;
		self.__appendVccToContactId(request);
		self.__GetAssignedInteractionByContact(request, function(result) {
			if (result.interaction) {
				request.id = result.interaction.id;
				self.__SendMessageToInteraction(request, callback);
			}
			else if (result.createIfNotFound) {
				self.__CreateOutgoingInteraction(request, callback);
			}
			else {
				callback(self.__makeError("FAIL", "an assigned interaction not found"));
			}
		});
	};

	AgentAPIServer.prototype.__GetSharedFiles = function(request, callback) {
		const self = this;
		request.extension = (request.extension && typeof(request.extension) === "string")? 
			[request.extension] : request.extension;
		let filter = {
			last : request.last || "",
			limit : request.rows || 20,
			query : {
				Name : request.name || "",
				Description : request.name || "",
				Type : request.type || "",
				Extensions : request.extension || null,
				MaxSize : request.maxSize || null,
				AgentVisible : "1"
			}
		};
		let filesAPI = inConcert.app.getSharedFilesAPI();
		filesAPI.searchFiles(filter, function(status, result) {
			callback(status? 
				self.__makeOk({
					files : result.rows,
					last : result.last,
					total : result.total_rows
				}) : self.__makeError("FAIL", "interaction not found or closed")
			);
		});
	};

/************************ Contact methods ************************/

	AgentAPIServer.prototype.__GetContact = function(request, callback) {
		const self = this;
		self.__getContact(request.id, function(contact) {
			callback(contact? self.__makeOk({ contact : contact }) : self.__makeError("FAIL", "contact not found"));
		});
	};

	AgentAPIServer.prototype.__GetContactByExternalId = function(request, callback) {
		const self = this;
		self.__getContactByExternalId(request.id, function(contact) {
			callback(contact? self.__makeOk({ contact : contact }) : self.__makeError("FAIL", "contact not found"));
		});
	};

	AgentAPIServer.prototype.__GetContactByAddress = function(request, callback) {
		const self = this;
		self.__getContactByAddressRaw(request.addressType, request.address, function(contact) {
			callback(contact? self.__makeOk({ contact : contact }) : self.__makeError("FAIL", "contact not found"));
		});
	};

	AgentAPIServer.prototype.__GetInteractionContact = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction && interaction.contact) {
				request.id = interaction.contact;
				self.__GetContact(request, callback);
			}
			else {
				callback(self.__makeError("FAIL", interaction? "interaction has not a contact" : "interaction not found"));
			}
		}, true);
	};

	AgentAPIServer.prototype.__GetContactInteractionsHistory = function(request, callback) {
		const self = this;
		self.__appendVccToContactId(request);
		let filter = self.__makeServerFilter(request, 
			[{ id : "contactId", elementId : request.isExternal? "contactExternalId" : "contact" }]);
		let im = inConcert.app.getInteractionManager();
		im.searchArchiveInteractions(filter, function(status, result) {
			if (status) {
				self.__replyServerInteractions(result.interactions, result.total, result.last, callback);
			}
			else {
				callback(self.__makeError(result && result.result, result && result.reason || "error searching interactions"));
			}
		});
	};

	AgentAPIServer.prototype.__GetInteractionContactInteractionsHistory = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction && interaction.contact) {
				request.contactId = interaction.contact;
				self.__GetContactInteractionsHistory(request, callback);
			}
			else {
				callback(self.__makeError("FAIL", interaction? "interaction has not a contact" : "interaction not found"));
			}
		}, true);
	};

/************************ Twitter methods ************************/

	AgentAPIServer.prototype.__SendTwitterPrivateMessageToInteraction = function(request, callback) {
		request.message = this.__buildMessage("TWITTER", request.message);
		request.message.privateMessage = true;
		request.message.replyEvent = request.event;
		this.__SendMessageToInteraction(request, callback);
	};
	
	AgentAPIServer.prototype.__SendTwitterTweetReplyMessageToInteraction = function(request, callback) {
		request.message = this.__buildMessage("TWITTER", request.message);
		request.message.replyTweet = true;
		request.message.replyEvent = request.event;
		this.__SendMessageToInteraction(request, callback);
	};

	AgentAPIServer.prototype.__RetweetTwitterMessage = function(request, callback) {
		request.message = this.__buildMessage("TWITTER", request.message);
		request.message.retweet = true;
		request.message.replyEvent = request.event;
		this.__SendMessageToInteraction(request, callback);
	};

	AgentAPIServer.prototype.__LikeTwitterMessage = function(request, callback) {
		request.message = this.__buildMessage("TWITTER", "");
		request.message.like = true;
		request.message.replyEvent = request.event;
		request.autosend = true;
		this.__SendMessageToInteraction(request, callback);
	};
	
	AgentAPIServer.prototype.__UnlikeTwitterMessage = function(request, callback) {
		request.message = this.__buildMessage("TWITTER", "");
		request.message.unlike = true;
		request.message.replyEvent = request.event;
		request.autosend = true;
		this.__SendMessageToInteraction(request, callback);
	};
	
/************************ Facebook methods ************************/

	AgentAPIServer.prototype.__SendFacebookCommentReplyMessageToInteraction = function(request, callback) {
		request.message = this.__buildMessage("FACEBOOK", request.message);
		request.message.replyComment = request.event == null? false : true;
		request.message.replyEvent = request.event;
		this.__SendMessageToInteraction(request, callback);
	};

	AgentAPIServer.prototype.__SendFacebookPrivateReplyMessageToInteraction = function(request, callback) {
		request.message = this.__buildMessage("FACEBOOK", request.message);
		request.message.privateReply = request.event == null? false : true;
		request.message.replyEvent = request.event;
		this.__SendMessageToInteraction(request, callback);
	};

	AgentAPIServer.prototype.__LikeFacebookMessage = function(request, callback) {
		request.message = this.__buildMessage("FACEBOOK", "");
		request.message.like = true;
		request.message.replyEvent = request.event;
		request.autosend = true;
		this.__SendMessageToInteraction(request, callback);
	};

	AgentAPIServer.prototype.__UnlikeFacebookMessage = function(request, callback) {
		request.message = this.__buildMessage("FACEBOOK", "");
		request.message.unlike = true;
		request.message.replyEvent = request.event;
		request.autosend = true;
		this.__SendMessageToInteraction(request, callback);
	};

/************************ Mail methods ************************/

	AgentAPIServer.prototype.__ReplyInteractionMail = function(request, callback) {
		request.message = this.__buildMessage("MAIL", request.message);
		request.message.mailReply = true;
		request.message.replyEvent = request.event;
		this.__SendMessageToInteraction(request, callback);
	};

	AgentAPIServer.prototype.__ReplyAllInteractionMail = function(request, callback) {
		request.message = this.__buildMessage("MAIL", request.message);
		request.message.mailReplyAll = true;
		request.message.replyEvent = request.event;
		this.__SendMessageToInteraction(request, callback);
	};

	AgentAPIServer.prototype.__ForwardInteractionMail = function(request, callback) {
		request.message = this.__buildMessage("MAIL", request.message);
		request.message.mailForward = true;
		request.message.replyEvent = request.event;
		this.__SendMessageToInteraction(request, callback);
	};

	AgentAPIServer.prototype.__PrintInteractionMail = function(request, callback) {
		request.message = this.__buildMessage("MAIL", request.message);
		request.message.print = true;
		request.message.replyEvent = request.event;
		request.autosend = true;
		this.__SendMessageToInteraction(request, callback);
	};

/************************ Phone methods ************************/

	AgentAPIServer.prototype.__GetPhoneState = function(request, callback) {
		let state = inConcert.app.getInteractionManager().getPhoneState();
		callback(this.__makeOk({ state : state }));
	};

	AgentAPIServer.prototype.__callPhoneMethod = function(request, method, callback) {
		const self = this;
		let phone = inConcert.app.getInteractionManager().getChannelClient("CALL");
		if (!phone) {
			return callback(self.__makeError("FAIL", "User has not phone"));
		}
		if (request.id) {
			self.__getAssignedInteractionObject(request.id, function(interaction) {
				if (!interaction) {
					return callback(self.__makeError("FAIL", "interaction not found or not taken"));
				}
				method(phone, interaction);
			}, true);
		}
		else {
			let interaction = phone.getInteraction();
			if (!interaction) {
				return callback(self.__makeError("FAIL", "phone has not active interactions"));
			}
			method(phone, interaction);
		}
	};

	AgentAPIServer.prototype.__HasVideoCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			callback(self.__makeOk({ video : phone.hasVideo() }));
		});
	};

	AgentAPIServer.prototype.__AnswerCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.answer(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__HoldCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.hold(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__UnholdCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.unhold(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__MuteCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.mute(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__UnmuteCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.unmute(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__DisableCallVideo = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.disableVideo(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__EnableCallVideo = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.enableVideo(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__SwitchCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone._switch(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__UnswitchCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.unswitch(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__StartCallRecord = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.record(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__StopCallRecord = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.stopRecord(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__GetInteractionRecords = function(request, callback) {
		const self = this;
		self.__getInteractionObject(request.id, function(interaction) {
			if (interaction && interaction.interactionData) {
				if(inConcert.app.has_allegro_integration())
				{
					inConcert.app.getAllegroAPI().getInteractionRecords(interaction.id,
						function( success, records ) {
							if (!success) {
								callback(self.__makeError("FAIL", "records not found"));
							}
							$.each(records, function(_, record) {
								let recordUrl = self.__getAllegroRecordUrl(interaction, record.Section, record.Part);
								record.url = recordUrl ? recordUrl : undefined;
							});								
							callback(self.__makeOk({ records : records }));
						}
					);
				}
				else
				{
					if(interaction.interactionData.records)
					{
						$.each(interaction.interactionData.records, function(_, record) {
							let recordUrl = self.__getRecordUrl(interaction, record.Part);
							record.url = recordUrl ? recordUrl : undefined;
						});				
						callback(self.__makeOk({ records : interaction.interactionData.records }));
					}
					else
					{
						callback(self.__makeOk({ records : [] }));
					}
				}
			}
			else {
				callback(self.__makeError("FAIL", "interaction not found"));
			}
		});
	};

	AgentAPIServer.prototype.__AcceptCallPreview = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.acceptPreview(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__CancelCallPreview = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.cancelPreview(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__ConfirmCallTransfer = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.confirmTransfer(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__CancelCallTransfer = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.cancelTransfer(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__ConfirmCallConference = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.confirmConference(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__CancelCallConference = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.cancelConference(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__RedialCall = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.redial(interaction, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	AgentAPIServer.prototype.__SendCallDigit = function(request, callback) {
		const self = this;
		self.__callPhoneMethod(request, function(phone, interaction) {
			phone.sendDigit(interaction, request.digit, function(status, error) {
				callback(status? self.__makeOk() : self.__makeError("FAIL", error));
			});
		});
	};

	/************************ Outbound Engine methods ************************/

	AgentAPIServer.prototype.__callOutboundContactMethod = function(request, callback) {
		inConcert.app.getOutboundEngineAPI().GetOutboundContact(request, (status, result) => {
			if (status && result?.contact) {
				request.processId = request.processId || result.contact.ProcessId;
				request.batchId = request.batchId || result.contact.BatchId;
			}
			callback();
		});
	};

	AgentAPIServer.prototype.__callOutboundEngineMethod = function (request, callback) {
		this.__getAssignedInteractionObject(request.id, (interaction) => {
			const dialer = interaction?.interactionData?.dialer || {};
			if (!request.processId && !request.campaignId) {
				request.campaignId = dialer.campaign || interaction?.campaign || undefined;
				request.processId = dialer.processId;
			}
			if (!request.contactId) request.contactId = dialer.contactId;
			if (!request.batchId) request.batchId = dialer.batchId;

			request.processId = this.__appendVcc(request.processId);
			request.campaignId = this.__appendVcc(request.campaignId);
			request.batchId = this.__appendProcessId(request.batchId, request.processId);

			if (request.oeRequired && request.contactId && !request.processId)
				return this.__callOutboundContactMethod(request, () =>
					callback(inConcert.app.getOutboundEngineAPI(), interaction));

			callback(inConcert.app.getOutboundEngineAPI(), interaction);
		}, true);
	};

	AgentAPIServer.prototype.__outboundEngineMethodCallback = function (callback) {
		return (status, result) => {
			if (status || result?.result) {
				if (!result.result) result.result = status ? "OK" : "FAIL";
				return callback(result);
			}

			callback(this.__makeError("FAIL", result));
		};
	};

	AgentAPIServer.prototype.__GetOutboundEngineCampaigns = function (_request, callback) {
		inConcert.app.getOutboundEngineAPI().GetCampaigns(
			this.__outboundEngineMethodCallback(callback)
		);
	};

	AgentAPIServer.prototype.__GetOutboundEngineCampaignBatches = function (request, callback) {
		inConcert.app.getOutboundEngineAPI().GetCampaignBatches(
			request, this.__outboundEngineMethodCallback(callback)
		);
	};

	AgentAPIServer.prototype.__GetOutboundEngineCampaignBatch = function (request, callback) {
		inConcert.app.getOutboundEngineAPI().GetCampaignBatch(
			request, this.__outboundEngineMethodCallback(callback)
		);
	};

	AgentAPIServer.prototype.__GetOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.GetOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__GetOutboundContactByName = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.GetOutboundContactByName(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__GetOutboundContactByAddress = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.GetOutboundContactByAddress(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__SearchOutboundContacts = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.SearchOutboundContacts(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__GetOutboundContactResults = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.GetOutboundContactResults(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__StoreOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.StoreOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__SetOutboundContactAddresses = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.SetOutboundContactAddresses(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__SetOutboundContactValues = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.SetOutboundContactValues(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__SetOutboundContactSkills = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.SetOutboundContactSkills(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__RemoveOutboundContactAddress = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.RemoveOutboundContactAddress(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__EnableOutboundContactAddress = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.EnableOutboundContactAddress(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__DisableOutboundContactAddress = function (request, callback) {
	this.__callOutboundEngineMethod(request, (api) =>
		api.DisableOutboundContactAddress(request, this.__outboundEngineMethodCallback(callback))
	);
	};

	AgentAPIServer.prototype.__RescheduleOutboundContact = function (request, callback) {
		this.__callOutboundEngineMethod(request, (api) =>
			api.RescheduleOutboundContact(request, this.__outboundEngineMethodCallback(callback))
		);
	};

	AgentAPIServer.prototype.__AddContactToOutboundBatch = function (request, callback) {
		this.__callOutboundEngineMethod(request, (api) =>
			api.AddContactToOutboundBatch(request, this.__outboundEngineMethodCallback(callback))
		);
	};

	AgentAPIServer.prototype.__PauseOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.PauseOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__ResumeOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.ResumeOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__CancelOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.CancelOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__FinishOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.FinishOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__RemoveOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.RemoveOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__RecycleOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.RecycleOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__ChangePriorityOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.ChangePriorityOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__SetEngineTypeOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.SetEngineTypeOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__SetAgentOutboundContact = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.SetAgentOutboundContact(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__GetOutboundEngineActionStatus = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.GetActionStatus(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__ValidateAddress = function (request, callback) {
	  this.__callOutboundEngineMethod(request, (api) =>
		api.ValidateAddress(request, this.__outboundEngineMethodCallback(callback))
	  );
	};

	AgentAPIServer.prototype.__GetAddressesBook = function (_, callback) {
		const self = this;
		inConcert.app.getContactManager().getAddressesBook(function(addresses, error){
			callback(addresses? self.__makeOk({addresses: addresses}) : self.__makeError("FAIL", error.result));
		});
	};

	AgentAPIServer.prototype.__HelpDeskReloadApps = function (_, callback) {
		inConcert.app.getMainWindow().HelpDeskReloadApps((resp)=> {
			callback(this.__makeOk(resp));
		});
	};
}());
