/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global $, inConcert, Mustache, _ */

// TO-DO: General TODO, remove long lines as much as possible.

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}

/**
 * @namespace
 * @name ContactWindow
 * @author Sebastian Moreno (smoreno@inconcertcc.com)
 */
(function () {

	'use strict';

	const format = inConcert.Utils.format;
	const nid = inConcert.Utils.getNormalizedId;

	/**
	 * Creates a new contact info instance.
	 *
	 * @constructor
	 * @memberOf ContactWindow
	 *
	 * @param {Object} options
	 * @param {Object} [options.interaction]
	 * @param {String} [options.name]
	 * @param {String} [options.contact_id]
	 * @param {String} [options.external_id]
	 * @param {Array}  [options.addresses]
	 * @param {Array}  [options.avatar]
	 *
	 */
	var ContactWindow = function( options )
	{
		this.m_interaction = options.interaction;
		this.m_templates = inConcert.ContactInformationTemplates;
		this.m_contact_manager = inConcert.app.getContactManager();

		this.m_contacts_per_page = 40;
		this.m_dialog = null;

		this.m_from_app = options.from_app;
		this.m_contact_name = options.name;
		this.m_contact_external_id = options.external_id;
		this.m_contact_avatar = options.avatar;
		this.m_contact = null;

		// normally, contact id and addresses are an exclusive OR (XOR),
		this.m_contact_id = options.contact_id;
		this.m_contact_addresses = options.addresses || [];

		this.m_main_basic_info = new inConcert.ContactWindowBasicInfo('main', this);
	};

	//Define functions and objects in the namespace in case they don't already exist
	if (typeof inConcert.ContactWindow !== 'function') {
		inConcert.ContactWindow = ContactWindow;
	}

	// TODO: Build a wrapper for async functions to show/hide automatically
	// the waitings or the messages.
	ContactWindow.prototype.__show_waiting = function(){
		$('#contact_window_waiting').show();
	};

	ContactWindow.prototype.__hide_waiting = function(){
		$('#contact_window_waiting').hide();
	};

	ContactWindow.prototype.__get_contact_search_container = function(){
		return $('#contact_window_search_container');
	};

	ContactWindow.prototype.__get_contact_search_input = function(){
		return $('#contact_window_search_input');
	};

	ContactWindow.prototype.__get_contact_merge_container = function(){
		return $('#contact_window_merge_container');
	};

	ContactWindow.prototype.__get_search_list_container = function(){
		return $('#contact_window_search_list');
	};

	ContactWindow.prototype.__get_conflict_view = function(){
		return $('#contact_window_conflict_view');
	};

	/**
	 * Show a message to the user
	 *
	 * @todo Convert options to an object
	 * @public
	 *
	 * @param  {String}  message     - Message to be shown.
	 * @param  {String}  type        - Message type, possible values are: [OK, alert, info, error, process]
	 * @param  {Boolean} [autoclose] - Optional. If the message should be closed automatically
	 * @param  {Number}  [timeout]   - Optional. Timeout in seconds. How much we need to wait to close the message, default 3 s.
	 */
	ContactWindow.prototype.show_message = function(message, type, autoclose, timeout)
	{
		// remove previous messages
		$('#contact_window_message_container').remove();

		// build new message
		var message_view = {
			message : message,
			type : inConcert.Utils.capitaliseFirstLetter(type),
			close : "<%=_T('Close')%>"
		};
		var $message_node = $(Mustache.render(this.m_templates.messages_tpl, message_view));
		this.m_main_basic_info.__get_contact_window_container().append($message_node);

		// setup close message
		$('#contact_window_message_close').click(function(){
			$message_node.remove();
		});

		// setup auto close message
		if (autoclose){
			window.setTimeout(function(){
				$message_node.remove();
			}, timeout ? timeout * 1000 : 3000);
		}
	};

	/**
	 * Hide message
	 *
	 * @public
	 */
	ContactWindow.prototype.hide_message = function(){
		$('#contact_window_message_container').remove();
	};

	/**
	 * Get the confirmation dialog container
	 *
	 * @return {Object} - jQuery object with the container.
	 */
	ContactWindow.prototype.__get_confirmation_dialog = function(){
		return $('#contact_window_confirmation_container');
	};

	/**
	 * Will display a confirmation dialog for danger actions.
	 *
	 * NOTE: Cancel will remove the confirmation dialog.
	 *
	 * @private
	 *
	 * @param  {String}   options
	 * @param  {String}   options.message           - Message to be shown.
	 * @param  {String}   [options.confirm_text]    - Confirm button text.
	 * @param  {String}   [options.cancel_text]     - Cancel button text.
	 * @param  {Boolean}  [options.hide_cancel]     - Hide cancel button.
	 *
	 * @param  {Object}   callbacks
	 * @param  {Function} callbacks.on_confirmed    - A callback when the user press confirm, it's not optional.
	 * @param  {Function} [callbacks.on_canceled]   - A callback when the user press cancel, it's optional.
	 */
	ContactWindow.prototype.__show_confirmation_dialog = function(options, callbacks)
	{
		var self = this;
		var $contact_window = this.m_main_basic_info.get_contact_window();

		var $confirm_dialog = this.__get_confirmation_dialog();
		if ($confirm_dialog.length > 0){
			$confirm_dialog.remove();
		}

		var previous_title = this.__get_window_title();
		this.__update_window_title("<%=_T('Confirm action')%>");

		var confirm_view = {
			confirm_label : options.confirm_text || "<%=_T('CONFIRM')%>",
			cancel_label : options.cancel_text || "<%=_T('CANCEL')%>",
			hide_cancel : options.hide_cancel,
			message : options.message
		};

		$confirm_dialog = $(Mustache.render(this.m_templates.confirmation_dialog_tpl, confirm_view));
		$contact_window.append($confirm_dialog);

		$('#contact_window_confirmation_confirm').click(function(){
			callbacks.on_confirmed();
			$confirm_dialog.remove();
		});

		$('#contact_window_confirmation_cancel').click(function(){
			$confirm_dialog.remove();
			self.__update_window_title(previous_title);
			if (typeof(callbacks.on_canceled) === 'function'){
				callbacks.on_canceled();
			}
		});
	};

	/**
	 * Will try to reload the iframe with cci integration after contact is manually changed
	 */
	ContactWindow.prototype.__reload_cci = function()
	{
		const iframe = document.getElementById('cc_integration_sandbox_iframe');
		if (!iframe) return;

		iframe.contentDocument.location.reload(true);
	};


	/**
	 * Will try to update the interaction contact (of course if we are dealing with an interaction)
	 *
	 * If the interaction does not have events we update the interaction contact without asking
	 *
	 * @param  {Object}   contact
	 * @param  {Function} on_interaction_updated
	 */
	ContactWindow.prototype.__update_interaction_contact = function(contact, on_interaction_updated)
	{
		var self = this;

		on_interaction_updated = on_interaction_updated || $.noop;

		// we only update the contact if:
		// * we are dealing with an interaction
		// * the interaction is not finished
		// * the interaction contact is different from the requested contact
		if (this.m_interaction && !this.m_interaction.isFinished() && this.m_interaction.contact !== contact.Id){
			var event_list = this.m_interaction.getEventList();

			var set_contact_callback = function( success, error ){
				if (!success){
					var error_msg = error && error.reason ? error.reason : "<%=_T('internal error')%>";
					self.show_message(format("<%=_T('Error updating conversation contact: {0}')%>", error_msg), 'error');
				}
				on_interaction_updated(success);
				self.__reload_cci();
			};
			var app_id = this.get_app_id();

			var __set_interaction_contact = function() {
				self.m_interaction.setContact(
					contact.Id,
					contact.Name,
					set_contact_callback,
					app_id
				);
			};

			// if we have more than 1 event or every events belongs to the same client
			// update the interaction contact automatically
			if ((event_list && event_list.length <= 1) || this.m_interaction.eventsAreFromTheSameClient()) {
				__set_interaction_contact();
			} else {
				this.__show_confirmation_dialog({
					message: "<%=_T('Do you want to replace the conversation contact with this one?')%>",
					confirm_text: "<%=_T('Yes')%>",
					cancel_text: "<%=_T('No')%>"
				}, {
					on_confirmed : function(){
						__set_interaction_contact();
					},
					on_canceled : function(){
						on_interaction_updated(false);
					}
				});
			}
		} else {
			self.__reload_cci();
		}
	};

	/**
	 * Will create the main contact window
	 *
	 * @private
	 */
	ContactWindow.prototype.__create_window = function()
	{
		var self = this;
		this.__hide_waiting();
		var dialog_options = { size : { min_width:755, width:755, min_height:560, height:560 } };
		this.m_dialog = inConcert.app.getWindowManager().createDialog(this.m_main_basic_info.get_contact_window(), dialog_options);
		this.m_dialog.on( 'dialogclose', function() {
			self.hide();
		});
	};

	/**
	 * Called to resolve a conflict creating a new contact.
	 *
	 * @private
	 * @param  {Object}   $conflict_node
	 * @param  {Object}   $container
	 * @param  {Function} callback
	 */
	ContactWindow.prototype.__conflict_handle_create_new = function($conflict_node, $container, callback)
	{
		var self = this;

		// show a disclaimer
		this.show_message("<%=_T('Watch out! Creating a new contact using the same address will remove the address from the existing contact')%>", 'alert');

		// we create a new instance of the basic info for our embedded purposes
		var basic_info_embedded = new inConcert.ContactWindowBasicInfo('embedded_conflict', this);

		// hide conflict node
		$conflict_node.hide();

		// we build a basic contact info embedded on our container
		// just the addresses are necessary, all the other things needs to be editable.
		var sham_contact = this.m_contact_manager.create({
			isNew : true,  /* we kinda fake this contact */
			Name : '',
			addresses : this.m_contact_addresses,
			Id: inConcert.Utils.generateGuid(),
			externalId : ''
		});
		this.m_contact = sham_contact;

		var $contact_node = basic_info_embedded.build_contact_info_container(sham_contact, $container);
		$contact_node.removeClass("existingContactContainer").addClass("existingContactContainer");
		basic_info_embedded.setup_contact_save_handlers(sham_contact, {
			on_success : function(ok){
				self.hide();
				if (typeof(callback) === 'function'){
					callback(ok);
				}
			}
		});

		basic_info_embedded.setup_contact_cancel_handlers(function(){
			self.hide_message();
			$contact_node.remove();
			$conflict_node.show();
		});

		basic_info_embedded.__setup_address_edition_handlers();
		basic_info_embedded.__setup_add_new_address_handlers();
		basic_info_embedded.__setup_name_edition_handlers();
		basic_info_embedded.__setup_external_id_edition_handlers();
	};

	/**
	 * Called to resolve a conflict moving an address or an external id
	 *
	 * @private
	 *
	 * @param  {Object}   conflict
	 * @param  {Object}   conflictive_contact
	 * @param  {Object}   $conflict_node
	 * @param  {Function} callback
	 */
	ContactWindow.prototype.__conflict_handle_move = function(conflict, conflictive_contact, $conflict_node, callback)
	{
		var self = this;

		var field_type = (conflict.field_type || '');

		this.hide_message();

		var callbacks = {
			/** Called when the user clicks "Confirm" dialog */
			on_confirmed : function(){
				if (field_type === 'address'){
					conflictive_contact.addresses.forEach(function(address, index){
						if (conflict.field === address.displayable){
							// this modify the contact addresses array, remove the address.
							conflictive_contact.addresses.splice(index, 1);
						}
					});
				}
				else if (field_type === 'externalId') {
					conflictive_contact.externalId = '';
				}
				var campaign_id = self.m_interaction ? self.m_interaction.campaign : undefined;
				// go and save conflictive contact
				const app_id = self.get_app_id();
				conflictive_contact.save(
					function(success_save, undefined /* saved_conflict_contact */ , error) {
						if (success_save && !error){
							self.m_contact.save(
								function(success, saved_contact, contact_saved_err) {
									if (success){
										self.show_message(format("<%=_T('Success resolving conflict, {0} removed from conflictive contact')%>", field_type), 'info', true);
										// hide this window and
										$conflict_node.remove();
										self.m_main_basic_info.__toggle_contact_basic_info_containers(true);
										// call on saved
										self.__on_contact_updated(saved_contact);
									}
									else {
										let msg = "<%=_T('Unrecognized')%>";
										if (contact_saved_err) {
											if (contact_saved_err.result === "E_ADDR_IN_USE") {
												msg = "<%=_T('Address belongs to another contact')%>";
											}
										}
										self.show_message(format("<%=_T('Error {0} contact: {1}')%>", self.m_contact ? "<%=_T('saving')%>" : "<%=_T('creating')%>", msg), 'error');
									}
									if (typeof(callback) === 'function'){
										callback(success);
									}
								},
								campaign_id, // optional, used by permissions
								app_id
							);
						}
						else {
							self.show_message(format("<%=_T('Error {0} contact: {1}')%>", self.m_contact ? "<%=_T('saving')%>" : "<%=_T('creating')%>", error && error.reason), 'error');
						}
					},
					campaign_id, // optional, used by permissions
					app_id
				);
			}
		};

		var confirmation_options = {
			message: format("<%=_T('Are you sure? This will remove the {0}: {1} from contact: {2}')%>", field_type, conflict.field, conflictive_contact.Name)
		};
		this.__show_confirmation_dialog(confirmation_options, callbacks);
	};


	/**
	 * Called to resolve a conflict merging two contacts
	 *
	 * @private
	 *
	 * @param  {Object}   conflict
	 * @param  {Object}   source_contact
	 * @param  {Object}   conflictive_contact
	 * @param  {Object}   $conflict_node
	 * @param  {Function} callback
	 */
	ContactWindow.prototype.__conflict_handle_merge = function(conflict, source_contact, conflictive_contact, $conflict_node, callback)
	{
		// TO-DO: This is completely necessary?
		$conflict_node.hide();

		var field_type = (conflict.field_type || '');

		// multiple matches conflicts allow edit everything
		var visual_settings = {
			addresses_disabled : conflict.multiple_conflicts ? field_type !== 'address' : false,
			name_disabled : conflict.multiple_conflicts,
			externalid_disabled : conflict.multiple_conflicts ? field_type !== 'externalId' : false
		};

		this.__switch_to_merge_view(source_contact, conflictive_contact, visual_settings, {
			on_back : function(){
				$conflict_node.show();
			},
			on_canceled : function(){
				$conflict_node.show();
			},
			on_success : function(){
				if (typeof(callback) === 'function'){
					callback(true);
				}
			}
		});
	};

	/*

	*/
	ContactWindow.prototype.__GetDisplayableAddresss = function(address) {
		let displayable = address.displayable;
		let addressType = (address.Type || address.type || "").toUpperCase();
		if (addressType === "PHONE") {
			if (!address.LocalNumber) {
				displayable = (
					(address.Country || "") + (address.Area || "") + (address.SimpleNumber || address.Number || "") ||
					address.address || ""
				);
				const first_char = displayable.substring(0, 1);
				displayable = first_char === "+" ? displayable : "+" + displayable;
			}
		}
		return displayable;
	};

	/**
	 * Will setup the conflict window, it can be used to show conflicts
	 * when saving contacts or when is an automatic conflict (auto merge)
	 *
	 * @todo Split into multiples functions, its too long.
	 * @todo Do not assume grabbing contacts from outside (i.e. contact member)
	 *
	 * @private
	 * @async
	 *
	 * @param {Object}  conflict
	 * @param {String}  conflict.message
	 * @param {String}  conflict.field
	 * @param {String}  conflict.field_type
	 * @param {String}  conflict.contact_id
	 * @param {String}  conflict.contact_external_id
	 * @param {Array}   conflict.addresses
	 * @param {Boolean} conflict.allow_merge
	 * @param {Boolean} conflict.allow_create_new
	 * @param {Boolean} conflict.allow_move
	 * @param {Boolean} conflict.allow_cancel
	 *
	 * @param {Function} [callback]
	 */
	ContactWindow.prototype.__show_conflict_window = function(conflict, callback)
	{
		var self = this;

		// TO-DO: Convert this to parameters
		var interaction_contact_name = this.m_interaction && this.m_interaction.interactionData && this.m_interaction.interactionData.contactName || '';

		// remove previous conflict window
		var previous_conflict = this.__get_conflict_view();
		if (previous_conflict.length > 0 ){
			previous_conflict.remove();
		}

		var field_type = (conflict.field_type || '');

		// go and get more info for this conflictive contact
		this.m_contact_manager.getContact(conflict.contact_id, function(conflictive_contact, error) {
			if (conflictive_contact && !error){

				// this will hide the contact containers
				self.m_main_basic_info.__toggle_contact_basic_info_containers(false);

				let translated_address;
				if (field_type.toUpperCase() === "ADDRESS") {
					translated_address = "<%=_T('Address')%>";
				}
				var conflict_view = {
					// interaction contact
					// TO-DO: This seems wrong, derp.
					interaction_contact_name : interaction_contact_name,
					interaction_contact_id : conflict.contact_external_id ? conflict.contact_external_id : "<%=_T('N/A')%>",
					has_id_conflict_interaction : field_type === 'externalId',

					// conflict contact
					conflict_contact_name : conflictive_contact.Name,
					conflictive_contact_id : conflictive_contact.externalId ? conflictive_contact.externalId : "<%=_T('N/A')%>",
					has_id_conflict_existing : field_type === 'externalId',

					// message
					message : conflict.message,

					// translations
					interaction_info_label : "<%=_T('Conversation Information')%>",
					existing_info_label : "<%=_T('System (pre-existing contact)')%>",
					contact_name_label : "<%=_T('Contact Name')%>",
					contact_id_label : "<%=_T('Contact ID')%>",
					conflict : "<%=_T('CONFLICT')%>",
					merge_label : "<%=_T('Merge')%>",
					merge_tooltip : "<%=_T('Show merge preview')%>",
					move_label : format("<%=_T('MOVE {0}')%>", translated_address),
					move_tooltip : format("<%=_T('Remove {0} from conflictive contact')%>", translated_address),
					nothing_label : "<%=_T('NOTHING')%>",
					nothing_tooltip : "<%=_T('Move back to the contact view')%>",
					create_new_label : "<%=_T('New')%>",
					create_new_tooltip : "<%=_T('Create a new contact with existing info')%>",

					// buttons on/off
					allow_merge : conflict.allow_merge,
					allow_move : conflict.allow_move,
					allow_cancel : conflict.allow_cancel,
					allow_create_new :conflict.allow_create_new
				};

				// create the conflict window
				var $conflict_node = $(Mustache.render(self.m_templates.conflict_view_tpl, conflict_view));
				var $container = self.m_main_basic_info.__get_contact_window_container();
				$container.append($conflict_node);

				const __get_address_node_markup = function(address) {
					const displayable = self.__GetDisplayableAddresss(address);
					const address_view = {
						address_displayable: displayable,
						address_type       : address.Type || address.type,
						has_conflict       : field_type === "address" && conflict.field === displayable,
						conflict_label     : "<%=_T('CONFLICT')%>"
					};
					return Mustache.render(self.m_templates.save_conflict_address_tpl, address_view);
				};

				// add interaction/contact addresses to the list
				// TO-DO: Find a way to pass this addresses to the main template (mustache)
				if (conflict.addresses && conflict.addresses.length > 0 ){
					var $interaction_address_list = $('#contact_window_conflict_interaction_contact_list');
					conflict.addresses.forEach(function(address) {
						$interaction_address_list.append(__get_address_node_markup(address));
					});
				}

				// add contact addresses to the list
				// TO-DO: Find a way to pass this addresses to the main template (mustache)
				if (conflictive_contact.addresses && conflictive_contact.addresses.length > 0 ){
					var $conflict_address_list = $('#contact_window_conflict_existing_contact_list');
					conflictive_contact.addresses.forEach(function(address) {
						$conflict_address_list.append(__get_address_node_markup(address));
					});
				}

				// MERGE
				if (conflict.allow_merge){
					$('#contact_window_conflict_merge').click(function(){
						// TO-DO: This seems pretty shitty
						// We override the id that comes from the interaction data
						// to do the merge thing
						var source_contact = self.m_contact;
						if (self.m_interaction){
							source_contact.Name = interaction_contact_name;
							if (field_type === 'externalId'){
								source_contact.externalId = conflict.field;
							}
						}
						self.__conflict_handle_merge(conflict, source_contact, conflictive_contact, $conflict_node, callback);
					});
				}

				// CREATE NEW
				if (conflict.allow_create_new){
					$('#contact_window_conflict_create_new').click(function(){
						self.__conflict_handle_create_new($conflict_node, $container, callback);
					});
				}
				
				// NOTHING
				if (conflict.allow_cancel){
					$('#contact_window_conflict_nothing').click(function(){
						$conflict_node.remove();
						// this will show the contact containers
						self.m_main_basic_info.__toggle_contact_basic_info_containers(true);
						if (typeof(callback) === 'function'){
							callback(false);
						}
					});
				}

				// MOVE ADDRESS/EXTERNAL ID
				if (conflict.allow_move){
					$('#contact_window_conflict_move').click(function(){
						self.__conflict_handle_move(conflict, conflictive_contact, $conflict_node, callback);
					});
				}
			}
			// Contact not found
			else if (!conflictive_contact && !error){
				// TO-DO: Check if this is ok
				// Conflictive contact not found, maybe someone delete it?
				// Remove interaction conflict
				self.m_interaction.removeContactConflict(function(ok, error_msg){
					if (ok){
						if (typeof(callback) === 'function'){
							callback(true);
						}
					} else {
						self.show_message(format("<%=_T('Failed to remove conflict from conversation: {0}')%>", error_msg || ''), 'error');
						if (typeof(callback) === 'function'){
							callback(false);
						}
					}
				});
			}
			else {
				self.show_message(format("<%=_T('Failed to get conflictive contact: {0}, error: {1}')%>", conflict.contact_id, error && error.reason || ''), 'error');
				if (typeof(callback) === 'function'){
					callback(false);
				}
			}
		});
	};

	/**
	 * Will set the title to the current window dialog
	 * NOTE: Assume a created dialog.
	 *
	 * @private
	 * @param  {String} title
	 */
	ContactWindow.prototype.__update_window_title = function(title){
		this.m_dialog.dialog( 'option', 'title', title );
	};

	/**
	 * Get the title of the current dialog
	 * NOTE: Assume a created dialog.
	 *
	 * @private
	 * @return  {String} title
	 */
	ContactWindow.prototype.__get_window_title = function(){
		return this.m_dialog.dialog( 'option', 'title');
	};

	/**
	 * Called when a contact is successfully saved or merged.
	 *
	 * NOTE: This function will update the contact member
	 * with the new contact
	 *
	 * @private
	 * @param  {Object} contact
	 * @param  {Object} contact.Id
	 */
	ContactWindow.prototype.__on_contact_updated = function(contact)
	{
		var self = this;

		this.m_contact = contact;

		// set new title
		this.__update_window_title("<%=_T('Contact Information')%>");

		// update contact actions buttons
		this.m_main_basic_info.__get_contact_save_button().val("<%=_T('SAVE')%>");
		this.m_main_basic_info.__disable_save_create_contact();
		this.m_main_basic_info.__get_contact_cancel_button().val("<%=_T('Close')%>");
		this.__setup_contact_delete_handler();
		this.m_main_basic_info.show_contact_favorite_button();

		// update name and avatar
		this.m_main_basic_info.__get_contact_avatar().attr('src', contact.getCompleteAvatarURI());
		this.m_main_basic_info.__get_contact_name().text(contact.Name);

		// update contact id's
		var external_id = this.m_main_basic_info.__get_contact_external_id();
		if (contact.externalId){
			external_id.val(contact.externalId);
		} else {
			external_id.val("<%=_T('N/A')%>");
		}

		// TO-DO: This should be a method of contact basic info.
		// update addresses (remove old ones and add new ones)
		var $address_list = this.m_main_basic_info.__get_address_list();
		this.m_main_basic_info.__clean_addresses_list($address_list);

		if (contact.addresses.length !== 0) {
			this.m_main_basic_info.__hide_address_empty_message();
			contact.addresses.forEach(function(address){
				var address_node = self.m_main_basic_info.$build_address_node(address);
				$address_list.append(address_node);
			});
			this.m_main_basic_info.__setup_address_edition_handlers();
		} else {
			this.m_main_basic_info.__show_address_empty_message();
		}

		this.__update_contact_members(contact);
		this.__update_interaction_contact(contact);
		this.m_main_basic_info.setup_contact_save_handlers(contact);

		// publish update
		inConcert.app.messagebus.publish( '/application/contacts/contact_updated', contact );
	};

	/**
	 * Called when a contact is successfully deleted.
	 *
	 * @private
	 */
	ContactWindow.prototype.__on_contact_deleted = function()
	{
		var self = this;

		inConcert.app.messagebus.publish( '/application/contacts/contact_updated', this.m_contact );

		var confirm_options = {
			message: "<%=_T('Contact successfully deleted!')%>",
			confirm_text: "<%=_T('Close')%>",
			hide_cancel : true
		};

		this.__show_confirmation_dialog(confirm_options, {
			on_confirmed : function(){
				self.hide();
			}
		});
	};

	/**
	 * Will handle the Change/Choose existing one.
	 *
	 * @param  {Object} $change_view
	 * @param  {Object} $change_view_container
	 */
	ContactWindow.prototype.__handle_change_existing_contact = function($change_view, $change_view_container)
	{
		var self = this;

		var __move_back_to_contact_view = function(){
			$change_view.remove();
			self.m_main_basic_info.__toggle_contact_basic_info_containers(true);
			self.__update_window_title("<%=_T('Contact Information')%>");
		};

		var callbacks = {
			on_success : function(contact_selected){
				if (!self.m_interaction){
					return;
				}
				var confirm_options = { message: format("<%=_T('Are you sure? This will override the conversation contact with selected contact: {0}')%>", contact_selected.Name) };
				self.__show_confirmation_dialog(confirm_options, {
					on_confirmed : function(){
						self.__update_interaction_contact(contact_selected, function(ok){
							if (ok){
								self.__on_contact_updated(contact_selected);
								__move_back_to_contact_view();
							}
						});
					}
				});
			__move_back_to_contact_view();
			},
			on_canceled : function(){
				__move_back_to_contact_view();
			},
			on_back : function(){
				__move_back_to_contact_view();
			}
		};
		this.__switch_to_search_view($change_view_container, "<%=_T('Change Contact')%>", callbacks);
	};

	/**
	 * Will handle the Change/Create new contact view.
	 *
	 * @param  {Object} $change_view
	 * @param  {Object} $change_view_container
	 */
	ContactWindow.prototype.__handle_change_create_new = function($change_view, $change_view_container)
	{
		var self = this;
		// we create another instance of the basic info for our embedded purposes
		var basic_info_embedded = new inConcert.ContactWindowBasicInfo('embedded', this);

		// we build a basic contact info embedded on our container
		// just the addresses are necessary, all the other things needs to be editable.
		var sham_contact = this.m_contact_manager.create({
			isNew : true,  /* we kinda fake this contact */
			Name : '',
			addresses : this.m_contact_addresses,
			Id: inConcert.Utils.generateGuid()
		});
		this.m_contact = sham_contact;
		basic_info_embedded.build_contact_info_container(sham_contact, $change_view_container);

		var save_callbacks = {
			on_saved : function(ok){
				if (ok){
					$change_view.remove();
				}
			}
		};
		// Save action its a little bit complicated, because involves a confirmation
		// dialog when the user is saving the same address.
		basic_info_embedded.setup_contact_save_handlers(sham_contact, save_callbacks, {
			confirm_function : function(contact_to_be_saved, callback){

				self.show_message("<%=_T('Changing conversation contact, please wait...')%>", 'process');

				var using_same_address = false;
				var address_in_discord;
				contact_to_be_saved.addresses.forEach(function(address_to_be_saved){
					self.m_contact_addresses.forEach(function(address){
						if (address_to_be_saved.displayable === address.displayable){
							using_same_address = true;
							address_in_discord = address_to_be_saved.displayable;
						}
					});
				});

				// if the user have the same address, show a confirmation dialog
				// allowing us to remove the address from the previous contact
				if (using_same_address){
					var options = { message : "<%=_T('You are using the same address for this new contact, this will remove the address from the previous contact. Are you sure you want to continue?')%>" };
					self.__show_confirmation_dialog(options, {
						on_confirmed : function(){
							// we save the previous contact removing the address in discord
							self.m_contact.addresses.forEach(function(address, index){
								if (address_in_discord === address.displayable){
									// this modify the contact addresses array, remove the address.
									self.m_contact.addresses.splice(index, 1);
								}
							});

							var campaign_id = self.m_interaction ? self.m_interaction.campaign : undefined;
							self.m_contact.save(
								function(success /*, saved_contact, error */) {
									self.__update_interaction_contact(contact_to_be_saved, function(ok){
										if (ok){
											self.show_message("<%=_T('Conversation contact successfully changed!')%>", 'info', true);
											callback(success);
										}
									});
								},
								campaign_id, // optional, used by permissions
								self.get_app_id()
							);
						},
						on_canceled : function(){
							callback(false);
						}
					});
				} else {
					callback(true);
				}
			}
		});

		basic_info_embedded.setup_contact_cancel_handlers(function(){
			$change_view.remove();
		});

		basic_info_embedded.__setup_address_edition_handlers();
		basic_info_embedded.__setup_add_new_address_handlers();
		basic_info_embedded.__setup_name_edition_handlers();
		basic_info_embedded.__setup_external_id_edition_handlers();
		// TO-DO: Cancel?
	};

	/**
	 * Will setup the *CHANGE* contact handlers. Change means that this contact
	 * is not the contact I am interacting with, we can assign this interaction to
	 * a new contact or we can create a new contact with the information available.
	 *
	 * @private
	 */
	ContactWindow.prototype.__setup_contact_change_handlers = function()
	{
		var self = this;

		var $change_button = this.m_main_basic_info.show_contact_change_button();
		if ( !$change_button ) {
			return;
		}

		$change_button.click(function()
		{
			self.m_main_basic_info.__toggle_contact_basic_info_containers(false);
			self.__update_window_title("<%=_T('Change Contact')%>");

			var $change_view = $('#contact_window_change_view');
			if ($change_view.length > 0){
				$change_view.show();
				return;
			}

			var change_window_view = {
				create_new_label : "<%=_T('New')%>",
				choose_existing_label : "<%=_T('Choose Existing One')%>",
				disassociate_label : "<%=_T('Just Disassociate')%>",
				// TODO: this option should only be visible if the account does not have "auto add contact" enabled
				show_disassociate : false
			};
			$change_view = $(Mustache.render(self.m_templates.change_container_tpl, change_window_view));
			self.m_main_basic_info.__get_contact_window_container().append($change_view);

			var $view_container = $('#contact_window_change_container');
			var $options = $change_view.find('.optionRadiobutton');

			$options.change(function(){
				switch($(this).val()) {
					case 'existing_contact':
						$view_container.children().remove();
						self.__handle_change_existing_contact($change_view, $view_container);
						break;
					case 'new_contact':
						$view_container.children().remove();
						self.__handle_change_create_new($change_view, $view_container);
						break;
					case 'disassociate_contact':
						// TO-DO:
						break;
					default:
				}
			});
			// first render, trigger change on the selected option.
			$change_view.find('.optionRadiobutton:checked').trigger('change');
		});
	};

	/**
	 * Will switch to contact merge view.
	 *
	 * NOTE: This assume the other contact to be merged is the previously built contact.
	 *
	 * @private
	 *
	 * @param  {Object}   contact_source                         - The source contact to be merged (normally, contact created)
	 * @param  {String}   contact_source.Name                    - Source contact's name (normally, contact created)
	 * @param  {String}   contact_source.externalId              - Source contact's external id (normally, contact created)
	 * @param  {Array}    contact_source.addresses               - Source contact's addresses (normally, contact created)
	 *
	 * @param  {Object}   contact_target                         - The contact to be merged
	 * @param  {String}   contact_target.Name                    - Contact's name
	 * @param  {String}   contact_target.externalId              - Contact's external id
	 * @param  {Array}    contact_target.addresses               - Contact's addresses
	 *
	 * @param  {Object}   [visual_settings]                      - Optional visual settings
	 * @param  {Object}   [visual_settings.addresses_disabled]   - Adding/Removing addresses disabled or not
	 * @param  {Object}   [visual_settings.name_disabled]        - Name choosing disabled or not
	 * @param  {Object}   [visual_settings.externalid_disabled]  - External Id choosing disabled or not
	 *
	 * @param  {Object}   [callbacks]                            - A bunch of optionals callbacks
	 * @param  {Function} [callbacks.on_back]                    - Called when the user hit "Back" button, optional callback.
	 * @param  {Function} [callbacks.on_canceled]                - Called when the user hit "Cancel" button, optional callback.
	 * @param  {Function} [callbacks.on_success]                 - Called when the merge is successful.
	 *
	 * @param  {Object}   [$container]                           - Optional container where we are going to append the search view
	 *                                                             default its contact window container
	 */
	ContactWindow.prototype.__switch_to_merge_view = function(contact_source, contact_target, visual_settings, callbacks, $container)
	{
		var self = this;
		var contacts_are_the_same = contact_source.Id === contact_target.Id;
		var visual_settings_enabled = !_.isEmpty(visual_settings);

		this.__update_window_title("<%=_T('Merge Contacts')%>");

		// override container if it's not passed as argument
		$container = $container ? $container : this.m_main_basic_info.__get_contact_window_container();

		// show external id container if at least
		// one of the contacts has external id.
		var show_external_id = (contact_target.externalId && contact_target.externalId.length > 0) || contact_source.externalId && contact_source.externalId.length > 0;

		var no_external_id = "<%=_T('Empty')%>";

		// build basic markup old/new contacts
		$container.append(Mustache.render(this.m_templates.merge_contacts_tpl, {
			// translations
			new_label : "<%=_T('NEW')%>",
			contact_name_label : "<%=_T('Contact Name')%>",
			contact_addresses_label : "<%=_T('Contact Addresses')%>",
			contact_id_label : "<%=_T('Contact ID')%>",
			merge_label : "<%=_T('MERGE')%>",
			cancel_label : "<%=_T('CANCEL')%>",
			back_label : "<%=_T('BACK')%>",
			// actual contacts information
			name_new : contact_target.Name,
			name_old : contact_source.Name,
			show_external_id_container : show_external_id,
			external_id_new : contact_target.externalId || no_external_id,
			external_id_old : contact_source.externalId || no_external_id,
			// visual settings
			name_disabled : visual_settings_enabled ? visual_settings.name_disabled : false,
			id_disabled : visual_settings_enabled ? visual_settings.id_disabled : false,
		}));

		var $merge_address_list = $('#contact_window_merge_address_list');
		var address_tpl = self.m_templates.merge_contact_address_tpl;

		var added = 0;
		// append current contact addresses
		contact_source.addresses.forEach(function(address){
			const displayable = self.__GetDisplayableAddresss(address);
			const address_normalized = nid(displayable);
			if ($('#contact_window_merge_address_' + address_normalized).length === 0) {
				const $address = Mustache.render(address_tpl, {
					address_normalized : address_normalized,
					address: displayable,
					address_type : address.Type,
					disabled: visual_settings_enabled ? visual_settings.addresses_disabled : false
				});
				$merge_address_list.append($address);
				added++;
			}
		});

		// append contact to merge addresses
		contact_target.addresses.forEach(function(address){
			const displayable = self.__GetDisplayableAddresss(address);
			const address_normalized = nid(displayable);
			if ($('#contact_window_merge_address_' + address_normalized).length === 0 ){
				$merge_address_list.append(Mustache.render(address_tpl, {
					address_normalized : address_normalized,
					address: displayable,
					address_type: address.Type,
					new_label : "<%=_T('NEW')%>",
					is_new : true,
					disabled: visual_settings_enabled ? visual_settings.addresses_disabled : false
				}));
				added++;
			}
		});

		if (added > 0) {
			this.m_main_basic_info.__hide_address_empty_message();
		}

		var __move_back_to_contact_view = function(){
			self.__get_contact_merge_container().remove();
			self.__get_contact_search_container().hide();
			self.__clean_contact_search_view();
		};

		// CANCEL
		$('#contact_window_merge_cancel').click(function(){
			__move_back_to_contact_view();
			if (callbacks.on_canceled && typeof(callbacks.on_canceled) === 'function'){
				callbacks.on_canceled();
			}
		});

		// BACK
		$('#contact_window_merge_back').click(function(){
			self.__get_contact_merge_container().remove();
			if (typeof(callbacks.on_back) === 'function'){
				callbacks.on_back();
			}
		});

		// MERGE
		$('#contact_window_merge_save').click(function(){
			var contact_already_exists = self.m_contact && !self.m_contact.isNew;
			var final_contact = contact_already_exists ? self.m_contact : contact_target;

			// update current contact information
			final_contact.Name = $.trim($('input[name=radioName]:radio:checked').next().text());
			var external_id = $.trim($('input[name=radioID]:radio:checked').next().text());
			if (external_id.length === 0){
				delete final_contact.externalId;
			} else {
				final_contact.externalId = external_id;
			}

			// get checked addresses
			final_contact.addresses = [];
			$('.mergeCheckbox:checked').each(function(undefined, e){
				var $address_label_node = $(e).next();
				var address = $address_label_node.attr('address');
				var address_type = $address_label_node.attr('address_type');
				var address_protocol = $address_label_node.attr('protocol');
				var address_normalized = self.m_contact_manager.normalizeContactAddress(address, address_type, address_protocol);
				final_contact.addresses.push(address_normalized);
			});

			var __on_merge_success = function(merged_contact){
				self.show_message("<%=_T('Contacts successfully merged')%>", 'info', true);
				self.__on_contact_updated(merged_contact);
				__move_back_to_contact_view();
			};

			var __on_merge_finished = function(merged_contact, error)
			{
				if (merged_contact){
					if (self.m_interaction){

						var __set_contact_callback = function( success_set, error ) {
							if (success_set){
								__on_merge_success(merged_contact);
							} else {
								self.show_message(format("<%=_T('Unable to associate contact: {0} to conversation: {1}')%>", merged_contact.Name, error && error.message), 'error');
							}
						};
						var app_id = self.get_app_id();

						var __set_interaction_contact = function(){
							self.m_interaction.setContact(
								merged_contact.Id,
								merged_contact.Name,
								__set_contact_callback,
								app_id
							);
						};

						// we update the interaction contact without asking
						// if the interaction does not have events
						if (!self.m_interaction.getEventList()){
							__set_interaction_contact(merged_contact);
						} else {
							self.__show_confirmation_dialog({
								message: "<%=_T('Do you want to update the conversation contact with the merged one?')%>",
								confirm_text: "<%=_T('Yes')%>",
								cancel_text: "<%=_T('No')%>"
							},
							{
								on_confirmed : function(){
									__set_interaction_contact(merged_contact);
								},
								on_canceled : function(){
									__on_merge_success(merged_contact);
								}
							});
						}
					} else {
						__on_merge_success(merged_contact);
					}
					if (typeof(callbacks.on_success) === 'function'){
						callbacks.on_success();
					}
				} else {
					if ( typeof error === 'object' ) {
						error = error.reason;
					}
					var message_error;
					switch( error ) {
						case 'forbidden':
							message_error = "<%=_T('You not has permission to merge contact.')%>";
							break;
						default:
							message_error = format("<%=_T('Error merging contacts: {0}')%>", error && error.reason ? error.reason : "<%=_T('internal error')%>");
					}
					self.show_message( message_error, 'error' );
				}
			};

			// async operation ahead, show waiting message
			self.show_message("<%=_T('Merging process started, please wait...')%>", 'process');

			var campaign_id = self.m_interaction ? self.m_interaction.campaign : undefined;
			// merge only happens between two different contacts
			// otherwise it's just a contact save
			const app_id = self.get_app_id();
			if (contact_already_exists && !contacts_are_the_same){
				// on merge, with take the current contact id as the new one.
				final_contact.Id = self.m_contact.Id;

				// go and merge this contacts
				// NOTE: contact_target.Id will be the obsolete contact
				self.m_contact_manager.merge(
					final_contact,
					contact_target.Id,
					function( merged_contact, error ){
						__on_merge_finished(merged_contact, error);
					},
					campaign_id, // optional, used by permissions,
					app_id
				);
			} else {
				final_contact.save(
					function(undefined /* success */ , saved_contact, error) {
						__on_merge_finished(saved_contact, error);
					},
					campaign_id, // optional, used by permissions
					app_id
				);
			}
		});
	};

	/**
	 * Will build the mergeable contact list,
	 * this function is called from the search contact window.
	 *
	 * @private
	 * @async
	 *
	 * @param  {String}   search_query - The search query inserted by the user
	 * @param  {String}   search_type  - The search type (i.e. 'name' or 'mail')
	 * @return {Function} [callback]   - An optional callback function
	 */
	ContactWindow.prototype.__build_search_contacts_lists = function(search_query, search_type, callback)
	{
		var self = this;
		var search_name = '';
		var search_external_id = '';
		var search_addresses = [];

		var $search_list = this.__get_search_list_container();
		var $associate_button = $('#contact_window_search_associate');

		// TO-DO: Get this options from somewhere else.
		switch(search_type) {
			case 'name':
				search_name = search_query;
				break;
			case 'external_id':
				search_external_id = search_query;
				break;
			case 'Phone':
			case 'Mail':
			case 'IM':
			case 'Twitter':
			case 'Facebook':
			case 'Instagram':
			case 'Youtube':
			case 'Appstore':
			case 'Googleplay':
				search_addresses.push(this.m_contact_manager.normalizeContactAddress(search_query, search_type));
				/* falls through */
		}

		// go and find contacts
		this.__search_contacts(search_name, search_external_id, search_addresses, function(contacts, error){
			if (contacts && contacts.length > 0){
				// generate the contact list
				contacts.forEach(function(contact){
					// make sure we exclude the current contact from the list
					if (self.m_contact && self.m_contact.Id !== contact.Id){
						var $contact_node = $(Mustache.render(self.m_templates.contact_item, {
							contact_avatar_uri : contact.getCompleteAvatarURI(),
							contact_name : contact.Name
						}));
						$contact_node.data('contact', contact);
						$search_list.append($contact_node);
					}
				});

				// if the contact that comes from the server is
				// the same the user is showing on the window, the list
				// will be empty, so show a message.
				if ($search_list.children().length === 0){
					$search_list.append(Mustache.render(self.m_templates.search_no_contact_msg_tpl, {
						message : "<%=_T('No matching contact found. We are excluding the current contact from the list.')%>"
					}));
				}

				var selectContact = function($row) {
					// enable associate button
					$associate_button.prop('disabled', false);
					$row.addClass('selected').siblings().removeClass('selected');
				};

				// bind selected handlers
				$search_list.find('.resultsListItem')
					.click(function() {
						selectContact($(this));
					})
					.dblclick(function() {
						selectContact($(this));
						$associate_button.trigger('click');
					});
			}
			else if (!contacts || contacts.length === 0 && !error){
				$search_list.append(Mustache.render(self.m_templates.search_no_contact_msg_tpl, {
					message : "<%=_T('No matching contact found.')%>"
				}));
			}
			else {
				self.show_message(format("<%=_T('Error searching for contacts: {0}')%>", error && error.message ? error.message : error), 'error');
			}
			if (typeof(callback) === 'function'){
				callback();
			}
		});
	};

	/**
	 * Will clean the search/merge contact list
	 *
	 * @private
	 * @param  {Object} [search_list]  - Optional cached jQuery object
	 * @param  {Object} [search_input] - Optional cached jQuery object
	 */
	ContactWindow.prototype.__clean_contact_search_view = function(search_list, search_input)
	{
		var $search_list = search_list ? search_list : this.__get_search_list_container();
		var $search_input = search_input ? search_input : this.__get_contact_search_input() ;

		// update all the things
		$search_list.children().remove();
		$search_input.val('');
	};

	/**
	 * Will create a new search view.
	 *
	 * @todo Change parameters to an object, we lose visibility from outside the function
	 *
	 * @private
	 * @param  {Object}   $container               - Container where we are going to append the search view.
	 *                                              default its contact window container
	 *
	 * @param  {String}   main_action_name         - Action name for the main button, called after the user selects a contact
	 *
	 * @param  {Object}   callbacks                - An object containing callbacks
	 * @param  {Function} callbacks.on_success     - Called when the contact is selected and the main action is pressed (merge/associate)
	 * @param  {Function} [callbacks.on_back]      - Called when the user hit "Back" button, optional callback.
	 * @param  {Function} [callbacks.on_canceled]  - A callback when the user press cancel, it's optional.
	 */
	ContactWindow.prototype.__switch_to_search_view = function($container, main_action_name, callbacks)
	{
		var self = this;

		// override container if it's not passed as argument
		$container = $container ? $container : this.m_main_basic_info.__get_contact_window_container();

		// remove previous searches
		var $previous_search_container = this.__get_contact_search_container();
		if ($previous_search_container.length > 0){
			$previous_search_container.remove();
		}

		// create search window
		var $search_container = $(Mustache.render(this.m_templates.search_container_tpl, {
			search_label : "<%=_T('SEARCH')%>",
			search_associate_label : main_action_name,
			search_cancel_label : "<%=_T('Cancel')%>",
			search_by_name : "<%=_T('Name')%>",
			search_by_address : "<%=_T('Address')%>",
			search_by_phone : "<%=_T('Phone Number')%>",
			search_by_id : "<%=_T('External ID')%>",
			search_by_mail : "<%=_T('E-mail')%>",
			search_by_twitter : "<%=_T('Twitter username')%>",
			search_by_instagram : "<%=_T('Instagram username')%>",
			search_by_youtube : "<%=_T('Youtube username')%>",
			search_by_appstore : "<%=_T('Appstore username')%>",
			search_by_googleplay : "<%=_T('Googleplay username')%>",
			search_by_facebook : "<%=_T('Facebook UID')%>",
			merge_title:  "<%=_T('Merge Contacts')%>"
		}));
		$container.append($search_container);

		var $search_input = this.__get_contact_search_input();
		var $search_options = $('#contact_window_search_options');
		var $search_button = $('#contact_window_search_button');
		var $search_list = this.__get_search_list_container();
		var $associate_button = $('#contact_window_search_associate');
		var $cancel_button = $('#contact_window_search_cancel');

		// focus cursor inside search
		$search_input.focus();

		// MERGE/ASSOCIATE INSIDE SEARCH
		$associate_button.click(function(){
			// hide search container
			$search_container.hide();
			// switch to merge view
			var contact_selected = $search_list.children('.selected').data('contact');
			// callback with the selected contact
			callbacks.on_success(contact_selected);
		});

		// CANCEL
		$cancel_button.click(function(){
			// hide the container and clean the search
			$search_container.hide();
			self.__clean_contact_search_view($search_list, $search_input);
			if (callbacks.on_canceled && typeof(callbacks.on_canceled) === 'function'){
				callbacks.on_canceled();
			}
		});

		const $merge_title_container = $("#contact_window_merge_title_container");
		// back button
		const $back = $merge_title_container.find("[action_back]");
		$back.click(function() {
			$cancel_button.trigger("click");
		});

		// OPTIONS CHANGE
		$search_options.change(function(){
			$search_input.val('');
			$search_list.children().remove();
		});

		// TO-DO: This kind of behavior is used a lot
		// we need to make something more generic to handle searches
		// like a jQuery plugin or something.
		$search_input.keyup(function(event){
			var search_term = $search_input.val();
			$search_button.prop('disabled', !search_term);
			if (search_term){
				// Enter trigger search
				if (event.keyCode === 13) {
					$search_button.click();
					return;
				}
				//Escape reset search
				if (event.keyCode === 27) {
					$search_list.children().remove();
					return;
				}
			} else {
				// backspace and no text, reset search.
				if (event.keyCode === 8 || event.keyCode === 46) {
					$search_list.children().remove();
					return;
				}
			}
		});

		// SEARCH
		$search_button.click(function(){
			// async operation ahead, show waiting
			self.__show_waiting();
			// remove previous results
			$search_list.children().remove();
			// build mergeable contact list
			var search_query = $search_input.val();
			var selected_option = $search_options.val();
			self.__build_search_contacts_lists(search_query, selected_option, function(){
				// all done, hide waiting.
				self.__hide_waiting();
			});
		});
	};

	/**
	 * Will setup the contact search feature, this is done
	 * to search contacts for merge or for other actions
	 *
	 * @private
	 */
	ContactWindow.prototype.__setup_contact_merge_handlers = function()
	{
		var self = this;

		var $merge_button = this.m_main_basic_info.show_contact_merge_button();
		if ( !$merge_button ) {
			return;
		}

		$merge_button.click(function(){
			self.__update_window_title("<%=_T('Search Contacts to Merge')%>");
			self.__switch_to_search_view(self.m_main_basic_info.__get_contact_window_container(), "<%=_T('Merge preview')%>", {
				on_success : function(contact_selected){
					self.__switch_to_merge_view(self.m_contact, contact_selected, {}, {
						on_success: function(){
							self.__update_window_title("<%=_T('Contact Information')%>");
						}
					});
				},
				on_canceled : function(){
					self.__update_window_title("<%=_T('Contact Information')%>");
				},
				on_back : function(){
					self.__get_contact_search_container().show();
					self.__update_window_title("<%=_T('Search Contacts')%>");
				}
			});
		});
	};

	/**
	 * Handling contact deletion
	 *
	 * @todo Move this to basic info
	 */
	ContactWindow.prototype.__setup_contact_delete_handler = function()
	{
		var self = this;

		var $delete_button = this.m_main_basic_info.show_contact_delete_button();
		if ( !$delete_button ) {
			return;
		}
		var campaign_id = this.m_interaction ? this.m_interaction.campaign : undefined;

		$delete_button.unbind('click').click(function(){
			self.__show_confirmation_dialog({ message : format("<%=_T('Are you sure you want to delete contact: {0}?')%>", self.m_contact_name) }, {
				on_confirmed : function(){
					self.m_contact.remove(
						function(ok, error) {
							if (ok){
								self.show_message("<%=_T('Contact successfully deleted')%>", 'info', true);
								// TO-DO: Deleting the confirmation is not responsibility of this function
								// seems really wrong, maybe the dialog can pass a function to the confirm callback? yo dawg.
								self.__get_confirmation_dialog().remove();
								self.__on_contact_deleted();
								return;
							}
							var message_error;
							switch ( error.result.toLowerCase() ) {
								case 'forbidden':
									message_error = "<%=_T('You not has permission to delete contact.')%>";
									break;
								default:
									message_error = format("<%=_T('There was an error deleting contact: {0}, please try again later')%>", self.m_contact.Name);
							}
							self.show_message( message_error, 'error');
						},
						campaign_id // optional, used by permissions
					);
				}
			});
		});
	};

	/**
	 * Will setup all event bindings for the first window
	 *
	 * @private
	 * @param  {Object} contact
	 */
	ContactWindow.prototype.__setup_main_event_bindings = function(contact)
	{
		var self = this;
		this.__setup_contact_merge_handlers(contact);
		this.__setup_contact_change_handlers(contact);
		this.__setup_contact_delete_handler();

		this.m_main_basic_info.setup_contact_save_handlers(contact, {
			on_saved : function(success, saved_contact){
				// TO-DO: Check if this is completely correct, I think we should check
				// if all the events (i.e. mail) are from the same contact, update the interaction contact.
				//
				// We are going to change the contact if:
				// - The interaction does not have events (sync)
				// - The interaction is not finished.
				if (success && self.m_interaction && !self.m_interaction.getEventList() && !self.m_interaction.isFinished()){

					var set_contact_callback = function( success_set, error ) {
						if (!success_set){
							self.show_message(format("<%=_T('Unable to associate contact: {0} to conversation: {1}')%>", saved_contact.Name, error && error.message), 'error');
						}
					};
					var app_id = self.get_app_id();

					self.m_interaction.setContact(
						saved_contact.Id,
						saved_contact.Name,
						set_contact_callback,
						app_id
					);
				}
			}
		});

		this.m_main_basic_info.setup_contact_cancel_handlers(function(){
			self.hide();
		});

		this.m_main_basic_info.__setup_contact_fav_handlers(contact);
		this.m_main_basic_info.__setup_create_outgoing_handlers(contact);
		this.m_main_basic_info.__setup_address_edition_handlers();
		this.m_main_basic_info.__setup_add_new_address_handlers(contact);
		this.m_main_basic_info.__setup_name_edition_handlers(contact);
		this.m_main_basic_info.__setup_external_id_edition_handlers(contact);
	};

	ContactWindow.prototype.__update_contact_members = function(contact)
	{
		if (!contact){
			return;
		}
		this.m_contact_name = contact.Name;
		this.m_contact_external_id = contact.externalId;
		this.m_contact_addresses = contact.addresses;
		this.m_contact = contact;
	};

	/**
	 * Search contact based on a few filters
	 *
	 * @todo  Change the parameters for an object with properties
	 *
	 * @private
	 * @param  {String}   contact_id
	 * @param  {String}   contact_name
	 * @param  {Array}    contact_addresses
	 * @param  {Function} callback
	 */
	ContactWindow.prototype.__search_contacts = function(contact_name, external_id, contact_addresses, callback)
	{
		var self = this;

		/**
		 * handle callbacks and errors
		 *
		 * @param  {Array} contacts
		 * @param  {Object} error   [description]
		 */
		var __finish_him = function(contacts, error){
			if (contacts && contacts.length > 0) {
				callback(contacts);
			} else if (contacts && contacts.length === 0){
				callback();
			} else {
				self.show_message(format("<%=_T('Unable to get contacts from server: {0}')%>", error.message), 'error');
				callback(null, error);
			}
		};

		if (external_id && external_id.length > 0){
			this.m_contact_manager.getContactByExternalId(external_id, function(contact, error){
				__finish_him(contact ? [contact] : [], error);
			});
		}
		else if (contact_addresses && contact_addresses.length > 0){
			this.m_contact_manager.getContactByAddress(contact_addresses[0], function(contact, error){
				__finish_him(contact ? [contact] : [], error);
			});
		}
		else if (contact_name && contact_name.length > 0){
			// this search will hit the index.
			var filter = {
				query : contact_name,
				last : '',
				rows : this.m_contacts_per_page
			};
			this.m_contact_manager.search(filter, function(contacts, undefined /* lastContact */ , error) {
				__finish_him(contacts, error);
			});
		}
	};

	ContactWindow.prototype.__can_modify = function(interaction)
	{
		interaction = interaction || this.m_interaction;
		if (!interaction || !interaction.interactionData) {
			return false;
		}
		if (!interaction.belongsToMe()) {
			return false;
		}
		return true;
	};

	ContactWindow.prototype.__can_merge = function(interaction)
	{
		interaction = interaction || this.m_interaction;
		if (!this.__can_modify(interaction)) {
			return false;
		}
		return true;
	};

	/**
	 * Will check if an interaction has conflict, if so, we
	 * are going to handle them, could be a field clash conflict or a
	 * multiple match conflict
	 *
	 * @private
	 * @param  {Object} interaction
	 * @param  {Object} contact
	 */
	ContactWindow.prototype.__handle_interaction_conflict = function(interaction /* , contact */)
	{
		var self = this;

		if (!this.__can_merge(interaction)) {
			return;
		}
		if (!interaction.interactionData.conflict) {
			return;
		}
		var conflict = interaction.interactionData.conflict;
		/**
		 * Final callback of conflict resolution
		 *
		 * @private
		 * @param  {Boolean} ok - If the conflict went well
		 */
		var __on_conflict_solved = function(ok){
			if (ok){
				self.m_interaction.removeContactConflict(function(success_remove_conflict, error_msg){
					if (success_remove_conflict){
						var $conflict_window = self.__get_conflict_view();
						if ($conflict_window){
							$conflict_window.remove();
						}
						self.__update_window_title("<%=_T('Conflict Information')%>");
						self.show_message("<%=_T('Congrats! All conflicts has been resolved')%>", 'info', true);
					} else {
						self.show_message(format("<%=_T('Failed to remove conflict from conversation: {0}')%>", error_msg || ''), 'error');
						self.__get_conflict_view.show();
					}
				});
			}
		};

		var default_options = {
			allow_merge : true,
			allow_cancel : false,
			allow_create_new : true,
			allow_move : false,
		};

		this.__update_window_title("<%=_T('Conflict Detected')%>");

		if (conflict.type === 'field_clash'){
			var options = {
				message : "<%=_T('A conflict has been detected merging this contact information, how do you want to proceed?')%>",
				contact_id : interaction.contact,
				contact_external_id : this.__get_interaction_external_id() || "<%=_T('N/A')%>",
				field : conflict["new"],
				field_type : conflict.field,
				addresses : this.__get_interaction_addresses()
			};
			this.__show_conflict_window( _.extend(default_options, options), __on_conflict_solved );
		}
		else if (conflict.type === 'multiple_matches'){
			var contacts_matches = conflict.matching_contacts.slice(0);
			var total_conflicts = contacts_matches.length;

			var __resolve_multiple_conflicts = function(){
				var current_conflictive_id = contacts_matches.pop();
				var conflicts_left = contacts_matches.length + 1;

				if (!current_conflictive_id){
					return __on_conflict_solved(true);
				}
				self.m_contact_manager.getContact(current_conflictive_id, function(contact, error){
					if (contact){
						var conflict_field, conflict_field_type;
						self.__get_interaction_addresses().forEach(function(address){
							contact.addresses.forEach(function(conflictive_address){
								if (address.displayable === conflictive_address.displayable){
									conflict_field = address.displayable;
									conflict_field_type = 'address';
								}
							});
						});
						var options = {
							message : format("<%=_T('{0} of {1} {2} left to be solved, how do you want to proceed?')%>", conflicts_left, total_conflicts, total_conflicts > 1 ? 'conflicts' : 'conflict'),
							contact_id : contact.Id,
							contact_external_id : self.__get_interaction_external_id() || contact.externalId || "<%=_T('N/A')%>",
							field : conflict_field,
							field_type : conflict_field_type,
							multiple_conflicts : true,
							addresses : self.m_interaction.interactionData.contactAddresses
						};
						self.__show_conflict_window( _.extend(default_options, options), __resolve_multiple_conflicts );
					}
					else if (!contact && !error){
						// contact not found, move along.
						__resolve_multiple_conflicts();
					} else {
						self.show_message(format("<%=_T('Failed to get contact: {0}, error: {1}')%>", current_conflictive_id, error.message), 'error');
					}
				});
			};
			__resolve_multiple_conflicts();
		}
	};

	/**
	 * Return interaction addresses
	 *
	 * @private
	 *
	 * @return {Array}
	 */
	ContactWindow.prototype.__get_interaction_addresses = function()
	{
		const data = this.m_interaction.interactionData || {};
		if (data.contactAddresses) {
			return data.contactAddresses;
		}
		if (data.completeAddress) {
			return [data.completeAddress];
		}
		return [];
	};

	/**
	 * Return interaction addresses
	 *
	 * @private
	 *
	 * @return {Array}
	 */
	ContactWindow.prototype.__get_interaction_external_id = function()
	{
		const data = this.m_interaction.interactionData || {};
		return data.gatheredContactData?.external_id;
	};

	/**
	 * Build a fake contact from info available on interaction
	 *
	 * @private
	 *
	 * @return {Object}
	 */
	ContactWindow.prototype.__build_sham_contact = function()
	{
		var contact = this.m_contact_manager.create({
			isNew : true,  /* internal flag, we kinda fake this contact */
			Name : this.m_contact_name || '',
			addresses : this.m_contact_addresses,
			Id: inConcert.Utils.generateGuid(),
			Company: ''
		});
		if (this.m_contact_external_id){
			contact.externalId = this.m_contact_external_id;
		}
		if (this.m_contact_avatar){
			contact.avatar = this.m_contact_avatar;
		}
		return contact;
	};

	/**
	 * It will show the contact information window, if we have a valid contact we will try to use it,
	 * but if we only have the address of the client, it will show useful info too.
	 *
	 * @public
	 * @async
	 */
	ContactWindow.prototype.show = function()
	{
		var self = this;
		var interaction = this.m_interaction;

		/**
		 * Called on every async callback, it will end the contact
		 * view building.
		 *
		 * @param  {Object} contact
		 * @param  {String} title - Contact window title
		 */
		var __finish_him = function(contact, title) {
			self.m_main_basic_info.build_contact_info_container(contact);
			self.__setup_main_event_bindings(contact);
			self.__create_window();
			self.__update_window_title(title);
			self.__update_contact_members(contact);
			self.__handle_interaction_conflict(interaction, contact);
		};

		// if the user give us a contact id, we
		// use it before anything else.
		if (this.m_contact_id) {
			this.m_contact_manager.getContact(interaction.contact, function(contact, error){
				if (contact){
					self.m_contact = contact;
					__finish_him(contact, "<%=_T('Contact Information')%>");
				}
				// contact not found
				else if (!contact && !error){
					self.m_contact = self.__build_sham_contact();
					__finish_him(self.m_contact, "<%=_T('New contact')%>");
				}
				else {
					self.show_message(format("<%=_T('Failed to search contact by address, error: {1}')%>", error.message));
				}
			});
		} else {
			if (this.m_contact_addresses.length === 0){
				// No addresses? show the minimal info available.
				this.m_contact = this.__build_sham_contact();
				__finish_him(this.m_contact, "<%=_T('New contact')%>");
			} else {
				// first and second parameter are empty strings because I don't want to
				// search by name or external id, just by address in this case.
				this.__search_contacts('', '', self.m_contact_addresses, function(contacts, error){
					if (contacts) {
						// TO-DO: Handle the case when multiple contacts are returned
						self.m_contact = contacts[0];
						__finish_him(self.m_contact, "<%=_T('Contact Information')%>");
					}
					else if (!contacts || contacts.length === 0 && !error){
						self.m_contact = self.__build_sham_contact();
						__finish_him(self.m_contact, "<%=_T('New contact')%>");
					}
					else {
						self.show_message(format("<%=_T('Failed to search contact by address, error: {1}')%>", error.message));
					}
				});
			}
		}
	};

	/**
	 * Will remove the contact window.
	 *
	 * @public
	 */
	ContactWindow.prototype.hide = function()
	{
		this.m_main_basic_info.get_contact_window().remove();
	};

	ContactWindow.prototype.is_new_contact = function()
	{
		if ( this.m_contact === null ) {
			return true;
		}
		return this.m_contact.isNew;
	};

	ContactWindow.prototype.get_app_id = function()
	{
		return this.m_from_app || 'interaction';
	};

	ContactWindow.prototype.__has_permission = function( action_type )
	{
		var actions_by_apps = {
			VIEW     : { interaction: 'int_contacts_mng'     , supervisor: 'sup_contacts_mng'     } ,
			SAVE     : { interaction: 'int_save_contact'     , supervisor: 'sup_save_contact'     } ,
			DELETE   : { interaction: 'int_delete_contact'   , supervisor: 'sup_delete_contact'   } ,
			FAVORITE : { interaction: 'int_favorite_contact' , supervisor: 'sup_favorite_contact' } ,
			// MERGE Y CHANGE, utilizan el mismo id que SAVE
			MERGE    : { interaction: 'int_save_contact'     , supervisor: 'sup_save_contact'     } ,
			CHANGE   : { interaction: 'int_save_contact'     , supervisor: 'sup_save_contact'     }
		};

		if ( !actions_by_apps[ action_type ] ) {
			return false;
		}
		var app_id = this.get_app_id();
		var action_id = actions_by_apps[ action_type ][ app_id ];
		if ( !action_id ) {
			return false;
		}

		var campaign_id = this.m_interaction.campaign;
		return inConcert.permissions.to_action( app_id, action_id, undefined /* category */, campaign_id, true /* no log */ );
	};

	ContactWindow.prototype.has_permission_to_view = function() {
		return this.__has_permission( 'VIEW' );
	};

	ContactWindow.prototype.has_permission_to_save = function() {
		if ( !this.__can_modify() ) { return false; }
		return this.__has_permission( 'SAVE' );
	};

	ContactWindow.prototype.has_permission_to_merge = function() {
		if ( !this.__can_merge() ) { return false; }
		return this.__has_permission( 'MERGE' );
	};

	ContactWindow.prototype.has_permission_to_change = function() {
		if ( !this.__can_modify() ) { return false; }
		return this.__has_permission( 'CHANGE' );
	};

	ContactWindow.prototype.has_permission_to_delete = function() {
		if ( !this.__can_modify() ) { return false; }
		return this.__has_permission( 'DELETE' );
	};

	ContactWindow.prototype.has_permission_to_favorite = function() {
		if ( !this.__can_modify() ) { return false; }
		return this.__has_permission( 'FAVORITE' );
	};
}());
