/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:false, browser:true, indent:false, maxerr:50, quotmark:double, esversion: 11 */

/*global $, inConcert, _ */

/*

Here you'll find contact management related stuff such as:
+ Searching & Finding contacts
+ Administration (create, update, merge, delete)
+ Getting contact information and specific profiles (facebook, twitter)
+ Contact history and past interactions

Exposes a single object inConcert.ContactManager where everything else derives from.

*/

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (!this.inConcert) {
	this.inConcert = {};
}

(function () {
	"use strict";

	var s_baseUrl;

	//Base constructor for Contact
	function Contact(rawContact) {
		$.extend(true, this, rawContact);
	}

	/*
		Saves the current contact
	*/
	Contact.prototype.save = function( onContactSaved, campaign_id, app_id = "interaction" )
	{
		inConcert.logger.action({action_name : "save_contact", id : this.Id, action_type:"contact"});

		var self = this;
		delete this.isNew;

		var cloned_contact = inConcert.Utils.cloneObject( this );
		cloned_contact.__campaign_id = campaign_id; // optional, used by permissions
		cloned_contact.__app_id = app_id;

		var url = s_baseUrl + "/contacts/save";
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify( cloned_contact ),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				self.isNew = false;
				var contact = new Contact(response);
				if (typeof(onContactSaved) === "function") {
					onContactSaved(true, contact);
				}
			},
			error: function(undefined, error) {
				if (typeof(onContactSaved) === "function") {
					onContactSaved(false, null, error);
				}
			}
		});
	};

	/*
		Retrieves all the archived interactions for a specific contact, paginating from lastKey
	*/
	Contact.prototype.getArchive = function(filter, onArchiveLoadedFn)
	{
		inConcert.logger.action({action_name : "get_contact_archive", id : this.Id, action_type:"contact"});
		var contactFilter = $.extend({}, filter);
		contactFilter.query.contact = this.Id;
		inConcert.app.getInteractionManager().searchArchiveInteractions(contactFilter, onArchiveLoadedFn);
	};


	/*
		Deletes the current contact
			campaign_id is optional, used by permissions
	*/
	Contact.prototype.remove = function( onContactRemoved, campaign_id, app_id = "interaction" )
	{
		inConcert.logger.action({action_name : "delete_contact", id : this.Id, action_type:"contact"});
		onContactRemoved = onContactRemoved || $.noop;

		var request = {
			id : this.Id,
			campaign_id:  campaign_id, // optional, used by permissions
			app_id : app_id
		};

		var url = s_baseUrl + "/contacts/delete";
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				if (typeof(onContactRemoved) === "function") {
					onContactRemoved(response);
				}
			},
			error: function(undefined, error) {
				if (typeof(onContactRemoved) === "function") {
					onContactRemoved( false, error);
				}
			}
		});
	};

	/*
		Returns the correctly formed avatar URI for this contact
	*/
	Contact.prototype.getAvatarURI = function() {
		return "/contacts/get_contact_avatar/" + this.Id;
	};

	/**
	 * Get the complete url avatar, normally used on src images.
	 *
	 * @return {String} The url
	 */
	Contact.prototype.getCompleteAvatarURI = function() {
		return inConcert.app.getBaseUrl() + "/contacts/get_contact_avatar/" + this.Id;
	};

	//Returns the display-able text for a contact address
	// TODOCONTACT: ver para que se usa esto
	Contact.prototype.getAddressDisplayText = function(address, fullPhoneNumber) {
		switch (address.Type) {
			case "Phone":
				if (address.GlobalNumber) return address.GlobalNumber;

				let phoneNumber = address.Number;
				if (!fullPhoneNumber) return phoneNumber;

				if (address.Area) {
					phoneNumber = address.Area + phoneNumber;
				}
				if (address.Country) {
					phoneNumber = address.Country + phoneNumber;
					if (!inConcert.app.has_allegro_integration()) {
						phoneNumber = "+" + phoneNumber;
					}
				}
				return phoneNumber;
			case "Facebook":
				return address.UID;
			case "Instagram":
				return address.User;
			case "Youtube":
				return address.User;
			case "Appstore":
				return address.User;
			case "Googleplay":
				return address.User;
			case "Twitter":
				return address.User;
			case "IM":
				return address.buddy;
			case "Mail":
				return address.address;
			case "Webchat":
				return address.customerId;
			default:
				return address.AddressId;
		}
	};

	Contact.prototype.addToCampaign = function(campaignId) {
		for (var i = 0; i < this.campaigns.length; i++) {
			if (this.campaigns[i] === campaignId) {
				return false;
			}
		}
		this.campaigns.push(campaignId);
		return true;
	};

	Contact.prototype.removeFromCampaign = function(campaignId) {
		for (var i = 0; i < this.campaigns.length; i++) {
			if (this.campaigns[i] === campaignId) {
				this.campaigns.splice(i, 1);
				return true;
			}
		}
		return false;
	};


	/*
		Get contact online status, returning only TRUE if the user is online on one of her IM addresses
	*/
	Contact.prototype.getOnlineStatus = function() {
		var status = false;
		if (this.addresses) {
			for (var i = 0; i < this.addresses.length; i++) {
				var address = this.addresses[i];
				if (address.Type === "IM" && address.status &&  address.status.online === true) {
					status = true;
					break;
				}
			}
		}
		return status;
	};

	/*
		Delete a particular contact address, MEGA HIPER ULTRA WARNING
		1) The address is not removed by key, it is removed by display text (this is recipe for disaster)
		2) This function is NOT invoking a server API, only changes locally, so .save() must be called in order for the changes to take effect
	*/
	Contact.prototype.deleteAddress = function(addressToDelete) {
		if (this.addresses) {
			if (addressToDelete) {
				for (var i = 0; i < this.addresses.length; i++) {
					var address = this.addresses[i];
					if (this.getAddressDisplayText(address) === this.getAddressDisplayText(addressToDelete)) {
						this.addresses.splice(i, 1);
						return;
					}
				}
			}
		}
	};

	/*
		Updates a particular contact address, MEGA HIPER ULTRA WARNING
		1) The address is not updated by key, it's removed by display text (this is recipe for disaster)
		2) This function is NOT invoking a server API, only changes locally, so .save() must be called in order for the changes to take effect
	*/
	Contact.prototype.updateAddress = function(newAddress, previousAddress) {
		if (this.addresses) {
			if (previousAddress) {
				for (var i = 0; i < this.addresses.length; i++) {
					var address = this.addresses[i];
					if (this.getAddressDisplayText(address) === this.getAddressDisplayText(previousAddress)) {
						this.addresses[i] = newAddress;
						return;
					}
				}
			}
		}
		else {
			this.addresses = {};
		}
		this.addresses = _.toArray(this.addresses);
		this.addresses.push(newAddress);
	};

	Contact.prototype.getUsersFavorite = function()
	{
		switch ( typeof this.UsersFavorite )
		{
			case "object":
				return this.UsersFavorite;
			case "string":
				this.UsersFavorite = JSON.parse( this.UsersFavorite );
				return this.UsersFavorite;
			default:
				return null;
		}
	};

	Contact.prototype.isUserFavorite = function(user /* optional value */) {
		var users_favorite = this.getUsersFavorite();
		if (!users_favorite) {
			return false;
		}
		user = (user || inConcert.app.getLoggedUserId()).toLowerCase();
		var simple_user_id = inConcert.Utils.getIdWithoutVCC(user);
		var is_favorite = false;
		$.each(users_favorite, function(_, user_favorite_id) {
			user_favorite_id = user_favorite_id.toLowerCase();
			if (user_favorite_id === simple_user_id) {
				is_favorite = true;
				return false; //break
			}
		});
		return is_favorite;
	};

	Contact.prototype.addUserFavorite = function(user /* optional value */) {
		user = (user || inConcert.app.getLoggedUserId()).toLowerCase();
		var simple_user_id = inConcert.Utils.getIdWithoutVCC(user);
		this.UsersFavorite = this.getUsersFavorite() || [];
		if (!this.isUserFavorite(simple_user_id)) {
			this.UsersFavorite.push(simple_user_id);
			return true;
		}
		return false;
	};

	Contact.prototype.removeUserFavorite = function(user /* optional value */) {
		var users_favorite = this.getUsersFavorite();
		if (!users_favorite) {
			return false;
		}
		user = (user || inConcert.app.getLoggedUserId()).toLowerCase();
		var simple_user_id = inConcert.Utils.getIdWithoutVCC(user);
		var index_to_remove = -1;
		$.each(users_favorite, function(i, user_favorite_id) {
			user_favorite_id = user_favorite_id.toLowerCase();
			if (user_favorite_id === simple_user_id) {
				index_to_remove = i;
				return false; //break
			}
		});
		if (index_to_remove === -1) {
			return false;
		}
		users_favorite.splice(index_to_remove, 1);
		return true;
	};

	/*
		ContactManager, entry point for everything happening in this module
	*/
	function ContactManager(baseUrl) {
		s_baseUrl = baseUrl;
		this.m_baseUrl = baseUrl;
		this.m_lastContactFilter = null;
		this.m_cachedRelationships = {};
	}

	//Define functions and objects in the namespace in case they don't already exist
	if (typeof inConcert.ContactManager !== "function") {
		inConcert.ContactManager = ContactManager;
	}

	/*
		Retrieves full contact data using a specific ID
	*/
	ContactManager.prototype.getContact = function(contactID, onContactFoundFn) {
		var url = this.m_baseUrl + "/contacts/get/" + contactID;
		inConcert.invoke({
			url: url,
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				// handle contact not found
				if (response.result && response.result === "CONTACT_NOT_FOUND") {
					if (onContactFoundFn) {
						onContactFoundFn();
					}
				} else {
					var contact = new Contact(response);
					if (onContactFoundFn) {
						onContactFoundFn(contact);
					}
				}
			},
			error: function(undefined, error) {
				if (onContactFoundFn) {
					onContactFoundFn(null, error);
				}
			}
		});
	};

	/**
	 * getContactByAddress: can look for a contact like "mail", "pepe@example.com" or using a proper Fullcontact address
	 *
	 * @param  {Object}   address             - A normalized address object
	 * @param  {Function} onContactFoundFn    - Callback
	 */
	ContactManager.prototype.getContactByAddress = function(address, onContactFoundFn) {
		inConcert.logger.action({action_name : "get_contact_by_address", address : address, action_type:"contact"});
		var url = this.m_baseUrl + "/contacts/get_by_address";

		inConcert.invoke({
			url: url,
			type: "POST",
			dataType: "json",
			contentType: "application/json",
			data: JSON.stringify({address : address}),
			success: function(response) {
				if (typeof(onContactFoundFn) === "function"){
					if (response.Id){
						var fullContact = new Contact(response);
						onContactFoundFn(fullContact);
					}
					else{
						onContactFoundFn("<%=_T('This does not look like a contact')%>");
					}
				}
			},
			error: function(/*response, error*/) {
				if (typeof(onContactFoundFn) === "function"){
					onContactFoundFn();
				}
			}
		});
	};

	ContactManager.prototype.getContactByExternalId = function(externalId, onContactFoundFn) {
		inConcert.logger.action({action_name : "get_contact_by_external_id", externalId : externalId, action_type:"contact"});
		var url = this.m_baseUrl + "/contacts/get_by_external_id";

		inConcert.invoke({
			url: url,
			type: "POST",
			dataType: "json",
			contentType: "application/json",
			data: JSON.stringify({ externalId: externalId}),
			success: function(response) {
				// handle contact not found
				if (response.result && response.result === "CONTACT_NOT_FOUND") {
					if (onContactFoundFn) {
						onContactFoundFn();
					}
				} else {
					var contact = new Contact(response);
					if (onContactFoundFn) {
						onContactFoundFn(contact);
					}
				}
			},
			error: function(undefined, error) {
				if (typeof(onContactFoundFn) === "function"){
					onContactFoundFn(null, error && error.message ? error.message : "<%=_T('unknown error')%>");
				}
			}
		});
	};

	/**
	 * Given an address type and an address value will build a proper address object.
	 *
	 * @param  {String}   addressType     - The address type (i.e. if mail: "Mail")
	 * @param  {String}   address         - The address object
	 * @param  {Function} onAddressBuilt  - Callback
	 */
	ContactManager.prototype.buildAddress = function(addressType, address_info, onAddressBuilt = $.noop)
	{
		const url = this.m_baseUrl + "/contacts/build_address";
		const request = {
			addressType : addressType,
			address : address_info,
			options: {
				originCountryCode : inConcert.app.getCountryCode(),
				originAreaCode : inConcert.app.getAreaCode()
			}
		};

		inConcert.invoke({
			url: url,
			type: "POST",
			dataType: "json",
			contentType: "application/json",
			data: JSON.stringify(request),
			success: function(address) {
				onAddressBuilt(address || null);
			},
			error: function(undefined, error) {
				onAddressBuilt(null, error.message);
			}
		});
	};

	ContactManager.prototype.addressTypeToInteractionType = function(addressType) {
		var interactionType;
		switch (addressType) {
			case "Phone":
				interactionType = "CALL";
				break;
			case "Facebook":
				interactionType = "FACEBOOK";
				break;
			case "Twitter":
				interactionType = "TWITTER";
				break;
			case "Instagram":
				interactionType = "INSTAGRAM";
				break;
			case "Youtube":
				interactionType = "YOUTUBE";
				break;
			case "Appstore":
				interactionType = "APPSTORE";
				break;
			case "Googleplay":
				interactionType = "GOOGLEPLAY";
				break;
			case "IM":
				interactionType = "IM";
				break;
			case "Mail":
				interactionType = "MAIL";
				break;
			case "Webchat":
				interactionType = "CHAT";
				break;
			default:
			var errorMsg = "ContactManager - Someone is trying convert an address type to an interaction type with an invalid type: " + addressType;
			inConcert.logger.error(errorMsg);
			throw(errorMsg);
		}
		return interactionType;
	};

	ContactManager.prototype.cancelCurrentRequest = function() {
		if (this.currentRequest) {
			this.currentRequest.abort();
		}
	};

	/*
		Returns all the favorite contacts for the logged user
	*/
	ContactManager.prototype.getFavorites = function(onFavoritesReady) {
		var self = this;
		this.cancelCurrentRequest();
		var url = this.m_baseUrl + "/contacts/get_favorites";
		this.currentRequest = inConcert.invoke({
			url: url,
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				var contactObjects = [];
				for (var i = 0; i < response.length; i++) {
					var contact = new Contact(response[i]);
					contactObjects.push(contact);
				}
				self.m_lastFavorites = contactObjects;
				if (onFavoritesReady) {
					onFavoritesReady(contactObjects);
				}
			},
			error: function(/*response, error*/) {
				onFavoritesReady();
			}
		});
	};

	/*
		Removes a specific contact from the logged user favorites
	*/
	ContactManager.prototype.removeFromFavorites = function( contact, onFavoriteRemoved, campaign_id, app_id = "interaction" ) {
		inConcert.logger.action({action_name : "remove_favorite", id : contact.Id, action_type:"contact"});

		var request = {
			contactID : contact.Id,
			campaign_id : campaign_id, // optional, used by permissions
			app_id : app_id
		};
		var url = this.m_baseUrl + "/contacts/remove_from_favorites";
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(undefined) {
				contact.removeUserFavorite();
				if (onFavoriteRemoved) {
					onFavoriteRemoved(true);
				}
			},
			error: function(undefined, error) {
				onFavoriteRemoved(false, error);
			}
		});
	};

	/*
		Adds a specific contact to the logged user favorites
	*/
	ContactManager.prototype.addToFavorites = function( contact, onFavoriteAdded, campaign_id, app_id = "interaction" ) {
		inConcert.logger.action({action_name : "add_favorite", id : contact.Id, action_type:"contact"});

		var request = {
			contactID : contact.Id,
			contactName : contact.Name,
			campaign_id : campaign_id, // optional, used by permissions
			app_id : app_id
		};
		var url = this.m_baseUrl + "/contacts/add_to_favorites";
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				if (response.result === "OK") {
					contact.addUserFavorite();
					if (typeof(onFavoriteAdded) === "function") {
						onFavoriteAdded(true);
					}
				}
			},
			error: function(undefined, error) {
				if (typeof(onFavoriteAdded) === "function") {
					onFavoriteAdded(false, error);
				}
			}
		});
	};


	/*
		Creates a new empty  contact
	*/
	ContactManager.prototype.create = function(newContact) {
		var contact = newContact ? newContact : {
			isNew: true,
			Id: inConcert.Utils.generateGuid() + "@" + inConcert.app.get_vcc_id(),
			Name: "",
			Company: "",
			addresses: [],
			campaigns: []
		};
		inConcert.logger.action({action_name: "create_contact", id: contact.Id, action_type: "contact"});
		return new Contact(contact);
	};

	/*
		Executes a contact search using the specified filter
	*/
	ContactManager.prototype.search = function(filter, onSearchCompleted) {
		inConcert.logger.action({action_name : "search_contacts", filter : filter, action_type:"contact"});
		this.cancelCurrentRequest();

		var url = this.m_baseUrl + "/contacts/search";
		this.currentRequest = inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(filter),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				var contacts = response.contacts;
				var contactObjects = [];
				if (!contacts || contacts.length === 0) {
					onSearchCompleted(contactObjects, "");
					return;
				}
				for (var i = 0; i < contacts.length; i++) {
					var contact = new Contact(contacts[i]);
					contactObjects.push(contact);
				}
				onSearchCompleted(contactObjects, response.last);
			},
			error: function(undefined, error) {
				//TODO switch according to message type and show a better friendlier message
				onSearchCompleted(undefined, undefined, error.result);
			}
		});
	};

	ContactManager.prototype.merge = function( contact, obsoleteContactId, onContactsMerged, campaign_id, app_id = "interaction" ) {
		if (typeof(onContactsMerged) !== "function") {
			throw("Callback needed for contact merge action");
		}

		var request = {
			id : contact.Id,
			contact: contact,
			obsoleteId : obsoleteContactId,
			campaign_id: campaign_id, // optional, used by permissions
			app_id : app_id
		};

		var url = s_baseUrl + "/contacts/merge";
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				onContactsMerged(new Contact(response));
			},
			error: function(undefined, error) {
				onContactsMerged(null, error.result);
			}
		});
	};

	//
	//	Twitter Specific Actions
	//

	/*
		Given a contact, will get the social information available,
		save that info to the contact and return the contact updated
	*/
	ContactManager.prototype.twitterRefreshSocialInformation = function(contact, twitterAccount, campaignId, onContactReady) {
		inConcert.logger.action({action_name : "refresh_social_info", contact : contact, action_type:"contact"});

		var url = this.m_baseUrl + "/contacts/twitter/refresh_social_info";
		var request = {
			contactID : contact.Id,
			campaignId : campaignId,
			twitterAccount : twitterAccount
		};

		this.currentRequest = inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				onContactReady(response);
			},
			error: function(response, error) {
				var errorMessage;
				switch (error.result) {
					case "USER_NOT_FOUND":
						errorMessage =  "<%=_T('Contact not found')%>";
						break;
					case "TWITTER_USER_NOT_FOUND":
						errorMessage =  "<%=_T('Twitter user not found!')%>";
						break;
					case "GET_TWITTER_USER_ERROR":
						errorMessage =  "<%=_T('There is an error trying to get twitter user information from this contact')%>";
						break;
					case "UNABLE_TO_SAVE_CONTACT":
						errorMessage =  "<%=_T('There is an error trying to save the contact')%>";
						break;
					default:
						errorMessage =  "<%=_T('There is an Internal Error refreshing social information, please try again later')%>";
						break;
				}
				onContactReady(response, errorMessage);
			}
		});
	};

	/*
		Shows the relationship between two accounts
	*/
	ContactManager.prototype.twitterGetRelationship = function(screen_name, accountId, campaignId, callbackFn = $.noop)
	{
		const self = this;
		inConcert.logger.action({action_name : "twitter_get_relationship", campaign : campaignId, screen_name : screen_name, account : accountId, action_type : "contacts"});

		const campaign = inConcert.app.getCampaign(campaignId);
		if (!campaign) {
			callbackFn(false);
			return;
		}
		const accountIsOperational = campaign.accountIsOperational(accountId);
		if (!accountIsOperational.success) {
			callbackFn(false);
			return;
		}

		if (self.m_cachedRelationships[screen_name]) {
			if (self.m_cachedRelationships[screen_name].status === "pending") {
				// significa que ya se esta solicitando la data para este usuario. Espero 2 secs.
				window.setTimeout(function() {
					if (self.m_cachedRelationships[screen_name]) {
						// Return cached relationship if relation was updated in less than an hour.
						if (inConcert.Utils.getTimeDifference(self.m_cachedRelationships[screen_name].lastUpdate, true) < 3600){
							callbackFn(self.m_cachedRelationships[screen_name]);
							return;
						}
					}
				},1000);
				return;
			} else {
				// Return cached relationship if relation was updated in less than an hour.
				if (inConcert.Utils.getTimeDifference(self.m_cachedRelationships[screen_name].lastUpdate, true) < 3600){
					callbackFn(self.m_cachedRelationships[screen_name]);
					return;
				}
			}
		}
		self.m_cachedRelationships[screen_name] = {status : "pending"};
		const url = self.m_baseUrl + "/contacts/twitter/show_relationship";
		const request = {
			campaignId: campaignId,
			account   : accountId,
			to_account: screen_name
		};

		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(relationship) {
				self.m_cachedRelationships[screen_name] = relationship;
				self.m_cachedRelationships[screen_name].lastUpdate = new Date().getTime();
				delete self.m_cachedRelationships[screen_name].status;
				callbackFn(relationship);
			},
			error: function(/*response, error*/) {
				delete self.m_cachedRelationships[screen_name].status;
				callbackFn(false);
			}
		});
	};

	ContactManager.prototype.twitterFollowUser = function(screen_name, account, campaignId, callbackFn, contactID, followers) {
		var self = this;
		inConcert.logger.action({action_name : "twitter_follow", campaign : campaignId, screen_name : screen_name, account : account, action_type : "contacts"});

		var url = this.m_baseUrl + "/contacts/twitter/follow_user";
		var request = {
			campaignId : campaignId,
			account : account,
			userId : screen_name
		};

		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				delete self.m_cachedRelationships[screen_name];

				if(contactID && !isNaN(followers))
				{
					self.updateContactfollowers(contactID, followers + 1, function () {		
						if (typeof(callbackFn) === "function") {
							callbackFn(true);
						}
					});
				}
				else if (typeof(callbackFn) === "function") {
					callbackFn(true);
				}
			},
			error: function(undefined, error) {
				inConcert.logger.action({action_name : "twitter_follow", error_message : error.result });
				if (typeof(callbackFn) === "function") {
					var message;
					switch (error.result) {
						case "PROXY_ERROR":
							message = "Unable to follow user: " + error.reason; //TODO: Ver que el reason sea algo entendible y no un JSON!
							break;
						default:
							message = "There was an error while applying this action, please try again later";
					}
					callbackFn(false, message);
				}
			}
		});
	};

	ContactManager.prototype.updateContactfollowers = function(contactID, followers, callbackFn = $.noop)
	{

		var url = this.m_baseUrl + "/contacts/update_contact_followers";
		var request = {
			contactID : contactID,
			followers : followers
		};

		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {

				callbackFn();
			},
			error: function(undefined, error) {
				inConcert.logger.action({action_name : "update_contact_followers", error_message : error.result });
				callbackFn();
			}
		});
	};

	ContactManager.prototype.twitterUnfollowUser = function(screen_name, account, campaignId, callbackFn, contactID, followers) {
		var self = this;
		inConcert.logger.action({action_name : "twitter_unfollow", campaign : this.campaign, screen_name : screen_name, account : account, action_type : "contacts"});

		var url = this.m_baseUrl + "/contacts/twitter/unfollow_user";
		var request = {
			campaignId : campaignId,
			account : account,
			userId : screen_name
		};

		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				delete self.m_cachedRelationships[screen_name];
				if(contactID && !isNaN(followers)) {
					followers = followers === 0 ? followers: followers - 1;
					self.updateContactfollowers(contactID, followers, function () {		
						if (typeof(callbackFn) === "function") {
							callbackFn(true);
						}
					});
				}
				else if (typeof(callbackFn) === "function") {
					callbackFn(true);
				}
					
			},
			error: function(undefined, error) {
				inConcert.logger.action({action_name : "twitter_unfollow", error_message : error.result });
				if (typeof(callbackFn) === "function") {
					var message;
					switch (error.result) {
						case "PROXY_ERROR":
							message = "Unable to unfollow user: " + error.reason;
							break;
						default:
							message = "There was an error while applying this action, please try again later";
					}
					callbackFn(false, message);
				}
			}
		});
	};

	ContactManager.prototype.twitterGetUser = function(screenName, campaignId, account, callbackFn) {
		inConcert.logger.action({action_name : "get_twitter_user", screenName : screenName, account : account ,campaign : campaignId, action_type : "contacts"});

		var url = this.m_baseUrl + "/contacts/twitter/get_user";
		var request = {
			userId : screenName,
			campaignId : campaignId,
			account : account
		};

		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				if (typeof(callbackFn) === "function") {
					callbackFn(response);
				}
			},
			error : function(/*response, error*/) {
				if (typeof(callbackFn) === "function") {
					callbackFn("<%=_T('Cannot retrieve twitter user info at this time, please try again later')%>");
				}
			}
		});
	};

	ContactManager.prototype.twitterBlockExists = function(screen_name, account, campaignId, callbackFn) {
		inConcert.logger.action({action_name : "twitter_block_exists", account : account, screen_name : screen_name, campaign : campaignId, action_type : "contacts"});

		var url = this.m_baseUrl + "/contacts/twitter/block_exists";
		var request = {
			campaignId : campaignId,
			account : account,
			userId : screen_name
		};

		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				if (typeof(callbackFn) === "function") {
					if (response.blocked_user){
						callbackFn(true);
					} else {
						// user not found
						callbackFn(false);
					}
				}
			},
			error : function(/*response, error*/) {
				if (typeof(callbackFn) === "function") {
					callbackFn(false, "<%=_T('Cannot check if twitter block exist at this time, please try again later')%>");
				}
			}
		});
	};

	ContactManager.prototype.twitterBlockUser = function(screen_name, account, campaignId, callbackFn) {
		inConcert.logger.action({action_name : "twitter_block_user", account : account, screen_name : screen_name, campaign : campaignId, action_type : "contacts"});

		var url = this.m_baseUrl + "/contacts/twitter/block_user";
		var request = {
			campaignId : campaignId,
			account : account,
			userId : screen_name
		};

		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(undefined) {
				if (typeof(callbackFn) === "function") {
					callbackFn(true);
				}
			},
			error : function(/*response, error*/) {
				if (typeof(callbackFn) === "function") {
					callbackFn(false, "<%=_T('Cannot block twitter user at this time, please try again later')%>");
				}
			}
		});
	};

	ContactManager.prototype.twitterUnBlockUser = function(screen_name, account, campaignId, callbackFn) {
		inConcert.logger.action({action_name : "twitter_unblock_user", account : account, screen_name : screen_name, campaign : campaignId, action_type : "contacts"});

		var url = this.m_baseUrl + "/contacts/twitter/unblock_user";
		var request = {
			campaignId : campaignId,
			account : account,
			userId : screen_name
		};

		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(undefined) {
				if (typeof(callbackFn) === "function") {
					callbackFn(true);
				}
			},
			error : function(/*response, error*/) {
				if (typeof(callbackFn) === "function") {
					callbackFn(false, "<%=_T('Cannot unblock twitter user at this time, please try again later')%>");
				}
			}
		});
	};

	ContactManager.prototype.normalizeContactAddress = function (address, addressType, addressProtocol, phone_kind, addressInfo = {})
	{
		let normalizedAddress;

		switch ((addressType || "").toUpperCase()) {
			// TO-DO: Review this, this is comming from Tweety this way.
			case "CELLULAR":
				normalizedAddress = {
					Type : "Phone",
					Number : address,
					displayable : address
				};
				break;
			case "PHONE":
				normalizedAddress = {
					// Estos atributos conforman parte de la clave que necesita la rocka "fullcontact", tienen que estar si o si. (helpers.BuildAddressId)
					Type       : "Phone",
					Country    : addressInfo.Country || "",
					Area       : addressInfo.Area || "",
					LocalNumber: addressInfo.LocalNumber,
					Number     : address,
					Extension  : addressInfo.Extension || "",
					SanitizedId: address,
					// -------------------
					displayable: address,
					Kind       : phone_kind || ""
				};
				break;
			case "IM":
				normalizedAddress = {
					Type : "IM",
					protocol : addressProtocol,
					buddy : address,
					displayable : address
				};
				break;
			case "MAIL":
				normalizedAddress = {
					Type : "Mail",
					address : address,
					displayable : address
				};
				break;
			case "FACEBOOK":
				normalizedAddress = {
					Type : "Facebook",
					UID : address,
					displayable : address
				};
				break;
			case "INSTAGRAM":
				normalizedAddress = {
					Type : "Instagram",
					User : address,
					displayable : address
				};
				break;
			case "YOUTUBE":
				normalizedAddress = {
					Type : "Youtube",
					User : address,
					displayable : address
				};
				break;
			case "APPSTORE":
				normalizedAddress = {
					Type : "Appstore",
					User : address,
					displayable : address
				};
				break;
			case "GOOGLEPLAY":
				normalizedAddress = {
					Type : "Googleplay",
					User : address,
					displayable : address
				};
				break;
			case "TWITTER":
				normalizedAddress = {
					Type : "Twitter",
					User : address,
					displayable : address
				};
				break;
			default:
				throw("Invalid address type provided:" + addressType);
		}
		return normalizedAddress;
	};

	/*
		Returns all the addresses book for channel
	*/
	ContactManager.prototype.getAddressesBookByChannel = function(channel, callback = $.noop) {
		const self = this;
		this.cancelCurrentRequest();
		let url = this.m_baseUrl + "/contacts/addresses_book/load_by_channel/" + channel;
		this.currentRequest = inConcert.invoke({
			url: url,
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				let contactObjects = [];
				for (let i = 0; i < response.length; i++) {
					let rawContact = response[i];
					rawContact.contactType = "ADDRESSES BOOK";
					let contact = new Contact(rawContact);
					contactObjects.push(contact);
				}
				self.m_lastFavorites = contactObjects;

				callback(contactObjects);
			},
			error: function(/*response, error*/) {
				callback();
			}
		});
	};

	/*
		Returns all the addresses book for channel
	*/
	ContactManager.prototype.getAddressesBook = function(callback = $.noop) {
		const url = this.m_baseUrl + "/contacts/addresses_book/get_all";
		this.currentRequest = inConcert.invoke({
			url: url,
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				let contactObjects = [];
				for (let i = 0; i < response.length; i++) {
					let rawContact = response[i];
					rawContact.contactType = "ADDRESSES BOOK";
					let contact = new Contact(rawContact);
					contactObjects.push(contact);
				}

				callback(contactObjects);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});
	};
}());
