/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 9 */

/* global inConcert */

if (!this.inConcert) {
	this.inConcert = {};
}
if (!inConcert.interactions) {
	inConcert.interactions = {};
}

/**
 * @namespace
 * @name AppstoreInteraction
 */
(function () {

	"use strict";
	const NOT_EXIST_IN_ARRAY = -1;
	const FORMAT = inConcert.Utils.format;

	/**
	 * Class AppstoreInteraction.
	 * @param {Object} rawInteraction
	 */
	function AppstoreInteraction (rawInteraction)
	{
		AppstoreInteraction.uber.constructor.call(this, rawInteraction);
		this.events = this.events || [];

		// Pending interactions have no properties yet. Provide some sane defaults
		if (this.isBeingComposed()) {
			this.properties = this.properties || {};
			this.properties.display_name = "";
			this.properties.is_private_message = false;
			this.properties.summary = "";
		}
	}

	inConcert.Utils.inherit(AppstoreInteraction, inConcert.interactions.base);

	if (typeof inConcert.interactions.appstore !== "function") {
		inConcert.interactions.appstore = AppstoreInteraction;
	}

	AppstoreInteraction.prototype.inflate = function (callback)
	{
		var self = this;
		AppstoreInteraction.uber.inflate.call(this, function(data){
			if(data) {
				// Building the events hierarchy has the side-effect of setting
				// createdTime in the events
				// TODO: do that in the server
				self.buildEventsHierarchy();
				self.sortEvents();
				callback(data);
			}
		});
	};

	//sends a message on the current conversation
	AppstoreInteraction.prototype.sendMessage = function(/*message*/) { };

	AppstoreInteraction.prototype.get_pending_origin_post = function()
	{
		if ( this.isPrivateMessage() ) {
			return;
		}
		var pending_events = this.get_pending_events();
		if ( Object.keys( pending_events ).length < 1 ) {
			return;
		}
		var origin_post;
		$.each( pending_events, function( _, pending_event )
		{
			if ( pending_event.is_origin_post ) {
				origin_post = pending_event;
				return false; // break
			}
		});
		return origin_post;
	};

	/*
		Esta funcion obtiene el evento/mensaje/post que ha orignado la interaccion
	*/
	AppstoreInteraction.prototype.get_origin_message_id = function()
	{
		if (!$.isArray(this.events) || this.events.length === 0) {
			return;
		}
		var origin_message_id = this.interactionData && this.interactionData.postId;
		if (!origin_message_id) {
			origin_message_id = this.events[0].message.id;
		}
		return origin_message_id;
	};

	/*
		Funcion que ordena los eventos primero el evento orgigen, luego por fecha de creacion
	*/
	AppstoreInteraction.prototype.__sort_by_creation_date = function()
	{
		var origin_message_id = this.get_origin_message_id();

		this.events.sort( function( a, b )
		{
			if (a.message && a.message.id === origin_message_id) {
				// En la primer posicion agrego el mensaje original
				return -1;
			}
			// Luego ordeno por fecha de creacion
			if (!a.original_timestamp || !b.original_timestamp) {
				return a.__index - b.__index;
			}
			if (a.original_timestamp === b.original_timestamp) {
				return a.__index - b.__index;
			}
			return a.original_timestamp - b.original_timestamp;
		});
	};

	/**
	* Build post and comments tree ordering by root post creation date.
	* This function must be called before sortEvents.
	*/
	//
	AppstoreInteraction.prototype.buildEventsHierarchy = function() {
		this.assert_inflated();

		var rootNodesIndices = {};
		var knownIds = [];
		var interactionEvents = this.getEventList();

		if (!interactionEvents) {
			return;
		}
		if (interactionEvents && interactionEvents.length) {
			var events = interactionEvents.length;

			for (var i = 0; i < events; i++) {
				var event = interactionEvents[i];
				event.children = [];
				if (!event.message) {
					continue;
				}
				event.creationDate = new Date(event.message.timestamp * 1000);
				if (event.message.parent_id && rootNodesIndices.hasOwnProperty(event.message.parent_id)) {
					if ( knownIds.indexOf(event.message.id) === NOT_EXIST_IN_ARRAY ) { // not repeated
						//set parent
						event.parentIndex = rootNodesIndices[event.message.parent_id];
						//tell parent to add current as children
						var parent = interactionEvents[event.parentIndex];
						parent.children.push(i);

						//add to known list
						knownIds.push(event.message.id);
					}
				}
				else {
					rootNodesIndices[event.message.id] = i;
				}
			}
		}
	};

	//------------------
	// http://fbdevwiki.com/wiki/Error_codes
	//
	AppstoreInteraction.prototype.get_error_message = function( error_code, type )
	{
		switch ( (error_code || "" ).toUpperCase() )
		{
		// SERVER ERRORS!!
			case "MISSING_FIELD":
				if ( type === "NEW_POST" ) {
					return "<%=_T('The post field is mandatory')%>";
				}
				else if ( type === "NEW_PRIVATE_MESSAGE" ) {
					return "<%=_T('The private message field is mandatory')%>";
				}
				else if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('The post comment field is mandatory')%>";
				}
				break;
						
			case "POST_CLOSED":
				if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('Unable to send message. Related post was closed or deleted.')%>";
				}
				break;

			case "IRSERVER_ERROR":
				if ( type === "NEW_POST" ) {
					return "<%=_T('There was an error while creating the new post')%>";
				}
				else if ( type === "NEW_PRIVATE_MESSAGE" ) {
					return "<%=_T('There was an error while creating the new private message')%>";
				}
				else if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('There was an error while creating the new post comment')%>";
				}
				else if ( type === "LIKE" ) {
					return "<%=_T('There was an error while liking/unliking')%>";
				}
				break;

			case "NOT_LICENSED":
				return "<%=_T('Unable to send review reply. Not has licensed for the Appstore channel.')%>";

			case "TIMEOUT":
				return "<%=_T('Waiting time expired to send message')%>";

			case "CHATSERVER_NOT_AVAILABLE":
				return "<%=_T('Chatserver not available')%>";

			case "INTERACTION_ALREADY_EXISTS":
				return "<%=_T('There was an error while creating the conversation')%>";

			case "AGENT_UNAVAILABLE":
				return "<%=_T('There was an error while creating the conversation')%>";

			case "FAIL_CREATING_OUTGOING":
				return "<%=_T('There was an error while creating the conversation')%>";

			case "INTERACTION_CANCELED_BEFORE_TAKING":
				return "<%=_T('The conversation was cancelled')%>";

	// DEFAULT errors
			default:
				if ( type === "NEW_POST" ) {
					return "<%=_T('There was an error while creating the conversation')%>";
				}
				else if ( type === "NEW_PRIVATE_MESSAGE" ) {
					return "<%=_T('There was an error while creating the private message')%>";
				}
				else if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('There was an error while creating the new post comment')%>";
				}
				break;
		}

		return "<%=_T('There was an error while sending message')%>";
	};

	AppstoreInteraction.prototype.commentPost = function( fake_event_id, postId, commentId, commentMessage, callback ) 
	{
		var self = this;
		callback = callback || $.noop;
		fake_event_id = fake_event_id || inConcert.Utils.generateGuid();

		if (commentId) {
			commentId = commentId.toString();
		}

		inConcert.logger.action({action_name : "comment_appstore_post", id : this.id, type : this.type,
			campaign : this.campaign, postId : postId, account : this.account, action_type : "conversation", fake_event_id: fake_event_id});

		var url = this.m_manager.getBaseUrl() + "/interactions/appstore/comment_post/";
		var request = {
			fake_event_id : fake_event_id,
			interactionId : this.id,
			postId        : postId,
			commentId     : commentId,
			campaign      : this.campaign,
			message       : commentMessage,
			account       : this.account
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				var message;
				var status = result.result === "OK";
				if ( result.result !== "OK" ) {
					message = self.get_error_message( result.result, "NEW_COMMENT_POST" );
				}
				callback( status, message );
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "new_appstore_comment_post", error_message : error.result });
				var message = self.get_error_message( error.result, "NEW_COMMENT_POST" );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};

	AppstoreInteraction.prototype.like = function(like, event_index, callback) {
		var self = this;
		callback = callback || $.noop;
		var fake_event_id = inConcert.Utils.generateGuid();

		inConcert.logger.action({
			action_name : "appstore_like",
			id : this.id,
			type : this.type,
			campaign : this.campaign,
			fake_event_id: fake_event_id
		});

		var event = this.getEvent(event_index);
		if (!event || !event.message || !event.message.id) {
			inConcert.logger.action({ action_name : "appstore_like", error_message : "Event not found" });
			var message = self.get_error_message( "IRSERVER_ERROR", "LIKE" );
			callback( false, message, "Event not found" );
			return;
		}

		var url = this.m_manager.getBaseUrl() + "/interactions/appstore/like/";
		var request = {
			fake_event_id : fake_event_id,
			interactionId : this.id,
			campaign      : this.campaign,
			account       : this.account,
			objectId      : event.message.id,
			like          : like
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback( true, "OK" );
			},
			error: function(_, error){
				inConcert.logger.action({ action_name : "appstore_like", error_message : error.result });
				var message = self.get_error_message( error.result, "LIKE" );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};

	// Just returns the summary
	AppstoreInteraction.prototype.getExtendedInfo = function( max_length )
	{
		return inConcert.Utils.truncateWord( this.properties.summary, max_length );
	};

	AppstoreInteraction.prototype.getEventShortDescription = function(index) {
		var message = this.getMessageContent(index);
		return this.getMessageText(message);
	};

	AppstoreInteraction.prototype.getMessageContent = function(index)
	{
		this.assert_inflated();
		var data = { type : "text", text : "" };
		if (!this.events || !this.events.length || index >= this.events.length) {
			return data;
		}
		if (!this.events[index] || !this.events[index].message) {
			return data;
		}
		let message = this.events[index].message;
		data.text = message.text;
		data.review = message.review;
		data.type = message.msgtype;
		switch(message.msgtype) {
			case "attachment":
				if (message.mediaFail) {
					data.mediaFail = message.mediaFail;
				}
				else {
					data.fileType = message.fileType;
					data.fileName = message.fileName;
					data.fileId = message.fileId || message.fileName;
					data.fileExtension = message.fileExtension ||
						inConcert.Utils.getFileExtension(data.fileId) ||
						inConcert.Utils.getFileExtensionFromMimeType(message.mimeType);
					data.sticker = message.mimeType === "image/webp";
					switch(message.attachmentType) {
						case "repository":
							data.link = inConcert.app.getBaseDomain() + "/inconcert/apps/repository/download/?id=" + encodeURIComponent(data.fileId) + "&section=-1&part=-1";
							if (data.fileName) {
								let fileName = "-";
								if ((data.fileName || "").toLowerCase() !== (data.fileId || "").toLowerCase()) {
									fileName = data.fileName;
									const max_length = 50;
									if (fileName.length > max_length) {
										// Si el mensaje (nombre del archivo), supera los "max_length" caracteres, lo trunco para que el nombre no se a muy extenso
										fileName = inConcert.Utils.truncateWord(fileName, max_length, "...");
									}
								}
								let timestampObject = new Date(message.timestamp * 1000);
								let userTimezoneObject = inConcert.app.getUserTimezoneObject();
								timestampObject = inConcert.Utils.convertLocalDateToTimezoneDate(timestampObject, userTimezoneObject);
								let stringDate = inConcert.Utils.getStringFromDate(timestampObject, false, "full", "ymd", "-");
								/*
									Al nombre sugerido lo conformo de la siguiente forma:
										1 - Con el prefjo del Canal
										2 - Con el texto del mensaje, si es que lo tiene, sino va vacio un "-".
										3 - Con la fecha y hora del mensaje, en el timezone del usuario.
										4 - La extension.
								*/
								let suggestedName = FORMAT("<%=_T('AppStore File-{0}-{1}.{2}')%>", fileName, stringDate, data.fileExtension);
								data.link += "&name=" + encodeURIComponent(suggestedName);
							}
							break;
						case "url":
							if (message.url) {
								data.link = message.url;
							}
							else {
								data.link = "https://www.appstore.com/watch?v=" + message.id;
							}
							break;
						default:
							data.mediaFail = true;
							break;
					}
					if (message.previewFileId) {
						data.previewLink = inConcert.app.getBaseDomain() + "/inconcert/apps/repository/download/?id=" + 
							encodeURIComponent(message.previewFileId) + "&section=-1&part=-1";
					}
					else if (message.previewUrl) {
						data.previewLink = message.previewUrl;
					}
					break;
				}			
		}
		return data;
	};

	/**
	* @param index
	*/
	AppstoreInteraction.prototype.getDateObject = function(index) {
		this.assert_inflated();

		index = index || 0;
		if ( !this.events || this.events.length === 0 || !this.events[ index ] ) {
			return "";
		}
		var event_date = this.events[ index ].creationDate;
		return this.convertDateStringToDate( event_date );
	};

	AppstoreInteraction.prototype.getRawAddress = function(index) {
		this.assert_inflated();
		return this.events[index].message.name;
	};

	AppstoreInteraction.prototype.getSource = function( sorted_index )
	{
		this.assert_inflated();
		if ( !this.events || sorted_index === undefined || sorted_index === -1 ) {
			return "";
		}
		var event = this.events[ sorted_index ];
		if ( event.creator === "agent" ) {
			return event.agent || inConcert.app.getCompleteId();
		}
		else {
			return event.message.name;
		}
	};

	AppstoreInteraction.prototype.getEventUserAvatarURI = function( event )
	{
		if (typeof(event) !== "object" || typeof(event.message) !== "object") {
			return "";
		}
		if ( event.creator === "agent" && event.agent ) {
			var user_id = event.agent;
			if ( !user_id ) {
				user_id = inConcert.app.getCompleteId();
			}
			if ( user_id.match(inConcert.app.idAtVCCRegExp) ) {
				user_id = RegExp.$1;
			}
			return inConcert.app.getBaseUrl() + "/users/get_avatar/" + user_id;
		}
		else if ( event.creator === "bot" ) {
			return "/inconcert/apps/agent/images/botAvatar.png";
		}
		else {
			return "";
		}
	};

	//returns the conversation contact or member display text
	AppstoreInteraction.prototype.getUserName = function( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}

		if ( event.creator === "agent" && event.agent ) {
			return this.get_agent_name( event );
		}
		if ( event.creator === "bot" ) {
			return this.getBotName();
		}
		else {
			if ( typeof( event.message ) !== "object" ) {
				return "";
			}
			return event.message.name || "";
		}
	};

	AppstoreInteraction.prototype.get_user_event_id = function( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}
		if ( event.creator === "agent" ) {
			return event.agent || inConcert.app.getCompleteId();
		}
		if ( typeof( event.message ) !== "object" ) {
			return "";
		}
		return event.message.name || "";
	};

	/**
	* Returns a displayable name, which was pre computed on the server. It extracts the user name from the first event 
	* of the interaction or the agent id, depending on which one originated the interaction.
	* The computed property has a prefixed value to it. If the prefix is "text:", then the rest is to be taken 
	* verbatim. It is the name of a appstore user.
	* If the prefix is "agent_id:", then we need to perform a lookup to translate the id to an agent name.
	* 
	* @return {String}
	*/
	AppstoreInteraction.prototype.getDisplayName = function()
	{
		var display_name = this.properties.display_name;
		if(display_name.match(/text:(.+)/)) {
			return RegExp.$1;
		}
		else if(display_name.match(/agent_id:/)) {
			var agent_id = RegExp.$1;
			if (agent_id === inConcert.app.getCompleteId()) {
				return inConcert.app.getLoggedUserName();
			}
			else {
				var agent_name;

				inConcert.app.getTeamMember( agent_id, function(agent)
					{
						if (!agent) {
							return;
						}
						// Esto por qué carajo no lo hago una vez sola?
						var have_name = false;
						if ((agent.Name || "").trim().length > 0) {
							agent_name = agent.Name;
							have_name = true;
						}
						if ((agent.FirstSurname || "").trim().length > 0) {
							agent_name = (have_name ? (agent_name + " ") : "") + agent.FirstSurname;
						}
					},
					true // from cache
				);

				agent_name = agent_name || agent_id;
				return agent_name;
			}
		}
		return display_name;
	};

	AppstoreInteraction.prototype.getFirstUnreadEventIndex = function() {
		this.assert_inflated();

		var eventIndex = inConcert.interactions.base.prototype.getFirstUnreadEventIndex.call(this);
		if (typeof(eventIndex) === "number" && this.events[eventIndex]) {
			var event = this.events[eventIndex];
			return (typeof(event.parentIndex) !== "undefined") ? event.parentIndex : eventIndex;
		}
		else {
			return eventIndex;
		}
	};
	
	/* @return {Boolean} */
	// IB: ESTA FUNCION SE PUEDE CALCULAR UNA VEZ EN EL SERVER Y LISTO
	AppstoreInteraction.prototype.eventsAreFromTheSameClient = function() {
		this.assert_inflated();

		var events_from_the_same_client = true;
		var last_event_client = null;

		this.events.forEach(function(current_event){
			var current_event_client = current_event.message.name;
			// first iteration
			if (!last_event_client){
				last_event_client = current_event_client;
			} else {
				if (current_event_client !== last_event_client){
					// IB: aca por que no para de iterar?
					// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/forEach
					// hay que usar Array.every o Array.some
					events_from_the_same_client = false;
				}
			}
		});
		return events_from_the_same_client;
	};

	//called by browser_windows.js to differentiate between posts and private messages
	AppstoreInteraction.prototype.isPrivateMessage = function() {
		//return true if at least one event is a private message
		return this.properties.is_private_message;
	};

	AppstoreInteraction.prototype.getMessageText = function(message) {
		if (message.webhookType === "REFERRAL") {
			return this.__getReferralMessage(message);
		}
		if (message.webhookType === "OPTIN") {
			return this.__getOptinMessage(message);
		}
		return message.text || 
			message.attachment && message.attachment.payload && message.attachment.payload.text || "";
	};

	//returns the extra information
	AppstoreInteraction.prototype.getMessage = function(index) {
		this.assert_inflated();

		if (this.events && this.events.length) {
			return this.getMessageText(this.events[index].message);
		}
		return "";
	};

	AppstoreInteraction.prototype.hasAttachments = function (event) {
		// IB: Esto se puede cambiar por Object.keys()
		var isEmpty = function (obj) {
			for(var prop in obj) {
				if (obj.hasOwnProperty(prop)) {
					return false;
				}
			}
			return true;
		};
		if (event) {
			return event.message && event.message.attachment && !isEmpty(event.message.attachment);
		}
		
		AppstoreInteraction.uber.hasAttachments.call(this);
	};

	AppstoreInteraction.prototype.isLiked = function (index) {
		var event = this.getEvent(index);
		var likes = this.interactionData && this.interactionData.likes || {};
		return event && event.message && event.message.id && likes[event.message.id] || false;
	};

	AppstoreInteraction.prototype.isValid = function() {
		/*this.assert_inflated();
		// IB: Revisar cual es el sentido de esto

		if (this.state === "PENDING") {
			return true;
		}
		if (!$.isArray(this.events) || this.events.length === 0) {
			return false;
		}*/
		return true;
	};
}());
