/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global inConcert */

if (!this.inConcert) {
	this.inConcert = {};
}
if (!inConcert.interactions) {
	inConcert.interactions = {};
}

/**
 * @namespace
 * @name InstagramInteraction
 */
(function () {

	"use strict";
	const NOT_EXIST_IN_ARRAY = -1;

	/**
	 * Class InstagramInteraction.
	 * @param {Object} rawInteraction
	 */
	function InstagramInteraction (rawInteraction)
	{
		InstagramInteraction.uber.constructor.call(this, rawInteraction);
		this.events = this.events || [];

		// Pending interactions have no properties yet. Provide some sane defaults
		if (this.isBeingComposed()) {
			this.properties = this.properties || {};
			this.properties.display_name = "";
			this.properties.is_private_message = false;
			this.properties.summary = "";
		}
	}

	inConcert.Utils.inherit(InstagramInteraction, inConcert.interactions.base);

	if (typeof inConcert.interactions.instagram !== "function") {
		inConcert.interactions.instagram = InstagramInteraction;
	}

	InstagramInteraction.prototype.inflate = function (callback)
	{
		var self = this;
		InstagramInteraction.uber.inflate.call(this, function(data){
			if(data) {
				// Building the events hierarchy has the side-effect of setting
				// createdTime in the events
				// TODO: do that in the server
				self.buildEventsHierarchy();
				self.sortEvents();
				callback(data);
			}
		});
	};

	//sends a message on the current conversation
	InstagramInteraction.prototype.sendMessage = function(/*message*/) { };

	InstagramInteraction.prototype.get_pending_origin_post = function()
	{
		if ( this.isPrivateMessage() ) {
			return;
		}
		var pending_events = this.get_pending_events();
		if ( Object.keys( pending_events ).length < 1 ) {
			return;
		}
		var origin_post;
		$.each( pending_events, function( _, pending_event )
		{
			if ( pending_event.is_origin_post ) {
				origin_post = pending_event;
				return false; // break
			}
		});
		return origin_post;
	};

	/*
		Esta funcion obtiene el evento/mensaje/post que ha orignado la interaccion
	*/
	InstagramInteraction.prototype.get_origin_message_id = function()
	{
		if (!$.isArray(this.events) || this.events.length === 0) {
			return;
		}
		var origin_message_id = this.interactionData && this.interactionData.postId;
		if (!origin_message_id) {
			origin_message_id = this.events[0].message.id;
		}
		return origin_message_id;
	};

	/*
		Funcion que ordena los eventos primero el evento orgigen, luego por fecha de creacion
	*/
	InstagramInteraction.prototype.__sort_by_creation_date = function()
	{
		var origin_message_id = this.get_origin_message_id();

		this.events.sort( function( a, b )
		{
			if (a.message && a.message.id === origin_message_id) {
				// En la primer posicion agrego el mensaje original
				return -1;
			}
			// Luego ordeno por fecha de creacion
			if (!a.original_timestamp || !b.original_timestamp) {
				return a.__index - b.__index;
			}
			if (a.original_timestamp === b.original_timestamp) {
				return a.__index - b.__index;
			}
			return a.original_timestamp - b.original_timestamp;
		});
	};

	/**
	* Build post and comments tree ordering by root post creation date.
	* This function must be called before sortEvents.
	*/
	//
	InstagramInteraction.prototype.buildEventsHierarchy = function() {
		this.assert_inflated();

		var rootNodesIndices = {};
		var knownIds = [];
		var interactionEvents = this.getEventList();

		if (!interactionEvents) {
			return;
		}
		if (interactionEvents && interactionEvents.length) {
			var events = interactionEvents.length;

			for (var i = 0; i < events; i++) {
				var event = interactionEvents[i];
				event.children = [];
				if (!event.message) {
					continue;
				}
				var dateArray = event.message.created_time.match(/(\d\d\d\d)-(\d\d)-(\d\d)[T ](\d\d):(\d\d):(\d\d)/);
				var dateString = dateArray[1].toString() + "/" + dateArray[2].toString() + "/" + dateArray[3].toString() + " " + dateArray[4].toString() + ":" + dateArray[5].toString() + ":" + dateArray[6].toString() +  " GMT-0000";
				event.creationDate = new Date(dateString);
				if (i === 0) {
					event.parentId = null;
					event.message.parent_id = null;
				}				
				if (event.parentId && rootNodesIndices.hasOwnProperty(event.parentId)) {
					/* WARNING / TODO
						-> instagram puede estar encolando comentarios repetidos por un problema en el proxy
						esto no deberia estar aca nunca ya que esta rota el *api* que se expone el cleanup deberia estar server side
					*/
					if ( knownIds.indexOf(event.message.id) === NOT_EXIST_IN_ARRAY ) { // not repeated
						//set parent
						event.parentIndex = rootNodesIndices[event.parentId];
						//tell parent to add current as children
						var parent = interactionEvents[event.parentIndex];
						parent.children.push(i);

						//add to known list
						knownIds.push(event.message.id);
					}
				}
				else {
					rootNodesIndices[event.message.id] = i;
				}
			}
		}
	};

	//------------------
	// http://fbdevwiki.com/wiki/Error_codes
	//
	InstagramInteraction.prototype.get_error_message = function( error_code, type )
	{
		switch ( (error_code || "" ).toUpperCase() )
		{
		// SERVER ERRORS!!
			case "MISSING_FIELD":
				if ( type === "NEW_POST" ) {
					return "<%=_T('The post field is mandatory')%>";
				}
				else if ( type === "NEW_PRIVATE_MESSAGE" ) {
					return "<%=_T('The private message field is mandatory')%>";
				}
				else if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('The post comment field is mandatory')%>";
				}
				break;
						
			case "POST_CLOSED":
				if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('Unable to send message. Related post was closed or deleted.')%>";
				}
				break;

			case "IRSERVER_ERROR":
				if ( type === "NEW_POST" ) {
					return "<%=_T('There was an error while creating the new post')%>";
				}
				else if ( type === "NEW_PRIVATE_MESSAGE" ) {
					return "<%=_T('There was an error while creating the new private message')%>";
				}
				else if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('There was an error while creating the new post comment')%>";
				}
				else if ( type === "LIKE" ) {
					return "<%=_T('There was an error while liking/unliking')%>";
				}
				break;

			case "NOT_LICENSED":
				if ( type === "NEW_POST" ) {
					return "<%=_T('Unable to send instagram new post. Not has licensed for the Instagram channel.')%>";
				}
				else if ( type === "NEW_PRIVATE_MESSAGE" ) {
					return "<%=_T('Unable to send instagram private message. Not has licensed for the Instagram channel.')%>";
				}
				else if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('Unable to send instagram comment. Not has licensed for the Instagram channel.')%>";
				}
				else if ( type === "LIKE" ) {
					return "<%=_T('Unable to like/unlike. Not has licensed for the Instagram channel.')%>";
				}
				break;

			case "TIMEOUT":
				return "<%=_T('Waiting time expired to send message')%>";

			case "CHATSERVER_NOT_AVAILABLE":
				return "<%=_T('Chatserver not available')%>";

			case "INTERACTION_ALREADY_EXISTS":
				return "<%=_T('There was an error while creating the conversation')%>";

			case "AGENT_UNAVAILABLE":
				return "<%=_T('There was an error while creating the conversation')%>";

			case "FAIL_CREATING_OUTGOING":
				return "<%=_T('There was an error while creating the conversation')%>";

			case "INTERACTION_CANCELED_BEFORE_TAKING":
				return "<%=_T('The conversation was cancelled')%>";

			case "FORBIDDEN":
				return "<%=_T('You not have permission for create outgoing instagram conversation.')%>";


	//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	// INSTAGRAM API ERRORS --> http://fbdevwiki.com/wiki/Error_codes
	//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

		//---------------
		// General Errors
		// ( 0 al 18 )
			case "1":     // 1 API_EC_UNKNOWN                   An unknown error occurred
				return "<%=_T('instagram error: An unknown error occurred')%>";

			case "2":     // 2 API_EC_SERVICE                   Service temporarily unavailable
				return "<%=_T('instagram error: Service temporarily unavailable')%>";

			case "4":     // 4  API_EC_TOO_MANY_CALLS           Application request limit reached
				return "<%=_T('instagram error: Application request limit reached')%>";

			case "10":    // 10 API_EC_PERMISSION_DENIED        Application does not have permission for this action
				return "<%=_T('instagram error: Application does not have permission for this action')%>";

		//---------------
		// Parameter Errors
		// ( 100 al 190 )
			case "100":   // 100 API_EC_PARAM                   Invalid parameter ( Error finding the requested story ? )
				return "<%=_T('instagram error: Error finding the requested story')%>";

			case "190":   // 190 API_EC_PARAM_ACCESS_TOKEN      Invalid OAuth 2.0 Access Token
				return "<%=_T('instagram error: Invalid OAuth Access Token')%>";

		//---------------
		// User Permission Errors
		// ( 200 al 299 )
			case "200":  // 200 API_EC_PERMISSION               Permissions error
				return "<%=_T('instagram error: Permissions error')%>";

			case "230":  // 230 API_EC_PERMISSION_MESSAGE       Permissions disallow message to user
				return "<%=_T('instagram error: Permissions disallow message to user')%>";

	// DEFAULT errors
			default:
				if ( type === "NEW_POST" ) {
					return "<%=_T('There was an error while creating the conversation')%>";
				}
				else if ( type === "NEW_PRIVATE_MESSAGE" ) {
					return "<%=_T('There was an error while creating the private message')%>";
				}
				else if ( type === "NEW_COMMENT_POST" ) {
					return "<%=_T('There was an error while creating the new post comment')%>";
				}
				break;
		}

		return "<%=_T('There was an error while sending message')%>";
	};

	InstagramInteraction.prototype.commentPost = function( fake_event_id, postId, commentId, commentMessage, callback ) 
	{
		var self = this;
		callback = callback || $.noop;
		fake_event_id = fake_event_id || inConcert.Utils.generateGuid();
		
		if (this.interactionData.subtype === "MentionComment") {
			commentId = this.interactionData.commentId;
			commentMessage.isMention = true;
		}
		else if (this.interactionData.subtype === "MentionPost") {
			commentMessage.isMention = true;
		}

		if (commentId) {
			commentId = commentId.toString();
		}

		inConcert.logger.action({action_name : "comment_instagram_post", id : this.id, type : this.type,
			campaign : this.campaign, postId : postId, account : this.account, action_type : "conversation", fake_event_id: fake_event_id});

		var url = this.m_manager.getBaseUrl() + "/interactions/instagram/comment_post/";
		var request = {
			fake_event_id : fake_event_id,
			interactionId : this.id,
			postId        : postId,
			commentId     : commentId,
			campaign      : this.campaign,
			message       : commentMessage,
			account       : this.account
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				var message;
				var status = result.result === "OK";
				if ( result.result !== "OK" ) {
					message = self.get_error_message( result.result, "NEW_COMMENT_POST" );
				}
				callback( status, message );
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "new_instagram_comment_post", error_message : error.result });
				var message = self.get_error_message( error.result, "NEW_COMMENT_POST" );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};

	InstagramInteraction.prototype.newPost = function( fake_event_id, destination, postText, destinationType, destinationName, callback ) {
		var self = this;

		fake_event_id = fake_event_id || inConcert.Utils.generateGuid();

		inConcert.logger.action({action_name : "create_instagram", id : self.id, type : self.type, campaign : self.campaign, destination : destination, action_type : "conversation", fake_event_id: fake_event_id});

		if (this.state === "PENDING") {

			var url = self.m_manager.getBaseUrl() + "/interactions/instagram/create_outgoing/";
			var request = {
				fake_event_id : fake_event_id,
				id : self.id,
				parentInteractionId : this.parentInteractionId,
				campaign : self.campaign,
				account : this.account,
				destination : destination,
				destinationType : destinationType,
				destinationName : destinationName
			};
			inConcert.invoke({
				type: "POST",
				url: url,
				data: JSON.stringify(request),
				dataType: "json",
				contentType: "application/json",
				success: function() {
					//200 OK
					self._newPost(fake_event_id, destination, postText, callback);
				},
				error: function(_, error){
					inConcert.logger.action({action_name : "create_instagram", error_message : error.result });
					if (typeof(callback) !== "function") {
						return;
					}
					var message = self.get_error_message( error.result, "NEW_POST" );
					callback(false, message, error);
				}
			});
		}
		else {
			this._newPost(fake_event_id, destination, postText, callback);
		}

		return fake_event_id;
	};

	//creates a new post related to this conversation
	InstagramInteraction.prototype._newPost = function( fake_event_id, destination, postText, callback ) 
	{
		var self = this;
		inConcert.logger.action({action_name : "new_instagram_post", id : this.id, type : this.type,
			campaign : this.campaign, destination : destination, account : this.account, action_type : "conversation"});

		var url = this.m_manager.getBaseUrl() + "/interactions/instagram/new_post/";
		var request = {
			fake_event_id : fake_event_id,
			interactionId : this.id,
			campaign : this.campaign,
			destination : destination,
			text : postText,
			account : this.account
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				if (typeof(callback) === "function") {
					callback(true, "OK");
				}
			},
			error: function(_, error){
				inConcert.logger.action({action_name : "new_instagram_post", error_message : error.result });
				if (typeof(callback) !== "function") {
					return;
				}
				var message = self.get_error_message( error.result, "NEW_POST" );
				callback(false, message, error);
			}
		});
	};

	InstagramInteraction.prototype.newPrivateMessage = function( fake_event_id, page, message, callback )
	{
		var self = this;

		callback = callback || $.noop;
		fake_event_id = fake_event_id || inConcert.Utils.generateGuid();

		inConcert.logger.action({
			action_name : "new_instagram_private_message",
			id : this.id,
			type : this.type,
			campaign : this.campaign,
			fake_event_id: fake_event_id
		});

		var url = this.m_manager.getBaseUrl() + "/interactions/instagram/new_private_message/";
		var request = {
			fake_event_id : fake_event_id,
			interactionId : this.id,
			campaign      : this.campaign,
			account       : this.account,
			page          : page || null,
			message       : message
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback( true, "OK" );
			},
			error: function(_, error){
				inConcert.logger.action({ action_name : "new_instagram_private_messaage", error_message : error.result });
				var message = self.get_error_message( error.result, "NEW_PRIVATE_MESSAGE" );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};

	InstagramInteraction.prototype.like = function(like, event_index, callback) {
		var self = this;
		callback = callback || $.noop;
		var fake_event_id = inConcert.Utils.generateGuid();

		inConcert.logger.action({
			action_name : "instagram_like",
			id : this.id,
			type : this.type,
			campaign : this.campaign,
			fake_event_id: fake_event_id
		});

		var event = this.getEvent(event_index);
		var page = this.interactionData && this.interactionData.page || event && event.message.page;
		if (!event || !event.message || !event.message.id || !page) {
			inConcert.logger.action({ action_name : "instagram_like", error_message : "Event not found" });
			var message = self.get_error_message( "IRSERVER_ERROR", "LIKE" );
			callback( false, message, "Event not found" );
			return;
		}

		var url = this.m_manager.getBaseUrl() + "/interactions/instagram/like_message/";
		var request = {
			fake_event_id : fake_event_id,
			interactionId : this.id,
			campaign      : this.campaign,
			account       : this.account,
			page          : page,
			recipientId   : event.message.from.id,
			messageId     : event.message.id,
			like          : like
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback( true, "OK" );
			},
			error: function(_, error){
				inConcert.logger.action({ action_name : "instagram_like", error_message : error.result });
				var message = self.get_error_message( error.result, "LIKE" );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};

	// Just returns the summary
	InstagramInteraction.prototype.getExtendedInfo = function( max_length )
	{
		return inConcert.Utils.truncateWord( this.properties.summary, max_length );
	};

	InstagramInteraction.prototype.__getReferralMessage = function(message) {
		switch(message.source) {
			case "ADS":
				return "<%=_T('Clicked on messenger ad.')%>";
			case "MESSENGER_CODE":
				return "<%=_T('Scanned a messenger code.')%>";
			case "DISCOVER_TAB":
				return "<%=_T('Open bot in messenger discover tab.')%>";
			case "SHORTLINK":
			default:
				return "<%=_T('Open a messenger short link.')%>";
		}
	};

	InstagramInteraction.prototype.__getOptinMessage = function(/*message*/) {
		return "<%=_T('Clicked on Send-to-messenger button.')%>";
	};

	InstagramInteraction.prototype.__getWebhookMessageContent = function(message) {
		if (message.webhookType === "REFERRAL") {
			return {
				type : "referral",
				text : this.__getReferralMessage(message)
			};
		}
		if (message.webhookType === "OPTIN") {
			return {
				type : "optin",
				text : this.__getOptinMessage(message)
			};
		}
		return {
			type : "default",
			text : message.text || message.message || message.attachment && message.attachment.payload && 
				message.attachment.payload.text || ""
		};
	};

	InstagramInteraction.prototype.getEventShortDescription = function(index) {
		var message = this.getMessageContent(index);
		return this.getMessageText(message);
	};

	InstagramInteraction.prototype.UpdatePostFirstMediaData = function(message) {
		let result_message = message;

		let media = message.id;
		const isCarousel = message.media_type === "CAROUSEL_ALBUM" && message.children;
		if (isCarousel) {
			media = message.children.map(obj => obj.id);
		}

		inConcert.invoke({
			url:       inConcert.app.getBaseUrl() + "/interactions/instagram/get_media_data/",
			type:      "POST",
			async: false,
			data: JSON.stringify({
				media: media,
				page: this.interactionData && this.interactionData.page
			}),
			dataType: "json",
			contentType: "application/json",
			timeout: 30000,
			success:   function (response)
			{ 
				const list_media_data = response.media_data && Object.values(response.media_data);
				if (!list_media_data || list_media_data.length === 0) {
					return;
				}

				if (isCarousel) {
					result_message.children = list_media_data;
				}
				else{
					const media_data = list_media_data[0];
					result_message.media_url = media_data.media_url;
					result_message.permalink = media_data.permalink;
				}
			},
			error: function(/*_, error*/){}
		});

		return result_message;
	};

	InstagramInteraction.prototype.getMessageContent = function(index)
	{
		this.assert_inflated();
		var data = { type : "default", text : "" };
		if (!this.events || !this.events.length || index >= this.events.length) {
			return data;
		}

		var message = this.events[index].message;
		if (message.type === "webhookMessage") {
			return this.__getWebhookMessageContent(message);
		}

		data.text = message.text || message.message;
		data.type = message.type;
		data.media = [];
		
		//Si es post y el index es 0, se obtiene la data de los media. Esto pq despues de X tiempo, expira la firma de los media y no se visualizan.
		if (this.isSNPost() && index === 0 && message.media_url && message.media_url !=="") {
			message = this.UpdatePostFirstMediaData(message);
		}
		
		(function processMessage(message) {
			switch(message.media_type) {
				case "IMAGE":
					data.media.push({
						type : "IMAGE",
						picture : message.media_url
					});
					break;
				case "VIDEO":
					data.media.push({
						type : "VIDEO",
						picture : message.thumbnail_url,
						link : message.media_url
					});
					break;
				case "CAROUSEL_ALBUM":
					$.each(message.children || [], function(_, child) {
						processMessage(child);
					});
					break;
			}
		})(message);
		return data;
	};

	/**
	* @param index
	*/
	InstagramInteraction.prototype.getDateObject = function(index) {
		this.assert_inflated();

		index = index || 0;
		if ( !this.events || this.events.length === 0 || !this.events[ index ] ) {
			return "";
		}
		var event_date = this.events[ index ].creationDate;
		return this.convertDateStringToDate( event_date );
	};

	InstagramInteraction.prototype.getRawAddress = function(index) {
		this.assert_inflated();
		return this.events[index].message.from.name;
	};

	InstagramInteraction.prototype.getSource = function( sorted_index )
	{
		this.assert_inflated();
		if ( !this.events || sorted_index === undefined || sorted_index === -1 ) {
			return "";
		}
		var event = this.events[ sorted_index ];
		if ( event.creator === "agent" ) {
			return event.agent || inConcert.app.getCompleteId();
		}
		else {
			var from = event.message.from;
			return from.name;
		}
	};

	InstagramInteraction.prototype.getEventUserAvatarURI = function( event )
	{
		if (typeof(event) !== "object" || typeof(event.message) !== "object") {
			return "";
		}
		var user = event.message.from;
		if (typeof(user) !== "object") {
			return "";
		}
		if ( event.creator === "agent" && event.agent ) {
			var user_id = event.agent;
			if ( !user_id ) {
				user_id = inConcert.app.getCompleteId();
			}
			if ( user_id.match(inConcert.app.idAtVCCRegExp) ) {
				user_id = RegExp.$1;
			}
			return inConcert.app.getBaseUrl() + "/users/get_avatar/" + user_id;
		}
		else if ( event.creator === "bot" ) {
			return "/inconcert/apps/agent/images/botAvatar.png";
		}
		else if (!event.outgoing_pending && user.id) {
			return inConcert.app.getBaseUrl() + "/interactions/instagram/get_avatar/" + 
				"?campaignId=" + encodeURIComponent(this.campaign) + 
				"&account=" + encodeURIComponent(this.account) + 
				"&page=" + encodeURIComponent(this.getPage()) + 
				"&userId=" + encodeURIComponent(user.id);
		}
		else {
			return "";
		}
	};

	//returns the conversation contact or member display text
	InstagramInteraction.prototype.getUserName = function( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}

		if ( event.creator === "agent" && event.agent ) {
			return this.get_agent_name( event );
		}
		if ( event.creator === "bot" ) {
			return this.getBotName();
		}
		else {
			if ( typeof( event.message ) !== "object" ) {
				return "";
			}
			var user = event.message.from;
			return user.label || user.name || user.username;
		}
	};

	InstagramInteraction.prototype.get_user_event_id = function( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}
		if ( event.creator === "agent" ) {
			return event.agent || inConcert.app.getCompleteId();
		}
		if ( typeof( event.message ) !== "object" ) {
			return "";
		}
		var user = event.message.from;
		return user.name;
	};

	/**
	* Returns a displayable name, which was pre computed on the server. It extracts the user name from the first event 
	* of the interaction or the agent id, depending on which one originated the interaction.
	* The computed property has a prefixed value to it. If the prefix is "text:", then the rest is to be taken 
	* verbatim. It is the name of a instagram user.
	* If the prefix is "agent_id:", then we need to perform a lookup to translate the id to an agent name.
	* 
	* @return {String}
	*/
	InstagramInteraction.prototype.getDisplayName = function()
	{
		var display_name = this.properties.display_name;
		if(display_name.match(/text:(.+)/)) {
			return RegExp.$1;
		}
		else if(display_name.match(/agent_id:/)) {
			var agent_id = RegExp.$1;
			if (agent_id === inConcert.app.getCompleteId()) {
				return inConcert.app.getLoggedUserName();
			}
			else {
				var agent_name;

				inConcert.app.getTeamMember( agent_id, function(agent)
					{
						if (!agent) {
							return;
						}
						// Esto por qué carajo no lo hago una vez sola?
						var have_name = false;
						if ((agent.Name || "").trim().length > 0) {
							agent_name = agent.Name;
							have_name = true;
						}
						if ((agent.FirstSurname || "").trim().length > 0) {
							agent_name = (have_name ? (agent_name + " ") : "") + agent.FirstSurname;
						}
					},
					true // from cache
				);

				agent_name = agent_name || agent_id;
				return agent_name;
			}
		}
		return display_name;
	};

	InstagramInteraction.prototype.getFirstUnreadEventIndex = function() {
		this.assert_inflated();

		var eventIndex = inConcert.interactions.base.prototype.getFirstUnreadEventIndex.call(this);
		if (typeof(eventIndex) === "number" && this.events[eventIndex]) {
			var event = this.events[eventIndex];
			return (typeof(event.parentIndex) !== "undefined") ? event.parentIndex : eventIndex;
		}
		else {
			return eventIndex;
		}
	};
	
	/* @return {Boolean} */
	// IB: ESTA FUNCION SE PUEDE CALCULAR UNA VEZ EN EL SERVER Y LISTO
	InstagramInteraction.prototype.eventsAreFromTheSameClient = function() {
		this.assert_inflated();

		var events_from_the_same_client = true;
		var last_event_client = null;

		this.events.forEach(function(current_event){
			var current_event_client = current_event.message.from.name;
			// first iteration
			if (!last_event_client){
				last_event_client = current_event_client;
			} else {
				if (current_event_client !== last_event_client){
					// IB: aca por que no para de iterar?
					// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/forEach
					// hay que usar Array.every o Array.some
					events_from_the_same_client = false;
				}
			}
		});
		return events_from_the_same_client;
	};

	//called by browser_windows.js to differentiate between posts and private messages
	InstagramInteraction.prototype.isPrivateMessage = function() {
		//return true if at least one event is a private message
		return this.properties.is_private_message;
	};

	InstagramInteraction.prototype.isSNPost = function() {
		return !this.isPrivateMessage() && this.interactionData && this.interactionData.postId;
	};

	InstagramInteraction.prototype.getMessageText = function(message) {
		if (message.webhookType === "REFERRAL") {
			return this.__getReferralMessage(message);
		}
		if (message.webhookType === "OPTIN") {
			return this.__getOptinMessage(message);
		}
		return message.text || message.message ||
			message.attachment && message.attachment.payload && message.attachment.payload.text || "";
	};

	//returns the extra information
	InstagramInteraction.prototype.getMessage = function(index) {
		this.assert_inflated();

		if (this.events && this.events.length) {
			return this.getMessageText(this.events[index].message);
		}
		return "";
	};
	
	InstagramInteraction.prototype.getFiles = function(event) {
		let message = event && event.message || {};
		return message.files? message.files : message.file? [message.file] : null;
	};

	InstagramInteraction.prototype.hasAttachments = function (event) {
		// IB: Esto se puede cambiar por Object.keys()
		var isEmpty = function (obj) {
			for(var prop in obj) {
				if (obj.hasOwnProperty(prop)) {
					return false;
				}
			}
			return true;
		};
		
		if (event) {
			return event.message && event.message.attachment && !isEmpty(event.message.attachment);
		}
		
		return InstagramInteraction.uber.hasAttachments.call(this);
	};

	InstagramInteraction.prototype.isLiked = function (index) {
		var event = this.getEvent(index);
		var likes = this.interactionData && this.interactionData.likes || {};
		return event && event.message && event.message.id && likes[event.message.id] || false;
	};
	
	InstagramInteraction.prototype.getPage = function() {
		return this.interactionData && this.interactionData.page || null;
	};

	InstagramInteraction.prototype.isValid = function() {
		/*this.assert_inflated();
		// IB: Revisar cual es el sentido de esto

		if (this.state === "PENDING") {
			return true;
		}
		if (!$.isArray(this.events) || this.events.length === 0) {
			return false;
		}*/
		return true;
	};
	
	InstagramInteraction.prototype.canPrivateReply = function(index) {
		let subtype = this.getSubtype();
		if (this.isPrivateMessage() || subtype === "FanPageRateFeed" || subtype !== "NewPostComment" && index > 0) {
			return false;
		}
		let event = this.getEvent(index);
		if (!event || !event.message || !event.message.id || !event.message.from || !event.message.from.name) {
			return false;
		}
		if (event.creator !== "contact") {
			return false;
		}
		return !this.interactionData || !this.interactionData.repliedObjects || 
			!this.interactionData.repliedObjects[event.message.id];
	};

	InstagramInteraction.prototype.privateReply = function( index, message, callback ) 
	{
		callback = callback || $.noop;
		let self = this;
		let fake_event_id = inConcert.Utils.generateGuid();
		if (!this.canPrivateReply(index)) {
			return callback(false, "<%=_T('Error replying message.')%>");
		}
		let event = this.getEvent(index);
		
		inConcert.logger.action({action_name : "instagram_private_reply", id : this.id, type : this.type,
			campaign : this.campaign, eventId : event.message.id, account : this.account, action_type : "conversation", fake_event_id: fake_event_id});

		let url = this.m_manager.getBaseUrl() + "/interactions/instagram/private_reply/";
		var request = {
			fake_event_id : fake_event_id,
			campaignId	  : this.campaign,
			account       : this.account,
			page          : this.getPage(),
			interactionID : this.id,
			object        : event,
			index		  : index,
			message		  : message
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				var message;
				var status = result.result === "OK";
				if (!status) {
					message = self.get_error_message( result.result, "PRIVATE_REPLY", result );
				}
				callback( status, message, result.result );
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "instagram_private_reply", error_message : error.result });
				var message = self.get_error_message( error.result, "PRIVATE_REPLY", error );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};
}());
