/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, quotmark:double, esversion: 11 */

/* global $, inConcert, _ */

if (!this.inConcert) {
	this.inConcert = {};
}
if (!inConcert.interactions) {
	inConcert.interactions = {};
}

/**
 * @namespace
 * @name MailInteraction
 */
(function () {

	"use strict";

	var format = inConcert.Utils.format;
	var esc_attr   = inConcert.Utils.escapeAttrib;
	var nl2br      = inConcert.Utils.nl2br;
	var html2text  = inConcert.Utils.html2text;
	const default_send_mail_error = "<%=_T('There was an error while creating the conversation, please try again later')%>";

	/**
	 * Class MailInteraction.
	 * @param {Object} rawInteraction
	 */
	function MailInteraction (rawInteraction)
	{
		MailInteraction.uber.constructor.call(this, rawInteraction);
		//if no events present it's assumed to be an outgoing mail being created
		if (!this.events) {
			this.events = [];
		}
		// Pending interactions have no properties yet. Provide some sane defaults
		if (this.isBeingComposed()) {
			this.properties = this.properties || {};
			this.properties.human_readable_address = "";
			this.properties.has_attachments = false;
			this.properties.summary = "";
		}
	}

	inConcert.Utils.inherit(MailInteraction, inConcert.interactions.base);

	if (typeof inConcert.interactions.mail !== "function") {
		inConcert.interactions.mail = MailInteraction;
	}

	MailInteraction.prototype.inflate = function (callback)
	{
		var self = this;
		MailInteraction.uber.inflate.call(this, function(data) {
			if(data) {
				self.sortEvents();
				callback(data);
			}
		});
	};

	MailInteraction.prototype.send = function(message, callback)
	{
		var self = this;
		callback = callback || $.noop;

		// if conversation is already created
		if ( this.state !== "PENDING" ) {
			return this.sendMail(message, callback);
		}

		if (this.sending) {
			callback(false, "<%=_T('Cannot send e-mail. Current e-mail being sent.')%>"); //not successful
			return;
		}
		this.sending = true;

		message.body = inConcert.Utils.makeBodyParts(message.body);

		// new mail from scratch
		var mail = {
			id                  : this.id,
			parentInteractionId : this.parentInteractionId,
			campaign            : this.campaign,
			account             : message.account,
			attention_level_id	: message.attention_level_id,
			contact             : message.contact,
			from                : message.from,
			to                  : message.to,
			cc                  : message.cc,
			bcc                 : message.bcc,
			subject             : message.subject,
			body                : message.body,
			attachments         : message.attachments
		};

		inConcert.invoke({
			type: "POST",
			timeout : 600 * 1000, // 600 seconds, this should go hand in hand with the value set on nginx
			url: this.m_manager.getBaseUrl() + "/interactions/mail/create_and_send/",
			data: JSON.stringify(mail),
			dataType: "json",
			contentType: "application/json",
			success: function(_) {
				delete self.sending;
				callback(true);
			},
			//"encodedError" contains a description of the error and a reason,
			error: function(_, error) {
				delete self.sending;
				inConcert.logger.action({action_name : "create_mail", error_message : error.result });
				let message_err;
				switch (error.result) {
					case "CHATSERVER_NOT_AVAILABLE":
						message_err = "<%=_T('Chatserver not available')%>";
						break;

					case "NOT_LICENSED":
						message_err = "<%=_T('Unable to send e-mail. Not has licensed for the e-mail channel.')%>";
						break;

					case "INTERACTION_ALREADY_EXISTS":
						message_err = default_send_mail_error;
						break;

					case "AGENT_UNAVAILABLE":
						message_err = "<%=_T('Unable to send e-mail. You have reached the limit of your attention capability.')%>";
						break;

					case "FAIL_CREATING_OUTGOING":
						message_err = default_send_mail_error;
						break;

					case "INTERACTION_CANCELED_BEFORE_TAKING":
						message_err = "<%=_T('The conversation was cancelled')%>";
						break;

					case "TO_ADDRESS_IS_EMPTY":
						message_err = "<%=_T('Error sending e-mail. TO e-mail address is empty.')%>";
						break;

					case "TO_ADDRESS_EXCEEDED_LIMIT_OF_ALLOWED_ADDRESSES":
						message_err = "<%=_T('Error sending e-mail. TO address has exceeded the limit of allowed addresses.')%>";
						break;

					case "FORBIDDEN":
						message_err = "<%=_T('You not have permission for create outgoing e-mail conversation.')%>";
						break;

					case "BLACKLIST":
						message_err = "<%=_T('Blacklisted addresses')%><%=_T(':')%> " +
							(error.addresses || []).join(", ");
						break;

					default:
						message_err = default_send_mail_error;
				}
				callback(false, message_err); //not successful
			}
		});
	};

	MailInteraction.prototype.sendMail = function(mail, callback)
	{
		var self = this;
		callback = callback || $.noop;

		if (this.sending) {
			callback(false, "<%=_T('Cannot send e-mail. Current e-mail being sent.')%>"); //not successful
			return;
		}
		this.sending = true;

		//responses go thru the same conversation id and campaign of the replied mail
		mail.id = this.id;
		mail.body = inConcert.Utils.makeBodyParts(mail.body);

		inConcert.invoke({
			type: "POST",
			timeout : 600 * 1000, // 600 seconds, this should go hand in hand with the value set on nginx
			url: this.m_manager.getBaseUrl() + "/interactions/mail/send_message/",
			data: JSON.stringify(mail),
			dataType: "json",
			contentType: "application/json",
			success: function(_) {
				delete self.sending;
				callback(true, "Message sent"); //successfully
			},
			error: function(_, error) {
				delete self.sending;
				inConcert.logger.action({action_name : "send_mail", error_message : error.result });
				var message;
				switch (error.result) {
					case "NOT_LICENSED":
						message = "<%=_T('Unable to send e-mail. Not has licensed for the e-mail channel.')%>";
						break;
					case "INTERACTION_NOT_FOUND":
						message = default_send_mail_error;
						break;
					case "INTERACTION_NOT_TAKEN":
						message = "<%=_T('Unable to send e-mail. The conversation is not taken.')%>";
						break;
					case "INTERACTION_NOT_TAKEN_BY_AGENT":
						message = "<%=_T('Unable to send e-mail. The conversation is not taken by you.')%>";
						break;
					case "CANNOT_DISPATCH_MAIL":
						message = "<%=_T('Cannot dispatch e-mail')%>";
						break;
					case "INVALID_PARAMETERS":
						message = "<%=_T('The mailing list must have valid addresses.')%>";
						break;
					case "BLACKLIST":
						message = "<%=_T('Blacklisted addresses')%><%=_T(':')%> " +
							(error.addresses || []).join(", ");
						break;
					default:
						message = default_send_mail_error;
				}
				callback(false, message); //not successful
			}
		});
	};

	MailInteraction.prototype.canMoveToFolder = function()
	{
		var campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		var account_id = this.getAccount();
		if (!campaign.can_move_folder_mail_account(account_id)) {
			return false;
		}
		if (!this.hasIncomingEvents()) {
			return false;
		}
		return true;
	};

	MailInteraction.prototype.loadFoldersAccount = function(callback)
	{
		callback = callback || $.noop;

		var uri = this.m_manager.getBaseUrl() + "/interactions/mail/load_account_folders/{0}/{1}";
		uri = format(uri, encodeURIComponent(this.campaign), encodeURIComponent(this.account));

		inConcert.invoke({
			type: "GET",
			timeout : 600 * 1000, // 600 seconds, this should go hand in hand with the value set on nginx
			url: uri,
			dataType: "json",
			contentType: "application/json",
			success: function(result) {
				callback(true, result);
			},
			error: function(_, error) {
				inConcert.logger.action({action_name : "loadFoldersAccount", error_message : error.result });
				var message;
				switch (error.result) {
					case "NOT_LICENSED":
						message = "<%=_T('Unable to load folders. Not has licensed for the E-mail channel.')%>";
						break;
					default:
						message = "<%=_T('There was an error while loading folders, please try again later')%>";
				}
				callback(false, message); //not successful
			}
		});
	};

	MailInteraction.prototype.moveToFolder = function(params, callback)
	{
		callback = callback || $.noop;

		if (!this.canMoveToFolder()) {
			callback(false, "<%=_T('Unable to move to folder. Option not configured in e-mail account.')%>");
			return;
		}

		var last_moved_folder = this.getLastMovedFolder();
		var fromFolder = "";
		if (last_moved_folder) {
			fromFolder = last_moved_folder.to || ""; // El TO pasa a ser el FROM
		}
		var uri = this.m_manager.getBaseUrl() + "/interactions/mail/move_to_folder/";
		var post_data = {
			interaction: {id: this.id, finalization_data: this.getFinalizationData()},
			campaign   : this.campaign,
			account    : this.account,
			fromFolder : fromFolder,
			toFolder   : params.toFolder,
			appId      : params.appId,
			emailsId   : params.emailsId // optional (son los eventos a mover, si no viene del lado server se mueven todos los eventos)
		};

		if (params.labelInfo) {
			post_data.labelInfo = {
				must_label_interaction: true,
				label                 : params.labelInfo.label,
				action                : params.labelInfo.action
			};
		}

		inConcert.invoke({
			type: "POST",
			timeout : 600 * 1000, // 600 seconds, this should go hand in hand with the value set on nginx
			url: uri,
			data: JSON.stringify(post_data),
			dataType: "json",
			contentType: "application/json",
			success: function(result) {
				callback(true, result);
			},
			error: function(_, error) {
				inConcert.logger.action({action_name: "moveToFolder", error_message: error.result});
				var message;
				switch (error.result) {
					case "NOT_LICENSED":
						message = "<%=_T('Unable to move to folder. Not has licensed for the E-mail channel.')%>";
						break;
					case "NOT_HAS_PERMISSION":
						message = "<%=_T('Unable to move to folder. You not has permissions.')%>";
						break;
					default:
						message = "<%=_T('There was an error while moving to folder, please try again later')%>";
				}
				callback(false, message); //not successful
			}
		});
	};

	MailInteraction.prototype.getFolders = function()
	{
		return this.folders;
	};

	MailInteraction.prototype.getDraft = function() {
		return this.draft;
	};

	MailInteraction.prototype.setDraft = function(draft) {
		this.draft = draft;
	};

	MailInteraction.prototype.update = function(data)
	{
		// Se sobrescribe metodo update, para eliminar draft en memoria, dado que se elimina del lado del servidor y no del cliente
		// Esto ocurre ejemplo si se elimina al transferir, finalizar o reencolar.
		if (data && !data.draft && this.hasDraft()) {
			delete this.draft;
		}
		
		MailInteraction.uber.update.call(this, data);
	};

	MailInteraction.prototype.hasDraft = function() {
		return !!this.draft;
	};

	MailInteraction.prototype.saveDrafts = function(draft, callback = $.noop)
	{
		const self = this;
		if (!draft) {
			return;
		}
		const post_data = this.isBeingComposed() ?
			this.serialize(undefined, {
					"m_manager": true,
					"m_session": true,
					"m_view": true,
					"events": true
				}, false
			) :
			{ id: this.id, type: this.type };

		post_data.draft = draft;
		const uri = this.m_manager.getBaseUrl() + "/interactions/drafts/save/";
		const interactionId = this.id;
		inConcert.invoke({
			type: "POST",
			timeout : 600 * 1000, // 600 seconds, this should go hand in hand with the value set on nginx
			url: uri,
			data: JSON.stringify(post_data),
			dataType: "json",
			contentType: "application/json",
			success: function(result) {
				self.draft = draft;
				callback(true, result, interactionId);
			},
			error: function(_, error) {
				var message;
				switch (error.result) {
					
					default:
						message = "<%=_T('There was an error saving draft, please try again later.')%>";
				}
				callback(false, message); //not successful
			}
		});
	};

	MailInteraction.prototype.deleteDrafts = function(callback, isPending)
	{
		callback = callback || $.noop;
		const self = this;
		const uri = this.m_manager.getBaseUrl() + "/interactions/drafts/delete/";
		const interactionId = this.id;
		inConcert.invoke({
			type: "POST",
			timeout : 600 * 1000, // 600 seconds, this should go hand in hand with the value set on nginx
			url: uri,
			data: JSON.stringify({
				id: this.id,
				isPending: isPending || self.isBeingComposed(),
				type: this.type
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(result) {
				if (self.draft) {
					delete self.draft;
				}
				callback(true, result, interactionId);
			},
			error: function(/*_, error*/) {
				const message = "<%=_T('There was an error deleting draft, please try again later.')%>";
				callback(false, message); //not successful
			}
		});
	};

	MailInteraction.prototype.dispose = function() // Cancel only apply for Pendings
	{
		if (this.state === "CANCELED" && this.hasDraft()) {
			this.deleteDrafts(function (success, message) {
				if (!success) {
					inConcert.logger.debug(message);
				}
			}, true);
		}
		MailInteraction.uber.dispose.call(this);
	};

	MailInteraction.prototype.getLastMovedFolder = function()
	{
		var folders = this.getFolders() || [];
		if (folders.length === 0) {
			return;
		}
		var last_moved_folder = folders[folders.length - 1];
		return last_moved_folder;
	};

	MailInteraction.prototype.is_auto_response_event = function( event_index ){
		this.assert_inflated();

		event_index = event_index || 0;
		if ( !this.events || !this.events[ event_index ] ) {
			return null;
		}
		var event = this.events[ event_index ];
		return event.creator === "automatic";
	};

	/**
	This *does* include automatic responses.
	*/
	MailInteraction.prototype.getAllRelatedEventCount = function() {
		return (this.events || []).length;
	};

	MailInteraction.prototype.getNextRelatedEventIndex = function( current_event_index ) {
		this.assert_inflated();

		var self = this;

		current_event_index = current_event_index || 0;
		if ( !this.events || !this.events[ current_event_index ] ) {
			return null;
		}

		// recursive function
		var next_event = function( current ) {
			var next = current + 1;
			var is_auto = self.is_auto_response_event( next );
			if ( is_auto )  {
				return next_event( next );
			}
			else if ( is_auto === null ) {
				return null;
			}
			return next;
		};

		return next_event( current_event_index );
	};

	MailInteraction.prototype.getPreviousRelatedEventIndex = function( current_event_index ) {
		this.assert_inflated();

		var self = this;
		current_event_index = current_event_index || 0;
		if ( current_event_index === 0  || !this.events || !this.events[ current_event_index ]  ) {
			return null;
		}

		// recursive function
		var previous_event = function( current ) {
			var previous = current - 1;
			var is_auto = self.is_auto_response_event( previous );
			if ( is_auto )  {
				return previous_event( previous );
			}
			else if ( is_auto === null ) {
				return null;
			}
			return previous;
		};

		return previous_event( current_event_index );
	};

	MailInteraction.prototype.getReadEventsCount = function() {
		var self = this;
		var readers_events = 0;
		var user = inConcert.app.getCompleteId();

		if ( this.readers && this.readers[user] ) {
			$.each( this.readers[user], function( index/*, reader_event*/ ) {
				if ( self.is_auto_response_event( index ) ) {
					return true; // continue
				}
				readers_events++;
			});
		}
		return readers_events;
	};

	/* @return {Boolean} */
	MailInteraction.prototype.eventsAreFromTheSameClient = function() {
		var events_from_the_same_client = true;
		var last_event_client = null;

		this.events.forEach(function(current_event){
			var current_event_client_from = _.findWhere(current_event.addresses, {"source" : "from"}); 
			var current_event_client;
			if (current_event_client_from && current_event_client_from.address && current_event_client_from.address.displayable)  {
				current_event_client = current_event_client_from.address.displayable;
			}
			// first iteration
			if (current_event_client) {
				if (!last_event_client){
					last_event_client = current_event_client;
				} else {
					if (current_event_client !== last_event_client){
						events_from_the_same_client = false;
					}
				}
			}
		});
		return events_from_the_same_client;
	};

	MailInteraction.prototype.getCompleteAddress = function(){
		if ( typeof this.completeAddress === "object" && this.completeAddress && this.completeAddress.address) {
			return this.completeAddress.address;
		}
		return this.completeAddress || "";
	};

	MailInteraction.prototype.getHumanReadableAddress = function() {
		// IB:
		return this.properties.human_readable_address;
	};

	MailInteraction.prototype.formatAddress = function(email, name) {
		if (name && name.length > 0 && name !== email) {
			return "'" + name + "' <" + email + ">";
		}
		else {
			return email;
		}
	};

	MailInteraction.prototype.getSource = function( sorted_index ) {
		return this.getFrom(sorted_index || 0);
	};

	MailInteraction.prototype.getFrom = function(sorted_index) {
		this.assert_inflated();

		if (!$.isArray(this.events) || !this.events[sorted_index]) {
			return "";
		}
		var event = this.events[sorted_index];
		if (event.agent) {
			return event.agent;
		}
		else {
			return this.formatAddress(event.from, event.name);
		}
	};

	MailInteraction.prototype.getAddressByAccount = function() {
		var campaign = inConcert.app.getCampaign( this.campaign );
		if ( !campaign ) {
			return "";
		}
		var account_id = this.getAccount() || "";
		var account = campaign.getAccount( account_id, "MAIL" );
		if ( !account ) {
			return "";
		}
		return account.address || "";
	};

	MailInteraction.prototype.getAddress = function(index) {
		// IB: contactAddress es un buen candidato?
		if(index > 0) {
			this.assert_inflated();

			if (!this.events || !this.events.length || this.events.length  === 0) {
				return "";
			}
			else {
				return this.events[index || 0].from;
			}
		}
		return this.contactAddress;
	};

	MailInteraction.prototype.getDisplayAddress = function() {
		const visualize_address = inConcert.app.canVisualizeAddress(null, this.campaign, null, true);
		let displayAddress = this.getAddress();
		if(displayAddress !== "" && !visualize_address){
			return "<%=_T('Hidden Address')%>";
		}
		return displayAddress;
	};

	MailInteraction.prototype.getFirstIndexEvent = function()
	{
		var self = this;
		this.assert_inflated();

		if (!$.isArray(this.events)) {
			return 0;
		}
		if (!this.hasSortedEvents()) {
			this.sortEvents();
		}
		var first_sorted_index;
		$.each(this.events, function(sorted_index)
		{
			var is_auto = self.is_auto_response_event(sorted_index);
			if (!is_auto) {
				first_sorted_index = sorted_index;
				return false; // break
			}
		});
		return first_sorted_index;
	};

	MailInteraction.prototype.getDateObject = function( index ) {
		index = index || 0;
		const event = this.getEvent(index);
		if (!event) {
			return "";
		}
		const event_date = event.date;
		return this.convertDateStringToDate(event_date);
	};

	MailInteraction.prototype.getEventShortDescription = function(index) {
		return this.getSubject(index);
	};

	MailInteraction.prototype.getSubject = function(index) {
		this.assert_inflated();
		index = index || 0;
		const event = this.getEvent(index);
		if (!event) {
			return "";
		}
		return event.subject;
	};

	MailInteraction.prototype.getFromEmail = function(index) {
		this.assert_inflated();
		const event = this.getEvent(index);
		if (!event) {
			return "";
		}
		return event.from;
	};

	MailInteraction.prototype.getRawAddress = function(index) {
		this.assert_inflated();

		return this.getFromEmail(index);
	};

	/*
		content_disposition = "attachment", "inline" o undefined
			"attachment" = retorna los adjuntos de los cuales el agente realizo el upload manualmente.
			"inline"     = retorna los adjuntos que se han subido en el parseo de las imagenes embebidas en el body del mensaje.
			"undefined   = reorna los 2 anteriores.
	*/
	MailInteraction.prototype.getAttachmentsEvent = function( event_index, content_disposition )
	{
		this.assert_inflated();

		if ( typeof event_index !== "number" || !this.events ) {
			return;
		}
		var event = this.events[ event_index ];
		if ( !event ) {
			return;
		}
		var attachments = event.attachments;
		if ( !$.isArray( attachments ) ) {
			if ( typeof attachments === "string" ) {
				try {
					attachments = JSON.parse( attachments );
				}
				catch(e) { return; }
			}
			else {
				return;
			}
		}

		if ( !content_disposition ) {
			// si no especifica el content disposition, retorna todos los attachments.
			return attachments;
		}
		var attachments_by_type = [];
		$.each(attachments, function(_, attachment)
		{
			// Si el adjunto no tiene content dispositin, le asigno "attachment" para la comparacion. Esto sirve paras las interacciones viejas.
			var current_content_disposition = attachment["content-disposition"] || "attachment";
			if ( current_content_disposition === content_disposition ) {
				attachments_by_type.push(attachment);
			}
		});

		return attachments_by_type;
	};

	MailInteraction.prototype.hasAttachments = function() {
		return this.properties.has_attachments;
	};

	MailInteraction.prototype.isValid = function() {
		//this.assert_inflated();

		if (this.state === "PENDING") {
			return true;
		}
		// IB: que problema...
		/*if (!$.isArray(this.events) || this.events.length === 0) {
			return false;
		}*/
		return true;
	};

	MailInteraction.prototype.convert_string_addresses_to_object = function( addresses ) {
		var destination_list = {}; // addresses object list
		if ( typeof addresses !== "string" || addresses.length === 0 ) {
			return destination_list;
		}

		var parse_address = function( mail )
		{
			var name = "";
			var matches = /^\s*("<[^<>]*>"\s+|[^<>]*\s+)?(<[^<>\s]*>|[^<>\s]*)$/g.exec( mail );
			if ( matches ) {
				name = matches[1] || "" ;
				mail = matches[2];
				//extract mail from text, remove < and >
				mail = mail.replace( /<|>/g, "" );
			}

			destination_list[ mail.trim() /* sample@aaa.com */ ] = name /* El mail de Ellio */;
		};

		var address;
		var current_position = 0;

		while ( true )
		{
			var separator_position = addresses.indexOf( ";", current_position );
			if ( separator_position === -1 ) {
				address = addresses.substring( current_position, addresses.length );
				parse_address( address );
				break ;
			}
			else {
				address = addresses.substring( current_position, separator_position );
				parse_address( address );
				current_position = separator_position + 1;
			}
		}

		return destination_list;
	};

	MailInteraction.prototype.parseAddresses = function( addresses, returnAsTable) {
		if ( !addresses ) {
			return "";
		}
		if ( typeof(addresses) === "object" || $.isArray( addresses ) ) {
			if (returnAsTable) {
				return addresses;
			}
			var toString = "";
			for (var email in addresses) {
				if (addresses.hasOwnProperty(email)) {
					toString += toString.length > 0 ? "; " : "";
					toString += this.formatAddress(email, addresses[email]);
				}
			}
			return toString;
		}
		else if (typeof(addresses) === "string") {
			if (returnAsTable) {
				return this.convert_string_addresses_to_object( addresses );
			}
			return addresses;
		}
	};

	MailInteraction.prototype.getTo = function( index, returnAsTable ) {
		this.assert_inflated();
		index = index || 0;
		const event = this.getEvent(index);
		if (!event) {
			return "";
		}
		return this.parseAddresses( event.to, returnAsTable );
	};

	MailInteraction.prototype.getCC = function( index, returnAsTable ) {
		this.assert_inflated();
		index = index || 0;
		const event = this.getEvent(index);
		if (!event) {
			return "";
		}
		return this.parseAddresses( event.cc, returnAsTable );
	};

	MailInteraction.prototype.getBCC = function( index, returnAsTable ) {
		this.assert_inflated();
		index = index || 0;
		const event = this.getEvent(index);
		if (!event) {
			return "";
		}
		return this.parseAddresses( event.bcc, returnAsTable );
	};

	//returns the extra information
	MailInteraction.prototype.getExtendedInfo = function(eventIndex)
	{
		var info = "";

		if(!this.properties){ throw new Error("Me faltan las properties"); }

		if(eventIndex > 0) {
			this.assert_inflated();

			eventIndex = eventIndex || 0;
			if (this.events && this.events.length > 0) {
				var subject = this.events[eventIndex].subject || "";
				var summary = this.getPlainBody(eventIndex).substring(0, 150);
				var plainBody;
				// TODO for some confusing reason on some stuff html2text blows up badly
				try {
					plainBody = html2text(summary);
				}
				catch(e) {
					plainBody = "";
				}
				info = subject + "\n\r" + plainBody;
			}
		}
		else {
			// Use what we've set in the initial load
			info = this.properties.summary;
		}
		try {
			info = html2text(info);
		}
		catch(e) {
			info = "";
		}
		return info || "";
	};

	// iterates over body parts and returns what's asked for
	MailInteraction.prototype.getBodyPart = function( index, content_type ) {
		this.assert_inflated();

		if ( typeof index !== "number" || typeof content_type !== "string" ) {
			return;
		}
		if ( !this.events[ index ] ) {
			return;
		}
		var body = this.events[ index ].body;
		if ( !body ) {
			return;
		}

		content_type = content_type.toLowerCase();

		for ( var i = 0; i < body.length; i++ ) {
			var part = body[i];
			if ( part.type && part.type.toLowerCase() === content_type ) {
				return part.content;
			}
		}
	};

	/*
		returns the plain text MIME part of the email body or the string content if not MIME parts are found
	*/
	MailInteraction.prototype.getPlainBody = function(index) {
		this.assert_inflated();

		var plainBody = this.getBodyPart(index, "text/plain");
		if (!plainBody && typeof(this.events[index].body) === "string") {
			return this.events[index].body;
		}
		return plainBody || "";
	};

	/*
		returns the html text MIME part of the email body
	*/
	MailInteraction.prototype.getHTMLBody = function( index )
	{
		this.assert_inflated();

		var body_part = this.getBodyPart( index, "text/html" );

		if ( body_part ) {
			body_part = inConcert.Utils.html_remove_malicious_code( body_part );
			return body_part;
		}

		body_part = this.getBodyPart( index, "text/plain" );
		if ( body_part ) {
			body_part = nl2br( esc_attr( body_part || "" ) );
		}
		else {
			var event = this.events[ index ];
			if ( event ) {
				body_part = typeof( event.body ) === "string" ? event.body : "";
			}
		}

		return body_part || "";
	};

	MailInteraction.prototype.getEventUserAvatarURI = function( event )
	{
		var uri = inConcert.app.getBaseUrl();
		if (this.initiative === "OUTBOUND" && this.state === "PENDING") {
			return uri + "/users/get_avatar/" + inConcert.app.getLoggedUserId();
		}
		if (event.creator === "agent" && event.agent) {
			var userId = event.agent;
			if (event.agent.match(inConcert.app.idAtVCCRegExp)) {
				userId = RegExp.$1;
			}
			return uri + "/users/get_avatar/" + userId;
		}
		else if (event.creator === "automatic") {
		// creator is campaign
			var campaign = inConcert.app.getCampaign( event.name || this.campaign);
			if (campaign) {
				return uri + campaign.getAvatarURI();
			}
		}
		if ( typeof event.__index !== undefined ) {
		// if creator is a contact
			var sorted_index = this.get_sorted_index_by_original_index( event.__index );
			return uri + "/contacts/get_address_avatar/MAIL/" + this.getRawAddress( sorted_index );
		}
		return uri;
	};
}());

