/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true, 
	strict:true, undef:true, unused:vars, curly:false, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global $, _, inConcert */

if (!this.inConcert) {
	this.inConcert = {};
}

/**
 * @namespace
 * @name InteractionManager
 */
(function () {

	"use strict";
	const format = inConcert.Utils.format;
	const ASSIGNED_PANEL_ID  = 'AssignedInteractions';
	const QUEUED_PANEL_ID    = 'QueuedInteractions';
	const SPAM_PANEL_ID      = 'SpamInteractions';
	const WAITING_PANEL_ID   = 'WaitingInteractions';
	const ABANDONED_PANEL_ID = 'AbandonedInteractions';

	/**
	 * InteractionManager, entry point for creating, binding and storing the current conversations the agent is handling.
	 * @param {String} baseUrl
	 * @param {String} user
	 * @param {String} domain
	 * @param {Object} clientsSettings
	 */
	function InteractionManager(baseUrl, user, domain, clientsSettings) {
		var self = _.extend(this, {
			m_urlBase               : baseUrl,
			m_user                  : user,
			m_domain                : domain,
			m_firstTimeCheckQueued  : false,
			m_interactions          : {}, // all the conversations available
			m_queuedInteractions    : {}, // all the conversations queued
			m_abandonedInteractions : {}, // all the conversations abandoned
			m_monitoredInteractions : {}, // current monitored conversations, can receive sessions and events for this ones
			m_interactionClients    : {}  // each conversation handler indexed by conversation type
		});
		// initialize conversation physical channels according to user settings
		_.each(clientsSettings, function(clientConfig, clientChannel)
		{
			self.setChannelClient(clientChannel, clientConfig);
		});
		self.startCheckingAppointment();
	}

	//Define functions and objects in the namespace in case they don't already exist
	if (typeof inConcert.InteractionManager !== "function") {
		inConcert.InteractionManager = InteractionManager;
	}

	InteractionManager.prototype.setChannelClient = function(clientChannel, clientConfig) {
		const self = this;

		if (clientConfig && !clientConfig.hasOwnProperty("automaticPollerDisabled")) {
			clientConfig.automaticPollerDisabled = inConcert.app.isAutomaticPollerDisabled();
		}

		switch (clientChannel) {
			case "CALL":
				this.m_interactionClients.CALL = new inConcert.PhoneClient(this, clientConfig, _.bind(this.onClientChangedStatus, this, "CALL"));
				this.m_interactionClients.VOICEMAIL = inConcert.Utils.cloneObject(this.m_interactionClients.CALL);
				break;
			case "CHAT": //the client used is the same for IM and CHAT, just point them in case both are required
				let client = new inConcert.ChatClient(this, clientConfig, function(status, message) {
					self.onClientChangedStatus("IM", status, message);
					self.onClientChangedStatus("CHAT", status, message);
				});
				_.each(["CHAT", "INTERNAL_CHAT", "IM"], function(channel) {
					self.m_interactionClients[channel] = client;
				});
				break;
			case "IM":
			case "MAIL":
			case "WEBCONTACT":
			case "FACEBOOK":
			case "INSTAGRAM":
			case "YOUTUBE":
			case "APPSTORE":
			case "GOOGLEPLAY":
			case "RCS":
			case "TWITTER":
			case "SMS":
			case "WHATSAPP":
				break;
			default:
				throw ("InteractionManager - channel not recognized " + (clientChannel || ""));
		}
	};

	InteractionManager.prototype.closeAllClients = function() {
		_.each(this.m_interactionClients, function(client) {
			client.close();
		});
	};

	// Given a channel name, will try to make a relogin
	InteractionManager.prototype.reconnectClient = function(channel, on_client_reconnected)
	{
		on_client_reconnected = on_client_reconnected || $.noop;
		if (this.m_interactionClients[channel]){
			this.m_interactionClients[channel].reconnect(function(undefined){
				on_client_reconnected(true);
			});
		}
		else{
			on_client_reconnected(false);
		}
	};

	InteractionManager.prototype.getBaseUrl = function() {
		return this.m_urlBase;
	};

	/*! cancels an conversation (pending or not) */
	InteractionManager.prototype.cancelInteraction = function(interaction) {
		if (!interaction || !interaction.id) {
			return false;
		}
		console.info("[InteractionManager.prototype.cancelInteraction] Cancel interaction: " + interaction.id + ", state: " + interaction.state + "', channel: "  + interaction.type);
		interaction.cancel();
		if (!this.m_interactions[interaction.id]) {
			console.debug("[InteractionManager.prototype.cancelInteraction]  -> Unable to cancel interaction: " + interaction.id + ". Interaction does not exist.");
			return false;
		}
		const updateResult = this.updateInteraction(interaction);
		inConcert.app.getMainWindow().updateInteractions([
			updateResult || {id: interaction.id, deleted: true}
		]);
		return true;
	};

	InteractionManager.prototype.cancelCallPendingInteraction = function(currentInteraction, callback = $.noop) {
		console.info("[InteractionManager.prototype.cancelCallPendingInteraction] Cancel call pending interaction: " + currentInteraction.id);
		let mustCanceled = false;
		for (let id in this.m_interactions) {
			if (this.m_interactions.hasOwnProperty(id)) {
				let interaction = this.m_interactions[id];
				if (interaction.type === "CALL" && interaction.state === "PENDING" && interaction.id === currentInteraction.id) {
					mustCanceled = true;
					break;
				}
			}
		}
		callback(mustCanceled);
		if (mustCanceled) {
			this.cancelInteraction(currentInteraction);
		}
		else {
			console.info("[InteractionManager.prototype.cancelCallPendingInteraction] NO cancel call pending interaction: " + currentInteraction.id);
		}
	};

	InteractionManager.prototype.cancelPendingInteractions = function(proxyType, excludedIds = {}, callback = $.noop) {
		console.info("[InteractionManager.prototype.cancelPendingInteractions] Cancel all pending interactions of proxyType: " + proxyType);
		const self = this;
		$.each(this.m_interactions, function(_, interaction) {
			if (interaction.type === proxyType && interaction.state === "PENDING" && !excludedIds[interaction.id]) {
				self.cancelInteraction(interaction);
			}
		});
		callback();
	};

	/*
		Triggered when an conversation channel changes it's state (for example phone lost connection with PBX)
	*/
	InteractionManager.prototype.onClientChangedStatus = function(channelId, available, message) {
		inConcert.app.getMainWindow().updateChannelStatus(channelId, available, message);
	};

	/*
		Retrieves the requested channel initialized client in case it exists, for telephony the PHONE for chat the CHATCLIENT, etc
	*/
	InteractionManager.prototype.getChannelClient = function(channel) {
		var client = this.m_interactionClients[channel];
		//TODO check for proper initialization
		return client;
	};

	/*
		Get team members suitable as targets for an conversation transfer
		#19143 Added param agent_status. If true then agent status is retrieved (i.e. check if the agent is online or not).
	*/
	InteractionManager.prototype.getMembersForTransfer = function(interactionID, callback, app_id, agent_status = false)
	{
		callback = callback || $.noop;
		var url = this.m_urlBase + "/interactions/get_members_for_transfer/";

		return inConcert.invoke({
			url: url,
			type: "POST",
			timeout: 300000, // 5 minutes
			data: JSON.stringify({
				interactionID : interactionID,
				appId: app_id,
				agentStatus: agent_status
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				var teamMembers = _.map(response, function(member) {
					return new inConcert.TeamMember(member);
				});
				callback(teamMembers);
			},
			error: function( /* response, error */ ) {
				callback();
			}
		});
	};

	/*
		Get campaign users
	*/
	InteractionManager.prototype.getCampaignUsers = function(campaignAtVCC, interactionPermissions, callback)
	{
		callback = callback || $.noop;
		var url = this.m_urlBase + "/interactions/get_campaign_users/";

		return inConcert.invoke({
			url: url,
			type: "POST",
			timeout: 300000, // 5 minutes
			data: JSON.stringify({ campaignAtVCC, interactionPermissions }),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				callback(response.users);
			},
			error: function( /* response, error */ ) {
				callback();
			}
		});
	};

	InteractionManager.prototype.isAllegroInteraction = function(interactionId) {
		return this.m_interactions[interactionId] && this.m_interactions[interactionId].isAllegro();
	};

	/*
		Get campaigns accounts suitable as targets for an conversation transfer
	*/
	InteractionManager.prototype.getCampaignsAccountsForTransfer = function( params )
	{	
		var callback = params.callback || $.noop;
		var url = this.m_urlBase + "/interactions/get_campaigns_accounts_for_transfer/";

		return inConcert.invoke({
			url: url,
			type: "POST",
			timeout: 300000, // 5 minutes
			data: JSON.stringify({
				app_id        : params.app_id,
				interactionID : params.interactions
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				callback(response);
			},
			error: function( /* response, error */ ) {
				callback();
			}
		});
	};

	/*
		Get accounts attention levels suitable as targets for an conversation transfer
	*/
	InteractionManager.prototype.getAccountsAttentionLevelsForTransfer = function( params )
	{
		var callback = params.callback || $.noop;
		var url = this.m_urlBase + "/interactions/get_accounts_attention_levels_for_transfer/";

		return inConcert.invoke({
			url: url,
			type: "POST",
			timeout: 300000, // 5 minutes
			data: JSON.stringify({
				app_id        : params.app_id,
				interactionID : params.interactions
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				callback(response);
			},
			error: function( /* response, error */ ) {
				callback();
			}
		});
	};

	/*
		Get accounts bots suitable as targets for an conversation transfer
	*/
	InteractionManager.prototype.getAccountsBotsForTransfer = function( params )
	{
		var callback = params.callback || $.noop;
		var url = this.m_urlBase + "/interactions/get_accounts_bots_for_transfer/";

		return inConcert.invoke({
			url: url,
			type: "POST",
			timeout: 300000, // 5 minutes
			data: JSON.stringify({
				app_id        : params.app_id,
				interactionID : params.interactions
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				callback(response);
			},
			error: function( /* response, error */ ) {
				callback();
			}
		});
	};

	/*
		Tries to get the next best conversation according to priority
		onTakeResponse if present receives the conversation id by parameter

		app_id is optional, default value "interaction"
	*/
	InteractionManager.prototype.takeNextBest = function( campaignId, interactionType, onTakeResponse, app_id ) {

		onTakeResponse = onTakeResponse || $.noop;

		inConcert.logger.action({action_name : "take_next_best", type : (interactionType ? interactionType : ""), campaign : (campaignId ? campaignId : ""), action_type : "conversation"});

		var getErrorMessage = function(error)
		{
			switch (error.result.toUpperCase())
			{
				case "FORBIDDEN":
					if(campaignId) {
						return format("<%=_T('You have no permissions to take conversations on campaign {0}.')%>", campaignId);
					}
					else {
						return format("<%=_T('No conversation can currently be taken due to lack of permissions.')%>");
					}
				break;

				case "NO_INTERACTIONS_FOUND":
					return "<%=_T('No conversations found.')%>";

				case "AGENT_UNAVAILABLE":
					if (inConcert.app.canTakeWithCurrentState()) {
						return "<%=_T('Please check you are not exceeding the maximum amount of conversations in your inbox')%>";
					}
					else {
						return format("<%=_T('You can not take this because your current status is {0}')%>", inConcert.app.getLoggedUserStatus());
					}
				break;

				default:
					return "<%=_T('Failed to take conversation, please try again in a few seconds...')%>";
			}
		};

		if (!inConcert.app.canTakeWithCurrentState()) {
			var message = format("<%=_T('You can not take this because your current status is {0}')%>", inConcert.app.getLoggedUserStatus());
			onTakeResponse(false, message);
			return;
		}

		var request = {
			campaignId : campaignId,
			interactionType : interactionType,
			app_id: app_id, // optional
			can_take_allegro: !inConcert.app.has_allegro_integration() || 
					inConcert.app.userHasAddress("CALL") &&
					inConcert.AllegroInteractionManager.canTakeAllegroInteractions()
		};

		var url = this.m_urlBase + "/interactions/take_next_best/";
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				if (inConcert.app.is_allegro_channel(response.type)) {
					inConcert.BaragentActions.take(response.interactionID, function(ok) {
						if (ok) {
							onTakeResponse(true, response.interactionID);
						}
						else {
							var message = getErrorMessage({ result : "ERROR" });
							onTakeResponse(false, message);
						}
					});
				}
				else {
					onTakeResponse(true, response.interactionID);
				}
			},
			error: function( undefined, error ) {
				var message = getErrorMessage(error);
				onTakeResponse(false, message);
			}
		});
	};

	/*
		Triggered when a new native session is received, will try to match it with a known conversation
	*/
	InteractionManager.prototype.onNewSession = function(channel, session) {
		//TODO can be solved faster, will perform  bad with lot of conversations (and the only ones pending are ACTIVE and without session)
		var channelClient = this.m_interactionClients[channel];

		_.some(this.m_interactions, function(interaction) {
			if (interaction.type === channel && channelClient.associationValid(interaction, session)) {
				interaction.setSession(session);
				return true; // break the loop
			}
		});

		//the session may be associated with a monitored conversation and not with an
		//conversation belonging to me
		_.some(this.m_monitoredInteractions, function(monitored_interaction) {
			if (monitored_interaction.type === channel && channelClient.associationValid(monitored_interaction, session)) {
				monitored_interaction.setSession(session);
				return true;
			}
		});
	};

	/**
	* Returns the data of an interaction from the server. It is not inflated.
	*/
	InteractionManager.prototype.getInteraction = function( interactionID, callback, inflate, searchArchive )
	{
		var self = this;
		var url = this.m_urlBase + "/interactions/get_interaction/";
		var request = {
			interactionID : interactionID,
			inflate : inflate || false
		};
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function( interactionRaw ) {
				let interaction = self.buildInteractionObject(interactionRaw);
				callback(interaction);
			},
			error: function( /* response, error */ ) {
				if (searchArchive) {
					self.getFromArchive(interactionID, callback, inflate);
				}
				else {
					callback();
				}
			}
		});
	};

	InteractionManager.prototype.canInflateInteraction = function(interactionID)
	{
		let current_interaction = this.findInteraction( interactionID );
		if (!current_interaction) {
			return true;
		}
		return current_interaction.canInflateInteraction();
	};

	/**
	* Retrieves the entire interaction. If the interaction is not found, it looks in the archive too.
	*/
	InteractionManager.prototype.inflateInteraction = function(interactionID, callback)
	{
		var self = this;
		callback = callback || $.noop;

		if (!this.canInflateInteraction(interactionID)) {
			callback([]);
			return;
		}

		var url = this.m_urlBase + "/interactions/get_interaction/";
		var request = {
			interactionID : interactionID,
			inflate : true
		};
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(data) {
				callback(data);
			},
			error: function(xhr) {
				if (xhr.status !== 404) {
					callback(); // Revisar
				}
				else {
					inConcert.invoke({
						url: self.m_urlBase + "/interactions/get_archived_interaction/",
						type: "POST",
						data: JSON.stringify(request),
						dataType: "json",
						contentType: "application/json",
						success: function(data) {
							callback(data);
						},
						error: function(xhr) {
							if(xhr.status === 404) {
								inConcert.logger.error("InteractionManager.inflateInteraction: interaction " + interactionID + " definitely not found");
							}
							callback(); // Revisar
						}
					});
				}
			}
		});
	};

	InteractionManager.prototype.inflateBotInteraction = function(interactionID, callback)
	{
		var self = this;
		callback = callback || $.noop;

		if (!this.canInflateInteraction(interactionID)) {
			callback([]);
			return;
		}

		var url = this.m_urlBase + "/interactions/get_bot_interaction/";
		var request = {
			interactionID : interactionID,
			inflate : true
		};
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(data) {
				callback(data);
			},
			error: function(xhr) {
				if (xhr.status !== 404) {
					callback();
				}
				else {
					inConcert.invoke({
						url: self.m_urlBase + "/interactions/get_archived_interaction/",
						type: "POST",
						data: JSON.stringify(request),
						dataType: "json",
						contentType: "application/json",
						success: function(data) {
							callback(data);
						},
						error: function(xhr) {
							if(xhr.status === 404) {
								inConcert.logger.error("InteractionManager.inflateInteraction: interaction " + interactionID + " definitely not found");
							}
							callback();
						}
					});
				}
			}
		});
	};

	/**
	* Returns the data of an archived interaction from the server. It is not currently inflated.
	*/
	InteractionManager.prototype.getFromArchive = function(interactionID, onInteractionFoundFn, inflate) {
		var self = this;
		onInteractionFoundFn = onInteractionFoundFn || $.noop;
		var url = this.m_urlBase + "/interactions/get_archived_interaction/";
		var request = {
			interactionID : interactionID,
			inflate : inflate || false
		};
		inConcert.invoke({
			url: url,
			type: "POST",
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(interactionRaw) {
				let interaction = self.buildInteractionObject(interactionRaw);
				onInteractionFoundFn(interaction);
			},
			error: function( /* response, error */ ) {
				onInteractionFoundFn();
			}
		});
	};

	InteractionManager.prototype.getFromRepository = function(interaction, onInteractionFoundFn) {
		var url = inConcert.app.getBaseDomain() + "/inconcert/apps/repository/download/?id=" + encodeURIComponent(interaction.id) + "&section=0&part=0";
		inConcert.invoke({
			url: url,
			dataType: "json",
			contentType: "application/json",
			success: function(interaction) {
				if (typeof(onInteractionFoundFn) === "function"){
					onInteractionFoundFn(interaction);
				}
			},
			error: function(response, error) {
				if (typeof(onInteractionFoundFn) === "function"){
					onInteractionFoundFn(null, response, error);
				}
			}
		});
	};
	
	InteractionManager.prototype.__processContactSearchFilter = function(filter, callback)
	{
		if (!filter.query || !filter.query.contactExternalId || !filter.query.contactExternalId.length) {
			return callback(filter);
		}
		if (typeof(filter.query.contactExternalId) === "string") {
			filter.query.contactExternalId = [filter.query.contactExternalId];
		}
		filter.query.contact = [];
		(function processContactFilter(index, onEnd) {
			if (index === filter.query.contactExternalId.length) {
				return onEnd(filter);
			}
			let externalId = filter.query.contactExternalId[index];
			inConcert.app.getContactManager().getContactByExternalId(externalId, function(contact) {
				if (contact) {
					filter.query.contact.push(contact.Id);
				}
				processContactFilter(index + 1, onEnd);
			});
		})(0, function() {
			filter.query.contactExternalId = null;
			callback(filter.query.contact.length? filter : null);
		});
	};
	
	InteractionManager.prototype.__processSearchFilter = function(filter, callback)
	{
		this.__processContactSearchFilter(filter, callback);
	};
	
	InteractionManager.prototype.__getSearchUrl = function(app, method) {
		let url = this.m_urlBase;
		switch(app) {
			case "supervisor":
				url = "/inconcert/apps/supervisor";
				break;
		}
		return url + "/interactions/" + method + "/";
	};
	
	InteractionManager.prototype.__search = function(method, filter, callback = $.noop, app = "agent") {
		let self = this;		
		self.__processSearchFilter(filter, function(filter, error) {
			if (!filter) {
				if (error) {
					//is an error
					return callback(false, error);
				}
				else {
					//is an empty result (like no contact found)
					return callback(true, { interactions : [], last : "", total : 0 });
				}
			}
			inConcert.invoke({
				url: self.__getSearchUrl(app, method),
				type: "POST",
				data: JSON.stringify(filter),
				dataType: "json",
				contentType: "application/json",
				success: function(response) {
					let result = {
						interactions : (response.interactions || []).map(function(interaction) {
							return self.buildInteractionObject(interaction);
						}),
						last : (response.interactions || []).length !== 0? response.last : "",
						total : response.total_interactions
					};
					callback(true, result);
				},
				error: function(_, error) {
					callback(false, error);
				}
			});
		});
	};

	//executes a search using the specified filter
	InteractionManager.prototype.searchMonitorInteractions = function(filter, callback = $.noop, app = "agent")
	{
		this.__search("search_monitor", filter, callback, app);
	};	

	//executes a search using the specified filter
	InteractionManager.prototype.searchOpenInteractions = function(filter, callback = $.noop, app = "agent")
	{
		this.__search("search_open", filter, callback, app);
	};	

	//executes a search using the specified filter
	InteractionManager.prototype.searchArchiveInteractions = function(filter, callback = $.noop, app = "agent")
	{
		this.__search("search_archive", filter, callback, app);
	};

	//executes a search using the specified filter
	InteractionManager.prototype.searchQueuedInteractions = function(filter, callback = $.noop, app = "agent")
	{
		this.__search("search_queued", filter, callback, app);
	};

	//executes a search using the specified filter
	InteractionManager.prototype.searchWaitingInteractions = function(filter, callback = $.noop, app = "agent")
	{
		this.__search("search_waiting", filter, callback, app);
	};

	//executes a search using the specified filter
	InteractionManager.prototype.searchAbandonedInteractions = function(filter, callback = $.noop, app = "agent")
	{
		this.__search("search_abandoned", filter, callback, app);
	};

	//executes a search using the specified filter
	InteractionManager.prototype.searchBotInteractions = function(filter, callback = $.noop, app = "agent")
	{
		this.__search("search_bot", filter, callback, app);
	};

	//executes a search using the specified filter
	InteractionManager.prototype.globalSearch = function(filter, callback = $.noop, app = "agent")
	{
		this.__search("global_search", filter, callback, app);
	};

	/*
		Retrieves loaded/known conversations
	*/
	InteractionManager.prototype.getInteractions = function() {
		return this.m_interactions;
	};

	InteractionManager.prototype.isInteractionAssigned = function(interaction) {
		return interaction && interaction.belongsToMe() && !interaction.isAppointment() && !interaction.isQueued() &&
			!interaction.isAbandoned() && !interaction.isFinished();
	};

	InteractionManager.prototype.getAssignedInteractions = function(filter = function() { return true; }) {
		let self = this;
		return Object.keys(self.m_interactions).reduce(function(assigned, interactionId) {
			let interaction = self.m_interactions[interactionId];
			if (self.isInteractionAssigned(interaction) && filter(interaction)) {
				assigned[interactionId] = interaction;
			}
			return assigned;
		}, {});
	};

	InteractionManager.prototype.getAssignedInteraction = function(interactionId) {
		let interaction = this.m_interactions[interactionId];
		return this.isInteractionAssigned(interaction)? interaction : null;
	};
	
	InteractionManager.prototype.getAssignedInteractionsWithDraftByChannel = function(channel) {
		return this.getAssignedInteractions(function(interaction) {
			return interaction.type === channel && typeof interaction.hasDraft === "function" && interaction.hasDraft();
		});
	};
	
	InteractionManager.prototype.getAssignedInteractionsByChannel = function(channel) {
		return this.getAssignedInteractions(function(interaction) {
			return interaction.type === channel;
		});
	};

	InteractionManager.prototype.getQueuedInteractionsSync = function() {
		let queued = [];
		$.each(this.m_interactions, function(_, interaction) {
				if (interaction && interaction.isQueued()) {
					queued.push(interaction);
				}
			}
		);
		return queued;
	};

	InteractionManager.prototype.getQueuedInteractions = function(callback, inflate) {
		const self = this;
		if (!this.m_firstTimeCheckQueued) {
			this.m_firstTimeCheckQueued = true;
			this.refreshQueued(function() {
				callback(self.getQueuedInteractionsSync());
			}, inflate);
		} else {
			callback(this.getQueuedInteractionsSync());
		}
	};

	InteractionManager.prototype.getAppointmentInteractions = function() {
		let self = this;
		return Object.keys(self.m_interactions).reduce(function(appointments, interactionId) {
			let interaction = self.m_interactions[interactionId];
			if (interaction && interaction.belongsToMe() && interaction.isAppointment()) {
				appointments[interactionId] = interaction;
			}
			return appointments;
		}, {});
	};

	InteractionManager.prototype.findInteraction = function(id)
	{
		if (typeof id !== "string") {
			return;
		}
		var interactions = this.getInteractions();
		if ($.isArray(interactions)) {
			return;
		}
		var match_interaction;
		$.each(interactions, function(undefined, interaction)
		{
			if (interaction.id === id) {
				match_interaction = interaction;
				return false; // break
			}
		});
		return match_interaction;
	};

	InteractionManager.prototype.getSpamError = function( response )
	{
		var result = response.result;
		if ( typeof result === "string" ) {
			try {
				response = JSON.parse( result ); // TODO: Ver esto por que el request me esta retornar un string aca adentro
				result = response.result;
			}
			catch ( err ) {}
		}
		switch( result.toUpperCase() )
		{
			case "FORBIDDEN":
				return "<%=_T('Unable to SPAM action: You not have permission for this action.')%>";

			case "TRAINING_WITH_NO_TYPE_MAIL":
				return "<%=_T('Trying to train spamfilter with an conversation with type != E-MAIL')%>";

			case "INTERACTIONS_MUST_BELONG_TO_SAME_VCC":
				return "<%=_T('All conversations to be trained must belong to the same VCC')%>";

			case "TRAINING_NON_INBOUND_INTERACTION":
				return "<%=_T('Trying to train spamfilter with a non inbound conversation')%>";

			case "INTERACTION_WITHOUT_EVENTS":
				return "<%=_T('Trying to train spamfilter with an conversation without events')%>";

			case "TRAINING_INTERACTION_CONFIRMED_HAM":
				return "<%=_T('Trying to train spamfilter with an conversation confirmed as Ham')%>";

			default:
				return "<%=_T('Internal error. Please retry later')%>";
		}
	};

	/*
		delete spam conversations
		app_id: optional -- default value = interaction
	*/
	InteractionManager.prototype.deleteSpam = function( interactionIDs, onDeleteSpamDoneFn, app_id )
	{
		var self = this;
		onDeleteSpamDoneFn = onDeleteSpamDoneFn || $.noop;

		inConcert.invoke({
			url: this.m_urlBase + "/interactions/spam/delete_spam/",
			type: "POST",
			data: JSON.stringify({ interactionIDs: interactionIDs, app_id: app_id }),
			dataType: "json",
			contentType: "application/json",
			success: function(undefined) {
				onDeleteSpamDoneFn(true);
			},
			error: function(undefined, error) {
				onDeleteSpamDoneFn(false, self.getSpamError( error ));
			}
		});
	};

	/*
		mark as not spam
		app_id: optional -- default value = interaction
	*/
	InteractionManager.prototype.markNotSpam = function( interactionIDs, onMarkNotSpamDoneFn, app_id )
	{
		var self = this;
		onMarkNotSpamDoneFn = onMarkNotSpamDoneFn || $.noop;

		inConcert.invoke({
			url: this.m_urlBase + "/interactions/spam/mark_not_spam/",
			type: "POST",
			data: JSON.stringify({ interactionIDs: interactionIDs, app_id: app_id }),
			dataType: "json",
			contentType: "application/json",
			success: function(undefined) {
				onMarkNotSpamDoneFn(true);
			},
			error: function(undefined, error) {
				onMarkNotSpamDoneFn(false, self.getSpamError( error ));
			}
		});
	};

	/*
		mark as spam
		app_id: optional -- default value = interaction
	*/
	InteractionManager.prototype.markSpam = function( interactionIDs, onMarkSpamDoneFn, app_id )
	{
		var self = this;
		onMarkSpamDoneFn = onMarkSpamDoneFn || $.noop;

		inConcert.invoke({
			url: this.m_urlBase + "/interactions/spam/mark_as_spam/",
			type: "POST",
			data: JSON.stringify({ interactionIDs: interactionIDs, app_id: app_id }),
			dataType: "json",
			contentType: "application/json",
			success: function(undefined) {
				onMarkSpamDoneFn(true);
			},
			error: function(undefined, error) {
				onMarkSpamDoneFn(false, self.getSpamError(error));
			}
		});
	};

	/*
		Retrieves spam conversations
	*/
	InteractionManager.prototype.getSpamInteractions = function(onSpamReadyFn)
	{
		onSpamReadyFn = onSpamReadyFn || $.noop;
		//only MAIL spam is currently supported
		inConcert.invoke({
			url: this.getBaseUrl() + "/interactions/spam/get_spam/",
			type: "GET",
			dataType: "json",
			contentType: "application/json",
			success: function(interactions) {
				onSpamReadyFn(interactions);
			},
			error: function( /* response, error */ ) {
				onSpamReadyFn();
			}
		});
	};

	/**
	* Retrieves conversations for a team member, will return only the ones in the same campaigns as I am.
	* The interactions are currently not inflated.
	*/
	InteractionManager.prototype.getMemberInteractions = function(member, onInteractionsLoadedFn) {
		var self = this;
		var url = this.getBaseUrl() + "/interactions/get_user_interactions/";
		var request = {
			memberId : member.Id
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(interactions) {
				var list = [];
				for (var i = 0; i < interactions.length; i++) {
					list.push(self.buildInteractionObject(interactions[i]));
				}
				if (typeof(onInteractionsLoadedFn) === "function") {
					onInteractionsLoadedFn(list);
				}
			},
			error: function( /* response, error */ ) {
				if (typeof(onInteractionsLoadedFn) === "function") {
					onInteractionsLoadedFn();
				}
			}
		});
	};

	/**
	* Retrieves conversations for a team member, will return only the ones in the same campaigns as I am.
	* The interactions are currently not inflated.
	*/
	InteractionManager.prototype.getCampaignInteractions = function( filter, onInteractionsLoadedFn )
	{
		var self = this;

		onInteractionsLoadedFn = onInteractionsLoadedFn || $.noop;

		var url = this.getBaseUrl() + "/interactions/get_campaign_interactions/";

		// por que el default a null?
		var posted_filter = {
			campaign         : filter.campaignId,
			account          : filter.account,
			attention_levels : filter.attention_level,
			type             : filter.interactionType,
			date_from        : filter.date_from,
			date_to          : filter.date_to,
			address          : filter.address, // contact address
			agent            : filter.agentId,
			label            : filter.label,
			page             : filter.page || 0,
			rows_per_page    : filter.rows_per_page || 100,
			text             : filter.text,
			ticket           : filter.ticket,
			exact            : filter.exact,
			sort             : filter.sort,
			workgroup        : filter.workgroup
		};

		if (filter.states) {
			posted_filter.states = filter.states;
		}

		var request = inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(posted_filter),
			dataType: "json",
			contentType: "application/json",
			success: function(response)
			{
				var list = [];
				if ( response && response.interactions ){
					for (var i = 0; i < response.interactions.length; i++) {
						list.push(self.buildInteractionObject(response.interactions[i]));
					}
				}
				onInteractionsLoadedFn(list, response.total_interactions);
			},
			error: function(undefined, error) {
				//this function does not return structured(JSON) error messages
				onInteractionsLoadedFn( null, 0, error );
			}
		});

		return request;
	};

	/*
		Removes all the known conversations cleanly
	 */
	InteractionManager.prototype.deleteAllInteractions = function() {
		var deletedInteractions = _.map(this.m_interactions, function(interaction) {
			return {
				id     : interaction.id,
				deleted: true
			};
		});
		this.m_interactions = {};
		return deletedInteractions;
	};

	InteractionManager.prototype.getCachedInteraction = function(interaction_id)
	{
		if (!this.m_interactions) {
			return;
		}
		return this.m_interactions[interaction_id];
	};

	/**
	Updates a bunch of conversations with notifications received from somewhere.
	'interactions' is an array of objects of the form:
		type: "interaction",
		channel: "AGENT",
		message: {}, // the actual interaction
		reference: "a string" (some value used to track the message)
	*/
	InteractionManager.prototype.updateInteractions = function(interactions)
	{
		let result = [];
		if (!$.isArray(interactions) || interactions.length === 0) {
			return result;
		}

		//console.log("[InteractionManager.updateInteractions] --> interactions " + JSON.stringify(interactions));

		for (let i = 0; i < interactions.length; i++) {
			//updates are received as JSON messages
			let parsedMessage = interactions[i].message;
			parsedMessage.notificationChannel = interactions[i].channel;
//			let total_new_events = 0;
			//was this an outgoing one?
			let match = this.matchToPendingInteraction(parsedMessage);
			if (!match) {
				// let old_interaction = this.getCachedInteraction(parsedMessage.id);
				// if (old_interaction) {
					// total_new_events = old_interaction.countNewEvents(parsedMessage);
				// }
				let interaction = this.updateInteraction(parsedMessage);
				let message = interaction || {id: parsedMessage.id, deleted: true};
				this.updateQueuedInteractions([message]);
				this.updateAbandonedInteractions([message]);
				this.cleanFinishedInteractions([message]);
				//deletions are returned as flagged ids only for proper cleanup
				result.push(message);
			}
			else {
				//if what I have is a pending switch treat it as a delete/insert
				result.push(match.interaction);
				result.push({id : match.previousId, deleted : true});
				inConcert.app.m_mainWindow.switchInteractions(match.previousId, match.interaction);
				this.updateQueuedInteractions([match.interaction]);
				this.updateAbandonedInteractions([match.interaction]);
				this.cleanFinishedInteractions([match.interaction]);
			}
		}
		return result;
	};

	/**
	* Retrieves the N best queued conversation and treat the response as if it were new notifications
	* The interactions are not inflated.
	*/
	InteractionManager.prototype.getTopQueued = function(onRefreshed, inflate)
	{
		onRefreshed = onRefreshed || $.noop;

		var url = this.getBaseUrl() + "/interactions/get_top_queued_interactions/";
		inConcert.invoke({
			type: "POST",
			url: url,
			dataType: "json",
			contentType: "application/json",
			data:  JSON.stringify({
				inflate : inflate || false
			}),
			success: function(response) {
				onRefreshed(response.interactions);
			},
			error: function() {
				onRefreshed();
			}
		});
	};

	InteractionManager.prototype.refreshQueued = function(onRefreshed, inflate)
	{
		var self = this;
		onRefreshed = onRefreshed || $.noop;
		
		self.getTopQueued(function(interactions) {
			if (interactions) {
				//update received conversations
				interactions = _.map(interactions || [], self.updateInteraction, self);
				//send them onto screen
				if (interactions.length > 0) {
					//only update Pending Interactions
					var panel = inConcert.app.m_mainWindow.getPanel(QUEUED_PANEL_ID);
					if (panel) {
						panel.updateInteractions(interactions);
					}
				}
				onRefreshed(true);
			}
			else {
				onRefreshed();
			}
		}, inflate);
	};

	/*
		Returns a summary of unread conversations by type
		{
			MAIL:12
			TWITTER:1
		}
	*/
	InteractionManager.prototype.summarizeUnreadInteractions = function() {
		let summary = {};
		for (let id in this.m_interactions) {
			if (this.m_interactions.hasOwnProperty(id)) {
				let interaction = this.m_interactions[id];
				let belongsToMe = interaction.belongsToMe();
				if (!belongsToMe) {
					continue;
				}
				let isLocked = interaction.isLocked();
				let isWrappedUp = interaction.isWrappedUp();
				let isAssigned = interaction.isAssigned();
				if (!isLocked && !isWrappedUp && !isAssigned) {
					continue;
				}
				let isReserved = interaction.isReserved();
				if (isReserved) {
					continue;
				}
				let isGone = interaction.isGone();
				if (isGone) {
					continue;
				}
				let isRead = interaction.isRead();
				if (isRead) {
					continue;
				}
				let channel = interaction.type;
				summary[channel] = (summary[channel] || 0) + 1;
			}
		}
		return summary;
	};

	/*
		Interaction factory method, just builds stuff
	*/
	InteractionManager.prototype.buildInteractionObject = function(data) {
		var interaction;
		switch (data.type) {
		case "CALL":
			interaction = (data.interactionData && data.interactionData.isInternal) ? new inConcert.interactions.internal_call(data) : new inConcert.interactions.telephony(data);
			break;
		case "VOICEMAIL":
			interaction = new inConcert.interactions.voicemail(data);
			break;
		case "CHAT":
			interaction = new inConcert.interactions.webchat(data);
			break;
		case "INTERNAL_CHAT":
			interaction = new inConcert.interactions.internal_chat(data);
			break;
		case "IM":
			interaction = new inConcert.interactions.im(data);
			break;
		case "SMS":
			interaction = new inConcert.interactions.sms(data);
			break;
		case "WHATSAPP":
			interaction = new inConcert.interactions.whatsapp(data);
			break;
		case "MAIL":
			interaction = new inConcert.interactions.mail(data);
			break;
		case "WEBCONTACT":
			interaction = new inConcert.interactions.webcontact(data);
			break;
		case "FACEBOOK":
			interaction = new inConcert.interactions.facebook(data);
			break;
		case "INSTAGRAM":
			interaction = new inConcert.interactions.instagram(data);
			break;
		case "YOUTUBE":
			interaction = new inConcert.interactions.youtube(data);
			break;
		case "APPSTORE":
			interaction = new inConcert.interactions.appstore(data);
			break;
		case "GOOGLEPLAY":
			interaction = new inConcert.interactions.googleplay(data);
			break;
		case "RCS":
			interaction = new inConcert.interactions.rcs(data);
			break;
		case "TWITTER":
			interaction = new inConcert.interactions.twitter(data);
			break;
		}
		if (interaction) {
			interaction.setManager(this);
		}
		return interaction;
	};

	InteractionManager.prototype.load = function(data) {
		this.create(data);
	};

	/*
	Initializes an interaction object according to its base type.
	*/
	InteractionManager.prototype.create = function(data, invoked_from)
	{
		/*
			Cuando se crea una interaccion, nunca puede estar inflada, ni marcar como inflada, si es que la misma no se ha inflado via invocacion al server.
			La misma se debe inflar invocando a la funcion .inflateInteraction de esta misma clase.
		*/
		let interaction = this.buildInteractionObject(data);
		let log = {action_name: null, id: interaction.id, type: interaction.type, campaign: interaction.campaign, state: interaction.state, initiative: interaction.initiative, action_type: "conversation", class_name: "INTERACTION_MANAGER"};
		if (typeof(interaction.id) === "undefined") {
			interaction.createOwnId();
			log.action_name = "CREATE_INTERACTION";
		}
		else {
			log.action_name = "NEW_INTERACTION";
		}
		log.action_name = invoked_from ? (log.action_name + "_" + invoked_from) : log.action_name;
		inConcert.logger.action(log);
		if (interaction.isQueued()) {
			this.updateQueuedInteractions([interaction]);
		}
		else if (interaction.isAbandoned()) {
			this.updateAbandonedInteractions([interaction]);
			return interaction;
		}
		if (!interaction.isFinished()) {
			this.cleanFinishedInteractions([interaction]);
		}
		this.m_interactions[interaction.id] = interaction;
		inConcert.app.messagebus.publish("/interactions/created", interaction);
		return interaction;
	};

	/**
	*/
	InteractionManager.prototype.discard = function(element) {
		var interaction = this.m_interactions[element.id];
		if (interaction) {
			delete this.m_interactions[element.id];
			interaction.dispose();
			// Publica el evento de remove de la interaction via agent poller. Por el momento solo lo usa el phoneclient.
			inConcert.app.messagebus.publish(`/interactions/removed/${interaction.id}`, interaction);
		}
	};

	/**
	Por que corchos hay dos metodos que borran ?? Por que uno se banca appointments y el otro no?
	*/
	InteractionManager.prototype.remove = function(interaction = {}, pending = false)
	{
		const log = {action_name: "REMOVE_INTERACTION", id: interaction.id, type: interaction.type, campaign: interaction.campaign, state: interaction.state, initiative: interaction.initiative, action_type: "CONVERSATION", class_name: "INTERACTION_MANAGER"};
		inConcert.logger.action(log);
		if (!interaction || !interaction.id || !this.m_interactions) {
			return false;
		}
		const inter = this.m_interactions[interaction.id];
		// Since the deletion is quite shitty (multiple code paths), we may have already removed the interaction.
		if (inter) {
			if (inter.type === "CALL") {
				const phoneClient = this.getChannelClient("CALL");
				if (phoneClient) {
					/*
						Ref Issue #15445: PhoneClientWindow con ventanas que NO se cerraban.

						Esta invocacion debe ser sincronica, ya que la interaccion si no se elimina antes y cuando se invocaba la funcion de forma asincronica
							daba N errores en las aperturas de las ventanas.
							Problemas con redial, etc.
							Ademas habia diferencias respecto a Allegro y Nativo.
					*/
					phoneClient.interactionFinished(interaction, pending);
				}
			}
			inter.dispose();
		}
		delete this.m_interactions[interaction.id];
		// Publica el evento de remove de la interaction via agent poller. Por el momento solo lo usa el phoneclient.
		inConcert.app.messagebus.publish(`/interactions/removed/${interaction.id}`, interaction);
		return true;
	};

	/**
	* Called when an conversation is modified. Receives the updated conversation data as parameter.
	*/
	InteractionManager.prototype.updateInteraction = function(data)
	{
		console.log("updateInteraction", data);
		
		var interaction = this.m_interactions[data.id] || this.m_monitoredInteractions[data.id];
		var old_interaction;
		if ( !interaction ) {
			interaction = this.create(data);
			if (!interaction) {
				return;
			}
		}
		else {
			old_interaction = inConcert.Utils.cloneObject(interaction);
			interaction.update(data);
			interaction.pending = false;
		}

		inConcert.app.messagebus.publish("/interactions/updated", interaction, old_interaction);

		//try to associate the conversation with its native client handler.
		var session;
		var client = this.getChannelClient(data.type);
		if ( client && client.isInitialized() ) {
			if (client.clientUpdate) {
				client.clientUpdate(interaction,function(sessionWindow, callback = $.noop) {
					if (interaction.type === 'CALL' && sessionWindow) {
						const panel = inConcert.app.getMainWindow().getPanel(ASSIGNED_PANEL_ID);
						let on_window_close;
						panel.onNewWindow(interaction, sessionWindow, on_window_close);
						callback(true);
					}
				});
			}
			session = interaction.getSession();
			if (!session) {
				session = client.findInteractionSession(interaction);
				if (session) {
					interaction.setSession(session);
				}
			}
			//TODO
			//can an interaction exist without session? if not callback error handler to be displayed by the main window
		}
		// Si la interaccion fue finalizada, se elimina la misma de la lista de interacciones del agente.
		//  Solo se cierra la sesion de chat cuando la misma es finalizada
		var is_finished = interaction.isFinished();
		if ( is_finished ) {
			// cierro la sesion de chat
			session = interaction.getSession();
			if ( session ) {
				session.close();
			}
			interaction.deleted = true;
			this.remove(interaction);
			return interaction;
		}
		return interaction;
	};

	InteractionManager.prototype.onBatchSuccess = function(response, batchAction, callbackFn) {
		if (typeof(callbackFn) === "function") {
			if (response && response.successful_nodes && response.unsuccessful_nodes){
				callbackFn(true, response.successful_nodes, response.unsuccessful_nodes);
			}
			else{
				inConcert.logger.error(
					format("(InteractionManager.prototype.{0}): Response from server does not have successful or unsuccessful nodes", batchAction)
				);
			}
		}
	};

	InteractionManager.prototype.onBatchError = function( _, _1, error, callbackFn, interactions ) {
		if (typeof(callbackFn) === "function") {
			if ( $.isArray( interactions ) ) {
				for ( var i = 0; i < interactions.length; i++ ) {
					var interaction = interactions[ i ];
					interaction.pending = false;
					inConcert.app.m_mainWindow.updateInteractions([interaction]);
				}
			}
			callbackFn(false, null, null, error.result);
		}
	};

	InteractionManager.prototype.extractIdsFromBatch = function(batch) {
		var batchIds = [];
		// to make sure only send conversations ids and not the complete object.
		batch.forEach(function( currentNode ){
			if ( currentNode && currentNode.id ){
				var finalization_data = currentNode.getFinalizationData();
				batchIds.push({
					id : currentNode.id,
					finalization_data: finalization_data
				});
			}
			else{
				inConcert.logger.error(
					format( "(InteractionManager.prototype.extractIdsFromBatch): A node doesn't have id: '{0}'", currentNode.toString() )
				);
			}
		});
		return batchIds;
	};

	InteractionManager.prototype._setDispositionComment = function (interactionId, dispositionCode, comment, fnCallback = $.noop) {
		
		let url = inConcert.app.getBaseUrl() + "/interactions/set_disposition/";
		let post_data = {
			interactionID : interactionId,
			disposition : dispositionCode || "",
			comment : comment || "",
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data:  JSON.stringify(post_data),
			dataType: "json",
			contentType: "application/json",
			success: function () {
				fnCallback(true);
			},
			error: function () {
				fnCallback(false);
			}
		});
	};

	/*	Given an batch of conversations, this function will finish all of them.
		'conversations' parameter, an array with N conversations to archive.
		if there's an error on the server, will return an array of unsuccessful conversations.

		app_id: optional -- default value = interaction
	*/
	InteractionManager.prototype.batchArchive = function(interactions, dispositionCode, comment, onArchived, app_id)
	{
		onArchived = onArchived || $.noop;

		// nothing to do without conversations
		if (!interactions || interactions.length === 0){
			inConcert.logger.error("Called batch archive conversations without conversations!");
			return onArchived(false, null, null, "<%=_T('There is no conversations to archive!')%>");
		}

		inConcert.logger.action({action_name : "archive_interactions", interactions_amount : interactions.length, action_type : "conversation"});

		const firstInteraction = interactions[0];
		if (firstInteraction.isAllegro()) {
			this._setDispositionComment(interactions[0].id, dispositionCode, comment || "", function (success) {
				if (!success){
					console.warn("Error adding disposition comment");
				}
				inConcert.AllegroInteractionManager.batchArchive(interactions, dispositionCode, comment, onArchived, app_id);
			});
			return;
		}
		if (firstInteraction.type === "CALL") {
			const phoneState = this.getPhoneState().toLowerCase();
			if (phoneState === "active") {
				inConcert.logger.error("Called batch archive conversations without conversations! phoneState: " + phoneState);
				return onArchived(false, null, null, "<%=_T('There is no conversations to archive!')%>");
			}
		}

		const interactionManager = inConcert.app.getInteractionManager();
		const url = inConcert.app.getBaseUrl() + "/interactions/archive_interactions/";
		const post_data = {
			interactions : interactionManager.extractIdsFromBatch(interactions),
			campaignId : interactions[0].campaign,
			disposition : dispositionCode || "",
			comment : comment || "",
			app_id : app_id
		};

		inConcert.invoke({
			type: "POST",
			url: url,
			data:  JSON.stringify(post_data),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				for (var i = 0; i < interactions.length; i++) {
					var interaction = interactions[i];
					// #7100 (IE-Fix) Se emite el evento cuando el usuario ha finalizado la interaccion.
					// Esto para solucionar el problema de que en IE queda referencia cuando tenemos Integracion, y los eventos ".on",
					// continuan escuchando y termian dando erroes runtime.
					//
					interaction.emit("archived_by_user");
				}
				interactionManager.onBatchSuccess(response, "batchArchive", onArchived);
			},
			error: function(response, error) {
				interactionManager.onBatchError(response, "batchArchive", error, onArchived);
			}
		});
	};

	/*
		Funcion encargada de marcar en que proceso de batch, se encuentran las interacciones.
		(De momento solo se hizo para REQUEUE)
	*/
	InteractionManager.prototype.markBatchInteractionsInProgress = function(interactions, inProgressType = "NOTHING") {
		interactions.forEach((interaction) => {
			interaction.inProgressType = inProgressType;
			this.updateInteraction({
				id: interaction.id, inProgressType: inProgressType, type: interaction.type
			});
		});
	};

	/*	Given an batch of conversations, this function will requeue all of them.
		'conversations' parameter, an array with N conversations to requeue.
		if there's an error on the server, will return an array of unsuccessful conversations.

		app_id: optional -- default value = interaction
	*/
	InteractionManager.prototype.batchRequeue = function( interactions, onRequeued, app_id ) {
		const interactionManager = inConcert.app.getInteractionManager();
		inConcert.logger.action({action_name : "requeue_interaction", interactions_amount: interactions.length, action_type : "conversation"});

		// nothing to do without conversations
		if (!interactions || interactions.length === 0){
			return onRequeued(false, null, null, "<%=_T('There is no conversations to requeue!')%>");
		}
		// CALL allegro
		if (interactions[0].isAllegro()) {
			return inConcert.AllegroInteractionManager.batchRequeue(interactions, onRequeued, app_id);
		}
		// NATIVO (todos los canales)

		// para nativo agrego un flag para identificar que las interacciones se estan encolando.
		interactionManager.markBatchInteractionsInProgress(interactions, "IN_PROGRESS_QUEUE");

		/*
		*/
		const onResponse = (success, response, err) => {
			// elimino el flag
			interactionManager.markBatchInteractionsInProgress(interactions, "NOTHING");
			// ok
			if (success) return interactionManager.onBatchSuccess(response, "batchRequeue", onRequeued);
			// error
			interactionManager.onBatchError(response, "batchRequeue", err, onRequeued);
		};

		// ejecuto request para encolar
		const url = inConcert.app.getBaseUrl() + "/interactions/requeue_interactions/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify({
				app_id      : app_id,
				interactions: interactionManager.extractIdsFromBatch(interactions)
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				onResponse(true, response);
			},
			error: function(response, err) {
				onResponse(false, response, err);
			}
		});
	};

	InteractionManager.prototype.batchTake = function( interactions, onTaken, app_id ) {
		var interactionManager = inConcert.app.getInteractionManager();
		inConcert.logger.action({action_name : "take_interaction", interactions_amount: interactions.length, action_type : "conversation"});

		if (!inConcert.app.canTakeWithCurrentState()) {
			onTaken(false, null, null, format("<%=_T('You can not take this because your current status is {0}')%>", inConcert.app.getLoggedUserStatus()));
			return;
		}

		// nothing to do without conversations
		if (!interactions || interactions.length === 0){
			return ;
		}
		
		if (interactions[0].isAllegro()) { //Solo existe una interaccion cuando se quiere tomar
			interactions[0].take(function (ok, message) {
				if (ok) {
					interactionManager.onBatchSuccess({ successful_nodes: interactions, unsuccessful_nodes: [] }, "batchTake", onTaken);
				} else {
					const args = { error_code: "UNABLE_TO_TAKE", error_message: message };
					const unsuccessful_node = { id: interactions[0].id, result: args.error_code, reason: args.error_message, raw_info: args };
					interactionManager.onBatchError({ successful_nodes: [], unsuccessful_nodes: [unsuccessful_node] }, "batchTake", onTaken);
				}
			});
			return;
		}
		
		var url = inConcert.app.getBaseUrl() + "/interactions/take_interactions/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify({
				interactions : interactionManager.extractIdsFromBatch(interactions),
				app_id : app_id
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				interactionManager.onBatchSuccess(response, "batchTake", onTaken);
			},
			error: function(response, error) {
				interactionManager.onBatchError(response, "batchTake", error, onTaken);
			}
		});
	};

	InteractionManager.prototype.batchCancel = function(interactions, onCanceled) {
		// nothing to do without conversations
		if (!interactions || interactions.length === 0){
			if (typeof(onCanceled) === "function"){
				return onCanceled(false, null, null, "<%=_T('There is no conversations to cancel!')%>");
			}
		}
		inConcert.logger.action({action_name : "cancel_interactions", interactions_amount : interactions.length, action_type : "conversation"});

		interactions.forEach(function(interaction){
			interaction.state = "CANCELED";
			if (interaction.m_session && interaction.m_session.close) {
				interaction.m_session.close();
			}
		});
		if (typeof(onCanceled) === "function"){
			onCanceled(true, interactions, []);
		}
	};

	// app_id: optional -- default value = interaction
	InteractionManager.prototype.batchUpdateLabels = function( interactions, labels, currentLabels, onInteractionsLabeled, app_id ) {
		var interactionManager = inConcert.app.getInteractionManager();
		inConcert.logger.action({action_name : "update_labels_interactions", interactions_amount : interactions.length, action_type : "conversation"});

		// nothing to do without conversations
		if (!interactions || interactions.length === 0){
			return onInteractionsLabeled(false, null, null, "<%=_T('There is no conversations to label!')%>");
		}

		var url = inConcert.app.getBaseUrl() + "/interactions/update_labels_interactions/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify({
				interactions : interactionManager.extractIdsFromBatch(interactions),
				newLabels: labels,
				currentLabels: currentLabels,
				app_id: app_id
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				interactionManager.onBatchSuccess(response, "batchUpdateLabels", onInteractionsLabeled);
			},
			error: function(response, error) {
				interactionManager.onBatchError(response, "batchUpdateLabels", error, onInteractionsLabeled, interactions );
			}
		});
	};

	// app_id: optional -- default value = interaction
	InteractionManager.prototype.batchDeleteLabels = function( interactions, labels, onLabelsDeleted, app_id ) {
		var interactionManager = inConcert.app.getInteractionManager();
		inConcert.logger.action({action_name : "delete_label_interactions", interactions_amount : interactions.length, action_type : "conversation"});

		// nothing to do without conversations
		if (!interactions || interactions.length === 0){
			return onLabelsDeleted(false, null, null, "<%=_T('There is no conversations to delete the label!')%>");
		}

		var url = inConcert.app.getBaseUrl() + "/interactions/delete_labels_interactions/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify({
				interactions : interactionManager.extractIdsFromBatch(interactions),
				deleteLabels: labels,
				app_id: app_id
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				interactionManager.onBatchSuccess(response, "batchDeleteLabels", onLabelsDeleted);
			},
			error: function(response, error) {
				interactionManager.onBatchError(response, "batchDeleteLabels", error, onLabelsDeleted);
			}
		});
	};

	InteractionManager.prototype.canTransferToMember = function(info, callback = $.noop)
	{
		const url = inConcert.app.getBaseUrl() + "/interactions/can_transfer_to_member/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify({
				interaction_id: info.interaction,
				target        : info.target,
				app_id        : info.appId
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				callback(true, response);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});
	};

	//transfer the conversations to another ( team member / campaign account / attention level )
	// app_id: optional -- default value = interaction
	// dispositonInfo es utilizado para las transferencias a campanas, de los canales (mail, sms, whatsapp)
	InteractionManager.prototype.batchTransfer = function( interactions, target, onTransferred, app_id, dispositonInfo)
	{
		var interactionManager = inConcert.app.getInteractionManager();
		inConcert.logger.action({action_name : "transfer_interactions", interactions_amount : interactions.length, action_type : "conversation", target: JSON.stringify(target)});

		// nothing to do without conversations
		if (!interactions || interactions.length === 0){
			return onTransferred(false, null, null, "<%=_T('There is no conversations to transfer!')%>");
		}

		if (interactions[0].isAllegro()) {
			return inConcert.AllegroInteractionManager.batchTransfer( interactions, target, onTransferred, app_id );
		}

		var url = inConcert.app.getBaseUrl() + "/interactions/transfer_interactions/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify({
				interactions : interactionManager.extractIdsFromBatch(interactions),
				target : target,
				app_id: app_id,
				dispositonInfo: dispositonInfo
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				interactionManager.onBatchSuccess(response, "batchTransfer", onTransferred);
			},
			error: function(response, error) {
				interactionManager.onBatchError(response, "batchTransfer", error, onTransferred);
			}
		});
	};

	/*
		'data' parameter, is an object with the following properties:
				- type:				interaction type.	(required)
				- campaigns:		campaign list.		(required)
				- currentAgent:		logged user id.		(required)
				- contact:			contact object		(optional, type: object)
				- address:			contact address		(optional, type: object)
				- parentInteractionId:	appoiment id.		(optional)
		'options:  is an object with particular options
	 */
	InteractionManager.prototype.createOutgoing = function( data, options )
	{
		options = options || {};
		var contact_id = null, contact_name = null;
		if ( data.contact ) {
			contact_id   = data.contact.Id;
			contact_name = data.contact.Name;
		}
		
		let campaignId = null;
		if (data.campaigns && !$.isArray(data.campaigns)) {
			campaignId = data.campaigns.Id;
			data.campaigns = [data.campaigns];
		}
		else {
			campaignId = data.parentInteractionId ? data.campaigns[0].Id : null;
		}
		
		var rawInteraction = {
			type                : data.type                                                               ,
			created_by_agent    : true                                                                    ,
			account             : data.account                                                            ,
			contact             : contact_id                                                              ,
			interactionData     : { contactName: contact_name, fromAPI: data.fromAPI }                    ,
			initiative          : "OUTBOUND"                                                              ,
			assignedAgents      : [ data.currentAgent ]                                                   ,
			contactAddress      : data.address ? inConcert.Utils.getAddressDisplayText(data.address) : "" ,
			completeAddress     : data.address                                                            ,
			state               : "PENDING"                                                               ,
			parentInteractionId : data.parentInteractionId                                                ,
			groupId             : data.groupId                                                            ,
			campaign            : campaignId
		};
		$.extend(true, rawInteraction, options);
		var interaction = this.create(rawInteraction);
		if (!interaction) {
			return false;
		}
		if ( interaction.parentInteractionId ) {
			var appointment_interaction = this.findInteraction( interaction.parentInteractionId );
			if ( appointment_interaction ) {
				appointment_interaction.addRelationshipPendingInteraction( interaction );
			}
		}
		interaction.setContactObject(data.contact);
		var campaigns = interaction.getAllowedCampaigns(data.campaigns);
		if (!campaigns || campaigns.length === 0) {
			this.remove(interaction);
			return false;
		}
		if (data.contact && (!interaction.interactionData || !interaction.interactionData.gatheredContactData)) {
			interaction.interactionData = interaction.interactionData || {};
			interaction.interactionData.gatheredContactData = {
				external_id : data.contact.Id
			};
		}
		var channelClient = this.getChannelClient(data.type);
		if (channelClient && typeof(channelClient.onNewOutgoing) === "function") {
			channelClient.onNewOutgoing(interaction);
		}

		if(interaction.interactionData && interaction.interactionData.fromAPI)
		{
			inConcert.invoke({
				url: this.m_urlBase + "/interactions/save_api_additional_data/",
				type: "POST",
				data: JSON.stringify({ id: interaction.id, data : { creator : "AgentAPI" }}),
				dataType: "json",
				contentType: "application/json",
				success: function() { },
				error: function() {
					inConcert.logger.error("Error saving API additional data");
				}
			});
		}
		
		return interaction;
	};

	/*
		Creates the dummy-stub conversation object for the internal chat, the main purpose is to be able
		to handle the internal chat just as if it were a proper conversation
	*/
	InteractionManager.prototype.createInternalChatInteraction = function(memberId, memberName, initiative) {
		var data = {
				type : "INTERNAL_CHAT",
				id : inConcert.Utils.generateGuid(),
				localInteraction : true,
				teamMember : memberId,
				teamMemberName : memberName,
				interactionData : {
					contactName : memberName
				},
				initiative : initiative,
				assignedAgents : [
					inConcert.app.getCompleteId()
				],
				contactAddress : memberName,
				completeAddress: memberId,
				campaign: "",
				state : "CREATED"
			};
		var interaction = this.buildInteractionObject(data);
		return interaction;
	};

	/*
		Triggered when an internal chat session has ended
	*/
	InteractionManager.prototype.onInternalChatEnded = function(session) {
		var interaction = session.getInteraction();
		if (interaction && interaction.isLocal()) {
			this.remove(interaction);
			interaction.deleted = true;
			inConcert.app.m_mainWindow.updateInteractions([interaction]);
		}
	};

	/*
		Triggered when a new internal chat is received (not one from a client, but one from a team member)
	*/
	InteractionManager.prototype.onInternalChat = function(session, inviter, localInteractionId)
	{
		if (inviter === inConcert.app.getCompleteId()) {
			var interaction = this.m_interactions[localInteractionId];
			if (interaction) {
				interaction.setSession(session);
			}
			return true;
		}
		else {
			var self = this;
			if(!inviter.match(inConcert.app.idAtVCCRegExp)) {
				return false;
			}
			inConcert.app.getTeamMember(
				RegExp.$1,
				function(teamMember) {
					if ( !teamMember ) {
						return;
					}
					var interaction = self.createInternalChatInteraction(teamMember.Id, teamMember.get_name_and_surname(), "INBOUND");
					self.m_interactions[interaction.id] = interaction;
					interaction.setSession(session);
					inConcert.app.m_mainWindow.updateInteractions([interaction]);
					inConcert.app.messagebus.publish("/interactions/updated", interaction);
				},
				true // from cache
			);
			return true;
		}
	};

	/*
		Creates an outgoing chat with a team member
	*/
	InteractionManager.prototype.createInternalChat = function( targetMemberId, targetMemberName, permissionFn, callbackOpenClientFn )
	{
		var self = this;
		callbackOpenClientFn = callbackOpenClientFn || $.noop;
		permissionFn = permissionFn || function() { return true; };

		if ( !permissionFn() ) {
			callbackOpenClientFn( false, { result: "FORBIDDEN" } );
			return;
		}
		var interaction;

		if ( this.m_interactionClients.CHAT.isInitialized() ) {
			//if there is an ongoing conversation, open that one
			var sessionId = this.m_interactionClients.CHAT.getUserConversation( targetMemberId );
			if ( !sessionId ) {
				interaction = this.createInternalChatInteraction( targetMemberId, targetMemberName, "OUTBOUND" );
				this.m_interactions[ interaction.id ] = interaction;

				this.m_interactionClients.CHAT.inviteUser( targetMemberId, interaction.id,
				{
					success: function( sessionId ) {
						var session = self.m_interactionClients.CHAT.getSession( sessionId );
						interaction.setSession( session );
						callbackOpenClientFn( true, interaction );
					},
					error : function( undefined, error ) {
						self.remove(interaction);
						callbackOpenClientFn( false, error );
					}
				});
			}
			else {
				//search for existing interaction and return that one in order to activate view
				var session = this.m_interactionClients.CHAT.getSession( sessionId );
				interaction = session.getInteraction();
				callbackOpenClientFn( true, interaction );
			}
		}
	};

	/*
		Creates the dummy-stub conversation object for the internal call, the main purpose is to be able
		to handle the internal call just as if it were a proper conversation
	*/
	InteractionManager.prototype.createInternalCallInteraction = function(memberId, memberName, memberAddress, initiative) {
		var data = {
				type : "CALL",
				id : inConcert.Utils.generateGuid(),
				localInteraction : true,
				teamMember : memberId,
				teamMemberName : memberName,
				inviterMemberName : inConcert.app.getLoggedUserName(),
				interactionData : {
					isInternal: true,
					contactName : memberName
				},
				initiative : initiative,
				assignedAgents : [
					inConcert.app.getCompleteId()
				],
				contactAddress : memberAddress,
				completeAddress: memberAddress,
				campaign: "",
				state : "CREATED"
			};
		var interaction = this.buildInteractionObject(data);
		return interaction;
	};

	/*
		Triggered when an internal call session has ended
	*/
	InteractionManager.prototype.onInternalCallEnded = function(/*session*/) {
		//Nothing to do
	};

	/*
		Triggered when a new internal call is received (not one from a client, but one from a team member)
	*/
	InteractionManager.prototype.onInternalCall = function(/*session, inviter, localInteractionId*/) {
		//Nothing to do
	};

	/*
		Creates an outgoing call with a team member
	*/
	InteractionManager.prototype.createInternalCall = function( targetMemberId, targetMemberName, permissionFn, callbackOpenClientFn, targetAddress ) {
		var self = this;
		callbackOpenClientFn = callbackOpenClientFn || $.noop;
		permissionFn = permissionFn || function() { return true; };

		if ( !permissionFn() ) {
			callbackOpenClientFn( false, { result: "FORBIDDEN" } );
			return;
		}

		if( !targetAddress ) {
			callbackOpenClientFn( false, { result: "NO_ADDRESS" } );
			return;
		}

		var interaction;

		if ( this.m_interactionClients.CALL.isInitialized() ) {
			if(this.getPhoneState() === "connected") {
				interaction = this.createInternalCallInteraction( targetMemberId, targetMemberName, targetAddress, "OUTBOUND" );
				this.m_interactions[ interaction.id ] = interaction;
				var targetPeer = targetAddress;
				var session = this.m_interactionClients.CALL.createSession(interaction);
				interaction.setSession( session );
				var dialCallback;
				dialCallback = function(success) {
					if(!success) {
						self.remove(interaction);
					}
					callbackOpenClientFn( success, interaction );
				};

				this.m_interactionClients.CALL.dial({
					interactionID: interaction.id,
					campaignAtVCC: "",
					prefix: "",
					destination: targetPeer,  
					contact: targetMemberName,
					isInternal: true,
					callback: dialCallback
				});
			}
		}
	};

	/** Monitoreo y coaching todo junto en la misma funcion?! me quiero matar!
	* TODO: Disentangle this shit
	*/
	InteractionManager.prototype.startMonitoringInteraction = function(interaction, data, onMonitorComplete) {
		var self = this;
		data = data || {};
		let coachedAgent = data.coachedAgent || null;
		switch (interaction.type) {
			case "CHAT":
			case "IM":
				var callback = function(result, response) {
					if (!result) {
						delete self.m_monitoredInteractions[interaction.id];
					}
					onMonitorComplete(result, response);
				};

				// Flag the conversation as being monitored in order to cope with the posibility
				// of the newSession message arriving before the callback returns.
				self.m_monitoredInteractions[interaction.id] = interaction;
				if(coachedAgent) {
					self.m_interactionClients.CHAT.coachSession(interaction.id, coachedAgent, callback);
				}
				else {
					self.m_interactionClients.CHAT.monitorSession(interaction.id, callback);
				}
			break;

			case "CALL":
				var alreadyMonitoring = _.some(self.m_monitoredInteractions, function(interaction) {
					return interaction.type === "CALL";
				});
				if (alreadyMonitoring) {
					return onMonitorComplete(false,  "<%=_T('Cannot supervise more than one conversation at the same time')%>");
				}

				self.m_monitoredInteractions[interaction.id] = interaction;

				var url = inConcert.app.getBaseUrl() + "/interactions/call/monitor/";
				inConcert.invoke({
					type: "POST",
					url: url,
					data: JSON.stringify({ id: interaction.id, data : data }),
					dataType: "json",
					contentType: "application/json",
					success: function( /* response*/ ) {
						onMonitorComplete(true);
					},
					error: function(undefined, error) {
						var errorMessage;
						var alreadyClosed = false;
						delete self.m_monitoredInteractions[interaction.id];
						switch (error.result) {
							case "ALREADY_MONITORING":
								errorMessage = "<%=_T('Conversation is already being supervised.')%>";
								break;
							case "INTERACTION_NOT_TAKEN":
								if (error.state === "FINISHED" || error.state === "CANCELED") {
									alreadyClosed = true;
									errorMessage = "<%=_T('Conversation is already closed.')%>";
								} else if (error.state === "WRAPUP") {
									alreadyClosed = true;
									errorMessage = "<%=_T('Conversation is in wrapup.')%>";
								} else {
									errorMessage = "<%=_T('Conversation must be taken by an agent.')%>";
								}
								break;
							case "AGENT_UNAVAILABLE": // this should not happen
								errorMessage = "<%=_T('Cannot supervise and talk at the same time.')%>";
								break;
							default:
								errorMessage = "<%=_T('Cannot start monitoring at this time, please try again later.')%>";
								break;
						}
						onMonitorComplete(false, errorMessage, alreadyClosed);
					}
				});
				break;

			default:
				throw( format("<%=_T('Unsupported conversation type {0}')%>", interaction.type) );
		}
	};

	/**
	*/
	InteractionManager.prototype.stopMonitoringInteraction = function(interaction, onMonitorFinished) {
		var self = this;
		delete this.m_monitoredInteractions[interaction.id];
		switch (interaction.type) {
			case "CHAT":
			case "IM":
				var doStopMonitoring = function() {
					if (interaction.belongsToMe()) {
						return onMonitorFinished();
					}
					self.m_interactionClients.CHAT.leaveSession(interaction.id, onMonitorFinished);
				};
				if (interaction.state === "WAITING" || interaction.isQueued()) {
					interaction.inflate(doStopMonitoring);
				}
				else {
					doStopMonitoring();
				}
			break;
			case "CALL":
				break; // call monitoring stops with a hangup
			default:
				throw( format("<%=_T('Unsupported conversation type {0}')%>", interaction.type) );
		}
	};

	/*
		this is done when an outgoing conversation was created but the ID isn't known yet, so each time a new conversation arrive
		it must be checked whether or not it corresponds to an outgoing-but-waiting-one (such as an outgoing call)
	*/
	InteractionManager.prototype.matchToPendingInteraction = function(newInteraction) {
		var pendingInteraction = _.find(this.m_interactions, function(interaction) {
			return interaction.matchPending(newInteraction);
		});
		if (pendingInteraction) {
			var previousId = pendingInteraction.id;

			this.remove(pendingInteraction, true);
			pendingInteraction.update(newInteraction);
			pendingInteraction.pending = false;
			this.m_interactions[pendingInteraction.id] = pendingInteraction;

			return {
				interaction: pendingInteraction,
				previousId : previousId
			};
		}
	};

	InteractionManager.prototype.hasPermissionToCreateOutgoingInteractions = function()
	{
		var canCreateOutgoingInteractions = function()
		{
			var allowed_app_interaction = inConcert.permissions.to_application("interaction");
			if ( !allowed_app_interaction ) {
				return false;
			}
			var has_permission = inConcert.permissions.to_action_on_any_campaign("interaction", "int_create_outgoing");
			if (!has_permission) {
				return false;
			}
			return true;
		};

		if (!canCreateOutgoingInteractions()) {
			return {
				success: false,
				message: "<%=_T('You not have permissions for create outgoing conversations.')%>",
				title: "<%=_T('Unable to create conversation.')%>"
			};
		}
		// success
		return {
			success: true
		};
	};

	InteractionManager.prototype.conflictsToCreateOutgoingInteractions = function(proxyType)
	{
		var response = this.hasPermissionToCreateOutgoingInteractions();

		if (!response.success) {
			return response;
		}

		if (proxyType === "CALL") {
			let phone = inConcert.app.getInteractionManager().getChannelClient("CALL");
			let error = phone && phone.canMakeAnOutgoingCall();
			if (error) {
				return error;
			}
		}

		return false;
	};

	InteractionManager.prototype.checkAppointment = function() {
		var self = this;
		try {
			$.each(self.m_interactions, function(_, interaction) {
				if (!interaction || !interaction.isAppointment()) {
					return;
				}
				var lastState = self.m_lastAppointmentStates[interaction.id] || false;
				var currentState = interaction.isAppointmentActive();
				if (lastState !== currentState) {
					//notify change
					inConcert.app.getMainWindow().updateInteractions([interaction]);
				}
				self.m_lastAppointmentStates[interaction.id] = currentState;
			});
		}
		catch(e) {
			inConcert.logger.error({ message : "[InteractionManager.prototype.checkAppointment] Error processing baragent polling", error : e.message });
		}
	};

	InteractionManager.prototype.startCheckingAppointment = function() {
		var self = this;
		this.m_lastAppointmentStates = {};
		this.m_appointmentCheckInterval = window.setInterval(function() {
			self.checkAppointment();
		}, 5000);
	};

	InteractionManager.prototype.stopCheckingAppointment = function() {
		if (this.m_appointmentCheckInterval) {
			window.clearInterval(this.m_appointmentCheckInterval);
			this.m_appointmentCheckInterval = null;
		}
	};

	//NOT IMPLEMENTED
	InteractionManager.prototype.getMembersForConference = function(_, callback) {
		callback({});
	};

	//NOT IMPLEMENTED
	InteractionManager.prototype.getCampaignsAccountsForConference = function( params ) {
		var callback = params.callback || $.noop;
		callback({});
	};

	//NOT IMPLEMENTED
	InteractionManager.prototype.getAccountsAttentionLevelsForConference = function( params ) {
		var callback = params.callback || $.noop;
		callback({});
	};

	//conference the conversations to another ( team member / campaign account / attention level )
	// app_id: optional -- default value = interaction
	InteractionManager.prototype.batchConference = function( interactions, target, onConference, app_id )
	{
		let self = this;
		inConcert.logger.action({action_name : "conference_interactions", interactions_amount : interactions.length, action_type : "conversation", target: JSON.stringify(target)});

		// nothing to do without conversations
		if (!interactions || interactions.length === 0){
			return onConference(false, null, null, "<%=_T('There is no conversations to conference!')%>");
		}
		
		if (interactions[0].isAllegro()) {
			return inConcert.AllegroInteractionManager.batchConference( interactions, target, onConference, app_id );
		}
		
		inConcert.invoke({
			type: "POST",
			url: inConcert.app.getBaseUrl() + "/interactions/conference_interactions/",
			data: JSON.stringify({
				interactions : self.extractIdsFromBatch(interactions),
				target : target,
				app_id: app_id
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				self.onBatchSuccess(response, "batchConference", onConference);
			},
			error: function(response, error) {
				self.onBatchError(response, "batchConference", error, onConference);
			}
		});
	};

	InteractionManager.prototype.updateQueuedInteractions = function(interactions, noNotify) {
		var self = this;
		var changes = false;
		$.each(interactions, function(_, interaction) {
			if (!interaction || !interaction.id) {
				return;
			}
			if (!interaction.isQueued) {
				interaction.deleted = true;
			}
			if (self.m_queuedInteractions[interaction.id] && (interaction.deleted || !interaction.isQueued() || !interaction.belongsToMe())) {
				delete self.m_queuedInteractions[interaction.id];
				changes = true;
				return;
			}
			if (!self.m_queuedInteractions[interaction.id] && !interaction.deleted && interaction.isQueued() && interaction.belongsToMe()) {
				self.m_queuedInteractions[interaction.id] = true;
				changes = true;
			}
		});
		if (!noNotify && changes) {
			var panel = inConcert.app.m_mainWindow.getPanel(QUEUED_PANEL_ID);
			if (panel) {
				var counter = 0;
				$.each(self.m_queuedInteractions, function(_, interaction) {
					if (interaction) {
						counter++;
					}
				});
				panel.updateAssignedCounter(counter);
			}
		}
	};

	InteractionManager.prototype.updateAbandonedInteractions = function(interactions, noNotify) {
		const self = this;
		let changes = false;

		var isAllegroDialerAbandoed = function(interaction) {
			return interaction.state === "ABANDONED" &&
				interaction.initiative === "OUTBOUND" &&
				interaction.interactionData &&
				interaction.interactionData.isAllegro &&
				interaction.interactionData.ender === "Contact" &&
				interaction.interactionData.dialer;
		};
		$.each(interactions, function(_, interaction) {
			if (!interaction) {
				return true; // continue
			}
			let abandonedInteraction = self.m_abandonedInteractions[interaction.id];

			if (abandonedInteraction && interaction.deleted) {
				delete self.m_abandonedInteractions[interaction.id];
				changes = true;
				return true; // continue
			}

			let is_abandoned = interaction.isAbandoned() && !isAllegroDialerAbandoed(interaction);
			let belongs_to_me = interaction.belongsToMe();

			if (abandonedInteraction && (!is_abandoned || !belongs_to_me)) {
				delete self.m_abandonedInteractions[interaction.id];
				changes = true;
				return true; // continue
			}

			if (!abandonedInteraction && is_abandoned && belongs_to_me) {
				self.m_abandonedInteractions[interaction.id] = true;
				changes = true;
			}
		});

		if (!noNotify && changes) {
			const panel = inConcert.app.m_mainWindow.getPanel(ABANDONED_PANEL_ID);
			if (panel) {
				let counter = 0;
				$.each(self.m_abandonedInteractions, function(_, interaction) {
					if (interaction) {
						counter++;
					}
				});
				panel.updateAssignedCounter(counter);
			}
		}
	};
	
	InteractionManager.prototype.cleanFinishedInteractions = function(interactions) {
		let panel = inConcert.app.getMainWindow().getPanel(ASSIGNED_PANEL_ID);
		if (panel) {
			interactions.forEach(function(interaction) {
				if (interaction.isFinished !== undefined && !interaction.isFinished()) {
					panel.remove_archive_interaction(interaction.id);
				}
			});
		}
	};

	InteractionManager.prototype.countQueuedInteractions = function() {
		var counter = 0;
		$.each(this.m_queuedInteractions, function(_, interaction) {
			if (interaction) {
				counter++;
			}
		});
		return counter;
	};

	InteractionManager.prototype.countAbandonedInteractions = function() {
		var counter = 0;
		$.each(this.m_abandonedInteractions, function(_, interaction) {
			if (interaction) {
				counter++;
			}
		});
		return counter;
	};

	// TODO: #9778 - CANY - Docked.
	InteractionManager.prototype.getInteractionPanelType = function(interaction) {
		let panel_type = ASSIGNED_PANEL_ID; // por defecto
		if (interaction.isAssigned()) {
			panel_type = ASSIGNED_PANEL_ID;
		}
		if (interaction.isArchived() || interaction.isWrappedUp() || interaction.isAppointment()) {
			panel_type = ASSIGNED_PANEL_ID;
		}
		else if (interaction.isSpam()) {
			panel_type = SPAM_PANEL_ID;
		}
		else if (interaction.isQueued()) {
			panel_type = QUEUED_PANEL_ID;
		}
		else if (interaction.isWaiting()) {
			panel_type = WAITING_PANEL_ID;
		}
		else if (interaction.isAbandoned()) {
			panel_type = ABANDONED_PANEL_ID;
		}
		return panel_type;
	};

	// TODO: #9778 - CANY - Docked.
	InteractionManager.prototype.getInteractionPanel = function(interaction, customPanel) {
		let panel_type;
		let panel = customPanel;
		if (!customPanel) {
			panel_type = this.getInteractionPanelType(interaction);
			panel = inConcert.app.getMainWindow().getPanel(panel_type);
		}
		if (!panel) {
			throw('Unrecognized panel ' + panel_type);
		}
		return panel;
	};

	InteractionManager.prototype.makeInteractionOpenOptions = function(interaction, options = {}) {
		if ( interaction.isFinished() ) {
			options.archive = true;
		}
		else if ( interaction.isSpam() ) {
			options.spam = true;
			options.readonly = true;
		}
		else if ( interaction.isQueued() ) {
			options.queued = true;
			options.readonly = true;
		}
		else if ( interaction.isAppointment() ) {
			options.appointment = true;
			options.readonly = true;
		}
		else if ( interaction.isWaiting() ) {
			options.waiting = true;
			options.readonly = true;
		}
		else if ( interaction.isWrappedUp() ) {
			options.wrapup = true;
			options.readonly = true;
		}
		options.monitor = interaction.isMonitorable();
		if ( options.monitor ) {
			options.readonly = true;
		}
		// TODO: #9778 - CANY - Docked.
		if (inConcert.app.is_enable_docked_windows_interactions()) {
			if (options.from_app === "supervisor") {
				let normalized_camp_id = inConcert.Utils.getNormalizedId(interaction.campaign);
				options.id_iframe_windows_dialog_container = format("iFramesWindowsDailogsContainerForSuperviorAppCampaign_{0}", normalized_camp_id);
			}
		}
		return options;
	};

	InteractionManager.prototype.openInteractionClient = function(interaction, $row, options) {
		let im = inConcert.app.getInteractionManager();
		options = im.makeInteractionOpenOptions(interaction, options);
		// TODO: #9778 - CANY - Docked.
		let panel = im.getInteractionPanel(interaction, options.panel);
		panel.openInteractionClient(interaction, $row, options);
	};

	InteractionManager.prototype.getInteractionClient = function(interaction) {
		let panel = this.getInteractionPanel(interaction);
		return panel.getClientWindow(interaction.id);
	};

	InteractionManager.prototype.isOpenInteractionClient = function(interaction) {
		let panel = this.getInteractionPanel(interaction);
		return panel.isOpenedClientWindow(interaction.id);
	};

	InteractionManager.prototype.closeInteractionClient = function(interaction) {
		if (!this.isOpenInteractionClient(interaction)) {
			return;
		}
		let win = this.getInteractionClient(interaction);
		if (win) {
			win.close();
		}
	};

	InteractionManager.prototype.getPhoneState = function() {
		if (this.m_interactionClients.CALL) {
			return this.m_interactionClients.CALL.getPhoneState();
		}
		return "disconnected";
	};
}());
