/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global inConcert */

if (!this.inConcert) {
	this.inConcert = {};
}
if (!inConcert.interactions) {
	inConcert.interactions = {};
}

/**
 * @namespace
 * @name WhatsappInteraction
 */
(function () {

	"use strict";
	const FORMAT = inConcert.Utils.format;

	const LIST_MESSAGES_MAX_ROWS = 3;

	/**
	 * Class WhatsappInteraction.
	 * @param {Object} rawInteraction
	 */
	function WhatsappInteraction (rawInteraction)
	{
		WhatsappInteraction.uber.constructor.call(this, rawInteraction);
		this.events = this.events || [];

		// Pending interactions have no properties yet. Provide some sane defaults
		if (this.isBeingComposed()) {
			this.properties = this.properties || {};
			this.properties.display_name = "";
			this.properties.summary = "";
		}
	}

	inConcert.Utils.inherit(WhatsappInteraction, inConcert.interactions.base);

	if (typeof inConcert.interactions.whatsapp !== "function") {
		inConcert.interactions.whatsapp = WhatsappInteraction;
	}

	WhatsappInteraction.prototype.inflate = function (callback)
	{
		var self = this;
		WhatsappInteraction.uber.inflate.call(this, function(data){
			if(data) {
				self.__processEvents();
				self.sortEvents();
				callback(data);
			}
		});
	};

	WhatsappInteraction.prototype.__processEvents = function() {
		$.each(this.events || [], function(_, event) {
			if (event.message) {
				//all messages need an id
				event.message.id = event.message.id || event.message.fake_id;
			}
		});
	};

	/*
		Funcion que ordena los eventos primero el evento orgigen, luego por fecha de creacion
	*/
	WhatsappInteraction.prototype.__sort_by_creation_date = function()
	{
		this.events.sort( function( a, b )
		{
			if (!a.message.timestamp || !b.message.timestamp) {
				return a.__index - b.__index;
			}
			if (a.message.timestamp === b.message.timestamp) {
				return a.__index - b.__index;
			}
			return a.message.timestamp - b.message.timestamp;
		});
	};

	WhatsappInteraction.prototype.get_error_message = function( error_code, reason, _, error )
	{
		switch ( (error_code || "" ).toUpperCase() )
		{
			case "INTERACTION_ALREADY_EXIST_ADDRESS":
				return "<%=_T('Already exists an open interaction with same phone number')%>";
			case "TIMEOUT":
				return "<%=_T('Waiting time expired to send message')%>";
			case "THREAD_OPEN":
				if (error.agent) {
					let agent = inConcert.Utils.getIdWithoutVCC(error.agent);
					return FORMAT("<%=_T('There is an open interaction with this user assigned to agent {0}.')%>", agent);
				}
				else {
					var state = error.state || "";
					switch(state) {
						case "WAITING":
							state = "<%=_T('Waiting')%>";
							break;
						case "QUEUED":
							state = "<%=_T('Queued')%>";
							break;
					}
					return FORMAT("<%=_T('There is an open interaction with this user in state {0}.')%>", state);
				}
				break;
			case "INVALID":
				return "<%=_T('Invalid phone number')%>";
			case "BLACKLIST":
				return "<%=_T('Blacklisted addresses')%><%=_T(':')%> " +
					(error.addresses || []).join(", ");
			default:
				switch(reason || "") {
					case "conversation is closed":
						return "<%=_T('We have not received any messages from this client in the last 24 hours. Only template messages can be sent.')%>";
					case "account template message limit reached":
						return "<%=_T('The limit for sending template messages in the last 24 hours from this account has been reached.')%>";
					default:
						return "<%=_T('There was an error while sending message')%>";
				}
		}
	};

	WhatsappInteraction.prototype.createAndSendMessage = function (params)
	{
		const self = this;
		const attention_level_id = params.attention_level_id;
		const address = params.address;
		const message = params.message;
		const fake_event_id = params.fake_event_id || inConcert.Utils.generateGuid();
		const callback = params.callback || $.noop;
		
		if (message.text) {
			message.text = message.text.trim();
		}

		inConcert.logger.action({
			action_name : "new_whatsapp_message",
			id : this.id,
			type : this.type,
			campaign : this.campaign,
			fake_event_id: fake_event_id
		});

		var url = this.m_manager.getBaseUrl() + "/interactions/whatsapp/create_and_send/";
		var request = {
			fake_event_id		: fake_event_id,
			interactionID		: this.id,
			parentInteractionId : this.parentInteractionId,
			campaign			: this.campaign,
			account				: this.account,
			attention_level_id  : attention_level_id,
			contactID			: this.contact || null,
			address				: address,
			message				: message
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback( true, "OK" );
			},
			error: function(_, error){
				inConcert.logger.action({ action_name : "new_whatsapp_message", error_message : error.result });
				var message = self.get_error_message( error.result, error.reason, "SEND_MESSAGE", error );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};

	WhatsappInteraction.prototype.sendMessage = function( fake_event_id, message, callback )
	{
		let self = this;

		callback = callback || $.noop;
		fake_event_id = fake_event_id || inConcert.Utils.generateGuid();
		
		if (message.text) {
			message.text = message.text.trim();
		}

		inConcert.logger.action({
			action_name : "new_whatsapp_message",
			id : this.id,
			type : this.type,
			campaign : this.campaign,
			fake_event_id: fake_event_id
		});

		let url = this.m_manager.getBaseUrl() + "/interactions/whatsapp/send_message/";
		let request = {
			fake_event_id		: fake_event_id,
			interactionID		: this.id,
			campaign			: this.campaign,
			account				: this.account,
			message				: message
		};
		
		let onError = function(result) {
			inConcert.logger.action({ action_name : "new_whatsapp_message", error_message : result.result });
			let message = self.get_error_message(result.result, result.reason, "SEND_MESSAGE", result);
			callback(false, message, result);
		};
		
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				self.__asyncSendMessageCallback(function(result) {
					if (result.result === "OK") {
						callback(true, result.result);
					}
					else {
						onError(result);
					}
				});
			},
			error: function(_, error){
				onError(error);
			}
		});

		return fake_event_id;
	};

	WhatsappInteraction.prototype.sendInternalMessage = function( fake_event_id, message, callback )
	{
		let self = this;

		callback = callback || $.noop;
		fake_event_id = fake_event_id || inConcert.Utils.generateGuid();
		
		if (message.text) {
			message.text = message.text.trim();
		}

		let url = this.m_manager.getBaseUrl() + "/interactions/whatsapp/send_internal_message/";
		let request = {
			fake_event_id		: fake_event_id,
			interactionID		: this.id,
			campaign			: this.campaign,
			account				: this.account,
			message				: message
		};
		
		let onError = function(result) {
            console.error("onError");
			inConcert.logger.action({ action_name : "new_whatsapp_message", error_message : result.result });
			let message = self.get_error_message(result.result, result.reason, "SEND_MESSAGE", result);
			callback(false, message, result);
		};
		
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
                callback(true);
			},
			error: function(_, error){
				onError(error);
			}
		});

		return fake_event_id;
	};
	
	WhatsappInteraction.prototype.sendTemplateMessage = function(params) {
		const self = this;
		const attention_level_id = params.attention_level_id;
		const address = params.address;
		const groupId = params.groupId;
		const messageId = params.messageId;
		const languageId = params.languageId;
		const textData = params.textData;
		const callback = params.callback || $.noop;
		const buttons = params.buttons;
		const mediaUrl = params.mediaUrl;
		const mediaFileName = params.mediaFileName;
		const mediaContentType = params.mediaContentType;
		
		if (!self.canSendTemplateMessages()) {
			return callback(false, "<%=_T('You dont have permissions to send template messages.')%>");
		}
		var message = {
			msgtype : "template",
			text : textData.text,
			template : {
				group : groupId,
				id : messageId,
				language : languageId,
				header_value_param : textData.header_value_param,
				parameters : textData.values,
				text: textData.template_text,
				hsm_buttons: buttons,
				mediaUrl: mediaUrl,
				mediaFileName: mediaFileName,
				mediaContentType: mediaContentType
			}
		};
		if (this.isBeingComposed()) {
			this.createAndSendMessage({
				attention_level_id : attention_level_id,
				address : address,
				message : message,
				callback : callback
			});
		}
		else {
			this.sendMessage(undefined, message, callback);
		}
	};

	// Just returns the summary
	WhatsappInteraction.prototype.getExtendedInfo = function( max_length )
	{
		let internal_consultation_flag = this.interactionData && this.interactionData.internal_consultation_flag && this.state === "CREATED";
		if (internal_consultation_flag) {
			return inConcert.Utils.truncateWord( this.interactionData.internal_consultation_question, max_length );
		} else {
			return inConcert.Utils.truncateWord( this.properties.summary, max_length );
		}
	};
	
	WhatsappInteraction.prototype.__getMessageContent = function(message)
	{
		var data = { type : message.msgtype || "text", text : "" };
		data.text = message.text || ((message.msgtype === "reply_buttons" || message.msgtype === "list_messages") && message.interactive && message.interactive.body && message.interactive.body.text) || "";
		
		switch(message.msgtype) {
			case "template":
				if (message.template?.template_type === "hsm" && message.template.parameters?.length > 0 && message.template.text) {
					let template_text =  message.template.text;
					$.each(message.template.parameters, function(i, n) {
						template_text = template_text.replace(new RegExp("\\{{" + (i + 1) + "\\}}", "g"), n);
					});
					data.text = template_text;
				}
				break;
			case "geo":
				data.longitude = message.longitude;
				data.latitude = message.latitude;
				break;
			case "contact":
				data.contacts = message.contacts;
				break;
			case "attachment":
				if (message.mediaFail) {
					data.mediaFail = message.mediaFail;
					break;
				}
				else {
					data.fileType = message.fileType;
					data.fileName = message.fileName;
					data.fileId = message.fileId || message.fileName;
					data.fileExtension = message.fileExtension ||
						inConcert.Utils.getFileExtension(data.fileId) ||
						inConcert.Utils.getFileExtensionFromMimeType(message.mimeType);
					data.sticker = message.mimeType === "image/webp";

					switch(message.attachmentType) {
						case "repository":
							data.link = inConcert.app.getBaseDomain() + "/inconcert/apps/repository/download/?id=" + encodeURIComponent(data.fileId) + "&section=-1&part=-1";
							if (data.fileName) {
								let fileName = "-";
								if ((data.fileName || "").toLowerCase() !== (data.fileId || "").toLowerCase()) {
									fileName = data.fileName;
									const max_length = 50;
									if (fileName.length > max_length) {
										// Si el mensaje (nombre del archivo), supera los "max_length" caracteres, lo trunco para que el nombre no se a muy extenso
										fileName = inConcert.Utils.truncateWord(fileName, max_length, "...");
									}
								}
								let timestampObject = new Date(message.timestamp * 1000);
								let userTimezoneObject = inConcert.app.getUserTimezoneObject();
								timestampObject = inConcert.Utils.convertLocalDateToTimezoneDate(timestampObject, userTimezoneObject);
								let stringDate = inConcert.Utils.getStringFromDate(timestampObject, false, "full", "ymd", "-");
								/*
									Al nombre sugerido lo conformo de la siguiente forma:
										1 - Con el prefjo del Canal
										2 - Con el texto del mensaje, si es que lo tiene, sino va vacio un "-".
										3 - Con la fecha y hora del mensaje, en el timezone del usuario.
										4 - La extension.
								*/
								let suggestedName = FORMAT("<%=_T('WhatsApp File-{0}-{1}.{2}')%>", fileName, stringDate, data.fileExtension);
								data.link += "&name=" + encodeURIComponent(suggestedName);
							}
							break;
						case "url":
							data.link = message.url;
							break;
						default:
							data.mediaFail = true;
							break;
					}
					break;
				}
			case "reply_buttons":
				var quickReplies = "<div class='whatsappQuickrepliesContainer'>";
				$.each(message.interactive.action.buttons, function(_, quickReply) {
					var text = quickReply.reply.title;
					quickReplies += FORMAT("".concat(
						"<button class='whatsappQuickreply'>",
							"{0}",
						"</button>"
					), text);
				});
				quickReplies += "</div>";
				data.media = quickReplies;
				break;
			case "list_messages":
				/* EJEMPLO:
                    "sections": [
                        {
                            "title": "section1_title", 
                            "rows": [
                                {
                                    "id": "section1_row1", 
                                    "title": "section 1 row 1", 
                                    "description": "This is the section 1 row 1"
                                }, 
                                {
                                    "description": "This is the section 1 row 2", 
                                    "title": "section 1 row 2", 
                                    "id": "section1_row2"
                                }
                            ]
                        }, 
                        {
                            "rows": [
                                {
                                    "description": "This is the section 2 row 1", 
                                    "title": "section 2 row 1", 
                                    "id": "section2_row1"
                                }, 
                                {
                                    "id": "section2_row2", 
                                    "title": "section 2 row 2", 
                                    "description": "This is the section 2 row 2"
                                }, 
                                {
                                    "description": "This is the section 2 row 3", 
                                    "title": "section 2 row 3", 
                                    "id": "section2_row3"
                                }
                            ], 
                            "title": "section2_title"
                        }
                    ], 
				*/

				let message_uid = 	inConcert.Utils.generateGuid();
				let listMessages = "<div id='" + message_uid + "_whatsappListMessagesContainer' class='whatsappListMessagesContainer'>";
				let num_rows = 0;
				let show_more_selector = "";
				let style = "";
				$.each(message.interactive.action.sections, function(_, section) {
					show_more_selector = num_rows >= LIST_MESSAGES_MAX_ROWS ? "show_more_selector" : "";
					style = num_rows >= LIST_MESSAGES_MAX_ROWS ? "style='display:none;'" : "";
					listMessages += FORMAT("".concat(
						"<h6 ", show_more_selector, " class='whatsappListSectionTitle' ", style, ">",
							"{0}",
						"</h6>"
					), section.title);
					$.each(section.rows, function(_, row) {
						num_rows++;
						show_more_selector = num_rows > LIST_MESSAGES_MAX_ROWS ? "show_more_selector" : "";
						style = num_rows > LIST_MESSAGES_MAX_ROWS ? "style='display:none;'" : "";
						listMessages += FORMAT("".concat(
							"<button ", show_more_selector, " class='whatsappListMessage' ", style, ">",
								"{0}",
							"</button>"
						), row.title);
					});
				});
				if(num_rows > LIST_MESSAGES_MAX_ROWS) {
					listMessages += "".concat(
						"<a href='javascript:void(0);' show_more_selector_link=", message_uid, " class='whatsappListLink'>",
							"<%=_T('See more')%>",
						"</a>"
					);
				}
				listMessages += "</div>";
				data.media = listMessages;
				break;
		}
		return data;
	};

	WhatsappInteraction.prototype.getQuotedMessage = function(quotedId)
	{
		if (quotedId) {
			for (let i = 0; i < this.events.length; i++) {
				if ( this.events[i].message && this.events[i].message.id && this.events[i].message.id === quotedId ) {
					return this.events[i];
				}
			}
		}
	};

	WhatsappInteraction.prototype.getMessageContent = function(event)
	{
		var message = event.message || {};
		var data = this.__getMessageContent(message);
		if (message.quoted) {
			const event_quoted = this.getQuotedMessage(message.quoted);
			if (event_quoted) {
				data.quoted = this.__getMessageContent(event_quoted.message);
			}
		}
		return data;
	};

	WhatsappInteraction.prototype.getMessageText = function(event)
	{
		return this.getMessageContent(event).text;
	};

	WhatsappInteraction.prototype.getEventShortDescription = function(sorted_index) {
		var ev = this.getEvent(sorted_index);
		return this.getMessageText(ev);
	};

	WhatsappInteraction.prototype.getMessageStatus = function(event) {
		return event.message && event.message.status || null;
	};

	WhatsappInteraction.prototype.getMessageStatusReason = function(event) {
		return event.message && event.message.reason || null;
	};

	WhatsappInteraction.prototype.getMessageStatusDate = function(event) {
		var status = this.getMessageStatus(event);
		var timestamp = status && event.message["timestamp_" + status] || null;
		return timestamp && new Date(timestamp * 1000) || null;
	};

	/**
	* @param index
	*/
	WhatsappInteraction.prototype.getDateObject = function(index) {
		index = index || 0;
		const event = this.getEvent(index);
		if (!event || !event.message) {
			return "";
		}
		const event_date = event.message.timestamp * 1000;
		return this.convertDateStringToDate(event_date);
	};

	WhatsappInteraction.prototype.getRawAddress = function(index) {
		this.assert_inflated();
		return this.events[index].address && this.events[index].address.Number || "";
	};

	WhatsappInteraction.prototype.getSource = function( sorted_index )
	{
		this.assert_inflated();
		if ( !this.events || sorted_index === undefined || sorted_index === -1 ) {
			return "";
		}
		var event = this.events[ sorted_index ];
		if ( event.creator === "agent" ) {
			return event.agent || inConcert.app.getCompleteId();
		}
		if (event.message && event.message.name) {
			return event.message.name;
		}
		if (event.message && event.message.chatname) {
			return event.message.chatname;
		}
		if (event.address && event.address.Number) {
			return event.address.Number;
		}
		return "";
	};

	WhatsappInteraction.prototype.getEventUserAvatarURI = function( event )
	{
		if (typeof(event) !== "object" || typeof(event.message) !== "object") {
			return null;
		}
		if ( event.creator === "agent" && event.agent ) {
			var user_id = event.agent;
			if ( !user_id ) {
				user_id = inConcert.app.getCompleteId();
			}
			if ( user_id.match(inConcert.app.idAtVCCRegExp) ) {
				user_id = RegExp.$1;
			}
			return inConcert.app.getBaseUrl() + "/users/get_avatar/" + user_id;
		}
		else if ( event.creator === "bot" ) {
			return "/inconcert/apps/agent/images/botAvatar.png";
		}
		else if ( event.creator === "contact" && event.message.chatid ) {
			if (this.integrationHasAvatars()) {
				return inConcert.app.getBaseDomain() + "/inconcert/apps/repository/download/?id=WhatsappAvatar_" + 
					encodeURIComponent(event.message.chatid) + "&section=-1&part=-1";
			}
			else {
				return null;
			}
		}
		else {
			return null;
		}
	};
	
	var integrationsWithAvatars = {
		WEB : true
	};
	WhatsappInteraction.prototype.integrationHasAvatars = function() {
		return integrationsWithAvatars[this.getIntegrationType()] || false;
	};

	//returns the conversation contact or member display text
	WhatsappInteraction.prototype.getUserName = function( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}
		
		if ( event.creator === "agent" && event.agent ) {
			return this.get_agent_name( event );
		}
		if ( event.creator === "bot" ) {
			return this.getBotName();
		}
		if (event.creator === "contact" && this.interactionData.batchAddressName) {
			return this.interactionData.batchAddressName;
		}
		if (event.message && event.message.name) {
			if (event.message.name === "T" || event.message.name === "You") {
				return "<%=_T('Me')%>";
			}
			return event.message.name;
		}
		if (event.message && event.message.chatname) {
			if (event.message.chatname === "T" || event.message.chatname === "You") {
				return "<%=_T('Me')%>";
			}
			return event.message.chatname;
		}
		if (event.address && event.address.Number) {
			return event.address.Number;
		}
		return "";
	};
	
	WhatsappInteraction.prototype.getUserNumber = function ( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}
		
		if (event.address && event.address.Number) {
			return event.address.Number;
		}
		return "";
	};

	WhatsappInteraction.prototype.get_user_event_id = function( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}
		if ( event.creator === "agent" ) {
			return event.agent || inConcert.app.getCompleteId();
		}
		if (event.userid) {
			return event.userid;
		}
		return "";
	};

	/**
	* Returns a displayable name, which was pre computed on the server. It extracts the user name from the first event 
	* of the interaction or the agent id, depending on which one originated the interaction.
	* The computed property has a prefixed value to it. If the prefix is "text:", then the rest is to be taken 
	* verbatim. It is the name of a FB user.
	* If the prefix is "agent_id:", then we need to perform a lookup to translate the id to an agent name.
	* 
	* @return {String}
	*/
	WhatsappInteraction.prototype.getDisplayName = function()
	{
		var display_name = this.properties.display_name;
		if(display_name.match(/text:(.+)/)) {
			return RegExp.$1;
		}
		else if(display_name.match(/phone:(.+)/)) {
			return RegExp.$1;
		}
		else if(display_name.match(/agent_id:/)) {
			var agent_id = RegExp.$1;
			if (agent_id === inConcert.app.getCompleteId()) {
				return inConcert.app.getLoggedUserName();
			}
			else {
				var agent_name;

				inConcert.app.getTeamMember( agent_id, function(agent)
					{
						if (!agent) {
							return;
						}
						// Esto por qué carajo no lo hago una vez sola?
						var have_name = false;
						if ((agent.Name || "").trim().length > 0) {
							agent_name = agent.Name;
							have_name = true;
						}
						if ((agent.FirstSurname || "").trim().length > 0) {
							agent_name = (have_name ? (agent_name + " ") : "") + agent.FirstSurname;
						}
					},
					true // from cache
				);

				agent_name = agent_name || agent_id;
				return agent_name;
			}
		}
		return display_name;
	};

	WhatsappInteraction.prototype.getFirstUnreadEventIndex = function() {
		this.assert_inflated();

		var eventIndex = inConcert.interactions.base.prototype.getFirstUnreadEventIndex.call(this);
		if (typeof(eventIndex) === "number" && this.events[eventIndex]) {
			var event = this.events[eventIndex];
			return (typeof(event.parentIndex) !== "undefined") ? event.parentIndex : eventIndex;
		}
		else {
			return eventIndex;
		}
	};

	/* @return {Boolean} */
	// IB: ESTA FUNCION SE PUEDE CALCULAR UNA VEZ EN EL SERVER Y LISTO
	WhatsappInteraction.prototype.eventsAreFromTheSameClient = function() {
		this.assert_inflated();

		let events_from_the_same_client = true;
		let last_event_client = null;

		for (let i = 0; i < this.events.length; i++) {
			let current_event_client = this.getRawAddress(i);
			// first iteration
			if (!last_event_client){
				last_event_client = current_event_client;
			} else {
				if (current_event_client !== last_event_client){
					events_from_the_same_client = false;
					break;
				}
			}
		}

		return events_from_the_same_client;
	};

	WhatsappInteraction.prototype.__getSendAttachmentError = function(response) {
		switch(response || "") {
			case "conversation is closed":
				return "<%=_T('We have not received any messages from this client in the last 24 hours. Only template messages can be sent.')%>";
			default:
				return "<%=_T('Error uploading file.')%>";
		}
	};

	WhatsappInteraction.prototype.isValid = function() {
		/*this.assert_inflated();
		// IB: Revisar cual es el sentido de esto

		if (this.state === "PENDING") {
			return true;
		}
		if (!$.isArray(this.events) || this.events.length === 0) {
			return false;
		}*/
		return true;
	};
	
	WhatsappInteraction.prototype.getIntegrationType = function() {
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return;
		}
		const account = this.getAccountConfiguration();
		return campaign.getWhatsappAccountIntegrationType(account);
	};
	
	WhatsappInteraction.prototype.hasGroups = function() {
		var campaign = inConcert.app.getCampaign(this.campaign);
		var account = this.getAccountConfiguration();
		return campaign.hasAccountWhatsappGroups(account);
	};
	
	var groupsCache = null;
	function makeGroupsCache() {
		groupsCache = new inConcert.Cache(
			function(fetchData, callback) {
				var url = fetchData.baseUrl + "/campaigns/whatsapp/list_groups/";
				var request = {
					campaignAtVCC		: fetchData.campaign,
					account				: fetchData.account
				};
				inConcert.invoke({
					type: "POST",
					url: url,
					data: JSON.stringify(request),
					dataType: "json",
					contentType: "application/json",
					success: function(response) {
						callback(response.groups);
					},
					error: function(_, error){
						callback(false, error.reason || "error getting groups");
					}
				});
			},
			function(fetchData) {
				return fetchData.campaign.Id + ":" + fetchData.account;
			},
			60
		);
	}
	
	WhatsappInteraction.prototype.listGroups = function(callback) {
		if (!this.hasGroups()) {
			return callback(false, "account not supports groups");
		}
		if (!groupsCache) {
			makeGroupsCache();
		}
		groupsCache.get({
				baseUrl : this.m_manager.getBaseUrl(),
				campaign : this.campaign, 
				account : this.account
			}, 
			function(groups, error) {
				callback(groups, error);
			});
	};
	
	WhatsappInteraction.prototype.getChatType = function() {
		return this.interactionData && this.interactionData.chatType || "user";
	};
	
	WhatsappInteraction.prototype.isGroup = function() {
		return this.getChatType() === "group";
	};

	WhatsappInteraction.prototype.forward = function( address, messages, callback ) 
	{
		callback = callback || $.noop;
		let self = this;
		let fake_event_id = inConcert.Utils.generateGuid();
		
		inConcert.logger.action({action_name : "whatsapp_forward", id : this.id, type : this.type,
			campaign : this.campaign, address : address, account : this.account, action_type : "conversation", fake_event_id: fake_event_id});

		let url = this.m_manager.getBaseUrl() + "/interactions/whatsapp/forward/";
		var request = {
			fake_event_id : fake_event_id,
			campaignId	  : this.campaign,
			account       : this.account,
			page          : this.getPage(),
			interactionID : this.id,
			address		  : address,
			messages      : messages
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				var message;
				var status = result.result === "OK";
				if (!status) {
					message = self.get_error_message( result.result, result.reason, "FORWARD", result );
				}
				callback( status, message, result.result );
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "whatsapp_forward", error_message : error.result });
				var message = self.get_error_message( error.result, error.reason, "FORWARD", error );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};
	
	let internal_message_keys = {
		chattype : true,
		chatid : true,
		wappid : true,
		id : true,
		name : true,
		forward : true,
		status : true,
		origin : true
	};
	
	WhatsappInteraction.prototype.__isForwardMessageKey = function(key) {
		return !internal_message_keys[key] && key.indexOf("timestamp") < 0;
	};
	
	WhatsappInteraction.prototype.getForwardMessage = function(index) {
		let self = this;
		let event = this.getEvent(index);
		return event && event.message && Object.keys(event.message).reduce(function(message, key) {
			if (self.__isForwardMessageKey(key)) {
				message[key] = event.message[key];
			}
			return message;
		}, {}) || null;
	};
	
	WhatsappInteraction.prototype.getForwarded = function(index) {
		let event = this.getEvent(index);
		return event && event.message && event.message.forward || null;
	};
	
	WhatsappInteraction.prototype.hasForward = function() {
		return true;
	};
	
	WhatsappInteraction.prototype.isSendSendAttachmentAsync = function() {
		return true;
	};
	
	WhatsappInteraction.prototype.isSendSharedFileAsync = function() {
		return true;
	};
	
}());
