/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  undef:true, unused:vars, curly:false, browser:true, indent:false, maxerr:50, quotmark: false,
  smarttabs:true, -W014, esversion: 11 */

/* global inConcert, EventEmitter, _, console */

/*
	This module contains everything related to interaction creation and management. Here you will find suff such as:
	+ Conversation request and creation
	+ Integration with IR operations
	+ Event reception handling & processing

	Exposes a single object "inConcert.interactions.base" where everything else derives from.
*/

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (!this.inConcert) {
	this.inConcert = {};
}
if (!inConcert.interactions) {
	inConcert.interactions = {};
}

(function () {

	"use strict";

	/**
	Copes with addresses in both "tweety" format and "Fullcontact" with old format.
	Mutates 'address' in place.
	@param address
	*/
	var normalize_old_format_address = function(address) {
		if(typeof(address) === "object" && !address.displayable) {
			var type = address.type ? address.type.toLowerCase() : address.Type.toLowerCase();
			switch(type) {
				case "home":
				case "office":
				case "cellular":
					address.displayable = address.Number || address.address;
					address.Type = "Phone";
				break;
				case "phone":
					address.displayable = address.Number || address.address;
					address.Type = "Phone";
				break;
				case "im":
					address.displayable = address.buddy || address.address;
					address.Type = "IM";
				break;
				case "facebook":
					address.displayable = address.UID || address.address;
					address.Type = "Facebook";
				break;
				case "instagram":
					address.displayable = address.User || address.address;
					address.Type = "Instagram";
				break;
				case "youtube":
					address.displayable = address.User || address.address;
					address.Type = "Youtube";
				break;
				case "appstore":
					address.displayable = address.User || address.address;
					address.Type = "Appstore";
				break;
				case "googleplay":
					address.displayable = address.User || address.address;
					address.Type = "Googleplay";
				break;
				case "twitter":
					address.displayable = address.User || address.address;
					address.Type = "Twitter";
				break;
				case "mail":
					address.displayable = address.address;
					address.Type = "Mail";
				break;
				default:
					address.displayable = "n/a";
				break;
			}
		}
	};

	var format = inConcert.Utils.format;

	// Defines a generic interaction object with its common behaviour()
	// constructor
	function Interaction (rawInteraction)
	{
		// actions for apps ( interaction / supervisor )
		this.m_actions = {
			COMMENT         : { interaction : "int_comment"        , supervisor : "sup_monitoring_comment"        },
			FINISH          : { interaction : "int_finish"         , supervisor : null                            },
			HANGUP          : { interaction : "int_hang_up"        , supervisor : null                            },
			LABEL           : { interaction : "int_label"          , supervisor : "sup_monitoring_label"          },
			REQUEUE         : { interaction : "int_requeue"        , supervisor : null                            },
			REOPEN          : { interaction : "int_archive_reopen" , supervisor : "sup_archive_reopen"            },
			SCHEDULE        : { interaction : "int_schedule"       , supervisor : null                            },
			TICKET_EMIT     : { interaction : "int_emit_ticket"    , supervisor : "sup_emit_ticket"               },
			TICKET_SEARCH   : { interaction : "int_search_ticket"  , supervisor : "sup_search_ticket"             },
			TAKE            : { interaction : "int_queued_take"    , supervisor : "sup_queued_take"               },
			TRANSFER        : { interaction : "int_transfer"       , supervisor : "sup_monitoring_transfer"       },
			TRANSFER_QUEUED : { interaction : null                 , supervisor : "sup_queued_transfer"           },
			CHAT_FILE_TRANSFER : { interaction : "int_chat_file_transfer", supervisor : null, no_category : true  },
			FORWARD_MESSAGES : { interaction : "int_forward_messages", supervisor : null, no_category : true  },
			SCREEN_SHARE    : { interaction : "int_screen_share"   , supervisor : null                            }
		};

		//the contact for this conversation
		this.m_contact = null;
		//the native session for this conversation
		this.m_session = null;
		//the channel view for this conversation, only valid when active
		this.m_view = null;
		this.m_has_sorted_events = false;
		this.m_inflated = false;

		if (rawInteraction && !rawInteraction.timestamp)
		{
			var zp = inConcert.Utils.zeropad;
			var now_utc = inConcert.Utils.getNowUTCDateByServer();
			rawInteraction.timestamp = format("{0}-{1}-{2} {3}:{4}:{5}",
				now_utc.getUTCFullYear(),
				zp(now_utc.getUTCMonth() + 1),
				zp(now_utc.getUTCDate()),
				zp(now_utc.getUTCHours()),
				zp(now_utc.getUTCMinutes()),
				zp(now_utc.getUTCSeconds())
			);
		}

		$.extend(true, this, rawInteraction);

		if (this.interactionData && this.interactionData.contactAddresses) {
			$.each( this.interactionData.contactAddresses || {}, function( undefined, address ) {
				normalize_old_format_address(address);
			});
		}

		// For interactions which are being composed, provide some default properties
		if (this.isBeingComposed() && !this.properties) {
			this.properties = {
				event_count: 0,
				unread_event_count: 0
			};
		}
		
		if (this.steps && this.steps.length > 0) {
			this.__setInflated();
		}
		
		this.__notifyNew();
	}

	inConcert.Utils.inherit(Interaction, EventEmitter);

	if (typeof inConcert.interactions.base !== "function") {
		inConcert.interactions.base = Interaction;
	}

	/**
	dispose: Perform cleanup tasks, mostly used to help the GC.
			 This must be called whenever the conversation is "destroyed"
	*/
	Interaction.prototype.dispose = function()
	{
		console.debug("Interaction.prototype.dispose", this.id);
		this.m_contact = null;
		this.m_session = null;
		this.m_view = null;
		this.removeRelationshipPendingAppointment();
	};

	/**
	Asserts that an interaction had been 'inflated' (that is, we've retrieved all its properties).
	Every function that operates on events, steps, analysisTimes or readers *MUST* call this function beforehand.
	If not inflated, an error will be thrown. That means you have a bug in your code.
	*/
	Interaction.prototype.assert_inflated = function()
	{
		if(!this.m_inflated) {
			throw new Error("Interaction not inflated. Id: " + this.id);
		}
	};

	Interaction.prototype.isInflated = function () { return this.m_inflated; };

	/*
		Esta funcion unicamente deber ser invocada si la interaccion se encuentra inflada.
	*/
	Interaction.prototype.__setInflated = function() {
		this.m_inflated = true;
	};

	/**
	* Retrieves all data (events and stuff) of an interaction, asynchronously.
	* @param callback
	*/
	Interaction.prototype.inflate = function (callback)
	{
		var self = this;

		if (self.interactionData.internal_consultation_flag && self.state === "CREATED") {
			// Tirar el request siempre. Luego que se infla, la interaccion igual puede cambiar.
			return this.m_manager.inflateBotInteraction(this.id, function inflateCallback(data) {
				if(data) {
					$.extend( self, data );
					self.__setInflated();
					self.__processUpdateNotifications();
					callback(data);
				}
			});
		} else {
			// Tirar el request siempre. Luego que se infla, la interaccion igual puede cambiar.
			return this.m_manager.inflateInteraction(this.id, function inflateCallback(data) {
				if(data) {
					$.extend( self, data );
					self.__setInflated();
					self.__processUpdateNotifications();
					callback(data);
				}
			});
		}
	};

	/**
	* After dealing with an interaction, deflate sets it back to normal. Tries to keep memory
	* usage down.
	*/
	Interaction.prototype.deflate = function (_ /*callback*/)
	{
		this.m_inflated = false;
		delete this.events;
	};

	/**
	* Most interactions (mail, facebook, twitter, etc) supports events. In a mail interaction, each email in the 
	* conversation is an event. However, other interactions (calls, webchats, webcontacts) don't have events.
	* Those will override this method.
	*/
	Interaction.prototype.supportsEvents = function ()
	{
		return true;
	};

	/**
	* An interaction being composed is, for instance, an agent writing a new email, or a new Facebook post, or 
	* a new message on Twitter. We introduced this method because the word "PENDING" has multiple (and different) 
	* meanings in the codebase. It applies to labels, or the state of a request.
	*/
	Interaction.prototype.isBeingComposed = function ()
	{
		// check. should I enforce 'initiative' to be OUTBOUND ?
		return this.state === "PENDING" || (this.state === "CREATED" && !this.isBotInternalConsultation());
	};

	Interaction.prototype.canInflateInteraction = function ()
	{
		// check. should I enforce 'initiative' to be OUTBOUND ?
		return !this.isBeingComposed();
	};	

	//Ends the conversation (called after hangup or physical finish)
	Interaction.prototype.end = function() {
	};

	Interaction.prototype.convertToJSON = function() {
		var self = this;
		var encoded = JSON.stringify(self, function(key, value) {
			if ("m_manager" === key) {
				return undefined;
			}
			if ("m_notificationBubble" === key) {
				return undefined;
			}
			return value;
		});
		return encoded;
	};
	
	let INTERNAL_KEYS = {
		_events : true
	};
	
	Interaction.prototype.toRaw = function() {
		let self = this;
		return Object.keys(self).reduce(function(table, key) {
			if (key.indexOf("m_") !== 0 && !INTERNAL_KEYS[key]) { 
				table[key] = self[key]; 
			} 
			return table;
		}, {});
	};

	Interaction.prototype.hasCCIntegration = function()
	{
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		return campaign.hasCCIntegration();
	};

	Interaction.prototype.getIntegrationData = function()
	{
		var campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		var integration = campaign.getIntegrationData();
		if (integration) {
			integration.isAllegro = this.isAllegro();
		}
		return integration;
	};

	Interaction.prototype.__getExtIntegrationOverrideAction = function(action, type) {
		if (typeof action !== "string" || typeof type !== "string") {
			return;
		}
		let integration;
		switch (type) {
			case "HELPDESK":
				integration = this.getHelpDeskIntegrationData();
				break;
			case "MAS":
				integration = this.getMasIntegrationData();
				break;
			default:
				break;
		}

		if (typeof integration === "object" && typeof integration.override_actions === "object"
		 && integration.override_actions[action.toLowerCase()]) {
			return integration.override_actions[action.toLowerCase()];
		}
		return;
	};

	Interaction.prototype.hasIntegrationOverrideActionPopupShow = function(actionType) {
		
		let override_action = this.getIntegrationOverrideAction(actionType);
		if (typeof override_action === 'object' && override_action.hasOwnProperty('show')) {
			if (override_action.show.toLowerCase() === 'popup') {
				return true;
			}
		}

		override_action = this.__getExtIntegrationOverrideAction(actionType, "HELPDESK");
		if (typeof override_action === 'object' && override_action.hasOwnProperty('show')) {
			if (override_action.show.toLowerCase() === 'popup') {
				return true;
			}
		}

		override_action = this.__getExtIntegrationOverrideAction(actionType, "MAS");
		if (typeof override_action === 'object' && override_action.hasOwnProperty('show')) {
			if (override_action.show.toLowerCase() === 'popup') {
				return true;
			}
		}

		return false;
	};

	Interaction.prototype.getIntegrationOverrideAction = function(action) {
		if (!this.hasCCIntegration() || typeof action !== "string") {
			return false;
		}
		var integration = this.getIntegrationData();
		if (typeof integration !== "object" || !integration.override_actions) {
			return false;
		}
		return integration.override_actions[action.toLowerCase()];
	};

	Interaction.prototype.hasIntegrationWithMarketingAndSales = function()
	{
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		return campaign.hasIntegrationWithMarketingAndSales();
	};

	Interaction.prototype.getMasIntegrationData = function()
	{
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		let mas_integration = campaign.getMasIntegrationData();
		if (mas_integration) {
			mas_integration.isAllegro = this.isAllegro();
		}
		return mas_integration;
	};

	Interaction.prototype.hasIntegrationWithHelpDesk = function()
	{
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		return campaign.hasIntegrationWithHelpDesk();
	};

	Interaction.prototype.getHelpDeskIntegrationData = function()
	{
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		let helpdesk_integration = campaign.getHelpDeskIntegrationData();
		if (helpdesk_integration) {
			helpdesk_integration.isAllegro = this.isAllegro();
		}
		return helpdesk_integration;
	};

	Interaction.prototype.hasIntegrationWithOccWidget = function(widget_id)
	{
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		return campaign.hasIntegrationWithOccWidget(widget_id);
	};

	Interaction.prototype.getWidgetIntegrationData = function(widget_id)
	{
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return false;
		}
		let widget_integration = campaign.getWidgetIntegrationData(widget_id);
		if (widget_integration) {
			widget_integration.isAllegro = this.isAllegro();
		}
		return widget_integration;
	};

	Interaction.prototype.hasPermission = function( action_type, app_id )
	{
		var log_warning = function( message ) {
			inConcert.logger.warn({
				action_name : "Interaction.hasPermission",
				message     : message                    ,
				action_type : action_type || ""          ,
				app_id      : app_id || ""
			});
		};

		if ( !action_type ) {
			log_warning( "interaction action type is undefined" );
			return false;
		}
		var campaign_id = this.campaign;
		if ( !campaign_id ) {
			log_warning( "campaign type is undefined" );
			return false;
		}
		if ( !this.m_actions[ action_type ] ) {
			log_warning( "unrecognized interaction action type" + action_type );
			return false;
		}
		app_id = app_id || "interaction";
		var action_id = this.m_actions[ action_type ][ app_id ];
		if ( typeof action_id !== "string" ) {
			if ( action_id !== null ) {
				log_warning( "unrecognized application " + app_id );
			}
			return false;
		}

		// La categoria aca siempre tiene que ser "interaction", ya que la validacion de uno de estos permisos a nivel de esta clase, siempre son de categoria "interaction", si no se encuentra en esta categoria, quiere decir que el permiso se ha definido incorrectamente en la BD de la CF Application.
		var category = !this.m_actions[ action_type ].no_category && "interaction" || undefined;
		return inConcert.permissions.to_action( app_id, action_id, category, campaign_id, true /* no log */ );
	};

	Interaction.prototype.hasContactConflict = function(undefined) {
		if (this.interactionData && this.interactionData.conflict){
			return true;
		}
		return false;
	};

	/**
	* Retrieves the assigned ticket, if any (i.e. Ticket#09091234)
	*/
	Interaction.prototype.getAssignedTicket = function()
	{
		if (!this.interactionData) {
			return;
		}
		return this.interactionData.ticket;
	};

	/**
	* Emits a new ticket and assigns it to this interaction
	* @param callback
	*/
	Interaction.prototype.emitTicket = function (callback = $.noop)
	{
		let self = this;

		inConcert.logger.action({action_name: "emit_ticket", id: this.id, type: this.type, campaign: this.campaign, action_type: "interaction"});

		if( this.isInternal() || this.isBeingComposed() || !this.isTaken() ) {
			callback(false, "UNABLE_TO_EMIT_TICKET");
			return;
		}
		
		let api = inConcert.app.getTicketsAPI();
		api.emit({ interaction_id : self.id }, function(status, ticket) {
			if (status) {
				self.interactionData.ticket = ticket;
				let view = self.getView();
				if (view) {
					view.updateTicketInfo(ticket);
				}
				callback(true, ticket);
			}
			else {
				callback(false, ticket || "An unexpected error occurred, please try again later...");
			}
		});
	};

	// finishes the conversation(called when wrapup is finished)
	// app_id is optional, default value "interaction"
	Interaction.prototype.archive = function( disposition, comment, onArchived, app_id )
	{
		var self = this;
		this.__cleanupAutoWrapupTimer();

		onArchived = onArchived || $.noop;

		this.pending = true;
		this.disposition = disposition;
		this.comment = comment;

		// call batch archive with one conversation
		this.m_manager.batchArchive([this],
			this.disposition,
			this.comment,
			function( success, successful_nodes, unsuccessful_nodes, error) {
				self.pending = false;
				onArchived( success, successful_nodes, unsuccessful_nodes, error );
			},
			app_id
		);
	};

	//reopens the current conversation
	// app_id is optional, default value "interaction"
	Interaction.prototype.reopen = function( callbackFn, app_id )
	{
		var self = this;

		inConcert.logger.action({action_name : "reopen_interaction", id : this.id, type : this.type, campaign : this.campaign, action_type : "conversation"});
		callbackFn = callbackFn || $.noop;
		this.pending = true;

		var getErrorMessage = function(err, message)
		{
			err = (err || "").toUpperCase();
			switch(err) {
				case "FORBIDDEN":
					return format("<%=_T('You dont have permissions to reopen conversations on {0} campaign.')%>", self.campaign);

				case "USER DOES NOT BELONG TO CAMPAIGN":
					return format("<%=_T('Unable to reopen conversation. You does not belong to {0} campaign.')%>", self.campaign);

				case "INTERACTION_DOES_NOT_BELONG_TO_YOU":
					return "<%=_T('Unable to reopen conversation. The conversation does not belong to you.')%>";

				case "USER DOES NOT BELONG TO ATTENTION LEVEL":
					return "<%=_T('Unable to reopen conversation. You does not belong to attention level.')%>";

				case "USER BELONGS TO ANOTHER LEVEL OF ATTENTION":
					return "<%=_T('Unable to reopen conversation. You belongs to another level of attention.')%>";

				case "INTERACTION HAS AN INVALID ATTENTION LEVEL":
					return "<%=_T('Unable to reopen conversation. Has an invalid attention level.')%>";

				default:
					return message || "<%=_T('Failed reopening conversation, please try again later...')%>";
			}
		};

		if (!this.isArchived()) {
			this.pending = false;
			callbackFn(false, getErrorMessage("DEFAULT"));
			return false;
		}

		/*
			Si la interaccion que se quiere reabrir, tiene relacionda una nueva interaccion que se ha creado a partir de esta, NO permitimos su reapertura.
		*/
		if (this.relatedInteractions) {
			if (this.relatedInteractions.newInteraction) {
				this.pending = false;
				callbackFn(false, getErrorMessage("DEFAULT"));
				return false;
			}
		}

		if (app_id === "interaction") {
			// Valida si desde la app "interaction/global search", tiene permisos para reabrir interacciones que no le pertenecen
			var reopen_not_belong_to_you = "int_global_search_reopen_interactions_not_belong_to_you";
			var has_permission_reopen_conv_not_belong_to_you = inConcert.permissions.to_action(app_id,
				reopen_not_belong_to_you,
				"global_search_item",
				this.campaign,
				true /* no log */
			);
			if (!has_permission_reopen_conv_not_belong_to_you) {
				// Si no tiene permisos para reabrir interacciones que no le pertenecen, valida que la interaccion sea de el.
				var archived_belongs_to_me = this.archivedBelongsToMe();
				if ( !archived_belongs_to_me ) {
					this.pending = false;
					callbackFn(false, getErrorMessage("INTERACTION_DOES_NOT_BELONG_TO_YOU"));
					return false;
				}
			}
		}

		var request = {
			id : this.id,
			type : this.type,
			app_id: app_id // optional
		};
		var url = this.m_manager.getBaseUrl() + "/interactions/reopen_interaction/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				self.pending = false;
				if (response.result === "OK") {
					callbackFn(true);
				}
				else {
					callbackFn(false, getErrorMessage(response.reason, response.reasonTraslated));
				}
			},
			error: function( response /*, error */ ) {
				self.pending = false;
				var errors = {
					"403": "FORBIDDEN"
				};
				callbackFn(false, getErrorMessage(errors[response.status]));
			}
		});
	};

	//Adds a comment to the conversation
	//	app_id: optional -- default value = interaction
	Interaction.prototype.comment = function( comment, callbackFn, app_id ) {
		inConcert.logger.action({action_name : "comment_interaction", id : this.id, type : this.type, campaign : this.campaign, action_type : "conversation"});

		callbackFn = callbackFn || $.noop;

		var request = {
			id : this.id,
			comment : comment,
			app_id : app_id
		};
		var url = this.m_manager.getBaseUrl() + "/interactions/comment_interaction/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function( /* commented */ ) {
				callbackFn(true);
			},
			error: function(response /*, error */ ) {
				var errors = {
					"403": "FORBIDDEN"
				};
				callbackFn(false, errors[response.status]);
			}
		});
	};

	Interaction.prototype.youHasUnreadComments = function()
	{
		var logged_user_id = inConcert.app.getCompleteId();
		//check if the comments are unread for the current user
		var readers = this.commentReaders || [];
		if (readers.indexOf(logged_user_id) < 0) {
			return true;
		}
		return false;
	};

	Interaction.prototype.readComments = function(callback)
	{
		callback = callback || $.noop;

		var current_comments = this.getComments();
		if (!current_comments) {
			callback(true);
			return;
		}

		if (!this.youHasUnreadComments()) {
			callback(true);
			return;
		}

		var url = this.m_manager.getBaseUrl() + "/interactions/read_comments/";
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify({
				id : this.id
			}),
			dataType: "json",
			contentType: "application/json",
			success: function () {
				//200 OK
				callback(true);
			},
			error: function(undefined, error) {
				var message;
				switch (error.result) {
					case "INTERACTION_NOT_FOUND":
						message = "Conversation not found.";
						break;
					case "AGENT_NOT_FOUND":
						message = "Agent not found";
						break;
					default:
						message = "An unexpected error occurred, please try again later...";
				}
				callback(false, message);
			}
		});
	};

	Interaction.prototype.getUnsortedEventList = function() {
		this.assert_inflated();
		return (this.__unsorted_events && this.__unsorted_events.length) ? this.__unsorted_events : null;
	};

	/**
	* Sort Events by original date. (not inConcert timestamp)
	*/
	Interaction.prototype.sortEvents = function()
	{
		var self = this;
		// Nothing to be done when the interaction is being composed
		if ( this.isBeingComposed() ) {
			return;
		}

		this.assert_inflated();

		if ( !this.events ) {
			// nothing to do without events
			return;
		}

		this.__unsorted_events = [];
		this.events.forEach( function( event, ev_index )
		{
			event.__index = ev_index;
			if ( !event.original_timestamp ) {
				var date_time = self.getDateObject( ev_index );
				if ( date_time ) {
					event.original_timestamp = date_time.getTime();
				}
			}
			self.__unsorted_events.push(event);
		});

		// Invoca a la fucion que ordena los eventos.
		// Esta funcion puede estar definida en las clases derivadas (ej: interaction facebook), sino va contra la clase base
		this.__sort_by_creation_date();

		this.m_has_sorted_events = true;
	};

	/*
		Funcion que ordena los eventos por fecha de creacion
	*/
	Interaction.prototype.__sort_by_creation_date = function()
	{
		this.events.sort( function( a, b ) {
			if ( !a.original_timestamp || !b.original_timestamp ) {
				return a.__index - b.__index;
			}
			if (a.original_timestamp === b.original_timestamp) {
				return a.__index - b.__index;
			}
			return a.original_timestamp - b.original_timestamp;
		});
	};

	Interaction.prototype.canBeTaken = function() {
		return true;
	};

	Interaction.prototype.hasAppointment = function() {
		return this.interactionData && this.interactionData.hasAppointment || false;
	};

	Interaction.prototype.takeAppointment = function(callback, app_id = "interaction") {
		callback = callback || $.noop;
		if (!this.isAppointmentActive()) {
			return callback(false, "<%=_T('Appointment is not active.')%>");
		}
		let url = this.m_manager.getBaseUrl() + "/interactions/take_appointment/";
		let request = {
			proxyType		: this.type,
			campaignAtVCC	: this.campaign,
			interactionID	: this.id,
			app_id			: app_id
		};

		var getErrorMessage = function(reason)
		{
			reason = (reason || "").toUpperCase();
			switch (reason) {
				case "INVALID_REQUEST":
					return "<%=_T('There was an error, please try again later')%>";

				case "FORBIDDEN":
					return "<%=_T('Unable to create outgoing conversation. You not have permission.')%>";

				case "AGENT_UNAVAILABLE":
					return format("<%=_T('You can not take this because your current status is {0}')%>", inConcert.app.getLoggedUserStatus());

				default:
					return "<%=_T('There was an error, please try again later')%>";
			}
		};

		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				callback(result.result === "OK", result.reason);
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "take_appointment", error_message : error.result });
				let error_message = getErrorMessage(error.result);
				callback(false, error_message);
			}
		});
	};

	//takes the conversation(called for queued conversations)
	// app_id is optional, default value "interaction"
	Interaction.prototype.take = function( callback, app_id )
	{
		if (this.isAppointment()) {
			return this.takeAppointment(callback, app_id);
		}
		
		var self = this;
		callback = callback || $.noop;

		inConcert.logger.action({action_name : "take_interaction", id : this.id, type : this.type, campaign : this.campaign, action_type : "conversation"});

		var getErrorMessage = function(result)
		{
			switch ((result || "").toUpperCase())
			{
				case "FORBIDDEN":
					return "<%=_T('You not have permission to take conversation')%>";

				case "CHATSERVER_NOT_AVAILABLE":
					return "<%=_T('Chatserver not available, please try again in a few seconds...')%>";

				case "INTERACTION_CLOSED_OR_TAKEN":
					return "<%=_T('The conversation was closed or taken')%>";

				case "AGENT_UNAVAILABLE":
					if (inConcert.app.canTakeWithCurrentState()) {
						return "<%=_T('Please check you are not exceeding the maximum amount of conversation of this type in your inbox')%>";
					}
					else {
						return format("<%=_T('You can not take this because your current status is {0}')%>", inConcert.app.getLoggedUserStatus());
					}
					break;
				case "INTERACTION_CLOSED_FROM_PROXY":
					return "<%=_T('The conversation was closed from the proxy')%>";

				case "FAIL_TRANSFERRING_TO_AGENT":
					return "<%=_T('There was an error while transferring the conversation, please try again later')%>";

				case "INTERACTION_CANCELED_BEFORE_TAKING":
					return "<%=_T('The conversation was canceled')%>";

				case "CONVERSATION_DOES_NOT_EXISTS":
					return "<%=_T('The conversation does not exists')%>";

				case "BUDDY_IS_OFFLINE":
					return "<%=_T('The requested buddy is offline')%>";

				case "NO_BUDDY":
					return "<%=_T('There is no buddy')%>";

				case "FAILED_TO_CREATE_IM_ACCOUNT":
					return "<%=_T('Failed to create IM account, please try again later')%>";

				case "CONNECTION_ERROR":
					return "<%=_T('Failed to connect, please try again later')%>";

				default:
					return "<%=_T('Failed to take conversation, please try again in a few seconds...')%>";
			}
		};

		var on_taken = function( success, code_type )
		{
			var message = code_type;
			if ( !success ) {
				message = getErrorMessage(code_type);
				inConcert.logger.error({action_name : "take_interaction", app: app_id, error_type : code_type || "", error_message : message });
			}
			self.pending = false;
			callback( success, message );
		};

		if (!this.canBeTaken()) {
			on_taken(false);
			return false;
		}

		if (!inConcert.app.canTakeWithCurrentState()) {
			on_taken(false, "AGENT_UNAVAILABLE");
			return;
		}

		this.pending = true;

		inConcert.invoke({
			type: "POST",
			url: this.m_manager.getBaseUrl() + "/interactions/take_interaction/",
			contentType: "application/json",
			data: JSON.stringify({
				id : this.id,
				campaign : this.campaign,
				type: this.type,
				app_id: app_id
			}),
			dataType: "json",
			success: function () {
				//200 OK
				on_taken(true, "OK");
			},
			error: function(undefined, error) {
				on_taken(false, error.result || error.error_code);
			}
		});
		return true;
	};
	
	Interaction.prototype.take_bot_interaction = function( callback, app_id )
	{	
		var self = this;
		callback = callback || $.noop;

		inConcert.logger.action({action_name : "take_bot_interaction", id : this.id, type : this.type, campaign : this.campaign, action_type : "conversation"});

		var on_taken = function( success, reason )
		{
			if ( !success ) {
				inConcert.logger.error({action_name : "take_interaction", app: app_id, error_type : "", error_message : reason });
			}
			self.pending = false;
			callback( success, reason );
		};

		if (!this.canBeTaken()) {
			on_taken(false);
			return false;
		}

		this.pending = true;

		inConcert.invoke({
			type: "POST",
			url: this.m_manager.getBaseUrl() + "/interactions/take_bot_interaction/",
			contentType: "application/json",
			data: JSON.stringify({
				id : this.id,
				campaign : this.campaign,
				type: this.type,
				app_id: app_id
			}),
			dataType: "json",
			success: function () {
				//200 OK
				on_taken(true, "OK");
			},
			error: function(_, error) {
				on_taken(false, error.reason);
			}
		});
		return true;
	};

	//transfer the conversation to another team member
	// app_id is optional, default value "interaction"
	Interaction.prototype.transfer = function( member, onTransferred, app_id )
	{
		var self = this;

		onTransferred = onTransferred || $.noop;
		this.pending = true;
		inConcert.logger.action({action_name : "transfer_interaction", id : this.id, target : member.Id, type : this.type, campaign : this.campaign, action_type : "conversation"});

		var getErrorMessage = function(reason)
		{
			switch (reason) {
				case "INVALID_REQUEST":
				case "LABEL_PENDING_IN_USE":
				case "LABEL_PENDING_IN_INTERACTION":
				case "INTERACTION_NOT_TAKEN":
				case "INTERACTION_NOT_TAKEN_BY_AGENT":
					return "<%=_T('There was an error, please try again later')%>";
				case "COULD_NOT_RESERVE_AGENT":
					return "<%=_T('Could not reserve the agent')%>";
				case "INTERACTION_CLOSED_OR_TAKEN":
					return "<%=_T('The conversation was closed or taken')%>";
				case "CHATSERVER_NOT_AVAILABLE":
					return "<%=_T('Chatserver not available')%>";
				case "FORBIDDEN":
					return format("<%=_T('You have not permissions to transfer conversations on {0} campaign.')%>", self.campaign);
				case "MEMBER_DOES_NOT_EXIST":
					return "<%=_T('User not exists')%>";
				default:
					return "<%=_T('There was an error, please try again later')%>";
			}
		};

		var transfer_callback = function(ok, undefined /*success_interactions*/, failed_interactions, request) {
			self.pending = false;
			var errorMessage = null;
			// at least one conversation failed
			if (ok && failed_interactions && failed_interactions.length > 0){
				errorMessage = getErrorMessage(failed_interactions.pop().result);
			}
			// server error
			else if (!ok && request){
				try {
					errorMessage = getErrorMessage(JSON.parse(request).result);
				}
				catch (err) {
					errorMessage = request && getErrorMessage(request) || 
						("Error parsing data (Interaction.prototype.transfer): " + err);
				}
			}
			// everything looks good.
			else{
				onTransferred(true);
			}
			if (errorMessage){
				inConcert.logger.error( format("There is an error transferring conversation {0}: {1}", self.id, errorMessage) );
				onTransferred(false, errorMessage);
			}
		};

		this.m_manager.batchTransfer( [this], member, transfer_callback, app_id );
	};
	
	//conference the conversation to another team member
	// app_id is optional, default value "interaction"
	Interaction.prototype.conference = function( member, onConference, app_id )
	{
		var self = this;

		onConference = onConference || $.noop;
		this.pending = true;
		inConcert.logger.action({action_name : "conference_interaction", id : this.id, target : member.Id, type : this.type, campaign : this.campaign, action_type : "conversation"});

		var getErrorMessage = function(reason)
		{
			switch (reason) {
				case "INVALID_REQUEST":
				case "LABEL_PENDING_IN_USE":
				case "LABEL_PENDING_IN_INTERACTION":
				case "INTERACTION_NOT_TAKEN":
				case "INTERACTION_NOT_TAKEN_BY_AGENT":
					return "<%=_T('There was an error, please try again later')%>";
				case "COULD_NOT_RESERVE_AGENT":
					return "<%=_T('Could not reserve the agent')%>";
				case "INTERACTION_CLOSED_OR_TAKEN":
					return "<%=_T('The conversation was closed or taken')%>";
				case "CHATSERVER_NOT_AVAILABLE":
					return "<%=_T('Chatserver not available')%>";
				case "FORBIDDEN":
					return format("<%=_T('You have not permissions to conference conversations on {0} campaign.')%>", self.campaign);
				case "MEMBER_DOES_NOT_EXIST":
					return "<%=_T('User not exists')%>";
				default:
					return "<%=_T('There was an error, please try again later')%>";
			}
		};

		var conference_callback = function(ok, undefined /*success_interactions*/, failed_interactions, request) {
			self.pending = false;
			var errorMessage = null;
			// at least one conversation failed
			if (ok && failed_interactions && failed_interactions.length > 0){
				errorMessage = getErrorMessage(failed_interactions.pop().result);
			}
			// server error
			else if (!ok && request){
				try {
					errorMessage = getErrorMessage(JSON.parse(request).result);
				}
				catch (err) {
					errorMessage = request && getErrorMessage(request) || 
						("Error parsing data (Interaction.prototype.conference): " + err);
				}
			}
			// everything looks good.
			else{
				onConference(true);
			}
			if (errorMessage){
				inConcert.logger.error( format("There is an error doing conference conversation {0}: {1}", self.id, errorMessage) );
				onConference(false, errorMessage);
			}
		};

		this.m_manager.batchConference( [this], member, conference_callback, app_id );
	};
	
	Interaction.prototype.requeue = function( onRequeued, app_id )
	{
		var self = this;

		onRequeued = onRequeued || $.noop;
		this.pending = true;
		inConcert.logger.action({action_name : "requeue_interaction", id : this.id, type : this.type, campaign : this.campaign, action_type : "conversation"});

		var getErrorMessage = function(reason){
			switch (reason) {
				case "INVALID_REQUEST":
				case "LABEL_PENDING_IN_USE":
				case "LABEL_PENDING_IN_INTERACTION":
				case "INTERACTION_NOT_TAKEN":
				case "INTERACTION_NOT_TAKEN_BY_AGENT":
					return "<%=_T('There was an error, please try again later')%>";
				case "COULD_NOT_RESERVE_AGENT":
					return "<%=_T('Could not reserve the agent')%>";
				case "INTERACTION_CLOSED_OR_TAKEN":
					return "<%=_T('The conversation was closed or taken')%>";
				case "CHATSERVER_NOT_AVAILABLE":
					return "<%=_T('Chatserver not available')%>";
				case "FORBIDDEN":
					return format("<%=_T('You have not permissions to requeue conversations on {0} campaign.')%>", self.campaign);
				default:
					return "<%=_T('There was an error, please try again later')%>";
			}
		};

		var on_requeued_callback = function(ok, undefined /*success_interactions*/, failed_interactions, request){
			self.pending = false;
			var errorMessage = null;
			// at least one conversation failed
			if (ok && failed_interactions && failed_interactions.length > 0){
				errorMessage = getErrorMessage(failed_interactions.pop().result);
			}

			// server error
			else if (!ok && request) {
				try {
					errorMessage = getErrorMessage(JSON.parse(request).result);
				}
				catch (err) {
					errorMessage = "Error parsing data (Interaction.prototype.requeue): " + err;
				}
			}
			// everything looks good.
			else{
				onRequeued(true);
			}
			if (errorMessage) {
				inConcert.logger.error( format("There was an error requeueing conversation {0}: {1}", self.id, errorMessage) );
				onRequeued(false, errorMessage);
			}
		};

		// call batch requeue with one conversation
		this.m_manager.batchRequeue( [this], on_requeued_callback, app_id );
	};

	Interaction.prototype.updateLabels = function(addedLabels, currentLabels, callback, app_id) {
		let interactions = [this];
		this.m_manager.batchUpdateLabels(interactions, addedLabels, currentLabels, function(status, _ /*success_interactions*/, failed_interactions) {
			if (status && (!failed_interactions || !failed_interactions.length)) {
				callback(true);
			}
			else {
				callback(false);
			}
		}, app_id);
	};

	Interaction.prototype.deleteLabels = function(labels, callback) {
		var self = this;
		inConcert.logger.action({action_name : "delete_label_interaction", id : self.id, type : self.type, campaign : self.campaign, action_type : "conversation"});
		inConcert.invoke({
			type: "POST",
			url: this.m_manager.getBaseUrl() + "/interactions/delete_labels_interaction/",
			data: JSON.stringify({
				interactionId: self.id,
				deleteLabels: labels
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(label) {
				self.pending = false;
				if (typeof(callback) === "function") {
					callback(label);
				}
			},
			error: function( /* response, error */ ) {
				self.pending = false;
				if (typeof(callback) === "function") {
					callback(false);
				}
			}
		});
	};

	Interaction.prototype.getLabels = function() {
		if (typeof(this.labels) === "string") {
			return JSON.parse(this.labels);
		}
		return this.labels;
	};

	Interaction.prototype.labeledWith = function(label) {
		if (!label || !this.labels || !this.labels[label]) {
			return false;
		}
		return true;
	};

	Interaction.prototype.hasInteractionParent = function() {
		var parent_it_id = this.parentInteractionId;
		if ( typeof parent_it_id !== "string" ) {
			return false;
		}
		return true;
	};

	Interaction.prototype.hasContactFormInteractionParent = function( callback ) {
		callback = callback || $.noop;
		this.getParentInteraction( function( parent_interaction ) {
			if ( !parent_interaction ) {
				return callback( false );
			}
			return callback( parent_interaction.type.toUpperCase() === "WEBCONTACT" );
		});
	};

	Interaction.prototype.getParentInteraction = function( callback ) {
		callback = callback || $.noop;
		if ( !this.hasInteractionParent() ) {
			callback( false );
			return;
		}
		var interaction_manager = this.m_manager;
		if ( !interaction_manager ) {
			callback( false );
			return;
		}
		var parent_it_id = this.parentInteractionId;
		var parent_interaction = interaction_manager.findInteraction( parent_it_id );
		if (parent_interaction) {
			setTimeout(function(){
				callback( parent_interaction );
			});
			return;
		}
		//
		const inflate = false;
		interaction_manager.getInteraction( parent_it_id, function( parent_interaction ) {
			if ( parent_interaction ) {
				callback( parent_interaction );
				return;
			}
			//
			interaction_manager.getFromArchive( parent_it_id, function( parent_interaction ) {
				if ( parent_interaction ) {
					callback( parent_interaction );
					return;
				}
				callback( false );
			}, inflate);
		}, inflate);
	};

	Interaction.prototype.getSiblingInteractions = function( callback ) {
		var self = this;
		callback = callback || $.noop;
		if ( !this.belongsToSomeGroup() ) {
			return callback ( false );
		}

		this.getParentInteraction( function( parent_interaction )
		{
			if ( !parent_interaction ) {
				return callback( false );
			}
			if ( !parent_interaction.interactionsGroup ) {
				return callback( false );
			}
			var current_it_id = this.id;
			var parent_it_id = parent_interaction.id;
			var sibling_interactions = {};

			// recursive search
			var search = function ( interactionsGroup )
			{
				$.each( interactionsGroup || {}, function( group_it_id, child_its_id)
				{
					if ( parent_it_id === group_it_id ) {
						$.each( child_its_id || {}, function( child_it_id )
						{
							if ( child_it_id === current_it_id ) {
							// la current conversation no va.
								return true; // continue
							}
							var sibling_it = self.m_manager.findInteraction( child_it_id );
							if ( sibling_it ) {
								sibling_interactions[ sibling_it.id ] = sibling_it;
							}
						});
					}
					else {
						search( child_its_id ) ; // find in child
					}
				});
			};

			search( parent_interaction.interactionsGroup );

			return callback( sibling_interactions );
		});
	};

	Interaction.prototype.getChildInteractions = function()
	{
		var self = this;
		if ( !this.belongsToSomeGroup() ) {
			return false;
		}

		var current_it_id = this.id;
		var child_interactions = {};

		// recursive search
		var search = function ( interactionsGroup )
		{
			$.each( interactionsGroup || {}, function( group_it_id, child_its_id)
			{
				if ( current_it_id === group_it_id ) {
					$.each( child_its_id || {}, function( child_it_id )
					{
						var child_it = self.m_manager.findInteraction( child_it_id );
						if ( child_it ) {
							child_interactions[ child_it.id ] = child_it;
						}
					});
				}
				else {
					search( child_its_id ) ; // find in child
				}
			});
		};

		search( this.interactionsGroup );

		return child_interactions;
	};

	Interaction.prototype.belongsToSomeGroup = function()
	{
		if ( !this.groupId ) {
			return false;
		}
		return true;
	};

	Interaction.prototype.getGroupId = function()
	{
		return this.groupId;
	};

	Interaction.prototype.setGroupId = function( groupId )
	{
		this.groupId = groupId;
	};

	Interaction.prototype.isSpam = function() {
		return this.state === "SPAM";
	};

	Interaction.prototype.isOutgoing = function() {
		return this.initiative === "OUTBOUND";
	};
	
	Interaction.prototype.isCreatedByAgent = function() {
		return this.created_by_agent || false;
	};

	Interaction.prototype.hasPendingLabel = function() {
		return this.labeledWith("PENDING");
	};

	Interaction.prototype.matchPending = function( /* interaction */ ) {
		return false;
	};

	Interaction.prototype.createOwnId = function() {
		this.id = inConcert.Utils.generateGuid();
	};

	//set conversation view
	Interaction.prototype.setView = function(view) {
		// TODO: #9778 - CANY - Docked.
		// if (this.m_view && view) {
		// 	return;
		// }
		this.m_view = view;
	};

	Interaction.prototype.getHumanReadableAddress = function() {
		return this.contactAddress;
	};

	//set manager handler
	Interaction.prototype.setManager = function(interactionManager) {
		this.m_manager = interactionManager;
	};

	//retrieves view associated with the conversation in case it exists
	Interaction.prototype.getView = function() {
		return this.m_view;
	};

	Interaction.prototype.getAddress = function() {
		return this.contactAddress || "";
	};

	Interaction.prototype.getDisplayAddress = function() {
		const visualize_address = inConcert.app.canVisualizeAddress(null, this.campaign, null, true);
		let displayAddress = this.getAddress();
		if(displayAddress !== "" && !visualize_address){
			return "<%=_T('Hidden Address')%>";
		}
		return displayAddress;
	};

	Interaction.prototype.getDisplayContactAddress = function() {
		return this.getDisplayAddress();
	};

	Interaction.prototype.hasReportingData = function()
	{
		var it_data = this.data;
		if ( !it_data || !it_data.reportingData ) {
			return false;
		}
		return true;
	};
	
	Interaction.prototype.__getReportDefaultField = function()
	{
		return "api";
	};

	Interaction.prototype.getReportFields = function(fields_type, json = false)
	{
		fields_type = fields_type || this.__getReportDefaultField();
		let fields = this.data && this.data.reportingData && 
				this.data.reportingData[fields_type] && this.data.reportingData[fields_type].fields || false;
		if (json) {
			fields = (fields || []).reduce(function(result, field) {
				result[field.name] = field.value;
				return result;
			}, {});
		}
		return fields;
	};
	
	Interaction.prototype.getReportEntry = function(fields_type, key, case_insensitive) {
		key = case_insensitive? key.toLowerCase() : key;
		return (this.getReportFields(fields_type) || []).find(function(field) {
			return (case_insensitive? field.name.toLowerCase() : field.name) === key;
		}) || null;
	};
	
	Interaction.prototype.getReportValue = function(fields_type, key, case_insensitive) {
		let entry = this.getReportEntry(fields_type, key, case_insensitive);
		return entry? entry.value : null;
	};

	/**
	* Returns an array with the addresses that have been collected by the proxy. This includes addresses entered
	* in a survey, login form, etc in the case of webcontact and webchat. The other type of interactions will have an array
	* with just one address, the contactAddress.
	* For compatibility with old contact info, we normalize the values and introduce a 'displayable' field if it is not
	* present.
	*/
	Interaction.prototype.getExternalContactAddresses = function()
	{
		var it_data = this.interactionData;
		if ( !it_data || !it_data.contactAddresses ) {
			return false;
		}
		if ($.isArray(it_data.contactAddresses)) {
			return inConcert.Utils.cloneArray( it_data.contactAddresses );
		}
		return inConcert.Utils.cloneObject( it_data.contactAddresses );
	};

	Interaction.prototype.loadContactAddresses = function( callback )
	{
		callback = callback || $.noop;

		var contact_addresses = this.getExternalContactAddresses() || [];

		if ( !this.contact ) {
			callback( contact_addresses );
			return;
		}

		inConcert.app.getContactManager().getContact( this.contact, function( contact ) {
			if ( contact ) {

				var hash_addresses = {};
				$.each( contact_addresses, function( undefined, address ) {
					hash_addresses[ ( address.displayable || "" ).toUpperCase() ] = true;
				});

				// merge contact addresses
				$.each( contact.addresses || [], function( undefined, contact_address ) {
					var address = {
						type    : ( contact_address.Type || "" ).toUpperCase()     ,
						address : contact.getAddressDisplayText( contact_address )
					};
					if ( typeof address.address !== "string" || address.address.length === 0 ) {
						return true; // continue
					}
					if ( hash_addresses[ address.address.toUpperCase() ] ) {
						return true; // continue
					}
					contact_addresses.push( address );
				});
			}
			callback( contact_addresses );
		});

	};

	// renombrar. no es appointment, es webcontact
	Interaction.prototype.addRelationshipPendingAppointment = function( appointment /* webcontact */ ) {
		// client side pending relationship, no deberia persistir en el server
		if ( typeof appointment !== "object" ) {
			return;
		}
		this.m_pending_relationship_appointment = {
			id: appointment.id
		};
	};

	// renombrar. no es appointment, es webcontact
	Interaction.prototype.removeRelationshipPendingAppointment = function() {
		// client side pending relationship, no deberia persistir en el server
		if ( !this.m_pending_relationship_appointment ) {
			return;
		}
		this.emit( "cancel_outgoing_interaction" );
		var interaction_id = this.m_pending_relationship_appointment.id;
		this.m_pending_relationship_appointment = null;
		var im = inConcert.app.getInteractionManager();
		var appointment = im.findInteraction( interaction_id );
		if ( appointment ) {
			appointment.removeRelationshipPendingInteraction();
		}
	};

	Interaction.prototype.getAvatarURI = function() {
		if (this.isLocal()) {
			return "/users/get_avatar/" + this.teamMember;
		}
		else {
			if (this.contact && this.contact.length > 0) {
				return "/contacts/get_contact_avatar/" + this.contact;
			}
			else {
				if (this.contactAddress && this.contactAddress.length > 0) {
					return "/contacts/get_address_avatar/" + this.type + "/" + this.getAddress();
				}
				else {
					if (this.campaign) {
						return "/campaigns/get_avatar/" + this.campaign;
					}
				}
			}
		}
		return "/contacts/get_generic_avatar";
	};

	Interaction.prototype.isReserved = function()
	{
		return this.state === "RESERVED";
	};

	Interaction.prototype.isArchived = function()
	{
		return this.state === "FINISHED" || this.state === "CANCELED";
	};

	//returns true if the conversation is done
	Interaction.prototype.isFinished = function()
	{
		return (this.isArchived() || this.isSpam() || this.isDiscarded());
	};

	//returns true if the conversation is LOCKED.
	Interaction.prototype.isLocked = function()
	{
		return this.state === "LOCKED" || this.state === "TRANSFER LOCKED" || this.state === "CONFERENCE LOCKED";
	};

	//returns true if the conversation is discarded
	Interaction.prototype.isDiscarded = function()
	{
		return this.state === "DISCARD";
	};

	//returns true if the conversation is done
	Interaction.prototype.isAbandoned = function()
	{
		return this.state === "ABANDONED";
	};

	//returns true if the conversation is transferring
	Interaction.prototype.isTransferring = function()
	{
		return this.state === "TRANSFERRING";
	};

	//returns true if the conversation is queued
	Interaction.prototype.isQueued = function()
	{
		return this.state === "QUEUED" || this.state === "TRANSFER QUEUED";
	};

	Interaction.prototype.isOpen = function()
	{
		return this.state === "OPEN";
	};

	Interaction.prototype.isAppointment = function()
	{
		return this.isOpen() && this.interactionData && this.interactionData.isAppointment;
	};

	Interaction.prototype.isBotInternalConsultation = function()
	{
		if (this.state !== "CREATED")
			return false;
		if (!this.interactionData)
			return false;
		if (this.interactionData.internal_consultation_flag)
			return true;
		if (!this.interactionData.isSentToCC && this.type !== "INTERNAL_CHAT")
			return true;
		return false;
	};

	Interaction.prototype.isAppointmentActive = function()
	{
		if (!this.isAppointment()) {
			return false;
		}
		var time = inConcert.Utils.getTimeFromString(this.getAppointmentDate());
		return time <= new Date();
	};
	
	Interaction.prototype.getAppointmentDate = function()
	{
		return (this.isAllegro() && this.startTimestampReadable) || this.interactionData && this.interactionData.fromDate || 
				this.lastEventTimestampStr || this.timestamp;
	};

	Interaction.prototype.isRedial = function()
	{
		return this.isTaken() && this.interactionData && this.interactionData.redialProcess;
	};

	Interaction.prototype.isWaiting = function()
	{
		return this.state === "WAITING";
	};

	Interaction.prototype.isWrappedUp = function()
	{
		return this.state === "WRAPUP";
	};

	Interaction.prototype.isTaken = function()
	{
		return this.state === "TAKEN";
	};

	Interaction.prototype.isBot = function()
	{
		return this.state === "BOT";
	};

	Interaction.prototype.isTransferTaken = function()
	{
		return this.state === "TRANSFER TAKEN";
	};
	
	Interaction.prototype.isAssigned = function()
	{
		return this.isTaken() || this.isTransferTaken();
	};

	Interaction.prototype.isUpdateTransferBot = function()
	{
		let beforeStep = this.getBeforeLastStep();
		return ((this.isWrappedUp() && beforeStep && (beforeStep.action === "TRANSFERRING" || beforeStep.action === "BOT")) ||
				(this.isTaken() && beforeStep && beforeStep.action === "BOT"));
	};

	//returns the extra information
	Interaction.prototype.getExtendedInfo = function() { return ""; };

	Interaction.prototype.hasAttachments = function() {
		return (this.properties && this.properties.has_attachments) || false;
	};

	Interaction.prototype.isValid = function() {
		return true;
	};

	Interaction.prototype.getEventList = function() {
		this.assert_inflated();
		return (this.events && this.events.length) ? this.events : null;
	};

	Interaction.prototype.getComments = function() {
		return (this.comments && this.comments.length) ? this.comments : null;
	};

	/**
	event_count is computed on the server. It does not include automatic responses.
	*/
	Interaction.prototype.getEventCount = function() {
		return this.properties && this.properties.event_count || 0;
	};

	Interaction.prototype.get_sorted_index_by_original_index = function( original_index )
	{
		this.assert_inflated();

		var sorted_index = false;
		$.each( this.events, function( current_index, event )
		{
			if ( event.__index === original_index ) {
				sorted_index = current_index;
				return false; // break
			}
		});
		return sorted_index;
	};

	Interaction.prototype.get_event_by_original_index = function( original_index )
	{
		var sorted_index = this.get_sorted_index_by_original_index( original_index );
		if (!sorted_index && sorted_index !== 0) {
			return;
		}
		return this.getEvent(sorted_index);
	};

	Interaction.prototype.getEvent = function( sorted_index )
	{
		if ( this.isBeingComposed() || sorted_index === - 1 ) {
			return null;
		}
		this.assert_inflated();
		return ( this.events && this.events.length > sorted_index ) ? this.events[ sorted_index ] : null;
	};

	Interaction.prototype.getRelatedEventIndex = $.noop;
	Interaction.prototype.getNextRelatedEventIndex = $.noop;
	Interaction.prototype.getPreviousRelatedEventIndex = $.noop;

	Interaction.prototype.getUnreadEventCount = function()
	{
		if( typeof this.properties.unread_event_count === "undefined" ) {
			// Track these errors
			console.error(this.id);
		}
		return this.properties.unread_event_count;
	};

	Interaction.prototype.getUnAnsweredEventCount = function()
	{
		if (typeof this.properties.unanswered_event_count === "undefined") {
			return 0;
		}
		return this.properties.unanswered_event_count;
	};

	/**
	* Returns an array containing the unread events.
	*/
	Interaction.prototype.getUnreadEvents = function()
	{
		this.assert_inflated();

		var self = this;
		var unread_events = [];
		if ( !$.isArray( this.events ) ) {
			return unread_events;
		}

		$.each( this.events, function( _, event )
		{
			$.each( self.readers_v2 || {}, function( reader_id, mark )
			{
				if ( reader_id === "I" + event.__index && mark === "U" ) {
					unread_events.push( event );
					return false;
				}
			});
		});

		return unread_events;
	};

	/**
	* 
	*/
	Interaction.prototype.getFirstUnreadEventIndex = function() {
		this.assert_inflated();

		for (var i = 0; i < this.events.length; i++) {
			var ev = this.events[i];
			var ev_index = ev.__index || i;
			if ( !this.isRead( ev_index ) ) {
				return i;
			}
		}
		return null;
	};

	Interaction.prototype.getFirstIncommingEvent = function() {
		if ( this.isBeingComposed() ) {
			return;
		}
		this.assert_inflated();

		for (var i = 0; i < this.events.length; i++) {
			var ev = this.events[i];
			var ev_index = ev.__index || i;
			if ( this.isContactEventIndex( ev_index ) ) {
				return this.events[ev_index];
			}
		}
		return;
	};

	Interaction.prototype.getFirstOutgoingEvent = function() {
		if ( this.isBeingComposed() ) {
			return;
		}
		this.assert_inflated();

		for (var i = 0; i < this.events.length; i++) {
			var ev = this.events[i];
			var ev_index = ev.__index || i;
			if ( this.isAgentEvent( ev_index ) ) {
				return this.events[ev_index];
			}
		}
		return;
	};

	Interaction.prototype.getRawAddress = function() {
		return this.contactAddress;
	};

	Interaction.prototype.getSource = function() {
		return this.contactAddress;
	};

	Interaction.prototype.hasSortedEvents = function() {
		return this.m_has_sorted_events;
	};

	Interaction.prototype.getFirstIndexEvent = function() {
		return 0;
	};

	Interaction.prototype.getDateObject = function() {
		return this.convertDateStringToDate(this.timestamp);
	};

	Interaction.prototype.getLocaleDateString = function(index)
	{
		index = index || 0;
		var date_time = this.getDateObject(index);
		if (!date_time) {
			return "";
		}
		return inConcert.Utils.userFormatDate(date_time);
	};

	Interaction.prototype.getAccount = function() {
		//TODO this is for backward copmatibility with previous conversations, should be gone soon
		var account = this.account;
		if (!account) {
			account = this.interactionData.account;
			if (!account) {
				account = "";
			}
		}
		return account;
	};

	Interaction.prototype.get_language = function()
	{
		var default_language = "es";
		if ( !this.interactionData ) {
			return default_language;
		}
		var detected_language = this.interactionData.detectedLanguage;
		if ( detected_language ) {
			// 1 retorno el lenguaje que se ha detectado en el IR
			return detected_language;
		}
		// si no tiene un lenguaje detectado, tengo que obtener y retornar el valor configurado en la cuenta
		var campaign = inConcert.app.getCampaign(this.campaign);
		if ( !campaign ) {
			// 2 si la campana ha sido eliminada retorno el x default
			return default_language;
		}
		var account_id = this.account;
		var account_language = campaign.get_account_language(account_id);
		if ( account_language ) {
			// 3 retorno el lenguaje configurado en la cuenta
			return account_language;
		}
		// si no existe configuracion en la cuenta retorno el lenguaje por defecto (es)
		return default_language;
	};

	Interaction.prototype.setAccount = function(accountId) {
		this.account = accountId;
	};

	Interaction.prototype.getAccountDisplayName = function() {
		let interaction = this;
		let campaign = inConcert.app.getCampaign(interaction.campaign);
		// campaigns can be removed with alive interactions
		if (!campaign) {
			return interaction.account;
		}
		// accounts can be removed with alive interactions
		let account = _.findWhere(campaign.accounts, { "id" : interaction.account });
		if (account){
			return account && account.name || interaction.account;
		}
		return interaction.account;
	};

	Interaction.prototype.getAttentionLevel = function() { return this.attention_level; };

	Interaction.prototype.hasAttentionLevel = function() { return typeof this.attention_level === "object"; };

	Interaction.prototype.updateAttentionLevel = function( new_attention_level )
	{
		if ( !new_attention_level ) {
			delete this.attention_level;
		}
		else {
			this.attention_level = new_attention_level;
		}
	};
	
	Interaction.prototype.getAttentionLevelEntity = function()
	{
		let al = this.getAttentionLevel();
		if (!al) {
			return null;
		}
		let campaign = inConcert.app.getCampaign( this.campaign );
		return campaign && campaign.getAttentionLevel(al.id) || null;
	};
	
	Interaction.prototype.getAttentionLevelDisplayName = function()
	{
		let al = this.getAttentionLevelEntity();
		return al && al.name || "";
	};

	Interaction.prototype.getAvailableAttentionLevels = function( account_id /* optional */)
	{
		account_id = account_id || this.account;
		if ( !this.isBeingComposed() ) {
			let al = this.getAttentionLevelEntity();
			return al? [al] : [];
		}
		else {
			let campaign = inConcert.app.getCampaign( this.campaign );
			let logged_user_id = inConcert.app.getLoggedUserId();
			return campaign && campaign.getUserAttentionLevelsInAccount(account_id, logged_user_id) || [];
		}
	};

	//Returns true if the specified index is a message sent from an internal agent
	Interaction.prototype.isInternalMessage = function(index)
	{
		this.assert_inflated();
		if (typeof(index) !== "undefined" && index >= 0) {
			return this.events &&
				this.events[index] &&
				this.events[index].agent;
		}
		else {
			return this.interactionData.agent;
		}
	};

	//returns true if the conversation is with a team member
	Interaction.prototype.isLocal = function() {
		return this.localInteraction;
	};

	// returns transferInfo or false
	Interaction.prototype.check_is_transfer = function()
	{
		// Este atributo llega desde la "interactionmanip", es un atributo calculado.
		if ( !this.transferInfo ) {
			return false;
		}
		return this.transferInfo;
	};

	Interaction.prototype.getBeforeLastStep = function()
	{
		return ($.isArray(this.steps) && this.steps.length > 2)? this.steps[ this.steps.length - 2 ]: null;
	};

	Interaction.prototype.getLastStep = function()
	{
		var last_step;
		if ($.isArray(this.steps) && this.steps.length > 0) {
			last_step = this.steps[ this.steps.length -1 ];
		}
		if (!last_step) {
		// Si el evento aun, no existe (para el caso en que la interaccion aun no se ha inflado)
		//   al valor lo obtiene del campo calculado "__last_step". Al mismo lo calcula la "lua_interaction_manip"
			last_step = this.__last_step;
		}
		return last_step;
	};

	/**
	 * returns a preview for the conversation to be displayed in notifications and the like
	 * is virtual should be implemented downward
	 *
	 * @return {String}
	 */
	Interaction.prototype.getContentPreview = function()
	{
		var preview = "";

		if (this.isLocked()) {
			var step = this.getLastStep();
			if (step) {
				var userId = step.from.match(new RegExp("([^@]+)@(.*)"));
				preview = format("<%=_T('Transferred by {0}')%>", userId ? userId[1] : step.from);
			}
		}

		return preview;
	};

	/**
	* Returns a displayable name which is computed according to whether the conversation is local or not, it has a
	* contactName or a contact associated.
	*
	* @return {String}
	*/
	Interaction.prototype.getDisplayName = function() {
		if (this.isLocal()) {
			if (this.teamMemberName && this.teamMemberName.length > 0) {
				return this.teamMemberName;
			}
			else {
				return (this.teamMember && this.teamMember.length > 0) ? this.teamMember : (this.teamMemberAddress || "");
			}
		}
		else {
			if (this.interactionData && this.interactionData.contactName && this.interactionData.contactName.length > 0) {
				return this.interactionData.contactName;
			}
			else {
				if (this.initiative === "OUTBOUND" && (!this.contact || this.contact === "") && (!this.contactAddress || this.contactAddress === "")) {
					var campaign = inConcert.app.getCampaign(this.campaign);
					if (campaign) {
						return campaign.getDescription();
					}
				}
				else {
					//TODO should do some contacto lookup?
					return (this.contact && this.contact.length > 0) ? this.getHumanReadableAddress() : (this.getHumanReadableAddress() || "");
				}
			}
		}
		return "";
	};

	Interaction.prototype.getCampaignDisplayName = function() {
		if (!this.campaign) {
			return "";
		}
		let campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return inConcert.app.get_id(this.campaign);
		}
		return campaign.getDescription();
	};

	/**
	* Returns the displayable client name (i.e. Juan Perez). Will try to use first 'contactName' which is present
	* in most conversations, and then display_name which is present in internal chats.
	*
	* @return {String}
	*/
	Interaction.prototype.getClientName = function()
	{
		if ( this.isBeingComposed() ) {
			return "";
		}
		if ( this.interactionData ) {
			if ( this.interactionData.contactName ) {
				return this.interactionData.contactName;
			}
			// TODO: Make internal chats use contactName too ?
			if ( this.interactionData.data && this.interactionData.data.display_name ) {
				return this.interactionData.data.display_name;
			}
		}
		if (this.contactObject && this.contactObject.Name) {
			return this.contactObject.Name;
		}
		var first_incomming_event = this.getFirstIncommingEvent();
		if ( first_incomming_event ) {
			var contact_name = this.getUserName( first_incomming_event );
			if ( contact_name ) {
				return contact_name;
			}
		}
		return "";
	};

	Interaction.prototype.getUserName = function( /*event*/ )
	{
		// implementar en clase base
		return "";
	};
	
	Interaction.prototype.getBotName = function()
	{
		return "<%=_T('Bot')%>";
	};

	/**
	* Returns the gathered contact name (data collected by the proxy)
	* This name will not change ever, so if you need an updated name, this is
	* not the function you are looking for
	*
	* @return {String}
	*/
	Interaction.prototype.getGatheredClientName = function() {
		if ( !this.interactionData || !this.interactionData.gatheredContactData ) {
			return "";
		}
		return this.interactionData.gatheredContactData.name;
	};

	/**
	* Returns the external id that may have been supplied when the interaction was created.
	* Note: If you need the current value, this is *not* the method you're looking for. You need to
	* retrieve the value from the contact itself.
	*
	* @return {String}
	*/
	Interaction.prototype.getClientExternalId = function() {
		if (this.interactionData){
			if ( this.interactionData.data && this.interactionData.data.contact && this.interactionData.data.contact.externalId ) {
				return this.interactionData.data.contact.externalId;
			}
			if (this.interactionData.gatheredContactData && this.interactionData.gatheredContactData.external_id){
				return this.interactionData.gatheredContactData.external_id;
			}
		}
		return "";
	};

	//Returns conversation formatted time on the current state, 00:00 s 2 days, etc
	Interaction.prototype.currentTime = function() {
		var strDate = this.timestamp;
		if (strDate) {
			return inConcert.Utils.getTimeDifference(inConcert.Utils.getTimeFromString(strDate));
		}
		else {
			return "00:00";
		}
	};

	Interaction.prototype.creationDate = function(creation_date)
	{
		creation_date = creation_date || this.startTimestampReadable || this.timestamp;
		return inConcert.Utils.userFormatDate(creation_date, true);
	};

	Interaction.prototype.lastStepTimestamp = function()
	{
		var step = this.getLastStep();
		if (!step) {
			return;
		}
		var timestamp = inConcert.Utils.getTimeFromString(step.timestamp);
		return this.__formatDate(timestamp);
	};

	/**
	* Return the last event time as unix timestamp: 1339636656180
	*/
	Interaction.prototype.lastEventTimestamp = function()
	{
		var str_time = this.lastEventTimestampStr || this.timestamp;
		var time = inConcert.Utils.getTimeFromString(str_time);
		return time;
	};

	Interaction.prototype.__formatDate = function( timestamp )
	{
		// if we are dealing with a live conversation and is not finished
		if (this.isLive() && !this.isFinished() && !this.isAbandoned() && !this.isAppointment()) {
			if (timestamp){
				return inConcert.Utils.getTimeDifference(timestamp);
			}
			return "00:00";
		}
		return inConcert.Utils.userFormatDate(timestamp, true);
	};

	Interaction.prototype.timeSinceLastEvent = function() {
		var timestamp = this.lastEventTimestamp();
		return this.__formatDate( timestamp );
	};

	Interaction.prototype.timeTransferOriginInteraction = function()
	{
		var transfer_info = this.check_is_transfer();
		if ( !transfer_info ) {
			return false;
		}
		if ( !transfer_info.origin_data || !transfer_info.origin_data.startTimestamp ) {
			return false;
		}
		var origin_interaction_timestamp = transfer_info.origin_data.startTimestamp * 1000;
		return origin_interaction_timestamp;
	};

	Interaction.prototype.formattedTimeTransferOriginInteraction = function()
	{
		var origin_interaction_timestamp = this.timeTransferOriginInteraction();
		if ( !origin_interaction_timestamp ) {
			return this.timeSinceLastEvent();
		}
		return this.__formatDate( origin_interaction_timestamp );
	};

	/*
		This round trip is in order to translate the date-string to the locale of the logged user
	*/
	Interaction.prototype.convertDateStringToDate = function(eventDateString)
	{
		let dateTime = new Date(eventDateString);
		if (isNaN(dateTime)) {
			let strDate = inConcert.Utils.parseDateTimeString(eventDateString);
			dateTime = new Date(strDate);
			if (isNaN(dateTime)) {
				//unable to parse it, return it as-is
				return eventDateString;
			}
		}
		return dateTime;
	};

	Interaction.prototype.getLastDispositionFromSteps = function() {
		if (!this.steps) {
			return;
		}
		for(var si = this.steps.length - 1; si >= 0; si--) {
			var step = this.steps[si];
			if (step.wrapupData){
				return step.wrapupData;
			}
		}
		return;
	};

	//Returns last modified date formatted as string
	Interaction.prototype.lastModified = function() {
		return this.timestamp;
	};

	/*
		Retorna el current time, con esta funcion se ordena las ineracciones en el inbox (active/live)
	*/
	Interaction.prototype.currentTimeRaw = function()
	{
		var time;
		var transfer_info = this.check_is_transfer();
		if ( !transfer_info ) {
			// 1 - Si no es una interaccion transferida, retorna la fecha del ultimo evento (step)
			time = this.lastEventTimestamp();
		}
		else {
			var origin_interaction_timestamp = this.timeTransferOriginInteraction();
			if ( origin_interaction_timestamp ) {
				// 2 Si es transferencia busca la fecha del timestamp de la interaccion original
				time = new Date( origin_interaction_timestamp ).getTime();
			}
			else {
				// 3 - Si es transferencia, pero la misma no tiene el timestamp de la interaccion original, retorna lo mismo que en el punto (1)
				//       la fecha del ultimo evento.
				time = this.lastEventTimestamp();
			}
		}
		return time;
	};

	//Returns true if the conversation should be considered active or live, to be implemented in derived classes
	Interaction.prototype.isLive = function() {
		return false;
	};

	//Returns true or false if the conversation is an internal one (like internal chat conversation)
	Interaction.prototype.isInternal = function() {
		return false;
	};

	Interaction.prototype.hasDraft = function() {
		return false;
	};

	Interaction.prototype.cancel = function() {
		this.state = "CANCELED";
		inConcert.logger.action({action_name : "cancel_interaction", id : this.id, type : this.type, campaign : this.campaign, action_type : "conversation"});
		if (this.m_session && this.m_session.close) {
			console.info("[Interaction.prototype.cancel] 1-> close session of interaction: " + this.id);
			this.m_session.close();
		}
		else {
			console.info("[Interaction.prototype.cancel] 2-> NO close session of interaction: " + this.id);
		}
		//this.dispose();
	};

	Interaction.prototype.iAmRequired = function() {
		var self = this;
		if (typeof(self.requiredAgent) === "string" && self.requiredAgent === inConcert.app.getCompleteId()) {
			return true;
		}
		return false;
	};

	Interaction.prototype.isMonitorable = function()
	{
		if ( this.isTaken() && !this.belongsToMe() ) {
			return true;
		}
		return false;
	};

	Interaction.prototype.belongsTo = function(member_id)
	{
		var isMember = function(user) {
			if (user && typeof user === "string" && user.match(inConcert.app.idAtVCCRegExp)) {
				var user_id = RegExp.$1;
				if (user_id === member_id) {
					return true;
				}
			}
			return false;
		};
		if (this.assignedAgents) {
			for (var i = 0; i <  this.assignedAgents.length; i++) {
				if (isMember(this.assignedAgents[i])) {
					return true;
				}
			}
		}
		if (isMember(this.requiredAgent)) {
			return true;
		}
		if (this.reservedAgents) {
			for (var j = 0; j <  this.reservedAgents.length; j++) {
				if (isMember(this.reservedAgents[j])) {
					return true;
				}
			}
		}
		return false;
	};

	Interaction.prototype.belongsToAny = function()
	{
		if (this.assignedAgents && this.assignedAgents.length) {
			return true;
		}
		if (this.requiredAgent) {
			return true;
		}
		if (this.reservedAgents && this.reservedAgents.length) {
			return true;
		}
		return false;
	};

	Interaction.prototype.belongsToMe = function() {
		var loggedUserId = inConcert.app.getLoggedUserId();
		return this.belongsTo(loggedUserId);
	};

	Interaction.prototype.archivedBelongsToMe = function()
	{
		if (!this.isArchived()) {
			return false;
		}
		var logged_user_id = inConcert.app.getLoggedUserId().toLowerCase();
		var last_agent = (this.getLastAgentEndsConversation() || "").toLowerCase();
		return logged_user_id === last_agent;
	};

	Interaction.prototype.getConversationOwner = function()
	{
		if (this.assignedAgents && this.assignedAgents.length && this.state === "CREATED") {
			return this.assignedAgents[0];
		}
		if ( this.isTaken() || this.isWrappedUp() ) {
			return this.getFirstAssignedAgent();
		}
		if ( this.isFinished() ) {
			return this.getLastAgentEndsConversation();
		}
		if (this.requiredAgent) {
			return this.requiredAgent;
		}
		if (this.reservedAgents && this.reservedAgents.length) {
			return this.reservedAgents[0];
		}
		return;
	};

	Interaction.prototype.getFirstAssignedAgent = function()
	{
		if (!this.assignedAgents || !this.assignedAgents.length || this.scheduledDate) {
			return;
		}
		var agent_with_vcc = this.assignedAgents[0];
		// just get the username ant not user + vcc
		if (agent_with_vcc.match(inConcert.app.idAtVCCRegExp)) {
			return RegExp.$1;
		}
	};

	Interaction.prototype.setTotalNewEvents = function(data)
	{
		if (!this.supportsEvents()) {
			return;
		}
		if (!data || !data.properties) {
			return;
		}
		var current_events_count = this.getEventCount();
		this.__total_new_events = data.properties.event_count - current_events_count;
	};

	Interaction.prototype.getTotalNewEvents = function()
	{
		if (!this.supportsEvents()) {
			return 0;
		}
		return this.__total_new_events || 0;
	};

	Interaction.prototype.getNewContactEvents = function()
	{
		this.assert_inflated();

		var all_events = this.getUnsortedEventList() || [];
		var total_all_events = all_events.length;
		if (total_all_events === 0) {
			return [];
		}
		var total_new_events = this.getTotalNewEvents();
		if (total_new_events === 0) {
			return [];
		}
		var total_old_events = total_all_events - total_new_events;
		var new_contact_events = [];

		for (var event_index = (total_all_events - 1); event_index > (total_old_events - 1); event_index--) {
			var evt = all_events[event_index];
			var is_contact_event = this.isContactEvent(evt);
			if (is_contact_event) {
				new_contact_events.unshift(evt);
			}
		}
		return new_contact_events;
	};

	Interaction.prototype.getEventShortDescription = function(/*event_index*/) {
		return "TODO: implement!! SHORT DESCRIPTION";
	};

	// Updates the conversation with the new server data
	Interaction.prototype.update = function(data)
	{
		let self = this;
		this.manualState = false;
		
		let previous_it_state = this.state;
		if ( previous_it_state === "PENDING" ) {
			this.removeRelationshipPendingAppointment();
		}

		let hasScreenRecording = this.hasScreenRecording();

		if(this.type === "CALL")
		{
			const transferrer = data?.interactionData?.asyncTaken?.transfer?.transferrer || "";

			if((data.state === "TAKEN" && (previous_it_state !== "TRANSFER TAKEN" || (previous_it_state === "TRANSFER TAKEN" && transferrer !== inConcert.app.getCompleteId())) && data.assignedAgents && data.assignedAgents.length === 1 && data.assignedAgents[0] === inConcert.app.getCompleteId()) ||
				(data.state === "CONFERENCE" && (previous_it_state !== "CONFERENCE TAKEN" || (previous_it_state === "CONFERENCE TAKEN" && transferrer !== inConcert.app.getCompleteId())) && data.assignedAgents && data.assignedAgents.length > 0 && data.assignedAgents.includes(inConcert.app.getCompleteId())))
			{
				var local_phone_client = (this.m_view && this.m_view.m_phoneClient) ? this.m_view.m_phoneClient : undefined;
				if (local_phone_client === undefined) {
					local_phone_client = inConcert.app.getInteractionManager().getChannelClient('CALL');
				}
				if(hasScreenRecording)
				{
					inConcert.app.StartScreenRecording(this.type, this.id,local_phone_client._sip_phone, function(success, error)
					{
						if(success)
						{
							self.setActiveScreenRecording(true);
						}
						else{
							console.log("StartScreenRecording: " + error);
						}
					});
				}
				else
				{
					inConcert.app.AddAudioToScreenCapture(this.type, this.id,local_phone_client._sip_phone);
				}
			}
			else
			{
				if(data.state !== "TAKEN" && previous_it_state !== "TRANSFER TAKEN" && data.state !== "TRANSFER LOCKED" && data.state !== "TRANSFER TAKEN" && data.state !== "CONFERENCE" && data.state !== "CONFERENCE LOCKED" && data.state !== "CONFERENCE TAKEN" && (data.state !== "TAKEN" || (previous_it_state !== "CONFERENCE TAKEN" && previous_it_state !== "CONFERENCE")))
				{
					if(hasScreenRecording && this.hasActiveScreenRecording())
					{
						inConcert.app.StopScreenRecording(this.id, function(success, error)
						{
							if(success)
							{
								self.setActiveScreenRecording(false);
							}
							else{
								console.log("StopScreenRecording: " + error);
							}
						});
					}
				}
			}
		}
		else if(this.type === "CHAT")
		{
			if(data.state === "TAKEN" && data.assignedAgents && data.assignedAgents.length === 1 && data.assignedAgents[0] === inConcert.app.getCompleteId())
			{
				if(hasScreenRecording)
				{
					inConcert.app.StartScreenRecording(this.type, this.id, undefined, function(success, error)
					{
						if(success)
						{
							self.setActiveScreenRecording(true);
						}
						else{
							console.log("StartScreenRecording: " + error);
						}
					});
				}
			}
			else
			{
				if(hasScreenRecording)
				{
					inConcert.app.StopScreenRecording(this.id, function(success, error)
					{
						if(success)
						{
							self.setActiveScreenRecording(false);
						}
						else{
							console.log("StopScreenRecording: " + error);
						}
					});
				}
			}
		}
		
		this.setTotalNewEvents(data);

		//just copy the values to the current conversation
		for (let member in data) {
			if (data.hasOwnProperty(member)) {
				this[member] = data[member];
			}
		}
		// actualizo de forma puntal el attention level, ya que el mismo no viene en "data", cuando se ha quitado
		this.updateAttentionLevel(data.attention_level);

		if( this.state === "WRAPUP" ) {
			let campaign = inConcert.app.getCampaign(this.getCampaign());
			let autoWrapupTimeout = campaign.getThreshold(this.type, "AutoWrapupTimeout");
			if(autoWrapupTimeout && autoWrapupTimeout > 0) {
				// Always send a default disposition code, but the CCI integration is free to ignore it.
				// Even when no disposition code is needed one is sent. The server side code just ignores it.
				let defaultDisposition = campaign.getDefaultDispositionTreePathForAutoWrapup();
				this.__auto_wrapup_timer = setTimeout(function() {
					self.emit("automatic_wrapup_timeout", defaultDisposition);
					if(!self.hasCCIntegration()) {
						self.archive(defaultDisposition);
					}
				}, autoWrapupTimeout * 1000);
				// The timer is cancelled when the conversation is finished
			}
		}
		if (this.state === "FINISHED") {
			if (this.__auto_wrapup_timer) {
				clearTimeout(this.__auto_wrapup_timer);
			}
		}
		
		self.__processUpdateNotifications();

		this.emit("update");
	};

	//relates the conversation with a native session according to the base type
	Interaction.prototype.setSession = function(session) {
		if (session) {
			this.m_session = session;
			session.setInteraction(this);
		}
		if (this.m_view && typeof(this.m_view.onSessionChanged) === "function") {
			this.m_view.onSessionChanged();
		}
	};

	//retrieves conversation native session
	Interaction.prototype.getSession = function( /* session */) {
		return this.m_session;
	};

	Interaction.prototype.get_pending_events = function()
	{
		return this.pending_events || {};
	};

	Interaction.prototype.get_pending_event = function( fake_event_id )
	{
		if ( !this.pending_events ) {
			return;
		}
		return this.pending_events[ fake_event_id ];
	};

	Interaction.prototype.get_pending_events_by_state = function( state )
	{
		var pending_events_by_state = {};

		if ( !state || !this.pending_events ) {
			return pending_events_by_state;
		}

		$.each( this.get_pending_events(), function( id, pending_event ) {
			if ( pending_event.state === state ) {
				pending_events_by_state[ id ] = pending_event;
			}
		});

		return pending_events_by_state;
	};

	Interaction.prototype.count_pending_events_by_states = function()
	{
		var count_by_states = {};

		$.each( this.get_pending_events(), function( _, pending_event ) {
			count_by_states[ pending_event.state ] = ( count_by_states[ pending_event.state ] || 0 ) + 1;
		});

		return count_by_states;
	};

	Interaction.prototype.count_total_pending_events = function()
	{
		var total_count = 0;
		var count_by_states = this.count_pending_events_by_states();

		$.each( count_by_states, function( _, state_count ) {
			total_count = total_count + state_count;
		});

		return total_count;
	};

	/*
		This function assign contact to conversation.
		app_id is optional, default value "interaction"
	*/
	Interaction.prototype.setContact = function( contactId, contactName, on_contact_set, app_id ) {
		var self = this;
		var callback = typeof(on_contact_set) === "function" ? on_contact_set : $.noop;

		var getErrorMessage = function(reason)
		{
			reason = ( reason || "").toUpperCase();
			switch (reason) {
				case "INVALID_REQUEST":
					return "<%=_T('There was an error, please try again later')%>";
				case "FORBIDDEN":
					return format("<%=_T('You have not permissions to set contact conversations on {0} campaign.')%>", self.campaign);
				default:
					return "<%=_T('There was an error, please try again later')%>";
			}
		};

		if (!contactId) {
			self.contact = "";
			self.pending = false;
			callback(false);
			return;
		}

		inConcert.logger.action({action_name : "set_interaction_contact", id: self.id, type: self.type, campaign: self.campaign, contactId: contactId, action_type: "conversation"});
		inConcert.invoke({
			type: "POST",
			url: self.m_manager.getBaseUrl() + "/interactions/set_interaction_contact/",
			data: JSON.stringify({
				interactionId: self.id,
				contactId: contactId,
				contactName: contactName || "",
				app_id: app_id
			}),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				self.setContactObject({ Id : contactId, Name : contactName });
				self.pending = false;
				callback(true);
			},
			error: function(response /*, error */) {
				self.contact = "";
				self.pending = false;
				var errors = {
					"403": "FORBIDDEN"
				};
				callback(false, getErrorMessage(errors[response.status]));
			}
		});
	};

	Interaction.prototype.removeContactConflict = function(on_conflict_removed) {
		var self = this;
		var callback = typeof(on_conflict_removed) === "function" ? on_conflict_removed : $.noop;

		// TO-DO: Check for permissions?

		inConcert.logger.action({action_name : "remove_contact_conflict", id: self.id, type: self.type, campaign: self.campaign, action_type: "interaction"});

		inConcert.invoke({
			type: "POST",
			url: self.m_manager.getBaseUrl() + "/interactions/remove_contact_conflict/",
			data: JSON.stringify({
				interactionId: self.id
			}),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback(true);
			},
			error: function(response /*, error*/ ) {
				switch (response.reason) {
					case "INVALID_REQUEST":
						callback(false, "<%=_T('There was an error, please try again later')%>");
						break;
					case "FORBIDDEN":
						callback(false, format("<%=_T('You have not permissions to add contact conversations on {0} campaign.')%>", self.campaign));
						break;
					default:
						callback(false, "<%=_T('There was an error, please try again later')%>");
				}
			}
		});
	};

	Interaction.prototype.discardFailedEvent = function( fake_event_id, callback )
	{
		callback = callback || $.noop;

		inConcert.logger.action({action_name : "discard_failed_event", id: this.id, type: this.type, campaign: this.campaign, action_type: "interaction"});

		inConcert.invoke({
			type: "POST",
			url: this.m_manager.getBaseUrl() + "/interactions/discard_failed_event/",
			data: JSON.stringify({
				interactionId : this.id,
				fake_event_id : fake_event_id 
			}),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback( true );
			},
			error: function( response /*, error*/ ) {
				callback( false, response );
			}
		});
	};

	Interaction.prototype.failSendingEvent = function( fake_event_id, callback )
	{
		callback = callback || $.noop;

		inConcert.logger.action({action_name : "fail_sending_event", id: this.id, type: this.type, campaign: this.campaign, action_type: "interaction"});

		inConcert.invoke({
			type: "POST",
			url: this.m_manager.getBaseUrl() + "/interactions/fail_sending_event/",
			data: JSON.stringify({
				interactionId : this.id,
				fake_event_id : fake_event_id 
			}),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback( true );
			},
			error: function( response /*, error*/ ) {
				callback( false, response );
			}
		});
	};

	Interaction.prototype.setContactObject = function(contact) {
		if (!contact) {
			return;
		}
		this.contact = contact.Id;
		this.contactObject = contact;
	};

	/* Sets the conversation as read, with the following rationale:
		- Live conversations do not update read marks to the server cause it make no sense at all
		- If an event is specified only that specific conversation event is marked as read
		- If no specific event is specified all the events are marked as read for the user (-1)
	*/
	Interaction.prototype.setRead = function( mark_as_read, event, onReadCompleted )
	{
		var self = this;
		onReadCompleted = onReadCompleted || $.noop;

		if ( this.isFinished() || !this.belongsToMe() || this.state === "PENDING" ) {
			onReadCompleted(true);
			return;
		}
		if ( this.isLive() ) {
			this.read = mark_as_read;
			onReadCompleted(true);
			return;
		}

		var get_read_events = function()
		{
			if ( $.isArray( event ) ) {
				return event;
			}
			event = event || 0;
			return event >= 0 ? [ event ] : [];
		};

		var events = get_read_events().map(function(index) {
			return parseInt(index);
		});
		var url = this.m_manager.getBaseUrl() + "/interactions/read_interaction/";

		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify({
				id     : this.id,
				read   : mark_as_read,
				events : events
			}),
			dataType: "json",
			contentType: "application/json",
			success: function( success )
			{
				if ( success ) {
					var readers = self.readers_v2 || {};
					$.each( events, function( _, event_index )
					{
						if ( mark_as_read ) {
							delete readers[ "I" + event_index ];
						}
					});
				}
				onReadCompleted( success );
			},
			error: function( /* response, error*/ ) {
				onReadCompleted( false );
			}
		});
	};

	Interaction.prototype.getReadEventsCount = function() {
		return this.getEventCount() - this.getUnreadEventCount();
	};

	Interaction.prototype.getContactsEvents = function() {
		var events = [];
		$.each( this.getEventList() || [], function( event_index, event ) {
			if ( event.creator === "contact" ) {
				events.push( event_index );
			}
		});
		return events;
	};

	Interaction.prototype.getUsersEvents = function() {
		var events = [];
		$.each( this.getEventList() || [], function( event_index, event ) {
			if ( event.creator === "agent" ) {
				events.push( event_index );
			}
		});
		return events;
	};

	Interaction.prototype.isAgentEvent = function( event_index ) {
		this.assert_inflated();
		return this.events[event_index].creator === "agent";
	};

	Interaction.prototype.isContactEventIndex = function(event_index) {
		this.assert_inflated();
		var evt = this.events[event_index];
		return this.isContactEvent(evt);
	};

	Interaction.prototype.isContactEvent = function(evt) {
		this.assert_inflated();
		return evt.creator === "contact";
	};

	Interaction.prototype.hasIncomingEvents = function()
	{
		var self = this;
		this.assert_inflated();

		var has_incoming_mails = false;

		$.each(this.getEventList() || [], function(event_index) {
			var is_contact_event = self.isContactEventIndex(event_index);
			if (is_contact_event) {
				has_incoming_mails = true;
				return false; //break
			}
		});

		return has_incoming_mails;
	};

	/**
	* Returns whether a specific event, or the whole interaction, has been read or not.
	* Remember, the read/unread mark is now by interaction instead of by agent.
	* If no 'event_index' is provided, returns true if all events have been read.
	* @param event_index
	*/
	Interaction.prototype.isRead = function( event_index )
	{
		if ( this.isLive() ) {
			return this.read;
		}

		if (! ("readers_v2" in this) ) {
			return false;
		}

		if ( event_index >= 0 ) {
			var readers = this.readers_v2;
			if (readers[ "I" + event_index ] === "U" ) {
				return false;
			}
		}
		else {
			return this.getUnreadEventCount() === 0;
		}

		return true;
	};

	Interaction.prototype.get_agent_name = function( event )
	{
		this.assert_inflated();

		if ( !event ) {
			return "";
		}
		var logged_user_id = inConcert.app.getCompleteId();
		var agent_id = event.agent || logged_user_id;

		if ( agent_id === logged_user_id ) {
			return inConcert.app.getLoggedUserName();
		}
		else {
			var agent_name;

			inConcert.app.getTeamMember(
				agent_id,
				function( agent ) {
					if ( !agent ) {
						return;
					}
					var can_name = false;
					if ( ( agent.Name || "" ).trim().length > 0 ) {
						agent_name = agent.Name;
						can_name = true;
					}
					if ( ( agent.FirstSurname || "" ).trim().length > 0 ) {
						agent_name = ( can_name ? ( agent_name + " " ) : "" ) + agent.FirstSurname;
					}
				},
				true // from cache
			);

			if ( !agent_name ) {
				agent_name = ( ( event.agentName || "" ).trim().length > 0 ) ? event.agentName : agent_id;
			}

			return agent_name;
		}
	};

	/*
		Esta funcion retorna el ultimo agente que ha finalizado la interaccion.
		(*) Solo para interacciones infladas.
	*/
	Interaction.prototype.getLastAgentEndsConversation = function()
	{
		if ( !this.isFinished() ) {
			return "";
		}
		let agent_with_vcc;
		let last_step = this.getLastStep();
		if ( last_step && last_step.action === "FINISHED" && last_step.agent ) {
			// (1) - (*) Busca en el ultimo STEP si es "finished" y obtiene el agente (solo para interacciones archivadas nuevas).
			agent_with_vcc = last_step.agent;
		}
		else {
			// (2) - (*) Si es una interaccion vieja, el agente no se encontraba en el step, entonces lo busca en la lista de agentes "analysisTimes.agentTakingTimestamps" que la han tomado (el ultimo)
			if ( this.analysisTimes && $.isArray(this.analysisTimes.agentTakingTimestamps) ) {
				let last_agent_taking = this.analysisTimes.agentTakingTimestamps[ this.analysisTimes.agentTakingTimestamps.length - 1 ];
				if ( last_agent_taking ) {
					agent_with_vcc = last_agent_taking.agentTaking;
				}
			}
			// (3) - (*) Si es una interaccion vieja, el agente no se encontraba en el step, ni en 2), entonces lo busca en la lista de agentes "agentHistory" que la han tomado (comparando por fecha)
			if (!agent_with_vcc) {
				agent_with_vcc = this.getLastHistoryAgent();
			}
		}
		// (4) - Si el agente no se obtuvo en (1) ni (2) ni en (3), obtiene el mismo del campo calculado "__last_agent_ends_conversation".
		//       Este campo calculado se obtiene en la "lua_interaction_manip". La forma de obtener el valor es la misma que se utiliza en los pasos (1) y (2) de esta funcion.
		agent_with_vcc = agent_with_vcc || this.__last_agent_ends_conversation;
		let separated_vcc = inConcert.app.separate_vcc( agent_with_vcc );
		if ( separated_vcc ) {
			// si se ha obtenido el agente retorna unicamente el ID sin el VCC
			return separated_vcc.id;
		}
		return "";
	};

	/*
		Funcion que retorna el ultomo "agentId" comparando por fecha.
		El mismo se obtiene de la lista "agentHistory"
	*/
	Interaction.prototype.getLastHistoryAgent = function()
	{
		if (!this.agentHistory) {
			return;
		}
		let last_agent_id, last_timestamp;
		for (let agent_id in this.agentHistory) {
			if (this.agentHistory.hasOwnProperty(agent_id)) {
				let current_timestamp = this.agentHistory[agent_id];
				current_timestamp = inConcert.Utils.getDateFromString(current_timestamp);
				if (!last_timestamp) {
					last_timestamp = current_timestamp;
					last_agent_id = agent_id;
				}
				else {
					if (current_timestamp > last_timestamp) {
						last_timestamp = current_timestamp;
						last_agent_id = agent_id;
					}
				}
			}
		}

		return last_agent_id;
	};

	/*
		returns an address object formatted to be used in a contact
		TODO this should be reviewed because there is a tight coupling unnecesary between the conversation and the contact address representation
	*/
	Interaction.prototype.buildContactAddress = function (contactAddress) {
		window.alert("Interaction.prototype.buildContactAddress - DEPRECATED, NO USAR MAS");
		var interaction = this;
		var address;

		contactAddress = contactAddress || interaction.contactAddress;

		switch (interaction.type) {
			case "SMS":
			case "RCS":
			case "WHATSAPP":
			case "CALL":
			case "VOICEMAIL":
				address = {
					Type : "Phone",
					Number : contactAddress
				};
				break;
			case "INTERNAL_CHAT":
			case "CHAT":
				address = {
					Type : "Webchat",
					customerId : interaction.interactionData.contactAddress
				};
				break;
			case "IM":
				address = {
					Type : "IM",
					protocol : interaction.interactionData.protocol,
					buddy : contactAddress
				};
				break;
			case "MAIL":
				address = {
					Type : "Mail",
					address : contactAddress
				};
				break;
			case "FACEBOOK":
				address = {
					Type : "Facebook",
					UID : contactAddress
				};
				break;
			case "INSTAGRAM":
				address = {
					Type : "Instagram",
					User : contactAddress
				};
				break;
			case "YOUTUBE":
				address = {
					Type : "Youtube",
					User : contactAddress
				};
				break;
			case "APPSTORE":
				address = {
					Type : "Appstore",
					User : contactAddress
				};
				break;
			case "GOOGLEPLAY":
				address = {
					Type : "Googleplay",
					User : contactAddress
				};
				break;
			case "TWITTER":
				address = {
					Type : "Twitter",
					User : contactAddress
				};
				break;
			case "WEBCONTACT":
				address = {
					Type : interaction.getAddressType(),
					address : interaction.getAddress()
				};
				break;
		}
		return address;
	};

	// set campaign id
	Interaction.prototype.setCampaign = function(campaign) {
		this.campaign = campaign;
	};

	// get campaign id
	Interaction.prototype.getCampaign = function() {
		return this.campaign;
	};

	Interaction.prototype.compatibleAccount = function() {
		return true;
	};

	// TODO: Ni campaignIsOperational ni accountIsOperational tienen que ser miembros
	// de esta clase. No tienen absolutamente nada que ver (tanto campaign como account estan o no operativas
	// independientemente de lo que pase con una interaccion).
	Interaction.prototype.campaignIsOperational = function( permissionFn )
	{
		permissionFn = permissionFn || function() { return { enabled: true }; };
		var campaignResponse = {
			success: false
		};
		if ( !this.campaign ) {
			return campaignResponse;
		}
		var campaign = inConcert.app.getCampaign( this.campaign );
		if ( !campaign ) {
			campaignResponse.errorMessage = "<%=_T('Removed campaign')%>";
			return campaignResponse;
		}
		var permission = permissionFn();
		if ( !permission.enabled ) {
			campaignResponse.errorMessage = permission.message || "<%=_T('You dont have permission on current campaign.')%>";
			return campaignResponse;
		}
		campaignResponse.success = true;
		return campaignResponse;
	};

	Interaction.prototype.accountIsOperational = function(current_account_id) {
		const accountResponse = {
			success: false
		};
		current_account_id = current_account_id || this.account;
		if (!this.campaign || !current_account_id) {
			return accountResponse;
		}
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return accountResponse;
		}
		return campaign.accountIsOperational(current_account_id);
	};

	/**
	 * Check if the interaction account has contact management enabled.
	 *
	 * NOTE: This will assume the interaction campaign and the interaction account
	 *
	 * @param app_id
	 * @return {Boolean}
	 */
	Interaction.prototype.accountHasContactManagementEnabled = function( app_id )
	{
		app_id = app_id || "interaction";
		var action_by_apps = { interaction: "int_contacts_mng", supervisor: "sup_contacts_mng" };
		var action_id = action_by_apps[ app_id ];
		if ( !action_id ) {
			return false;
		}

		var interaction = this;
		var campaign_id = interaction.campaign;

		var has_permission = inConcert.permissions.to_action( app_id, action_id, undefined /* category */, campaign_id, true /* no log */ );
		if ( !has_permission ) {
			return false;
		}

		var campaign = inConcert.app.getCampaign( campaign_id );
		// campaigns can be removed with alive interactions
		if (!campaign){
			return false;
		}

		var account = _.findWhere(campaign.accounts, { "id" : interaction.account });
		// accounts can be removed with alive interactions
		if (!account){
			return false;
		}
		var has_contact_management_enabled = account && account.ContactManagement && account.ContactManagement.Enabled === "1";
		return has_contact_management_enabled;
	};

	/*
		Filters and retrieve a list of campaigns having of the contact an account of a particular proxy type.
	*/
	Interaction.prototype.getAllowedCampaigns = function(campaigns) {
		const self = this;
		if (!campaigns) {
			return [];
		}
		// valid if contact belongs to campaign.
		// removed condition
		const contactBelongsToCampaign = function(/*campaignId*/) {
			return true;
			/*var contact = self.contactObject;
			if (!contact.campaigns || contact.campaigns.length === 0) {
				return false;
			}
			for (var i = 0; i < contact.campaigns.length; i++) {
				var campaignObject = inConcert.app.getCampaign(contact.campaigns[i]);
				if (campaignObject && campaignObject.Id === campaignId) {
					return true;
				}
			}
			return false;*/
		};
		// valid if the account is available.
		const accountAvailable = function(account) {
			if (!account || account.state === "0" || !account.status || account.status.status !== "OK") {
				return false;
			}
			return true;
		};
		// iterates the accounts and validates
		const validateAccounts = function(campaign, accounts) {
			for (let ai = 0; ai < accounts.length; ai++) {
				let account = accounts[ai];
				if (account.proxyType === self.type && accountAvailable(account) && 
					self.compatibleAccount(account) && campaign.canCreateOutboundInteraction(account)
				) {
					return true; //only one account is enough to pass the filter
				}
			}
			return false;
		};

		const validateCampaign = function(campaign) {
			if (self.isLive() && !campaign.isOnSchedule()) {
				return false;
			}
			return true;
		};

		let result = [];
		for (let ci = 0; ci < campaigns.length; ci++) {
			let campaign = campaigns[ci];
			if (self.contactObject && !contactBelongsToCampaign(campaign.Id)) {
				continue;
			}
			if (validateAccounts(campaign, campaign.accounts)) {
				if (validateCampaign(campaign)) {
					result.push(campaign);
				}
			}
		}
		return result;
	};

	Interaction.prototype.getFinalizationData = function() {
		return this.finalizationData;
	};

	Interaction.prototype.setFinalizationData = function( finalizationData ) {
		this.finalizationData = finalizationData;
	};

	/*
		Given two events, returns the event with the longest wait for the customer
	*/
	Interaction.prototype.selectBestCustomerEvent = function (lastEvent, currentEvent){
		if (lastEvent.creator === "agent"){
			// if current event is newer
			if (currentEvent.timestamp > lastEvent.timestamp){
				return currentEvent;
			}
		}
		else {
			// if current is from a client, return the older
			if (currentEvent.creator === "contact"){
				if (currentEvent.timestamp < lastEvent.timestamp){
					return currentEvent;
				}
			}
			else {
				if (currentEvent.timestamp > lastEvent.timestamp){
					return currentEvent;
				}
			}
		}
		// If any conditions are met, return the last event
		return lastEvent;
	};

	Interaction.prototype.serialize = function(include_attributes, exclude_attributes, recursiveArray)
	{
		return inConcert.Utils.plainOldObject(this, include_attributes, exclude_attributes, recursiveArray);
	};

	Interaction.prototype.interactionTypeToAddressType = function() {
		var interactionType = this.type;
		var addressType;
		switch (interactionType) {
			case "CALL":
			case "VOICEMAIL":
			case "SMS":
			case "RCS":
			case "WHATSAPP":
				addressType = "Phone";
				break;
			case "FACEBOOK":
				addressType = "Facebook";
				break;
			case "INSTAGRAM":
				addressType = "Instagram";
				break;
			case "YOUTUBE":
				addressType = "Youtube";
				break;
			case "APPSTORE":
				addressType = "Appstore";
				break;
			case "GOOGLEPLAY":
				addressType = "Googleplay";
				break;
			case "TWITTER":
				addressType = "Twitter";
				break;
			case "IM":
				addressType = "IM";
				break;
			case "MAIL":
			case "WEBCONTACT": //TODO: by now it's only mail
				addressType = "Mail";
				break;
			case "CHAT":
				addressType = "Webchat";
				break;
			default:
				var errorMsg = "Conversations - Someone is trying convert an conversation type to an address type with an invalid type: " + interactionType;
				inConcert.logger.error(errorMsg);
				throw errorMsg;
		}
		return addressType;
	};

	/**
	Private helper. Cancels the current auto wrapup timer (if any).
	*/
	Interaction.prototype.__cleanupAutoWrapupTimer = function()
	{
		if(this.__auto_wrapup_timer) {
			clearTimeout(this.__auto_wrapup_timer);
			delete this.__auto_wrapup_timer;
		}
	};

	Interaction.prototype.canSchedule = function()
	{
		return false;
	};

	Interaction.prototype.canMultiSchedule = function()
	{
		return false;
	};

	Interaction.prototype.isScheduleEnabled = function()
	{
		return (this.canMultiSchedule() || !this.hasAppointment()) && this.canSchedule();
	};

	Interaction.prototype.schedule = function(data, callback = $.noop)
	{
		let self = this;
		if (!self.isScheduleEnabled()) {
			return callback(false, "<%=_T('You dont have permissions to schedule interaction.')%>");
		}
		
		let getErrorMessage = function(result) {
			result = result || {};
			switch(result.result || "") {
				case "OK":
					return null;
				case "CANNOT VALIDATE NUMBER":
					return "<%=_T('Invalid phone number.')%>";
				default:
					return result.reason || "<%=_T('Error scheduling conversation.')%>";
			}
		};
		
		let url = this.m_manager.getBaseUrl() + "/interactions/make_appointment/";
		let request = {
			proxyType		: data.proxyType || self.type,
			campaignAtVCC	: self.campaign,
			account			: self.getAccount(),
			interactionID	: self.id,
			addressTable	: data.addressTable,
			fromDate		: data.date? inConcert.Utils.getStringFromDate(data.date, true, "iso") : null
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				callback(result.result === "OK", getErrorMessage(result));
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "make_appointment", error_message : error.result });
				callback(false, error);
			}
		});
	};

	Interaction.prototype.cancelInteraction = function() {
	};

	Interaction.prototype.isTelephonyInteraction = function()
	{
		return false;
	};
	
	Interaction.prototype.isRecording = function()
	{
		return false;
	};
	/**
	 * Indicates if is an allegro campaign
	 */
	Interaction.prototype.isAllegro = function()
	{
		return inConcert.app.has_allegro_integration() && (this.type === "CALL" || this.type === "VOICEMAIL");
	};
	
	/**
	 * 
	 * Indicates if has to show integration
	 */
	Interaction.prototype.isIntegrationEnabled = function()
	{
		return this.hasPreview() || this.isAssigned() || this.isWrappedUp();
	};
	
	Interaction.prototype.getSNPostId = function()
	{
		return this.interactionData && this.interactionData.postId;
	};
	
	Interaction.prototype.isSNPostReply = function()
	{
		return !!this.getSNPostId();
	};
	
	Interaction.prototype.getBatchId = function()
	{
		return this.interactionData && this.interactionData.batchId;
	};
	
	Interaction.prototype.isBatchReply = function()
	{
		return !!this.getBatchId();
	};
	
	Interaction.prototype.canSendAttachment = function() {
		return this.isTaken() && this.hasPermission("CHAT_FILE_TRANSFER");
	};
	
	Interaction.prototype.canSendSharedFile = function() {
		return this.canSendAttachment();
	};
	
	Interaction.prototype.isSendSendAttachmentAsync = function() {
		return false;
	};
	
	Interaction.prototype.isSendSharedFileAsync = function() {
		return false;
	};
	
	Interaction.prototype.getAgentDisplayName = function() {
		return inConcert.app.getLoggedUserFirstName();
	};
	
	Interaction.prototype.__getSendAttachmentError = function(/*response*/) {
		return "<%=_T('Error uploading file.')%>";
	};
	
	Interaction.prototype.sendAttachment = function(data, file, callback) {
		var self = this;
		if (!self.canSendAttachment()) {
			return callback(false, "<%=_T('You dont have permissions to send attachment.')%>");
		}
		var url = this.m_manager.getBaseUrl() + "/interactions/send_attachment/";
		var request = {
			proxyType		: self.type,
			campaignAtVCC	: self.campaign,
			account			: self.getAccount(),
			interactionID	: self.id,
			data			: data,
			displayName		: self.getAgentDisplayName(),
			vcc             : inConcert.Utils.md5(inConcert.app.get_vcc_id()).toUpperCase()
		};
		var fileUpload = new inConcert.FileUpload();
		fileUpload.uploadFile(url, file, request, function(status, response) {
			if (status) {
				if (self.isSendSendAttachmentAsync()) {
					self.__asyncSendMessageCallback(function(result) {
						if (result.result === "OK") {
							callback(true, result);
						}
						else {
							callback(false, self.__getSendAttachmentError(result.reason));
						}
					});
				}
				else {
					callback(true, JSON.parse(response));
				}
			}
			else {
				callback(false, self.__getSendAttachmentError(response));
			}
		});
	};
	
	Interaction.prototype.sendRepositoryFile = function(data, fileData, callback) {
		let self = this;
		if (!self.canSendAttachment()) {
			return callback(false, "<%=_T('You dont have permissions to send attachment.')%>");
		}
		let url = this.m_manager.getBaseUrl() + "/interactions/send_repository_file/";
		let request = {
			proxyType		: self.type,
			campaignAtVCC	: self.campaign,
			account			: self.getAccount(),
			interactionID	: self.id,
			data			: data,
			file			: fileData,
			displayName		: self.getAgentDisplayName()
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				if (result.result === "OK") {
					if (self.isSendSendAttachmentAsync()) {
						self.__asyncSendMessageCallback(function(result) {
							if (result.result === "OK") {
								callback(true);
							}
							else {
								let message = self.get_error_message(result.result, result.reason, "SEND_REPOSITORY_FILE", result);
								callback(false, message, result.result);
							}
						});
					}
					else {
						callback(true);
					}
				}
				else {
					let message = self.get_error_message(result.result, result.reason, "SEND_REPOSITORY_FILE", result);
					callback(false, message, result.result);
				}
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "send_repository_file", error_message : error.result });
				let message = self.get_error_message( error.result, error.reason, "SEND_REPOSITORY_FILE", error );
				callback( false, message, error );
			}
		});
	};
	
	Interaction.prototype.sendSharedFile = function(message, callback) {
		let self = this;
		if (!self.canSendSharedFile()) {
			return callback(false, "<%=_T('You dont have permissions to send attachment.')%>");
		}
		let url = this.m_manager.getBaseUrl() + "/interactions/send_shared_file/";
		let request = {
			proxyType		: self.type,
			campaignAtVCC	: self.campaign,
			account			: self.getAccount(),
			interactionID	: self.id,
			message			: message,
			displayName		: self.getAgentDisplayName()
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				if (result.result === "OK") {
					if (self.isSendSharedFileAsync()) {
						self.__asyncSendMessageCallback(function(result) {
							if (result.result === "OK") {
								callback(true);
							}
							else {
								let message = self.get_error_message(result.result, result.reason, "SEND_SHARED_FILE", result);
								callback(false, message, result.result);
							}
						});
					}
					else {
						callback(true);
					}
				}
				else {
					let message = self.get_error_message(result.result, result.reason, "SEND_SHARED_FILE", result);
					callback(false, message, result.result);
				}
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "send_shared_file", error_message : error.result });
				let message = self.get_error_message( error.result, error.reason, "SEND_SHARED_FILE", error );
				callback( false, message, error );
			}
		});
	};

	Interaction.prototype.getPage = function() {
		return null;
	};

	Interaction.prototype.getAccountConfiguration = function() {
		const campaign = inConcert.app.getCampaign(this.campaign);
		if (!campaign) {
			return null;
		}
		return campaign.getAccount(this.account);
	};

	Interaction.prototype.getAccountConfigGUI = function() {
		const account_config = this.getAccountConfiguration();
		if (!account_config) {
			return {};
		}
		return account_config.GUI || {};
	};

	Interaction.prototype.hasTemplateMessages = function() {
		var campaign = inConcert.app.getCampaign(this.campaign);
		return campaign && campaign.hasAccountTemplateMessageOptions(this.getAccountConfiguration()) || false;
	};

	Interaction.prototype.canSendTemplateMessages = function() {
		return (this.isTaken() || this.isBeingComposed()) && this.hasTemplateMessages();
	};

	let m_templateMessageOptionsCache = null;
	function makeTemplateMessageOptionsCache() {
		m_templateMessageOptionsCache = new inConcert.Cache(
			function(fetchData, callback) {
				fetchData.campaign.loadTemplateMessageOptions(fetchData.account, fetchData.page, true, 
					function(data, error) {
						callback(data, error);
					});
			},
			function(fetchData) {
				return fetchData.campaign.Id + ":" + fetchData.account + (fetchData.page? (":" + fetchData.page) : "");
			},
			60
		);
	}
	
	Interaction.prototype.loadTemplateMessageOptions = function(callback) {
		if (!this.hasTemplateMessages()) {
			return callback(false, "account not supports template messages");
		}
		if (!m_templateMessageOptionsCache) {
			makeTemplateMessageOptionsCache();
		}
		var campaign = inConcert.app.getCampaign(this.campaign);
		m_templateMessageOptionsCache.get({
				campaign : campaign,
				account : this.account,
				page : this.getPage()
			},
			function(options, error) {
				callback(options, error);
			}
		);
	};

	Interaction.prototype.sendTemplateMessage = function(/*params*/) {
		throw "not implemented";
	};

	Interaction.prototype.getSubtype = function() {
		return this.interactionData && this.interactionData.subtype || "";
	};

	Interaction.prototype.canForward = function() {
		return this.hasForward() && this.isTaken() && this.hasPermission("FORWARD_MESSAGES");
	};

	Interaction.prototype.hasForward = function() {
		return false;
	};
	
	Interaction.prototype.forward = function(/*address, messages, callback*/) {
		throw "not implemented";
	};

	Interaction.prototype.getForwardMessage = function(/*index*/) {
		throw "not implemented";
	};

	Interaction.prototype.get_error_message = function(/*error_code, reason, type, error*/) {
		return "<%=_T('There was an error while sending message')%>";
	};

	Interaction.prototype.setDraftMessage = function(draftMessage, autosend, callback) {
		this.draftMessage = draftMessage;
		this.autoSendDraftMessage = autosend;
		this.draftMessageCallback = callback;
	};

	Interaction.prototype.removeDraftMessage = function() {
		this.draftMessage = null;
		this.autoSendDraftMessage = null;
		this.draftMessageCallback = null;
	};

	Interaction.prototype.getDraftMessage = function() {
		return this.draftMessage;
	};

	Interaction.prototype.hasDraftMessage = function() {
		return !!this.draftMessage;
	};

	Interaction.prototype.autosendDraftMessage = function() {
		return !!this.autoSendDraftMessage;
	};

	Interaction.prototype.getDraftMessageCallback = function() {
		return this.draftMessageCallback;
	};
	
	Interaction.prototype.__notifyNew = function() {
		this.m_beforeNotificationsData = this.__makeNotificationsBeforeData();
		let im = inConcert.app.getInteractionManager();
		if (!im.isInteractionAssigned(this)) {
			return;
		}
		if (!this.isBeingComposed()) {
			this.__fireInteractionAssigned();
		}	
	};
	
	Interaction.prototype.__makeNotificationsBeforeData = function() {
		let im = inConcert.app.getInteractionManager();
		let data = {
			state : this.state,
			assigned : im.isInteractionAssigned(this),
			draft : this.isBeingComposed(),
			eventCount : this.getEventCount()
		};
		return data;
	};

	Interaction.prototype.__processUpdateNotifications = function() {
		let im = inConcert.app.getInteractionManager();
		if (!this.m_beforeNotificationsData) {
			return;
		}
		let beforeData = this.m_beforeNotificationsData;
		
		if (im.isInteractionAssigned(this)) {
			//assigned interaction
			if (!beforeData.assigned || (beforeData.draft && !this.isBeingComposed())) {
				//new interaction
				this.__notifyNew();
			}
			else {
				//existing interaction
				if (beforeData.state !== this.state) {					

					//state change
					if (this.isAssigned()) {
						let interaction = im.findInteraction(this.id);
						if (interaction && interaction.isTelephonyInteraction()) {
							this.__fireCallOnRecording(interaction.isRecording());						
						}
					} 
					else if (this.isWrappedUp()) {
						this.__fireInteractionInWrapup();
						if (beforeData.state === "TRANSFERRING") {
							this.__fireNotification("InteractionTransferred");
						}
					} 
				}
				if (this.isRedial()) {
					this.__fireCallTaken();
				}				
				let eventCount = this.getEventCount();
				if (beforeData.eventCount > 0 && eventCount > beforeData.eventCount) {
					this.__fireNewInteractionMessages(beforeData.eventCount, eventCount);
				}
			}
		}
		else {
			//interaction not assigned
			if (beforeData.assigned && !beforeData.draft) {
				//interaction removed
				if (this.state === "QUEUED") {
					this.__fireInteractionRequeued();
				}
				else if (this.state.indexOf("TRANSFER") >= 0) {
					this.__fireInteractionTransferred();
				}
				else if (beforeData.state.indexOf("TRANSFER") >= 0 && this.state === "TAKEN") {
					this.__fireInteractionTransferred();
				}
				else if (this.state === "FINISHED") {
					this.__fireInteractionArchived();
				}
				else {
					this.cancelInteraction();
					this.__fireInteractionCanceled();
				}
			}
		}

		//update notifications data to compare with next event
		this.m_beforeNotificationsData = this.__makeNotificationsBeforeData();
	};

	Interaction.prototype.__fireNotification = function(eventName, ...args) {
		let api = inConcert.app.getAgentApiServer();
		if (api && api[eventName]) {
			api[eventName](this, ...args);
		}
	};

	Interaction.prototype.__fireCallOnRecording = function(recording) {
		this.__fireNotification("CallOnRecording", recording);
	};

	Interaction.prototype.__fireCallTaken = function(/*recording*/) {		
		this.__fireNotification("CallTaken");
		this.__fireNotification("InteractionAssigned");
	};

	Interaction.prototype.__fireInteractionAssigned = function() {
		this.__fireNotification("InteractionAssigned");
	};

	Interaction.prototype.__fireInteractionRequeued = function() {
		this.__fireNotification("InteractionRemovedFromAgent");
		this.__fireNotification("InteractionRequeued");
	};

	Interaction.prototype.__fireInteractionTransferred = function() {
		this.__fireNotification("InteractionRemovedFromAgent");
		this.__fireNotification("InteractionTransferred");
	};
	
	Interaction.prototype.__fireInteractionArchived = function() {
		this.__fireNotification("InteractionRemovedFromAgent");
		this.__fireNotification("InteractionArchived");
	};

	Interaction.prototype.__fireInteractionCanceled = function() {
		this.__fireNotification("InteractionRemovedFromAgent");
		this.__fireNotification("InteractionCanceled");
	};

	Interaction.prototype.__fireInteractionInWrapup = function() {
		this.__fireNotification("InteractionInWrapup");
	};

	Interaction.prototype.__fireNewInteractionMessages = function(sinceEvent, eventCount) {
		this.__fireNotification("NewInteractionMessages", sinceEvent, eventCount);
	};

	Interaction.prototype.getClientId = function() {
		return this.interactionData && this.interactionData.clientId || "";
	};

	Interaction.prototype.getBrowsingHistory = function(callback = $.noop) {
		let client_id = this.getClientId();
		inConcert.logger.action({action_name : 'get_browsing_history', client_id : client_id, action_type:'trackme'});

		if (!client_id) {
			setTimeout(function() { callback({}); }, 0);
			return;
		}

		const url = inConcert.app.getTrackmeUrl() + '/get_visited_pages/' + client_id;
		inConcert.invoke({
			url: url,
			type: 'GET',
			dataType: 'json',
			contentType: 'application/json',
			success: function(response) {
				callback(response.page_list);
			},
			error: function(response, error){
				if (response.status === 404){
					callback([]);
				}
				else{
					callback(null, error);
				}
			}
		});
	};

	Interaction.prototype.getFinalizationMessageConfiguration = function(callback = $.noop) {
		let self = this;
		let campaign = inConcert.app.getCampaign( self.campaign );
		if ( !campaign ) {
			return callback(false);
		}
		campaign.getAccountMessages(self.account, function( message, error ) {
			if ( error || !message ) {
				return callback(false);
			}
			let config = message.finalization_message;
			callback(true, config);
		}, "administration");
	};

	Interaction.prototype.canHangup = function() {
		return false;
	};

	Interaction.prototype.getEventUserId = function(/*eventIndex*/)
	{
		return "TODO getEventUserId implement";
	};

	Interaction.prototype.getEventUserlIdByChannel = function(/*eventIndex*/)
	{
		if (!this.interactionData) {
			return "";
		}
		return this.getDisplayName();
	};

	Interaction.prototype.getEventUserAvatarURI = function(/*event*/)
	{
		return "";
	};

	Interaction.prototype.__asyncPollerCallback = function(event, callback)
	{
		inConcert.app.addPollerEventCallback(event, this.id, callback);
	};

	Interaction.prototype.__asyncSendMessageCallback = function(callback)
	{
		this.__asyncPollerCallback("async_message_result", callback);
	};

	Interaction.prototype.getAnalysisTimesInfo = function()
	{
		if (this.properties && this.properties.analysisTimes) {
			return this.properties.analysisTimes;
		}
		if (this.analysisTimes) {
			return this.analysisTimes;
		}
		return null;
	};

	Interaction.prototype.getAttentionTime = function()
	{
		let analysisTimes = this.getAnalysisTimesInfo();
		if (!analysisTimes) {
			return 0;
		}
		if (analysisTimes.ReporterInfo && analysisTimes.ReporterInfo.AttentionTime) {
			return analysisTimes.ReporterInfo.AttentionTime;
		}
		if (analysisTimes.IRInfo && analysisTimes.IRInfo.AttentionTime) {
			return analysisTimes.IRInfo.AttentionTime;
		}
		return analysisTimes.attentionTime || 0;
	};

	Interaction.prototype.setPreInteractionData = function(data, callback)
	{
		inConcert.invoke({
			url: this.m_manager.getBaseUrl() + "/interactions/set_preinteraction_data/",
			type: "POST",
			data: JSON.stringify({ 
				id: this.id, 
				data : data
			}),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback(true);
			},
			error: function() {
				callback(false);
			}
		});
	};

	Interaction.prototype.isGone = function()
	{
		return false;
	};

	Interaction.prototype.getRecords = function(callback, records = [])
	{
		records = inConcert.Utils.sortArray(records, "Part", "asc");
		callback(true, records);
	};

	Interaction.prototype.isDialer = function() {
		return !!this.interactionData?.dialer;
	};

	Interaction.prototype.hasPreview = function() {
		return this.isDialer() && !!this.interactionData.dialer.hasPreview;
	};

	Interaction.prototype.isPreview = function() {
		return this.isDialer() && !!this.interactionData.dialer.inPreview;
	};

	Interaction.prototype.acceptPreview = function(callback)
	{
		callback = callback || $.noop;
		inConcert.invoke({
			url: this.m_manager.getBaseUrl() + "/interactions/dialer/preview_confirm/",
			type: "POST",
			data: JSON.stringify({
				interactionID: this.id,
				proxyType: this.type
			}),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback(true);
			},
			error: function() {
				callback(false, "Error accepting preview");
			}
		});
	};

	Interaction.prototype.cancelPreview = function(callback)
	{
		callback = callback || $.noop;
		inConcert.invoke({
			url: this.m_manager.getBaseUrl() + "/interactions/dialer/preview_cancel/",
			type: "POST",
			data: JSON.stringify({
				interactionID: this.id,
				proxyType: this.type
			}),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback(true);
			},
			error: function() {
				callback(false, "Error canceling preview");
			}
		});
	};

	Interaction.prototype.isVirtualhold = function()
	{
		return !!this.interactionData?.virtualHoldOriginInteraction;
	};

	Interaction.prototype.isVirtualholdStarted = function()
	{
		return !!this.interactionData?.virtualHoldStarted;
	};

	Interaction.prototype.isVirtualholdStarting = function()
	{
		return this.isVirtualhold() && !this.isVirtualholdStarted;
	};

	Interaction.prototype.hasVirtualhold = function()
	{
		return !!this.interactionData?.virtualHoldInteraction;
	};

	Interaction.prototype.hasScreenRecording = function() {
		return !!this.interactionData?.hasScreenRecording;
	};

	Interaction.prototype.hasActiveScreenRecording = function() {
		return this.hasScreenRecording() && this.isActiveScreenRecording;
	};

	Interaction.prototype.setActiveScreenRecording = function(active) {
		this.isActiveScreenRecording = active;
	};

}());
