/*jshint bitwise:true, browser:true, curly:true, eqeqeq:true, noarg:true, noempty:true, strict:true, undef:true, unused:true, -W014, esversion: 11 */
/*jshint quotmark:false*/
/*global _, EventEmitter, inConcert, console, $, SIP, MediaStream*/
this.inConcert = this.inConcert || {};

(function () {
	"use strict";

	/*!
	 * Possible phone states
	 */
	var PhoneStates = {
		IDLE       : "idle",
		CONNECTING : "connecting",
		CONNECTED  : "connected",
		OUTBOUND   : "outbound",
		INBOUND    : "inbound",
		ACTIVE     : "active"
	};

	/*!
	 * A wrapper around sip.js
	 *
	 * Events:
	 * \li rejected(reason)
	 * \li warning(message)
	 * \li error(message)
	 * \li invited(callerid)
	 * \li connected()
	 * \li connecting()
	 * \li dialing()
	 * \li connected() -> after connecting()
	 * \li accepted()  -> after dialing()
	 * \li cancelled()
	 * \li byed()
	 * \li ringing()
	 */
	function SipPhone(params) {
		// Initial setup
		_.extend(this, {
			// public, handles the state machine
			currentState    : PhoneStates.IDLE,
			muted: false,
			videoDisabled: false,
			switched: false,

			m_initialized   : false,
			m_error         : null,
			m_defaultHeaders: [
				"Organization: inConcert"
			],
			m_hasPermission: false,
			m_hasLibPermission: true,
			m_libPermissionAccepted : false,
			m_libPermissionRequesting : false,
			m_callConfig   : undefined,
			
			localVideo	   : undefined,
			remoteVideo	   : undefined,
			remoteAudio	   : undefined
		}, params);

		this.logger = this.logger || {
			debug: function () { console.debug.call(null, arguments); },
			info: function () { console.info.call(null, arguments); },
			warn: function () { console.warn.call(null, arguments); },
			error: function () { console.error.call(null, arguments); }
		};

		//this.logger.info("SipPhone", params);
		//SIPml.setDebugLevel("warn");
	}
	inConcert.Utils.inherit(SipPhone, EventEmitter);

	/*!
	 * Connects and registers with Asterisk (aka the SIP server)
	 */
	SipPhone.prototype.connect = function() {
		this.logger.debug("SIPPhone: Connect");
		if (this.currentState !== PhoneStates.IDLE) {
			return this.__invalidState("connect");
		}
		this.__setState(PhoneStates.CONNECTING);
		this.__connect();
	};

	/*!
	 * Disconnects from the SIP server, closes all sessions
	 */
	SipPhone.prototype.disconnect = function() {
		this.logger.debug("SIPPhone: Disconnect");
		this.__setState(PhoneStates.IDLE);
		if (this.m_userAgent) {
			this.m_userAgent.disconnect(); // shutdown all sessions
			this.m_userAgent = null;
		}
		this.logger.debug("SIPPhone: Disconnected");
	};

	/*!
	 * Makes a call
	 */
	SipPhone.prototype.invite = function(params) {
		const interactionID = params.interactionID || "";
		const accountId     = params.accountId || "";
		//const campaignAtVCC = params.campaignAtVCC;	// Not used
		const prefix = params.prefix;
		const destination = params.destination;
		const contact = params.contact;
		const isInternal = params.isInternal || false;
		const callback = params.callback || $.noop;

		if (this.currentState !== PhoneStates.CONNECTED) {
			this.emit("rejected", "Cannot call while " + this.currentState);
			this.__invalidState("invite");
			return callback(false, "Cannot call while " + this.currentState);
		}

		var number = prefix + destination;
		this.__setState(PhoneStates.OUTBOUND);
		
		const sipUrl = `sip:${encodeURIComponent(number)}@${this.rtmpServer}`;
		this.m_userAgent.call(sipUrl,
			{
				extraHeaders : [
					...this.m_defaultHeaders,
					`ICC_INTERACTIONID: ${interactionID}`,
					`INTERNAL_CALL: ${isInternal}`,
					`CALL_ACCOUNT_ID: ${accountId}`,
					`ICC_CONTACT_DATA: ${contact ? JSON.stringify({
						Id: contact.Id,
						Name: contact.Name,
						externalId: contact.externalId
					}) : ""}`,
				],
				earlyMedia : true,
				sessionDescriptionHandlerOptions : {
					iceGatheringTimeout : 3000
				}
			},
			{
				requestDelegate : {
					onProgress : (response) => this.__onProgress(response),
					onTrying : (response) => this.__onTrying(response, false),
					onReject : (response) => this.__onReject(response, false)
				}
			})
			.then(() => {
				//this.emit("ringing");
				callback(true);
			})
			.catch((error) => {
				//Another event change status phone and it was emitted
				if (this.currentState === PhoneStates.OUTBOUND) {
					this.__setState(PhoneStates.CONNECTED);
					this.emit("rejected", "Failed to make call");
					callback(false, "Failed to make call");
					console.error(error);
				}
			});
	};
	
	SipPhone.prototype.__onProgress = function (response) {
		this.logger.debug(`SIPPhone: Progress, Code: ${response.message.statusCode} Reason: ${response.message.reasonPhrase}`);
		if (response.message.statusCode === 180) {
			this.emit("ringing");
		} 
		//else if (response.message.statusCode === 183) {}	// Session progress extra info not used
	};

	SipPhone.prototype.__onTrying = function (response, hasVideo) {
		this.logger.debug(`SIPPhone: Trying, Code: ${response.message.statusCode} Reason: ${response.message.reasonPhrase}`);

		 this.setupLocalMedia(hasVideo);
		 this.setupRemoteMedia(hasVideo);
	};

	SipPhone.prototype.setupLocalMedia = function (hasVideo) {
		var self = this;
		//const mediaElement = $("#telephony-local-video")[0];
		const mediaElement = hasVideo ? this.localVideo : undefined;

		if (mediaElement) {
			if (!this.m_userAgent.session.sessionDescriptionHandler) {
				this.__emitError("Local media sessionDescriptionHandler undefined.");
				return;
			}
			const localStream = this.m_userAgent.session.sessionDescriptionHandler.localMediaStream;
			if (!localStream) {
				this.__emitError("Local media stream undefined.");
				return;
			}
			mediaElement.srcObject = localStream;
			mediaElement.volume = 0;
			mediaElement.onloadedmetadata = function (/*e*/) {
				mediaElement.play().catch((error) => {
					self.logger.error(`[${this.id}] Failed to play remote media`);
					self.logger.error(error.message);
					console.error(error);
				});
			};
		}
	};

	SipPhone.prototype.setupRemoteMedia = function (hasVideo) {
		const self = this;
		// Gets remote tracks
		const pc = this.m_userAgent.session?.sessionDescriptionHandler?.peerConnection;
		if (!pc) {
			this.__emitError("Remote media stream undefined.");
			return;
		}
		const remoteAudioStream = new MediaStream();
		const remoteVideoStream = new MediaStream();
		pc.getReceivers().forEach(function (receiver) {
			if(receiver.track){
				if(receiver.track.kind === "audio"){
					remoteAudioStream.addTrack(receiver.track);
				}
				if(receiver.track.kind === "video"){
					remoteVideoStream.addTrack(receiver.track);
				}
			}
		});
		const remoteAudio = this.remoteAudio;  //$("#telephony-remote-audio")[0];
		remoteAudio.autoplay = true; // Safari hack, because you cannot call .play() from a non user action
		remoteAudio.srcObject = remoteAudioStream;
		remoteAudio.onloadedmetadata = function(/*e*/) {
		remoteAudio.play().catch((error) => {
			self.logger.error(`[${this.id}] Failed to play remote media`);
			self.logger.error(error.message);
			console.error(error);
			});
		};
		if (hasVideo) {
			const remoteVideo = this.remoteVideo;
			remoteVideo.srcObject = remoteVideoStream;
			remoteVideo.onloadedmetadata = function(/*e*/) {
				remoteVideo.play();
			};
		}
	};

	SipPhone.prototype.cleanupMedia = function (hasVideo) {

		//const mediaElement = $("#telephony-remote-audio")[0];
		var mediaElement = this.remoteAudio;
		mediaElement.srcObject = null;
		mediaElement.pause();

		if (hasVideo) {
			mediaElement = this.remoteVideo;
			mediaElement.srcObject = null;
			mediaElement.pause();

			var mediaElementLocal = this.localVideo;
			mediaElementLocal.srcObject = null;
			mediaElementLocal.pause();
		}
	};

	SipPhone.prototype.__onReject = function (response, hasVideo) {
		this.logger.debug(`SIPPhone: Reject, Code: ${response.message.statusCode} Reason: ${response.message.reasonPhrase}`);
		this.cleanupMedia(hasVideo);
		//Another event change status phone and it was emitted
		if (this.currentState === PhoneStates.OUTBOUND) {
			this.__setState(PhoneStates.CONNECTED);
			let message = {};
			message.result = "REJECT CALL";
			message.reason = `Code: ${response.message.statusCode} Reason: ${response.message.reasonPhrase}`;
			this.emit("message", JSON.stringify(message));
		}
	};

	/*!
	 * Checks if has permission on library
	 */
	SipPhone.prototype.__hasLibPermission = function(callback) {
		let self = this;
		inConcert.Utils.polling({
			interval : 500,
			timeout : 60 * 1000,
			onPolling : function() {
				if (self.m_hasLibPermission || !self.m_libPermissionRequesting) {
					callback(self.m_hasLibPermission);
					return false;
				}
			},
			onTimeout : function() {
				callback(self.m_hasLibPermission);
			}
		});
	};

	/*!
	 * Accepts an incoming call
	 */
	SipPhone.prototype.accept = function(callback = $.noop) {
		const self = this;
		if (this.currentState !== PhoneStates.INBOUND) {
			self.__invalidState("accept");
			return callback(false, "Can not accept on: " + this.currentState);
		}
		this.__hasLibPermission(function(hasPermission) {
			if (!hasPermission) {
				self.__permissionPending('accept');
				return callback(false, "Has not permissions");
			}
			const state = self.m_userAgent.session ? self.m_userAgent.session.state : "Empty";
			if (state === 'Establishing') {
				callback(false, "failed to accept");
				return;
			}
			self.m_userAgent.answer({
				extraHeaders : self.m_defaultHeaders,
				sessionDescriptionHandlerOptions: {
					constraints: {
						audio: true,
						video: self.hasVideo()
					}
				}
			})
				.then(() => {
					callback(true);
				})
				.catch((error) => {
					callback(false, "failed to accept");
					console.error(error);
				});
		});
	};

	/*!
	 * Rejects an incoming call
	 */
	SipPhone.prototype.reject = function(reason, callback = $.noop) {
		if (this.currentState !== PhoneStates.INBOUND) {
			this.__invalidState("reject");
			return callback(false, "Can not reject on: " + this.currentState);
		}
		this.m_userAgent.decline({
			extraHeaders : this.m_defaultHeaders
		})
			.then(() => {
				callback(true);
			})
			.catch((error) => {
				this.__emitError("failed to reject");
				callback(false, "failed to reject");
				console.error(error);
			});
	};

	/*!
	 * Hangs-up the active call
	 */
	SipPhone.prototype.bye = function(callback = $.noop) {
		if (!_.contains([PhoneStates.ACTIVE, PhoneStates.OUTBOUND], this.currentState)) {
				this.__invalidState("hangup");
			return callback(false, "Can not bye on: " + this.currentState);
		}
		this.m_userAgent.hangup()
			.then(() => {
				callback(true);
			})
			.catch((error) => {
				this.__emitError("failed to bye");
				callback(false, "failed to bye");
				console.error(error);
			});
	};

	/*!
	 * Sends a DTMF digit
	 */
	SipPhone.prototype.sendDigit = function(digit, callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("dtmf");
			return callback(false, "Can not send dtmf on: " + this.currentState);
		}
		if ( "1234567890ABCD*#".indexOf(digit) < 0 ) {
			return callback(false, "Invalid dtmf");
		}
		this.m_userAgent.sendDTMF(digit)
			.then(() => {
				callback(true);
			})
			.catch((error) => {
				callback(false, "error sending dtmf");
				console.error(error);
			});
	};

	/*!
	 * Puts the current call on hold
	 */
	SipPhone.prototype.hold = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("hold");
			return callback(false, "Can not hold on: " + this.currentState);
		}
		this.m_userAgent.hold()
			.then(() => {
				callback(true);
			})
			.catch((error) => {
				this.__emitError("failed to hold");
				callback(false, "failed to hold");
				console.error(error);
			});
	};

	/*!
	 * Resumes the current call
	 */
	SipPhone.prototype.resume = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("resume");
			return callback(false, "Can not resume on: " + this.currentState);
		}
		this.m_userAgent.unhold()
			.then(() => {
				callback(true);
			})
			.catch((error) => {
				this.__emitError("failed to resume");
				callback(false, "failed to resume");
				console.error(error);
			});
	};

	/*!
	 * Puts the current call on mute
	 */
	SipPhone.prototype.mute = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("mute");
			return callback(false, "Can not mute on: " + this.currentState);
		}
		if (this.hasVideo()) {
			this.toggleSender({ type : "audio", enabled : false });
		}
		else {
			this.m_userAgent.mute();
		}
		this.muted = true;
		callback(true);
	};

	/*!
	 * Removes the current call on mute
	 */
	SipPhone.prototype.unmute = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("unmute");
			return callback(false, "Can not unmute on: " + this.currentState);
		}
		if (this.hasVideo()) {
			this.toggleSender({ type : "audio", enabled : true });
		}
		else {
			this.m_userAgent.unmute();
		}
		this.muted = false;
		callback(true);
	};

	/*!
	 * Disable video
	 */
	SipPhone.prototype.disableVideo = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("disableVideo");
			return callback(false, "Can not disable video on: " + this.currentState);
		}
		else if (!this.hasVideo()) {
			return callback(false, "has not video");
		}
		this.toggleSender({ type : "video", enabled : false });
		this.videoDisabled = true;
		callback(true);
	};

	/*!
	 * Enable video
	 */
	SipPhone.prototype.enableVideo = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("enableVideo");
			return callback(false, "Can not enable video on: " + this.currentState);
		}
		else if (!this.hasVideo()) {
			return callback(false, "has not video");
		}
		this.toggleSender({ type : "video", enabled : true });
		this.videoDisabled = false;
		callback(true);
	};

	/*!
	 * Start screen share
	 */
	SipPhone.prototype.screenShare = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("screenShare");
			return callback(false, "Can not share screen on: " + this.currentState);
		}
		else if (!this.hasVideo()) {
			return callback(false, "has not video");
		}
		
		let self = this;
		var switchToShareScreen = function(callback) {
		    var session = self.m_userAgent.session;
		    var localStream = new MediaStream();
		    var pc = session.sessionDescriptionHandler.peerConnection;

		    // TODO: Remove legasy ones
		    if (navigator.getDisplayMedia) {
		        // EDGE, legasy support
		        let screenShareConstraints = { video: true, audio: false };
		        navigator.getDisplayMedia(screenShareConstraints).then(function(newStream) {
		            console.log("navigator.getDisplayMedia");
		            var newMediaTrack = newStream.getVideoTracks()[0];
		            pc.getSenders().forEach(function (RTCRtpSender) {
		                if(RTCRtpSender.track && RTCRtpSender.track.kind === "video") {
		                    console.log("Switching Video Track : "+ RTCRtpSender.track.label + " to Screen");
		                    RTCRtpSender.track.stop();
		                    RTCRtpSender.replaceTrack(newMediaTrack);
		                    localStream.addTrack(newMediaTrack);
		                }
		            });

		            // Set Preview
		            // ===========
		            console.log("Showing as preview...");
		            var localVideo = self.localVideo;
		            localVideo.srcObject = localStream;
		            localVideo.onloadedmetadata = function(/*e*/) {
		                localVideo.play();
		            };
				    // Restore Audio Stream is it was changed
				    if(session.data && session.data.AudioSourceTrack && session.data.AudioSourceTrack.kind === "audio"){
				        pc.getSenders().forEach(function (RTCRtpSender) {
				            if(RTCRtpSender.track && RTCRtpSender.track.kind === "audio") {
				                RTCRtpSender.replaceTrack(session.data.AudioSourceTrack).then(function(){
				                    if(session.data.ismute){
				                        RTCRtpSender.track.enabled = false;
				                    }
				                }).catch(function(error){
				                    console.error(error);
				                });
				                session.data.AudioSourceTrack = null;
				            }
				        });
				    }
				    // Agrego evento para notificar al phone_client, por si termina de compartir con el boton nativo del browser
				    newStream.oninactive = function() {
				    	self.emit("stopscreenshare");
				    };

				    callback(true);
		        }).catch(function (error) {
		            console.error("Error on getUserMedia");
		            callback(false, "Error on getUserMedia");
		            console.error(error);
		            return;		            
		        });
		    } 
		    else if (navigator.mediaDevices.getDisplayMedia) {
		        // New standard
		        let screenShareConstraints = { video: true, audio: false, displaySurface: ['application', 'browser', 'monitor', 'window'], };
		        navigator.mediaDevices.getDisplayMedia(screenShareConstraints).then(function(newStream) {
		            console.log("navigator.mediaDevices.getDisplayMedia");
		            var newMediaTrack = newStream.getVideoTracks()[0];
		            pc.getSenders().forEach(function (RTCRtpSender) {
		                if(RTCRtpSender.track && RTCRtpSender.track.kind === "video") {
		                    console.log("Switching Video Track : "+ RTCRtpSender.track.label + " to Screen");
		                    RTCRtpSender.track.stop();
		                    RTCRtpSender.replaceTrack(newMediaTrack);
		                    localStream.addTrack(newMediaTrack);
		                }
		            });

		            // Set Preview
		            // ===========
		            console.log("Showing as preview...");
		            var localVideo = self.localVideo;
		            localVideo.srcObject = localStream;
		            localVideo.onloadedmetadata = function(/*e*/) {
		                localVideo.play();
		            };
				    // Restore Audio Stream is it was changed
				    if(session.data && session.data.AudioSourceTrack && session.data.AudioSourceTrack.kind === "audio"){
				        pc.getSenders().forEach(function (RTCRtpSender) {
				            if(RTCRtpSender.track && RTCRtpSender.track.kind === "audio") {
				                RTCRtpSender.replaceTrack(session.data.AudioSourceTrack).then(function(){
				                    if(session.data.ismute){
				                        RTCRtpSender.track.enabled = false;
				                    }
				                }).catch(function(error){
				                    console.error(error);
				                });
				                session.data.AudioSourceTrack = null;
				            }
				        });
				    }
				    // Agrego evento para notificar al phone_client, por si termina de compartir con el boton nativo del browser
				    newStream.oninactive = function() {
				    	self.emit("stopscreenshare");
				    };
				    
				    callback(true);
		        }).catch(function (error) {
		            console.error("Error on getUserMedia");
		            console.error(error);
		            callback(false, "Error on getUserMedia");
		            return;		            
		        });
		    } 
		    else {		    	
		        // Firefox, apparently
		        let screenShareConstraints = { video: { mediaSource: 'screen' }, audio: false };
		        navigator.mediaDevices.getUserMedia(screenShareConstraints).then(function(newStream) {
		            console.log("navigator.mediaDevices.getUserMedia");
		            var newMediaTrack = newStream.getVideoTracks()[0];
		            pc.getSenders().forEach(function (RTCRtpSender) {
		                if(RTCRtpSender.track && RTCRtpSender.track.kind === "video") {
		                    console.log("Switching Video Track : "+ RTCRtpSender.track.label + " to Screen");
		                    RTCRtpSender.track.stop();
		                    RTCRtpSender.replaceTrack(newMediaTrack);
		                    localStream.addTrack(newMediaTrack);
		                }
		            });

		            // Set Preview
		            console.log("Showing as preview...");
		            var localVideo = self.localVideo;
		            localVideo.srcObject = localStream;
		            localVideo.onloadedmetadata = function(/*e*/) {
		                localVideo.play();
		            };
				    // Restore Audio Stream is it was changed
				    if(session.data && session.data.AudioSourceTrack && session.data.AudioSourceTrack.kind === "audio"){
				        pc.getSenders().forEach(function (RTCRtpSender) {
				            if(RTCRtpSender.track && RTCRtpSender.track.kind === "audio") {
				                RTCRtpSender.replaceTrack(session.data.AudioSourceTrack).then(function(){
				                    if(session.data.ismute){
				                        RTCRtpSender.track.enabled = false;
				                    }
				                }).catch(function(error){
				                    console.error(error);
				                });
				                session.data.AudioSourceTrack = null;
				            }
				        });
				    }
				    // Agrego evento para notificar al phone_client, por si termina de compartir con el boton nativo del browser
				    newStream.oninactive = function() {
				    	self.emit("stopscreenshare");
				    };
				    
				    callback(true);
		        }).catch(function (error) {
		            console.error("Error on getUserMedia");
		            console.error(error);
		            callback(false, "Error on getUserMedia");
		            return;
		        });
		    }
		};
		switchToShareScreen(callback);		
	};

	/*!
	 * Stop screen share
	 */
	SipPhone.prototype.StopScreenShare = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("screenShare");
			return callback(false, "Can not share screen on: " + this.currentState);
		}
		else if (!this.hasVideo()) {
			return callback(false, "has not video");
		}
		
		let self = this;
		var session = this.m_userAgent.session;
		var localStream = new MediaStream();
		var pc = session.sessionDescriptionHandler.peerConnection;
		navigator.mediaDevices.getUserMedia({video: true}).then(function(newStream) {
			var newMediaTrack = newStream.getVideoTracks()[0];
			pc.getSenders().forEach(function (RTCRtpSender) {
				if(RTCRtpSender.track && RTCRtpSender.track.kind === "video") {
					console.log("Switching Video Track : Screen to "+ RTCRtpSender.track.label);
					RTCRtpSender.track.stop();
					RTCRtpSender.replaceTrack(newMediaTrack);
					localStream.addTrack(newMediaTrack);
				}
			});
					
			self.localVideo.srcObject = localStream;
			callback(true);
		});
	};		

	/*!
	 * Switch the current call conference to consultation_call 
	 */
	SipPhone.prototype._switch = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("switch");
			return callback(false, "Can not switch on: " + this.currentState);
		}
		this.switched = true;
		callback(true);
	};

	/*!
	 * Switch the conference consultation_call to call 
	 */
	SipPhone.prototype.unswitch = function(callback = $.noop) {
		if (this.currentState !== PhoneStates.ACTIVE) {
			this.__invalidState("unswitch");
			return callback(false, "Can not unswitch on: " + this.currentState);
		}
		this.switched = false;
		callback(true);
	};
	
	SipPhone.prototype.reinvite = function(callback = $.noop) {
		this.m_userAgent.session.invite()
			.then(() => {
				callback(true);
			})
			.catch((error) => {
				this.__emitError("failed to reinvite");
				callback(false, "failed to reinvite");
				console.error(error);
			});
	};
	
	SipPhone.prototype.refreshVideo = function(callback = $.noop) {
		if (!this.hasVideo()) {
			return callback(true);
		}
		this.reinvite(callback);
	};

	SipPhone.prototype.reconnect = function() {
		var self = this;
		setTimeout(function() {
			self.logger.debug("SIPPhone: Reconnecting");
			self.connect();
		}, 1000);
	};

	/*****************************/
	/**    PRIVATE PROPERTY     **/
	/**                         **/
	/**     NO TRESPASSING      **/
	/**                         **/
	/**        KEEP OUT         **/
	/**                         **/
	/**        VIOLATORS        **/
	/**         WILL BE         **/
	/**        PROSECUTED       **/
	/*****************************/

	SipPhone.prototype.__invalidState = function(action) {
		var message = "Cannot " + action + " while " + this.currentState;
		this.logger.warn("SIPPhone: " + message);
		//this.logger.warn(message);
		this.emit("warning", message, action);
		return false;
	};

	SipPhone.prototype.__permissionPending = function(/*action*/) {
		var message = "You must gran permission to use microphone";
		this.logger.warn("SIPPhone: " + message);
		//this.logger.warn(message);
		this.emit("warning", message, 'microphone');
		return false;
	};

	SipPhone.prototype.__emitError = function(message) {
		this.logger.error("SIPPhone: " + message);
		//this.logger.error(arguments);
		this.emit("error", message);
	};
	
	SipPhone.prototype.__checkPermissions = function(callback) {
		let self = this;
		if (self.m_hasPermission) {
			return callback();
		}
		else if (!window.hasOwnProperty("navigator") || !navigator.getUserMedia) {
			return callback();
		}
		navigator.getUserMedia(
			{ 
				audio : true 
			},
			function() {
				self.m_hasPermission = true;
				callback();
			},
			function() {
				self.m_hasPermission = false;
				let error = {
					description : "<%=_T('Microphone permissions are required to make and receive phone calls.')%>"
				};
				callback(error);
			}
		);
	};

	SipPhone.prototype.__init = function(callback) {
		let self = this;
		self.logger.debug("SIPPhone: Init");
		if (self.m_initialized) {
			self.logger.debug("SIPPhone: Init - already initialized");
			return callback();
		}
		self.m_initialized = true;
		callback();
	};

	/*!
	 * Initializes the SIP Stack, connects to the server but does
	 * not send anything
	 */
	SipPhone.prototype.__connect = function() {
		var self = this;
		// initialize lib
		self.__init(function(error) {
			if (error) {
				self.__setState(PhoneStates.IDLE);
				return self.__emitError("<%=_T('Error initializing phone')%><%=_T(':')%> " + error.description);
			}
			// check permissions
			self.__checkPermissions(function(error) {
				if (error) {
					self.__setState(PhoneStates.IDLE);
					return self.__emitError(error.description);
				}
				if (self.m_userAgent) {
					self.m_userAgent.attemptReconnection();
				}
				else {
					self.__createUser();
				}
			});
		});
	};
	
	SipPhone.prototype.__createUser = function() {
		// create user agent
		this.m_userAgent = new SIP.Web.SimpleUser(this.rtmpServer, {
			aor : `sip:${this.name}@${this.rtmpServer}`,
			userAgentOptions: {
				displayName : this.name,
				authorizationUsername : this.name,
				authorizationPassword : this.password,
				transportOptions : {
					server: `wss://${this.rtmpServer}:8089/ws`
				},
				logLevel : 'debug', // debug | log | warn | error
				//hacks for asterisk
				contactParams: { transport: "wss" },
				hackIpInContact : true,
				reconnectionDelay : 1,
				reconnectionAttempts : Number.MAX_VALUE,
				sessionDescriptionHandlerFactoryOptions : {
					iceGatheringTimeout: 500, // Currently, the smallest allowed value
					peerConnectionConfiguration: {
						iceServers : inConcert.app.getSettings().iceServers
					}
				}
			},
			delegate : {
				onServerConnect : () => {
					//phone connected
					this.__setState(PhoneStates.CONNECTED);
					this.emit("connected");
				},
				onServerDisconnect : () => {
					//phone disconnected
					this.__setState(PhoneStates.IDLE);
					this.isRegistered = false;
				},
				onCallReceived : () => this.__onInvite(),
				onCallAnswered : () => this.__onAnswered(),
				onCallHangup : () => this.__onHangup(),
				onMessageReceived : (message) => this.__onMessageReceived(message),
				onRegistered: () => {
					this.isRegistered = true;
				}
			},
			registererOptions: {
				expires : 30
			},
			media: {
				local: {
					video: this.localVideo
				},
				remote: {
					audio: this.remoteAudio,
					video: this.remoteVideo
				}
			}
		});
		// connect and register
		this.m_userAgent.connect().then(() => {
			this.m_userAgent.register();
		})
		.catch(_ => {
			//Error connecting
			this.__setState(PhoneStates.IDLE);
		});
	};
	
	SipPhone.prototype.__onInvite = function() {
		const session = this.m_userAgent.session;
		const name = session.remoteIdentity.friendlyName;
		this.logger.debug(`SIPPhone: invitation from ${
			name}, current state: ${this.currentState}`);
		//change state
		this.__setState(PhoneStates.INBOUND);		
		//emit invited
		this.emit("invited", {
			id: name
		});
	};

	SipPhone.prototype.isUserRegistered = function() {
		return this.isRegistered;
	};
	
	SipPhone.prototype.__onAnswered = function() {
		this.logger.debug(`SIPPhone: call answered, current state: ${
			this.currentState}`);
		this.__setState(PhoneStates.ACTIVE);
		this.setMaxBitrate();
		this.emit("accepted");
	};
	
	SipPhone.prototype.__onHangup = function() {
		this.logger.debug(`SIPPhone: call hangup, current state: ${
			this.currentState}`);
		this._cleanCallState();
		this.cleanupMedia(false);
		const previousState = this.__setState(PhoneStates.CONNECTED);
		switch(previousState) {
			case PhoneStates.OUTBOUND:
				this.emit("rejected");
				break;
			case PhoneStates.INBOUND:
			case PhoneStates.ACTIVE:
				this.emit("byed");
				break;
			default:
				this.logger.warn(
					`SIPPhone: Session terminated while in state ${
						previousState}`);
				break;
		}
	};

	SipPhone.prototype.__onMessageReceived = function(text) {
		this.logger.debug(`SIPPhone: message received : ${text}`);
		let message;
		try {
			message = JSON.parse(text);
		}
		catch(error) {
			console.log(error);
			return; //ignore non json message
		}
		message.msgtype = message.msgtype || "information";
		switch(message.msgtype) {
			case "notification":
				this._notificationReceived(message.message);
				break;
			case "information": //processed in phone_client
				this.emit("message", text);
				break;
			default:
				this.emit("message", text);
		}
	};

	SipPhone.prototype._notificationReceived = function(/*notification*/) {
	//	switch(notification.event) {
	//		case "ringing":
	//			this.emit("ringing");
	//			break;
	//		default:
	//			break;
	//	}
	};

	SipPhone.prototype.__setState = function(state) {
		var previousState = this.currentState;
		this.logger.debug('SIPPhone: Change state = ' + state + ', previous = ' + previousState);
		this.currentState = state;
		return previousState;
	};
	
	SipPhone.prototype.hasVideo = function() {
		const inviteRequest = this.m_userAgent.session.incomingInviteRequest;
		return inviteRequest && 
			inviteRequest.message.body.indexOf("m=video") >= 0;
	};
	
	SipPhone.prototype.toggleSender = function(options) {
		const track = options.type === "video" ? this.m_userAgent.localVideoTrack : this.m_userAgent.localAudioTrack;
		if (track) {
			track.enabled = options.enabled;
		}
		else {
			this.logger.warn(`SIPPhone:toggleSender ${
				options.type} track not found`);
			return false;
		}
	};
	
	SipPhone.prototype.setMaxBitrate = function() {
		if (!this.hasVideo()) {
			return;
		}
		const description = this.m_userAgent.session.sessionDescriptionHandler;
		const senders = description && description.peerConnection &&
			description.peerConnection.getSenders() || [];
		senders.forEach((sender) => {
			const type = sender.track && sender.track.kind;
			if (type !== "video") {
				return;
			}
			const bitrate = this.bitrate;
			const parameters = sender.getParameters() || {};
			parameters.encodings = parameters.encodings || [];
			if (!parameters.encodings.length) {
				parameters.encodings.push({});
			}
			parameters.encodings[0].maxBitrate = bitrate;
			sender
				.setParameters(parameters)
				.then(() => {
					this.logger.debug(`RTCP bitrate for ${type} set to ${
						bitrate}`); 
				})
				.catch(e => {
					this.logger.error(`RTCP bitrate for ${type} set to ${
						bitrate} - ${e}`); 
				});
		});
	};
	
	SipPhone.prototype.getPhoneState = function() {
		return this.currentState;
	};
	
	SipPhone.prototype._cleanCallState = function() {
		this.muted = false;
		this.videoDisabled = false;
		this.switched = false;
	};

	SipPhone.prototype.getRTCStatistics = function(statsCallback = $.noop)
	{
		if (this.m_userAgent 
			&& this.m_userAgent.session
			&& this.m_userAgent.session.sessionDescriptionHandler
			&& this.m_userAgent.session.sessionDescriptionHandler.peerConnection
			&& this.m_userAgent.session.sessionDescriptionHandler.peerConnection.getStats)
		{
			var pc = this.m_userAgent.session.sessionDescriptionHandler.peerConnection;
			pc.getStats(null).then(stats => {
				/* Llamo el callback 'statsCallback' */
				statsCallback(true, stats);
			});
		}
		else
		{
			statsCallback(false, null);
		}
	};

	if (!inConcert.SipPhone) {
		inConcert.SipPhone = SipPhone;
		inConcert.PhoneStates = PhoneStates;
	}
}());
