local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local allegro = require "apps.allegro.proxy"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

function MapUris (module, htmlRender, prefix, appLocation)
	m_htmlRender = htmlRender
	m_appLocation = appLocation
end

function HasIntegration (vcc_id) return allegro.HasIntegration(vcc_id) end

local function __UpdateValues (groups_omnichannel, vcc_id)

	local function HasPermissionToAction (appId, actionId, full_group_id, grouppermissions)
		--LogDebug("--allegro.groups.__UpdateValues--------------------------------------------------")
		local has_permission_to_action = permissions.RoleHasPermissionToApplicationAction(
			appId,
			actionId,
			full_group_id,
			grouppermissions
		)
		--LogDebug(" ---> app: '%s', actionId: '%s', full_group_id: '%s', HAS_PERMISSION_TO_ACTION: '%s'", appId, actionId, full_group_id, has_permission_to_action)
		return has_permission_to_action
	end

	for group_id, group in pairs(groups_omnichannel) do
		local full_group_id = utils.AppendVcc(group_id, vcc_id)
		-- si tiene acceso a las apps
		group.AllegroIntegration = group.AllegroIntegration or {}
		group.AllegroIntegration.allowed_apps = {
			{
				app_id = "administration",
				has_permission = permissions.RoleHasPermissionToApplication( "administration", full_group_id, group.grouppermissions ),
				--[[
					La tabla "isAllowedByOmnichannelAction" es utilizada para realizar validaciones contra acciones de apps
					de i6 que habilitan o inhabilitan una app de Allegro.

					Ej: con la acción "view_flows" de la app "administration" de i6, se habilita o inhabilita la app "designer" de "Allegro"
				
					Estos valores se definen y calculan en Allegro en:
						\Web\inConcert\apps\omnichannelintegration\templates.lua 
						__GetMappedAppsPermissions
						MakeAllegroGroups
				]]--
				allowed_actions = {
					{
						actionId = "view_flows",
						has_permission = false,
						validate_permission = function(self, appId)
							self.has_permission = HasPermissionToAction(appId, self.actionId, full_group_id, group.grouppermissions)
						end
					},
					{
						actionId = "view_apps_launcher",
						has_permission = false,
						validate_permission = function(self, appId)
							self.has_permission = HasPermissionToAction(appId, self.actionId, full_group_id, group.grouppermissions)
						end
					}
				}
			},
			{
				app_id = "interaction",
				has_permission = permissions.RoleHasPermissionToApplication( "interaction", full_group_id, group.grouppermissions )
			},
			{
				app_id = "reports",
				has_permission = permissions.RoleHasPermissionToApplication( "reports", full_group_id, group.grouppermissions )
			},
			{
				app_id = "supervisor",
				has_permission = permissions.RoleHasPermissionToApplication( "supervisor", full_group_id, group.grouppermissions )
			},
			{
				app_id = "settings",
				has_permission = permissions.RoleHasPermissionToApplication( "settings", full_group_id, group.grouppermissions )
			}
		}

		for _, appInfo in ipairs(group.AllegroIntegration.allowed_apps) do
			for _, actionInfo in ipairs(appInfo.allowed_actions or {}) do
				if actionInfo.validate_permission then
					if appInfo.has_permission then
						actionInfo:validate_permission(appInfo.app_id)
					end
					actionInfo.validate_permission = nil
				end
			end
		end

		-- esto no es necesario que viaje a 5.5
		group.grouppermissions = nil
	end
	return true
end

--[[
  Guarda groups en 5.5
]]--
function Save (groups_omnichannel, vcc_id)

	assert(type(groups_omnichannel) == "table")

	if not HasIntegration(vcc_id) then
		LogWarning("[ALLEGRO.GROUPS:Save] Error saving allegro groups. VCC '%s' Without integration", vcc_id)
		return true -- no es un error
	end

	local success, err_type = __UpdateValues(groups_omnichannel, vcc_id)
	if not success then
		local err = { type = "ALLEGRO_INTEGRATION_ERROR", sub_type = err_type }
		LogWarning("[ALLEGRO.GROUPS:Save] Error saving allegro groups. Err: %s.", json.encode(err))
		return false, err
	end

	local encoded_groups_omnichannel = json.encode(groups_omnichannel)
	LogInfo("[ALLEGRO.GROUPS:Save] encoded groups: %s", encoded_groups_omnichannel)

	local success, allegro_err = allegro.Post(
		vcc_id,
		encoded_groups_omnichannel,
		"administration/groups/save/" -- la URL de 5.5
	)

	local err
	if not success then
		err = { type = "ALLEGRO_INTEGRATION_ERROR", sub_type = allegro_err }
		LogWarning("[ALLEGRO.GROUPS:Save] Error saving allegro groups: %s. Err: %s.", encoded_groups_omnichannel, json.encode(err))
	end

	return success, err
end

--[[
  Elimina los groups en 5.5
]]--
function Delete (groups_omnichannel, vcc_id)

	assert(type(vcc_id) == "string")

	local encoded_groups_omnichannel = json.encode(groups_omnichannel)

	LogInfo("[ALLEGRO.GROUPS:Delete] Deleting allegro groups '%s'", encoded_groups_omnichannel)

	if not HasIntegration(vcc_id) then
		LogWarning("[ALLEGRO.GROUPS:Delete] Error deleting allegro groups. VCC '%s' Without integration", vcc_id)
		return true -- no es un error
	end

	local success, allegro_err = allegro.Post(
		vcc_id,
		encoded_groups_omnichannel,
		"administration/groups/delete/" -- la URL de 5.5
	)

	local err
	if not success then
		err = { type = "ALLEGRO_INTEGRATION_ERROR", sub_type = allegro_err }
		if allegro_err.result == "NOT_FOUND" then
			err.type = "NOT_A_ERROR"
		end
		LogWarning("---> [ALLEGRO.GROUPS:Delete] Error deleting allegro groups: %s. Err: %s.", encoded_groups_omnichannel, json.encode(err))
	end

	return success, err
end
