module(..., package.seeall, inconcert.orbit.application.Decorate)

local utils = require "shared.apps.utils"
local proxy = require "apps.allegro.proxy"
local allegroRepository = require "apps.allegro.controllers.repository"

function LoadPaginatedImports (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE:LoadPaginatedImports] Loading Importations.")

	local success, data = proxy.outboundEngine.LoadImports(encodedInfo)
	if success and data then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:LoadPaginatedImports] Successfully loaded %d of %s Importations.", #data.rows, data.total_rows)
		return true, data
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_LOADING_IMPORTS"
	}
	return false, errType
end

function GetImport (importId)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE:GetImport] Retrieving Import: '%s'.", importId)

	local response, err = proxy.outboundEngine.GetImport(importId)

	if response then
		local import = response.import
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:GetImport] Successfully retrieving %s Import. Info: %s", importId, import)
		return true, import
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_IMPORT"
	}
	LogError("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:GetImport] Error retrieving %s Import. Info: %s", importId, json.encode(errType))
	return false, errType
end

function DeleteImport (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:DeleteImport] Deleting Import. Info: %s", encodedInfo)

	local success, data = proxy.outboundEngine.DeleteImport(encodedInfo)
	--LogDebug("---> DeleteImport success: %s, data: %s", success, data)
	if success then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:DeleteImport] Successfully deleted %s Import.", encodedInfo)
		return true
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_DELETING_IMPORT"
	}
	return false, errType
end

local function AddAditionaInfoToPost (encodedInfo)
	local info = json.decode(encodedInfo)
	-------------------------------------
	-- Se le agrega el "logged user" a la informacion que se va a postear a 5.5, para que en el campo "ContactImportDetail.ImportUser" quede el usuario
	-- que ha realizado la accion en i6 y no el usuario "integrationuser"
	info.__user = inconcert.authUser.userName
	info.__vcc = inconcert.authUser.virtualContactCenter
	encodedInfo = json.encode(info)
	-------------------------------------
	return info, encodedInfo
end

function StartUpload (encodedInfo)
	local info, encodedInfo = AddAditionaInfoToPost(encodedInfo)
	info.FileInfo.fileName = utils.SanitizeFileName(info.FileInfo.fileName)
	encodedInfo = json.encode(info)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:StartUpload] Starting Uploading Contacts File. Info: %s", encodedInfo)
	local vcc = inconcert.authUser.virtualContactCenter
	local repositoryModel = allegroRepository.GetModel()

	local temporalFile = repositoryModel.GetTemporalUploadedFile({
		FileId = info.FileInfo.id
	})

	local success, data
	if temporalFile then
		--[[
			Aca se inovca al repositorio para que este realice directamente el upload derecho contra la app "outboundengine" de 5.5
			y no que pase contra "omnichannel" ya que sino es mucha pasarela y no funciona bien. Se puede revisar, pero hay que darle una vuelta.
		]]--
		success, data = allegroRepository.OutboundEngineUploadImportFile(encodedInfo, temporalFile, vcc)
		if success then
			LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:StartUpload] Successfully Uploaded Contacts File. Info: %s", encodedInfo)
			return true
		end
	else
		LogError("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:StartUpload] Error Starting Uploading Contacts File. Contact File does not exist.", encodedInfo)
	end

	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_UPLOADING_CONTACTS_FILE"
	}
	return false, errType
end

function StartImport (encodedInfo)
	local info, encodedInfo = AddAditionaInfoToPost(encodedInfo)
	info.FileInfo.fileName = utils.SanitizeFileName(info.FileInfo.fileName)
	encodedInfo = json.encode(info)


	LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:StartImport] Starting Imports. Info: %s", encodedInfo)

	--local success, data = proxy.outboundEngine.StartImport(encodedInfo)
	local success, data = allegroRepository.OutboundEngineStartImport(encodedInfo, info.__vcc)

	LogDebug("---> StartImport success: %s, data: %s", success, data)
	if success then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:StartImport] Successfully started %s Imports.", encodedInfo)
		return true
	end

	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_STARTING_IMPORTS"
	}
	return false, errType
end

function CancelImport (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:CancelImport] Canceling Imports. %s", encodedInfo)

	local success, data = proxy.outboundEngine.CancelImport(encodedInfo)
	--LogDebug("---> CancelImport success: %s, data: %s", success, data)
	if success then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:CancelImport] Successfully cancelled %s Imports.", encodedInfo)
		return true
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_CANCELING_IMPORTS"
	}
	return false, errType
end

function DepurateImport (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:DepurateImport] Depurating Imports. %s", encodedInfo)

	local success, data = proxy.outboundEngine.DepurateImport(encodedInfo)
	--LogDebug("---> DepurateImport success: %s, data: %s", success, data)
	if success then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:DepurateImport] Successfully depurated %s Imports.", encodedInfo)
		return true
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_DEPURATING_IMPORTS"
	}
	return false, errType
end

function DownloadFailedContacts (importId, update, fileName)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:DownloadFailedContacts] Downloading Failed Contacts on ImportationId: '%s', Update: '%s', fileName: '%s'.", importId, update, fileName)
	local vcc = inconcert.authUser.virtualContactCenter
	return allegroRepository.OutboundEngineImportDownloadFailedContacts(importId, update, vcc, fileName)
end

function LoadPaginatedImportFormats (encodedInfo)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE:LoadPaginatedImportFormats] Loading Export Layout.")

	local success, data = proxy.outboundEngine.LoadImportFormats(encodedInfo)
	if success and data then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:LoadPaginatedImportFormats] Successfully loaded %d of %s Export Layout.", #data.rows, data.total_rows)
		return true, data
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_LOADING_IMPORT_FORMAT"
	}
	return false, errType
end

function DeleteImportFormat (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:DeleteImportFormat] Deleting Import Format.", encodedInfo)

	local success, data = proxy.outboundEngine.DeleteImportFormat(encodedInfo)
	--LogDebug("---> DeleteImportFormat success: %s, data: %s", success, data)
	if success then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:DeleteImportFormat] Successfully deleted %s Import Format.", encodedInfo)
		return true
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_DELETING_IMPORT_FORMAT"
	}
	return false, errType
end

function LoadImportsFormats ()

	LogInfo("[ALLEGRO.OUTBOUNDENGINE:LoadImportsFormats] Loading Imports Formats.")

	local response, err = proxy.outboundEngine.LoadImportsFormats()

	if response then
		local importsFormats = response.importsFormats
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:LoadImportsFormats] Successfully Loaded Imports Formats. Info: %s", importsFormats)
		return true, importsFormats
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_IMPORTS_FORMATS"
	}
	LogError("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:LoadImportsFormats] Error loading Imports Formats. Info: %s", json.encode(errType))
	return false, errType
end

function GetImportFormat (formatId)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE:GetImportFormat] Retrieving Import Format: '%s'.", formatId)

	local response, err = proxy.outboundEngine.GetImportFormat(formatId)

	if response then
		local importFormat = response.importFormat
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:GetImportFormat] Successfully Retrieved %s Import Format. Info: %s", formatId, importFormat)
		return true, importFormat
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_IMPORT_FORMAT"
	}
	LogError("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:GetImportFormat] Error retrieving %s Import Format. Info: %s", formatId, json.encode(errType))
	return false, errType
end

function SaveImportFormat (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE:SaveImportFormat] Saving Import Format.", encodedInfo)

	local success, data = proxy.outboundEngine.SaveImportFormat(encodedInfo)
	--LogDebug("---> SaveImportFormat success: %s, data: %s", success, data)
	if success then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:SaveImportFormat] Successfully saved %s Import Format.", encodedInfo)
		return true, data
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_SAVING_IMPORT_FORMAT"
	}
	return false, errType
end

function HasSomeImportRunning ()

	LogInfo("[ALLEGRO.OUTBOUNDENGINE:HasSomeImportRunning] Validating if has some import running.")

	local response, err = proxy.outboundEngine.HasSomeImportRunning()
	--LogDebug("---HasSomeImportRunning------> response: %s", json.encode(response))

	if response then
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:HasSomeImportRunning] Successfully. Not has some import running.")
		return true, response
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_HAS_SOME_IMPORT_RUNNING"
	}
	LogError("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:HasSomeImportRunning] Exists another importation in process. Info: %s", json.encode(errType))
	return false, errType
end

function GetUploadedCSVHeader (fileId)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE:GetUploadedCSVHeader] Retrieving CSV Header file '%s'.", fileId)

	local repositoryModel = allegroRepository.GetModel()
	local csvFile = repositoryModel.GetTemporalUploadedFile({
		FileId = fileId
	})

	--LogDebug("---GetUploadedCSVHeader------> csvFile: %s", csvFile)

	if csvFile then
		local csvHeader = utils.ParseCSVHeader(csvFile)
		LogInfo("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:GetUploadedCSVHeader] Successfully retrieved CSV Header: '%s'.", json.encode(csvHeader))
		return true, csvHeader
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = "ERROR_HAS_RETRIEVING_UPLOADED_FILE_INFO"
	}
	LogError("[ALLEGRO.OUTBOUNDENGINE_IMPORTS:GetUploadedCSVHeader] Error retrieving CSV Header file '%s'. Info: %s", fileId, json.encode(errType))
	return false, errType
end
