/*jshint bitwise:true, browser:true, curly:true, eqeqeq:true, noarg:true, noempty:true, strict:true, undef:true, unused:true, esversion: 9 */
/*jshint quotmark:false*/
/*global _, EventEmitter, inConcert, console, $*/
this.inConcert = this.inConcert || {};

(function () {
	"use strict";

	/*!
	 * Possible phone states
	 */
	var PhoneStates = {
		IDLE								: "idle",
		CONNECTING							: "connecting",
		CONNECTED							: "connected",
		OUTBOUND							: "outbound",
		INBOUND								: "inbound",
		ACTIVE								: "active",
		WRAPUP								: "wrapup",
		TRANSFERRING						: "transferring",
		TRANSFERRING_CONNECTED				: "transferring_connected",
		CONFERENCE_CONSULTATION_TRYING		: "conference_consultation_trying",
		CONFERENCE_CONSULTATION_CONNECTED	: "conference_consultation_connected"
	};
	
	var InteractionTypes = {
		INBOUND		: "ENTDIRECTA",
		DIALER		: "SALDIRECTA",
		MANUAL		: "TPOQUECALL",
		TRANSFER	: "TPIDIRCALL",
		VOICEMAIL	: "VOICEMAIL"
	};

	/*!
	 * A wrapper around baragent
	 *
	 * Events:
	 * \li rejected(reason)
	 * \li warning(message)
	 * \li error(message)
	 * \li invited(callerid)
	 * \li connected()
	 * \li connecting()
	 * \li dialing()
	 * \li connected() -> after connecting()
	 * \li accepted()  -> after dialing()
	 * \li cancelled()
	 * \li byed()
	 * \li ringing()
	 */
	function SipPhone(params) {
		// Initial setup
		_.extend(this, {
			// public, handles the state machine
			currentState    : PhoneStates.IDLE,
			recording		: false,
			in_conference	: false,
			in_hold			: false,
			muted			: false,
			switched		: false,
			with_preview	: false,
			preview			: false,
			is_redial		: false,
			
			m_initialized   : false,
			m_callSession   : null,
			m_hasPermission: true
		}, params);

		this.logger = this.logger || {
			debug : _.bind(console.debug, console),
			info  : _.bind(console.info, console),
			warn  : _.bind(console.warn, console),
			error : _.bind(console.error, console)
		};
	}
	inConcert.Utils.inherit(SipPhone, EventEmitter);
	
	SipPhone.prototype.emit = function(eventName, ...args) {
		this.logger.debug("emit " + eventName);
		SipPhone.uber.emit.call(this, eventName, ...args);
	};

	/*!
	 * Connects and registers with Asterisk (aka the SIP server)
	 */
	SipPhone.prototype.connect = function() {
		if (this.currentState !== PhoneStates.IDLE) {
			return this.__invalidState("connect");
		}
		this.__setState(PhoneStates.CONNECTING);
		this.__connect();
	};

	/*!
	 * Disconnects from the SIP server, closes all sessions
	 */
	SipPhone.prototype.disconnect = function() {
		this.__setState(PhoneStates.IDLE);
	};

	/*!
	 * Makes a call
	 */
	SipPhone.prototype.invite = function(params) {
		const self = this;
		const interactionID = params.interactionID;
		const campaignAtVCC = params.campaignAtVCC;
		const destination = params.destination;
		const contact = params.contact;
		const callback = params.callback || $.noop;

		if (self.currentState !== PhoneStates.CONNECTED) {
			self.emit("rejected", "Cannot call while " + this.currentState);
			self.__invalidState("invite");
			return callback(false, "Cannot call while " + this.currentState);
		}
		
		self.__setState(PhoneStates.OUTBOUND);
		self.emit("dialing");
		
		//dialing flag to avoid baragent connected state during dialing
		self.m_dialing = true;		
		inConcert.BaragentActions.makeCall(campaignAtVCC, destination, 
			contact && contact.externalId || null, contact && contact.Name || null, 
			function(ok) {
				self.m_dialing = false;
				if (ok) {
					callback(true);
				}
				else {
					self.__setState(PhoneStates.CONNECTED);
					self.emit("rejected", "Failed to make call");
					callback(false, "Failed to make call");
				}
			},
			interactionID);
	};

	/*!
	 *
	 */
	SipPhone.prototype.returnErr = function(method, errType, callback = $.noop) {
		this.logger.info(`[SipPhone.returnErr] ERROR. Method: ${method}, errType: ${errType}.`);
		return callback(false, errType);
	};

	/*!
	 * Accepts an incoming call
	 */
	SipPhone.prototype.accept = function(callback = $.noop) {
		const self = this;
		if (this.currentState !== PhoneStates.INBOUND) {
			this.__invalidState("accept");
			return this.returnErr(
				"accept/answer",
				"Can not accept on: " + this.currentState,
				callback
			);
		}
		if (this.m_hasPermission === false) {
			this.__permissionPending('accept');
			return this.returnErr(
				"accept/answer",
				"Has not permissions",
				callback
			);
		}
		inConcert.BaragentActions.answer(function(ok) {
			if (ok) {
				return callback(true, null);
			}
			return self.returnErr(
				"accept/answer",
				"Error doing answer",
				callback
			);
		});
	};

	/*!
	 * Rejects an incoming call
	 */
	SipPhone.prototype.reject = function(reason, callback = $.noop) {
		const self = this;
		if (this.currentState !== PhoneStates.INBOUND) {
			this.__invalidState("reject");
			return this.returnErr(
				"reject",
				"Can not reject on: " + this.currentState,
				callback
			);
		}
		inConcert.BaragentActions.hangup(function(ok) {
			if (ok) {
				return callback(true);
			}
			self.__invalidState("reject");
			return self.returnErr(
				"reject",
				"Error doing reject",
				callback
			);
		});
	};
	
	SipPhone.prototype.record = function(callback = $.noop) {
		inConcert.BaragentActions.startRecording(function(ok) {
			callback(ok, !ok? "Error starting record" : null);
		});
	};
	
	SipPhone.prototype.stopRecord = function(callback = $.noop) {
		inConcert.BaragentActions.stopRecording(function(ok) {
			callback(ok, !ok? "Error stopping record" : null);
		});
	};
	
	SipPhone.prototype.acceptPreview = function(callback = $.noop) {
		inConcert.BaragentActions.acceptPreview(function(ok) {
			callback(ok, !ok? "Error accepting preview" : null);
		});
	};
	
	SipPhone.prototype.cancelPreview = function(callback = $.noop) {
		inConcert.BaragentActions.cancelPreview(function(ok) {
			callback(ok, !ok? "Error canceling preview" : null);
		});
	};
	
	SipPhone.prototype.confirmTransfer = function(callback = $.noop) {
		inConcert.BaragentActions.transferCommit(function(ok) {
			callback(ok, !ok? "Error confirming transfer" : null);
		});
	};
	
	SipPhone.prototype.cancelTransfer = function(callback = $.noop) {
		inConcert.BaragentActions.transferCancel(function(ok) {
			callback(ok, !ok? "Error canceling transfer" : null);
		});
	};
	
	SipPhone.prototype.confirmConference = function(callback = $.noop) {
		inConcert.BaragentActions.conferenceCommit(function(ok) {
			callback(ok, !ok? "Error confirming conference" : null);
		});
	};
	
	SipPhone.prototype.cancelConference = function(callback = $.noop) {
		inConcert.BaragentActions.conferenceCancel(function(ok) {
			callback(ok, !ok? "Error canceling conference" : null);
		});
	};
	
	SipPhone.prototype.replayVoicemail = function(callback = $.noop) {
		inConcert.BaragentActions.replayVoicemail(function(ok) {
			callback(ok, !ok? "Error replaying voicemail" : null);
		});
	};

	SipPhone.prototype.redial = function(callback) {
		const self = this;
		let number = "";
		//needed for callback's redial (when client never answer)
		//it apply for allegro (and is a baragent's bug)
		const statusInteraction = inConcert.BaragentPolling.getStatusInteraction();
		if (statusInteraction && statusInteraction.workitemType === "TPOQUECALL") {
			const interaction = inConcert.app.getInteractionManager().getChannelClient("CALL").getInteraction();
			number = interaction ? interaction.getRedialNumber() : "";
		}
		this.__setState(PhoneStates.OUTBOUND);
		inConcert.BaragentActions.redial(number, function(ok) {
			if (ok) {
				return callback(true, null);
			}
			return self.returnErr(
				"redial",
				"Error doing redial",
				callback
			);
		});
	};

	/*!
	 * Hangs-up the active call
	 */
	SipPhone.prototype.bye = function(callback = $.noop) {
		var self = this;
		if (!_.contains([PhoneStates.ACTIVE, PhoneStates.OUTBOUND], this.currentState)) {
			self.__invalidState("hangup");
			return callback(false, "Can not bye on: " + this.currentState);
		}
		self.m_doingHangup = true;
		inConcert.BaragentActions.hangup(function(ok) {
			if (ok) {
				callback(true);
			}
			else {
				self.__invalidState("hangup");
				self.m_doingHangup = false;
				callback(false, "Error doing bye");
			}
		});
	};

	/*!
	 * Sends a DTMF digit
	 */
	SipPhone.prototype.sendDigit = function(digit, callback) {
		var self = this;
		if (self.currentState !== PhoneStates.ACTIVE && self.currentState !== PhoneStates.TRANSFERRING_CONNECTED && self.currentState !== PhoneStates.CONFERENCE_CONSULTATION_CONNECTED) {
			self.__invalidState("dtmf");
			return callback(false, "Can not send dtmf on: " + this.currentState);
		}
		if ( "1234567890ABCD*#".indexOf(digit) >= 0 ) {
			inConcert.BaragentActions.sendDigits(digit, function(ok) {
				if (!ok) {
					self.__invalidState("dtmf");
				}
				callback(ok);
			});
		}
	};

	/*!
	 * Puts the current call on hold
	 */
	SipPhone.prototype.hold = function(callback) {
		var self = this;
		if (this.currentState !== PhoneStates.ACTIVE) {
			self.__invalidState("hold");
			return callback(false, "Can not hold on: " + this.currentState);
		}
		inConcert.BaragentActions.hold(function(ok) {
			if (!ok) {
				self.__invalidState("hold");
			}
			callback(ok);
		});
	};

	/*!
	 * Resumes the current call
	 */
	SipPhone.prototype.resume = function(callback) {
		var self = this;
		//when transferring, baragent auto unhold the call
		if (this.currentState === PhoneStates.TRANSFERRING) {
			return callback(true);
		}
		if (this.currentState !== PhoneStates.ACTIVE) {
			self.__invalidState("resume");
			return callback(false, "Can not resume on: " + this.currentState);
		}
		inConcert.BaragentActions.unhold(function(ok) {
			if (!ok) {
				self.__invalidState("resume");
			}
			callback(ok);
		});
	};

	/*!
	 * Puts the current call on mute
	 */
	SipPhone.prototype.mute = function(callback = $.noop) {
		inConcert.BaragentActions.mute(function(ok) {
			callback(ok, !ok? "Error doing mute" : null);
		});
	};

	/*!
	 * Removes the current call on mute
	 */
	SipPhone.prototype.unmute = function(callback = $.noop) {
		inConcert.BaragentActions.unmute(function(ok) {
			callback(ok, !ok? "Error doing unmute" : null);
		});
	};

	/*!
	 * Switch the current call conference to consultation_call 
	 */
	SipPhone.prototype._switch = function(callback = $.noop) {
		var self = this;
		inConcert.BaragentActions._switch(function(ok) {
			self.switched = true;
			self.emit("switched", self.switched);
			callback(ok, !ok? "Error doing switch" : null);
		});
	};

	/*!
	 * Switch the conference consultation_call to call
	 */
	SipPhone.prototype.unswitch = function(callback = $.noop) {
		var self = this;
		inConcert.BaragentActions.unswitch(function(ok) {
			self.switched = false;
			self.emit("switched", self.switched);
			callback(ok, !ok? "Error doing unswitch" : null);
		});
	};	

	/*****************************/
	/**    PRIVATE PROPERTY     **/
	/**                         **/
	/**     NO TRESPASSING      **/
	/**                         **/
	/**        KEEP OUT         **/
	/**                         **/
	/**        VIOLATORS        **/
	/**         WILL BE         **/
	/**        PROSECUTED       **/
	/*****************************/

	SipPhone.prototype.__invalidState = function(action) {
		var message = "Cannot " + action + " while " + this.currentState;
		this.emit("warning", message, action);
		return false;
	};

	SipPhone.prototype.__permissionPending = function() {
		var message = "You must gran permission to use microphone";
		this.emit("warning", message, 'microphone');
		return false;
	};

	SipPhone.prototype.__emitError = function(message) {
		this.emit("error", message);
	};
	
	//TODO: Implementame
	SipPhone.prototype.__parsePhoneNumber = function(number) {
		if (!number) {
			return "";
		}
		return number.replace(/[\[|\]|-]/g, "");
	};
	
	SipPhone.prototype.__onStatusChange = function(status) {
		var self = this;
		var wasRedial = self.is_redial;
		self.is_redial = inConcert.BaragentPolling.isRedial();
		
		function loadSession() {
			if (!hasInteraction()) {
				self.m_callSession = null;
				return;
			}
			var interaction = status.interaction;
			var campaignId = inConcert.app.getDomainId(interaction.campaignId);
			var accounts = inConcert.app.getCampaignAddresses(campaignId, "CALL");
			var accountId = accounts && accounts.length? accounts[0].id : null;
			var session = {
				id : interaction.id,
				campaign : campaignId,
				account : accountId,
				type : interaction.workitemType,
				callerId : self.__parsePhoneNumber(interaction.callerId),
				calledId : self.__parsePhoneNumber(interaction.dnis || interaction.calledId),
				contactId : "",
				contactName : interaction.contactName || ""
			};
			if ($.inArray(session.type, [InteractionTypes.INBOUND, InteractionTypes.VOICEMAIL, InteractionTypes.TRANSFER]) >= 0) {
				//reencolado de llamada saliente
				if (session.type === InteractionTypes.TRANSFER && !interaction.dnis) {
					session.initiative = "OUTBOUND";
					session.contactName = session.contactName || session.calledId;
				}
				else {
					//transferencia de llamada saliente
					if (session.type === InteractionTypes.INBOUND && !interaction.dnis) {
						if (session.contactName && session.callerId.indexOf(session.contactName) >= 0) {
							session.callerId = session.contactName;
						}
					}
					//en algunos casos como los voicemail reencolados muchas veces, el callerId llega vacio
					if (!session.callerId && interaction.contactName.indexOf("Origen : ") >= 0) {
						session.callerId = interaction.contactName.replace("Origen : ", "");
					}
					session.initiative = "INBOUND";
				}
			}
			else {
				session.initiative = "OUTBOUND";
				session.contactName = session.contactName || session.calledId;
			}
			//en caso de telefonos ocultos se setea un contactName unknown porque de arriba se espera tener algo que mostrar
			if (!session.contactName) {
				session.contactName = "<%=_T('unknown')%>";
			}
			$.each(status.activeInteractions, function(_, activeInteraction) {
				if (activeInteraction.id === session.id) {
					session.startDate = activeInteraction.startDate;
				}
			});
			
			self.m_callSession = session;
		}
		
		function getInteractionState() {
			if (!hasInteraction()) {
				return PhoneStates.CONNECTED;
			}
			var interaction = status.interaction;
			if (interaction.interactionState === "TAKED") {
				switch(interaction.phoneState) {
					case "TRANSFERRING":
						return PhoneStates.TRANSFERRING;
					case "TRANSFERRING_CONNECTED":
						return PhoneStates.TRANSFERRING_CONNECTED;
					case "CONFERENCE_CONSULT_CALL":
						return PhoneStates.CONFERENCE_CONSULTATION_TRYING;
					case "CONFERENCE_CONSULT_CALL_CONNECTED":
						return PhoneStates.CONFERENCE_CONSULTATION_CONNECTED;
					default:
						return PhoneStates.ACTIVE;
				}
			}
			if (interaction.interactionState === "ENDED" || interaction.interactionState === "FINISHED") {
				if (interaction.phoneState === "REDIAL") {
					return PhoneStates.OUTBOUND;
				}
				return PhoneStates.WRAPUP;
			}
			if (interaction.interactionState === "LOCKED") {
				return interaction.workitemType === InteractionTypes.MANUAL? PhoneStates.OUTBOUND : PhoneStates.INBOUND;
			}
			//existen algunos estados intermedios al comenzar la llamada (todos se interpretan como ringing)
			//y otros al cerrarla (todos se interpretan como active hasta que termina y queda en ENDED)
			if (self.currentState === PhoneStates.ACTIVE) {
				return PhoneStates.ACTIVE;
			}
			return interaction.workitemType === InteractionTypes.MANUAL? PhoneStates.OUTBOUND : PhoneStates.INBOUND;
		}
		
		function newInbound() {
			self.logger.info("[newInbound] emit: 'invited'");
			self.emit("invited", self.m_callSession);
		}

		function newOutbound() {
			self.emit("outbound", self.m_callSession);
		}

		function emitEvent(previousState, newState) {
			var event = "";
			switch(newState) {
				//call end
				case PhoneStates.CONNECTED:
					switch(previousState) {
						case PhoneStates.OUTBOUND:
							event = self.m_doingHangup? "cancelled" : "rejected";
							break;
						case PhoneStates.INBOUND:
						case PhoneStates.ACTIVE:
						case PhoneStates.WRAPUP:
							event = "byed";
							break;
						default:
							break;
					}
					self.m_doingHangup = false;
					break;
				case PhoneStates.INBOUND:
					/*-1---------------------------
						OK lo nuevo
					*/
					 if (previousState === PhoneStates.WRAPUP) {
						//if (true) {  // SI usamos esto en lugar del if de la linea anterior, podemos generar siempre como que estuviera en WRAPUP.
						/*
							Si Emitimos un "onByed" hacia el "PhoneClient", el metodo "newInboud()" NO se puede ejecutar de forma inmediata
								sino que debe ejecutarse una vez que hayan culminado las ejecuciones asincrónicas de la funcion "onByed".
							Especificamente una vez que la ventana se ha cerrado (si es que se estaba cargando).
							Esto queda escuchando con subscribe del objeto "messagebus" bajo la clave 'live_phone_client_window/interaction/EL_ID_DE_LA_INTERACTION/windows_closed/'.
							Una vez que ha culminado de cerrar la ventana, procede a invocar al callback, que ahora si ejecuta al metodo "newInboud()"

							Ref Issue #18744
						*/
						self.logger.info("[emitEvent] 1- PhoneStates.INBOUND: 'inbound' (previousState == 'wrapup')");
						self.emit("byed", undefined, {
							invokedFrom: "SipPhone state: " + PhoneStates.INBOUND,
							callback   : () => {
								// este callback se ejecuta cuando se cierra el phoneclient.
								self.logger.info("[emitEvent] 1.2- (byed -> callback) PhoneStates.INBOUND: 'inbound' (previousState == 'wrapup')");
								newInbound();
							}
						});
						self.logger.info("[emitEvent] 1.1-  PhoneStates.INBOUND: 'inbound' (previousState == 'wrapup')");
						newInbound();
					}
					else {
						self.logger.info("[emitEvent] 2-  PhoneStates.INBOUND: 'inbound' (previousState == '" + previousState + "')");
						newInbound();
					}
					// mev voy
					return;

					//----------------------------

														//-2---------------------------
														// ERROR - LO VIEJO LO DE SIEMPRE
														// TODO: BORRAME CUANDO TODO ESTE OK!!!
														//if (previousState === PhoneStates.WRAPUP) {
														// 	self.emit("byed");
														// }
														// newInbound()
														// return;
														//----------------------------

															// //-3---------------------------
															// // FORZAR EL ERROR CON LO VIEJO
															// // TODO: BORRAME CUANDO TODO ESTE OK!!!
															// newInbound()
															// if (true) { // para forzar siempre el error
															// 	self.emit("byed");
															// }
															// newInbound()
															// return;
															// //----------------------------
				case PhoneStates.OUTBOUND:
					event = "dialing";
					break;
				case PhoneStates.ACTIVE:
					if (previousState === PhoneStates.CONNECTED) {
						newInbound();
					}
					if (wasRedial) {
						event = "redial_accepted";
					}
					else {
						event = "accepted";
					}
					break;
				case PhoneStates.WRAPUP:
					event = "wrapup";
					break;
				case PhoneStates.TRANSFERRING:
					event = "transferring";
					break;
				case PhoneStates.TRANSFERRING_CONNECTED:
					event = "transferring_connected";
					break;
				case PhoneStates.CONFERENCE_CONSULTATION_TRYING:
					event = "conference_consultation_trying";
					break;
				case PhoneStates.CONFERENCE_CONSULTATION_CONNECTED:
					event = "conference_consultation_connected";
					break;
				default:
					break;
			}
			if (event) {
				self.emit(event, self.m_callSession, {
					invokedFrom  : "SipPhone default",
					previousState: previousState,
					newState     : newState
				});
			}
		}
		
		if (!status) {
			return;
		}
		
		function hasInteraction() {
			return status.interaction && status.interaction.id && true || false;
		}
		
		function setRecording() {
			if (!hasInteraction()) {
				self.recording = false;
				return;
			}
			var old = self.recording;
			self.recording = status.interaction.recording === "1";
			if (old !== self.recording) {
				self.emit("recording", self.recording);
			}
		}
		
		function setInConference() {
			if (!hasInteraction()) {
				self.in_conference = false;
				return;
			}
			var old = self.in_conference;
			self.in_conference = status.interaction.phoneState === "CONFERENCING";
			if (old !== self.in_conference) {
				self.emit("in_conference", self.in_conference);
			}
		}
		
		function setInHold() {
			if (!hasInteraction()) {
				self.in_hold = false;
				return;
			}
			var old = self.in_hold;
			self.in_hold = status.interaction.phoneState === "HOLD";
			if (old !== self.in_hold) {
				self.emit("in_hold", self.in_hold);
			}
		}
		
		function setMuted() {
			if (!hasInteraction()) {
				self.muted = false;
				return;
			}
			var old = self.muted;
			self.muted = status.agent.phoneMute === "1";
			if (old !== self.muted) {
				self.emit("muted", self.muted);
			}
		}
		
		function setWithPreview() {
			if (!hasInteraction()) {
				self.with_preview = false;
				return;
			}
			self.with_preview = status.interaction && status.interaction.hasPreview === "1" || false;
		}
		
		function setPreview() {
			if (!hasInteraction()) {
				self.preview = false;
				return;
			}
			var old = self.preview;
			self.preview = status.buttons && status.buttons.PreviewAccept && 
						  status.buttons.PreviewAccept.enabled === "1" || false;
			if (old !== self.preview) {
				self.emit("preview", self.preview);
			}
		}
		
		function setRedial() {
			if (!hasInteraction()) {
				return;
			}
			if (!wasRedial && self.is_redial) {
				self.emit("redial");
			}
		}
		
		function checkChangedInteraction() {
			if (!self.is_redial && hasInteraction() && self.m_callSession && status.interaction.id !== self.m_callSession.id) {
				self.logger.debug("[checkChangedInteraction] changed interaction detected");
				self.emit("byed", undefined, {
					invokedFrom  : "SipPhone checkChangedInteraction"
				});
				self.__setState(PhoneStates.CONNECTED);
			}
		}
		
		var newState = getInteractionState();
		if (self.m_dialing || status.agent.makingCall) {
			if (newState === PhoneStates.CONNECTED) {
				self.logger.debug("Dialing call: waiting for baragentstatus state change");
				return;
			}
			else {
				self.m_dialing = false;
				self.logger.debug("Dialing call: state changed to " + PhoneStates.OUTBOUND);
			}
		}
		
		setRecording();
		setInConference();
		setInHold();
		setMuted();
		setWithPreview();
		setPreview();
		setRedial();
		
		var previousState = self.__setState(newState);
		self.logger.debug("new State " + newState + " and previous state " + previousState);
		//start outbound interaction
		if (previousState === PhoneStates.OUTBOUND && !self.m_callSession) {
			loadSession();
			if (self.m_callSession) {
				newOutbound();
			}
		}
		if (newState === previousState) {
			//check changed interaction
			checkChangedInteraction();
			//check popup
			if (newState === PhoneStates.ACTIVE || newState === PhoneStates.WRAPUP || self.with_preview) {
				if (status.interaction.popup) {
					self.emit("check_allegro_popup", status.interaction.popup);
				}
				else {
					self.logger.debug("Popup info not found for interaction " + status.interaction.id);
				}
			}
			return;
		}
		loadSession();
		emitEvent(previousState, newState);
	};

	SipPhone.prototype.__init = function(callback) {
		if (this.m_initialized) {
			return callback();
		}
		
		inConcert.BaragentPolling.addObserver("SIP_PHONE", this, "__onStatusChange");
		this.m_initialized = true;
		callback();
	};

	/*!
	 * Initializes the SIP Stack, connects to the server but does
	 * not send anything
	 */
	SipPhone.prototype.__connect = function() {
		var self = this;
		self.emit("connecting");
		self.__init(function() {
			self.__setState(PhoneStates.CONNECTED);
			self.emit("connected");
		});
	};

	SipPhone.prototype.__setState = function(state) {
		var previousState = this.currentState;
		this.currentState = state;
		return previousState;
	};
	
	SipPhone.prototype.isTransferring = function() {
		return this.currentState === PhoneStates.TRANSFERRING;
	};
	
	SipPhone.prototype.isTransferringConnected = function() {
		return this.currentState === PhoneStates.TRANSFERRING_CONNECTED;
	};
	
	SipPhone.prototype.isConferenceTrying = function() {
		return this.currentState === PhoneStates.CONFERENCE_CONSULTATION_TRYING;
	};
	
	SipPhone.prototype.isConferenceConnected = function() {
		return this.currentState === PhoneStates.CONFERENCE_CONSULTATION_CONNECTED;
	};
	
	SipPhone.prototype.getPhoneState = function() {
		return this.currentState;
	};

	if (!inConcert.AllegroSipPhone) {
		inConcert.AllegroSipPhone = SipPhone;
		inConcert.PhoneStates = PhoneStates;
	}
}());
