require "json"
require "orbit"
require "inconcert.orbit.application"
require "uuid"
local relativeRequire = relativeRequire
require "inconcert.orbit.application"
local cassandra = require "lua_cassandra"
require "LuaOrm.Date"
require "sidereal"

local htmlRender = require "inconcert.render".new()
local LuaConfigurationControl = require "LuaConfigurationControl"

--//////////////////////////////////////////////////
local __module_location = select(2, ...)
assert(__module_location, "should be running under inconcertWebHandler")

local modulename = select(1, ...)

module(modulename, package.seeall, orbit.new, inconcert.orbit.application.Decorate)

-- es necesario 'tapar' apps.repository.model que agrega Orbit, porque tenemos un módulo llamado igual
_M.model = nil

local configControl = nil
local m_redis_host
local m_redis_port
local m_redis_pwd

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-- funcion que hace la inicialización del modulo, la misma es llamada cuando se inicializa el webserver en el config.lua
function Initialize(env)
	LogDebug("initialize Analytics Application")

	configControl = LuaConfigurationControl.Client {
		type = "ANALYTICS",
		address = inconcert.environment.apps.oauth.config_server.ip,
		port = inconcert.environment.apps.oauth.config_server.port,
		node_address = inconcert.environment.apps.oauth.config_server.myIp,
		node_port = inconcert.environment.apps.oauth.config_server.myPort,
	}
	
	configControl:TriggerNodeReady()
	return true
end

-- Returns the date as the browser likes it
local function FormatDate(d)
	return inconcert.authUser:DisplayDate(LuaOrm.Date(d))
end

-- Returns a time as Xd HH:MM:SS (removing the days and hours if 0)
local function FormatTime(seconds)
	if seconds < 0 then
		return "-"..FormatTime(-seconds)
	end

	local function Phase(origin, phase)
		return math.floor(origin/phase), (origin % phase)
	end

	local minutes, seconds = Phase(seconds, 60)
	if minutes < 60 then
		return ("%02d:%02d"):format(minutes, seconds)
	end
	local hours, minutes = Phase(minutes, 60)
	if hours < 24 then
		return ("%02d:%02d:%02d"):format(hours, minutes, seconds)
	end
	local days, hours = Phase(hours, 24)
	return ("%dd %02d:%02d:%02d"):format(days, hours, minutes, seconds)
end

-- Returns a first level copy of a table
local function FirstLevelCopy(t)
	local result = {}
	for k, v in pairs(t) do
		result[k] = v
	end
	return result
end

-- Checks if a date is a valid working date. If it is, the date is returned. If not, the next working date is returned.
-- This is uglily hardcoded, it should depend on the campaign somewhat.
local function WorkingDate(d)
	-- TODO: Rework this part by campaign
	local holidays = {
		["01-01"] = true,
		["05-01"] = true,
		["07-18"] = true,
		["08-25"] = true,
		["12-24"] = true,
		["12-25"] = true,
		["12-31"] = true
	}
	local weekend = {
		[1] = true,
		[7] = true
	}
	local dayStart = LuaOrm.Date(d:toString("%Y-%m-%d 08:00:00"))
	local dayEnd = LuaOrm.Date(d:toString("%Y-%m-%d 20:00:00"))

	local oldDate = d:toString()
	local advanced
	if d:time() < dayStart:time() then
		d = dayStart
		advanced = true
	elseif d:time() >= dayEnd:time() then
		d = dayStart:add(0,0,1) -- Next start
		advanced = true
	end
	while weekend[d:date().wday] or holidays[d:toString("%m-%d")] do
		if not advanced then
			d = dayStart
			advanced = true
		end
		d = d:add(0,0,1)
	end
	if advanced then
		LogInfo("Mail on date %s will count as %s", oldDate, d:toString())
	end
	return d
end

-- Returns an array with all the campaigns of a VCC, in elements { id, name }
local function GetVCCCampaigns(vcc)
	local keyspace = cassandra:CassandraConnection("Agent")
	
	local campaigns = {}
	local basicColumns = keyspace:GetSliceRange(vcc, "VCCCampaigns", "", {
		slice_range = { start = "", finish = "", reversed = false, count = 200 } 
	}) 
	for _, column in ipairs(basicColumns or {}) do
		table.insert(campaigns, {id = column.name, name = column.value ~= "" and column.value or column.name})
	end
    keyspace:Close();
	return campaigns
end

-----------------------------------------------------
-- HTML and CSV generation
-----------------------------------------------------

-- Returns an error to the user explaining what went wrong
local function RenderUsage(web, errorMsg)
	web.status = 400
	return "Error: "..(errorMsg or "unspecified.")
end

-- Returns a string escaped to put on a CSV
local function CSVEscape(s)
	return s and '"'..tostring(s):gsub('"', '""')..'"' or ""
end

-- Renders a summary or detail page, with an export to CSV link
local function RenderPage(params)
	local web = params.web
	local campaignId = params.campaignId
	local VCC = params.VCC
	local page = params.page
	local report = params.report
	local description = params.description
	local startDate = params.startDate
	local endDate = params.endDate
	local getEntriesFn = params.getEntriesFn

	local entries, processed, more = getEntriesFn(params)

	local env = {
		title = description .. " For " .. (campaignId or VCC) .. 
			(startDate and (" From "..startDate) or "")..
			(endDate and (" To "..endDate) or ""),
		content = function()
			local env = {
				Description = description,
				Truncated = more,
				Processed = processed,
				StartDate = startDate or "the beginning of time",
				EndDate = endDate or "the present",
				ExportUrl = function()
					local filter = ""
					local campaign = campaignId and string.match(campaignId, "^(.+)@[^@]+$")
					filter = 'filter?output=csv'
					if startDate then
						filter = filter .."&startDate=" .. startDate;
					end
					if endDate then
						filter = filter.."&endDate=" .. endDate;
					end
					return "/inconcert/apps/analytics/" .. page .. "/" .. (campaign and (campaign .. "/") or "") .. filter
				end,
				List = function()
					for index, entry in ipairs(entries) do
							entry.index = index
							cosmo.yield(entry)
					end
				end
			}
			return htmlRender:IncludeCosmo("pages/"..report..".html", env)
		end
	}
	return htmlRender:RenderPage("pages/masterpage.html", web, env)
end

-- Export a summary or detail to a CSV string
local function ExportToCSV(params)
	local web = params.web
	local campaignId = params.campaignId
	local VCC = params.VCC
	local description = params.description
	local startDate = params.startDate
	local endDate = params.endDate
	local csvHeaders = params.csvHeaders
	local getEntriesFn = params.getEntriesFn

	local csv = {}
	table.insert(csv, table.concat(csvHeaders, ";"))
	
	params.limit = nil -- CSVs have no limit
	local entries = getEntriesFn(params)

	for _, entry in ipairs(entries) do
		local csvLineTable = {}
		for _, header in ipairs(csvHeaders) do
			table.insert(csvLineTable, CSVEscape(entry[header]))
		end
		table.insert(csv, table.concat(csvLineTable, ";"))
	end

	local csv = table.concat(csv, "\n")
	local upperDesc = string.upper(string.gsub(description, " ", "_"))
	local fileName =
		LuaOrm.Date():toString("%Y-%m-%d_%H-%M") .. "_" .. 
		(campaignId or VCC) .. "_" .. upperDesc ..
		(startDate and ("_FROM_"..startDate) or "") ..
		(endDate and ("_TO_"..endDate) or "") ..
		".csv"
	web.headers["Content-Type"] = "application/vnd.ms-excel"
	web.headers["Content-Length"] = #csv
	web.headers["Content-Disposition"] = [[attachment; filename="]] .. fileName .. '"'
	return csv
end

-- Creates either a page or a CSV, calling the provided callbacks to get the info
local function GeneratePage(params)
	local web = params.web
	local campaignId = params.campaignId
	local description = params.description
	local renderFn = params.renderFn
	local exportFn = params.exportFn

	local vcc = inconcert.authUser.virtualContactCenter
	if campaignId and campaignId ~= "" then
		params.campaignId = campaignId .. "@" .. vcc
		LogDebug(description .. " for campaign " .. params.campaignId)
	else
		params.campaignId = nil
		params.VCC = vcc
		LogDebug(description .. " for VCC " .. params.VCC)
	end
	
	params.startDate = web.input["startDate"]
	params.endDate = web.input["endDate"]
	if params.startDate then
		params.utcStartDate = LuaOrm.Date():newFromDifferentTimezone(params.startDate, inconcert.authUser.timezone):toString("!%Y-%m-%dT%H:%M:%SZ")
	end
	if params.endDate then
		params.utcEndDate = LuaOrm.Date():newFromDifferentTimezone(params.endDate, inconcert.authUser.timezone):toString("!%Y-%m-%dT%H:%M:%SZ")
	end
	
	local outputFormat = web.input["output"] or "render"
	if outputFormat == "render" then
		return RenderPage(params)
	elseif outputFormat == "csv" then
		return ExportToCSV(params)
	else
		return RenderUsage(web, "Unknown type of output '"..tostring(outputFormat).."'")
	end
end

-----------------------------------------------------
-- Interaction retrieval
-----------------------------------------------------

local interactionTypes = {
	"CALL",
	"MAIL",
	"IM",
	"CHAT",
	"FACEBOOK",
	"INSTAGRAM",
	"YOUTUBE",
	"APPSTORE",
	"GOOGLEPLAY",
	"RCS",
	"SMS",
	"WHATSAPP",
	"TWITTER"
}

local function send_auth(client_redis, pwd)
	if not pwd or type(pwd) ~= 'string' or pwd == "" then return true end
	local authresult = client_redis:auth(pwd)
	return authresult == "OK"
end	

-- Retrieves all the live interactions for the campaign, which were started in a certain period (which may be open ended).
-- Returns a set of interaction IDs, an array of interactions sorted by date, and how many open interactions started before the period.
local function GetLiveInteractions(params)
	local campaignId = params.campaignId
	local interactionType = params.interactionType
	local utcStartDate = params.utcStartDate
	local utcEndDate = params.utcEndDate
	local fetchPrevious = params.fetchPrevious

	local utcStartSeconds = utcStartDate and LuaOrm.Date(utcStartDate):time()
	local utcEndSeconds = utcEndDate and LuaOrm.Date(utcEndDate):time()
	local utcEarlySeconds = utcStartDate and fetchPrevious and LuaOrm.Date(utcStartDate):add(unpack(fetchPrevious)):time()

	local interactionSet = {}
	local interactionArray = {}
	local previousOpen = 0

	LogInfo("Connecting to redis on %s", m_redis_host)
	local redis = sidereal.connect(m_redis_host, m_redis_port)

	-- Retrieves all interactions of a type from redis and inserts the open ones
	-- on interactionSet and interactionArray, always inside the period.
	-- Increments previousOpen with those found previous to the period.
	local function FillWithInteractionsOfType(interactionType)
		local queue = "InteractionsByTypeAndCampaign:".. interactionType ..":"..campaignId
		local members = redis:zrange(queue, 0, -1)
		if type(members) == "table" then
			--members has format index/key
			for _, id in ipairs(members) do
				local data = redis:get("InteractionDetail:"..id)
				local res, interaction = pcall(json.decode, data)
				if 
					res and 
					interaction and 
					interaction.state ~= "FINISHED" and 
					interaction.state ~= "CANCELED" and
					(not utcEndSeconds or utcEndSeconds > interaction.startTimestamp)
				then
					-- Conversations earlier than utcStartSeconds are added to previousOpen
					if utcStartSeconds and utcStartSeconds > interaction.startTimestamp then
						previousOpen = previousOpen + 1
					end
					-- Conversations before utcStartSeconds (or utcEarlySeconds if it exists) are added to the result.
					if not utcStartSeconds or (utcEarlySeconds or utcStartSeconds) <= interaction.startTimestamp then
						interactionSet[id] = true
						table.insert(interactionArray, interaction)
					end
				end
			end
		end
	end

	if send_auth(redis, m_redis_pwd) then
		if interactionType then
			FillWithInteractionsOfType(interactionType)
		else
			for _, t in ipairs(interactionTypes) do
				FillWithInteractionsOfType(t)
			end
		end
	end

	table.sort(interactionArray, function(lhs, rhs)
		if lhs.startTimestamp < rhs.startTimestamp then 
			return true
		elseif lhs.startTimestamp > rhs.startTimestamp then 
			return false
		else
			return lhs.id < rhs.id
		end
	end)

	return interactionSet, interactionArray, previousOpen
end

-- Same as GetLiveInteractions, but for archived interactions, in a paged fashion.
-- An empty handle starts from the beginning.
-- Returns an array of interactions sorted by date, and the handle for the next page.
local function GetArchivedInteractions(handle, rows, params)
	local campaignId = params.campaignId
	local interactionType = params.interactionType
	local utcStartDate = params.utcStartDate
	local utcEndDate = params.utcEndDate
	if utcStartDate and params.fetchPrevious then
		utcStartDate = LuaOrm.Date(utcStartDate):add(unpack(params.fetchPrevious)):toString("!%Y-%m-%dT%H:%M:%SZ")
	end

	local interactionsKS = cassandra:CassandraConnection("Interactions")
	local stubs = {}
	local interactions = {}

	local function GetFromArchive(interactionID)
		local interaction = {}
		local basicColumns = interactionsKS:GetSliceRange(interactionID, "Interactions", "", {
			slice_range = { start = "", finish = "", reversed = false, count = 200 }
		}) 
		for _, column in ipairs(basicColumns or {}) do
			-- Most archived interactions have an "original" field which keeps all the interaction, 
			-- while some very very old ones have all their values scattered through the columns.
			if column.name == "original" then
				interaction = json.decode(column.value)
				break
			elseif column.name == "interactionData" or column.name == "steps" or column.name == "events" or column.name == "comments" then
				interaction[column.name] = json.decode(column.value)
			else
				interaction[column.name] = tonumber(column.value) or column.value
			end
		end
		return interaction
	end

	while not rs:atEnd() do
		local interactionID = rs:getCurrentValue()
		local interaction = GetFromArchive(interactionID, interactionsKS)
		if interaction.startTimestamp then
			table.insert(interactions, interaction)
		else
			table.insert(stubs, interaction)
		end
		rs:moveToNext()
	end
	interactionsKS:Close()

	return interactions, (#stubs + #interactions) == rows and rs:getLastHandle() or ""
end

-- Retrieves all interactions, both live and in archive, sorted by date inside a specific period, and calls the processFn function on each of them.
-- The processFn function should return false when no more interactions should be processed.
-- Returns the number of open interactions that started before the period.
local function ProcessInteractions(processFn, params)
	local archiveArray, lastArchiveFound = GetArchivedInteractions("", 2, params)
	local liveSet, liveArray, previousOpen = GetLiveInteractions(params)

	local liveCursor = 1
	local archiveCursor = 1

	-- Returns if lhs is better than rhs
	local function Better(lhs, rhs)
		return not rhs or lhs.startTimestamp < rhs.startTimestamp
	end

	-- Generate a sorted interaction array based on the sorted live and archived interactions, getting more archived ones if needed.
	while true do
		local candidate, candidateFrom
		if liveCursor <= #liveArray then
			candidate = liveArray[liveCursor]
			candidateFrom = "Live"
		end
		if archiveCursor <= #archiveArray then
			local archivedInt = archiveArray[archiveCursor]
			if liveSet[archivedInt.id] then
				-- Archived interactions that are live (probably reopened) are processed in the live version, because it's more up to date.
			elseif Better(archivedInt, candidate) then
				candidate = archivedInt
				candidateFrom = "Archive"
			end
		end
		if not candidate then
			break
		end
		if not processFn(candidate) then
			break
		end
		if candidateFrom == "Live" then
			liveCursor = liveCursor + 1
		else -- "Archive"
			archiveCursor = archiveCursor + 1
			if archiveCursor > #archiveArray and lastArchiveFound ~= "" then
				archiveArray, lastArchiveFound = GetArchivedInteractions(lastArchiveFound, 2, params)
				archiveCursor = 1
			end
		end
	end

	return previousOpen
end

-----------------------------------------------------
-- Agent state retrieval
-----------------------------------------------------

-- Retrieves all agent state changes sorted by date inside a specific period, and calls the processFn function on each of them.
-- The processFn function should return false when no more state changes should be processed.
local function ProcessAgentStates(processFn, params)
	local utcStartDate = params.utcStartDate
	local utcEndDate = params.utcEndDate
	local VCC = params.VCC

	-- The date parameters come in ISO form, and I need them in canonical
	local firstDate = utcStartDate and LuaOrm.Date(utcStartDate):toString("%Y-%m-%d %H:%M:%S") or ""
	local lastDate = utcEndDate and LuaOrm.Date(utcEndDate):toString("%Y-%m-%d %H:%M:%S") or ""

	local slice = 100
	local nextDate = firstDate
	local allStatesObtained

	local keyspace = cassandra:CassandraConnection("Historical")
	local VCCkey = "StatusChange:" .. VCC

	while not allStatesObtained do
		local rows = keyspace:GetSuperRangeSlices("AgentService", "", VCCkey, VCCkey,
			{ slice_range = { start = nextDate, finish = lastDate, reversed = false, count = slice + 1 } }, 1)
		allStatesObtained = true
		if not rows then
			keyspace:Close()
			return
		end
		for key, vccRow in rows:Iterator() do
			assert(key == VCCkey, "Cassandra is acting weird, GetSuperRangeSlices returned something unrequested.") -- Paranoid check
			local index = 0
			for _, superColumn in pairs(vccRow) do
				index = index + 1
				if index <= slice then
					for _, column in pairs(superColumn.columns) do
						local entry = json.decode(column.value)
						if not processFn(superColumn.name, entry.Agent, entry.NewStatus) then
							keyspace:Close()
							return
						end
					end
				else
					nextDate = superColumn.name
					allStatesObtained = false
					break
				end
			end
		end
	end

	keyspace:Close()
end

---------------------------------------------
-- All Conversations
---------------------------------------------

-- Retrieves all interactions from a period, with an optional limit.
-- If more than one agent took an interaction, an entry is generated for each agent.
-- Returns the list of interaction entries, the number of interaction entries, and a bool specifying if there are more interactions to retrieve (only when limit is used).
local function GetInteractions(params)
	local limit = params.limit

	-- Result contains mails as passed on the ProcessInteraction callback
	local result = {}
	local more

	-- Adds the interaction to result
	local function ProcessInteraction(interaction)
		local entry = {
			id = interaction.id,
			state = interaction.state,
			type = interaction.type,
			campaign = interaction.campaign,
			contact = interaction.contact or "",
			contact_name = interaction.interactionData and interaction.interactionData.contactName or "",
			contact_address = interaction.contactAddress or "",
			thread_events = interaction.events and #interaction.events or 0,
			last_modified = FormatDate(interaction.timestamp),
			creation_date = FormatDate(interaction.startTimestamp),
			first_taken = "",
			agent_id = "",
			agent_taken = "",
			end_date = "",
			queued_times = 0
		}

		local agentList = {}
		for _, step in ipairs(interaction.steps or {}) do
			if step.agent and step.action == "TAKEN" then
				if entry.first_taken == "" then
					entry.first_taken = FormatDate(step.timestampDouble or step.timestamp)
				end
				table.insert(agentList, {
					agent = step.agent, 
					timestamp = step.timestampDouble or step.timestamp
				})
			end
			if step.action == "FINISHED" or step.action == "CANCELED" then
				entry.end_date = FormatDate(step.timestamp)
			else
				-- It may be reopened, so erase whatever a previous closed step may have set
				entry.end_date = ""
			end
			if step.action == "QUEUED" then
				entry.queued_times = entry.queued_times + 1
			end
		end

		-- If having agents, generate one line per agent, if not, only one line
		if #agentList > 0 then
			for _, agent in ipairs(agentList) do
				if limit == #result then -- This may only happen if limit is a number >= 0
					more = true
					return false -- Process no more
				end
				entry.agent_id = agent.agent
				entry.agent_taken = FormatDate(agent.timestamp)
				table.insert(result, FirstLevelCopy(entry))
			end
		else
			if limit == #result then -- This may only happen if limit is a number >= 0
				more = true
				return false -- Process no more
			end
			table.insert(result, entry)
		end
		return true -- Keep on processing
	end

	ProcessInteractions(ProcessInteraction, params)
	return result, #result, more
end

local function ConversationsDetail(web, campaignId)
	return GeneratePage {
		web = web, 
		campaignId = campaignId, 
		description = "Conversations Detail",
		limit = 100,
		page = "conversations",
		report = "report_conversations",
		csvHeaders = {
			"id", "state", "type", "campaign",
			"contact", "contact_name", "contact_address",
			"last_modified", "creation_date", "first_taken",
			"agent_id", "agent_taken",
			"end_date", "queued_times"
		},
		getEntriesFn = GetInteractions
	}
end

----------------------------------------------
-- Agent State Changes (detail)
----------------------------------------------

-- Retrieves all agent state changes from a period, with an optional limit
-- Returns the list of changes, the number of changes, and a bool specifying if there are more changes to retrieve (only when limit is used).
local function GetAgentStateChanges(params)
	local limit = params.limit

	-- Result contains mails as passed on the ProcessMails callback
	local result = {}
	local more
	local function AddToChanges(date, agent, state)
		if #result == limit then
			more = true
			return false -- Stop processing
		end
		table.insert(result, {
			date = date,
			agent = agent,
			state = state
		})
		return true -- Processing more
	end
	ProcessAgentStates(AddToChanges, params)
	return result, #result, more
end

local function AgentStatesDetail(web)
	return GeneratePage {
		web = web, 
		description = "Agent States Detail",
		limit = 100,
		page = "agent_states",
		report = "report_agent_states",
		csvHeaders = { "date", "agent", "state" },
		getEntriesFn = GetAgentStateChanges
	}
end

----------------------------------------------
-- Agent State Changes (summary)
----------------------------------------------

-- Generates an array of agent state summaries. Each entry has an agent name, a state, and how much time it spent on that state.
-- Returns the array and the number of state changes that were processed to create it.
local function GetAgentStateSummary(params)
	local utcEndDate = params.utcEndDate

	local function AddTimeToPreviousState(date, agentSummary)
		local summaryState = agentSummary.state
		if summaryState == "OFFLINE" then
			-- Don't accumulate offline time
			return
		end
		local currentDate = LuaOrm.Date(date)
		local prevDate = LuaOrm.Date(agentSummary.timestamp)
		local interval = prevDate:interval(currentDate)
		if not agentSummary.accumulated then
			agentSummary.accumulated = {}
		end
		agentSummary.accumulated[summaryState] = (agentSummary.accumulated[summaryState] or 0) + interval
	end

	local summary = {}
	local processed = 0
	-- There is an improvements that can be done to this summarization. For each agent, there should be a search for
	-- the last state after the period, so it can be calculated what the agent was doing from the beginning of the period.
	-- But the customers are happy with this implementation, so it's left to freeze.
	local function AddToSummary(date, agent, state)
		if state:match("[^ ]+ CAMPAIGN ") then
			-- Ignore specific campaign activity
			return true -- Always keep processing
		end
		if summary[agent] then
			local dateOrm = LuaOrm.Date(date)
			AddTimeToPreviousState(dateOrm, summary[agent])
		else
			summary[agent] = {}
		end
		summary[agent].state = state
		summary[agent].timestamp = date
		processed = processed + 1
		return true -- Always keep processing
	end
	ProcessAgentStates(AddToSummary, params)

	if utcEndDate then
		local utcEndDateOrm = LuaOrm.Date(utcEndDate)
		-- Go agent per agent and set the last state of each agent as happened between the last timestamp and the period end
		for _, agentData in pairs(summary) do
			AddTimeToPreviousState(utcEndDateOrm, agentData)
		end
	end

	local orderedSummary = {}
	for agent, agentData in pairs(summary) do
		for state, seconds in pairs(agentData.accumulated or {}) do
			table.insert(orderedSummary, {
				agent = agent,
				state = state,
				time_in_seconds = seconds,
				time = FormatTime(seconds)
			})
		end
	end
	table.sort(orderedSummary, function(a, b)
		if a.agent < b.agent then
			return true
		elseif a.agent > b.agent then
			return false
		end
		return a.state < b.state
	end)
	return orderedSummary, processed
end

local function AgentStatesSummary(web)
	return GeneratePage {
		web = web, 
		description = "Agent States Summary",
		page = "agent_states_summary",
		report = "report_agent_states_summary",
		csvHeaders = { "agent", "state", "time_in_seconds", "time" },
		getEntriesFn = GetAgentStateSummary
	}
end

----------------------------------------------
-- Mail Details (one mail per row)
----------------------------------------------

-- Retrieves all mails from a period, with an optional limit
-- Returns the list of mails, the number of mails, and a bool specifying if there are more mails to retrieve (only when limit is used).
local function GetMails(params)
	local limit = params.limit

	-- Result contains mails as passed on the ProcessMails callback
	local result = {}
	local more

	-- Extracts the mails from an interaction and adds them to result
	local function ProcessMails(interaction)
		local contactName = interaction.interactionData and interaction.interactionData.contactName or ""
		for _, event in ipairs(interaction.events or {}) do
			if limit == #result then -- This may only happen if limit is a number >= 0
				more = true
				return false -- Process no more
			end
			local to = event.to
			if type(to) == "table" then
				to = ""
				for k, mail in pairs(event.to) do
					to = to .. (string.len(to) > 0 and ", " or "") .. mail
				end
			end
			local d = LuaOrm.Date(event.timestamp)
			local system_date = d:add(0,0,0,0,interaction.campaignOffsetMinutes or 0)
			table.insert(result, {
				id = interaction.id,
				original_date = event.date,
				utc_date = d:toString(),
				from = event.from,
				to = to,
				subject = event.subject,
				agent = event.agent or "",
				campaign = interaction.campaign,
				system_date = system_date
			})
		end
		return true -- Keep on processing
	end

	ProcessInteractions(ProcessMails, params)
	return result, #result, more
end

local function MailConversationsDetail(web, campaignId)
	return GeneratePage {
		web = web, 
		campaignId = campaignId, 
		interactionType = "MAIL",
		description = "Mail Conversations",
		limit = 100,
		page = "mail_conversations",
		report = "report_mail_detail",
		csvHeaders = {
			"id", "system_date", "utc_date",
			"from", "subject", "to", "agent",
			"campaign", "original_date"
		},
		getEntriesFn = GetMails
	}
end

----------------------------------------------
-- Mail Summary (totals by day)
----------------------------------------------

-- Generates an array of mail summary entries. Each entry has numbers for a specific day in a period.
-- Returns the array and the number of interactions that were processed to create it.
local function PrepareMailSummary(params)
	local startDate = params.startDate
	local endDate = params.endDate

	local startDateOrm = startDate and LuaOrm.Date(startDate)
	params.fetchPrevious = { 0, -3, 0 } -- Queries will be made against three months earlier.
	local endDateOrm = endDate and LuaOrm.Date(endDate)

	local summary_by_day = {}
	local function IncrementInSummary(date, indicator, increment)
		-- Don't increment values for dates outside the range
		if startDateOrm and startDateOrm:time() > date:time() then
			return
		end
		if endDateOrm and endDateOrm:time() <= date:time() then
			return
		end
		increment = increment or 1
		local key = date:toString("%Y-%m-%d")
		if not summary_by_day[key] then
			summary_by_day[key] = {
				unique_ids = 0,
				received_mails = 0,
				received_mails_on_schedule = 0,
				mail_responses = 0,
				mails_24hs = 0,
				mails_48hs = 0,
				mails_answered = 0,
				finished = 0,
				finished_unresponded = 0,
				pending = 0,
				punctuality_24hs = 0,
				punctuality_48hs = 0
			}
		end
		summary_by_day[key][indicator] = summary_by_day[key][indicator] + increment
	end

	local processed = 0
	-- Iterate through the conversations, filling the by-day aggregates
	local function AddToConversations(conversation)
		-- Outbound and SPAM mail conversations are not reflected on the summary
		if conversation.initiative ~= "INBOUND" or conversation.state == "SPAM" then
			LogInfo("Analytics report: Conversation %s is ignored for being %s", conversation.id, (conversation.state == "SPAM" and "spam" or "not inbound"))
			return true -- Always keep processing
		end

		-- Check if it's unfinished without response.
		-- If so, just increase finished_unresponded.
		local finishDate
		if conversation.state == "CANCELED" or conversation.state == "FINISHED" then
			local lastStep = conversation.steps[#conversation.steps]
			finishDate = LuaOrm.Date(FormatDate(lastStep.timestampDouble or lastStep.timestamp))
			local hasAgentEvent
			if conversation.events then
				for _, event in ipairs(conversation.events) do
					if event.agent then
						hasAgentEvent = true
						break
					end
				end
			end
			if not hasAgentEvent then
				IncrementInSummary(finishDate, "finished_unresponded")
				LogInfo("Analytics report: Conversation %s (finishing on %s) is ignored for being unresponded", conversation.id, finishDate)
				return true -- Always keep processing.
			end
		end

		-- Unique IDs go at creation date
		local creationDate = LuaOrm.Date(FormatDate(conversation.startTimestamp))
		LogInfo("Analytics report: Conversation %s started on %s", conversation.id, creationDate)
		IncrementInSummary(creationDate, "unique_ids")

		-- Finished go at finish date
		if finishDate then
			IncrementInSummary(finishDate, "finished")
			LogInfo("Analytics report: Conversation %s finished on %s", conversation.id, finishDate)
		end

		-- Process mails (both from the client and the agent) at the times of each mail
		local best_response_date
		if conversation.events then
			for _, event in ipairs(conversation.events) do
				local mailDate = LuaOrm.Date(FormatDate(event.timestamp))
				if not event.agent then
					--this is a mail from a client
					LogInfo("Conversation %s has client mail on %s", conversation.id, mailDate)
					IncrementInSummary(mailDate, "received_mails")
					local workingDate = WorkingDate(mailDate)
					IncrementInSummary(workingDate, "received_mails_on_schedule")
				else
					LogWarning("Conversation %s has response on date %s by agent %s", conversation.id, mailDate, event.agent)
					IncrementInSummary(mailDate, "mail_responses")
					--is this the best response?
					if not best_response_date or best_response_date:time() > mailDate:time() then
						best_response_date = mailDate
					end
				end
			end
		end

		-- Response time goes at creation time
		if best_response_date then
			local workingDate = WorkingDate(creationDate)
			local time_frame = (best_response_date:time() - workingDate:time() ) / 3600 -- In hours
			LogWarning("Best response for conversation %s was on %s that is %s hours", conversation.id, best_response_date:toString(), time_frame)
			IncrementInSummary(creationDate, "mails_answered")
			-- If it counts for the less than 24, it also counts for the less than 24
			if time_frame <= 48 then
				IncrementInSummary(creationDate, "mails_48hs")
				if time_frame <= 24 then
					IncrementInSummary(creationDate, "mails_24hs")
				end
			end
		end

		processed = processed + 1
		return true -- Always keep processing
	end
	local previousPending = ProcessInteractions(AddToConversations, params)

	-- Convert to an array and sort it, get the rates for 24/48hs
	local ordered_summary = {}
	for date, summary in pairs(summary_by_day) do
		summary.date = date
		summary.punctuality_24hs = ((summary.unique_ids > 0 and summary.mails_24hs / summary.unique_ids or 0) * 100).."%"
		summary.punctuality_48hs = ((summary.unique_ids > 0 and summary.mails_48hs / summary.unique_ids or 0) * 100).."%"
		table.insert(ordered_summary, summary)
	end
	table.sort(ordered_summary, function(a, b)
		return a.date < b.date
	end)

	-- Go date by date, setting the pending mails
	local total_pending = previousPending
	for _, summary in ipairs(ordered_summary) do
		total_pending = total_pending + (summary.unique_ids - summary.finished)
		summary.pending = total_pending
	end

	return ordered_summary, processed
end

local function MailConversationSummary(web, campaignId)
	return GeneratePage {
		web = web, 
		campaignId = campaignId, 
		interactionType = "MAIL",
		description = "Mail Summary",
		page = "mail_summary",
		report = "report_mail_summary",
		csvHeaders = {
			"date", "unique_ids", "received_mails", "received_mails_on_schedule",
			"mail_responses", "mails_24hs", "mails_48hs", "mails_answered", 
			"finished", "finished_unresponded",
			"pending", "punctuality_24hs", "punctuality_48hs"
		},
		getEntriesFn = PrepareMailSummary
	}
end

-----------------------------------------------
-- Chat Summary (totals by day/hour)
-----------------------------------------------

-- This one is also used by SMS summary (and may be used for IM if implemented)
local function PrepareChatSummary(params)
	local statuses = {}
	-- Date filtering is not needed, because everything is indexed by startTimestamp

	local processed = 0
	local function AddToConversations(conversation)
		local taken = false
		local creationDate = LuaOrm.Date(FormatDate(conversation.startTimestamp))
		local waitingTime = 0
		local attentionTime = 0
		local state = "Wait"
		local lastChange = conversation.startTimestamp
		for k, step in ipairs(conversation.steps) do
			local change = step.timestampDouble or LuaOrm.Date(step.timestamp):time() -- To deal with very very old interactions
			if state == "Wait" then
				waitingTime = waitingTime + change - lastChange
				if step.action == "TAKEN" then
					taken = true
					state = "Attention"
				end
			elseif state == "Attention" then
				attentionTime = attentionTime + change - lastChange
				if step.action ~= "TAKEN" then
					state = "Hold"
				end
			elseif state == "Hold" then
				if step.action == "TAKEN" then
					state = "Attention"
				end
			end
			if step.action == "CANCELED" or step.action == "FINISHED" then
				state = "Closed"
			end
			lastChange = change
		end
		local key = creationDate:toString("%Y-%m-%d-%H")
		if not statuses[key] then
			statuses[key] = {
				day =  creationDate:toString("%Y-%m-%d"),
				hour = creationDate:toString("%H"),
				ATTENDED = 0,
				ABANDONED = 0,
				TOTAL = 0,
				DURATIONS = 0,
				POSITIVE = 0
			}
		end
		statuses[key].TOTAL = statuses[key].TOTAL + 1
		statuses[key].DURATIONS = statuses[key].DURATIONS + attentionTime
		LogWarning("Interaction waiting time: "..waitingTime)
		if taken then
			statuses[key].ATTENDED = statuses[key].ATTENDED + 1
		elseif state == "Closed" then
			statuses[key].ABANDONED = statuses[key].ABANDONED + 1
		end
		if waitingTime < 30 then
			statuses[key].POSITIVE = statuses[key].POSITIVE + 1
		end
		processed = processed + 1
		return true -- Always keep processing
	end
	local previousPending = ProcessInteractions(AddToConversations, params)

	local orderedStatuses = {}
	for _, status in pairs(statuses) do
		table.insert(orderedStatuses, {
			day = status.day,
			hour = status.hour,
			attended = status.ATTENDED,
			abandoned = status.ABANDONED,
			total = status.TOTAL,
			av_duration = status.ATTENDED > 0 and status.DURATIONS / status.ATTENDED or 0,
			service_level = status.TOTAL > 0 and (status.POSITIVE / status.TOTAL) * 100 .. "%" or 0
		})
	end
	table.sort(orderedStatuses, function(a, b)
		if a.day < b.day then
			return true
		elseif a.day > b.day then
			return false
		end
		return a.hour < b.hour
	end)
	return orderedStatuses, processed
end

local function ChatConversationSummary(web, campaignId)
	return GeneratePage {
		web = web, 
		campaignId = campaignId, 
		interactionType = "CHAT",
		description = "Webchat Conversations Summary",
		page = "chat_summary",
		report = "report_chat_summary",
		csvHeaders = {
			"day", "hour",
			"attended", "abandoned", "total",
			"av_duration", "service_level"
		},
		getEntriesFn = PrepareChatSummary
	}
end

-----------------------------------------------
-- SMS Summary (totals by day/hour)
-----------------------------------------------
--
local function SMSConversationSummary(web, campaignId)
	return GeneratePage {
		web = web, 
		campaignId = campaignId, 
		interactionType = "SMS",
		description = "SMS Conversations Summary",
		page = "sms_summary",
		report = "report_chat_summary", -- It uses the same page as webchat
		csvHeaders = {
			"day", "hour",
			"attended", "abandoned", "total",
			"av_duration", "service_level"
		},
		getEntriesFn = PrepareChatSummary -- And the same collector
	}
end

---------------------------------------------------------
-- Main page
---------------------------------------------------------

local function RenderMain(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local env = {
		title = "Filter Export Data",
		content = function()
			local env = {}	
			env.Campaigns = function()
				local campaignList = GetVCCCampaigns(vcc)
				for k, v in ipairs(campaignList) do
					cosmo.yield(v)
				end	
			end
			return htmlRender:IncludeCosmo("pages/main.html", env)
		end
	}
	return htmlRender:RenderPage("pages/masterpage.html", web, env)
end

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",
	
	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		
		LogDebug("Checking analytics apps status")

		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end

		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"cassandra", "redis"})

		app_status.current = newConfig

		-- Configure Cassandra
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				LogInfo("Connecting to cassandra instance at %s:%s", value.host, tostring(value.port))
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port
			end
		end

		m_redis_host = newConfig["redis"]["host"]
		m_redis_port = tonumber(newConfig["redis"]["port"]
		m_redis_pwd = newConfig["redis"]["pwd"]
		
		app_status.actualStatus = "ready"
		return "ready"
	end
}

local function Ping(web)
	web.headers["Content-Type"] = "application/json"
	return json.encode({result = "pong"})
end

_M:dispatch_get(AgentStatesSummary, "/agent_states_summary/(.*)")
_M:dispatch_get(AgentStatesDetail, "/agent_states/(.*)")
_M:dispatch_get(ConversationsDetail, "/conversations/([^/]+)/(.*)")
_M:dispatch_get(MailConversationsDetail, "/mail_conversations/([^/]+)/(.*)")
_M:dispatch_get(ChatConversationSummary, "/chat_summary/([^/]+)/(.*)")
_M:dispatch_get(SMSConversationSummary, "/sms_summary/([^/]+)/(.*)")
_M:dispatch_get(MailConversationSummary, "/mail_summary/([^/]+)/(.*)")
_M:dispatch_get(RenderMain, "/")
_M:dispatch_post(Ping, "/ping")

_M:dispatch_static("/images/.+", "/styles/.+" , "/scripts/.+")
