local oo = require "loop.base"
local date = require "LuaOrm.Date"
require "LuaOrm.Mapper"
require "inconcert.translators"
local lua_globaltimer = require "lua_globaltimer"
local ir = require "lua_iragentlink"
local permissions = require "shared.permissions.permissions"
local http_mapping = require "apps.agent.result_codes_to_http_codes"
local helpers = require "apps.agent.helpers"
local model_perm = require "shared.permissions.models.permissions"
local model = require "apps.batch_management.model"
local models = model:GetModels()
local utils = require "shared.apps.utils"
local moses = require "moses"
local model_campaigns = require "apps.agent.models.campaigns"
local model_batches = require "apps.batch_management.models.Batches"
local utf8 = require "icu.utf8"

local wsapi = require "wsapi"
wsapi.request = require "wsapi.request"
wsapi.request_stream = require "wsapi.request_stream"

require "uuid"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil

local m_globaltimer
local function getNow()
	if not m_globaltimer then
		m_globaltimer = lua_globaltimer.LuaGlobalTimer(inconcert.environment.apps.agent.cluster)
		local timeServers = inconcert.environment.apps.agent.timeServers
		for _, server in pairs(timeServers or {}) do
			m_globaltimer:AddKnownServer(server.IP, server.port)
		end
	end
	return LuaOrm.Date(m_globaltimer:GetTime())
end

local function hasLicensesForChannel (channel, ksAgent)
	local agent = require "apps.agent.agent"
	return agent.hasLicensesForChannel(channel, ksAgent)
end

local function to_string (value)
	return moses.isTable(value) and table.concat(value, " ") or tostring(value)
end

-- wraps a single value in an array (or returns the array)
local function to_table (value)
	return moses.isTable(value) and value or { tostring(value) }
end

local function is_nonempty_string (value)
	return (moses.isString(value) and not moses.isEmpty(value))
end

local function IsPinned (batch)
	return batch and batch.properties and batch.properties.pinned or false
end

local function SortPinnedBatches (batches, last, rows, sort)
	local sort_type, sort_order = next(sort or {})
	table.sort( batches, function( a, b )
		if IsPinned(a) and IsPinned(b) then
			return a.publish_timestamp > b.publish_timestamp
		elseif IsPinned(a) then
			return true
		elseif IsPinned(b) then
			return false
		elseif sort_type and a and b then
			if sort_order.order == "desc" then
				return a[sort_type] > b[sort_type]
			else
				return a[sort_type] < b[sort_type]
			end
		end
		return false
	end )
	local filtered = {}
	for i = 1, rows do
		local index = i + last
		if index > #batches then
			break
		end
		table.insert(filtered, batches[index])
	end
	return filtered
end

function SearchBatchesHelper (request, skipViewPermission, allCampaigns, datasource, elasticSearchConnectionConfig)
	LogInfo("[BatchManagement.BACHTES:SearchPostsHelper] (START) Executing batches mng request: %s", json.encode(request))

	local _T = m_htmlRender and m_htmlRender.extension.makeTextTranslator() or function(text) return text end
	local user_name, vcc = inconcert.authUser.userName, inconcert.authUser.virtualContactCenter
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")

	-- validate the user is filtering for valid campaigns, in case empty add all the campaigns available
	local userCampaigns = {}
	if allCampaigns then
		local userCampaigns_a = model_campaigns.GetVCCCampaigns(vcc, ksAgent)
		for campaignId, _ in pairs(userCampaigns_a) do
			userCampaigns[utils.AppendVcc(campaignId, vcc)] = true
		end
	else
		userCampaigns = model_campaigns.GetUserCampaigns(user_name, vcc, ksAgent)
	end
	if not next(userCampaigns) then
		ksAgent:Close()
		LogInfo("[BatchManagement.BACHTES:SearchBatchesHelper] User %s does not belong to a campaign, no batches are returned", user_name)
		return 403, { reason = _T("User does not belong to any campaign") }
	end

	local request_query = request.query
	if not request_query then
		ksAgent:Close()
		LogWarning("[BatchManagement.BACHTES:SearchBatchesHelper] Missing query")
		return 400, { reason = _T("Missing query") }
	end

	local query = {
		query = {
			bool = {
				must = { },
				filter = { }
			}
		}
	}

	local sortPinned = false

	--view filter
	local viewPermission
	local now = getNow():toString("%Y-%m-%d %H:%M:%S")
	if request.view == "ACTIVE" then
		viewPermission = "active_batches"
		request_query.state = { "ACTIVE", "PAUSED", "PROCESSING" }
		request_query.publish_timestamp = { format = "yyyy-MM-dd HH:mm:ss", lte = now }
		sortPinned = true
	elseif request.view == "SCHEDULED" then
		viewPermission = "scheduled_batches"
		request_query.state = { "ACTIVE", "PAUSED", "PROCESSING" }
		request_query.publish_timestamp = { format = "yyyy-MM-dd HH:mm:ss", gte = now }
		sortPinned = true
	elseif request.view == "ACTIVE_SCHEDULED" then
		viewPermission = nil
		request_query.state = { "ACTIVE", "PAUSED", "PROCESSING" }
	elseif request.view == "ARCHIVE" then
		viewPermission = "batches_history"
		request_query.state = { "CLOSED", "DELETED" }
		request_query.publish_timestamp = nil
	elseif request.view == "ALL" then
		skipViewPermission = true
	else
		ksAgent:Close()
		LogWarning("[BatchManagement.BACHTES:SearchBatchesHelper] Unrecognized view: %s", request.view)
		return 400, { reason = _T("Invalid view") }
	end

	local applicationsActions = model_perm.GetApplicationsActions(ksAgent, vcc)

	--conditions
	local mustFilters = {
		type = {
			default = model_batches.GetTypes(),
			filter = function(channel)
				return hasLicensesForChannel(channel, ksAgent)
			end,
			empty = "There are no licenses to any outgoing batches channel"
		},
		campaign = {
			default = userCampaigns and moses.keys(userCampaigns) or nil,
			filter = function(campaign)
				return skipViewPermission or not viewPermission or model_perm.IsPermissionEnabledForUser("batch_management", viewPermission, campaign, ksAgent, applicationsActions)
			end,
			empty = "User has not permissions to any campaign in this view"
		},
		account = {},
		page = {},
		state = {},
		created_by_user = {},
		publish_timestamp = {
			condition = function(name, filter)
				return { range = { [name] = filter } }
			end
		}
	}

	--add all accounts and pages to avoid getting deleted pages
	if not request_query.account then
		mustFilters.account.default = {}
		mustFilters.page.default = {}
		for _, campaignId in pairs(mustFilters.campaign.default or {}) do
			local accounts = model_campaigns.GetCampaignAccounts(campaignId, ksAgent)
			for _, account in pairs(accounts) do
				if account.proxyType == "TWITTER" then
					table.insert(mustFilters.account.default, account.id)
					mustFilters.page.emptyCondition = true
				elseif account.proxyType == "FACEBOOK" or account.proxyType == "INSTAGRAM" then
					table.insert(mustFilters.account.default, account.id)
					if type(account.pages) == "string" then
						account.pages = json.decode(account.pages)
					end
					for _, pageId in pairs(account.pages) do
						table.insert(mustFilters.page.default, pageId)
					end
				elseif account.proxyType == "SMS" then
					table.insert(mustFilters.account.default, account.id)
					mustFilters.page.emptyCondition = true
				elseif account.proxyType == "RCS" then
					table.insert(mustFilters.account.default, account.id)
					mustFilters.page.emptyCondition = true
				elseif account.proxyType == "WHATSAPP" then
					table.insert(mustFilters.account.default, account.id)
					mustFilters.page.emptyCondition = true
				elseif account.proxyType == "MAIL" then
					table.insert(mustFilters.account.default, account.id)
					mustFilters.page.emptyCondition = true
				end
			end
		end
	end

	local must = query.query.bool.filter
	for name, config in pairs(mustFilters) do
		local filter = request_query[name] and to_table(request_query[name]) or config.default or {}
		if config.filter then
			filter = moses.filter(filter, function(_, value)
				return config.filter(value)
			end)
		end
		if not utils.TableIsEmpty(filter) then
			local condition
			if config.condition then
				condition = config.condition(name, filter)
			else
				local filterType
				if #filter > 1 then
					filterType = "terms"
				else
					filterType = "term"
					filter = filter[1]
				end
				condition = { [filterType] = { [name] = filter } }
			end
			if config.emptyCondition then
				condition = {
					["bool"] = {
						["should"] = {
							{
								["bool"] = {
									must_not = {
										exists = { field = name }
									}
								}
							},
							condition
						}
					}
				}
			end
			table.insert(must, condition)
		elseif config.empty then
			ksAgent:Close()
			LogWarning("[BatchManagement.BACHTES:SearchBatchesHelper] Empty config: %s", config.empty)
			return 403, { reason = _T(config.empty) }
		end
	end

	--content
	if request_query.content then
		local value = to_string(request_query.content)
		if is_nonempty_string(value) then
			table.insert(query.query.bool.must, {
				wildcard = {
					content = string.format("%s*", value)
				}
			})
			if request.highlight then
				query.highlight = { fields = { content = { foo = 1 } } } -- foo is there to avoid {} to be encoded as []
			end
		end
	end

	-- This field can't be empty, so remove it no query was added
	if #query.query.bool.must == 0 then
		query.query.bool.must = nil
	end

	--sort
	query.sort = request.sort or { { publish_timestamp = { order = "desc" } } }

	-- Determine if pagination is needed
	query.size = request.rows or 20
	query.from = (tonumber(request.from) or 0)
	if sortPinned then
		query.size = 500
		query.from = 0
	end

	local ok, result = model_batches.search_elastic(inconcert.authUser.virtualContactCenter, query, elasticSearchConnectionConfig)
	if not ok then
		ksAgent:Close()
		LogWarning("[BatchManagement.BACHTES:SearchBatchesHelper] Elasticsearch result: %s", json.encode(result))
		return 500, { result = "REQUEST_FAILED", reason = result.error }
	end
	if result.timed_out then
		ksAgent:Close()
		LogWarning("[BatchManagement.BACHTES:SearchBatchesHelper] Query against elasticsearch timed out. %s", json.encode(query))
		return 502, { result = "REQUEST_TIMEDOUT", reason = _T("Query timed out") }
	end
	LogInfo("[BatchManagement.BACHTES:SearchBatchesHelper] Query took %s ms and returned %s hits out of %s", result.took, #result.hits.hits, result.hits.total.value)
	
	local batches, highlights = {}, {}
	local ksInteractions = utils.GetDatasource():CassandraConnection( "Interactions" )
	-- retrieve the full batch
	moses.each(result.hits.hits, function(_, hit)
		local batch = model_batches.GetBatch(hit._id, true, ksInteractions, datasource)
		if not utils.TableIsEmpty(batch) then 
			batch.translatedState = batch.state and _T(batch.state) or batch.state
			table.insert(batches, batch) 
		end
		if request.highlight then
			highlights[hit._id] = hit.highlight.content
		end
	end)
	if sortPinned then
		batches = SortPinnedBatches(batches, tonumber(request.from) or 0, request.rows or 20, query.sort[1])
	end
	local last = (tonumber(request.from) or 0) + #batches

	local account_statistics = nil
	if request.account_statistics then
		local accountId = request_query.account .. (request_query.page and (":" .. request_query.page) or "")
		account_statistics = model_batches.GetAccountStatistics(accountId, ksInteractions, datasource) or {}
	end
	ksInteractions:Close()
	ksAgent:Close()

	local res = {
		batches = batches, 
		total = result.hits.total.value,
		last = last,
		highlights = request.highlight and highlights,
		account_statistics = account_statistics
	}

	--LogDebug("[BatchManagement.BACHTES:SearchPostsHelper] --> (END) Executed BATCHES. res: %s", json.encode(res))

	LogInfo("[BatchManagement.BACHTES:SearchPostsHelper] (END) Executed BATCHES. total: %d", res.total)

	return 200, res
end

local function SearchBatches (web)
	local status, response = SearchBatchesHelper( web.request )
	web.status = status
	return response
end

local function GetBatches (web)
	local post_data =  web.input["post_data"]
	local res, ids = pcall(json.decode, post_data)
	if not res or not ids then
		LogWarning("[BatchManagement:GetBatches] Invalid request: %s", post_data)
		web.status = 400
		return {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	end
	local batches = {}
	local ksInteractions = utils.GetDatasource():CassandraConnection( "Interactions" )
	for _, id in pairs(ids) do
		local batch = models.Batches.GetBatch(id, true, ksInteractions)
		if not utils.TableIsEmpty(batch) then 
			batches[id] = batch
		end
	end
	ksInteractions:Close()
	web.status = 200
	return batches
end

--Actions

local actions = {
	PAUSE = {
		permission		= "batches_pause",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId" },
		method			= "PauseOutgoingBatch"
	},
	RESUME = {
		permission		= "batches_resume",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId" },
		method			= "ResumeOutgoingBatch"
	},
	CLOSE = {
		permission		= "batches_close",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId" },
		method			= "CloseOutgoingBatch"
	},
	--REOPEN = {
	--	permission		= "batches_reopen",
	--	required_params = { "proxyType", "campaignAtVCC", "account", "batchId" },
	--	method			= "ReopenOutgoingBatch"
	--},
	DELETE = {
		permission		= "batches_delete",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId" },
		method			= "DeleteOutgoingBatch"
	},
	EDIT = {
		permission		= "batches_edit",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId", "batch" },
		method			= "EditOutgoingBatch"
	},
	EDIT_PROPERTIES = {
		permission		= "batches_edit",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId", "properties" },
		method			= "EditOutgoingBatchProperties"
	},
	CREATE = {
		permission		= "batches_create",
		required_params = { "proxyType", "campaignAtVCC", "account", "page", "batch" },
		method			= "NewOutgoingBatch"
	},
	PIN = {
		permission		= "batches_edit",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId" },
		method			= "PinOutgoingBatch"
	},
	UNPIN = {
		permission		= "batches_edit",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId" },
		method			= "UnpinOutgoingBatch"
	},
	ADD_ADDRESSES = {
		permission		= "batches_edit",
		required_params = { "proxyType", "campaignAtVCC", "account", "batchId", "addresses"},
		method			= "AddAddressesToOutgoingBatch"
	},
	
}

local function RunAction (data, action, applicationsActions)
	data.user = utils.GetCurrentLoggedUser()
	local definition = actions[action]
	if not definition then
		LogWarning("[OutgoingBatch] Invalid action: %s for data %s", action, json.encode(data))
		return 400, {result = "INVALID_REQUEST", reason = "Invalid action: " .. action}
	end
	--validate params
	for _, param in pairs(definition.required_params or {}) do
		if not data[param] then
			LogWarning("[OutgoingBatch] Invalid request: %s in action %s", json.encode(data), action)
			return 400, {result = "INVALID_REQUEST", reason = "Required param not found: " .. param}
		end
	end
	--validate type
	if definition.types and not definition.types[data.proxyType] then
		LogWarning("[OutgoingBatch] Invalid request: %s in action %s", json.encode(data), action)
		return 400, {result = "INVALID_REQUEST", reason = "Action " .. action .. " not supported for type: " .. data.proxyType}
	end
	--validate permissions
	if not model_perm.IsPermissionEnabledForUser("batch_management", definition.permission, data.campaignAtVCC, nil, applicationsActions) then
		LogWarning("[OutgoingBatch] Invalid action: %s for data %s", action, json.encode(data))
		return 403, {result = "FORBIDDEN", reason = "User has not permissions to action: " .. action}
	end
	--run action
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local response = irLink[definition.method](irLink, data)
	local status = http_mapping.getHTTPCode(response["result"])
	return status, response
end

local function RunActionFromWeb (web, action)
	local post_data =  web.input["post_data"]
	local res, data = pcall(json.decode, post_data)
	if not res or not data then
		LogWarning("[OutgoingBatch] Invalid request: %s in action %s", post_data, action)
		return 400, {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	end
	return RunAction(data, action)
end

local function RunBatchAction (batch, action)
	LogInfo("[OutgoingBatch] Running batch action %s for %d batches", action, #batch)
	local vcc = inconcert.authUser.virtualContactCenter
	local applicationsActions = model_perm.GetApplicationsActions(nil, vcc)
	local successful_nodes = {}
	local unsuccessful_nodes = {}

	for _, data in pairs(batch) do
		local id = (data.account or "") .. ":" .. (data.batchId or "")
		local status, response = RunAction(data, action, applicationsActions)
		if status == 200 then
			table.insert(successful_nodes, id)
		else
			response.id = id
			table.insert(unsuccessful_nodes, { id = id, status = status, result = response })
		end
	end
	LogInfo("[OutgoingBatch] Batch action %s finished with %d successful batches and %d unsuccessful", action, #successful_nodes, #unsuccessful_nodes)
	return 200, { result = "OK", successful_nodes = successful_nodes, unsuccessful_nodes = unsuccessful_nodes }
end

local function RunBatchActionFromWeb(web, action)
	local post_data =  web.input["post_data"]
	local res, batch = pcall(json.decode, post_data)
	if not res or #batch == 0 then
		LogWarning("[OutgoingBatch] Invalid request: %s in action %s", post_data, action)
		return 400, {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	end
	return RunBatchAction(batch, action)
end

local function PauseBatches (web)
	local status, response = RunBatchActionFromWeb(web, "PAUSE")
	web.status = status
	return response
end

local function ResumeBatches (web)
	local status, response = RunBatchActionFromWeb(web, "RESUME")
	web.status = status
	return response
end

local function CloseBatches (web)
	local status, response = RunBatchActionFromWeb(web, "CLOSE")
	web.status = status
	return response
end

local function ReopenBatches (web)
	local status, response = RunBatchActionFromWeb(web, "REOPEN")
	web.status = status
	return response
end

local function DeleteBatches (web)
	local status, response = RunBatchActionFromWeb(web, "DELETE")
	web.status = status
	return response
end

local function EditBatches(web)
	local status, response = RunBatchActionFromWeb(web, "EDIT")
	web.status = status
	return response
end

local function EditBatchesProperties(web)
	local status, response = RunBatchActionFromWeb(web, "EDIT_PROPERTIES")
	web.status = status
	return response
end

local function PinBatches (web)
	local status, response = RunBatchActionFromWeb(web, "PIN")
	web.status = status
	return response
end

local function UnpinBatches (web)
	local status, response = RunBatchActionFromWeb(web, "UNPIN")
	web.status = status
	return response
end

local function CreateFromSearch(web)
	local status, response = RunBatchActionFromWeb(web, "CREATE_FROM_SEARCH")
	web.status = status
	return response
end


local function NewBatch(web)
	local status, response = RunActionFromWeb(web, "CREATE")
	web.status = status
	return response
end

local function AddAddressesToBatch(web)
	local post_data =  web.input["post_data"]
	local res, data = pcall(json.decode, post_data)
	local response, status
	if not res or #data == 0 then
		LogWarning("[AddAddressesToBatch] Invalid request: %s in action %s", post_data, "ADD_ADDRESSES")
		status = 400 
		response = {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	else
		status, response = RunBatchAction(data, "ADD_ADDRESSES")
	end
	
	web.status = status
	return response
end

local function GetBatchRequest(web, requestId)
	local request = models.Batches.GetBatchRequest(requestId)
	web.status = 200
	return { result = "OK", request = request }
end

local function ExistsInElastic(web, batchId)
	local exists = models.Batches.ExistsInElastic(inconcert.authUser.virtualContactCenter, batchId)
	web.status = 200
	return { result = "OK", exists = exists }
end

local function PostImage(web)

	local response = wsapi.response.new(200, { ["Content-Type"]= "text/plain" })
	response:write("")
	response.headers["Response"] = "defaultError"

	local request = wsapi.request_stream.new(web)
	local fileSize = tonumber(web.CONTENT_LENGTH)
	
	if fileSize <= 0 then
		LogError("[OutgoingBatch] Unable to upload image. Current file size is %d", fileSize)
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write("Invalid file data")
		return response:finish()
	end

	local res, data = pcall(json.decode, request.params.p)
	if not res or not data or not data.proxyType or not data.campaignAtVCC or not data.account or not data.batch then
		LogError("[OutgoingBatch] Unable to upload image. Incomplete file information.")
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write("Invalid file data")
		return response:finish()
	end
	data.page = data.page or ""

	if not model_perm.IsPermissionEnabledForUser( "batch_management", "batches_create", data.campaignAtVCC ) then
		LogError("[OutgoingBatch] Unable to upload image. User has not permissions.")
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write("User has not permissions")
		return response:finish()
	end
	
	LogInfo("[OutgoingBatch] Creating batch from image in campaign %s account %s page %s", 
		data.campaignAtVCC, data.account, data.page)
	
	-- aca piso la metatabla para poder encargarme yo de la leida del POST data
	setmetatable(request.POST, nil)

	local chunkSize = 2048
	local table_file = {}

	local left = fileSize
	local blocksize = math.min(chunkSize, left)
	while left > 0 do
		local chunk = web.input:read(blocksize)
		if not chunk then break end
		table.insert(table_file, chunk)
		left = left - #chunk
		blocksize = math.min(blocksize, left)
	end
	
	local fileId = uuid.new()
	local fileContent = table.concat(table_file)
	models.Batches.SetTemporalFile(fileId, fileContent)
	data.batch.file.id = fileId

	local status, data = RunAction(data, "CREATE")
	response.status = status
	response:write(json.encode(data))
	
	return response:finish()
end

local reason_messages = {
	["error sending message"] = "Error sending message.",
	["chat not found"] = "Only can send messages to users with an existent whatsapp chat session.",
	["error creating chat"] = "Error starting chat. Check phone is valid and has a whatsapp account.",
	["interaction active"] = "Can't send batch messages when exists an active conversation with the address.",
}

local function makeAddressStatusRow(address, data, message)
	local status = data.status or ""
	local timestamp = data["timestamp_" .. status] or data.timestamp or ""
	local reason = data.reason or ""
	reason = reason_messages[reason] or reason
	if timestamp ~= "" then
		local timezone_date = utils.ConvertUTCDateToDate(timestamp, nil, nil, true --[[ is full date --]])
		timestamp = timezone_date:toString("%Y-%m-%dT%H:%M:%S.000")
	end
	local text = (data.message and data.message.text) or (message and message.text) or ""
	return '"' .. address .. '","' .. timestamp .. '","' .. text .. '","' .. status .. '","' .. reason .. '"\n'
end

local function DownloadAddressesStatus(web, account, batchId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[DownloadAddressesStatus] User '%s' Downloading batch '%s' in account '%s' vcc '%s'.", 
		userId, batchId, account, vcc)
	
	local batch = models.Batches.GetBatch(account .. ":" .. batchId)
	local message = batch and batch.data or nil
	local addresses = models.Batches.GetAddressesStatus(account .. ":" .. batchId)
	local text = '"address","date","text","status","reason"\n'
	for address, data in pairs(addresses) do
		for _, data_h in pairs(data.history or {}) do
			text = text .. makeAddressStatusRow(address, data_h, message)
		end
		text = text .. makeAddressStatusRow(address, data, message)
	end
	web.headers["Content-Disposition"] = [[attachment; filename="]] .. 'addresses_status.csv"'
	return text
end

local function WhatsappGetAccountLimitData (web)
	LogDebug("[OutgoingBatch] - WhatsappGetAccountLimitData - request:'%s'", json.encode(web.request))
	local campaignAtVCC = web.request.campaignAtVCC
	local account = web.request.account
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local response = irLink:WhatsappGetAccountLimitData({
		campaignAtVCC = campaignAtVCC,
		account = account
	})
	web.status = http_mapping.getHTTPCode(response["result"])
		
	return response
end

function MapUris( m, htmlRender, prefix )
	local perm = permissions
	local p = prefix
	local Log = require("apps.agent.requestFilters").LogRequest
	local jsonValidator = utils.JsonValidator
	local vr = utils.JsonValidatorReply
	local r = utils.JsonReply

	m:dispatch_post( Log .. vr	.. SearchBatches, p ..				"/search_batches/(.*)")
	m:dispatch_post( Log .. vr	.. GetBatches, p ..					"/get_batches/(.*)")
	m:dispatch_post( Log .. vr	.. PauseBatches, p ..				"/pause_batches/(.*)")
	m:dispatch_post( Log .. vr	.. ResumeBatches, p ..				"/resume_batches/(.*)")
	m:dispatch_post( Log .. vr	.. CloseBatches, p ..				"/close_batches/(.*)")
	m:dispatch_post( Log .. vr	.. ReopenBatches, p ..				"/reopen_batches/(.*)")
	m:dispatch_post( Log .. vr	.. DeleteBatches, p ..				"/delete_batches/(.*)")
	m:dispatch_post( Log .. vr	.. EditBatches, p ..				"/edit_batches/(.*)")
	m:dispatch_post( Log .. vr	.. EditBatchesProperties, p ..		"/edit_batches_properties/(.*)")
	m:dispatch_post( Log .. vr	.. PinBatches, p ..					"/pin_batches/(.*)")
	m:dispatch_post( Log .. vr	.. UnpinBatches, p ..				"/unpin_batches/(.*)")
	m:dispatch_post( Log .. vr	.. NewBatch, p ..					"/new_batch/(.*)")
	m:dispatch_post( Log .. vr	.. AddAddressesToBatch, p ..	    	"/add_addresses_to_batch/(.*)")
	m:dispatch_post( Log .. vr	.. SearchTweets, p ..				"/search_tweets/(.*)")
	m:dispatch_post( Log .. vr	.. CreateFromSearch, p ..			"/create_from_search/(.*)")
	m:dispatch_post( Log .. vr	.. WhatsappGetAccountLimitData, p ..		"/whatsapp_get_account_limit_data/(.*)")
	m:dispatch_get ( Log .. r	.. GetBatchRequest, p ..			"/get_batch_request/(.+)")
	m:dispatch_get ( Log .. r	.. ExistsInElastic, p ..			"/exists_in_elastic/(.+)")
	m:dispatch_get ( Log ..		   DownloadAddressesStatus, p ..	"/download_addresses_status/(.+)/(.+)")
	m:dispatch_wsapi(			   PostImage, p ..					"/post_image/([^.]*)")

	m_htmlRender = htmlRender
end
