local json = require "cjson"
local utils = require "shared.apps.utils"

module(..., package.seeall)

json.encode_empty_table_as_array("on")

local m_datasource

function GetName ()
	return "Batches"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

function GetTypes()
	return { "SMS", "RCS", "WHATSAPP", "MAIL" }
end

local function IsClosed(batch)
	return batch.state == "CLOSED" or batch.state == "DELETED"
end

---
-- Helper to perform the actual search on ES.
--
function search_elastic (vcc, query, elasticSearchConnectionConfig)
	elasticSearchConnectionConfig = elasticSearchConnectionConfig or inconcert.environment.apps.agent.elasticsearch
	if not elasticSearchConnectionConfig or not elasticSearchConnectionConfig.host or not elasticSearchConnectionConfig.port then
		LogError("[BATCHES.models.search_elastic] --> Unable to execute ES Request. Invalid elasticsearch config: %s", json.encode(elasticSearchConnectionConfig or {}))
		return false, {error = "invalid elasticsearch config", status = 500}
	end

	local requestInfo = utils.GetElasticSearchInstance():MakeRequestInfo({
		query = query
	})

	local http = require "socket.http"
	local ltn12 = require "ltn12"
	local response_body = {}
	local index_name = ("vcc_batches-%s"):format(vcc)
	local url = ("/%s/_search"):format(index_name)

	LogInfo("[BATCHES.models.search_elastic] --> Executing elastic search. '%s:%s - url: %s'\nPOST_DATA: %s",
		elasticSearchConnectionConfig.host,
		elasticSearchConnectionConfig.port,
		url,
		requestInfo.encodedData
	)

	local ok, response_code, response_headers, response_status_line = http.request({
		host = elasticSearchConnectionConfig.host,
		port = elasticSearchConnectionConfig.port,
		url = url,
		headers = requestInfo.headers,
		method = "POST",
		source = requestInfo.source,
		sink = ltn12.sink.table(response_body),
		redirect = false
	})
	if not ok then
		LogError("[BATCHES.models.search_elastic]  %s\n%s\n%s\n%s", response_code, json.encode(response_headers), response_status_line, json.encode(response_body))
		return false, { error = response_status_line, status = response_code }
	end

	local raw_response = table.concat(response_body)
	if not response_headers["content-type"]:match("^application/json") then
		LogError("[BATCHES.models.search_elastic] Unknown response from elasticsearch. headers: %s\nresponse: %s", json.encode(response_headers), json.encode(raw_response))
		return false, { error = "inconcert unknown", status = 500 }
	end
	local response = json.decode(raw_response)
	if response.error then
		LogError("[BATCHES.models.search_elastic] Request failed with the following error: '%s'", json.encode(response.error))
		return false, response
	end

	--LogDebug("[BATCHES.models.search_elastic] -> Ok search. response: %s", json.encode(response))
	return true, response
end

function ExistsInElastic (vcc, id)
	local query = {
		query = {
			bool = {
				filter = { term = { _id = id } } 
			}
		}
	}
	local status, result = search_elastic(vcc, query)
	return status and result and result.hits and result.hits.total.value == 1
end

function GetAccountStatistics(id, ks, datasource)
	datasource = datasource or m_datasource
	local release = false
	if not ks then
		release = true
		ks = datasource.cassandra:CassandraConnection("Interactions")
	end

	local data = ks:GetColumnValue(id, "OutgoingBatchesAccounts", "statistics")
	local statistics = data and json.decode(data) or nil

	if release then
		ks:Close()
	end
	return statistics
end

function GetBatchStatistics(id, ks, datasource)
	datasource = datasource or m_datasource
	local release = false
	if not ks then
		release = true
		ks = datasource.cassandra:CassandraConnection("Interactions")
	end

	local data = ks:GetColumnValue(id, "OutgoingBatches", "statistics")
	local statistics = data and json.decode(data) or nil

	if release then
		ks:Close()
	end
	return statistics
end

local function GetBatchAddressesFromRedis(id, datasource)
	datasource = datasource or m_datasource
	return datasource.redis:hgetall("OutgoingBatchAddressesStatus::" .. id)
end

function GetBatch(id, includeStatistics, ks, datasource)
	local batch = GetBatchFromRedis(id, datasource) or GetBatchFromArchive(id, ks, datasource)
	if not batch or type(batch) ~= "table" then
		return nil
	end
	if includeStatistics then
		batch.statistics = GetBatchStatistics(id, ks, datasource)
	end
	local redisAddresses = GetBatchAddressesFromRedis(id, datasource)
	batch.addressesStatus = not utils.TableIsEmpty(redisAddresses) and redisAddresses or batch.addressesStatus or {}
	return batch
end

--retrieves a full batch JSON from redis
function GetBatchFromRedis (id, datasource)
	datasource = datasource or m_datasource
	local data = datasource.redis:get("OutgoingBatch:"..id)
	if not data or data == "" then
		return
	end
	local res, decoded = pcall(json.decode, data)
	if res and decoded then
		return decoded
	else
		LogError("batch '%s' detail is undecodable: '%s'", batchId, data)
	end
end

function GetBatchFromArchive (id, ks, datasource)
	datasource = datasource or m_datasource
	local release = false
	if not ks then
		release = true
		ks = datasource.cassandra:CassandraConnection("Interactions")
	end

	local data = ks:GetColumnValue(id, "OutgoingBatches", "original")
	local batch = data and json.decode(data) or nil

	if release then
		ks:Close()
	end
	return batch
end

function GetAddressesStatus (id, ks, addresses)
	return utils.GetAllColumns {
		keyspace = ks or "Interactions",
		columnFamily = "OutgoingBatchesAddresses",
		key = id,
		columns = addresses
	} or {}
end

function GetAddressStatus (id, address, ks, datasource)
	datasource = datasource or m_datasource
	local release = false
	if not ks then
		release = true
		ks = datasource.cassandra:CassandraConnection("Interactions")
	end

	local data = ks:GetColumnValue(id, "OutgoingBatchesAddresses", address)
	local address = data and json.decode(data) or nil

	if release then
		ks:Close()
	end
	return address
end

function GetBatchRequest (id, datasource)
	datasource = datasource or m_datasource
	local data = datasource.redis:get("OutgoingBatchRequest:"..id)
	if not data or data == "" then
		return
	end
	local res, decoded = pcall(json.decode, data)
	if res and decoded then
		return decoded
	else
		LogError("Batch request '%s' is undecodable: '%s'", id, data)
	end
end

function SetTemporalFile (id, data, datasource)
	datasource = datasource or m_datasource
	datasource.redis:set("OutgoingBatchTemporalFile:" .. id, data)
	datasource.redis:expire("OutgoingBatchTemporalFile:" .. id, 180)
end

function GetTemporalUploadedFile (file_info, datasource)
	datasource = datasource or m_datasource
	return datasource.redis:get("OutgoingBatchTemporalFile:" .. id)
end

function DeleteTemporalUploadedFile (file_info, datasource)
	datasource = datasource or m_datasource
	datasource.redis:del("OutgoingBatchTemporalFile:" .. id)
end
