/*  jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:true, unused:true, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/*global $, inConcert*/

if (!this.inConcert) {
	this.inConcert = {};
}

(function () {
/*jshint validthis:true*/
"use strict";

	var path = "/inconcert/apps/batch_management/";
	var format = inConcert.Utils.format;
	
	var BATCH_REQUEST_POOLING = 1000;
	var BATCH_REQUEST_TIMEOUT = 90000;
	
	var BATCH_ELASTIC_POOLING = 1000;
	var BATCH_ELASTIC_TIMEOUT = 10000;

	function BatchesManager () {
		const self = this;
		this.dataProcessors = {};
		this.m_api = inConcert.app.getBatchesMngAPI();
		this.fileUpload = new inConcert.FileUpload();
		$.each(inConcert.batchManagementData, function(type, constructor) {
			if (type !== "base") {
				self.dataProcessors[type] = new constructor();
			}
		});
	}
	
	if (typeof inConcert.batchManagementManager !== "function") {
		inConcert.batchManagementManager = BatchesManager;
	}

	BatchesManager.prototype.GetAccountConfiguration = function(campaignId, accountId) {
		var campaign = inConcert.app.getCampaign(campaignId);
		if (!campaign) {
			return null;
		}
		var account = campaign.getAccount(accountId);
		return account;
	};
	
	BatchesManager.prototype.__makeUrl = function(method, controller) {
		controller = controller || "batches";
		return inConcert.app.getBaseDomain() + path + controller + "/" + method;
	};
	
	BatchesManager.prototype.BuildBatch = function(data) {
		var batchClass = inConcert.batchManagement[data.type.toLowerCase()];
		if (!batchClass) {
			throw "unsupported batch: " + JSON.stringify(data);
		}
		return new batchClass(data);
	};
	
	BatchesManager.prototype.BuildBatchStatistics = function(data) {
		var batchClass = inConcert.outgoingBatchesStatistics.batch[data.batch.type.toLowerCase()];
		if (!batchClass) {
			throw "unsupported batch: " + JSON.stringify(data);
		}
		return new batchClass(data);
	};
	
	BatchesManager.prototype.BuildAccountStatistics = function(data) {
		var statsClass = inConcert.outgoingBatchesStatistics.account[data.type.toLowerCase()];
		if (!statsClass) {
			throw "unsupported account: " + JSON.stringify(data);
		}
		
		return new statsClass(data);
	};
	
	BatchesManager.prototype.ShowCreatePopup = function(account) {
		var popupClass = inConcert.batchManagementCreate[account.type.toLowerCase()];
		if (!popupClass) {
			throw "unsupported create batch: " + JSON.stringify(account);
		}
		
		var popup = new popupClass(account);
		popup.Show();
	};
	
	BatchesManager.prototype.ShowEditPopup = function(batch) {
		var popupClass = inConcert.batchManagementEdit[batch.type.toLowerCase()];
		if (!popupClass) {
			throw "unsupported edit batch: " + JSON.stringify(batch);
		}
		
		var popup = new popupClass(batch);
		popup.Show();
	};

	BatchesManager.prototype.ShowAddAddressesPopup = function(batch) {
		var popupClass = inConcert.batchManagementAddresses[batch.type.toLowerCase()];
		if (!popupClass) {
			throw "unsupported add addresses to batch: " + JSON.stringify(batch);
		}
		
		var popup = new popupClass(batch);
		popup.Show();
	};
	
	BatchesManager.prototype.ShowReschedulePopup = function(batch) {
		var popupClass = inConcert.batchManagementReschedule[batch.type.toLowerCase()];
		if (!popupClass) {
			throw "unsupported edit batch: " + JSON.stringify(batch);
		}
		
		var popup = new popupClass(batch);
		popup.Show();
	};
	
	BatchesManager.prototype.ShowSetAttributesPopup = function(batch) {		
		var popup = new inConcert.batchManagementSetAttributes(batch);
		popup.Show();
	};
	
	BatchesManager.prototype.CanCreateNewBatch = function(campaignId, accountId) {
		if (!inConcert.permissions.to_action("batch_management", "batches_create", null, campaignId)) {
			return {success: false, error: "<%=_T('You have no permissions to create batches.')%>"};
		}
		var account = this.GetAccountConfiguration(campaignId, accountId);
		if (account && account.state && account.state !== "0") {
			return {success: true};
		}

		return {success: false, error: "<%=_T('The account is disabled.')%>"};
	};
	
	/**
	 * Search batches by conditions
	 *
	 * @public
	 *
	 * @param  {Object}			conditions
	 *		   {string}			conditions.view				- ACTIVE|SCHEDULED|ARCHIVE		required
	 *		   {object}			conditions.query											required
	 *		   {string|array}	conditions.query.type		- Batch type					optional
	 *		   {string|array}	conditions.query.campaign									optional
	 *		   {string|array}	conditions.query.account									optional
	 *		   {string|array}	conditions.query.page										optional
	 *		   {string|array}	conditions.query.created_by_user							optional
	 *		   {string|array}	conditions.query.content	- batch text					optional
	 *		   {array}			conditions.sort				- Sort conditions				optional	Sample: [ { publish_timestamp : { order : "desc" } } ]
	 *		   {bool}			conditions.exact			- Find by text exactly			optional
	 *		   {bool}			conditions.highlight		- Return highlights				optional
	 *		   {integer}		conditions.rows				- Number of rows to return		optional
	 *		   {integer}		conditions.from				- Index of first row 			optional
	 *
	 * @param  {Function}		callback					- Callback function
	 */
	BatchesManager.prototype.SearchBatches = function(conditions, callback = $.noop) {
		const self = this;

		return this.m_api.searchBatches(conditions, function(success, response) {
			if (!success) {
				callback(false, response);
				return;
			}
			let batches = [];
			$.each(response.batches || [], function(_, batch) {
				batches.push(self.BuildBatch(batch));
			});
			response.batches = batches;
			callback(response);
		});
	};

	BatchesManager.prototype.GetBatchActionErrorMessage = function(result, action) {
		if (result.status === 403) {
			return format("<%=_T('You have no permissions to {0}.')%>", action? action.name : "");
		}
		result.result = result.result || {};
		if (result.result.reason === "interactions open") {
			return "<%=_T('Batch has open interactions.')%>";
		}
		if (result.result.result === "OPEN_LIMIT") {
			switch(action.id) {
				case "REOPEN":
					return "<%=_T('Open batches limit exceeded.')%>";
			}
		}
		return "<%=_T('Server error running action.')%>";
	};
	
	/**
	 * Run batch action
	 *
	 * @public
	 *
	 * @param  {Array<Batch>}	batches						- batches to apply action
	 * @param  {string}			actionName					- Name of action to run
	 * @param  {Function}		callback					- Callback function
	 * @param  {object}			data						- Extra data to send in batch action
	 * @param  {bool}			fromList					- Indicates if action is run from list
	 */	
	BatchesManager.prototype.RunBatchAction = function(batches, actionName, callback, data, fromList) {
		callback = callback || $.noop;
		data = data || {};
		var self = this;
		var batchEntities = [];
		var error = null;
		var action = inConcert.batchManagementActions.all[actionName];
		if (!action) {
			callback(false, "<%=_T('Unexpected error running action.')%>");
			return;
		}
		$.each(batches, function(_, batch) {
			if (!batch.CanRunAction(actionName)) {
				callback(false, format("<%=_T('You have no permissions to {0}.')%>", action? action.name : actionName));
				error = true;
				return false;
			}
			var batchData = {
				proxyType		: batch.type,
				campaignAtVCC	: batch.campaign,
				account			: batch.account,
				batchId			: batch.id
			};
			$.each(data, function(k, v) {
				batchData[k] = v;
			});
			batchEntities.push(batchData);
		});
		if (error) {
			return;
		}
		
		if (fromList && action.batchPopup) {
			self[action.batchPopup](batches);
			callback(true, "ABORTED");
			return;
		}

		function onFinished(status, result) {
			self.UpdateBatches(batches, function() {
				callback(status, result);
			});
		}

		function runAction() {
			// TODO: Todas las invocaciones de AJAX deberian estar en un nuevo metodo del archivo del API "batches_mng_api.js". Ahora no se realiza, ya que sino hay que revisar todo por completo.
			inConcert.invoke({
				url: self.__makeUrl(action.method),
				type: "POST",
				data: JSON.stringify(batchEntities),
				dataType: "json",
				contentType: "application/json",
				success: function(response) {
					onFinished(true, response);
				},
				error: function(/*_, error*/) {
					onFinished(false, "<%=_T('Server error running action.')%>");
				}
			});
		}

		if (action.confirmation) {
			$.inconcertPopup.Show(
				action.confirmation.title,
				action.confirmation.text,
				$.inconcertPopup.ButtonSets.BS_YESNO,
				function(response) {
					if (response.toLowerCase() !== "yes") {
						return callback(true, "ABORTED");
					}
					runAction();
				},
				undefined,
				{
					popup_id        : "conf_inconcertPopup",
					container_class : "confirmationPopUp",
					overlay_class   : "confirmationOverlay"
				}
			);
		}
		else {
			runAction();
		}
	};
	
	BatchesManager.prototype.__getBatchRequestError = function(actionName, request) {
		if (!request.status) {
			return "<%=_T('Unexpected error.')%>";
		}
		switch(actionName) {
			case "CREATE":
				switch(request.status) {
					case "TIMEOUT":
						return "<%=_T('Timeout creating batch.')%>";
					default:
						return "<%=_T('Error creating batch.')%>";
				}
		}
		return "<%=_T('Unexpected error.')%>";
	};
 	
	BatchesManager.prototype.__processBatchSingleRequest = function(requestId, callback, time) {
		callback = callback || $.noop;
		time = time || 0;
		var self = this;
		if (time > BATCH_REQUEST_TIMEOUT) {
			return callback(false, self.__getBatchRequestError("CREATE", { status : "TIMEOUT" }));
		}
		time += BATCH_REQUEST_POOLING;
		// TODO: Todas las invocaciones de AJAX deberian estar en un nuevo metodo del archivo del API "batches_mng_api.js". Ahora no se realiza, ya que sino hay que revisar todo por completo.
		inConcert.invoke({
			url: this.__makeUrl("get_batch_request/" + encodeURIComponent(requestId)),
			type: "GET",
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				var request = response.request || {};
				switch(request.status) {
					case "OK":
						return callback(true, request.message);
					case "PENDING":
						return window.setTimeout(function() {
							self.__processBatchSingleRequest(requestId, callback, time);
						}, BATCH_REQUEST_POOLING);
					default:
						return callback(false, self.__getBatchRequestError("CREATE", request));
				}
			},
			error: function(/*_, error*/) {
				callback(false, self.__getBatchRequestError("CREATE", {}));
			}
		});
	};
 	
	BatchesManager.prototype.__existsInElastic = function(batchId, callback, time) {
		callback = callback || $.noop;
		time = time || 0;
		var self = this;
		if (time > BATCH_ELASTIC_TIMEOUT) {
			return callback(false);
		}
		time += BATCH_ELASTIC_POOLING;
		// TODO: Todas las invocaciones de AJAX deberian estar en un nuevo metodo del archivo del API "batches_mng_api.js". Ahora no se realiza, ya que sino hay que revisar todo por completo.
		inConcert.invoke({
			url: this.__makeUrl("exists_in_elastic/" + encodeURIComponent(batchId)),
			type: "GET",
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				if (response.exists) {
					return callback(true);
				}
				window.setTimeout(function() {
					self.__existsInElastic(batchId, callback, time);
				}, BATCH_REQUEST_POOLING);
			},
			error: function(/*_, error*/) {
				callback(false);
			}
		});
	};
	
	BatchesManager.prototype.MakeImageUploadUrl = function(account) {
		return inConcert.app.getBaseDomain() + path + 
			format("image_upload/?type={0}&campaign={1}&account={2}&page={3}" ,
				encodeURIComponent(account.type),
				encodeURIComponent(account.campaign),
				encodeURIComponent(account.account),
				encodeURIComponent(account.page));
	};
	
	BatchesManager.prototype.__CreateBatchFromImage = function(data, callback) {
		var url = this.__makeUrl("post_image/");
		var file = data.batch.file;
		data.batch.file = {
			name : file.name,
			type : file.type
		};
		this.fileUpload.uploadImage(url, file, data, function(status, data) {
			if (status) {
				callback(true, JSON.parse(data));
			}
			else {
				callback(false, "<%=_T('Error uploading image.')%>");
			}
		});
	};

	BatchesManager.prototype.CreateBatch = function(account, data, isScheduled, callback) {
		callback = callback || $.noop;
		var self = this;
		if (!self.CanCreateNewBatch(account.campaign, account.account).success) {
			return callback(false, "<%=_T('You have no permissions to create batches.')%>");
		}
		var batchData = {
			proxyType : account.type,
			campaignAtVCC : account.campaign,
			account : account.account,
			page : account.page || "",
			batch : data
		};

		function makeErrorMessage(result) {
			switch(result && result.result || "") {
				case "OPEN_LIMIT":
					return "<%=_T('Open batches limit exceeded.')%>";
				case "ALREADY_EXISTS":
					return "<%=_T('Already exists a batch with same id.')%>";
				default:
					return "<%=_T('Server error running action.')%>";
			}
		}

		function processResult(status, data) {
			if (status && data) {
				self.__existsInElastic(data, function() {
					inConcert.batch_management.MoveToCategory("ACTIVE", account, function() {
						inConcert.batch_management.RefreshList(data);
						callback(status, data);
					});
				});
			}
			else {
				callback(status, data);
			}
		}

		function processResponse(status, response) {
			if (status) {
				if (response.requestId) {
					return self.__processBatchSingleRequest(response.requestId, function(status, data) {
						if (status) {
							data = account.account + ":" + data;
						}
						processResult(status, data);
					});
				}
				else if (response.async) {
					//#16277: Llamo al processResult para liberar la interface en cuanto llega la respuesta
					processResult(response.result === "OK", response.result === "OK" ? response.batchId : makeErrorMessage(response));
					inConcert.app.addPollerEventCallback("async_new_outgoing_batch_result", response.batchId, function(result) {
						if (result.result === "OK") {
							processResult(true, response.batchId);
						}
						else {
							processResult(false, makeErrorMessage(result));
						}
					});
				}
				else {
					processResult(true, response.batchId);
				}
			}
			else {
				processResult(false, response);
			}
		}

		if (data.file) {
			return this.__CreateBatchFromImage(batchData, processResponse);
		}
		// TODO: Todas las invocaciones de AJAX deberian estar en un nuevo metodo del archivo del API "batches_mng_api.js". Ahora no se realiza, ya que sino hay que revisar todo por completo.
		return inConcert.invoke({
			url: this.__makeUrl("new_batch/"),
			type: "POST",
			data: JSON.stringify(batchData),
			dataType: "json",
			contentType: "application/json",
			timeout: 300000, // 5 minutes
			success: function(response) {
				processResponse(true, response);
			},
			error: function(_, error) {
				processResult(false, makeErrorMessage(error));
			}
		});
	};
	
	BatchesManager.prototype.UpdateBatches = function(batches, callback) {
		callback = callback || $.noop;
		var ids = [];
		$.each(batches, function(_, batch) {
			ids.push(batch.GetId());
		});
		// TODO: Todas las invocaciones de AJAX deberian estar en un nuevo metodo del archivo del API "batches_mng_api.js". Ahora no se realiza, ya que sino hay que revisar todo por completo.
		return inConcert.invoke({
			url: this.__makeUrl("get_batches/"),
			type: "POST",
			data: JSON.stringify(ids),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				$.each(batches, function(_, batch) {
					var data = response[batch.GetId()];
					if (data) {
						batch.Update(data);
					}
				});
				//update batches list
				inConcert.batch_management.UpdateBatches();
				callback(true);
			},
			error: function(_, error) {
				//update batches list
				inConcert.batch_management.UpdateBatches();
				callback(false, error);
			}
		});
	};
	
	BatchesManager.prototype.SearchTweets = function(campaign, account, conditions, callback) {
		callback = callback || $.noop;
		// TODO: Todas las invocaciones de AJAX deberian estar en un nuevo metodo del archivo del API "batches_mng_api.js". Ahora no se realiza, ya que sino hay que revisar todo por completo.
		return inConcert.invoke({
			url: this.__makeUrl("search_tweets/"),
			type: "POST",
			data: JSON.stringify({
				campaignAtVCC : campaign,
				account : account,
				conditions : conditions
			}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				callback(true, response.tweets && JSON.parse(response.tweets) || [], response.limit, response.reset_time);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});
	};
	
	BatchesManager.prototype.HasPermission = function(campaign, permission) {
		if (campaign) {
			return inConcert.permissions.to_action("batch_management", permission, null, campaign);
		}
		return inConcert.permissions.to_action_on_any_campaign("batch_management", permission);
	};
	
	BatchesManager.prototype.GetAccount = function(campaignId, accountId/*, pageId*/) {
		var account = this.GetAccountConfiguration(campaignId, accountId);
		if (!account) {
			return null;
		}
		switch(account.proxyType) {
			case "SMS":
				return {
					name		: account.name, 
					className	: "accountItemSms",
					type		: account.proxyType,
					campaign	: campaignId,
					account		: account.id,
					page		: ""
				};
			case "RCS":
				return {
					name		: account.name, 
					className	: "accountItemRcs",
					type		: account.proxyType,
					campaign	: campaignId,
					account		: account.id,
					page		: ""
				};
			case "WHATSAPP":
				return {
					name		: account.name, 
					className	: "accountItemWhatsapp",
					type		: account.proxyType,
					campaign	: campaignId,
					account		: account.id,
					page		: ""
				};
			case "MAIL":
				return {
					name		: account.name, 
					className	: "accountItemMail",
					type		: account.proxyType,
					campaign	: campaignId,
					account		: account.id,
					page		: ""
				};
		}
		return null;
	};
	
	BatchesManager.prototype.GetAccountName = function(campaignId, accountId, pageId) {
		var account = this.GetAccount(campaignId, accountId, pageId);
		return account? account.name : "";
	};
	
	BatchesManager.prototype.__getCampaignAccounts = function(campaign, permission) {
		var self = this;
		var accounts = [];
		if (permission && !self.HasPermission(campaign.Id, permission)) {
			return accounts;
		}
		$.each(campaign.accounts, function(_, account) {
			switch(account.proxyType) {
				case "SMS":
					accounts.push({ 
						name		: account.name, 
						className	: "accountItemSms",
						type		: account.proxyType,
						account		: account.id,
						page		: ""
					});
					break;
				case "RCS":
					accounts.push({ 
						name		: account.name, 
						className	: "accountItemRcs",
						type		: account.proxyType,
						account		: account.id,
						page		: ""
					});
					break;
				case "WHATSAPP":
					accounts.push({ 
						name		: account.name, 
						className	: "accountItemWhatsapp",
						type		: account.proxyType,
						account		: account.id,
						page		: ""
					});
					break;
				case "MAIL":
					accounts.push({ 
						name		: account.name, 
						className	: "accountItemMail",
						type		: account.proxyType,
						account		: account.id,
						page		: ""
					});
					break;
			}
		});
		return accounts;
	};

	BatchesManager.prototype.GetAccounts = function(sorted, permission) {
		var self = this;
		var accounts = [];
		var campaigns = inConcert.app.getCampaigns();
		$.each(campaigns, function(_, campaign) {
			var campaignAccounts = self.__getCampaignAccounts(campaign, permission);
			$.each(campaignAccounts, function(_, account) {
				account.campaign = campaign.Id;
				accounts.push(account);
			});
		});
		if (sorted) {
			accounts.sort(function(a, b) {
				if (a.type !== b.type) {
					return a.type.localeCompare(b.type);
				}
				return a.name.localeCompare(b.name);
			});
		}
		return accounts;
	};

	BatchesManager.prototype.GetDataProcessor = function(type) {
		var processor = this.dataProcessors[type.toLowerCase()];
		if (!processor) {
			throw "unsupported type: " + type;
		}
		return processor;
	};

	BatchesManager.prototype.GetBatchText = function(type, data) {
		return this.GetDataProcessor(type).GetText(data);
	};

	BatchesManager.prototype.GetBatchContent = function(type, data) {
		return this.GetDataProcessor(type).GetContent(data);
	};

	BatchesManager.prototype.DownloadBatchAddressesStatus = function(batch) {
		if (!batch.HasAddresses()) {
			return;
		}
		var url = this.__makeUrl("download_addresses_status/") + 
				encodeURIComponent(batch.account) + "/" + encodeURIComponent(batch.id);
		window.open(url, "_blank");
	};

	BatchesManager.prototype.WhatsappGetAccountLimitData = function(data, callback) {
		callback = callback || $.noop;

		return this.m_api.WhatsappGetAccountLimitData(data, function(success, response) {
			callback(success, response);
		});		
	};

}());