module(..., package.seeall, inconcert.orbit.application.Decorate)
local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local m_rasa = require("shared.rasa.rasa")
local apiRasa = m_rasa.GetAPI() -- rocka

m_htmlRender = nil
m_appLocation = nil

local function LowerAttr(row, attrName)
	attrName = attrName or "name"
	return row and row[attrName] and row[attrName]:lower() or ""
end

-------------------------------------------------------
-- APPS
-------------------------------------------------------

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/get/17c545c3-98cd-4e80-8201-15e8d24c4234

	Metodo: GET
		Parametros:
			1) App Id: 17c545c3-98cd-4e80-8201-15e8d24c4234

	Respuesta:
		Status:200 OK

		{"id":"17c545c3-98cd-4e80-8201-15e8d24c4234","name":"App 2","description":"Rocio 2","culture":"es-es","usageScenario":"","domain":"","versionsCount":1,"createdDateTime":"2017-05-16T13:44:12Z","endpoints":{},"endpointHitsCount":0}

]]--
local function GetApp (web, appId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:GetApp] User '%s' Retrieving RASA app '%s'.", userId, appId)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	-- 1 - Se lo pide al API de RASA.
	local success, response = apiRasa:GetApp(appId, vcc)
	LogDebug("[RASA:GetApp] From API appId: '%s', vcc: '%s', success: '%s' app: '%s'.", appId, vcc, success, response)
	if success then
		local encodedApp = response
		return encodedApp
	else
		web.status = 400
		LogError("[RASA:GetApp] Error Retrieving RASA app: '%s' for vcc: '%s'. Error: '%s'", appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/load_all/

	Metodo: POST
		Parametros:
			{}

	Respuesta:
		Status:200 OK
			{
			  "rows": {
				"6749b29c-9065-4b98-990e-174efa9effbc": {
				  "id": "6749b29c-9065-4b98-990e-174efa9effbc",
				  "culture": "es-es",
				  "description": "Application2 Rocio",
				  "name": "Application2"
				},
				"d469528f-a521-4035-a9ea-36a346b83df8": {
				  "ID": "App 1",
				  "Culture": "es-es",
				  "Name": "App 1",
				  "Description": "Rocio 1"
				}
			  },
			  "total_rows": 2,
			  "last": ""
			}
]]--

local function LoadApps (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local startTime = os.time()

	LogInfo("[RASA:LoadApps] 1 User '%s' Loading all apps for vcc '%s'. postData: '%s'", userId, vcc, web.input["post_data"])

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local apps, totalUnfilteredRows = apiRasa:LoadApps(vcc)

	local totalAdded = 0
	local filteredRows = {}
	local postData = web.request

	LogInfo("[RASA:LoadApps] 2 - Filtering rows")

	for appId, app in pairs(apps) do
		app.Id = appId
		app.Name = app.name or appId
		if utils.MatchFilter(app, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, app)
		end
	end

	LogInfo("[RASA:LoadApps] 3 - Sorting rows")

	--include only rows that belong to logged user vcc
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	postData.__order_attr_name = "name"
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[RASA:LoadApps] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/add/

	Metodo: POST
		Data:
			{"name":"App EMMA & ROCIO","culture":"es-es"}

	Respuesta:
		Status:200 OK
		{"appId":"70bffde2-1369-49ac-a200-77f0fef50b22"}
]]--
local function AddApp (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:AddApp] Error Adding RASA app for vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.name or not info.culture then
		web.status = 400
		LogError("[RASA:AddApp] Error Adding RASA app for vcc: '%s'. Invalid Posted Information '%s'", vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	LogInfo("[RASA:AddApp] User '%s' Adding RASA app.", userId)
	info.userId = userId

	local success, response = apiRasa:AddApp(info, vcc)
	if success then
		-- Elimino caracteres del Id
		local newId = utils.SanitizeId(response)
		return {appId = newId}
	else
		web.status = 400
		LogError("[RASA:AddApp] Error Adding RASA app for vcc: '%s'. Error: '%s'", vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/rename/

	Metodo: PUT

		Data:
		{
			"id": "string", -- requerido
			"name": "string", -- requerido
			"description": "string", -- optional
		}

	Respuesta:
		Status:200 OK
		[]
]]--
local function RenameApp (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:RenameApp] Error Renaming RASA app for vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.id or not info.name then
		web.status = 400
		LogError("[RASA:RenameApp] Error Renaming RASA app for vcc: '%s'. Invalid Posted Information '%s'", vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	LogInfo("[RASA:RenameApp] User '%s' Renaming RASA app.", userId)
	info.userId = userId

	local success, response = apiRasa:RenameApp(info, vcc)
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:RenameApp] Error Renaming RASA app for vcc: '%s'. Error: '%s'", vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/delete/

	Metodo: POST
		Data:
			["48b866be-f747-45bb-8a80-332a5cf6a0f8", "f23db312-1ff6-47d5-be25-f6a739daf5ee", ...]

	Respuesta:
		Status:200 OK
		[]
]]--
local function DeleteApps (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local appsToDelete = web.request

	LogInfo("[RASA:DeleteApps] User '%s' Deleting RASA %d apps '%s'.", userId, #appsToDelete, web.input["post_data"])

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response

	for _, appId in ipairs(appsToDelete) do
		--success, response = apiRasa:DeleteApp(appId, vcc)
		success, response = apiRasa:DeleteApp(appId, vcc)
		if not success then
			web.status = 400
			LogError("[RASA:DeleteApps] Error Deleting RASA app: '%s' for vcc: '%s'. Error: '%s'", appId, vcc, json.encode(response))
			return {result = "NOT_FOUND", rasaError = response}
		end
	end

	return response -- success
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/train/

	Metodo: POST
		Data:
			{
				"appId":"3a080c31-a44d-4b67-b586-ca51b0e15446"
			}

	Respuesta:
		Status:200 OK
			null
]]--
local function TrainApp (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:TrainApp] Error RASA TRAINING app for vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.appId then
		web.status = 400
		LogError("[RASA:TrainApp] Error RASA TRAINING app for vcc: '%s'. Invalid Posted Information '%s'", vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	info.vcc = vcc
	info.userId = userId
	info.makeNLUFile = true -- para que genere el archivo .NLU antes de entrenar.
	info.restart = true -- para que reinicie el modelo de rasa (la instancia de rasa con el modelo)

	LogInfo("[RASA:TrainApp] User '%s' RASA TRAINING app '%s' vcc '%s'.", userId, info.appId, info.vcc)

	local success, trainStatus, app = apiRasa:TrainApp(info)
	trainStatus = trainStatus or {}
	if success then
		app = app or {}
		LogInfo("[RASA:TrainApp] User '%s' RASA TRAINING app '%s' vcc '%s'. Trained Status '%s'.", userId, info.appId, info.vcc, json.encode(trainStatus))
		return {trainStatus = trainStatus, app = app}
	else
		web.status = 400
		local err = {result = "NOT_FOUND", rasaError = trainStatus.status or trainStatus}
		LogError("[RASA:TrainApp] Error TRAINING RASA app '%s' for vcc: '%s'. Error: '%s'", info.appId, vcc, json.encode(err))
		return err
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/load_cultures/

	Metodo: GET
		Parametros:
			...

	Respuesta:
		Status:200 OK
		[
			{
				"name": "English",
				"code": "en-us"
			},
			{
				"name": "Chinese",
				"code": "zh-cn"
			},
			..
			..
		]
]]--
local function LoadAppCultures (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:LoadAppCultures] User '%s' Loading RASA app cultures.", userId)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response = apiRasa:LoadAppCultures()
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:LoadAppCultures] Error Loading RASA app cultures for vcc: '%s'. Error: '%s'", vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

-------------------------------------------------------
-- INTENTS - APPS
-------------------------------------------------------

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/load_all/d469528f-a521-4035-a9ea-36a346b83df8/0/0.1

	Metodo: GET

	Respuesta:
		Todos los registros, o los de la pagina solicitada.
]]--
local function LoadAllAppIntents (web, appId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:LoadAllAppIntents] User '%s' Loading RASA. All APP INTENTS on appId '%s' for vcc '%s'.", userId, appId, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response = apiRasa:LoadAllAppIntents(appId, vcc)

	if success then
		local appIntents = utils.HashToArray(success)
		local total = response

		LogInfo("[RASA:LoadAllAppIntents] Loaded %d APP INTENTS on appId '%s' for vcc '%s'.", total, appId, vcc)

		-- 1 - Obtengo el index en el array del INTENT NONE
		local indexNone
		for index, intent in ipairs(appIntents) do
			if LowerAttr(intent) == "none" then
				indexNone = index
				break
			end
		end
		-- 2 - Remuevo el INTENT NONE del array y me quedo con el, para luego insertarlo en la primera posicion
		local intentNone
		if indexNone then
			intentNone = table.remove(appIntents, indexNone)
		end
		-- 3 - Ordeno el array alfabetico (sin el INTENT NONE)
		table.sort(appIntents, function (rowA, rowB)
			local aName = LowerAttr(rowA)
			local bName = LowerAttr(rowB)
			return aName < bName
		end)
		-- 4 - Se inserta en el array ordenado, el INTENT NONE en la primer posicion
		if intentNone then
			table.insert(appIntents, 1, intentNone)
		end

		local encodedAppIntents = json.encode(appIntents)
		return encodedAppIntents
	else
		web.status = 400
		LogError("[RASA:LoadAllAppIntents] Error loading RASA. All APP INTENTS on appId '%s' for vcc: '%s'. Error: '%s'", appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/get/d469528f-a521-4035-a9ea-36a346b83df8/7b1d9a5f-2804-42a1-9b58-03b07788477a

	Metodo: GET

		Parametros:
			1) App ID    : d469528f-a521-4035-a9ea-36a346b83df8
			2) Intent ID : 7b1d9a5f-2804-42a1-9b58-03b07788477a

	Respuesta:
		{"id":"d04834af-1e1b-44f1-884e-e454e7cda020","name":"INTENTS1","typeId":0,"readableType":"Intent Classifier"}
]]--
local function GetAppIntent (web, appId, intentId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:GetAppIntent] User '%s' getting RASA APP INTENTS '%s' on app: '%s'.", userId, intentId, appId)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response = apiRasa:GetAppIntent(appId, intentId)
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:GetAppIntent] Error getting RASA APP INTENTS '%s' on app: '%s', for vcc: '%s'. Error: '%s'", intentId, appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/create/

	Metodo: POST

		DATA:
			{
				"appId":"d469528f-a521-4035-a9ea-36a346b83df8",
				"name":"Intents 2.0"
			}

	Respuesta:
		{"intentId":"bdf01c0a-67c9-4b39-9ecf-f7185d0120b7"}
]]--
local function CreateAppIntent (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:CreateAppIntent] Error Creating RASA APP INTENTS for vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.appId or not info.name then
		web.status = 400
		LogError("[RASA:CreateAppIntent] Error Creating RASA APP INTENTS for vcc: '%s'. Invalid Posted Information '%s'", vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	LogInfo("[RASA:CreateAppIntent] User '%s' creating RASA APP INTENTS on app: '%s' vcc: '%s'.", userId, info.appId, vcc)

	info.userId = userId
	info.vcc = vcc
	local success, response, info = apiRasa:CreateAppIntent(info)
	if success then
		-- Elimino caracteres del Id
		local newId = utils.SanitizeId(response)
		return {intentId = newId, info = info}
	else
		local err = response
		web.status = 400
		LogError("[RASA:CreateAppIntent] Error creating RASA APP INTENTS on app: '%s', for vcc: '%s'. Error: '%s'", info.appId, vcc, json.encode(err))
		return {result = "NOT_FOUND", rasaError = err}
	end
end

--[[
	Ejemplo URL:
http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/rename/d469528f-a521-4035-a9ea-36a346b83df8/7b1d9a5f-2804-42a1-9b58-03b07788477a

	Metodo: PUT

		Parametros:
			1) App ID    : d469528f-a521-4035-a9ea-36a346b83df8
			2) Intent ID : 7b1d9a5f-2804-42a1-9b58-03b07788477a
		Data:
		{
			"name": "string" -- requerido
		}

	Respuesta:
		[]
]]--
local function RenameAppIntent (web, appId, intentId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:RenameAppIntent] Error renaming RASA APP INTENTS '%s' on app: '%s' for vcc: '%s'. Invalid Request Information '%s'", intentId, appId, vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.name then
		web.status = 400
		LogError("[RASA:RenameAppIntent] Error renaming RASA APP INTENTS '%s' on app: '%s' for vcc: '%s'. Invalid Posted Information '%s'", intentId, appId, vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	LogInfo("[RASA:RenameAppIntent] User '%s' renaming RASA APP INTENTS '%s' on app: '%s' vcc: '%s'.", userId, intentId, appId, vcc)

	info.appId = appId
	info.intentId = intentId
	info.vcc = vcc
	info.userId = userId

	local success, response = apiRasa:RenameAppIntent(info)
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:RenameAppIntent] Error renaming RASA APP INTENTS '%s' on app: '%s' for vcc: '%s'. Error: '%s'", intentId, appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/delete/d469528f-a521-4035-a9ea-36a346b83df8/7b1d9a5f-2804-42a1-9b58-03b07788477a

	Metodo: DELETE

		Parametros:
			1) App ID    : d469528f-a521-4035-a9ea-36a346b83df8
			2) Intent ID : 7b1d9a5f-2804-42a1-9b58-03b07788477a

	Respuesta:
		[]
]]--
local function DeleteAppIntent (web, appId, intentId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:DeleteAppIntent] User '%s' deleting RASA APP INTENTS '%s' on app: '%s' vcc: '%s'.", userId, intentId, appId, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response = apiRasa:DeleteAppIntent(appId, intentId, vcc)
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:DeleteAppIntent] Error deleting RASA APP INTENTS '%s' on app: '%s', for vcc: '%s'. Error: '%s'", intentId, appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/utterances/load_all/d469528f-a521-4035-a9ea-36a346b83df8/INTENT_NAME/0.1

	Metodo: GET

		Parametros:
			1) App ID    : f23db312-1ff6-47d5-be25-f6a739daf5ee
			2) Intent Name : Intent Name 1
			3) Version ID : 0.1 (opcional - por defecto 0.1)

	Respuesta:
	[
		{"entityPredictions":[],"intentPredictions":[],"tokenizedText":["1","hola","emma","como","estas","?"],"id":-8010780,"text":"1 hola emma como estas ?","intentLabel":"INTENTS2","entityLabels":[{"entityName":"ENTITY2","endTokenIndex":4,"startTokenIndex":4}]},

		{"entityPredictions":[],"intentPredictions":[],"tokenizedText":["6","hola","rocio","como","estas","?"],"id":-8001776,"text":"6 hola rocio como estas ?","intentLabel":"INTENTS1","entityLabels":[{"entityName":"ENTITY2","endTokenIndex":1,"startTokenIndex":1}]},
		..
		..
	]
]]--
local function LoadUtterancesByAppIntent (web, appId, intentId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:LoadUtterancesByAppIntent] User '%s' getting RASA APP UTTERANCES INTENTS '%s' on app: '%s' vcc: '%s'.", userId, intentId, appId, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response, utterances = apiRasa:LoadUtterancesByIntentId(appId, intentId, vcc)
	if success then
		utterances = utils.HashToArray(utterances)
		--Ordeno el array alfabetico
		local attrName = "text"
		table.sort(utterances, function (rowA, rowB)
			local aName = LowerAttr(rowA, attrName)
			local bName = LowerAttr(rowB, attrName)
			return aName < bName
		end)
		--LogDebug("[RASA:LoadUtterancesByAppIntent] App: '%s' vcc: '%s' intentId: '%s'. Utterances: %s",  intentId, appId, vcc, json.encode(utterances))
		return utterances
	else
		web.status = 400
		LogError("[RASA:LoadUtterancesByAppIntent] Error getting RASA APP UTTERANCES INTENTS '%s' on app: '%s', for vcc: '%s'. Error: '%s'", intentName, appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/utterances/add/

	Metodo: POST

		DATA:
			{
			-- Requeridos -----
			  "appId": "d469528f-a521-4035-a9ea-36a346b83df8",
			  "utterance": "1 Hola Me puede informar el saldo ?",
			  "intentName": "IntentsName1",
			}

	Respuesta:
		{"UtteranceText":"1 hola me puede informar el saldo ?","ExampleId":-8038796}
]]--
local function AddUtteranceAppIntent (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:AddUtteranceAppIntent] Error Adding RASA UTTERANCE TO INTENT on vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.appId or not info.utterance or not info.intentId then
		web.status = 400
		LogError("[RASA:AddUtteranceAppIntent] Error Adding RASA UTTERANCE TO INTENT on vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	info.userId = userId
	info.vcc = vcc

	LogInfo("[RASA:AddUtteranceAppIntent] User '%s' Adding RASA UTTERANCE TO INTENT '%s' on app: '%s' vcc: '%s'.", userId, info.intentId, info.appId, vcc)

	local success, response = apiRasa:AddUtteranceAppIntent(info)
	if success then
		local newUtteranceId = response
		local success, newUtterance, encodedUtterance = apiRasa:GetUtteranceByAppIntent(info.appId, info.intentId, newUtteranceId, info.vcc)

		if not success then
			web.status = newUtterance.code
			LogError("[RASA:AddUtteranceAppIntent] Error Adding RASA UTTERANCE TO INTENT '%s' on app: '%s', for vcc: '%s'. Error: '%s'", info.intentId, info.appId, vcc, encodedUtterance)
			return {result = "NOT_FOUND", rasaError = newUtterance}
		end

		--LogDebug("3 [AddUtteranceAppIntent] (GetUtteranceByAppIntent) encodedUtterance: %s", encodedUtterance)

		return encodedUtterance
	else
		web.status = 400
		LogError("[RASA:AddUtteranceAppIntent] Error Adding RASA UTTERANCE TO INTENT '%s' on app: '%s', for vcc: '%s'. Error: '%s'", info.intentId, info.appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/utterances/add/

	Metodo: POST

		DATA:
			{
			-- Requeridos -----
			  "appId": "d469528f-a521-4035-a9ea-36a346b83df8",
			  "utterance": "1 Hola Me puede informar el saldo ?",
			  "intentId": "IntentsName1",

			-- Opcional -----
			  "entityLabels": [
				{
				  "entityName": "EntityName1",
				-- Inicio y final de la palabra "puede"
				  "startCharIndex": 3,
				  "endCharIndex": 7
				}
			  ]
			}

	Respuesta:
		{"UtteranceText":"1 hola me puede informar el saldo ?","ExampleId":-8038796}
]]--
local function UpdateUtteranceLabels (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:UpdateUtteranceLabels] Error Updating RASA UTTERANCE TO INTENT on vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.appId or not info.utterance or not info.intentId then
		web.status = 400
		LogError("[RASA:UpdateUtteranceLabels] Error Updating RASA UTTERANCE TO INTENT on vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	info.userId = userId
	info.vcc = vcc

	LogInfo("[RASA:UpdateUtteranceLabels] User '%s' Updating RASA UTTERANCE TO INTENT '%s' on app: '%s' vcc: '%s'.", userId, info.intentId, info.appId, vcc)

	local success, response = apiRasa:UpdateUtteranceLabels(info)
	if success then
		local currentUtteranceId = response
		local success, currentUtterance, encodedUtterance = apiRasa:GetUtteranceByAppIntent(info.appId, info.intentId, currentUtteranceId, info.vcc)

		if not success then
			web.status = currentUtterance.code
			LogError("[RASA:UpdateUtteranceLabels] Error Updating RASA UTTERANCE TO INTENT '%s' on app: '%s', for vcc: '%s'. Error: '%s'", info.intentId, info.appId, vcc, encodedUtterance)
			return {result = "NOT_FOUND", rasaError = currentUtterance}
		end

		--LogDebug("3 [UpdateUtteranceLabels] (GetUtteranceByAppIntent) encodedUtterance: %s", encodedUtterance)

		return encodedUtterance
	else
		web.status = 400
		LogError("[RASA:UpdateUtteranceLabels] Error Adding RASA UTTERANCE TO INTENT '%s' on app: '%s', for vcc: '%s'. Error: '%s'", info.intentId, info.appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/intents/utterances/delete/f23db312-1ff6-47d5-be25-f6a739daf5ee/-8010780/0.1

	Metodo: DELETE

		Parametros:
			1) App ID    : f23db312-1ff6-47d5-be25-f6a739daf5ee
			2) UtteranceId ID : -8010780 (Example ID)
			3) Version ID : 0.1 (opcional - por defecto 0.1)

	Respuesta:
		[]
]]--
local function DeleteAppUtterance (web, appId, utteranceId, intentId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:DeleteAppUtterance] User '%s' deleting RASA UTTERANCE ID: '%s' on app: '%s' intentId: '%s' vcc: '%s'.", userId, utteranceId, appId, intentId, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response = apiRasa:DeleteAppUtterance(appId, utteranceId, intentId, vcc)
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:DeleteAppUtterance] Error deleting RASA UTTERANCE ID: '%s' on app: '%s' intentId: '%s', for vcc: '%s'. Error: '%s'", utteranceId, appId, intentId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

-------------------------------------------------------
-- ENTITIES - APPS (SIMPLE)
-------------------------------------------------------

--[[
	Esta función realiza 3 request para obtener las 3 listas por separado.

	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/entities/load_all/d469528f-a521-4035-a9ea-36a346b83df8

	Metodo: GET

		Parametros:
			1) App ID : d469528f-a521-4035-a9ea-36a346b83df8

	Respuesta:
		Todos los registros, o los de la pagina solicitada.
]]--
local function LoadAllAppEntities (web, appId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:LoadAllAppEntities] User '%s' Loading RASA. All APP ENTITIES (SIMPLE) on appId '%s' for vcc '%s'.", userId, appId,  vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	-------------------------------------------------------------------------------------------------------------------------
	-- 1 - Obtengo las Entidades SIMPLEs
	-------------------------------------------------------------------------------------------------------------------------
	LogInfo("[RASA:LoadAllAppEntities] User '%s' Loading SIMPLE ENTITIES on appId '%s' for vcc '%s'.", userId, appId, vcc)
	local list, total = apiRasa:LoadAllAppEntities(appId, vcc)

	local appSimpleEntities
	if list then
		appSimpleEntities = utils.HashToArray(list)
		-- Ordeno el array alfabetico
		table.sort(appSimpleEntities, function (rowA, rowB)
			local aName = LowerAttr(rowA)
			local bName = LowerAttr(rowB)
			return aName < bName
		end)
		LogInfo("[RASA:LoadAllAppEntities] Loaded %d APP SIMPLE ENTITIES on appId '%s' for vcc '%s'.", total, appId, vcc)
	else
		web.status = 400
		LogError("[RASA:LoadAllAppEntities] Error loading RASA. All APP SIMPLE ENTITIES on appId '%s' for vcc: '%s'. Error: '%s'", appId, vcc, json.encode(appSimpleEntities))
		return {result = "NOT_FOUND", rasaError = total}
	end

	return json.encode(appSimpleEntities)
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/simple/get/f23db312-1ff6-47d5-be25-f6a739daf5ee

	Metodo: GET

		Parametros:
			1) App ID    : f23db312-1ff6-47d5-be25-f6a739daf5ee

	Respuesta:
		{
		  "id": "3d73462d-98f0-4dcd-8d30-fab9e65f2e73",
		  "name": "DayOfWeek",
		  "typeId": 1,
		  "readableType": "Entity Extractor"
		}
]]--
local function GetAppEntitySimple (web, appId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:GetAppEntitySimple] User '%s' getting RASA APP SIMPLE ENTITIES '%s' on app: '%s' vcc: '%s'.", userId, entityId, appId, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response = apiRasa:GetAppEntity(appId, entityId, vcc)
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:GetAppEntitySimple] Error getting RASA APP SIMPLE ENTITIES '%s' on app: '%s', for vcc: '%s'. Error: '%s'", entityId, appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
		http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/entities/simple/create/

	Metodo: POST
		DATA:
			{
				"appId":"f23db312-1ff6-47d5-be25-f6a739daf5ee",
				"name":"Entity6"
			}

	Respuesta:
		{"entityId":"734783da-ca2f-4fac-893f-394eec461b83"}
]]--
local function CreateAppEntitySimple (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:CreateAppEntitySimple] Error Creating RASA APP SIMPLE ENTITIES for vcc: '%s'. Invalid Request Information '%s'", vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.appId or not info.name then
		web.status = 400
		LogError("[RASA:CreateAppEntitySimple] Error Creating RASA APP SIMPLE ENTITIES for vcc: '%s'. Invalid Posted Information '%s'", vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	LogInfo("[RASA:CreateAppEntitySimple] User '%s' creating RASA APP ENTITIES on app: '%s' vcc: '%s'.", userId, info.appId, vcc)
	info.vcc = vcc
	info.userId = userId

	local success, response = apiRasa:CreateAppEntity(info)
	if success then
		-- Elimino caracteres del Id
		local newId = utils.SanitizeId(response)
		local result = {entityId = newId}
		local success, encodedEntity = apiRasa:GetAppEntity(info.appId, result.entityId, vcc)
		if success then
			result.entity = json.decode(encodedEntity)
		end
		return result
	else
		web.status = 400
		LogError("[RASA:CreateAppEntitySimple] Error creating RASA APP SIMPLE ENTITIES on app: '%s', for vcc: '%s'. Error: '%s'", info.appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/entities/simple/rename/f23db312-1ff6-47d5-be25-f6a739daf5ee

	Metodo: PUT

		Parametros:
			1) App ID    : f23db312-1ff6-47d5-be25-f6a739daf5ee
		DATA:
		{
			"name": "string", -- requerido
		}
]]--
local function RenameAppEntitySimple (web, appId, entityId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local ok, info = pcall(json.decode, web.input["post_data"])
	if not ok then
		LogError("[RASA:RenameAppEntitySimple] Error Updating RASA APP SIMPLE ENTITIES '%s' on app: '%s' for vcc: '%s'. Invalid Request Information '%s'", entityId, appId, vcc, web.input["post_data"])
		web.status = 400
		return {result = "INVALID_REQUEST", code = 400}
	end

	if not info.name then
		web.status = 400
		LogError("[RASA:RenameAppEntitySimple] Error Updating RASA APP SIMPLE ENTITIES '%s' on app: '%s' for vcc: '%s'. Invalid Posted Information '%s'", entityId, appId, vcc, web.input["post_data"])
		return {result = "INVALID_REQUEST", code = 400}
	end

	LogInfo("[RASA:RenameAppEntitySimple] User '%s' Updating RASA APP SIMPLE ENTITIES '%s' on app: '%s' vcc: '%s'.", userId, entityId, appId, vcc)

	info.appId = appId
	info.entityId = entityId
	info.vcc = vcc
	info.userId = userId

	local success, response = apiRasa:RenameAppEntity(info)
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:RenameAppEntitySimple] Error Updating RASA APP SIMPLE ENTITIES '%s' on app: '%s' for vcc: '%s'. Error: '%s'", entityId, appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

--[[
	Ejemplo URL:
http://i6.inconcert.cc/inconcert/apps/cognitive_services/rasa/apps/entities/simple/delete/f23db312-1ff6-47d5-be25-f6a739daf5ee/entidad_id

	Metodo: DELETE

		Parametros:
			1) App ID    : f23db312-1ff6-47d5-be25-f6a739daf5ee
]]--
local function DeleteAppEntitySimple (web, appId, entityId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[RASA:DeleteAppEntitySimple] User '%s' deleting RASA APP SIMPLE ENTITIES '%s' on app: '%s' vcc: '%s'.", userId, entityId, appId, vcc)

	local success, response = utils.IsAvailableNLUInVCC("RASA", vcc)
	if not success then
		web.status = 400
		return {result = "NOT_AVAILABLE_NLU", err = response}
	end

	local success, response = apiRasa:DeleteAppEntity(appId, entityId, vcc)
	if success then
		return response
	else
		web.status = 400
		LogError("[RASA:DeleteAppEntitySimple] Error deleting RASA APP SIMPLE ENTITIES '%s' on app: '%s', for vcc: '%s'. Error: '%s'", entityId, appId, vcc, json.encode(response))
		return {result = "NOT_FOUND", rasaError = response}
	end
end

---------------------------------------------------
--
---------------------------------------------------
function Render (web)
	_M.response.DontCache(web)
	local env = { _T = m_htmlRender.extension.makeTextTranslator() }
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("[RASA:Render] Render form RASA with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local sections = {
		{id = "application"      , file = "frm_rasa_content_application.html"      },
		{id = "intents"          , file = "frm_rasa_content_intents.html"          },
		{id = "entities"         , file = "frm_rasa_content_entities.html"         },
		{id = "train_and_publish", file = "frm_rasa_content_train_and_publish.html"}
	}
	local page_dir = "pages/"
	local html_content = ""
	for _, section in ipairs(sections) do
		local html_file = page_dir .. section.file
		LogDebug("[RASA:Render] Open form File '%s' of Section: '%s'.", section.file, section.id)
		local html_section = m_htmlRender:RenderPageLP(html_file, web, env)
		if html_section then
			html_content = html_content ..
				"<div form_section='" .. section.id .. "' class='pageFormContainerAux' hidden>" ..
					html_section ..
				"</div>"
		else
			LogError("HTML RASA section does not exist: %s", html_file)
		end
	end

	return html_content
end

--[[
****************************************************************
	MAPEO DE URLS
****************************************************************
]]--

function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest

	local perm_view  = "view_cognitive_services"
	local perm_save  = "save_cognitive_services"
	local perm_erase = "delete_cognitive_services"

	local replyFn = utils.JsonReply
	local replyValFn = utils.JsonValidatorReply

	local function HasPermission (fn, actionId, responseFn)
		actionId = actionId or perm_save
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. permissions:CheckPermission("cognitive_services", actionId, fn)
	end

	local p = prefix
	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- APPS
	module:dispatch_post  (HasPermission(LoadApps, perm_view, replyValFn)     , p .. "/apps/load_all/"                                     )
	module:dispatch_get   (HasPermission(GetApp, perm_view)                   , p .. "/apps/get/(.*)"                                      )
	module:dispatch_post  (HasPermission(AddApp)                              , p .. "/apps/add/"                                          )
	module:dispatch_put   (HasPermission(RenameApp)                           , p .. "/apps/rename/"                                       )
	module:dispatch_post  (HasPermission(DeleteApps, perm_erase, replyValFn)  , p .. "/apps/delete/"                                       )
	module:dispatch_post  (HasPermission(TrainApp)                            , p .. "/apps/train/"                                        )
	module:dispatch_get   (HasPermission(LoadAppCultures, perm_view)          , p .. "/apps/load_cultures/"                                )

	-- APPS - ENTITIES SIMPLE
	module:dispatch_get   (HasPermission(LoadAllAppEntities, perm_view)       , p .. "/apps/entities/load_all/(.*)"                        )
	module:dispatch_get   (HasPermission(GetAppEntitySimple, perm_view)       , p .. "/apps/entities/simple/get/([^/]*)/(.*)"              )
	module:dispatch_post  (HasPermission(CreateAppEntitySimple)               , p .. "/apps/entities/simple/create/"                       )
	module:dispatch_put   (HasPermission(RenameAppEntitySimple)               , p .. "/apps/entities/simple/rename/([^/]*)/(.*)"           )
	module:dispatch_delete(HasPermission(DeleteAppEntitySimple)               , p .. "/apps/entities/simple/delete/([^/]*)/(.*)"           )

	-- APPS - INTENTS
	module:dispatch_get   (HasPermission(LoadAllAppIntents, perm_view)        , p .. "/apps/intents/load_all/(.*)"                         )
	module:dispatch_get   (HasPermission(GetAppIntent, perm_view)             , p .. "/apps/intents/get/([^/]*)/(.*)"                      )
	module:dispatch_post  (HasPermission(CreateAppIntent)                     , p .. "/apps/intents/create/"                               )
	module:dispatch_put   (HasPermission(RenameAppIntent)                     , p .. "/apps/intents/rename/([^/]*)/(.*)"                   )
	module:dispatch_delete(HasPermission(DeleteAppIntent)                     , p .. "/apps/intents/delete/([^/]*)/(.*)"                   )

	-- APPS - INTENTS / UTTERANCES
	module:dispatch_get   (HasPermission(LoadUtterancesByAppIntent, perm_view), p .. "/apps/intents/utterances/load_all/([^/]*)/(.*)"      )
	module:dispatch_post  (HasPermission(AddUtteranceAppIntent)               , p .. "/apps/intents/utterances/add/"                       )
	module:dispatch_post  (HasPermission(UpdateUtteranceLabels)               , p .. "/apps/intents/utterances/update/"                    )
	module:dispatch_delete(HasPermission(DeleteAppUtterance)                  , p .. "/apps/intents/utterances/delete/([^/]*)/([^/]*)/(.*)")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end
