local cjson = require "cjson"
local TIMEOUT_MIN = 5
local TIMEOUT_MAX = 1800

local function InvalidParameterVariable (stateName, activityId, variableName, aux, reason)
	return aux.invalidParameterActivity(stateName, activityId, "variable " .. variableName .. " " .. reason)
end

local function MakeExpression (stateName, activityId, variableName, aux, expression)
	if expression == "" then
		return InvalidParameterVariable(stateName, activityId, variableName, aux, "empty value")
	end
	return 200, expression
end

local function MakeVariable (stateName, activityId, variable, aux)
	if not variable.name then
		return aux.invalidParameterActivity(stateName, activityId, "invalid variable definition")
	end
	local text = variable.name .. " = "
	local status, value
	status, value = MakeExpression(stateName, activityId, variable.name, aux, variable.value)

	if status ~= 200 then
		return status, value
	end
	text = text .. value .. "\n"
	text = text .. 'logTrace("DEBUG", string.format("Variable ' .. aux.scapeText(variable.name) .. ' assigned with value %s", '
	text = text .. 'type(' .. variable.name .. ') == "table" and json.encode(' .. variable.name .. ') or tostring(' .. variable.name .. ')))'
	return 200, text
end

local function makeFields (fields, aux)
	local code = "{"
	for _, field in pairs(fields) do
		code = code .. string.format("\n\t{ name = '%s', value = %s, hidden = %s, sendToApplication = '%s' },", field.name, field.value, "true", "")
	end
	code = code .. "\n}"
	return code
end

local function MakeExecutable (stateName, activityId, data, aux)
	if aux.empty(data.message) or aux.empty(data.timeoutRetry) then
		return aux.invalidParameterActivity(stateName, activityId, "invalid data")
	end
	--validate limits client in code
	local timeout_n = tonumber(data.timeoutRetry)
	if timeout_n and (timeout_n < TIMEOUT_MIN or timeout_n > TIMEOUT_MAX) then
		return aux.invalidParameterActivity(stateName, activityId, "invalid timeout")
	end

	if aux.empty(data.llmProfile) then
		return aux.invalidParameterActivity(stateName, activityId, "invalid data")
	end

	local text = ""

	-- internal_consultation_flag
	local status, variableText = MakeVariable(stateName, activityId, {
		name = "internal_consultation_id",
		value = "\"" .. uuid.new() .. "\""
	}, aux)
	if status ~= 200 then
		return status, variableText
	end
	text = text .. "\n" .. variableText

	local variableText = [[
		local _fields_agent = %s
		SetReportingData(__interactionId, 'process_flow_data', nil, _fields_agent)
		logTrace('DEBUG', 'Processing activity (end)')
	]]

	local fields = {}
	table.insert(fields, { name = "internal_consultation_flag", value = "true" })
	table.insert(fields, { name = "internal_consultation_question", value = data.question })
	table.insert(fields, { name = "internal_consultation_id", value = "internal_consultation_id" })
	table.insert(fields, { name = "internal_consultation_timeout", value = data.timeoutRetry })
	table.insert(fields, { name = "internal_consultation_startdate", value = "getTime()" })
	table.insert(fields, { name = "internal_consultation_kb", value = "\"" .. data.knowledgeBase .. "\"" })

	local make_fields_result = makeFields(fields, aux)
	text = text .. "\n" .. string.format(variableText, make_fields_result)

	--process message
	local message = data.message
	if type(message) == "table" then
		message = cjson.encode(cjson.encode(message))
		text = text .. "\n" .. string.format('local _message = cjson.decode(%s)', message)
		if data.message.optionsVariable then
			text = text .. '\n_message.optionsVariable = nil'
			text = text .. '\n_message.optionsTemplate = nil'
			text = text .. '\n_message.options = {}'
			text = text .. string.format('\nlocal _options = %s', data.message.optionsVariable)
			text = text .. '\nfor _, _option in pairs(_options or {}) do'
				text = text .. string.format('\nlocal _element = cjson.decode(%s)\t', cjson.encode(cjson.encode(data.message.optionsTemplate)))
				text = text .. '\nsetTemporalVariable("element", _option)'
				text = text .. '\nprocessBrackets(_element)'
				text = text .. '\ntable.insert(_message.options, _element)'
			text = text .. '\nend'
			text = text .. '\nremoveTemporalVariable("element")'
		end
		text = text .. '\nprocessBrackets(_message)'
		if message.writingWait then
			text = text .. string.format('\n_message.writingWait = %s', message.writingWait)
		end
	else
		text = text .. string.format('local _message = %s', message)
	end
	--error message
	local errorMessage = data.errorMessage
	if errorMessage then
		if type(errorMessage) == "table" then
			errorMessage = cjson.encode(cjson.encode(errorMessage))
			text = text .. string.format('local _errorMessage = cjson.decode(%s)', errorMessage)
			text = text .. '\nprocessBrackets(_errorMessage)'
		else
			text = text .. string.format('local _errorMessage = %s', errorMessage)
		end
	end
	-- validate limits in runtime
	text = text .. string.format("\nassert(%s >= %d and %s <= %d, 'invalid timeout value')", 
		data.timeoutRetry, TIMEOUT_MIN, data.timeoutRetry, TIMEOUT_MAX)

	--run method
	text = text .. string.format(
		"\nAskForHelp(__interactionId, _message, { ".. 
			"timeout = %s, " .. 
			"errorMessage = _errorMessage, " ..
			"llmProfile = \"" ..data.llmProfile.. "\"," ..
			"question = " ..data.question.. "," ..
			"knowledgeBase = \"" ..data.knowledgeBase.. "\"" ..
		"}, __vccName)",
		data.timeoutRetry)
	
	return 200, text
end

return {
	MakeExecutable = MakeExecutable
}
