local cjson = require "cjson"
require "uuid"

local searchMethods = {
	contact_id = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
	{
		contactId = %s,
	}
]],
			aux.valueOrNil(data.contactId)
			)
		end,
		makeMethod = function(data)
			return "GetOutboundContact"
		end,
	},
	external_id = {
		validate = function(data, aux)
			return not aux.empty(data.externalId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
	{
		externalId = %s,
		campaignId = %s,
	}
]],
			aux.valueOrNil(data.externalId),
			aux.valueOrNil(data.campaignDefinition)
			)
		end,
		makeMethod = function(data)
			return "GetOutboundContactByExternalId"
		end,
	},
	phone_number = {
		validate = function(data, aux)
			return not aux.empty(data.searchPhone)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
	{
		addressType = "Phone",
		address = %s,
		campaignId = %s,
	}
]],
			aux.valueOrNil(data.searchPhone),
			aux.valueOrNil(data.campaignDefinition)
			)
		end,
		makeMethod = function(data)
			return "GetOutboundContactByAddress"
		end,
	},
}

local methods = {
	search_contact = {
		validate = function(data, aux)
			return searchMethods[data.searchCriteria].validate(data, aux)
		end,
		makeParams = function(data, aux)
			return searchMethods[data.searchCriteria].makeParams(data, aux)
		end,
		makeMethod = function(data)
			return searchMethods[data.searchCriteria].makeMethod(data, aux)
		end,
	},
	create_contact = {
		validate = function(data, aux)
			return not aux.empty(data.name)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
	{
		contactId = uuid.new(),
		mix = true,
		contact = {
			ImportId = "ProcessFlow",
			Name = %s,
			ExternalId = %s,
			Addresses = %s,
			NameValuesSimple = %s,
			NameValuesSearchDate = %s,
			NameValuesSearchNumeric = %s,
			NameValuesSearchText = %s,
		}
	}
]],
			aux.valueOrNil(data.name),
			aux.valueOrNil(data.externalId),
			aux.makeDialerAddresses(data.contactPhones or {}),
			aux.makeDialerNameValues(data.contactAttributes or {}, "simple"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "date"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "number"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "text")
			)
		end,
		makeMethod = function(data)
			return "StoreOutboundContact"
		end,
	},
	edit_contact = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
	{
		contactId = %s,
		mix = true,
		contact = {
			Name = %s,
			ExternalId = %s,
			Addresses = %s,
			NameValuesSimple = %s,
			NameValuesSearchDate = %s,
			NameValuesSearchNumeric = %s,
			NameValuesSearchText = %s,
		}
	}
]],
			aux.valueOrNil(data.contactId),
			aux.valueOrNil(data.name),
			aux.valueOrNil(data.externalId),
			aux.makeDialerAddresses(data.contactPhones or {}),
			aux.makeDialerNameValues(data.contactAttributes or {}, "simple"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "date"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "number"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "text")
			)
		end,
		makeMethod = function(data)
			return "StoreOutboundContact"
		end,
	},
}

local function makeParams(data, aux)
	local code = string.format(
[[
{
	type = "Phone",
	country = %s,
	area = %s,
	number = %s,
}
]],
	aux.valueOrNil(data.country),
	aux.valueOrNil(data.areaCode),
	aux.valueOrNil(data.phone)
	)
	return code
end

local function makeResult(data, aux)
	local code = 'logTrace("DEBUG", string.format("Processing activity (end) result: %s", cjson.encode(_result)))'
	if not aux.empty(data.result) then
		code = code .. string.format('\n\n%s = _result', data.result)
	end
	return code
end

local function MakeExecutable (stateName, activityId, data, aux)
	if not methods[data.contactManagementMethod].validate(data, aux) then
		return aux.invalidParameterActivity(stateName, activityId, "invalid data")
	end

	--run method
	local code = string.format(
[[
local _params = %s
_result = DialerAction("%s", _params)
%s
]],
	methods[data.contactManagementMethod].makeParams(data, aux),
	methods[data.contactManagementMethod].makeMethod(data, aux),
	makeResult(data, aux)
	)
	return 200, code
end

return {
	MakeExecutable = MakeExecutable
}
