local cjson = require "cjson"
require "uuid"

local function makePhoneCode(aux, phoneType, phone)
	if aux.empty(phone) then return "nil" end

	return string.format(
[[
{
	Type = "Phone",
	Channels = { "CALL" },
	Kind = "%s",
	Number = %s,
}
]],
	aux.valueOrNil(phoneType),
	aux.valueOrNil(phone)
	)
end

local function getPriority(data)
	local mode = data.priorityType
	if mode == "priority_relative" then
		return "relative", data.priority_relativeNumber_input
	elseif mode == "incremental" then
		return "increment", data.priority_absoluteNumber
	else
		return "absolute", data.priority_absoluteNumber
	end
end

local methods = {
	reschedule_contact = {
		validate = function(data, aux)
			return not aux.empty(data.contactId) or
				   not aux.empty(data.campaign) or
				   not aux.empty(data.lot)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
{
	campaignId = %s,
	batchId = %s,
	contactId = %s,
	configuration = {
		scheduleDate = %s,
		agent = %s,
		address = %s,
	}
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.lot),
			aux.valueOrNil(data.contactId),
			aux.valueOrNil(data.scheduledDate),
			aux.valueOrNil(data.agent),
			makePhoneCode(aux, data.phoneType, data.phone)
			)
		end,
		makeMethod = function(data)
			return "RescheduleOutboundContact"
		end,
	},
	add_to_batch = {
		validate = function(data, aux)
			return not aux.empty(data.contactId) or
				   not aux.empty(data.campaign) or
				   not aux.empty(data.lot) or
				   not aux.empty(data.name)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
	{
		campaignId = %s,
		batchId = %s,
		contactId = %s,
		configuration = {
			scheduleDate = %s,
			agent = %s,
			address = %s,
			contactData = {
				Name = %s,
				ExternalId = %s,
				Addresses = %s,
				NameValuesSimple = %s,
				NameValuesSearchDate = %s,
				NameValuesSearchNumeric = %s,
				NameValuesSearchText = %s,
			},
		}
	}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.lot),
			aux.valueOrNil(data.contactId),
			aux.valueOrNil(data.scheduledDate),
			aux.valueOrNil(data.agent),
			makePhoneCode(aux, data.phoneType, data.phone),
			aux.valueOrNil(data.name),
			aux.valueOrNil(data.externalId),
			aux.makeDialerAddresses(data.contactPhones or {}),
			aux.makeDialerNameValues(data.contactAttributes or {}, "simple"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "date"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "number"),
			aux.makeDialerNameValues(data.contactAttributes or {}, "text")
			)
		end,
		makeMethod = function(data)
			return "AddContactToOutboundBatch"
		end,
	},
	cancel_contact = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
{
	campaignId = %s,
	contactId = %s,
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.contactId)
			)
		end,
		makeMethod = function(data)
			return "CancelOutboundContact"
		end,
	},
	recycle_contact = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
{
	campaignId = %s,
	contactId = %s,
	configuration = {
		scheduleDate = %s,
		agent = %s,
		restartRetries = %s,
	}
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.contactId),
			aux.valueOrNil(data.scheduledDate),
			aux.valueOrNil(data.agent),
			aux.booleanValue(data.restartAttempts)
			)
		end,
		makeMethod = function(data)
			return "RecycleOutboundContact"
		end,
	},
	pause_contact = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
{
	campaignId = %s,
	contactId = %s,
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.contactId)
			)
		end,
		makeMethod = function(data)
			return "PauseOutboundContact"
		end,
	},
	resume_contact = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
{
	campaignId = %s,
	contactId = %s,
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.contactId)
			)
		end,
		makeMethod = function(data)
			return "ResumeOutboundContact"
		end,
	},
	finish_contact = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
{
	campaignId = %s,
	contactId = %s,
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.contactId)
			)
		end,
		makeMethod = function(data)
			return "FinishOutboundContact"
		end,
	},
	remove_contact = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
{
	campaignId = %s,
	contactId = %s,
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.contactId)
			)
		end,
		makeMethod = function(data)
			return "removeOutboundContact"
		end,
	},
	agent_assign = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			return string.format(
[[
{
	campaignId = %s,
	contactId = %s,
	agent = %s,
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.contactId),
			aux.valueOrNil(data.agent)
			)
		end,
		makeMethod = function(data)
			return "SetAgentOutboundContact"
		end,
	},
	change_priority = {
		validate = function(data, aux)
			return not aux.empty(data.contactId)
		end,
		makeParams = function(data, aux)
			local priorityMode, priority = getPriority(data)
			return string.format(
[[
{
	campaignId = %s,
	contactId = %s,
	priorityMode = "%s",
	priority = %s,
}
]],
			aux.valueOrNil(data.campaign),
			aux.valueOrNil(data.contactId),
			aux.valueOrNil(priorityMode),
			aux.valueOrNil(priority)
			)
		end,
		makeMethod = function(data)
			return "ChangePriorityOutboundContact"
		end,
	},
}

local function makeParams(data, aux)
	local code = string.format(
[[
{
	type = "Phone",
	country = %s,
	area = %s,
	number = %s,
}
]],
	aux.valueOrNil(data.country),
	aux.valueOrNil(data.areaCode),
	aux.valueOrNil(data.phone)
	)
	return code
end

local function makeResult(data, aux)
	local code = 'logTrace("DEBUG", string.format("Processing activity (end) result: %s", cjson.encode(_result)))'
	if not aux.empty(data.result) then
		code = code .. string.format('\n\n%s = _result', data.result)
	end
	return code
end

local function MakeExecutable (stateName, activityId, data, aux)
	if not methods[data.actionType].validate(data, aux) then
		return aux.invalidParameterActivity(stateName, activityId, "invalid data")
	end

	--run method
	local code = string.format(
[[
local _params = %s
_result = DialerAction("%s", _params)
%s
]],
	methods[data.actionType].makeParams(data, aux),
	methods[data.actionType].makeMethod(data, aux),
	makeResult(data, aux)
	)
	return 200, code
end

return {
	MakeExecutable = MakeExecutable
}
