local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local helpDeskLogin = require "apps.helpdesk.login"
local m_request = require "apps.helpdesk.request"
local cjson = require "cjson"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

local m_enabledUpdateGroupInHelpDesk = true -- flag para cuando estamos desarrollando. (SI, esta en false, NO realiza REQUEST CONTRA HELPDESK)

------------------------------------------
--[[
	URL PARA GUARDAR GROUPS
	------------------------
	https://168.62.161.232:2443/ichelpdeskdesarrollo/rest/wsroleupdateocc
]]--
local m_updateGroupUriTemplate = "rest/wsroleupdateocc"

local function MakeUpdateGroupSuffixURI ()
	local loggedUserId = utils.GetCurrentLoggedUser()
	LogInfo("[HELPDESK.GROUPS:MakeUpdateGroupSuffixURI] Making Update Group URL, User: '%s'\nurl base: '%s'.", loggedUserId, m_updateGroupUriTemplate)
	return true, m_updateGroupUriTemplate
end

------------------------------------------

local function UpdateValues (groups, vccId)
	-- Hacia HELPDESK, tiene que viajar un array ya que Genexus, no soporta una lista con índices.
	local targetGroups = {
		Roles = {}
	}
	local sourceGroups = utils.DeepCopy(groups)

	for groupId, group in pairs(sourceGroups) do
		local fullGroupId = utils.AppendVcc(groupId, vccId)
		-- si tiene acceso a las apps
		group.VCC = vccId
		group.VccGroup = tonumber(group.IsSystemGroup) == 1 and "system" or vccId
		group.HelpDeskIntegration = group.HelpDeskIntegration or {}
		group.HelpDeskIntegration.AllowedApps = {
			administration = {
				has_permission = permissions.RoleHasPermissionToApplication("administration", fullGroupId, group.grouppermissions)
			},
			interaction = {
				has_permission = permissions.RoleHasPermissionToApplication("interaction", fullGroupId, group.grouppermissions)
			},
			supervisor = {
				has_permission = permissions.RoleHasPermissionToApplication("supervisor", fullGroupId, group.grouppermissions)
			},
			helpdesk_tickets = {
				has_permission = permissions.RoleHasPermissionToApplicationAction("apps_integrations", "helpdesk_tickets_in_app_interaction", fullGroupId, group.grouppermissions)
			},
			helpdesk_full_app = {
				has_permission = permissions.RoleHasPermissionToApplicationAction("apps_integrations", "helpdesk_full_access", fullGroupId, group.grouppermissions)
			}
		}
		-- esto no es necesario que viaje a HelpDesk
		group.grouppermissions = nil
		-- Inserto en el array de Groups
		table.insert(targetGroups.Roles, group)
	end

	return true, targetGroups
end

--[[
	Guarda Groups en HelpDesk
]]--
function Save (groups, vccId)
	assert(type(groups) == "table")

	local enabled, vccHelpDeskConfig = utils.HasHelpDeskIntegration(vccId)
	if not enabled then
		LogError("[HELPDESK.GROUPS.Save] Unable saving HelpDesk Groups. VCC '%s' Without integration", vccId)
		return true -- no es un error
	end

	local function ProcesHelpdeskError(requestError)
		local info = requestError and ((requestError.response or requestError.Error) or "{}")
		local full_sub_type = utils.JSON_Decode(info)
		local errorMessage = full_sub_type and ((full_sub_type.error and full_sub_type.error.message) or full_sub_type.Description)
		return errorMessage
	end

	local success, response = UpdateValues(groups, vccId)
	if not success then
		local encodedGroups = cjson.encode(groups)
		local errType = response
		local err = {type = "HELPDESK_INTEGRATION_ERROR", sub_type = ProcesHelpdeskError(errType)}
		LogWarning("[HELPDESK.GROUPS:Save] Error saving HelpDesk Groups: %s. Err: %s.", encodedGroups, cjson.encode(err))
		return false, err
	end

	local omnichannelGroups = response
	local encodedOmnichannelGroups = cjson.encode(omnichannelGroups)
--LogDebug("[HELPDESK.GROUPS:Save] ----------------> ENCODED GROUPS: %s", encodedOmnichannelGroups)

	if not m_enabledUpdateGroupInHelpDesk then
	-- Esto es por si estamos desarrollando y no queremos que actualice el mismo en HELPDESK.
	-- TODO: cambiar, poner algo en un flag en REDIS o en algún lugar.
		LogInfo("[HELPDESK.GROUPS:Save] Update Groups in HELPDESK is disabled.")
		return true
	end

	local success, suffixURI = MakeUpdateGroupSuffixURI()
	if not success then
		local encodedGroups = cjson.encode(groups)
		local errType = suffixURI
		local err = { type = "HELPDESK_INTEGRATION_ERROR", sub_type = ProcesHelpdeskError(errType) }
		LogWarning("[HELPDESK.GROUPS:Save] Error saving HelpDesk Groups: %s. Err: %s.", encodedGroups, cjson.encode(err))
		return false, err
	end

	local saveSuccess, saveResponse
	local try = 0

	local function SaveInHelpDesk ()
		try = try + 1
		local noGetFromCache = try > 1

		LogInfo("[HELPDESK.GROUPS.Save] Saving HelpDesk Group. Try: '%d'.", try)

		local success, jwtUserToken = helpDeskLogin.GetJWTUserToken(vccHelpDeskConfig)
		if not success then
			saveSuccess = false
			saveResponse = ProcesHelpdeskError(jwtUserToken)
			LogError("[HELPDESK.GROUPS:Save] Error Retrieving JWT Token. Groups: '%s' Err: '%s'.", encodedGroups, saveResponse)
			return
		end

		local headers = m_request.GetHeaderOAuthAuthorization(jwtUserToken, noGetFromCache)
		local helpdeskSuccess, helpdeskResponse = m_request.Post(
			suffixURI, -- la suffix URL de HelpDesk
			omnichannelGroups,
			nil,
			vccHelpDeskConfig,
			headers
		)

		saveSuccess = helpdeskResponse
		--LogDebug("[HELPDESK.GROUPS:Save]-----> success '%s', helpdeskResponse.Success '%s'", saveSuccess, helpdeskResponse.Success)

		if not helpdeskSuccess or not helpdeskResponse.Success then
			saveSuccess = false
			saveResponse = {type = "HELPDESK_INTEGRATION_ERROR", sub_type = ProcesHelpdeskError(helpdeskResponse)}
			LogWarning("[HELPDESK.GROUPS:Save] Error saving HelpDesk groups: %s. Err: '%s'", encodedOmnichannelGroups, cjson.encode(saveResponse))
			--LogDebug("[HELPDESK.GROUPS:Save] helpdeskResponse.response: '%s'", helpdeskResponse.response)
			if helpdeskResponse.response then
				local helpdeskErr = type(helpdeskResponse.response) == "string" and json.decode(helpdeskResponse.response) or helpdeskResponse.response
				--LogDebug("[HELPDESK.GROUPS:Save] helpdeskErr: '%s'", cjson.encode(helpdeskErr))
				if helpdeskErr and helpdeskErr.error and tonumber(helpdeskErr.error.code) == 112 then
					if try == 1 then
						SaveInHelpDesk()
					end
				end
			end
		else
			saveSuccess = true
			LogInfo("[HELPDESK.GROUPS.Save] Success Saved HelpDesk Group.")
		end
	end

	SaveInHelpDesk()

	return saveSuccess, saveResponse
end

function MapUris (module, htmlRender, prefix, appLocation)
	m_htmlRender = htmlRender
	m_appLocation = appLocation
end
