local Synchronize = require "Synchronize"
local https = Synchronize("luasec_mutex", function()
	return require ("ssl.https")
end)
local json  = require "cjson"
local ltn12 = require "ltn12"

module(..., package.seeall)

-- *******************************************************************************
-- API REQUEST (POST, GET)
-- *******************************************************************************

local m_urlBase = "{HOST}/" -- https://168.62.161.232:2443/ichelpdeskdesarrollo/

--[[
]]--
local function RestoreDefaultHTTPtimeout ()
	if not https.__DEFAULT_TIMEOUT then
		return https.TIMEOUT
	end
	local currentTimeoutSec = https.TIMEOUT
	if currentTimeoutSec == https.__DEFAULT_TIMEOUT then
		return https.TIMEOUT
	end
	https.TIMEOUT = https.__DEFAULT_TIMEOUT
	LogDebug("[HELPDESK.REQUEST:RestoreDefaultHTTPtimeout] --> Current https.TIMEOUT: '%s' SEC. Restore with https.TIMEOUT: '%s' SEC.", currentTimeoutSec, https.TIMEOUT)
	return https.TIMEOUT
end

--[[
	Esto referencia ya no es útil, pero la dejo como referencia del problema/solución que hubo en algún momento.

		Función encargada de agregar un timeout en segundos al Request.
		"ssl.https" tiene un bug al momento de agregar un timeout, el cual no funciona,
		para solucionar este problema, se utiliza "socket.http", SOLO para asignar el timeout.

		Referencia:
			https://stackoverflow.com/questions/20193454/lua-https-timeout-is-not-working
			Ejemplos:
				1) NO funciona:
					local https = require "ssl.https"
					https.TIMEOUT = 0.01
					b, c, h = https.request("https://www.google.fr/")

				2) SI funciona:
					local http = require "socket.http"
					local https = require "ssl.https"
					http.TIMEOUT = 0.01
					b, c, h = https.request("https://www.google.fr/")
]]--
local function SetHTTPtimeout (options)
	if not https.__DEFAULT_TIMEOUT then
		https.__DEFAULT_TIMEOUT = https.TIMEOUT
	end
	local timeoutSec = nil
	if options then
		timeoutSec = tonumber(options.timeoutSec)
		if type(timeoutSec) ~= "number" then
			return RestoreDefaultHTTPtimeout()
		end
	else
		return RestoreDefaultHTTPtimeout()
	end
	LogDebug("[HELPDESK.REQUEST:SetHTTPtimeout] --> Current https.TIMEOUT: '%s' SEC. New https.TIMEOUT: '%s' SEC.", https.TIMEOUT, timeoutSec)
	https.TIMEOUT = timeoutSec
	return timeoutSec
end

local function MakeBasicHeader (headers)
	headers = headers or {}
	return headers
end

local function MakeContentTypeHeader (body, contentType, headers)
	body = body or {}
	headers = MakeBasicHeader(headers)
	local encodedBody
	if contentType then
		headers["Content-Type"] = contentType
		encodedBody = body
	else
		if type(body) == "table" then
			headers["Content-Type"] = "application/json"
			encodedBody = json.encode(body)
		else
			headers["Content-Type"] = "text/plain"
			encodedBody = body
		end
	end
	headers["Content-Length"] = #encodedBody
	return headers, encodedBody
end

local function ExecuteRequest (RequestFn)
	local result, code, header, statusline
	local response = {}

	result, code, header, statusline, response = RequestFn()

	-- OK o CREATED o ACCEPTED
	if code == 200 or code == 201 or code == 202 then
		--LogDebug("[HELPDESK.REQUEST:ExecuteRequest] request CODE: '%s' URI: '%s'\nresponse: '%s'.", code, uri, response)
		return true, response
	end

	local err = response
	return false, {code = code, header = header, statusline = statusline, response = err}
end

--[[
	Esta función se encarga de armar la URL completa, es decir utiliza la BASE Url de la App, más el sufijo de la URL específica.
]]--
function MakeFullURL (suffixURL, server)
	assert(suffixURL)
	assert(server)
	assert(server.HOST)
	-- 1 - Reemplazo el HOST
	local urlBase = m_urlBase:gsub("{HOST}", server.HOST)
	return urlBase .. suffixURL
end

function Post (uri, body, contentType, config, headers, options)
	assert(uri)
	assert(config)
	local fullURI = MakeFullURL(uri, config.Server)
	body = body or {}
	local encodedBody
	headers, encodedBody = MakeContentTypeHeader(body, contentType, headers)

	local function DecodeJSONResponse (response)
		--LogDebug("----> content type %s",  headers["Content-Type"])
		if headers["Content-Type"]:lower() == "application/json" then
			--LogDebug("1 %s", response)
			local ok, tempResponse = pcall(json.decode, response)
			--LogDebug("2 %s ", ok)
			--LogDebug("3 %s", tempResponse)
			if ok then
				response = tempResponse
			end
		end
		return response
	end

	-- despues de ejecutar el Request, debo de dejar los valores por defecto ya que sino todos los request de cualquier app que usen la rocka "ssl" van a quedar con este valor en los request.
	local timeoutSec = SetHTTPtimeout(options)

	local function Request ()
		local startTime = os.time()

		LogInfo("[HELPDESK.REQUEST:Post] POST URI: '%s', https.TIMEOUT: '%s' SEC, contentType: '%s', headers: '%s', body: '%s'.", fullURI, timeoutSec, contentType, json.encode(headers), encodedBody)

		local source = ltn12.source.string(encodedBody)
		local response = {}
		local result, code, header, statusline = https.request({
			url = fullURI,
			method = "POST",
			headers = headers,
			source = source,
			--respuesta
			sink = ltn12.sink.table(response)
		})
		-- LogDebug("----> content type %s",  headers["Content-Type"])
		-- LogDebug("----> response: %s", json.encode(response))

		local logResponse
		if type(response) == "table" then
			if #response > 0 then
				response = response[1]
				logResponse = response
				response = DecodeJSONResponse(response)
			else
				response = response
				logResponse = json.encode(response)
			end
		else
			logResponse = response
			response = DecodeJSONResponse(response)
		end

		local endTime = os.time()
		local elapsedTimeSec = endTime - startTime
		if result then
			LogInfo("[HELPDESK.REQUEST:Post] POST Success. ELAPSED Time '%s' seconds, RESULT: '%s', CODE: '%s' HEADERS: '%s' , STATUSLINE: '%s', RESPONSE: '%s'", elapsedTimeSec, result, code, json.encode(headers or {}), statusline, logResponse)
		else
			LogError("[HELPDESK.REQUEST:Post] POST Error. ELAPSED Time '%s' seconds, RESULT: '%s', CODE: '%s' HEADERS: '%s' , STATUSLINE: '%s', RESPONSE: '%s'", elapsedTimeSec, result, code, json.encode(headers or {}), statusline, logResponse)
		end
		-- despues de ejecutar el Request, debo de dejar los valores por defecto ya que sino todos los request de cualquier app que usen la rocka "ssl" van a quedar con este valor en los request.
		RestoreDefaultHTTPtimeout()
		--
		return result, code, header, statusline, response
	end

	return ExecuteRequest(Request)
end

function Get (uri, config, headers, options)
	assert(uri)
	assert(config)
	local startTime = os.time()

	local fullURI = MakeFullURL(uri, config.Server)
	headers = headers or {}
	-- despues de ejecutar el Request, debo de dejar los valores por defecto ya que sino todos los request de cualquier app que usen la rocka "ssl" van a quedar con este valor en los request.
	local timeoutSec = SetHTTPtimeout(options)

	LogInfo("[HELPDESK.REQUEST:Get] GET URI: '%s', https.TIMEOUT: '%s' SEC, headers: '%s'.", fullURI, timeoutSec, json.encode(headers))

	local responseBody = {}
	local result, code, header = https.request{
		url = fullURI,
		method = "GET",
		headers = headers,
		sink = ltn12.sink.table(responseBody)
	}
	local response = table.concat(responseBody, "")
	local endTime = os.time()
	local elapsedTimeSec = endTime - startTime
	-- despues de ejecutar el Request, debo de dejar los valores por defecto ya que sino todos los request de cualquier app que usen la rocka "ssl" van a quedar con este valor en los request.
	RestoreDefaultHTTPtimeout()
	--
	if code == 200 then
		LogInfo("[HELPDESK.REQUEST:Get] GET Success. ELAPSED Time '%s' seconds, RESULT: '%s', CODE: '%s', HEADER: '%s', RESPONSE: '%s'.", elapsedTimeSec, result, code, json.encode(header), response)
		return true, response
	end
	LogError("[HELPDESK.REQUEST:Get] GET Error. ELAPSED Time '%s' seconds, RESULT: '%s', CODE: '%s', HEADER: '%s', RESPONSE: '%s'.", elapsedTimeSec, result, code, json.encode(header), response)

	return false, response, code
end

function GetHeaderBasicAuthorization ()
	local headers ={}
	headers["Authorization"] = "Basic RUNPX0lOQ09OQ0VSVDpFQ09fSU5DT05DRVJU"
	LogDebug("[HELPDESK.REQUEST:GetHeaderBasicAuthorization] Header Authorization '%s'", json.encode(headers))
	return headers
end

function GetHeaderOAuthAuthorization (jwtUserToken)
	assert(type(jwtUserToken) == "table")
	local headers ={}
	headers["Authorization"] = string.format("OAuth %s", jwtUserToken.access_token)
	LogDebug("[HELPDESK.REQUEST:GetHeaderOAuthAuthorization] Header Authorization '%s'", json.encode(headers))
	return headers
end