/* jshint smarttabs:true, quotmark:false, esversion: 11 */
/* global $, inConcert */


if (!this.inConcert) {
	this.inConcert = {};
}

(function () {

	"use strict";

	const FORMAT = inConcert.Utils.format;
	const REQUEST_TIMEOUT_MS = 300000; // 5 minutes
	const RELATED_REQUEST_TIMEOUT_MS = 2000; // 2 segundos

	function HelpDeskAPI (params)
	{
		if(typeof(params) !== "object") {
			throw("HelpDeskAPI.Constructor - expected an object with configuration");
		}
		if(typeof(params.uri ) !== "string") {
			throw("HelpDeskAPI.Constructor - base uri is required");
		}

		this.uri = params.uri;
		this.m_cached_tickets_uri = false;
		this.m_cached_add_tickets_uri = false;
		this.m_cached_fullapp_uri = false;
		this.m_retrieving_uri = false;
	}

	if (typeof inConcert.HelpDeskAPI !== "function") {
		inConcert.HelpDeskAPI = HelpDeskAPI;
	}

/********************************************************************************************************
	FUNCIONES GENERICAS
********************************************************************************************************/

	/*
	
	*/
	HelpDeskAPI.prototype.__Get = function(uri, callback = $.noop)
	{
		if (typeof(uri) !== "string") {
			throw("HelpDeskAPI.prototype.__Get: uri variable must be string type.");
		}
		if (typeof(callback) !== "function") {
			throw("HelpDeskAPI.prototype.__Get: callback variable must be function type.");
		}

		const request_handler = inConcert.invoke({
			url    : uri,
			type   : "GET",
			timeout: REQUEST_TIMEOUT_MS,
			success: function (list, result, server_response)
			{
				if (typeof(server_response) === "object") {
					server_response = JSON.stringify(server_response);
				}
				//console.log(uri + "\nlist: " + list + "\nresult: " + result + "\nserver_response: " + server_response);
				if (result.toLowerCase() === "success") {
					callback(true, list);
				}
				else {
					callback(false, server_response);
				}
			},
			error: function (server_response, err) {
				console.warn("HelpDeskAPI.__Get: Error. URL: " + uri + ", server_response: " + JSON.stringify(server_response)+ ", err: " + JSON.stringify(err));
				err.code = server_response.status;
				callback(false, err);
			}
		});
		return request_handler;
	};


/********************************************************************************************************
	TICKETS
********************************************************************************************************/

	/*
		Este Request "getTicketsURI" se encuentra sincronizado con el request "getFullAppURI" mediante el flag "m_retrieving_uri".
		Es decir, mientras un request se encuentre en curso, el otro debe de esperar a que termine.
		Esto se realiza así para que no existan problemas en las cargas visuales de los iframes.
		Ya que para HelpDesk es el mismo contexto, y en i6 podemos tener 2 iframes al mismo tiempo sobre la misma instancia de helpdesk.
	*/
	HelpDeskAPI.prototype.getTicketsURI = function(callback, retry)
	{
		/*
		 * sample URI:
		 *
		 */
		const self = this;

		callback = callback || $.noop;
		retry = retry || 0;

		let log_preffix = "-----> [HelpDeskAPI:getTicketsURI] Retrieving Tickets URI.";
		console.debug(log_preffix + " Retry " + retry);

		if (this.m_cached_tickets_uri) {
			callback(true, this.m_cached_tickets_uri);
			console.debug(log_preffix + "from cache.");
			return;
		}

		if (this.m_retrieving_uri) {
			console.debug(log_preffix + "another request in progress. Wait " + RELATED_REQUEST_TIMEOUT_MS / 1000 + " seconds.");
			window.setTimeout(function() {
				retry = retry + 1;
				self.getTicketsURI(callback, retry);
			}, RELATED_REQUEST_TIMEOUT_MS);
			return;
		}

		this.m_retrieving_uri = true;
		console.debug(log_preffix + "from server.");

		const uri = FORMAT("{0}/tickets/get_uri/", this.uri);
		const request_handler = this.__Get(uri, function(success, tickets_uri) {
			self.m_retrieving_uri = false;
			if (!success) {
				let err = tickets_uri;
				callback(false, err);
				return;
			}
			self.m_cached_tickets_uri = tickets_uri;
			callback(true, tickets_uri);
		});
		return request_handler;
	};

	HelpDeskAPI.prototype.getAddTicketsURI = function(callback, retry)
	{
		/*
		 * sample URI:
		 *
		 */
		const self = this;

		callback = callback || $.noop;
		retry = retry || 0;

		let log_preffix = "-----> [HelpDeskAPI:getAddTicketsURI] Retrieving Tickets URI.";
		console.debug(log_preffix + " Retry " + retry);

		if (this.m_cached_add_tickets_uri) {
			callback(true, this.m_cached_add_tickets_uri);
			console.debug(log_preffix + "from cache.");
			return;
		}

		if (this.m_retrieving_uri) {
			console.debug(log_preffix + "another request in progress. Wait " + RELATED_REQUEST_TIMEOUT_MS / 1000 + " seconds.");
			window.setTimeout(function() {
				retry = retry + 1;
				self.getAddTicketsURI(callback, retry);
			}, RELATED_REQUEST_TIMEOUT_MS);
			return;
		}

		this.m_retrieving_uri = true;
		console.debug(log_preffix + "from server.");

		const uri = FORMAT("{0}/tickets/add/get_uri/", this.uri);
		const request_handler = this.__Get(uri, function(success, add_tickets_uri) {
			self.m_retrieving_uri = false;
			if (!success) {
				let err = add_tickets_uri;
				callback(false, err);
				return;
			}
			self.m_cached_add_tickets_uri = add_tickets_uri;
			callback(true, add_tickets_uri);
		});
		return request_handler;
	};

/********************************************************************************************************
	FULL APP
********************************************************************************************************/

	/*
		Este Request "getFullAppURI" se encuentra sincronizado con el request "getTicketsURI" mediante el flag "m_retrieving_uri".
		Es decir, mientras un request se encuentre en curso, el otro debe de esperar a que termine.
		Esto se realiza así para que no existan problemas en las cargas visuales de los iframes.
		Ya que para HelpDesk es el mismo contexto, y en i6 podemos tener 2 iframes al mismo tiempo sobre la misma instancia de helpdesk.
	*/
	HelpDeskAPI.prototype.getFullAppURI = function(callback, retry)
	{
		/*
		 * sample URI:
		 *
		 */
		const self = this;

		callback = callback || $.noop;
		retry = retry || 0;

		let log_preffix = "-----> [HelpDeskAPI:getFullAppURI] Retrieving Full App URI.";
		console.debug(log_preffix + " Retry " + retry);

		if (this.m_cached_fullapp_uri) {
			callback(true, this.m_cached_fullapp_uri);
			console.debug(log_preffix + "from cache.");
			return;
		}

		if (this.m_retrieving_uri) {
			console.debug(log_preffix + "another request in progress. Wait " + RELATED_REQUEST_TIMEOUT_MS / 1000 + " seconds.");
			window.setTimeout(function() {
				retry = retry + 1;
				self.getFullAppURI(callback, retry);
			}, RELATED_REQUEST_TIMEOUT_MS);
			return;
		}

		this.m_retrieving_uri = true;
		console.debug(log_preffix + "from server.");

		const uri = FORMAT("{0}/app/get_uri/", this.uri);
		const request_handler = this.__Get(uri, function(success, fullapp_uri) {
			self.m_retrieving_uri = false;
			if (!success) {
				let err = fullapp_uri;
				callback(false, err);
				return;
			}
			self.m_cached_fullapp_uri = fullapp_uri;
			callback(true, fullapp_uri);
		});
		return request_handler;
	};
}());
