local orbit = require "orbit"
require "inconcert.orbit.application"
require("inconcert.translators")
local LuaConfigurationControl = require "LuaConfigurationControl"
local cassandra = require "lua_cassandra"
local utils = require "shared.apps.utils"
local api = relativeRequire("api", ...)
local m_integrations = require "apps.integrations.controllers.integrations"
local m_appLocation = select(2, ...)
assert(m_appLocation, "should be running under inconcertWebHandler")

module(..., package.seeall, orbit.new, inconcert.orbit.application.Decorate)
_M.model = nil

local model = relativeRequire("model", ...)
local m_models = model.GetModels()
local configControl
local modelBuilt = false

---
-- Some components expect a lualiblogger compatible logger instance.
-- The webhandler does not provide one, but we have the needed methods
-- to fake one.
--
local function make_compatible_logger ()
	return {
		debug = LogDebug,
		info = LogInfo,
		warn = LogWarning,
		error = LogError,
		fatal = LogFatal
	}
end

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",

	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		--LogDebug("1---------> CheckStatus")
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()

		--LogDebug("2---------> CheckStatus --> hasChanged: '%s', app_status.actualStatus: '%s'", hasChanged, app_status.actualStatus)

		if not hasChanged and app_status.actualStatus ~= "invalid" then
		--LogDebug("3---------> CheckStatus")
			return app_status.actualStatus
		end

		--LogDebug("4---------> CheckStatus")

		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"agent", "redis", "cassandra"})
		app_status.current = newConfig

		if not inconcert.environment.apps.integrations then
			inconcert.environment.apps.integrations = {}
		end

		--Configure Redis
		inconcert.environment.apps.integrations.redis = inconcert.environment.apps.integrations.redis or {}
		inconcert.environment.apps.integrations.redis.host = newConfig["redis"]["host"]
		inconcert.environment.apps.integrations.redis.port = tonumber(newConfig["redis"]["port"])
		inconcert.environment.apps.integrations.redis.pwd = newConfig["redis"]["pwd"] or ""

		--Configure Cassandra
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				LogInfo("Connecting to cassandra on %s:%s", value.host, tostring(value.port))
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port
			end
		end

		if modelBuilt == false then
			model:BuildModels()
			modelBuilt = true
		end

		app_status.actualStatus = "ready"
		return "ready"
	end
}

-- Returns a function bound to a translator for this application
local function MakeTextTranslator ()
	return utils.MakeTextTranslator("integrations")
end

-- This function returns the title of the application. It needs to use an explicit translator because _T could be bound to a
-- different application.
function GetApplicationTitle ()
	local _T = MakeTextTranslator()
	return _T("Widgets Integrations")
end

local htmlRender = require"inconcert.render".new("integrations", {
	MainTitle = _M.GetApplicationTitle,
	Title = function() return _T("Integrations Applications") end,
	model = m_models,
	makeTextTranslator = MakeTextTranslator
})

-- funcion que es llamada cuando el webserver se inicializa.
function Initialize ()
	LogInfo("Starting application 'integrations'.")
	--OJO que si se cambia en controllers\interactions.lua se debe cambiar aqui tambien!
	configControl = LuaConfigurationControl.Client {
		type = "INTEGRATIONS",
		address = inconcert.environment.apps.agent.config_server.ip,
		port = inconcert.environment.apps.agent.config_server.port,
		node_address = inconcert.environment.apps.agent.config_server.myIp,
		node_port = inconcert.environment.apps.agent.config_server.myPort,
	}
	configControl:AddServiceDependency("Agent")
	configControl:TriggerNodeReady()
	return true
end

--
-- The user has been authenticated and has permissions for this application. Should we let him enter anyway? (returns false if not)
function CanInitializeSession(user)
	return true
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function InitializeSession (authUser) end

--
-- This function gets called when the user logs out. This gets called reliably. The Logout function below may not
-- be called if the user logs off first (apps/login/logout)
function FinalizeSession ()
	if app_status:CheckStatus() ~= "ready" then
		LogWarning("App status is not ready. Can not logout agent %s on session %s", inconcert.authUser.userName, inconcert.authUser.authToken)
		return false
	end
	LogInfo("apps.integrations: Detail of the logged out user:  userName: (%s) - userGroup: (%s) - virtualContactCenter: (%s)", 
		inconcert.authUser.userName, inconcert.authUser.userGroup, inconcert.authUser.virtualContactCenter)
	local sessionId = "INTEGRATIONSSESSION:" .. inconcert.authUser.userName .. "@" .. inconcert.authUser.virtualContactCenter
	SharedSessions.Discard(sessionId)
	return true
end

function CheckStatus ()
	if not configControl then
		Initialize()
	end
	app_status:CheckStatus()
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetAppLocation ()
	return m_appLocation
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetApplicationName ()
	return "integrations"
end

--///////////////////////////////////////////////////////////////////////////////////
--
function serve_file (app_module, cache)
	return function (web)
		local filename = string.sub(web.path_info, 2, #web.path_info)
		if not cache then
			app_module.response.DontCache(web)
		end
		return _M:serve_static(web, m_appLocation.."/"..filename)
	end
end

--///////////////////////////////////////////////////////////////////////////////////
--

---
-- Returns a string with a suitable value to be used as version in files which needs it.
-- For instance, the offline.appcache, or version fields in querystrings.
--
local function get_version_string ()
	local hourAlignedTime = os.time()
	hourAlignedTime = hourAlignedTime - (hourAlignedTime % 3600)
	local version = os.getenv("ICC_VERSION") or tostring(hourAlignedTime)
	return version
end

---
--
function render_main (web)
	local url = "pages/integrationsMaster.html"
	local _T, _L = MakeTextTranslator()
	local manifest = ""
	local urlParameters = web.GET or {}

	LogInfo("[INTEGRATIONS:render_main] -************----------> Rendering Main Integrations HTML. url: '%s', urlParameters: %s", url, json.encode(urlParameters))

	local widgetId = urlParameters.widget_id:lower() -- "example" "sample"
	urlParameters.widget_id = widgetId
	local interactionId = urlParameters.interaction_id

	local env = {
		Title            = "Titulo de la pagina",
		ImageTitleClass  = 'Image Title Class',
		AppCacheManifest = manifest,
		ICC_VERSION      = get_version_string(),
		_T               = _T,
		_L               = _L,
		m_utils          = utils,
		m_permissions    = permissions,
		m_api            = api,
		m_widgetId       = widgetId,
		m_interactionId  = interactionId,
		m_urlParameters  = urlParameters,
		m_initialData    = m_integrations.Initialize(urlParameters),
		content          = function()
			local url = url
			local captures = {}
			return self.htmlRender:IncludeLP(url, captures)
		end
	}
	_M.response.DontCache(web)

	return htmlRender:RenderPageLP(url, web, env)
end

--///////////////////////////////////////////////////////////////////////////////////////////////////
--Mapeo urls /apis(vistas, controlles, scripts y styles) --------------------------------------------------------

--main url
_M:dispatch_get(render_main, "/%d*")

-----------------------
--  (cotrolles) urls/apis
local function GetControllers (controllerDir)
	local controllers = {}
	for file in lfs.dir(controllerDir) do
		if file ~= "." and file ~= ".." then
			local f = controllerDir..'/'..file
			local attr = lfs.attributes (f)
			if attr.mode ~= "directory" then
				local controllerName = string.match(file, [[(.+).lua]])
				if controllerName then
					controllers[controllerName] = { url = controllerName }
				end
			end
		end
	end
	--LogDebug("[INTEGRATIONS:GetControllers] controllerDir: %s, controllers: %s", controllerDir, json.encode(controllers))
	return controllers
end

------------------------------------------------------------------------------------
-- 1 - Creo los controllers y mapeos de URLs/funciones al controller principal "integrations"
--
local function MapControllersUris (controllers, controllerPath, urlBasePath)
	-- LogDebug("----------> controllerPath %s", controllerPath)
	-- LogDebug("----------> urlBasePath %s", urlBasePath)
	for key, controllerDef in pairs(controllers) do
		--requires y mapeo funciones
		--local controllerName = "apps.integrations.controllers.integrations"
		local controllerName = controllerPath .. "." .. key
		--LogDebug("------------> controllerName %s", controllerName)
		local controller = require(controllerName)
		if type(controller.MapUris) == "function" then
			controller.MapUris(_M, htmlRender, "/" .. controllerDef.url, urlBasePath)
		end
	end
end

local dir = m_appLocation .. "/controllers"
local controllers = GetControllers(dir)
local controllerPath = "apps.integrations.controllers"
local urlBasePath ="/apps/integrations"
MapControllersUris(controllers, controllerPath, urlBasePath)
------------------------------------------------------------------------------------

------------------------------------------------------------------------------------
-- 1 - Creo los controllers y mapeos de URLs/funciones a los controllers de cada uno de los Widgets, Ej /integrations/widgets/sample/...
--
local function GetWidgetsDirectories (widgetDirectoryBase)
	--LogDebug("[INTEGRATIONS:GetWidgetsDirectories] widgetDirectoryBase: %s", widgetDirectoryBase)
	local widgetsDirectories = {}
	for file in lfs.dir(widgetDirectoryBase) do
		if file ~= "." and file ~= ".." then
			--LogDebug("[INTEGRATIONS:GetWidgetsDirectories]    file: %s", file)
			local f = widgetDirectoryBase..'/'..file
			local attr = lfs.attributes(f)
			if attr.mode == "directory" then
				local widgetDirectory = file
				widgetsDirectories[widgetDirectory] = {url = widgetDirectory:lower()}
			end
		end
	end
	--LogDebug("[INTEGRATIONS:GetWidgetsDirectories] widgetsDirectories: %s", json.encode(widgetsDirectories))
	return widgetsDirectories
end

local widgetsBaseDir = dir .. "/widgets"
local widgetsControllerBasePath = controllerPath .. ".widgets"
local widgetsDirectories = GetWidgetsDirectories(widgetsBaseDir)

for widgetDir, info in pairs(widgetsDirectories) do
	local fullWidgetDir = widgetsBaseDir .. "/" .. widgetDir
	local widgetsControllers = GetControllers(fullWidgetDir)
	local widgetsControllerPath = widgetsControllerBasePath .. "." .. widgetDir
	m_integrations.MapWidgetsURIs(_M, widgetsControllers, widgetsControllerPath, widgetDir)
end
------------------------------------------------------------------------------------


--------------
api.MapUris(_M, htmlRender, "/api", "/apps/integrations")


--scripts and pages

--paginas
local function render_pages (web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = MakeTextTranslator()
	local env = {
		_T = _T,
		_L = _L,
		m_utils = utils
	}
	return htmlRender:RenderPageLP("pages/" .. name .. ".html", web, env)
end

--scripts
function render_script (web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	--LogDebug("[INTEGRATIONS:render_script] ---------------------------------> inconcert.authUser.virtualContactCenter: %s", inconcert.authUser.virtualContactCenter)

	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = MakeTextTranslator()
	local env = {
		_T = _T,
		_L = _L,
		m_utils = utils
	}
	return htmlRender:ScriptLp("scripts/" .. name .. ".js", web, env)
end

_M:dispatch_get(render_script, "/scripts/([^.]+)%.([^.]+)%.?([^.]*)%.js")
_M:dispatch_get(render_pages, "/pages/([^.]+)%.([^.]+)%.?([^.]*)%.html")

--
--  styles and resources
--  NOTA: ESTO SE PROCESA EN XAVANTE
_M:dispatch_get(_M:serve_file(), "/images/.+", "/styles/.+", "/css/.+", ".+%.mp3")
