local cassandra = require "lua_cassandra"
require "crypto"
require "uuid"
require "base64"
require "json"


args = require ('lapp') [[
KeyMigration testing parameters
	-O,--oldPrivateKey (string)		path to the old privateKey file
	-N,--newPrivateKey (string)		path to the new privateKey file
	-C,--cassandraHost (string)		cassandra host
	-c,--cassandraPort (number default 9160)		cassandraPort
]]


cassandra.seed(args.cassandraHost, tonumber(args.cassandraPort))

function ReadMasterKey(path)
	local f = io.open (path, "rb")
	assert (f, "Can not open private key file, please check the keystore.privateKey path in apps_config.lua")
	local kpriv = crypto.hmac.digest("sha256", f:read("*a"), "inConcert6-AES256-MasterKey-HMAC-SHA256", true)
	f:close()
	return kpriv
end

local oldKey = ReadMasterKey(args.oldPrivateKey)
local newKey = ReadMasterKey(args.newPrivateKey)

function GetCampaignKey(campaign, version, key)
	--Campaign no necesariamente tiene que ser una campaa valida
	--puede ser cualquier string que tenga sentido para el que pida la key
	assert(key, "You must pass a valid key")
	assert (type (campaign) == "string", "Invalid campaign")
	assert (type (version) == "number", "Invalid version")
	return crypto.hmac.digest("sha256", campaign.."_"..version, key, true)
end

function EncryptData (campaign, plain)
	
	--Cada vez que se re encriptan datos, cambia la keyversion y por lo tanto la key es diferente tambien
	local keyVersion = os.time()
	
    local kpriv = GetCampaignKey (campaign, keyVersion, newKey)
	
	initVector = crypto.hmac.digest("md5", uuid.new(), kpriv, true)
	
	local function givePaddingSize ()
		local size = math.fmod(string.byte(crypto.rand.pseudo_bytes(1)), 64) + 16
		return size
	end
	
	local leftPadding = crypto.rand.pseudo_bytes(givePaddingSize())
	local rightPadding = crypto.rand.pseudo_bytes(givePaddingSize())
	local paddingInfo = {
		leftPadding = #leftPadding,
		rightPadding = #rightPadding,
		noise = uuid.new(),
	}
	
	local encrypted = crypto.encrypt ("aes256", leftPadding.. plain..rightPadding, kpriv, initVector)
	local encryptedPaddingInfo = crypto.encrypt ("aes256", json.encode (paddingInfo), kpriv, initVector)
	
	local result = {
		algorithm = "aes256",
		keyVersion = keyVersion,
		initVector = base64.encode(initVector),
		paddingInfo = base64.encode(encryptedPaddingInfo),
		encrypted = base64.encode(encrypted),
	}
	
	return "@"..json.encode(result)
	
end

function DecryptData (campaign, text)
	data = text:match ("^@(.*)")
	
	if not data then
		return nil, "data is not encrypted"
	end
	
	data = json.decode (data)
	assert (data, "Invalid encrypted data")
	
	local kpriv = GetCampaignKey (campaign, data.keyVersion, oldKey)
	
	if data.algorithm == "aes256" then
		local encrypted = base64.decode(data.encrypted)
		local initVector = base64.decode(data.initVector)
		local encryptedPaddingInfo = base64.decode(data.paddingInfo)
		
		local paddingInfo, err = crypto.decrypt("aes256", encryptedPaddingInfo, kpriv, initVector)
		if paddingInfo == nil then
		    return nil, err
		end	
		paddingInfo = json.decode(paddingInfo)
		
		local decrypted, err = crypto.decrypt("aes256", encrypted, kpriv, initVector)
		if decrypted == nil then
			return nil, err
		end
		
		decrypted = decrypted:sub(paddingInfo.leftPadding + 1)
		decrypted = decrypted:sub(1, #decrypted - paddingInfo.rightPadding)
		return decrypted
	elseif data.algorithm == "rsa" then
		error ("RSA not implemented")
	end 
end


function GetAllColumns (keyspaceName, columnFamily, options)
	local keyspace = cassandra:CassandraConnection(keyspaceName)
	
	local func = "GetRangeSlices"
	if options and options.superColumn then
		func = "GetSuperRangeSlices"
	end	
	
	local data = keyspace[func] (keyspace, columnFamily,  --column family
								   "",  --supercolumn
								   "", --startkey
								   "", --endkey
									{ --predicate
									column_names = {}, 
									slice_range = { 
												start = "", 
												finish = "", 
												reversed = false, 
												count = 100}
									},
								300 --total keys
								)
	
	function tryJson (value)
		local ok, result = pcall (function ()
			local result = json.decode(value)
			if type (result) == "table" then
				return result
			else
				return value
			end
		end)
		return ok and result or value
	end
	
	result = {}
	for k,v in data:Iterator() do
		result[k] = {}
		for k2, v2 in ipairs (v) do
			if options and options.superColumn then
				result[k][k2] = {}
				for k3, v3 in pairs (v2.columns) do
					result[k][k2][v3.name] = tryJson(v3.value)
				end
			else
				result[k][v2.name] = tryJson (v2.value)
			end
		end
	end
	
	keyspace:Close()
	return result
end


local campaignAccounts = GetAllColumns ("Agent", "CampaignAccounts", {superColumn = true})

local keyspace = cassandra:CassandraConnection("Agent")

for campaign, accounts in pairs(campaignAccounts) do
	for _, account in ipairs(accounts) do
		if account.password then
			local password, err = DecryptData(campaign, account.password)
			
			if password then
				
				local encrypted = EncryptData(campaign, password)
				keyspace:InsertSuperColumn(campaign, "CampaignAccounts", account.Id, "password", encrypted)
				
			else
				print ("Warning: campaign "..campaign.." account "..account.Id.. " => "..err..". Account will not be touched.")
			end
		end
	end
end
