require "orbit"
require "inconcert.orbit.application"
local relativeRequire = relativeRequire

require "crypto"
require "uuid"
require "base64"
require "json"
require "redis"
require "socket"

local cassandra = require "lua_cassandra"

local LuaConfigurationControl = require "LuaConfigurationControl"

local WrapRedis = require "wrapredis"

require "lualiblogger"
local logger = lualiblogger.new("KeyStore")

local function LogInfo (...) logger.Info(...) end
local function LogDebug (...) logger.Debug(...) end
local function LogWarning (...) logger.Warning(...) end
local function LogError (...) logger.Error(...) end


--//////////////////////////////////////////////////
local __module_location = select(2, ...)
assert(__module_location, "should be running under inconcertWebHandler")

local modulename = select(1, ...)

module(modulename, package.seeall, orbit.new, inconcert.orbit.application.Decorate)

-- es necesario 'tapar' apps.repository.model que agrega Orbit, porque tenemos un módulo llamado igual
_M.model = nil

local redis = nil
local configControl = nil


--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",
	
	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end

		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"redis"})

		app_status.current = newConfig
		
		--Configure Redis
		redis = WrapRedis(newConfig["redis"]["host"], tonumber(newConfig["redis"]["port"]), newConfig["redis"]["pwd"])
		
		--Configure Cassandra
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				LogInfo("Connecting to cassandra on %s:%s", value.host, tostring(value.port))
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port
			end
		end

		app_status.actualStatus = "ready"
		return "ready"
	end
}


---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------


--Reads the master key from the private file, suited to AES encription
function ReadMasterKeyAES()
	local f = io.open (inconcert.environment.apps.keystore.privateKey, "rb")
	assert (f, "Can not open private key file, please check the keystore.privateKey path in apps_config.lua")
	local kpriv = crypto.hmac.digest("sha256", f:read("*a"), "inConcert6-AES256-MasterKey-HMAC-SHA256", true)
	f:close()
	return kpriv
end

function ReadPrivateMasterKeyRSA ()
	return assert(crypto.pkey.read(inconcert.environment.apps.keystore.privateKey, true))
end

function ReadPublicMasterKeyRSA ()
	return assert(crypto.pkey.read(inconcert.environment.apps.keystore.publicKey))
end

function ReadPublicPeerValidatorKeyRSA ()
	return assert(crypto.pkey.read(inconcert.environment.apps.keystore.peerValidatorPublicKey))
end

function GetCampaignKey(campaign, version)
	--Campaign no necesariamente tiene que ser una campa�a valida
	--puede ser cualquier string que tenga sentido para el que pida la key
	assert (type (campaign) == "string", "Invalid campaign")
	assert (type (version) == "number", "Invalid version")
	return crypto.hmac.digest("sha256", campaign.."_"..version, ReadMasterKeyAES(), true)
end

function EncryptData (campaign, plain)
	--Cada vez que se re encriptan datos, cambia la keyversion y por lo tanto la key es diferente tambien
	local keyVersion = os.time()
	
    local kpriv = GetCampaignKey (campaign, keyVersion)
	if inconcert.environment.apps.keystore.algorithm == "aes256" then
		
        initVector = crypto.hmac.digest("md5", uuid.new(), kpriv, true)
		
		local function givePaddingSize ()
			local size = math.fmod(string.byte(crypto.rand.pseudo_bytes(1)), 64) + 16
			return size
		end
		
		local leftPadding = crypto.rand.pseudo_bytes(givePaddingSize())
		local rightPadding = crypto.rand.pseudo_bytes(givePaddingSize())
		local paddingInfo = {
			leftPadding = #leftPadding,
			rightPadding = #rightPadding,
			noise = uuid.new(),
		}
		
		local encrypted = crypto.encrypt ("aes256", leftPadding.. plain..rightPadding, kpriv, initVector)
		local encryptedPaddingInfo = crypto.encrypt ("aes256", json.encode (paddingInfo), kpriv, initVector)
		
		local result = {
			algorithm = "aes256",
			keyVersion = keyVersion,
			initVector = base64.encode(initVector),
			paddingInfo = base64.encode(encryptedPaddingInfo),
			encrypted = base64.encode(encrypted),
		}
		
		return "@"..json.encode(result)
	elseif inconcert.environment.apps.keystore.algorithm == "rsa" then
		error ("RSA not implemented")
	end 
end

function DecryptData (campaign, data)
	data = data:match ("^@(.*)")
	data = json.decode (data)
	assert (data, "Invalid encrypted data")
	
	local kpriv = GetCampaignKey (campaign, data.keyVersion)
	
	if data.algorithm == "aes256" then
		local encrypted = base64.decode(data.encrypted)
		local initVector = base64.decode(data.initVector)
		local encryptedPaddingInfo = base64.decode(data.paddingInfo)
		
		local paddingInfo, err = crypto.decrypt("aes256", encryptedPaddingInfo, kpriv, initVector)
		if paddingInfo == nil then
		    return nil, err
		end	
		paddingInfo = json.decode(paddingInfo)
		
		local decrypted, err = crypto.decrypt("aes256", encrypted, kpriv, initVector)
		if decrypted == nil then
			return nil, err
		end
				
		decrypted = decrypted:sub(paddingInfo.leftPadding + 1)
		decrypted = decrypted:sub(1, #decrypted - paddingInfo.rightPadding)
		return decrypted
	elseif data.algorithm == "rsa" then
		error ("RSA not implemented")
	end 
end


--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-- funcion que hace la inicializaci�n del modulo, la misma es llamada cuando se inicializa el webserver en el config.lua
function Initialize(env)
	LogInfo("initialize KeyStore")
	
	if inconcert.environment.apps.keystore.algorithm == "aes256" then
		ReadMasterKeyAES()
	elseif inconcert.environment.apps.keystore.algorithm == "rsa" then
		ReadPrivateMasterKeyRSA()
		ReadPublicMasterKeyRSA()		
	else
		error ("Invalid algorithm, please check keystore.algorithm in apps_config.lua")
	end
	
	configControl = LuaConfigurationControl.Client {
		type = "KEYSTORE",
		address = inconcert.environment.apps.keystore.config_server.ip,
		port = inconcert.environment.apps.keystore.config_server.port,
		node_address = inconcert.environment.apps.keystore.config_server.myIp,
		node_port = inconcert.environment.apps.keystore.config_server.myPort,
	}
	
	configControl:TriggerNodeReady()
	return true
end


--Esta funcion verifica que la clave publica recibida contenga la firma propia local
--Si contiene dicha firma, entonces la verifica con su clave privada
--Si no contiene la firma, entonces la verifica con la clave de validation y luego se a�ade la firma propia local
--Si en alguna verificacion falla, entonces se considera que hubo error
function MaintainKey(nodeId, kpub)
	local hostName = socket.dns.gethostname()
	if hostName == "" or hostName == nil then
		error("Can not obtain local host name")
	end
	
	--Primero se comprueba si existe una signature propia con la que se haya verificado previamente la clave
	if kpub["signature_"..hostName] then
		LogInfo("Verifying %s certificate with local signature for hostname '%s'", nodeId, hostName)
		local verified = crypto.verify("sha256", kpub.key, base64.decode(kpub["signature_"..hostName]), ReadPublicMasterKeyRSA())
		if not verified then
			error("Public key obtained from redis does not verify the local signature")
		end
	else
		--no hay signature propia, se debe intentar con la signature de validacion
		LogInfo("Verifying %s certificate for the first time with the peer validator key", nodeId)
		verified = crypto.verify("sha256", kpub.key, base64.decode(kpub.signature), ReadPublicPeerValidatorKeyRSA())
		
		if verified then
			LogInfo("%s certificate is valid. Signing with local (%s) private key", nodeId, hostName)
			
			local sig = assert(crypto.sign('sha256', kpub.key, ReadPrivateMasterKeyRSA()))
			
			kpub["signature_"..hostName] = base64.encode(sig)
			
			redis:hset("KeyStore::PublicKeys", nodeId, json.encode(kpub))
			redis:save()
		else
			error("Public key obtained from redis does not verify signature from actual peer validator key")
		end
	end
end


function APIGetCampaignKey(web)
	_M.response.DontCache(web)
	web.headers["Content-Type"] = "application/json"
	
	local function Error (message)
		LogError ("%s", message)
		web.status = 403
		return json.encode {error = message}
	end
	
	local data = json.decode(web.input["post_data"])
	
	LogInfo ("Received request: %s", data.request)
	
	local request = json.decode(data.request)
	local kpub = redis:hget("KeyStore::PublicKeys", request.nodeId)
	
	if not kpub then
		return Error("Can not find public key for nodeId "..request.nodeId)
	end
	
	kpub = json.decode(kpub)
	
	local ok, err = pcall(MaintainKey, request.nodeId, kpub)
	if not ok then
		return Error(err)
	end
	
	--TODO: Para no hacer esto, se deberia cambia en la luacrypto para que pudiera crear una pkey
	--		solamente sabiendo su contenido y no teniendo que ir a leer de disco cada vez
	local tmpName = os.tmpname ()
	local f = io.open (tmpName, "wb")
    f:write (kpub.key)
	f:close()
	
	local kpub = assert(crypto.pkey.read(tmpName))
	
	verified = crypto.verify("sha256", data.request, base64.decode(data.signature), kpub)
	
	os.remove(tmpName)
	
	if not verified then
		return Error("Received request does not verify the signature")
	end
	
	local data, ek, iv = crypto.seal("aes256", GetCampaignKey (request.campaign, request.version), kpub)
	
	local result = {
		data = base64.encode(data),
		ek = base64.encode(ek),
		iv = base64.encode(iv),
	}
	return json.encode(result)
end

function APICipherContent(web)
	_M.response.DontCache(web)
	web.headers["Content-Type"] = "application/json"
	
	local data = json.decode(web.input["post_data"])
	
	local cipheredContent = base64.decode(data.cipheredContent)
	local ek = base64.decode(data.ek)
	local iv = base64.decode(data.iv)
	local content = crypto.open ("aes256", cipheredContent, ReadPrivateMasterKeyRSA(), ek, iv)
	
	if content == nil then
	    local errMessage = "can not open sealed content. Sealed with bad public key?"
        LogError ("%s", errMessage)
		web.status = 403
		return json.encode {error = errMessage}
	end
	
	local encrypted = EncryptData (data.campaign, content)
	
	local result = {encrypted = encrypted}
	
	return json.encode(result)
end

function APIServiceOK(web)
	_M.response.DontCache(web)
	web.headers["Content-Type"] = "application/json"
    return json.encode(true)
end


function APIMaintainKeys(web)
	_M.response.DontCache(web)
	web.headers["Content-Type"] = "application/json"
	
	local keys = redis:hgetall("KeyStore::PublicKeys")
	
	for k, v in pairs(keys) do
		local ok, err = pcall(MaintainKey, k, json.decode(v))
		if not ok then
			LogWarning ("Error maintaining key for node %s. error was: %s", k, err)
		end
	end
	
    return json.encode(true)
end

_M:dispatch_post(APIGetCampaignKey, "/get_campaign_key")
_M:dispatch_post(APICipherContent, "/cipher_content")
_M:dispatch_post(APIServiceOK, "/serviceok/")
_M:dispatch_post(APIMaintainKeys, "/maintainkeys/")
