local orbit = require "orbit"
local zlib = require "zlib"
local lualiblogger = require "lualiblogger"
local cosmo = require "cosmo"

local agentModel = require "apps.agent.model"

local logger = lualiblogger.new("agentlogstorage")

local mail = relativeRequire "mail"

module(..., package.seeall, orbit.new)




---
-- Makes sure that the compressed data contains the needed header: 0x78da. Air does not send it.
--
local function ensure_proper_header (input_data)
	if string.byte(input_data, 1) ~= 120 or string.byte(input_data, 2) ~= 218 then
		input_data = string.char(120) .. string.char(218) .. input_data
	end
	return input_data
end

---
-- Logs the uncompressed data using lualiblogger, prepending the id of the agent who uploaded it.
--
local function log_data (data, agent_id)
	for line in data:gmatch("[^\r\n]+") do
		logger.info("%s\t%s", agent_id, line)
	end
end

---
-- Receives the log file sent by the agent before closing the application.
-- The file is sent compressed with deflate algorithm.
-- We store the file uncompressed.
--
local function upload_log (web)

	web.headers["Content-Type"] = "text/plain"

	local agent_id = inconcert.authUser.userName .. "@" .. inconcert.authUser.virtualContactCenter

	LogInfo("Uploading agent log file - agent_id: %s", agent_id)

	local input_data = ensure_proper_header(web.input["post_data"])

	local ok, uncompressed_data = pcall(function()
		return assert( zlib.inflate(input_data):read("*a") )
	end)
	if not ok then
		LogError("Log upload for agent '%s' failed with error: %s", agent_id, uncompressed_data)
		web.status = 400
		return "Invalid input data"
	end

	log_data(uncompressed_data, agent_id)

	web.status = 200
	return "ok"
end


---
-- Mail template
--
local mail_template_error = [[
Agent id: $agent_id
Error information:
$data
]]

---
-- Receives a single error that an agent has recorded and sent to us.
-- The message is sent using Content-Type: application/x-www-form-urlencoded
--
local function log_error (web)

	web.headers["Content-Type"] = "text/plain"

	local agent_id = inconcert.authUser.userName .. "@" .. inconcert.authUser.virtualContactCenter

	LogInfo("Recording error sent by agent_id: %s", agent_id)

	local input_data = web.POST.data
	if not input_data then
		LogError("Recording error sent by agent '%s' failed. No data was sent (or the format was not recognized)", agent_id)
		web.status = 400
		return "Invalid input data"
	end

	LogError("%s\tError captured by the client: %s", agent_id, input_data)

	--[[ ESTA VUELTA DE ACA NO ANDA
	logger.error("%s\tError captured by the client: %s", agent_id, input_data)
	
	local ok, data = pcall(json.decode, input_data)
	if ok then
		local ok, err = true, ""
	
		local errText = "" --errText will contain part of the error logged by the client, and will be used as a key in redis
		if data.data and data.data.exception then
			--with exception I can identify each error
			errText = json.encode(data.data.exception)
		elseif data.data and data.data.message then
			--try to parse until "Stack Traceback" is found
			errText = data.data.message:match("(.*)Stack Traceback")
			
			if not errText then
				--last chance, get first 200 characters
				errText = data.data.message:sub(1, 200)
			end
			
		else
			--I have no idea how to handle this situation
			ok = false
			err = "I don't know how to handler this particular case: "..input_data
		end	
		
		
		if ok then
			local ok, redis = pcall(agentModel.GetRedisClient)
			
			if ok then
				local count = redis:incr("LogStorage:"..errText)
				
				if redis:ttl("LogStorage:"..errText) == -1 then
					redis:expire("LogStorage:"..errText, 60 * 60 * 24)
				end
				
				if count > 1 then
					web.status = 200
					return "ok"
				end
			
			else
				LogWarning("LogStorage: %s", redis)
			end
			
		else
			LogWarning("LogStorage: %s", err)
		end
	end
	--]]
	--
	-- Only send mails if we're on PRODUCTION mode, to avoid development installs to spam our inbox.
	--
	--[[
	if os.getenv("INCONCERT_ENV_MODE") == "PRODUCTION" then
		-- if for whatever reason, the mail fails to be sent, I don't mind because it has been logged to file
		mail.send_mail{
			first_name = "inConcert logStorage",
			email_address = "no-reply@inconcertcc.com",
			body = cosmo.fill(mail_template_error, { agent_id = agent_id, data = input_data })
		}
	end
	--]]

	web.status = 200
	return "ok"
end


_M:dispatch_post(upload_log, "/upload")

_M:dispatch_post(log_error, "/log_error")
