module(..., package.seeall, inconcert.orbit.application.Decorate)


local utils = require "shared.apps.utils"
require "LuaOrm.Date"
local JSON_Encode                  = utils.JSON_Encode;
local JSON_Decode                  = utils.JSON_Decode;
local Map                          = utils.Map;
local Reduce                       = utils.Reduce;
local queryStats                   = require "apps.monitoring.controllers.agents.querystats"
local queryLogs                    = require "apps.monitoring.controllers.agents.querylogs"
local SPEEDTEST_RESULTS_KEY_PREFIX = "AgentMonitorService::AgentSpeedTest::Results";
local SPEEDTEST_INDEX_KEY_PREFIX   = "AgentMonitorService::AgentSpeedTest";

function js2unixts(ts)
	return math.floor(ts / 1000);
end

function unix2jsts(ts)
	return math.floor(ts * 1000)
end

function queryAgents(queryText, queryFields)
	local moduleAdmUsers = require "apps.administration.controllers.users"
	local moduleAdmModels = require("apps.administration.model"):GetModels()
	local moduleUsersModel = require "apps.agent.models.users"
	local agentservice = require "apps.agent.agentservice"
	local vcc = inconcert.authUser.virtualContactCenter;
	local redis = utils.GetRedisClient();
	local stdstates = {
		["ACTIVE"] = true,
		["AWAY"] = true,
		["LUNCH"] = true,
		["BATHROOM"] = true,
		["BREAK"] = true,
		["BUSY"] = true,
		["DISABLE"] = true,
		["OFFLINE"] = true,
		["NOACD"] = true
	}

	queryFields = queryFields or { Id = true, Name = true, FirstSurname = true, UserGroup = true }
	if queryFields.Id then queryFields.Id = queryText end
	if queryFields.Name then queryFields.Name = queryText end
	if queryFields.FirstSurname then queryFields.FirstSurname = queryText end
	if queryFields.UserGroup then queryFields.UserGroup = queryText end
	queryFields.Status = "1"

	local filterData = {
		last = "",
		limit = 1e5,
		numberPage = 1,
		query = queryFields,
		columns = {
			"Id",
			"Name",
			"FirstSurname",
			"UserGroup",
			"Status",
			"HasRoleOverride",
			"StatusReason"
		}
	}
	local success, searchResult = pcall(moduleAdmUsers.SearchUsers, filterData);
	if success then
		searchResult.rows = Map(searchResult.rows,
			function(_, user)
				local agentCaptureStatus = moduleUsersModel.GetUserCaptureStatisticsStatus(user.Id, vcc);
				local agent_at_vcc = user.Id .. "@" .. vcc
				local geoData = redis:hget("AgentMonitorService::AgentGeolocation:" .. agent_at_vcc, "position")
				if geoData then
					geoData = JSON_Decode(geoData);
					user.GeoLocation = {
						lat = geoData.latitude,
						lon = geoData.longitude
					}
				end
				local stateId = agentservice.GetAgentStatus(user.Id .. "@" .. vcc).state
				user.IsLoggedIn = agentservice.IsAgentLoggedIn(user.Id, vcc);
				user.LoggedStatus = stateId
				if user.LoggedStatus == "" then
					user.LoggedStatus = "OFFLINE"
				end
				if not stdstates[stateId] then
					local states = moduleAdmModels.States.GetStates(stateId)
					local state = states and states[stateId]
					if state then
						user.StatusDescription = state.Description
					end
				end

				user.MonitoringStatus = agentCaptureStatus.browser_health;
				return user;
			end)
		return searchResult
	else
		return {}
	end
end

function GetAllAgents(web)
	local queryText = web.input.q or ""
	return json.encode(queryAgents(queryText, { Id = true }))
end

function devCall(fnname, ...)
	package.loaded['apps.monitoring.controllers.dev.agents-stats-dev'] = nil;
	local f
	while not f do
		local p = require("apps.monitoring.controllers.dev.agents-stats-dev")
		f = p[fnname];
	end
	local s, r = pcall(f, ...);
	if not s then
		LogInfo("HEALTH ERROR calling %s, %s", fnname, r)
		return {}
	end
	return r or {};
end

function AgentStats(web, agentId)
	local tsfrom = tonumber(web.input.from or 0)
	local tsto = tonumber(web.input.to or 0)
	local result = queryStats.queryAgentStats(agentId, tsfrom, tsto)
	return JSON_Encode(result)
end

function AgentInfo(web, agentId)
	-- TODO : Usar otra api para traer el agente en lugar del search
	-- porque este trae mas de uno cuando los id son adminfull y adminfull2 buscando adminfull
	local searchResult = queryAgents(agentId, { Id = true })
	local result = {}
	-- Filtramos manualmente
	for _, v in ipairs(searchResult.rows) do
		if v.Id == agentId then
			result = v;
		end
	end
	return JSON_Encode(result);
end

function AgentLogs(web, agentId)
	local tsfrom = tonumber(web.input.from or 0)
	local tsto = tonumber(web.input.to or 0)
	local row = tonumber(web.input.row or 0)
	local size = tonumber(web.input.size or 20)
	local result = queryLogs.agentLogs(agentId, tsfrom, tsto, row, size);
	return JSON_Encode(result)
end

-- Esto viene desde tweety csvutils
function escapeCSV(s, delimiter)
	if s == nil then
		return
	end

	s = tostring(s)
	if string.find(s, '[' .. delimiter .. '"\r\n]') then
		s = '"' .. string.gsub(s, '"', '""') .. '"'
	end
	return s
end

function AgentLogsExport(web, agentId)
	local tsfrom = tonumber(web.input.from or 0)
	local tsto = tonumber(web.input.to or 0)
	local vcc = inconcert.authUser.virtualContactCenter;
	local result = queryLogs.agentLogs(agentId, tsfrom, tsto, 0, -1);
	local rows = Map(result.rows, function(_, row)
		row.timestamp = os.date("%Y-%m-%dT%H:%M:%SZ", math.floor(row.timestamp / 1000))
		row.message = escapeCSV(row.message, ",")
	end)

	local csv = utils.Table2Csv(rows, { timestamp = true, method = true, message = true }, nil, ",",
		{ "timestamp", "method", "message" })
	local suggestedFileName = string.format("%s-logs-%s-%s-%s.csv", agentId, vcc,
		os.date("%Y-%m-%d %H:%M:%SZ", math.floor(tsfrom / 1000)),
		os.date("%Y-%m-%d %H:%M:%SZ", math.floor(tsto / 1000)))
	web.status = 200;
	web.headers["Content-Disposition"] = [[attachment; filename="]] .. suggestedFileName .. '"'

	web.headers["Content-Type"] = "application/octet-stream"
	return csv;
end

function MonitoringPeriods(web, agentId)
	local vcc = inconcert.authUser.virtualContactCenter;
	local agentMonitoringEventsKey = "AgentMonitorService::CaptureData::" .. agentId .. "@" .. vcc;
	local redis = utils.GetRedisClient();
	local data = redis:hget(agentMonitoringEventsKey, "periods")
	local periods = {}

	if data then
		periods = JSON_Decode(data)
	end
	return JSON_Encode(Map(periods, function(_, period)
		period.started_ts = period.started;
		if not period.finished then
			period.finished = os.time() * 1000;
			period.realtime = true;
		end
		period.duration_ms = period.finished - period.started_ts;
		return period
	end))
end

function agentCalls(agentId, from_timestamp, to_timestamp)
	local vcc = inconcert.authUser.virtualContactCenter;
	local agentAtVcc = agentId .. "@" .. vcc;
	local interactions = require("apps.agent.controllers.interactions");
	local from_date = os.date("%Y-%m-%d %H:%M:%S", js2unixts(from_timestamp));
	local to_date = os.date("%Y-%m-%d %H:%M:%S", js2unixts(to_timestamp));
	local request = {
		query = {
			condition = "AND",
			fixedAgentRule = agentAtVcc,
			fixedStateRule = "FINISHED",
			rules = {
				{ id = "start_timestamp",
					operator = "between",
					value = from_date,
					value_to = to_date },
				{ id = "type",
					operator = "equal",
					value = "CALL" }
			}
		},
		rows = 20
	}
	local status, response = interactions.SearchInteractionsHelper(request, nil, nil, nil, true)
	if status == 200 then
		return Map(response.interactions,
			function(_, el)
				return {
					id = el.id,
					callerId = el.contact_address,
					date = el.start_timestamp .. "Z",
					ts = unix2jsts(LuaOrm.Date(el.start_timestamp):time()),
					duration = math.floor(el.total_records_time_sec * 1000),
				}
			end)
	else
		return {}
	end
end

function Calls(web, agentId)
	local r = agentCalls(agentId, web.input.from, web.input.to)
	return JSON_Encode(r);
end

function speedTests(agentId)
	local vcc = inconcert.authUser.virtualContactCenter;
	local agent_at_vcc = agentId .. "@" .. vcc
	local redis = utils.GetRedisClient();
	local speedTests = redis:zrange(SPEEDTEST_INDEX_KEY_PREFIX .. "::" .. agent_at_vcc, 0, -1)
	speedTests = Map(speedTests, function(_, s) return JSON_Decode(s) end)
	return Map(speedTests, function(_, test)
		local function avgkey(results, k)
			local c = 0;
			local t = 0;
			for _, r in ipairs(results) do
				if r[k] and tonumber(r[k]) then
					c = c + 1
					t = t + tonumber(r[k])
				end
			end
			if c > 0 then
				return t / c;
			else
				return 0
			end
		end

		local results = redis:zrangebyscore(SPEEDTEST_RESULTS_KEY_PREFIX .. ":" .. agent_at_vcc, test.id, test.id);

		results = Map(results, function(_, s) return JSON_Decode(s).speedTest end)
		test.downloadMbps = avgkey(results, "download")
		test.uploadMbps = avgkey(results, "upload")
		test.ping = avgkey(results, "ping")
		test.jitter = avgkey(results, "jitter");
		return test;
	end)
end

function SpeedTests(web, agentId)
	local r = speedTests(agentId)
	return JSON_Encode(r);
end

function MonitoringStatus(web)
	local vcc = inconcert.authUser.virtualContactCenter;
	local redis = utils.GetRedisClient();
	local monitored_agents = {}
	for _, k in pairs(redis:hkeys("AgentMonitorService::ControlSet::MonitoringStatus")) do
		local _, _, user_vcc = string.find(k, "%@(.+)")
		if user_vcc == vcc and tonumber(redis:hget("AgentMonitorService::ControlSet::MonitoringStatus", k)) > 0 then
			table.insert(monitored_agents, k)
		end
	end
	return JSON_Encode({
		monitored_list = monitored_agents,
		monitored_count = #monitored_agents,
		limit = 5
	})
end

function MapUris(module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest;
	module:dispatch_get(LogRequest .. GetAllAgents, prefix .. "/all");

	module:dispatch_get(LogRequest .. AgentInfo, prefix .. "/(.+)/info");

	module:dispatch_get(LogRequest .. AgentStats, prefix .. "/(.+)/stats");
	module:dispatch_get(LogRequest .. AgentLogs, prefix .. "/(.+)/logs");
	module:dispatch_get(LogRequest .. AgentLogsExport, prefix .. "/(.+)/logs/export")
	module:dispatch_get(LogRequest .. MonitoringPeriods, prefix .. "/(.+)/monitoringperiods");
	module:dispatch_get(LogRequest .. MonitoringStatus, prefix .. "/monitoringstatus");
	module:dispatch_get(LogRequest .. Calls, prefix .. "/(.+)/calls");
	module:dispatch_get(LogRequest .. SpeedTests, prefix .. "/(.+)/speedtests");
	m_htmlRender = htmlRender;
	m_appLocation = appLocation;
end
