module(..., package.seeall, inconcert.orbit.application.Decorate)

local utils = require "shared.apps.utils"
local Map = utils.Map;
local Reduce = utils.Reduce;
local JSON_Encode = utils.JSON_Encode;
local JSON_Decode = utils.JSON_Decode;
require("socket")
local http = require "socket.http"

local LOGGING_INDEX_NAME = "agentmonitor-logging-capture";


function elastic_get(index_name, query, withScroll, elasticSearchConnectionConfig)
    local string_data

    if withScroll then
        query.size = 5000;
    end

    local http = require "socket.http"
    local ltn12 = require "ltn12"


    local equery = function(url, currentQuery)
        local requestInfo = utils.GetElasticSearchInstance():MakeRequestInfo({
            query = currentQuery
        })

        local response_body = {}
        local ok, response_code, response_headers, response_status_line = http.request({
            host = elasticSearchConnectionConfig.host,
            port = elasticSearchConnectionConfig.port,
            url = url,
            headers = requestInfo.headers,
            method = "POST",
            source = requestInfo.source,
            sink = ltn12.sink.table(response_body),
            redirect = false
        })
        if not ok then
            LogError("[AgentMonitoring:elastic_get] %s\n%s\n%s\n%s", response_code, json.encode(response_headers),
                response_status_line,
                json.encode(response_body))
            return false, { error = response_status_line, status = response_code }
        end
        if not response_headers["content-type"]:match("^application/json") then
            LogError("[AgentMonitoring:elastic_get] Unknown response from elasticsearch. headers: %s\nresponse: %s",
                json.encode(response_headers))
            return false, { error = "inconcert unknown", status = 500 }
        end

        local raw_response = table.concat(response_body)
        local response = json.decode(raw_response)
        if response.error then
            LogError("[AgentMonitoring:elastic_get] Request failed with the following error: '%s'",
                json.encode(response.error));
            return false, response
        end
        return true, response

    end
    local url
    if withScroll then
        url = ("/%s/_search?scroll=3m"):format(index_name);
    else
        url = ("/%s/_search"):format(index_name)
    end
    local ok, response = equery(url, query)
    if ok then
        local scroll_id = response._scroll_id;
        return true, { next = function(_)
            if response then
                local r = response
                response = nil;
                return true, r.hits.hits, r.hits.total.value;
            else
                if scroll_id then
                    local scroll_response;
                    ok, scroll_response = equery("/_search/scroll", { scroll = "3m", scroll_id = scroll_id })
                    if ok then
                        return ok, scroll_response.hits.hits
                    else
                        return ok, scroll_response
                    end
                else
                    return true, {}
                end
            end
        end }, response.hits.total.value
    else
        return ok, response
    end
end

function search_elastic(index_name, query)
    local elasticSearchConnectionConfig = inconcert.environment.apps.agent.elasticsearch
    if not elasticSearchConnectionConfig or not elasticSearchConnectionConfig.host or
        not elasticSearchConnectionConfig.port then
        LogError("[AgentMonitoring:search_elastic] Unable to execute ES Request. Invalid elasticsearch config: %s",
            json.encode(elasticSearchConnectionConfig or {}))
        return false, { error = "invalid elasticsearch config", status = 500 }
    end
    local ok, es, total = elastic_get(index_name, query, query.size == nil, elasticSearchConnectionConfig)
    if ok then
        local records = {}
        local page
        ok, page = es:next()
        while ok and #page > 0 do
            for _, v in ipairs(page) do
                table.insert(records, v)
            end
            ok, page = es:next()
        end
        if ok then
            return ok, records, total
        else
            return ok, page
        end
    else
        return ok, es -- es = error response
    end
end

function agentLogs(agentId, tsfrom, tsto, from, size)
    local vcc = inconcert.authUser.virtualContactCenter;
    local index_name = (LOGGING_INDEX_NAME .. "-%s"):format(vcc);
    local agent_at_vcc = agentId .. "@" .. vcc
    local query = {
        from = from,
      
        sort = {
            {
                timestamp = {
                    order = "asc"
                }
            }
        },
        query = {
            bool = {
                must = {
                    {
                        term = {
                            agent = agent_at_vcc
                        }
                    },
                    {
                        range = {
                            timestamp = {
                                gte = tsfrom,
                                --        format = "yyyy-MM-dd HH:mm:ss",
                                lte = tsto
                            }
                        }
                    }
                },
                filter = {}
            }
        }
    }
    if size > 0 then
        query.size = size
    else
        query.size = nil
    end

    local success, records, total = search_elastic(index_name, query);
    if success then
        LogInfo("AgentMonitoring: TOTAL %d , SIZE %d", total, #records)

        return {
            tsfrom = tsfrom,
            tsto = tsto,

            total = total,
            from = from or 0,
            size = #records,
            rows = Map(records,
                function(_, line)
                    return line._source;
                end
            )
        }
    end

end
