-- Facebook OAuth controller

require "inconcert.orbit.application"
require "AsyncTasks"

local http = require "socket.http"
local ssl = ssl
local Synchronize = require "Synchronize"
local https = Synchronize("luasec_mutex", function()
	return require ("ssl.https")
end)

local cosmo = require "cosmo"
local json = require "json"
local wsapi_util = require "wsapi.util"

local campaigns = require("apps.agent.models.campaigns")

module(..., package.seeall, inconcert.orbit.application.Decorate)

local m_htmlRender = nil
local m_config = nil
local m_redis_client

local DEFAULT_SCOPE = 
	"pages_show_list," ..						-- show pages list
	"pages_manage_metadata," ..					-- read page data
	"instagram_basic," ..						-- manage instagram basic data
	"instagram_manage_comments," ..				-- manage instagram comments
	"instagram_manage_insights," ..				-- manage instagram statistics
	"instagram_manage_messages"					-- manage instagram messages

---
-- Given a 'authorization_code' (granted when the user gave permissions to this application)
-- obtain a valid access token
--
-- @tparam string authorization_code
-- @tparam table codes
-- @tparam string codes.client_id
-- @tparam string codes.redirect_uri
-- @tparam string codes.client_secret
--
-- @treturn boolean ok           - If the renew went well or not
-- @treturn string message       - If the renew went well wrong, an error message (optional)
-- @treturn string access_token
-- @treturn table values
--
local function RenewAccessToken(_T, authorization_code, codes)
	local new_url = cosmo.fill([[https://graph.facebook.com/oauth/access_token?format=json&client_id=$client_id&redirect_uri=$redirect_uri&code=$authorization_code&client_secret=$client_secret]], {
		client_id = codes.client_id,
		redirect_uri = codes.redirect_uri,
		authorization_code = authorization_code,
		client_secret = codes.client_secret
	})
	local response, code, headers, status = https.request(new_url)
	if code ~= 200 then
		local message = _T("Error in https request ") .. "(graph.facebook.com/oauth/access_token): " .. (response or "")
		LogError("[oauth] - %s", message )
		return false, message
	end
	
	local values = json.decode(response)
	
	new_url = cosmo.fill([[https://graph.facebook.com/me?access_token=$access_token]], {access_token = values.access_token})
	response, code, headers, status = https.request(new_url)
	if code ~= 200 then
		local message = _T("Error in https request ") .. "(graph.facebook.com/me): " .. response
		LogError("[oauth] - %s", message )
		return false, message
	end
	
	local userData = json.decode(response)
	values.uid = userData.id
	values.name = userData.name
	
	return true, "", values.access_token, values
end

local function RegisterAuthWebhook(webhook_host, subdomain_host, tracking_code)

	local url = cosmo.fill([[https://$webhook_proxy_host/registeroauth?callback_env=$subdomain_host&state=$state]], {
    	webhook_proxy_host = webhook_host,
        subdomain_host = subdomain_host,
        state = tracking_code
	})

    response, code, headers, status = https.request(url)
    if code ~= 200 then
    	LogError("[oauth] - Error in https request RegisterAuthWebhook: %s , code: %s", response,code )
        return false
	end
	return true
end

--
-- Shows the authorization form for a given oauth provider
-- Probablememte deva devolver acá markup, por ejemplo un iframe para que se inserte en una página o algo así...
-- https://graph.facebook.com/oauth/authorize?client_id=139099499445685&redirect_uri=http://localhost:8082/test/sniff.lp&display=popup&scope=user_photos,user_videos,publish_stream
local function RenderAuthorizationForm(web)
	local _T = m_htmlRender.extension.makeTextTranslator()
	
	if m_config == nil then
		local message = "Error in RenderAuthorizationForm: Must configure application oauth in apps_config.lua"
		LogError("[oauth] - %s ", message)
		return m_htmlRender:RenderPageLP("pages/instagramAccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end
	
	local scope = web.input.scope or DEFAULT_SCOPE
	
	local client_id = m_config.client_id
	local client_secret = m_config.client_secret
	local postData = json.decode(web.input.post_data)
	local webhook_host = m_config.webhook_host
	
	
	-- We use the tracking code to identify the data to be saved in the session
	local tracking_code = uuid.new()

	if not RegisterAuthWebhook(webhook_host, postData.host, tracking_code) then
		web.status = 500
		LogError("[oauth] - Error register twitter token, code %s body %s headers %s", err, body, json.encode(headers))
		return "Error requesting instagram token"
	end
	
	local redirect_uri = "https://" .. webhook_host .. "/instagram/auth_redirect"
	
	LogInfo("[oauth] - Requesting token for instagram access, postback data is %s", web.input.post_data)
		
	m_redis_client:set("OAuthTokens:"..tracking_code, json.encode({
		client_id = client_id,
		client_secret = client_secret,
		redirect_uri = redirect_uri,
		post_data = web.input.post_data,
		SID = inconcert.authUser and inconcert.authUser.authToken or nil
	}))
	
	local params = "display=popup"
	
	local new_url = cosmo.fill([[https://graph.facebook.com/oauth/authorize?client_id=$client_id&redirect_uri=$redirect_uri&state=$state&scope=$scope&$params]], {
		client_id = client_id,
		redirect_uri = redirect_uri,
		state = tracking_code,
		scope = scope,
		params = params
	})
	
	web.status = 302
	web.headers["Location"] = new_url
	return ""
end

--
-- Shows the page that the oauth provider redirects to when the user gaves permission
-- http://localhost:8082/inconcert/apps/oauth/facebook_auth_redirect?code=2.CjGFq5z2oF7xjOGfyQX71A__.86400.1280008800-1418538467|nngpd8EUkrSiBWWwwYYcWcklFQY.
local function RenderFormAuthRedirect(web)
	local _T = m_htmlRender.extension.makeTextTranslator()
	local tracking_code = web.input.state

	if not tracking_code then
		local message = "Can not obtain instagram tracking code from postback"
		LogError("[oauth] - %s", message)
		return m_htmlRender:RenderPageLP("pages/facebookAccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end
	
	LogInfo("[oauth] - Got instagram postback redirect with tracking_code: '%s'", tracking_code)
	
	local stored_code = m_redis_client:get("OAuthTokens:"..tracking_code)
	if not stored_code then
		local message = "Can not obtain instagram tracking code from Redis"
		LogError("[oauth] - %s: %s", message, tracking_code)
		return m_htmlRender:RenderPageLP("pages/instagramAccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end

	local decoded_transaction = json.decode(stored_code)
	local data = json.decode(decoded_transaction.post_data)

	if web.input.error_reason == "user_denied" then
		m_redis_client:del("OAuthTokens:"..tracking_code)
		return m_htmlRender:RenderPageLP("pages/instagramAccountError.lp", web, {
			messageError = "Auth rejected by the user",
			_T = _T
		})
	end

	local authorization_code = web.input.code
	local ok, message, access_token, values = RenewAccessToken(_T, authorization_code, decoded_transaction)
	if not ok then
		return m_htmlRender:RenderPageLP("pages/instagramAccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end
	
	LogInfo("[oauth] - Account: '%s' authorization succeeded.", values.uid)

	return m_htmlRender:RenderPageLP("pages/instagramAccountOk.lp", web, {
		authorization_code = authorization_code,
		access_token = values.access_token,
		uid = values.uid,
		name = values.name,
		origin = data.origin,
		_T = _T
	})
end

--
-- Given a 'authorization_code' (granted when the user gave permissions to this application), obtain a valid access token
function HandleRenewAccessToken(web)
	local _T = m_htmlRender.extension.makeTextTranslator()
	local code = web.input.authorization_code
	local ok, access_token, values = RenewAccessToken(_T, code)
	
	web:content_type("application", "json")
	return json.encode(values)
end


function MapUris(module, prefix)
	module:dispatch_get(RenderAuthorizationForm, prefix.."/authorize_user/(.*)")
	module:dispatch_get(RenderFormAuthRedirect, prefix.."/auth_redirect")
	module:dispatch_get(HandleRenewAccessToken, prefix.."/renew_access_token")
end

function SetViewEnvironment(env)
	assert(env and env.htmlRender)
	m_htmlRender = env.htmlRender
	
end

function SetConfig (redis_client, conf)
	m_config = conf
	m_redis_client = redis_client
end
