-- Micrsoft OAuth controller

require "inconcert.orbit.application"
require "AsyncTasks"

local cosmo = require "cosmo"
local http_request = require "shared.http.request"
local json = require "json"
local wsapi_util = require "wsapi.util"

local campaigns = require("apps.agent.models.campaigns")

module(..., package.seeall, inconcert.orbit.application.Decorate)

local m_htmlRender = nil
local m_config = nil
local m_redis_client
local DEFAULT_SCOPE = 
"offline_access "..
"https://outlook.office365.com/IMAP.AccessAsUser.All "..
"https://outlook.office365.com/POP.AccessAsUser.All "..
"https://outlook.office365.com/SMTP.Send "..
"https://outlook.office365.com/User.Read "..
"https://outlook.office365.com/profile"

local GRAPH_SCOPE = "https://graph.microsoft.com/User.Read"

---
-- Given a 'authorization_code' (granted when the user gave permissions to this application)
-- obtain a valid access token
--
-- @tparam string authorization_code
-- @tparam table codes
-- @tparam string codes.client_id
-- @tparam string codes.redirect_uri
-- @tparam string codes.client_secret
--
-- @treturn boolean ok           - If the renew went well or not
-- @treturn string message       - If the renew went well wrong, an error message (optional)
-- @treturn string access_token
-- @treturn table values
--
local function RenewAccessToken(_T, authorization_code, codes)
	--get token
	local url = "https://login.microsoftonline.com/".. m_config.tenant .."/oauth2/v2.0/token"
	local params = {
		grant_type = "authorization_code",
		client_id = codes.client_id,
		client_secret = codes.client_secret,
		redirect_uri = codes.redirect_uri,
		code = authorization_code
	}
	LogInfo("[Oauth] - Request Token, params = %s", json.encode(params))
	local code, values = http_request.postObject(url, nil, "url", params, nil, "json", "text")
	if code ~= 200 then
		local message = _T("Error in https request ") .. "(login.microsoftonline.com/"..m_config.tenant.."/oauth2/v2.0/token): " .. (values or "")
		LogError("[oauth] - %s", message )
		return false, message
	end
	LogInfo("[Oauth] - Request Resp Token, values = %s", json.encode(values))

	--get refresh token with scope GRAPH
        local params_refresh = {
                grant_type = "refresh_token",
                scope = GRAPH_SCOPE,
                client_id = codes.client_id,
                client_secret = codes.client_secret,
                redirect_uri = codes.redirect_uri,
                refresh_token = values.refresh_token
        }
        LogInfo("[Oauth] - Refresh Request Token, params = %s", json.encode(params_refresh))
        local code_refresh, values_refresh = http_request.postObject(url, nil, "url", params_refresh, nil, "json", "text")
        if code_refresh ~= 200 then
                local message = _T("Error in https refresh request ") .. "(login.microsoftonline.com/"..m_config.tenant.."/oauth2/v2.0/token): " .. (values_refresh or "")
                LogError("[oauth] - %s", message )
                return false, message
        end
        LogInfo("[Oauth] - Request Resp Refresh Token, values = %s", json.encode(values_refresh))
	
	-- Get name and email from https://graph.microsoft.com/v1.0/me/
	local tokenInfoUrl = "https://graph.microsoft.com/v1.0/me/"
	local headers = {
		authorization = "Bearer ".. values_refresh.access_token
	}
	local tokenInfoCode, tokenInfoValues = http_request.getObject(tokenInfoUrl, nil, headers, "json", "text")
	if tokenInfoCode ~= 200 then
		local message = _T("Error in https request ") .. "(https://graph.microsoft.com/v1.0/me/): " .. (tokenInfoValues or "")
		LogError("[oauth] - %s", message )
		return false, message
	end
	
	if not tokenInfoValues.mail or not tokenInfoValues.userPrincipalName then
		local message = _T("Error in https request ") .. "(https://graph.microsoft.com/v1.0/me/): " .. json.encode(tokenInfoValues)
		LogError("[oauth] - %s", message )
		return false, message
	end
	values.uid  = tokenInfoValues.mail
	values.name = tokenInfoValues.userPrincipalName

	LogInfo("[Oauth] - TokenInfoValues = %s", json.encode(tokenInfoValues))
	return true, "", values.access_token, values
end

local function RegisterAuthWebhook(webhook_host, subdomain_host, tracking_code)
	local url = string.format("https://%s/registeroauth", webhook_host)
	local params = {
		callback_env = subdomain_host,
		state = tracking_code
	}
	local code, response, headers = http_request.get(url, params)
    if code ~= 200 then
    	LogError("[oauth] - Error in https request RegisterAuthWebhook: %s , code: %s", response,code )
        return false
	end
	return true
end

local function GetStoredValues(webhook_host, uid)
	local url = string.format("https://%s/auth_get", webhook_host)
	local params = {
		id = uid
	}
	local code, response, headers = http_request.getObject(url, params, nil, "json")
    if code ~= 200 then
    	LogError("[oauth] - Error in https request GetStoredValues: %s , code: %s", response, code )
        return false
	end
	return response
end

local function StoreValues(webhook_host, uid, data)
	local url = string.format("https://%s/auth_set", webhook_host)
	local params = {
		id = uid,
		data = data
	}
	local code, response, headers = http_request.postObject(url, nil, "json", params)
    if code ~= 200 then
    	LogError("[oauth] - Error in https request StoreValues: %s , code: %s", response,code )
        return false
	end
	return true
end

--
-- Shows the authorization form for a given oauth provider
-- Probablememte deva devolver acá markup, por ejemplo un iframe para que se inserte en una página o algo así...
local function RenderAuthorizationForm(web)
	local _T = m_htmlRender.extension.makeTextTranslator()
	
	if m_config == nil then
		local message = "Error in RenderAuthorizationForm: Must configure application oauth in apps_config.lua"
		LogError("[oauth] - %s ", message)
		return m_htmlRender:RenderPageLP("pages/office365AccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end
	
	local scope = web.input.scope or DEFAULT_SCOPE
	
	local client_id =  m_config.client_id or web.input.client_id
	local client_secret =  m_config.client_secret or web.input.client_secret
	local postData = json.decode(web.input.post_data)
	local webhook_host = m_config.webhook_host
	local oauthTenant = m_config.tenant
	
	-- We use the tracking code to identify the data to be saved in the session
	local tracking_code = uuid.new()

	if not RegisterAuthWebhook(webhook_host, postData.host, tracking_code) then
		web.status = 500
		LogError("[oauth] - Error register office365 token, code %s body %s headers %s", err, body, json.encode(headers))
		return "Error requesting microsoft token"
	end
	
	local redirect_uri = "https://" .. webhook_host .. "/office365/auth_redirect"
	LogInfo("[oauth] - Requesting token for office365 access, postback data is %s", web.input.post_data)
		
	m_redis_client:set("OAuthTokens:"..tracking_code, json.encode({
		client_id = client_id,
		client_secret = client_secret,
		redirect_uri = redirect_uri,
		post_data = web.input.post_data,
		SID = inconcert.authUser and inconcert.authUser.authToken or nil
	}))
	
	local new_url = cosmo.fill([[
		https://login.microsoftonline.com/$tenant/oauth2/v2.0/authorize?client_id=$client_id&response_type=code&redirect_uri=$redirect_uri&scope=$scope&state=$state]], {
		client_id = client_id,
		redirect_uri = redirect_uri,
		state = tracking_code,
		scope = scope,
		tenant = oauthTenant
	})
	
	web.status = 302
	web.headers["Location"] = new_url
	return ""
end

--
-- Shows the page that the oauth provider redirects to when the user gaves permission
local function RenderFormAuthRedirect(web)
	local _T = m_htmlRender.extension.makeTextTranslator()
	local tracking_code = web.input.state

	if not tracking_code then
		local message = "Can not obtain office365 tracking code from postback"
		LogError("[oauth] - %s", message)
		return m_htmlRender:RenderPageLP("pages/office365AccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end
	
	LogInfo("[oauth] - Got office365 postback redirect with tracking_code: '%s'", tracking_code)
	
	local stored_code = m_redis_client:get("OAuthTokens:"..tracking_code)
	if not stored_code then
		local message = "Can not obtain office365 tracking code from Redis"
		LogError("[oauth] - %s: %s", message, tracking_code)
		return m_htmlRender:RenderPageLP("pages/office365AccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end

	local decoded_transaction = json.decode(stored_code)
	local data = json.decode(decoded_transaction.post_data)

	if web.input.error then
		LogError("[oauth] office365 error: %s", web.input.error)
		m_redis_client:del("OAuthTokens:"..tracking_code)
		return m_htmlRender:RenderPageLP("pages/office365AccountError.lp", web, {
			messageError = "Auth rejected by the user",
			_T = _T
		})
	end

	local authorization_code = web.input.code
	LogInfo("[oauth] - Account: authorization code %s.", authorization_code)
	local ok, message, access_token, values = RenewAccessToken(_T, authorization_code, decoded_transaction)
	if not ok then
		return m_htmlRender:RenderPageLP("pages/office365AccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end
	
	LogInfo("[oauth] - Account: '%s' authorization succeeded.", values.uid)

	local credentialsId = decoded_transaction.client_id .. "_" .. values.uid

	if values.refresh_token and values.refresh_token ~= "" then
		StoreValues(m_config.webhook_host, credentialsId, values)
	else
		local storedValues = GetStoredValues(m_config.webhook_host, credentialsId)
		if storedValues and storedValues.refresh_token then
			values.refresh_token = storedValues.refresh_token
		else
			LogError("[oauth] office365 refresh_token not found for uid: %s", values.uid)
			return m_htmlRender:RenderPageLP("pages/office365AccountError.lp", web, {
				messageError = _T("Refresh token not found"),
				_T = _T
			})
		end
	end

	return m_htmlRender:RenderPageLP("pages/office365AccountOk.lp", web, {
		client_id = decoded_transaction.client_id,
		client_secret = decoded_transaction.client_secret,
		access_token = values.access_token,
		refresh_token = values.refresh_token,
		tenant = m_config.tenant,
		uid = values.uid,
		name = values.name,
		origin = data.origin,
		_T = _T
	})
end

--
-- Given a 'authorization_code' (granted when the user gave permissions to this application), obtain a valid access token
function HandleRenewAccessToken(web)
	local _T = m_htmlRender.extension.makeTextTranslator()
	local code = web.input.authorization_code
	local ok, access_token, values = RenewAccessToken(_T, code)
	
	web:content_type("application", "json")
	return json.encode(values)
end


function MapUris(module, prefix)
	module:dispatch_get(RenderAuthorizationForm, prefix.."/authorize_user/(.*)")
	module:dispatch_get(RenderFormAuthRedirect, prefix.."/auth_redirect")
	module:dispatch_get(HandleRenewAccessToken, prefix.."/renew_access_token")
end

function SetViewEnvironment(env)
	assert(env and env.htmlRender)
	m_htmlRender = env.htmlRender
	
end

function SetConfig (redis_client, conf)
	m_config = conf
	m_redis_client = redis_client
end
