-- OAuth application

local orbit = require "orbit"
require "inconcert.orbit.application"
require("inconcert.translators")
local json = require "json"
local cosmo = require "cosmo"
local lfs = require "lfs"
require "webHandler.path"

local m_appLocation = select(2, ...)

local m_requireArgs = { ... }

local http = require "socket.http"
local ssl = ssl
local https = require "ssl.https"

local LuaConfigurationControl = require "LuaConfigurationControl"
local WrapRedis = require "wrapredis"
local cassandra = require "lua_cassandra"
local m_appConfig = {}

local controllers = {}

local configControl = nil
local m_redis_client

-- TODO:
-- ver si los permisos los configuro as� tambien o vienen como par�metro en la url del auth_request
-- hacer una p�gina para manejar cuando un usuario NO da permiso


module(..., package.seeall, orbit.new, inconcert.orbit.application.Decorate)

local function makeTextTranslator()
	local translator = inconcert.translators.oauth
	return function(...)
		return translator:Text(...)
	end, i18nlib.LCID[translator.language]
end

function GetApplicationTitle()
	local _T = makeTextTranslator()
	return _T("oauth")
end

local htmlRender = require"inconcert.render".new("oauth", {
	MainTitle = _M.GetApplicationTitle,
	Title = function() return _T("oauth") end,
	model = nil,
	makeTextTranslator = makeTextTranslator
})

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",
	
	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,
	
	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end
		
		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"twitter", "facebook", "instagram", "youtube", "googleplay", "redis", "whatsapp"})

		app_status.current = newConfig

		local redis_host, redis_port, redis_pwd = newConfig["redis"]["host"], tonumber(newConfig["redis"]["port"]), newConfig["redis"]["pwd"]
		
		LogInfo("apps.oauth: Connecting to Redis instance %s:%d", redis_host, redis_port)
		m_redis_client = WrapRedis(redis_host, redis_port, redis_pwd)

		if controllers["facebook"].SetConfig then
			controllers["facebook"].SetConfig(m_redis_client, {
				client_id = newConfig["facebook"]["appId"],
				client_secret = newConfig["facebook"]["appSecret"],
				webhook_host = newConfig["facebook"]["webhookHost"],
			})
		end

		if controllers["instagram"].SetConfig then
			controllers["instagram"].SetConfig(m_redis_client, {
				client_id = newConfig["instagram"]["appId"],
				client_secret = newConfig["instagram"]["appSecret"],
				webhook_host = newConfig["instagram"]["webhookHost"],
			})
		end

		if controllers["youtube"].SetConfig then
			controllers["youtube"].SetConfig(m_redis_client, {
				client_id = newConfig["youtube"]["appId"],
				client_secret = newConfig["youtube"]["appSecret"],
				webhook_host = newConfig["youtube"]["webhookHost"],
			})
		end

		if controllers["googleplay"].SetConfig then
			controllers["googleplay"].SetConfig(m_redis_client, {
				webhook_host = newConfig["googleplay"]["webhookHost"],
			})
		end

		if controllers["gmail"].SetConfig then
			controllers["gmail"].SetConfig(m_redis_client, {
				webhook_host = newConfig["gmail"]["webhookHost"],
			})
		end
		
		if controllers["twitter"].SetConfig then
			controllers["twitter"].SetConfig(m_redis_client, {
				consumerKey = newConfig["twitter"]["consumerKey"],
				consumerSecret = newConfig["twitter"]["consumerSecret"],
				webhook_host = newConfig["twitter"]["webhookHost"]
			})
		end
		
		if controllers["office365"].SetConfig then
			controllers["office365"].SetConfig(m_redis_client, {
				webhook_host = newConfig["office365"]["webhookHost"],
				client_id = newConfig["office365"]["client_id"],
				client_secret = newConfig["office365"]["client_secret"],
				tenant = newConfig["office365"]["tenant"],
			})
		end

		if controllers["whatsapp"].SetConfig then
			controllers["whatsapp"].SetConfig(m_redis_client, {
				client_id = newConfig["whatsapp"]["icc_cloud_api_client_id"],
				client_secret = newConfig["whatsapp"]["icc_cloud_api_app_secret"],
				webhook_host = newConfig["whatsapp"]["icc_cloud_api_webhook"],
				icc_business_id = newConfig["whatsapp"]["icc_cloud_api_icc_business_id"],
				system_user_token = newConfig["whatsapp"]["icc_cloud_api_system_user_token"],
			})
		end
		
		app_status.actualStatus = "ready"
		return "ready"
	end
}


--
-- Helper function to load config files
function LoadConfig(filename)
	return loadfile( webHandler.path.normalize(m_appLocation .. "\\" .. filename) )
end

--
-- Helper function, to filter values (should use Penlight instead...)
local function filter(condition, iterator, t, initial)
	return function(value, b, c)
		local output = iterator(value)
		while output and not condition(output) do
			output = iterator(value)
		end
		return output
	end, t, initial
end



-- 
-- This function gets called when the web server initializes this application (once per VM)
function Initialize(config)
	LogInfo("apps.oauth: Starting application OAuth")
    
	configControl = LuaConfigurationControl.Client {
		type = "OAUTH",
		address = inconcert.environment.apps.oauth.config_server.ip,
		port = inconcert.environment.apps.oauth.config_server.port,
		node_address = inconcert.environment.apps.oauth.config_server.myIp,
		node_port = inconcert.environment.apps.oauth.config_server.myPort,
	}
	
	configControl:TriggerNodeReady()
	
    --  map controllers. The url where they'll be mapped is the controller name (or check for a value supplied by the controller itself)
    for name in filter( function(v) return v:match("%.lua$") end, lfs.dir( webHandler.path.normalize(m_appLocation .. "\\controllers") ) ) do
        name = name:match("(.*)%.lua$")
     	local controller = relativeRequire("controllers."..name, unpack(m_requireArgs))
		controllers[name] = controller
		
        if controller.SetViewEnvironment then
            controller.SetViewEnvironment{ htmlRender = htmlRender }
        end
		
        controller.MapUris(_M, "/" ..name)
    end
     
	return true
end

--scripts and pages

--paginas
local function render_pages(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:RenderPageLP("pages/" .. name .. ".html", web, env)
end

local function render_pages_lp(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:RenderPageLP("pages/" .. name .. ".lp", web, env)
end

--scripts
function render_script(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:ScriptLp("scripts/" .. name .. ".js", web, env)
end

_M:dispatch_get(render_script, "/scripts/([^.]+)%.([^.]+)%.?([^.]*)%.js")
_M:dispatch_get(render_pages, "/pages/([^.]+)%.([^.]+)%.?([^.]*)%.html")
_M:dispatch_get(render_pages_lp, "/pages/([^.]+)%.([^.]+)%.?([^.]*)%.lp")

--
--  styles and resources
--  NOTA: ESTO SE PROCESA EN XAVANTE

local function serve_file(app_module, cache)
	return function (web)
		local filename = string.sub(web.path_info, 2, #web.path_info)
		if not cache then
			app_module.response.DontCache(web)
		end
		return _M:serve_static(web, "/apps/oauth/"..filename)
	end
end

_M:dispatch_get(serve_file(_M), "/images/.+", "/styles/.+", "/css/.+", ".+%.mp3")
