local utils = require "shared.apps.utils"
local modelGateways = require("apps.provisioning.models.gateways")
local m_permissions = require "shared.permissions.permissions"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

----------------------------------
-- Gateways
----------------------------------

local function Render (web)
	_M.response.DontCache(web)
	local env = { _T = m_htmlRender.extension.makeTextTranslator() }
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("Render form Gateway with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local sections = {
		{ id = "basic"                    , file = "frm_gateway_content_basic.html"                    		   },
		{ id = "carrier_trunks"           , file = "frm_gateway_content_carrier_trunks.html"    			   },
		{ id = "node_trunks"              , file = "frm_gateway_content_node_trunks.html"       			   },
		{ id = "dialplan_contexts"        , file = "frm_gateway_content_dialplan_contexts.html" 			   },
	}
	local page_dir = "pages/"
	local html_content = ""
	for _, section in ipairs(sections) do
		local html_file = page_dir .. section.file
		local html_section = m_htmlRender:RenderPageLP(html_file, web, env)
		if html_section then
			html_content = html_content .. 
				"<div form_section='" .. section.id .. "' class='pageFormContainerAux' hidden>" ..
					html_section ..
				"</div>"
		else
			LogError("HTML Gateway section does not exist: %s", html_file)
		end
	end

	return html_content
end

local function GetGateway (web, gatewayId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:GetGateway] Retrieving GATEWAY '%s'. User %s.", gatewayId, userId)

	local success, gatewayInfo = modelGateways.Get(gatewayId)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:GetGateway] Error Retrieving GATEWAY %s: %s", gatewayId, gatewayInfo)
		web.status = 404
		return { result = gatewayInfo }
	end

	web.status = 200
	return gatewayInfo
end

local function ValidateAndMakeInfoToUpdate (gatewayId, gatewayInfo)
	assert(gatewayId)
	assert(gatewayInfo)

	LogInfo("[PROVISIONING.GATEWAYS:ValidateAndMakeInfoToUpdate] Validating GATEWAY '%s' info to updated.", gatewayId)

	local gatewayExist, currentGatewayInfo = modelGateways.Get(gatewayId)

	-- 1 - clean up valores que llegan del lado del cliente

	if gatewayExist and gatewayInfo.isNew then
		-- Si estoy creando el Gateway y el mismo ya exsite en la BD, fallo!!
		return false, "MUST_HAVE_UNIQUE_NAME"
	end

	gatewayInfo.isNew = nil

	return true, gatewayInfo
end

local function SaveGateway (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	--LogInfo("[PROVISIONING.GATEWAYS:SaveGateway] Saving GATEWAY. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, gateways = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:SaveGateway] Error Updating GATEWAY. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local _, gatewayInfo = next(gateways)
	-- LogDebug("--> _ %s", _)
	-- LogDebug("--> gatewayInfo %s", gatewayInfo)

	local gatewayId = gatewayInfo.Id
	local success, gatewayInfo = ValidateAndMakeInfoToUpdate(gatewayId, gatewayInfo)
	if not success then
		local errMsg = gatewayInfo
		LogError("[PROVISIONING.GATEWAYS:SaveGateway] Error Validating GATEWAY '%s'. Err: %s", gatewayId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local response = modelGateways.Save(gatewayId, gatewayInfo)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:SaveGateway] Error Updating GATEWAY '%s'. Err: %s", gatewayId, response.errMsg or "")
		web.status = 500
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:SaveGateway] '%s' saved successfully.", gatewayId)

	web.status = 200
	result.result = "OK"
	result.configuration = response.configuration
	return result
end

local function DeleteGateway (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:DeleteGateway] Deleting GATEWAY. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, gateways = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGateway] Error Deleting GATEWAY. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local gatewayId = gateways[1]
	local success, vccInfo = modelGateways.Get(gatewayId)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGateway] Error Deleting GATEWAY %s. Not Found.", gatewayId)
		web.status = 404
		result.result = "GATEWAY DOES NOT EXIST"
		return result
	end

	local response = modelGateways.Delete(gatewayId)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGateway] Error Deleting GATEWAY %s. Error: %s.", gatewayId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:DeleteGateway] GATEWAY '%s' deleting successfully.", gatewayId)

	web.status = 200
	result.result = "OK"
	return result
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.GATEWAYS:Search] 1 - (START) Loading Gateways. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, gateways = modelGateways.ListAll(keyspace)

	if not success then
		keyspace:Close()
		local err = gateways
		LogWarning("[PROVISIONING.GATEWAYS:Search] Error loading Gateways. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	local totalUnfilteredRows = utils.GetRowCount(gateways)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.GATEWAYS:SEARCH] 2 - Filtering rows")

	for id, gatewayData in pairs(gateways) do
		local gatewayId = utils.GetId(id)
		local gateway = gatewayData.configuration
		if gateway then
            gateway.Id = gatewayId
            gateway.Name = gateway.Name or gatewayId
            if utils.MatchFilter(gateway, postData) then
                totalAdded = totalAdded + 1
                table.insert(filteredRows, gateway)
            end
        end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.GATEWAYS:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user gateways
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.GATEWAYS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadGateways (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS.:LoadGateways] Loading Gateways, user %s, vcc %s", userId, vcc)

	local success, gateways = modelGateways.ListAll()
	if not success then
		local err = gateways
		LogWarning("[PROVISIONING.GATEWAYS:LoadGateways] Error loading Gateways. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return gateways
end

----------------------------------
-- Node Trunks
----------------------------------

local function GetGatewayNodeTrunk (web, gatewayId, trunkId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:GetGatewayNodeTrunk] Retrieving GATEWAY '%s' trunk '%s'. User %s.",
			gatewayId, trunkId, userId)

	local success, trunkInfo = modelGateways.GetNodeTrunk(gatewayId, trunkId)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:GetGatewayNodeTrunk] Error Retrieving GATEWAY %s trunk %s. Not Found.",
				 trunkId, gatewayId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return trunkInfo
end

local function ValidateAndMakeNodeTrunkInfoToUpdate (gatewayId, trunkId, trunkInfo)
	assert(gatewayId)
	assert(trunkId)
	assert(trunkInfo)

	LogInfo("[PROVISIONING.GATEWAYS:ValidateAndMakeNodeTrunkInfoToUpdate] Validating TRUNK '%s' info to updated.", trunkId)

	local trunkExist, currentTrunkInfo = modelGateways.GetNodeTrunk(gatewayId, trunkId)

	-- 1 - clean up valores que llegan del lado del cliente

	if trunkExist and trunkInfo.isNew then
		-- Si estoy creando el trunk y el mismo ya exsite en la BD, fallo!!
		return false, "MUST_HAVE_UNIQUE_NAME"
	end

	trunkInfo.isNew = nil

	return true, trunkInfo
end

local function SaveGatewayNodeTrunk (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:SaveGatewayNodeTrunk] Saving TRUNK. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, trunks = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayNodeTrunk] Error Updating TRUNK. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local _, trunkInfo = next(trunks)
	-- LogDebug("--> _ %s", _)
	-- LogDebug("--> trunkInfo %s", trunkInfo)

	local trunkId = trunkInfo.Id
	local success, trunkInfo = ValidateAndMakeNodeTrunkInfoToUpdate(gatewayId, trunkId, trunkInfo)
	if not success then
		local errMsg = gatewayInfo
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayNodeTrunk] Error Validating TRUNK '%s'. Err: %s", trunkId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local response = modelGateways.SaveNodeTrunk(gatewayId, trunkId, trunkInfo)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayNodeTrunk] Error Updating TRUNK %s. Error: %s.", trunkId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:SaveGatewayNodeTrunk] '%s' saved successfully.", trunkId)

	web.status = 200
	result.result = "OK"
	result.trunk = response.trunk
	return result
end

local function DeleteGatewayNodeTrunk (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:DeleteGatewayNodeTrunk] Deleting TRUNK. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, trunks = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayNodeTrunk] Error Deleting TRUNK. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local trunkId = trunks[1]
	local success, vccInfo = modelGateways.GetNodeTrunk(gatewayId, trunkId)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayNodeTrunk] Error Deleting TRUNK %s. Not Found.", gatewayId)
		web.status = 404
		result.result = "TRUNK DOES NOT EXIST"
		return result
	end

	local response = modelGateways.DeleteNodeTrunk(gatewayId, trunkId)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayNodeTrunk] Error Deleting TRUNK %s. Error: %s.", trunkId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:DeleteGatewayNodeTrunk] TRUNK '%s' deleting successfully.", trunkId)

	web.status = 200
	result.result = "OK"
	return result
end

local function SearchNodeTrunks (web, gatewayId)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.GATEWAYS:SearchNodeTrunks] 1 - (START) Loading Trunks. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, trunks = modelGateways.ListAllNodeTrunks(keyspace, gatewayId)

	if not success then
		keyspace:Close()
		local err = trunks
		LogWarning("[PROVISIONING.GATEWAYS:SearchNodeTrunks] Error loading trunks. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	local totalUnfilteredRows = utils.GetRowCount(trunks)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.GATEWAYS:SearchNodeTrunks] 2 - Filtering rows")

	for id, trunk in pairs(trunks) do
		local trunkId = utils.GetId(id)
		trunk.Id = trunkId
		trunk.Name = trunk.Name or trunkId
		if utils.MatchFilter(trunk, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, trunk)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.GATEWAYS:SearchNodeTrunks] 3 - Sorting rows")
	--include only rows that belong to logged user trunks
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.GATEWAYS:SearchNodeTrunks] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadGatewayNodeTrunks (web, gatewayId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS.:LoadGatewayNodeTrunks] Loading trunks, user %s, vcc %s", userId, vcc)

	local success, trunks = modelGateways.ListAllNodeTrunks(gatewayId)
	if not success then
		local err = trunks
		LogWarning("[PROVISIONING.GATEWAYS:LoadGatewayNodeTrunks] Error loading trunks. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return trunks
end

----------------------------------
-- Carrier Trunks
----------------------------------

local function GetGatewayCarrierTrunk (web, gatewayId, trunkId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:GetGatewayCarrierTrunk] Retrieving GATEWAY '%s' trunk '%s'. User %s.",
			gatewayId, trunkId, userId)

	local success, trunkInfo = modelGateways.GetCarrierTrunk(gatewayId, trunkId)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:GetGatewayCarrierTrunk] Error Retrieving GATEWAY %s trunk %s. Not Found.",
				 trunkId, gatewayId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return trunkInfo
end

local function ValidateAndMakeCarrierTrunkInfoToUpdate (gatewayId, trunkId, trunkInfo)
	assert(gatewayId)
	assert(trunkId)
	assert(trunkInfo)

	LogInfo("[PROVISIONING.GATEWAYS:ValidateAndMakeCarrierTrunkInfoToUpdate] Validating TRUNK '%s' info to updated.", trunkId)

	local trunkExist, currentTrunkInfo = modelGateways.GetCarrierTrunk(gatewayId, trunkId)

	-- 1 - clean up valores que llegan del lado del cliente

	if trunkExist and trunkInfo.isNew then
		-- Si estoy creando el trunk y el mismo ya exsite en la BD, fallo!!
		return false, "MUST_HAVE_UNIQUE_NAME"
	end

	trunkInfo.isNew = nil

	return true, trunkInfo
end

local function SaveGatewayCarrierTrunk (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:SaveGatewayCarrierTrunk] Saving TRUNK. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, trunks = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayCarrierTrunk] Error Updating TRUNK. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local _, trunkInfo = next(trunks)
	-- LogDebug("--> _ %s", _)
	-- LogDebug("--> trunkInfo %s", trunkInfo)

	local trunkId = trunkInfo.Id
	local success, trunkInfo = ValidateAndMakeCarrierTrunkInfoToUpdate(gatewayId, trunkId, trunkInfo)
	if not success then
		local errMsg = gatewayInfo
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayCarrierTrunk] Error Validating TRUNK '%s'. Err: %s", trunkId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local response = modelGateways.SaveCarrierTrunk(gatewayId, trunkId, trunkInfo)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayCarrierTrunk] Error Updating TRUNK %s. Error: %s.", trunkId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:SaveGatewayCarrierTrunk] '%s' saved successfully.", trunkId)

	web.status = 200
	result.result = "OK"
	result.trunk = response.trunk
	return result
end

local function DeleteGatewayCarrierTrunk (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:DeleteGatewayCarrierTrunk] Deleting TRUNK. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, trunks = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayCarrierTrunk] Error Deleting TRUNK. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local trunkId = trunks[1]
	local success, vccInfo = modelGateways.GetCarrierTrunk(gatewayId, trunkId)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayCarrierTrunk] Error Deleting TRUNK %s. Not Found.", gatewayId)
		web.status = 404
		result.result = "TRUNK DOES NOT EXIST"
		return result
	end

	local response = modelGateways.DeleteCarrierTrunk(gatewayId, trunkId)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayCarrierTrunk] Error Deleting TRUNK %s. Error: %s.", trunkId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:DeleteGatewayCarrierTrunk] TRUNK '%s' deleting successfully.", trunkId)

	web.status = 200
	result.result = "OK"
	return result
end

local function SearchCarrierTrunks (web, gatewayId)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.GATEWAYS:SearchCarrierTrunks] 1 - (START) Loading Trunks. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, trunks = modelGateways.ListAllCarrierTrunks(keyspace, gatewayId)

	if not success then
		keyspace:Close()
		local err = trunks
		LogWarning("[PROVISIONING.GATEWAYS:SearchCarrierTrunks] Error loading trunks. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	local totalUnfilteredRows = utils.GetRowCount(trunks)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.GATEWAYS:SearchCarrierTrunks] 2 - Filtering rows")

	for id, trunk in pairs(trunks) do
		local trunkId = utils.GetId(id)
		trunk.Id = trunkId
		trunk.Name = trunk.Name or trunkId
		if utils.MatchFilter(trunk, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, trunk)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.GATEWAYS:SearchCarrierTrunks] 3 - Sorting rows")
	--include only rows that belong to logged user trunks
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.GATEWAYS:SearchCarrierTrunks] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadGatewayCarrierTrunks (web, gatewayId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS.:LoadGatewayCarrierTrunks] Loading trunks, user %s, vcc %s", userId, vcc)

	local success, trunks = modelGateways.ListAllCarrierTrunks(gatewayId)
	if not success then
		local err = trunks
		LogWarning("[PROVISIONING.GATEWAYS:LoadGatewayCarrierTrunks] Error loading trunks. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return trunks
end

----------------------------------
-- Dialplan Contexts
----------------------------------

local function GetGatewayDialplanContext (web, gatewayId, contextId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:GetGatewayDialplanContext] Retrieving GATEWAY '%s' context '%s'. User %s.",
			gatewayId, contextId, userId)

	local success, contextInfo = modelGateways.GetDialplanContext(gatewayId, contextId)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:GetGatewayDialplanContext] Error Retrieving GATEWAY %s context %s. Not Found.",
				 contextId, gatewayId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return contextInfo
end

local function ValidateAndMakeDialplanContextInfoToUpdate (gatewayId, contextId, contextInfo)
	assert(gatewayId)
	assert(contextInfo)
	assert(contextInfo)

	LogInfo("[PROVISIONING.GATEWAYS:ValidateAndMakeDialplanContextInfoToUpdate] Validating CONTEXT '%s' info to updated.", contextId)

	local contextExist, currentContextInfo = modelGateways.GetDialplanContext(gatewayId, contextId)

	-- 1 - clean up valores que llegan del lado del cliente

	if contextExist and contextInfo.isNew then
		-- Si estoy creando el context y el mismo ya exsite en la BD, fallo!!
		return false, "MUST_HAVE_UNIQUE_NAME"
	end

	contextInfo.isNew = nil

	return true, contextInfo
end

local function ValidateGatewayDialplanContext (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:ValidateGatewayDialplanContext] Validating CONTEXT. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, contexts = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:ValidateGatewayDialplanContext] Error validating CONTEXT. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local _, contextInfo = next(contexts)
	-- LogDebug("--> _ %s", _)
	-- LogDebug("--> contextInfo %s", contextInfo)

	local contextId = contextInfo.Id
	local success, contextInfo = ValidateAndMakeDialplanContextInfoToUpdate(gatewayId, contextId, contextInfo)
	if not success then
		local errMsg = gatewayInfo
		LogError("[PROVISIONING.GATEWAYS:ValidateGatewayDialplanContext] Error Validating CONTEXT '%s'. Err: %s", contextId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local response = modelGateways.ValidateDialplanContext(gatewayId, contextId, contextInfo)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:ValidateGatewayDialplanContext] Error validating CONTEXT %s. Error: %s.", contextId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		result.context = response.context
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:ValidateGatewayDialplanContext] '%s' validated successfully.", contextId)

	web.status = 200
	result.result = "OK"
	return result
end

local function SaveGatewayDialplanContext (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:SaveGatewayDialplanContext] Saving CONTEXT. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, contexts = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayDialplanContext] Error Updating CONTEXT. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local _, contextInfo = next(contexts)
	-- LogDebug("--> _ %s", _)
	-- LogDebug("--> contextInfo %s", contextInfo)

	local contextId = contextInfo.Id
	local success, contextInfo = ValidateAndMakeDialplanContextInfoToUpdate(gatewayId, contextId, contextInfo)
	if not success then
		local errMsg = gatewayInfo
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayDialplanContext] Error Validating CONTEXT '%s'. Err: %s", contextId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local response = modelGateways.SaveDialplanContext(gatewayId, contextId, contextInfo)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:SaveGatewayDialplanContext] Error Updating CONTEXT %s. Error: %s.", contextId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		result.context = response.context
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:SaveGatewayDialplanContext] '%s' saved successfully.", contextId)

	web.status = 200
	result.result = "OK"
	return result
end

local function DeleteGatewayDialplanContext (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:DeleteGatewayDialplanContext] Deleting CONTEXT. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, contexts = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayDialplanContext] Error Deleting CONTEXT. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local contextId = contexts[1]
	local success, vccInfo = modelGateways.GetDialplanContext(gatewayId, contextId)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayDialplanContext] Error Deleting CONTEXT %s. Not Found.", gatewayId)
		web.status = 404
		result.result = "CONTEXT DOES NOT EXIST"
		return result
	end

	local response = modelGateways.DeleteDialplanContext(gatewayId, contextId)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:DeleteGatewayDialplanContext] Error Deleting CONTEXT %s. Error: %s.", contextId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:DeleteGatewayDialplanContext] CONTEXT '%s' deleting successfully.", contextId)

	web.status = 200
	result.result = "OK"
	return result
end

local function SearchDialplanContexts (web, gatewayId)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.GATEWAYS:SearchDialplanContexts] 1 - (START) Loading Contexts. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, contexts = modelGateways.ListAllDialplanContexts(keyspace, gatewayId)

	if not success then
		keyspace:Close()
		local err = contexts
		LogWarning("[PROVISIONING.GATEWAYS:SearchDialplanContexts] Error loading contexts. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	local totalUnfilteredRows = utils.GetRowCount(contexts)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.GATEWAYS:SearchDialplanContexts] 2 - Filtering rows")

	for id, context in pairs(contexts) do
		local contextId = utils.GetId(id)
		context.Id = contextId
		context.Name = context.Name or contextId
		if utils.MatchFilter(context, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, context)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.GATEWAYS:SearchDialplanContexts] 3 - Sorting rows")
	--include only rows that belong to logged user contexts
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.GATEWAYS:SearchDialplanContexts] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadGatewayDialplanContexts (web, gatewayId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS.:LoadGatewayDialplanContexts] Loading contexts, user %s, vcc %s", userId, vcc)

	local success, contexts = modelGateways.ListAllDialplanContexts(gatewayId)
	if not success then
		local err = contexts
		LogWarning("[PROVISIONING.GATEWAYS:LoadGatewayDialplanContexts] Error loading contexts. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return contexts
end


----------------------------------
-- Monitoring
----------------------------------

local function SaveMonitoring (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:SaveMonitoring] User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, monitoring = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:SaveMonitoring] Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local response = modelGateways.SaveMonitoring(gatewayId, monitoring)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:SaveMonitoring] Error: %s.", response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:SaveMonitoring] '%s' successfully.", gatewayId)

	web.status = 200
	response.result = "OK"
	return response
end

local function MonitoringGet (web, gatewayId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringGet] User %s.", userId)

	local result = {}

	local status, response = modelGateways.Get(gatewayId)
	local monitoring = status and response.monitoring or {}

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringGet] '%s' successfully.", gatewayId)

	web.status = 200
	return { result = "OK", monitoring = monitoring }
end

local function MonitoringGrafanaWidgets (web, gatewayId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringGrafanaWidgets] User %s.", userId)

	local result = {}

	local response = modelGateways.MonitoringGrafanaWidgets(gatewayId)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:MonitoringGrafanaWidgets] Error: %s.", response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringGrafanaWidgets] '%s' successfully.", gatewayId)

	web.status = 200
	response.result = "OK"
	return response
end

local function MonitoringSipCaptureWidgets (web, gatewayId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringSipCaptureWidgets] User %s.", userId)

	local result = {}

	local response = modelGateways.MonitoringSipCaptureWidgets(gatewayId)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:MonitoringSipCaptureWidgets] Error: %s.", response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringSipCaptureWidgets] '%s' successfully.", gatewayId)

	web.status = 200
	response.result = "OK"
	return response
end

local function MonitoringStopSipCaptureWidgets (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringStopSipCaptureWidgets] User %s posted data '%s'.", userId, postedData)

	local result = {}

	local response = modelGateways.MonitoringStopSipCaptureWidgets(gatewayId)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:MonitoringStopSipCaptureWidgets] Error: %s.", response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringStopSipCaptureWidgets] '%s' successfully.", gatewayId)

	web.status = 200
	response.result = "OK"
	return response
end

local function MonitoringSendConsoleCommand (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringSendConsoleCommand] User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, monitoring = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:MonitoringSendConsoleCommand] Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local response = modelGateways.MonitoringSendConsoleCommand(gatewayId, monitoring)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:MonitoringSendConsoleCommand] Error: %s.", response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringSendConsoleCommand] '%s' successfully.", gatewayId)

	web.status = 200
	response.result = "OK"
	return response
end

local function MonitoringConfigurationFile (web, gatewayId)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringConfigurationFile] User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, monitoring = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.GATEWAYS:MonitoringConfigurationFile] Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local response = modelGateways.MonitoringConfigurationFile(gatewayId, monitoring.fileName)
	if not response.success then
		LogError("[PROVISIONING.GATEWAYS:MonitoringConfigurationFile] Error: %s.", response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.reason = response.errMsg
		return result
	end

	LogInfo("[PROVISIONING.GATEWAYS:MonitoringConfigurationFile] '%s' successfully.", gatewayId)

	web.status = 200
	response.result = "OK"
	return response
end

----------------------------------
-- Mapas de las URLS
----------------------------------
function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix

	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")
	-- GATEWAYS i6
	module:dispatch_post(fn .. Search, p .. "/search/"                                          )
	module:dispatch_get (HasPermission(GetGateway   , "view_gateways"     ), p .. "/get/(.+)"   )
	module:dispatch_post(HasPermission(SaveGateway  , "save_gateway", fn  ), p .. "/save/"      )
	module:dispatch_post(HasPermission(DeleteGateway, "delete_gateway", fn), p .. "/delete/(.*)")
	module:dispatch_get (replyFn .. LoadGateways, p .. "/load/"                                 )

	-- GATEWAYS monitoring
	module:dispatch_post(HasPermission(SaveMonitoring, "save_gateway", fn  ), p .. "/monitoring/save/(.+)")
	module:dispatch_get (HasPermission(MonitoringGet, "view_gateways"), p .. "/monitoring/get/(.+)"   )
	module:dispatch_get (HasPermission(MonitoringGrafanaWidgets, "view_gateways"), p .. "/monitoring/grafana_widgets/(.+)"   )
	module:dispatch_get (HasPermission(MonitoringSipCaptureWidgets, "view_gateways"), p .. "/monitoring/sip_capture_widgets/(.+)"   )
	module:dispatch_post(HasPermission(MonitoringStopSipCaptureWidgets, "view_gateways"), p .. "/monitoring/stop_sip_capture_widgets/(.+)"   )
	module:dispatch_post(HasPermission(MonitoringSendConsoleCommand, "view_gateways", fn  ), p .. "/monitoring/send_console_command/(.+)")
	module:dispatch_post(HasPermission(MonitoringConfigurationFile, "view_gateways", fn  ), p .. "/monitoring/configuration_file/(.+)")

	-- GATEWAYS node trunks
	module:dispatch_post(fn .. SearchNodeTrunks, p .. "/node_trunks/search/(.+)"                                       		)
	module:dispatch_get (HasPermission(GetGatewayNodeTrunk   , "view_gateways"     ), p .. "/node_trunks/get/(.+)/(.+)"   	)
	module:dispatch_post(HasPermission(SaveGatewayNodeTrunk  , "save_gateway", fn  ), p .. "/node_trunks/save/(.+)"      	)
	module:dispatch_post(HasPermission(DeleteGatewayNodeTrunk, "delete_gateway", fn), p .. "/node_trunks/delete/(.+)"		)
	module:dispatch_get (replyFn .. LoadGatewayNodeTrunks, p .. "/node_trunks/load/(.+)"                               		)

	-- GATEWAYS carrier trunks
	module:dispatch_post(fn .. SearchCarrierTrunks, p .. "/carrier_trunks/search/(.+)"                                     		)
	module:dispatch_get (HasPermission(GetGatewayCarrierTrunk   , "view_gateways"     ), p .. "/carrier_trunks/get/(.+)/(.+)"   )
	module:dispatch_post(HasPermission(SaveGatewayCarrierTrunk  , "save_gateway", fn  ), p .. "/carrier_trunks/save/(.+)"      	)
	module:dispatch_post(HasPermission(DeleteGatewayCarrierTrunk, "delete_gateway", fn), p .. "/carrier_trunks/delete/(.+)"		)
	module:dispatch_get (replyFn .. LoadGatewayCarrierTrunks, p .. "/carrier_trunks/load/(.+)"                                 	)

	-- GATEWAYS dialplan contexts
	module:dispatch_post(fn .. SearchDialplanContexts, p .. "/dialplan_contexts/search/(.+)"                                       	)
	module:dispatch_get (HasPermission(GetGatewayDialplanContext   , "view_gateways"     ), p .. "/dialplan_contexts/get/(.+)/(.+)" )
	module:dispatch_post(HasPermission(ValidateGatewayDialplanContext  , "save_gateway", fn  ), p .. "/dialplan_contexts/validate/(.+)" )
	module:dispatch_post(HasPermission(SaveGatewayDialplanContext  , "save_gateway", fn  ), p .. "/dialplan_contexts/save/(.+)"     )
	module:dispatch_post(HasPermission(DeleteGatewayDialplanContext, "delete_gateway", fn), p .. "/dialplan_contexts/delete/(.+)"	)
	module:dispatch_get (replyFn .. LoadGatewayDialplanContexts, p .. "/dialplan_contexts/load/(.+)"                               	)

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end