local utils = require "shared.apps.utils"
local modelIceServers = require("apps.provisioning.models.ice_servers")
local m_permissions = require "shared.permissions.permissions"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING:Render] Render form Ice Server with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local htmlFile = m_appLocation .. "/pages/" .. "frm_ice_server_content.html"

	local htmlContent = m_htmlRender:RenderPageLP(htmlFile, web, env)
	if not htmlContent then
		assert(htmlContent, "[PROVISIONING:Render] HTML ICE SERVER does not exist: %s", htmlFile)
	end

	return htmlContent
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local encodedInfo           = web.input["post_data"]
	local postData              = web.request
	local startTime             = os.time()

	LogInfo("[PROVISIONING.ICE_SERVERS:Search] 1 - (START) Loading Ice Servers. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, result = modelIceServers.ListAll(keyspace)

	if not success then
		keyspace:Close()
		LogWarning("[PROVISIONING.ICE_SERVERS:Search] Error loading Ice Servers. Err'%s'.", result)
		web.status = 500
		return {result = "SERVER ERROR", err = result}
	end

	local iceServers          = result.userConfigs or {}
	local totalUnfilteredRows = utils.GetRowCount(iceServers)
	local totalAdded          = 0
	local filteredRows        = {}

	LogInfo("[PROVISIONING.ICE_SERVERS:SEARCH] 2 - Filtering rows")

	for _, server in ipairs(iceServers) do
		if utils.MatchFilter(server, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, server)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.ICE_SERVERS:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user hosts
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.domain)--:upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.ICE_SERVERS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

local function GetIceServer (web, serverIdStr)

	local response = {}
	local serverId = tonumber(serverIdStr)

	LogInfo("[PROVISIONING.ICE_SERVERS:GetIceServer] Retrieving Ice Server '%d'.", serverId)

	local success, result = modelIceServers.Get(serverId)
	if not success or not result.success then
		LogError("[PROVISIONING.ICE_SERVERS:GetIceServer] Error Retrieving Ice Server %d. Error: %s", serverId, json.encode(result))
		web.status      = 404
		response.result = result.errCode and result.errCode or result

		return response
	end

	web.status = 200
	return result.data
end

local function SaveIceServer (web)

	local result
	local response             = {}
	local postedData           = web.input["post_data"]
	local success, decodedData = pcall(json.decode, postedData)

	LogInfo("[PROVISIONING.ICE_SERVERS:SaveIceServer] Saving ICE SERVER. Posted data '%s'.", postedData)

	if not success or not decodedData.user or not decodedData.password or not decodedData.domain or not decodedData.port then
		LogError("[PROVISIONING.ICE_SERVERS:SaveIceServer] Error Saving ICE SERVER. Invalid Request.")
		web.status      = 400
		response.result = "INVALID REQUEST"

		return response
	end

	success, result = modelIceServers.Save(decodedData)
	if not success or not result.success then
		LogError("[PROVISIONING.ICE_SERVERS:SaveIceServer] Error Saving Ice Server. Error: %s", json.encode(result))
		web.status      = 500
		response.result = result.errCode and result.errCode or "SERVER ERROR"

		return response
	end

	web.status    = 200
	result.result = "OK"

	return result
end

local function DeleteIceServer (web)
	local postedData = web.input["post_data"]

	LogInfo("[PROVISIONING.ICE_SERVERS:DeleteIceServer] Deleting ICE SERVER. Posted data '%s'.", postedData)

	local response = {}

	local success, iceServerData = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.ICE_SERVERS:DeleteIceServer] Error Deleting ICE SERVER. Invalid Request.")
		web.status      = 400
		response.result = "INVALID REQUEST"
		return response
	end

	local result
	local serverId  = tonumber(iceServerData[1])
	
	success, result = modelIceServers.Delete(serverId)
	if not success or not result.success then
		LogError("[PROVISIONING.ICE_SERVERS:DeleteIceServer] Error Deleting Ice Server: '%d'. Error: %s.", serverId, json.encode(result))
		web.status      = 400
		response.result = result.errCode and result.errCode or result

		return response
	end

	LogInfo("[PROVISIONING.ICE_SERVERS:DeleteIceServer] Ice Server: '%d' deleting successfully.", serverId)

	web.status      = 200
	response.result = "OK"

	return response
end

----------------------------------
-- Mapas de las URLS
----------------------------------
function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix

	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- HOST i6
	module:dispatch_post(fn .. Search, p .. "/search/"                                    )
	module:dispatch_get (HasPermission(GetIceServer   , "view_ice_servers"     ), p .. "/get/(.+)"   )
	module:dispatch_post(HasPermission(SaveIceServer  , "save_ice_server", fn  ), p .. "/save/"      )
	module:dispatch_post(HasPermission(DeleteIceServer, "delete_ice_server", fn), p .. "/delete/(.*)")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end
