local utils = require "shared.apps.utils"
local modelRoutes = require("apps.provisioning.models.routes")
local m_permissions = require "shared.permissions.permissions"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.ROUTES:Render] Render form ROUTES with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local htmlFile = m_appLocation .. "/pages/" .. "frm_route_content.html"

	local htmlContent = m_htmlRender:RenderPageLP(htmlFile, web, env)
	if not htmlContent then
		assert(htmlContent, "[PROVISIONING.ROUTES:Render] HTML ROUTES does not exist: %s", htmlFile)
	end

	return htmlContent
end

local function ValidateAndMakeInfoToUpdate (id, info)
	LogInfo("[PROVISIONING.ROUTES:ValidateAndMakeInfoToUpdate] Validating ROUTES '%s' info to updated. Info %s", id, info)

	if id then
		local exist, currentInfo = modelRoutes.Get(id)
		-- 1 - clean up valores que llegan del lado del cliente
		local isNew = info.isNew

		if exist then
			-- Si estoy creando la Entidad y el mismo ya exsite en la BD, fallo!!
			if isNew then
				return false, "MUST_HAVE_UNIQUE_NAME"
			end
		end
	end
	-- id
	if not info.Id or info.Id == "" then
		return false, "ID_IS_REQUIRED"
	end
	-- description
	if not info.description or info.description == "" then
		return false, "DESCRIPTION_IS_REQUIRED"
	end
	-- schedulers 
	local schedulers = info.schedulers
	if not schedulers or not next(schedulers) then
		return false, "SCHEDULERS_IS_REQUIRED"
	end

	info.isNew = nil
	info.schedulers = nil

	return true, info, schedulers
end

local function SaveRoute (web)
	local postedData = web.input["post_data"]

	LogInfo("[PROVISIONING.ROUTES:SaveRoute] Saving ROUTE. User %s@%s posted data '%s'.", inconcert.authUser.userName, inconcert.authUser.virtualContactCenter, postedData)

	local result = {}

	local success, rows = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.ROUTES:SaveRoute] Error Updating Route. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local rowId, rowInfo = next(rows)
	-- LogDebug("--> rowId %s", rowId)
	-- LogDebug("--> rowInfo %s", rowInfo)

	local success, rowInfo, schedulers = ValidateAndMakeInfoToUpdate(rowId, rowInfo)
	if not success then
		local errMsg = rowInfo
		LogError("[PROVISIONING.ROUTES:Route] Error Validating Route '%s'. Err: %s", rowId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local success, err = modelRoutes.Save(rowId, rowInfo)
	if not success then
		LogError("[PROVISIONING.ROUTES:Route] Error Updating Route '%s'. Err: %s", rowId, err)
		web.status = 500
		result.result = err.code -- Este código de error llega desde la Rocka "provisioningmanager"
		result.rawData = err.rawData
		return result
	end

	local success, err = modelRoutes.SaveSchedulers(rowId, schedulers)
	if not success then
		LogError("[PROVISIONING.ROUTES:Route] Error Updating Route '%s'. Err: %s", rowId, err)
		web.status = 500
		result.result = err.code -- Este código de error llega desde la Rocka "provisioningmanager"
		result.rawData = err.rawData
		return result
	end

	LogInfo("[PROVISIONING.ROUTES:SaveRoute] '%s' saved successfully.", rowId)

	result.result = "OK"
	return result
end

local function DeleteRoute (web)
	local postedData = web.input["post_data"]
	local success, routeInfo = pcall(json.decode, postedData)

	LogInfo("[PROVISIONING.ROUTES:DeleteRoute] Deleting Route. User %s@%s posted data '%s'.", inconcert.authUser.userName, inconcert.authUser.virtualContactCenter, postedData)

	local result = {}

	local success, routes = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.ROUTES:DeleteRoute] Error Deleting Route. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local routeId = routes[1]
	local ksProv = utils.GetDatasource():CassandraConnection("Provisioning")

	local success, vccInfo = modelRoutes.Get(routeId, ksProv)
	if not success then
		ksProv:Close()
		LogError("[PROVISIONING.ROUTES:DeleteRoute] Error Deleting Route %s. Not Found.", routeId)
		web.status = 404
		result.result = "Route DOES NOT EXIST"
		return result
	end

	local response = modelRoutes.Delete(routeId, ksProv)
	if not response.success then
		LogError("[PROVISIONING.ROUTES:DeleteRoute] Error Deleting Route %s. Error: %s.", routeId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.ROUTES:DeleteRoute] Route '%s' deleting successfully.", routeId)

	ksProv:Close()
	web.status = 200
	result.result = "OK"

	return result
end

local function GetRoute (web, routeId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.ROUTES:GetRoute] Retrieving Route '%s'. User %s.", routeId, userId)

	local success, routeInfo = modelRoutes.Get(routeId)
	if not success then
		LogError("[PROVISIONING.ROUTES:GetRoute] Error Retrieving Route %s. Not Found.", routeId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return routeInfo
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.ROUTES:Search] 1 - (START) Loading Routes. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, routes = modelRoutes.ListAll(keyspace)

	if not success then
		keyspace:Close()
		local err = routes
		LogWarning("[PROVISIONING.ROUTES:Search] Error loading Routes. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	--routes = utils.FilterEntities(routes, {IsSystemVCC = "0"})

	local totalUnfilteredRows = utils.GetRowCount(routes)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.ROUTES:SEARCH] 2 - Filtering rows")

	for id, route in pairs(routes) do
		local routeId = utils.GetId(id)
		route.Id = routeId
		route.Name = route.Name or routeId
		if utils.MatchFilter(route, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, route)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.ROUTES:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user routes
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.ROUTES:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadRoutes (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.ROUTES.:LoadRoutes] Loading Routes, user %s, vcc %s", userId, vcc)

	local success, routes = modelRoutes.ListAll()
	if not success then
		local err = routes
		LogWarning("[PROVISIONING.ROUTES:LoadRoutes] Error loading Routes. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return routes
end

--[[
]]--
local function LoadRoutesTrunksPriorityList (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local _T = m_htmlRender.extension.makeTextTranslator()

	LogInfo("[PROVISIONING.ROUTES.:LoadRoutesTrunksPriorityList] Loading Routes Trunks Priority List, user %s, vcc %s", userId, vcc)

	local success, routesSchedulers = modelRoutes.LoadRoutesSchedulers()
	if not success then
		local err = routesSchedulers
		LogWarning("[PROVISIONING.ROUTES:LoadRoutesTrunksPriorityList] Error loading Routes Trunks Priority List. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	local trunksPriorityList = {}
	local i = 0
	for routeId, routeInfo in pairs(routesSchedulers) do
		for schTimeId, trunksPriorityLst in pairs(routeInfo) do
			local trunksPriorityId = ""
			for trunkId, priority in pairs(trunksPriorityLst) do
				trunksPriorityId = trunksPriorityId .. (trunksPriorityId:len() > 0 and "_" or "")
				trunksPriorityId = trunksPriorityId .. tostring(trunkId) .. "_" .. tostring(priority)
			end
			if trunksPriorityId:len() > 0 then
				trunksPriorityId = trunksPriorityId:lower()
				if not trunksPriorityList[trunksPriorityId] then
					i = i + 1
					trunksPriorityList[trunksPriorityId] = {
						Id = trunksPriorityId,
						Name = _T("List") .. " " .. tostring(i),
						Trunks = trunksPriorityLst,
						Routes = {}
					}
				end

				trunksPriorityList[trunksPriorityId].Routes[routeId] = trunksPriorityList[trunksPriorityId].Routes[routeId] or {}
				table.insert(trunksPriorityList[trunksPriorityId].Routes[routeId], schTimeId)
			end
		end
	end

	web.status = 200
	return trunksPriorityList
end

function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix
	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- Route i6
	module:dispatch_post(fn .. Search, p .. "/search/")
	module:dispatch_get (HasPermission(GetRoute   , "view_routes"     ), p .. "/get/(.+)"   )
	module:dispatch_post(HasPermission(DeleteRoute, "delete_route", fn), p .. "/delete/(.*)")
	module:dispatch_post(HasPermission(SaveRoute  , "save_route", fn  ), p .. "/save/"      )
	module:dispatch_get (replyFn .. LoadRoutes, p .. "/load/")
	module:dispatch_get (replyFn .. LoadRoutesTrunksPriorityList, p .. "/load/list/trunks_priority/")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end