local utils = require "shared.apps.utils"
local modelTrunks = require("apps.provisioning.models.trunks")
local m_permissions = require "shared.permissions.permissions"
require "lua_telephonyproxyclient"
require "uuid"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil
local m_clientInstance = nil

local function GetClient ()
	if not m_clientInstance then
		m_clientInstance = lua_telephonyproxyclient.LuaTelephonyProxyClient(
			inconcert.environment.apps.agent.cluster, "webhandler")
	end
	local tpAddresses = inconcert.environment.apps.provisioning.call or {}
	for _, address in pairs(tpAddresses) do
		m_clientInstance:AddKnownTelephonyProxy(address.host, address.port)
	end
	m_clientInstance:ConfigureCassandra({
		{
			host = inconcert.environment.apps.provisioning.cassandra.ip,
			port = inconcert.environment.apps.provisioning.cassandra.port
		}
	})
	m_clientInstance:ConfigureTimeClient(
		inconcert.environment.apps.provisioning.timeServers)
	return m_clientInstance
end

local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING:Render] Render form TRUNK with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local htmlFile = m_appLocation .. "/pages/" .. "frm_trunk_content.html"

	local htmlContent = m_htmlRender:RenderPageLP(htmlFile, web, env)
	if not htmlContent then
		assert(htmlContent, "[PROVISIONING:Render] HTML TRUNK does not exist: %s", htmlFile)
	end

	return htmlContent
end

local function ValidateAndMakeInfoToUpdate (id, info)
	assert(id)
	assert(info)

	LogInfo("[PROVISIONING.TRUNKS:ValidateAndMakeInfoToUpdate] Validating TRUNK '%s' info to updated. Info %s", id, info)

	local exist, currentInfo = modelTrunks.Get(id)

	-- 1 - clean up valores que llegan del lado del cliente
	local isNew = info.isNew

	if exist then
		-- Si estoy creando la Entidad y el mismo ya exsite en la BD, fallo!!
		if isNew then
			return false, "MUST_HAVE_UNIQUE_NAME"
		end
	end

	info.isNew = nil

	return true, info
end

local function SaveTrunk (web)
	local postedData = web.input["post_data"]
	LogInfo("[PROVISIONING.TRUNKS:SaveTrunk] Saving Trunk. User %s@%s posted data '%s'.", inconcert.authUser.userName, inconcert.authUser.virtualContactCenter, postedData)

	local result = {}

	local success, rows = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.TRUNKS:SaveTrunk] Error Updating Trunk. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local rowId, rowInfo = next(rows)
	--LogDebug("--> rowId %s", rowId)
	--LogDebug("--> rowInfo %s", rowInfo)

	local success, rowInfo = ValidateAndMakeInfoToUpdate(rowId, rowInfo)
	if not success then
		local errMsg = rowInfo
		LogError("[PROVISIONING.TRUNKS:SaveTrunk] Error Validating Trunk '%s'. Err: %s", rowId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local success, err = modelTrunks.Save(rowId, rowInfo)
	if not success then
		LogError("[PROVISIONING.TRUNKS:SaveTrunk] Error Updating Trunk '%s'. Err: %s", rowId, err)
		web.status = 500
		result.result = err.code -- Este código de error llega desde la Rocka "provisioningmanager"
		result.rawData = err.rawData
		return result
	end

	LogInfo("[PROVISIONING.TRUNKS:SaveTrunk] '%s' saved successfully.", rowId)

	web.status = 200
	result.result = "OK"
	return result
end

local function DeleteTrunk (web)
	local postedData = web.input["post_data"]
	local success, trunkInfo = pcall(json.decode, postedData)

	LogInfo("[PROVISIONING.TRUNKS:DeleteTrunk] Deleting Trunk. User %s@%s posted data '%s'.", inconcert.authUser.userName, inconcert.authUser.virtualContactCenter, postedData)

	local result = {}

	local success, trunks = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.TRUNKS:DeleteTrunk] Error Deleting TRUNK. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local trunkId = trunks[1]

	local ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, trunkInfo = modelTrunks.Get(trunkId, ksProv)
	if not success or not trunkInfo then
		ksProv:Close()
		LogError("[PROVISIONING.TRUNKS:DeleteTrunk] Error Deleting TRUNK %s. Not Found.", trunkId)
		web.status = 404
		result.result = "TRUNK DOES NOT EXIST"
		return result
	end

	local response = modelTrunks.Delete(trunkId, ksProv)
	if not response.success then
		LogError("[PROVISIONING.TRUNKS:DeleteTrunk] Error Deleting TRUNK %s. Error: %s.", trunkId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.TRUNKS:DeleteTrunk] Trunk '%s' deleting successfully.", trunkId)

	ksProv:Close()
	web.status = 200
	result.result = "OK"

	return result
end

local function GetTrunk (web, trunkId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TRUNKS:GetTrunk] Retrieving TRUNK '%s'. User %s.", trunkId, userId)

	local success, trunkInfo = modelTrunks.Get(trunkId)
	if not success then
		LogError("[PROVISIONING.TRUNKS:GetTrunk] Error Retrieving TRUNK %s. Not Found.", trunkId)
		web.status = 404
		local result = {}
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return trunkInfo
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.TRUNKS:Search] 1 - (START) Loading Trunks. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, trunks = modelTrunks.ListAll(keyspace)

	if not success then
		keyspace:Close()
		local err = trunks
		LogWarning("[PROVISIONING.TRUNKS:Search] Error loading Trunks. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	--trunks = utils.FilterEntities(trunks, {IsSystemVCC = "0"})

	local totalUnfilteredRows = utils.GetRowCount(trunks)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.TRUNKS:SEARCH] 2 - Filtering rows")

	for id, trunk in pairs(trunks) do
		local trunkId = utils.GetId(id)
		trunk.Id = trunkId
		trunk.Name = trunk.Name or trunkId
		if utils.MatchFilter(trunk, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, trunk)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.TRUNKS:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user trunks
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.TRUNKS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadTrunks (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TRUNKS.:LoadTrunks] Loading Trunks, user %s, vcc %s", userId, vcc)

	local success, trunks = modelTrunks.ListAll()
	if not success then
		local err = trunks
		LogWarning("[PROVISIONING.TRUNKS:LoadTrunks] Error loading Trunks. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return trunks
end

--[[
]]--
local function LoadTrunksByCarriers (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TRUNKS.:LoadTrunksByCarriers] Loading Trunks By Carriers, user %s, vcc %s", userId, vcc)

	local success, trunks = modelTrunks.ListTrunksByCarrier()
	if not success then
		local err = trunks
		LogWarning("[PROVISIONING.TRUNKS:LoadTrunksByCarriers] Error loading Trunks By Carriers. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return trunks
end

local function GetAmdResults (web, trunkId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TRUNKS:GetAmdResults] Retrieving TRUNK '%s' amd results. User %s.", trunkId, userId)
	local trunkResults = modelTrunks.GetAmdResults(trunkId)

	web.status = 200
	return trunkResults
end

local function StartAmdTest (web, trunkId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	postData.status = "RUNNING"
	postData.timestamp = os.time(os.date("!*t"))

	LogInfo("[PROVISIONING.TRUNKS:StartAmdTest] Starting TRUNK '%s' test '%s'. User %s.",
			trunkId, encodedInfo, userId)
	local client = GetClient()
	local result = client:ValidatePhone({ VCC = vcc, number = postData.number })
	if result.result ~= "OK" then
		result.result = "PHONE_INVALID"
		web.status = 500
		return result
	end

	local destination = result.data
	local channelID = uuid.new()
	result = client:MakeCall({
		interactionID = channelID,
		channelID = channelID,
		trunksPackages = { [trunkId] = {} },
		destination = destination,
		trunks = { trunkId },
		amdTest = postData,
		amd = true
	})
	if result.result ~= "OK" then
		result.result = "ERROR_MAKING_CALL"
		web.status = 500
		return result
	end

	modelTrunks.StartAmdResult(trunkId, postData)
	web.status = 200
	return { result = "OK" }
end

local function ResetFailsMonitor (web, trunkId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TRUNKS:ResetFailsMonitor] Starting TRUNK '%s'. User %s.",
			trunkId, userId)
	local client = GetClient()
	local result = client:ResetFailsMonitor({ trunks = { trunkId } })
	if result.result ~= "OK" then
		web.status = 500
	end
	return result
end

function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix
	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- TRUNK i6
	module:dispatch_post(fn .. Search, p .. "/search/")
	module:dispatch_get (replyFn .. LoadTrunks, p .. "/load/")
	module:dispatch_get (replyFn .. LoadTrunksByCarriers, p .. "/load/by_carriers/")
	module:dispatch_get (HasPermission(GetTrunk   , "view_trunks"     ), p .. "/get/(.+)"   )
	module:dispatch_get (HasPermission(GetAmdResults   , "view_trunks"     ), p .. "/amd_results/(.+)"   )
	module:dispatch_post(HasPermission(DeleteTrunk, "delete_trunk", fn), p .. "/delete/(.*)")
	module:dispatch_post(HasPermission(SaveTrunk  , "save_trunk", fn  ), p .. "/save/"      )
	module:dispatch_post(HasPermission(StartAmdTest  , "save_trunk", fn  ), p .. "/amd_start_test/(.+)"  )
	module:dispatch_post(HasPermission(ResetFailsMonitor  , "save_trunk", fn  ), p .. "/reset_fails_monitor/(.+)"  )

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end