local utils = require "shared.apps.utils"
local modelTtsAsrEngines = require("apps.provisioning.models.ttsasrengines")
local m_permissions = require "shared.permissions.permissions"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

local uniMrcpLogLevel = {
	DEBUG = "DEBUG",
	INFO = "INFO",
	NOTICE = "NOTICE",
	WARNING = "WARNING",
	ERROR = "ERROR",
	CRITICAL = "CRITICAL",
	ALERT = "ALERT",
	EMERGENCY = "EMERGENCY"
}

local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING:Render] Render form TtsAsrEngine for user %s, group %s", userId, inconcert.authUser.userGroup)

	local htmlFile = m_appLocation .. "/pages/" .. "frm_ttsasrengines_content.html"

	local htmlContent = m_htmlRender:RenderPageLP(htmlFile, web, env)
	if not htmlContent then
		assert(htmlContent, "[PROVISIONING:Render] HTML TtsAsrEngine does not exist: %s", htmlFile)
	end

	return htmlContent
end

local function GetTtsAsrEngine (web, ttsAsrEngineId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.UNIMRCPPROFILES:GetTtsAsrEngine] Retrieving TtsAsrEngine '%s'. User %s.", ttsAsrEngineId, userId)

	local result = {}
	
	local success, ttsasrEngineInfo = modelTtsAsrEngines.Get(ttsAsrEngineId)
	if not success then
		LogError("[PROVISIONING.UNIMRCPPROFILES:GetTtsAsrEngine] Error Retrieving TtsAsrEngine %s. Not Found.", uniMrcpProfileId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end
	
	ttsasrEngineInfo.general = json.decode(ttsasrEngineInfo.general or "")
	ttsasrEngineInfo.profiles = json.decode(ttsasrEngineInfo.profiles or {})

	web.status = 200
	return ttsasrEngineInfo
end

local function ValidateAndMakeInfoToUpdate (ttsAsrEngineId, ttsasrEngineInfo)
	assert(ttsAsrEngineId)
	assert(ttsasrEngineInfo)

	LogInfo("[PROVISIONING.TTSASRENGINES:ValidateAndMakeInfoToUpdate] Validating TtsAsrEngine '%s' info to updated.", ttsAsrEngineId)

	local ttsasrEngineExist, currentUniMrcpProfileInfo = modelTtsAsrEngines.Get(ttsAsrEngineId)

	-- 1 - clean up valores que llegan del lado del cliente
	local isNew = ttsasrEngineInfo.isNew

	if ttsasrEngineExist then
		-- Si estoy creando el TtsAsrEngine y el mismo ya existe en la BD, fallo!!
		if isNew then
			return false, "MUST_HAVE_UNIQUE_NAME"
		end
		
		-- Si estoy actualizando un TtsAsrEngine existente, los valores que NO vengan en el POST y ya estaban en la BD deben almacenarse.
		for name, value in pairs(currentUniMrcpProfileInfo) do
			if ttsasrEngineInfo[name] == nil then
				ttsasrEngineInfo[name] = value
			end
		end
	end
	
	--Required fields
	-- TtsAsrEngine id
	if not ttsasrEngineInfo.ttsAsrEngineId or ttsasrEngineInfo.ttsAsrEngineId == "" then
		return false, "TTSASR_ENGINE_ID_IS_REQUIRED"
	end
	
	local enabled = ttsasrEngineInfo.enabled
	if enabled == "0" then
		local vcc = modelTtsAsrEngines.GetVCCsByTtsAsrEngine(ttsAsrEngineId)
		if vcc then
			LogError("[PROVISIONING.TTSASRENGINES:ValidateAndMakeInfoToUpdate] Tried to disable engine %s, wich is in use by vcc %s.", ttsAsrEngineId, vcc)
			return false, "CANNOT_DISABLE_ENGINE_WHILE_IN_USE_BY_VCC"
		end
	end
	
	local general = ttsasrEngineInfo.general or {}
	local profiles = ttsasrEngineInfo.profiles or {}
	
	if ttsasrEngineInfo.ttsAsrEngineId == "unimrcp" then
		--Log Level
		if not general["log-level"] or general["log-level"] == "" then
			general["log-level"] = "INFO"
		else
			if uniMrcpLogLevel[general["log-level"]] == nil then
				return false, "LOG_LEVEL_INVALID"
			end
		end
		-- max-connection-count
		if not general["max-connection-count"] or general["max-connection-count"] == "" then
			general["max-connection-count"] = "100"
		end
		-- max-shared-count
		if not general["max-shared-count"] or general["max-shared-count"] == "" then
			general["max-shared-count"] = "100"
		end
		-- offer-new-connection
		if not general["offer-new-connection"] or general["offer-new-connection"] == "" then
			general["offer-new-connection"] = "1"
		end
		-- request-timeout
		if not general["request-timeout"] or general["request-timeout"] == "" then
			general["request-timeout"] = "5000"
		end
		-- speech-channel-timeout
		if not general["speech-channel-timeout"] or general["speech-channel-timeout"] == "" then
			general["speech-channel-timeout"] = "30000"
		end
		-- rx-buffer-size
		if not general["rx-buffer-size"] or general["rx-buffer-size"] == "" then
			general["rx-buffer-size"] = "1024"
		end
		-- tx-buffer-size
		if not general["tx-buffer-size"] or general["tx-buffer-size"] == "" then
			general["tx-buffer-size"] = "1024"
		end
	end
		
	ttsasrEngineInfo.general = json.encode(general)
		
	--Comprobar que no se haya eliminado un profile de este engine que se este usando en un vcc
	local profilesToSave = {}
	for _, profile in pairs(profiles) do
		profilesToSave[profile.id] = true
	end
	local success, profilesInUse = modelTtsAsrEngines.GetTtsAsrProfilesInUse(ttsasrEngineInfo.ttsAsrEngineId)
	if not success then
		return false, "COULD_NOT_GET_PROFILES_IN_USE"
	end
	for profile, vcc in pairs(profilesInUse) do
		if not profilesToSave[profile] then
			LogError("[PROVISIONING.TTSASRENGINES:ValidateAndMakeInfoToUpdate] Tried to delete %s profile, wich is in use by vcc %s.", profile, vcc)
			return false, "CAN_NOT_DELETE_PROFILES_IN_USE"
		end
	end
	
	ttsasrEngineInfo.profiles = json.encode(profiles)

	ttsasrEngineInfo.isNew = nil

	return true, ttsasrEngineInfo
end

local function SaveTtsAsrEngine (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:SaveTtsAsrEngine] Saving TtsAsrEngine. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, ttsasrEngines = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.TTSASRENGINES:SaveTtsAsrEngine] Error Updating TtsAsrEngine. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local _, ttsasrEngineInfo = next(ttsasrEngines)
	-- LogDebug("--> _ %s", _)
	-- LogDebug("--> ttsasrEngineInfo %s", ttsasrEngineInfo)

	assert(ttsasrEngineInfo.ttsAsrEngineId)
	ttsasrEngineInfo.ttsAsrEngineId = ttsasrEngineInfo.ttsAsrEngineId:lower()
	local ttsAsrEngineId = ttsasrEngineInfo.ttsAsrEngineId
	local success, ttsasrEngineInfo = ValidateAndMakeInfoToUpdate(ttsAsrEngineId, ttsasrEngineInfo)
	if not success then
		local errMsg = ttsasrEngineInfo
		LogError("[PROVISIONING.TTSASRENGINES:SaveTtsAsrEngine] Error Validating TtsAsrEngine '%s'. Err: %s", ttsAsrEngineId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local success, errMsg = modelTtsAsrEngines.Save(ttsAsrEngineId, ttsasrEngineInfo)
	if not success then
		LogError("[PROVISIONING.TTSASRENGINES:SaveTtsAsrEngine] Error Updating TtsAsrEngine '%s'. Err: %s", ttsAsrEngineId, errMsg)
		web.status = 500
		result.result = errMsg -- Este código de error llega desde la Rocka "provisioningmanager"
		return result
	end

	LogInfo("[PROVISIONING.TTSASRENGINES:SaveTtsAsrEngine] '%s' saved successfully.", ttsAsrEngineId)

	web.status = 200
	result.result = "OK"
	return result
end

local function DeleteTtsAsrEngine (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:DeleteTtsAsrEngine] Deleting TtsAsrEngine. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, ttsasrEngines = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.TTSASRENGINES:DeleteTtsAsrEngine] Error Deleting TtsAsrEngine. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local ttsAsrEngineId = ttsasrEngines[1]
	local success, vccInfo = modelTtsAsrEngines.Get(ttsAsrEngineId)
	if not success then
		LogError("[PROVISIONING.TTSASRENGINES:DeleteTtsAsrEngine] Error Deleting TtsAsrEngine %s. Not Found.", ttsAsrEngineId)
		web.status = 404
		result.result = "TTS/ASR ENGINE DOES NOT EXIST"
		return result
	end

	local response = modelTtsAsrEngines.Delete(ttsAsrEngineId)
	if not response.success then
		LogError("[PROVISIONING.TTSASRENGINES:DeleteTtsAsrEngine] Error Deleting TtsAsrEngine %s. Error: %s.", ttsAsrEngineId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.TTSASRENGINES:DeleteTtsAsrEngine] TtsAsrEngine '%s' deleting successfully.", ttsAsrEngineId)

	web.status = 200
	result.result = "OK"
	return result
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.TTSASRENGINES:Search] 1 - (START) Loading TtsAsr Engines. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, ttsasrEngines = modelTtsAsrEngines.ListProvisioningAll(keyspace)

	if not success then
		keyspace:Close()
		local err = ttsasrEngines
		LogWarning("[PROVISIONING.TTSASRENGINES:Search] Error loading UniMrcp Profiles. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	--ttsasrEngines = utils.FilterEntities(ttsasrEngines, {IsSystemVCC = "0"})

	local totalUnfilteredRows = utils.GetRowCount(ttsasrEngines)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.TTSASRENGINES:SEARCH] 2 - Filtering rows")

	for id, ttsasrEngine in pairs(ttsasrEngines) do
		local ttsAsrEngineId = utils.GetId(id)
		ttsasrEngine.Id = ttsAsrEngineId
		ttsasrEngine.Name = ttsasrEngine.Name or ttsAsrEngineId
		if utils.MatchFilter(ttsasrEngine, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, ttsasrEngine)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.TTSASRENGINES:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user ttsasrEngines
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.TTSASRENGINES:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadTtsAsrVccEngines (web, vccId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:LoadTtsAsrVccEngines] Loading TtsAsr Engines, user %s, vcc %s", userId, vcc)

	local success, engines = modelTtsAsrEngines.ListTtsAsrVccEngines(vccId)
	if not success then
		local err = engines
		LogWarning("[PROVISIONING.TTSASRENGINES:LoadTtsAsrVccEngines] Error loading TtsAsr Engines. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return engines
end

--[[
]]--
local function LoadTtsAsrProvisionigEngines (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:LoadTtsAsrProvisionigEngines] Loading TtsAsr Engines, user %s, vcc %s", userId, vcc)

	local success, engines = modelTtsAsrEngines.ListProvisioningAll()
	if not success then
		local err = engines
		LogWarning("[PROVISIONING.TTSASRENGINES:LoadTtsAsrProvisionigEngines] Error loading TtsAsr Engines. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return engines
end

--[[
]]--
local function ListProvisionedTtsEngines (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES.ListProvisionedTtsEngines] Loading Engines containing TTS Profiles, user %s, vcc %s", userId, vcc)

	local success, ttsEngines = modelTtsAsrEngines.ListProvisionedTtsEngines()
	if not success then
		local err = ttsEngines
		LogWarning("[PROVISIONING.TTSASRENGINES:ListProvisionedTtsEngines] Error loading Engines containing TTS Profiles. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return ttsEngines
end

--[[
]]--
local function ListProvisionedAsrEngines (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES.ListProvisionedAsrEngines] Loading Engines containing ASR Profiles, user %s, vcc %s", userId, vcc)

	local success, asrEngines = modelTtsAsrEngines.ListProvisionedAsrEngines()
	if not success then
		local err = asrEngines
		LogWarning("[PROVISIONING.TTSASRENGINES:ListProvisionedAsrEngines] Error loading Engines containing ASR Profiles. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return asrEngines
end

--[[
]]--
local function LoadTtsAsrProviders (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:LoadTtsAsrProviders] Loading TtsAsr Providers, user %s, vcc %s", userId, vcc)

	local success, ttsasrProviders = modelTtsAsrEngines.ListTtsAsrProviders()
	if not success then
		local err = ttsasrProviders
		LogWarning("[PROVISIONING.TTSASRENGINES:LoadTtsAsrProviders] Error loading TtsAsr Providers. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return ttsasrProviders
end

--[[
]]--
local function ListTtsAsrEnginesInUse(web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:ListTtsAsrEnginesInUse] User %s.", userId)

	local result = {}
	local success, enginesInUse = modelTtsAsrEngines.ListTtsAsrEnginesInUse()
	if not success then
		local errMsg = enginesInUse
		LogError("[PROVISIONING.TTSASRENGINES:ListTtsAsrEnginesInUse] Error Retrieving Engines In Use. Err: %s", errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	-- OK
	web.status = 200
	return enginesInUse
end

--[[
]]--
local function LoadTtsAsrProviderTypes (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:LoadTtsAsrProviderTypes] Loading TtsAsr Provider Types, user %s, vcc %s", userId, vcc)

	local success, ttsasrProviders = modelTtsAsrEngines.ListTtsAsrProviderTypes()
	if not success then
		local err = ttsasrProviders
		LogWarning("[PROVISIONING.TTSASRENGINES:LoadTtsAsrProviderTypes] Error loading TtsAsr Provider Types. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return ttsasrProviders
end

local function LoadUniMrcpLogLevels (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:LoadUniMrcpLogLevels] Loading UniMrcp LogLevels, user %s, vcc %s", userId, vcc)

	-- OK
	web.status = 200
	return uniMrcpLogLevel
end

local function LoadUsedTtsAsrProfiles(web, ttsAsrEngineId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.TTSASRENGINES:LoadUsedTtsAsrProfiles] User %s ttsAsrEngineId '%s'.", userId, ttsAsrEngineId)
	local result = {}

	local success, profilesInUse = modelTtsAsrEngines.GetTtsAsrProfilesInUse(ttsAsrEngineId)
	if not success then
		local errMsg = profilesInUse
		LogError("[PROVISIONING.TTSASRENGINES:LoadUsedTtsAsrProfiles] Error Retrieving Profiles In Use by TtsAsrEngine'%s'. Err: %s", ttsAsrEngineId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end
	
	-- OK
	web.status = 200
	return profilesInUse
end

----------------------------------
-- Mapas de las URLS
----------------------------------
function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix

	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- HOST i6
	module:dispatch_post(fn .. Search, p .. "/search/"                                                     )
	module:dispatch_get (HasPermission(GetTtsAsrEngine   , "view_ttsasrengines"     ), p .. "/get/(.+)"    )
	module:dispatch_post(HasPermission(SaveTtsAsrEngine  , "save_ttsasrengines", fn  ), p .. "/save/"      )
	module:dispatch_post(HasPermission(DeleteTtsAsrEngine, "delete_ttsasrengines", fn), p .. "/delete/(.*)")
	module:dispatch_get (replyFn .. LoadTtsAsrVccEngines, p .. "/vcc/load/(.+)"                            )
	module:dispatch_get (replyFn .. LoadTtsAsrProvisionigEngines, p .. "/provisioning/load/"               )
	module:dispatch_get (replyFn .. ListProvisionedTtsEngines, p .. "/loadttsprovisioned/"                 )
	module:dispatch_get (replyFn .. ListProvisionedAsrEngines, p .. "/loadasrprovisioned/"                 )
	module:dispatch_get (replyFn .. LoadUniMrcpLogLevels, p .. "/load/loglevels/"                          )
	module:dispatch_get (replyFn .. LoadTtsAsrProviders, p .. "/load/providers/"                           )
	module:dispatch_get (replyFn .. LoadTtsAsrProviderTypes, p .. "/load/provider_types/"                  )
	module:dispatch_get (replyFn .. ListTtsAsrEnginesInUse, p .. "/listttsasrenginesinuse/"                )
	module:dispatch_get (replyFn .. LoadUsedTtsAsrProfiles, p .. "/loadusedprofiles/(.*)"                  )

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end