local utils = require "shared.apps.utils"
local modelVCCs = require("apps.provisioning.models.vccs")
local model_config_restrictions = require "apps.agent.models.config_restrictions"
local modelPeers = require("apps.provisioning.models.peers")
local modelPackages = require("apps.provisioning.models.packages")
local modelUsers = require "apps.provisioning.models.users"
local m_permissions = require "shared.permissions.permissions"
local jwt = require "shared.jwt"
local audits = require "audits"
local m_helpers = require "helpers"
local m_validations = require("apps.provisioning.controllers.validations.vccs").GetDefinitions()

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

--[[
	MIGRATIONS - ALLEGRO
]]--
local m_migrations = {allegro = {}}

--[[
	INTEGRATIONS - HELPDESK
]]--
local m_integrations = {helpdesk = {}, widgets = {}}

--[[
	SUPER USERS
]]--
local m_superUsers = {}

--[[
	Queda expuesto para su uso de forma externa.
	Ej: desde la app "allegro" cuando se crea un vcc desde allegro.
]]--
function SuperUser ()
	return m_superUsers
end

-----------------------
local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING:Render] Render form VCC with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local sections = {
		{id = "basic"                , file = "frm_vccs_content_basic.html"                },
		{id = "cognitive_services"   , file = "frm_vccs_content_cognitive_services.html"   },
		{id = "omnichannel_telephony", file = "frm_vccs_content_omnichannel_telephony.html"}
	}
	local pageDir = "pages/"
	local htmlContent = ""
	for _, section in ipairs(sections) do
		local htmlFile = pageDir .. section.file
		--LogDebug("[PROVISIONING:Render] Open form File '%s' of Section: '%s'.", section.file, section.id)
		local htmlSection = m_htmlRender:RenderPageLP(htmlFile, web, env)
		if htmlSection then
			htmlContent = htmlContent .. 
				"<div form_section='" .. section.id .. "' class='pageFormContainerAux' hidden>" ..
					htmlSection ..
				"</div>"
		else
			assert(htmlSection, "[PROVISIONING:Render] HTML VCC section does not exist: %s", htmlFile)
		end
	end

	return htmlContent
end

function GetVCC (web, vccId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("[PROVISIONING.VCCS:GetVCC] Retrieving VCC '%s'. User %s.", vccId, userId)
	-- 1. obtengo informacion del VCC invocando a provisioningmanager
	local withCustomFields = true
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local success, vccInfo = modelVCCs.Get(vccId, ksAgent, withCustomFields)
	if not success then
		ksAgent:Close()
		LogError("[PROVISIONING.VCCS:GetVCC] Error Retrieving VCC %s. Not Found.", vccId)
		web.status = 404
		local err = {
			vcc = vccId,
			errType = vccInfo
		}
		return err
	end
	-- 2. migration information - si se encuentra running ya sea una migracion o un rollback.
	local isRunning, currentMigration = m_migrations.allegro.IsRunning()
	vccInfo.Migration = {
		IsRunning        = isRunning,
		CurrentMigration = currentMigration
	}
	-- 3. migration allegro information: si es un vcc que tiene integracion con allegro tengo que obtener cierta informacion.
	local webParameters = utils.ParseWebQueryString(web.vars.QUERY_STRING)
	--LogDebug("[PROVISIONING.VCCS:GetVCC] ---------> webParameters %s.", utils.JSON_Encode(webParameters))
	if webParameters and tostring(webParameters.with_migration_data) == "1" then
		local success, allegroInfo = m_migrations.allegro.GetAllegroInfo(vccId, ksAgent)
		if success then
			vccInfo.Migration.AllegroInfo = allegroInfo
		end
	end
	-- ok
	ksAgent:Close()
	web.status = 200
	return vccInfo
end

-- Se obtienen los rangos disponibles que se pueden configurar como limites de configuracions del VCC, teniendo encuenta los limites maximos del ambiente
local function GetMaxNumberConfigurationsAvailableByVcc (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("[PROVISIONING.VCCS:GetMaxNumberConfigurationsAvailableByVcc] User %s.", userId)

	local success, configLimits = modelVCCs.GetMaxNumberConfigurationsAvailableByVcc()
	if not success then
		LogError("[PROVISIONING.VCCS:GetMaxNumberConfigurationsAvailableByVcc] Error retrieving info. error %s.", configLimits)
		web.status = 404
		local err = {
			errType = configLimits
		}
		return err
	end
	web.status = 200
	LogInfo("[PROVISIONING.VCCS:GetMaxNumberConfigurationsAvailableByVcc] configLimits %s.", json.encode(configLimits))
	return {result = "OK", configLimits = configLimits}
end

local function LoadAllVCCs (keyspace)
	local releaseConnection = false
	if not keyspace then
		releaseConnection = true
		keyspace = utils.GetDatasource():CassandraConnection("Agent")
	end
	local success, vccs = modelVCCs.ListAll(keyspace)
	if releaseConnection then
		keyspace:Close()
	end
	if not success then
		-- error
		local err = vccs
		LogWarning("[PROVISIONING.VCCS:LoadAllVCCs] Error loading vccs. Err'%s'.", err)
		return false, err
	end
	-- ok
	return true, vccs
end

local function SaveVCC (vccInfo)
	assert(vccInfo)
	--LogInfo("[PROVISIONING.VCCS:SaveVCC] Updating VCC '%s'.", json.encode(vccInfo))
	local vccId = vccInfo.Id
	local success, errMsg = modelVCCs.Save(vccId, vccInfo)
	if success then
		LogInfo("[PROVISIONING.VCCS:SaveVCC] '%s' updated successfully.", vccId)
	end
	return success, errMsg
end

--[[
]]--
local function __MakeTemplateNewUser (userInfo)
	assert(type(userInfo) == "table")
	assert(userInfo.userId)
	assert(userInfo.vccId)

	local userId = utils.GetId(userInfo.userId)
	local fullUserId = utils.AppendVcc(userId, userInfo.vccId)
	local userName = userId
	local plainPassword = userInfo.plainPassword or "12345678"

	local templateUser = {
		newUser         = true,
		Id              = fullUserId,
		Name            = userName,
		FirstSurname    = userName,
		Password        = plainPassword,
		Mail            = userInfo.email or nil,
		IsPlainPassword = true,
		Enabled         = "1",
		UserGroup       = "administrator",
		VccGroup        = "system",
		TimeZoneId      = "UYT",
		countryCode     = "598",
		areaCode        = "",
		notify          = false
	}

	--LogDebug("[PROVISIONING.VCCS:__MakeTemplateNewUser] template User VCC '%s'.", json.encode(templateUser))

	return templateUser
end

--[[
]]--
local function __MakeTemplateSuperUser (userInfo)
	assert(type(userInfo) == "table")
	assert(userInfo.Id)
	assert(userInfo.vccId)
	userInfo.Id = userInfo.Id:lower()
	local userId = utils.GetId(userInfo.Id)
	local fullUserId = utils.AppendVcc(userId, userInfo.vccId)
	local userName = userId
	local plainPassword = utils.alltrim(userInfo.plainPassword):len() > 0 and userInfo.plainPassword or "12345678"
	local templateSuperUser = {
		newUser             = userInfo.newUser, -- true/false
		Id                  = fullUserId,
		Name                = userInfo.Name or userName,
		FirstSurname        = userInfo.FirstSurname or userName,
		Password            = plainPassword,
		Mail                = userInfo.Mail or nil,
		IsPlainPassword     = true,
		Enabled             = "1",
		UserGroup           = "full-system-access",
		VccGroup            = "system",
		TimeZoneId          = userInfo.TimeZoneId or "UYT",
		countryCode         = userInfo.countryCode or "598",
		areaCode            = userInfo.areaCode or "",
		LanguageId          = userInfo.LanguageId or "es",
		notify              = false
	}
	LogDebug("[PROVISIONING.VCCS:__MakeTemplateSuperUser] template Sueper User: '%s'.", json.encode(templateSuperUser))
	return templateSuperUser
end

-- Se utiliza para resetear la pwd del usuario administrador del vcc
local function __MakeTemplateUpdatedUser (userInfo)
	assert(type(userInfo) == "table")
	assert(userInfo.userId)
	assert(userInfo.vccId)
	assert(userInfo.jwtToken)
	assert(userInfo.Password)

	local userId = utils.GetId(userInfo.userId)
	local fullUserId = utils.AppendVcc(userId, userInfo.vccId)

	local templateUser = {
		newUser         = false,
		Id              = fullUserId,
		Password        = userInfo.Password,
		IsPlainPassword = false,
		Status          = "1",
		notify          = false,
		jwtToken        = userInfo.jwtToken
	}

	--LogDebug("[PROVISIONING.VCCS:__MakeTemplateUpdatedUser] template User VCC '%s'.", json.encode(templateUser))

	return templateUser
end

local function __MakeTemplateNewApiUsers (apiUserInfo)
	assert(type(apiUserInfo) == "table")
	assert(type(apiUserInfo.Vcc) == "string")
	assert(apiUserInfo.Vcc:lower() ~= "system", "unable to create Api User for System Vcc.")

	local success, apiSDKSystemConfig = modelVCCs.GetSystemConfiguration("API_SDK")
	if not success then
		local errType = "API SDK - SYSTEM CONFIGURATION DOES NOT EXIST"
		LogError("[VCCS:__MakeTemplateNewApiUsers] API SDK Validation error: '%s'", errType)
		return false, errType
	end
	local encodedSystemUsers = apiSDKSystemConfig.SystemUsers
	--LogDebug("-----> apiSDKSystemConfig.SystemUsers: %s", encodedSystemUsers)
	if type(encodedSystemUsers) ~= "string" then
		local errType = "API SDK - SYSTEM USERS CONFIGURATION DOES NOT EXIST"
		LogError("[VCCS:__MakeTemplateNewApiUsers] API SDK Validation error: '%s'", errType)
		return false, errType
	end

	local systemUsers = json.decode(encodedSystemUsers)
	local apiUsers = {}
	for _, systemUserApi in pairs(systemUsers) do
		local apiUserId = systemUserApi.Id:lower()
		local templateApiUser = {
			isNew          = apiUserInfo.isNew,
			Id             = apiUserId,
			Password       = systemUserApi.Password,
			IsSystem       = 1,
			Type           = "full",
			ExpirationTime = 0,
			ExpirationUnit = "",
			Vcc            = apiUserInfo.Vcc:lower()
		}
		apiUsers[apiUserId] = templateApiUser
	end

	--LogDebug("[PROVISIONING.VCCS:__MakeTemplateNewApiUsers] --> Template API Users '%s'.", json.encode(apiUsers))

	return true, apiUsers
end

--[[
	Sample:
]]--
local function __LoadVCCSessionTimeouts (vccInfo)
	assert(type(vccInfo) == "table")
	local sessionTimeouts = modelVCCs.LoadVCCSessionTimeouts(vccInfo)
	local encodedSessionTimeouts
	if sessionTimeouts then
		encodedSessionTimeouts = json.encode(sessionTimeouts)
	end
	--LogDebug("[PROVISIONING.VCCS:__LoadVCCSessionTimeouts] Loaded VCC '%s' sessionTimeouts: '%s'.", vccInfo.Id, encodedSessionTimeouts)
	return sessionTimeouts, encodedSessionTimeouts
end

--[[
	Sample:
		{"captcha":{"enabled":"0","secretKey":"6Lf-J0UUAAAAAFFGy0pjfQgF0tZRtaGYOc2NuWdw","publicKey":"6Lf-J0UUAAAAAGmR_IKYvcsFhjsu9v_AXeTnsg87"}}
]]--
local function __LoadVCCSecurity (vccInfo)
	assert(type(vccInfo) == "table")
	local security = modelVCCs.LoadVCCSecurity(vccInfo)
	local encodedSecurity
	if security then
		encodedSecurity = json.encode(security)
	end
	--LogDebug("[PROVISIONING.VCCS:__LoadVCCSecurity] Loaded VCC '%s' security: '%s'.", vccInfo.Id, encodedSecurity)
	return security, encodedSecurity
end

--[[
	Sample:
		{Enabled = "1"}
]]--
local function __LoadVCCInteractionNotifications (vccInfo)
	assert(type(vccInfo) == "table")
	local isEnabledNotifications = modelVCCs.LoadVCCInteractionNotifications(vccInfo)
	--LogDebug("[PROVISIONING.VCCS:__LoadVCCInteractionNotifications] Loaded VCC '%s' interaction notifications enabled: '%s'.", vccInfo.Id, isEnabledNotifications)
	return isEnabledNotifications
end

--[[
	Sample:
		{"HelpDesk":{"Enabled":"1","UserAuthentication":{"Password":"Inc0nc3r710","User":"admin","Scope":"FullControl","GrantType":"password","ClientId":"51507d0c6a714f54bb23375c5eb4cabc"},"Notifications":{"Url":"https://168.62.161.232:2443/ichelpdeskdesarrollo/notifications"},"Licenses":{"Users":"1"},"Server":{"HOST":"https://168.62.161.232:2443/ichelpdeskdesarrollo"}},"MarketingAndSales":{"Notifications":{"Url":"https://webapp01.inconcertcc.com/public/occ/process"},"AppRender":{"Url":"https://webapp01.inconcertcc.com/form/render"},"AccessToken":{"Token":"eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VySW5zdGFuY2UiOiJkZWZhdWx0IiwidXNlcklkIjoiX19BdXhJbnRlZ3JhdGlvblVzZXIiLCJpYXQiOjE1ODgxNzg4NDF9.NaPbner0OXLcgc4Mt4iV_x5ZVrPmC2q3t6OpNRyRZb4","Url":"https://webapp01.inconcertcc.com/public/occ/get_session_token"},"Enabled":"1"},"WorkForceManagement":{"Enabled":"1","Notifications":{"Url":"https://wfm_site.com/notifications"}}}
]]--
local function __LoadVCCAppsIntegrations (vccInfo, enabledNotifications, widgetsDefinitions)
	assert(type(vccInfo) == "table")
	modelVCCs.LoadVCCAppsIntegrations(vccInfo, enabledNotifications, widgetsDefinitions)
	--LogDebug("[PROVISIONING.VCCS:__LoadVCCAppsIntegrations] Loaded VCC '%s' Apps Integrations.", vccInfo.Id)
end

local function __ValidateAndMakeInfoToUpdate (vccId, vccInfo, authToken)
	assert(vccId)
	assert(vccInfo)

	LogInfo("[PROVISIONING.VCCS:__ValidateAndMakeInfoToUpdate] Validating VCC '%s' info to updated.", vccId)

	local vccExist, currentVccInfo = modelVCCs.Get(vccId)

	-- 1 - clean up valores que llegan del lado del cliente
	local isNew = vccInfo.isNew
	
	if isNew then
		if vccId:lower() == "systemlicense" then
			LogError("[PROVISIONING.VCCS:__ValidateAndMakeInfoToUpdate] - The Contact Center Name is invalid - vccId:'%s'", vccId)
			return false, "INVALID_VCC_NAME"
		end			
		if #vccId < 3 or #vccId > 20 then
			LogError("[PROVISIONING.VCCS:__ValidateAndMakeInfoToUpdate] - The Contact Center Name length must be between 3 and 20 - vccId:'%s'", vccId)
			return false, "INVALID_VCC_NAME_LEN"
		end
		if string.find(vccId, "[^%a%d_]") then
			LogError("[PROVISIONING.VCCS:__ValidateAndMakeInfoToUpdate] - Only letters, numbers and dashes are allowed in vccId:'%s'", vccId)
			return false, "INVALID_CHARS_ON_VCC_NAME"
		end
		if tonumber(vccId) and true or nil then
			LogError("[PROVISIONING.VCCS:__ValidateAndMakeInfoToUpdate] - The contact center name cannot be all numbers. in vccId:'%s'", vccId)
			return false, "INVALID_ALL_NUMBER_ON_VCC_NAME"
		end
	end

	local temporalUserInfo = vccInfo.__userInfo
	vccInfo.__userInfo = nil

	local hasAllegroIntegration = false
	local function HasAllegroVcc (info)
		if info and info.AllegroIntegration then
			local allegroIntegration = info.AllegroIntegration
			if type(allegroIntegration) == "string" then
				allegroIntegration = json.decode(allegroIntegration)
			end
			if allegroIntegration and allegroIntegration.enabled then
				hasAllegroIntegration = true
			end
		end
	end

	if vccExist then
		-- Si estoy creando el VCC y el VCC ya exsite en la BD, fallo!!
		if isNew then
			return false, "MUST_HAVE_UNIQUE_NAME"
		end
		-- Funcion que valida si el atributo que NO llega en el POST pero SI existia en la BD se debe de incluir.
		local function MustIncludeNil (name)
			-- Este es el caso en el cual NO configuran nigun valor en los limites, en la BD NO se tiene que almacenar nada.
			local notIncludeAttributes = {OutgoingBatchesLimit = true, SocialNetworkPostsLimit = true}
			if notIncludeAttributes[name] then
				return false
			end
			return true
		end
		-- 2 - Si estoy actualizando un VCC existente, los valores que NO vengan en el POST y ya estaban en la BD deben almacenarse.
		for name, value in pairs(currentVccInfo) do
			if vccInfo[name] == nil then
				if MustIncludeNil(name) then
					vccInfo[name] = value
				end
			end
		end

		HasAllegroVcc(currentVccInfo)

		-- Si es un VCC con Allegro los siguientes valores NO se modifican desde la interfaz de i6, sino que se modifican desde la interfaz de Allegro
		if hasAllegroIntegration then
			vccInfo.CALL_licensed = currentVccInfo.CALL_licensed and "1" or "0"
		end
	else
		HasAllegroVcc(vccInfo)
		-- Cuando estoy creando un VCC, el BOT por defecto es LUIS y el mismo se encuentra disabled.
		vccInfo.BotInfo = {Type = "LUIS", Enabled = false}
		-- Cuando estoy creando un VCC con allegro, la info de Security no viene en el request. Le pongo la de por defecto
		if hasAllegroIntegration then
			vccInfo.Security = {
				captcha = {
					enabled = "0"
				},
				ldap = {
					enabled = 0
				},
				mfa = {
					enabled = 0
				},
				authentication = {
					passwordChecks = {
						minLowerChars = 1,
						minNumericChars = 1,
						expirationNotificationDays = 10,
						history = 10,
						minUpperChars = 1,
						blockOnLoginFailures = 5,
						minLen = 8,
						maxLen = 15,
						minSpecialChars = 1,
						expirationDays = 45
					},
					type = "BASIC"
				}
			}
		end
	end

	-- Funcion que valida los limites de Post de la App Social NetWorks y de Outgoing Batches
	local function ValidateChannelsLimits (channelsConfig, columnName)
		if not vccInfo[columnName] then
			return true
		end
		local values = {}
		for channel, value in pairs(vccInfo[columnName]) do
			channel = channel:upper()
			local channelConfig = channelsConfig[channel]
			if not channelConfig then
				return false, "CHANNEL_NOT_ALLOWED"
			end
			value = tonumber(value)
			if not value then
				return false, "CHANNEL_INVALID_LIMIT"
			end
			if value < channelConfig.minLimit or value > channelConfig.maxLimit then
				return false, "CHANNEL_OVERFLOW"
			end
			values[channel] = value
		end
		vccInfo[columnName] = values
		return true
	end

	-- Outgoing Batches Limit
	local channelsConfig = {
		MAIL     = {minLimit = 2, maxLimit = 300, default = 50},
		SMS      = {minLimit = 2, maxLimit = 300, default = 50},
		WHATSAPP = {minLimit = 2, maxLimit = 300, default = 50}
	}
	local success, err = ValidateChannelsLimits(channelsConfig, "OutgoingBatchesLimit")
	if not success then
		return false, "OUTGOING_BATCHES_" .. tostring(err)
	end
	-- Social Networks Post Limit
	local channelsConfig = {
		FACEBOOK  = {minLimit = 2, maxLimit = 300, default = 50},
		INSTAGRAM = {minLimit = 2, maxLimit = 300, default = 50},
		TWITTER   = {minLimit = 2, maxLimit = 300, default = 50},
		YOUTUBE   = {minLimit = 2, maxLimit = 300, default = 50}
	}
	local success, err = ValidateChannelsLimits(channelsConfig, "SocialNetworkPostsLimit")
	if not success then
		return false, "SOCIAL_NETWORKS_" .. tostring(err)
	end

	-- Si estoy creando el VCC o el VCC no tiene usuarios tengo que crear el primer Usuario Administrator del VCC.
	local firstAdminUser, firstApiUsers, firstSuperUser

	if type(temporalUserInfo) == "table" then
		local userId = utils.alltrim(temporalUserInfo.userId or "")
		if userId:len() == 0 then
			return false, "USER IS REQUIRED"
		end

		if not authToken then
			LogError("[VCCS:__ValidateAndMakeInfoToUpdate] invalid authToken: '%s'", authToken)
			return false, "INVALID AUTHTOKEN"
		end
	
		local jwtData, jwtErr = jwt.decode(temporalUserInfo.jwtToken, md5.sumhexa(authToken), "HS256")
		if jwtErr then
			LogError("[VCCS:__ValidateAndMakeInfoToUpdate] Error decoding jwtToken: '%s'", jwtErr)
			return false, "JWT TOKEN INVALID"
		end
		local plainPassword = jwtData.password

		local userInfo = {
			userId = userId,
			vccId = vccId,
			plainPassword = plainPassword
		}

		local helpDeskIntegration = vccInfo.AppsIntegrations and vccInfo.AppsIntegrations.HelpDesk
		if helpDeskIntegration and tonumber(helpDeskIntegration.Enabled) == 1 then
			local email = utils.alltrim(temporalUserInfo.email or "")
			if email:len() == 0 then
				return false, "EMAIL IS REQUIRED FOR VCC WITH HELPDESK INTEGRATION"
			end
			userInfo.email = email
		end
		-- #12289: En un vcc existente se permite ahora resetear la pwd del usuario admin, por lo que tengo temporalUserInfo pero en este caso no creo el user, sino solo cambio su pwd 
		if isNew then
			-- El primer Usuario Administrador que se debe de crear cuando se crea el VCC.
			firstAdminUser = __MakeTemplateNewUser(userInfo)
			firstSuperUser = __MakeTemplateSuperUser({
				Id    = "superuser",
				vccId = vccId
			})
			-- El primer API Usuario que se debe de crear cuando se crea el VCC.
			local success
			success, firstApiUsers = __MakeTemplateNewApiUsers({Vcc = vccId, isNew = isNew})
			if not success then
				local errType = firstApiUsers
				return false, errType
			end
		else
			-- password reset
			temporalUserInfo.vccId = vccId
			firstAdminUser = __MakeTemplateUpdatedUser(temporalUserInfo)
		end
	end

	--El password para conectar con el server externo de autenticacion viene con jwtoken
	local authType = vccInfo.Security.authentication.type
	if authType ~= "BASIC" and authType ~= "MFA" then
		authType = authType:lower()
		if vccInfo.Security[authType].adminPassword then
			local jwtData, jwtErr = jwt.decode(vccInfo.Security[authType].adminPassword, md5.sumhexa(authToken), "HS256")
			if jwtErr then
				LogError("[VCCS:__ValidateAndMakeInfoToUpdate] Error decoding jwtToken: '%s'", jwtErr)
				return false, "JWT TOKEN INVALID"
			end
			local fakePassword = modelVCCs.GetExternalAuthFakePassword() 
			if jwtData.password == fakePassword then 
				--El password no ha cambiado, recuperamos el que tenemos en BD
				vccInfo.Security[authType].adminPassword = modelVCCs.GetCurrentExternalAuthPassword(vccId)
			else
				--El password ha cambiado
				vccInfo.Security[authType].adminPassword = jwtData.password
			end
		else
			LogError("[VCCS:__ValidateAndMakeInfoToUpdate] adminPassword for external authentication '%s' is required", vccInfo.Security.authentication.type)
			return false, "EXTERNAL AUTHENTICATION PASSWORD INVALID"
		end
	end

	if authType:lower() == "openid" then
		local openid_configuration_url = vccInfo.Security.openid.openid_configuration_url
		local openIdConfigSuccess, openIdConfig = modelVCCs.AzureAD().GetOpenIdConfiguration(openid_configuration_url)
		if not openIdConfigSuccess or not openIdConfig.success then
			LogError("[VCCS:__ValidateAndMakeInfoToUpdate] Get openid_configuration_url error: '%s'", openIdConfig.error)
			return false, "OPENID CONFIGURATION URL ERROR"
		end
		vccInfo.Security.openid.openid_configuration_values = openIdConfig.result
	end

	local function MakeDefaultValue (currentValue, defaultValue, allowEmptyValues)
		--LogDebug("MakeDefaultValue '%s' - '%s' - '%s'", currentValue, defaultValue, noAllowEmptyValues)
		if not currentValue then
		--	LogDebug("a")
			return defaultValue
		end
		if utils.alltrim(currentValue) == "" then
		--	LogDebug("b")
			if not allowEmptyValues then
		--		LogDebug("c")
				return defaultValue
			end
		end
		--LogDebug("d")
		return currentValue
	end

	local function CheckExternalEntities (vccSecurity)
		local authenticationType = vccSecurity.authentication.type
		if authenticationType == "BASIC" or authenticationType == "MFA" then
			return true 	--Internal security
		end 

		local externalAuthSystemName = vccSecurity.authentication.type:lower()
		local externalAuthSystem = vccSecurity[externalAuthSystemName]
		if not externalAuthSystem then
			return true --No deberia ocurrir, pero por si las flies
		end

		if not externalAuthSystem.sync_enabled then
			return true --No hay grupos que sincronizar
		end

		local groupFilterValues = {}
		for groupName, groupFilter in pairs(externalAuthSystem.groupFilters) do
			if groupFilterValues[groupFilter] then
				return false, string.format("GROUP FILTER '%s' DUPLICATED", groupFilter)
			end
			groupFilterValues[groupFilter] = true
		end

		return true
	end

	--LogDebug("1-----> vccInfo.ApiConfiguration: %s", json.encode(vccInfo.ApiConfiguration or {}))
	if vccInfo.ApiConfiguration and vccInfo.ApiConfiguration.enabled and vccInfo.ApiConfiguration.urlsConfig then
		local success, apiSDKSystemConfig = modelVCCs.GetSystemConfiguration("API_SDK")
		if not success then
			local errType = "API SDK - SYSTEM CONFIGURATION DOES NOT EXIST"
			LogError("[VCCS:__ValidateAndMakeInfoToUpdate] API SDK Validation error: '%s'", errType)
			return false, errType
		end
		--LogDebug("2-----> apiSDKSystemConfig: %s", json.encode(apiSDKSystemConfig or {}))
		--LogDebug("3-----> apiSDKSystemConfig.URLsDefaultConfig: %s", apiSDKSystemConfig.URLsDefaultConfig)

		--
		local success, allVccs = LoadAllVCCs()
		if not success then
			local errType = "API SDK - ERROR LOADING VCCS"
			LogError("[VCCS:__ValidateAndMakeInfoToUpdate] API SDK err. Error loading VCCs. Err '%s'", errType)
			return false, errType
		end

		local vccResouresMax = {}
		local allVccsMax = {}
		local emptyRateLimits = {maxRequests = 0}
		-- system
		local systemResources = json.decode(apiSDKSystemConfig.URLsDefaultConfig).resources
		for systemResourceName, systemResource in pairs(systemResources) do
			--LogDebug("--- systemResourceName: %s ---------", systemResourceName)
			local systemRateLimits = systemResource.rateLimits
			local systemMax = tonumber(systemRateLimits.maxRequests)
			--LogDebug("-----> systemMax: %s", systemMax)
			--LogDebug("-----> systemLowMax: %s", systemLowMax)
			-- vcc
			local vccResources = vccInfo.ApiConfiguration.urlsConfig.resources
			--LogDebug("4-----> vccResources: %s", json.encode(vccResources or {}))
			local vccRateLimits = vccResources[systemResourceName] and vccResources[systemResourceName].rateLimits or emptyRateLimits
			vccResouresMax[systemResourceName] = tonumber(vccRateLimits.maxRequests)
			--LogDebug("-----> vccResouresMax[%s]: %s", systemResourceName, vccResouresMax[systemResourceName])
			if vccResouresMax[systemResourceName] > systemMax then
				local errType = "API SDK - RATE LIMIT HAS EXCEEDED WHAT IS ALLOWED IN THE SYSTEM"
				LogError("[VCCS:__ValidateAndMakeInfoToUpdate] API SDK err. The rate limit of the %s resource in the specified time exceed what is allowed in the system. Err '%s'", systemResourceName, errType)
				return false, errType, {vccRateLimits = vccRateLimits, systemRateLimits = systemRateLimits, resourceType = systemResourceName}
			end
			--
			allVccsMax[systemResourceName] = vccResouresMax[systemResourceName]
			for existingVccId, existingVccInfo in pairs(allVccs) do
				--LogDebug("[VCCS:__ValidateAndMakeInfoToUpdate] ---->  existingVccId '%s' - existingVccInfo: %s.", existingVccId, json.encode(existingVccInfo))
				if existingVccId:lower() ~= vccId:lower() then
					if existingVccInfo.ApiConfiguration then
						local apiConfiguration = json.decode(existingVccInfo.ApiConfiguration)
						if apiConfiguration.enabled and apiConfiguration.urlsConfig then
							-- vcc
							local existingVccResources = apiConfiguration.urlsConfig.resources
							local existingVccRateLimits = existingVccResources[systemResourceName] and existingVccResources[systemResourceName].rateLimits or emptyRateLimits
							local existingVccMax = tonumber(existingVccRateLimits.maxRequests)
							allVccsMax[systemResourceName] = allVccsMax[systemResourceName] + existingVccMax
						end
					end
				end
			end
			--LogDebug("-----> allVccsMax[%s]: %s", systemResourceName, allVccsMax[systemResourceName])
			if allVccsMax[systemResourceName] > systemMax then
				local available = systemMax - (allVccsMax[systemResourceName] - vccResouresMax[systemResourceName])
				--LogDebug("-----> allVccsMax[%s]: %s", systemResourceName, allVccsMax[systemResourceName])
				local errType = "API SDK - RATE LIMIT OF ALL VCSS HAS EXCEEDED WHAT IS ALLOWED IN THE SYSTEM"
				LogError("[VCCS:__ValidateAndMakeInfoToUpdate] API SDK err. The rate limit of all vcss of %s resource in the specified time exceed what is allowed in the system. Err '%s'", systemResourceName, errType)
				return false, errType, {available = available, systemRateLimits = systemRateLimits, resourceType = systemResourceName}
			end
		end
		---
		-- OK
	end

	--[[
	----------------------------------------------------------------------------------------------------------------
		Ref #11602: Se agrega esta función para mantener el modelo de licencia antiguo por compatibilidad con Debian
	----------------------------------------------------------------------------------------------------------------
	]]--
	local function SetLicensesDefaultValues ()
		-- Las licencias de los diferentes Canales
		local licenses = {
			APPSTORE   = 0,
			CHAT       = 0,
			FACEBOOK   = 0,
			GOOGLEPLAY = 0,
			INSTAGRAM  = 0,
			MAIL       = 0,
			RCS        = 0,
			SMS        = 0,
			TWITTER    = 0,
			WEBCONTACT = 0,
			WHATSAPP   = 0,
			YOUTUBE    = 0
		}
		vccInfo.IsSystemVCC                 = "0"
		if not hasAllegroIntegration then
			-- El default de CALL para VCCs con integración viene dado desde "allegro/templates.lua"
			licenses.CALL = 0
		end
		for channel, licenseValue in pairs(licenses) do
			local name = ("%s_licensed"):format(channel)
			vccInfo[name] = tostring(vccExist and currentVccInfo[name] or licenseValue)
		end

		-- Los Default salen de la rocka provisioningmanager
		local licencesDefaultValues = modelVCCs.Licenses().LoadDefaultLicensesValues()
		local function GetDefaultLicense (licenseType)
			licenseType = (licenseType or ""):upper()
			return licencesDefaultValues and licencesDefaultValues[licenseType] or ""
		end

		-- Los Usuarios en línea (solo para VCCs sin integracion con Allegro)
		-- Los VCCs con integracion con Allegro, se actualizan desde la interfaz de Allegro.
		if not hasAllegroIntegration then
			vccInfo.OnlineUsers = tostring(vccExist and currentVccInfo.OnlineUsers or GetDefaultLicense("ONLINE_USERS"))
		end

		-- Metrics, Filters, Quotas
		vccInfo.CustomMetricsLimit    = tostring(vccExist and currentVccInfo.CustomMetricsLimit or GetDefaultLicense("CUSTOM_METRICS_LIMIT"))
		vccInfo.CustomFiltersLimit    = tostring(vccExist and currentVccInfo.CustomFiltersLimit or GetDefaultLicense("CUSTOM_FILTERS_LIMIT"))
		vccInfo.ExportSpaceQuota      = tostring(vccExist and currentVccInfo.ExportSpaceQuota or GetDefaultLicense("EXPORT_SPACE_QUOTA"))
		vccInfo.SharedFilesSpaceQuota = tostring(vccExist and currentVccInfo.SharedFilesSpaceQuota or GetDefaultLicense("SHARED_FILES_SPACE_QUOTA"))

		-- HelpDesk (Usuarios en línea)
		local helpDeskIntegration = vccInfo.AppsIntegrations and vccInfo.AppsIntegrations.HelpDesk
		if helpDeskIntegration and tonumber(helpDeskIntegration.Enabled) == 1 then
			local currentValue
			if vccExist then
				local currentHelpDeskIntegration = currentVccInfo.AppsIntegrations and json.decode(currentVccInfo.AppsIntegrations).HelpDesk
				--LogDebug("------> currentHelpDeskIntegration: %s", json.encode(currentHelpDeskIntegration))
				if currentHelpDeskIntegration and tonumber(currentHelpDeskIntegration.Enabled) == 1 then
					currentValue = currentHelpDeskIntegration.Licenses and currentHelpDeskIntegration.Licenses.Users or GetDefaultLicense("HELPDESK_USERS_LICENSES")
				end
			end
			vccInfo.AppsIntegrations.HelpDesk.Licenses = vccInfo.AppsIntegrations.HelpDesk.Licenses or {}
			vccInfo.AppsIntegrations.HelpDesk.Licenses.Users = tostring(currentValue or "0")
		end
	end
	----------------------------------------------------------------------------------------------------------------

	if m_helpers.utils.HasLicenseSystem() then
		-- 3 - Se le agrega valores por default
		local GetChannelValueString = function (value)
			if type(value) == "boolean" then
				return (value and "1" or "0")
			end
			return value
		end

		--Estos datos cuando es un vcc de allegro vienen numericos. Hay que pasarlos a string
		vccInfo.CustomMetricsLimit = tostring(vccInfo.CustomMetricsLimit)
		vccInfo.CustomFiltersLimit = tostring(vccInfo.CustomFiltersLimit)
		vccInfo.ExportSpaceQuota = tostring(vccInfo.ExportSpaceQuota)
		vccInfo.SharedFilesSpaceQuota = tostring(vccInfo.SharedFilesSpaceQuota)

		--Estos datos cuando es un vcc de allegro y se edita desde allegro vienen boolean. Hay que pasarlos a string
		vccInfo.APPSTORE_licensed = GetChannelValueString(vccInfo.APPSTORE_licensed)
		vccInfo.CHAT_licensed = GetChannelValueString(vccInfo.CHAT_licensed)
		vccInfo.FACEBOOK_licensed = GetChannelValueString(vccInfo.FACEBOOK_licensed)
		vccInfo.GOOGLEPLAY_licensed = GetChannelValueString(vccInfo.GOOGLEPLAY_licensed)
		vccInfo.INSTAGRAM_licensed = GetChannelValueString(vccInfo.INSTAGRAM_licensed)
		vccInfo.MAIL_licensed = GetChannelValueString(vccInfo.MAIL_licensed)
		vccInfo.SMS_licensed = GetChannelValueString(vccInfo.SMS_licensed)
		vccInfo.TWITTER_licensed = GetChannelValueString(vccInfo.TWITTER_licensed)
		vccInfo.WEBCONTACT_licensed = GetChannelValueString(vccInfo.WEBCONTACT_licensed)
		vccInfo.WHATSAPP_licensed = GetChannelValueString(vccInfo.WHATSAPP_licensed)
		vccInfo.YOUTUBE_licensed = GetChannelValueString(vccInfo.YOUTUBE_licensed)

		vccInfo.Id                          = vccId
		vccInfo.IsSystemVCC                 = "0"
		vccInfo.Description                 = MakeDefaultValue(vccInfo.Description                , "" , true )
		vccInfo.Company                     = MakeDefaultValue(vccInfo.Company                    , "" , true )
		vccInfo.CompanyAddress              = MakeDefaultValue(vccInfo.CompanyAddress             , "" , true )
		vccInfo.Telephone1                  = MakeDefaultValue(vccInfo.Telephone1                 , "" , true )
		vccInfo.Telephone2                  = MakeDefaultValue(vccInfo.Telephone2                 , "" , true )
		vccInfo.ContactName                 = MakeDefaultValue(vccInfo.ContactName                , "" , true )
		vccInfo.ContactEmail                = MakeDefaultValue(vccInfo.ContactEmail               , "" , true )
		vccInfo.FreeSecondsAfterInteraction = MakeDefaultValue(vccInfo.FreeSecondsAfterInteraction, "0", false)

		-- licencias
		vccInfo.OnlineUsers                 = MakeDefaultValue(vccInfo.OnlineUsers                , "0", false)
		vccInfo.CustomMetricsLimit          = MakeDefaultValue(vccInfo.CustomMetricsLimit         , "0", false)
		vccInfo.CustomFiltersLimit          = MakeDefaultValue(vccInfo.CustomFiltersLimit         , "0", false)
		vccInfo.ExportSpaceQuota            = MakeDefaultValue(vccInfo.ExportSpaceQuota           , "0", false)
		vccInfo.SharedFilesSpaceQuota       = MakeDefaultValue(vccInfo.SharedFilesSpaceQuota      , "0", false)
		
		-- #11602 - Default de channels a 0, porque ya indicara el gestor de licencia si puede o no licenciarse el channel en este vcc
		vccInfo.APPSTORE_licensed           = MakeDefaultValue(vccInfo.APPSTORE_licensed          , "0", false)
		vccInfo.CALL_licensed               = MakeDefaultValue(vccInfo.CALL_licensed              , "0", false)
		vccInfo.CHAT_licensed               = MakeDefaultValue(vccInfo.CHAT_licensed              , "0", false)
		vccInfo.FACEBOOK_licensed           = MakeDefaultValue(vccInfo.FACEBOOK_licensed          , "0", false)
		vccInfo.GOOGLEPLAY_licensed         = MakeDefaultValue(vccInfo.GOOGLEPLAY_licensed        , "0", false)
		vccInfo.INSTAGRAM_licensed          = MakeDefaultValue(vccInfo.INSTAGRAM_licensed         , "0", false)
		vccInfo.MAIL_licensed               = MakeDefaultValue(vccInfo.MAIL_licensed              , "0", false)
		vccInfo.RCS_licensed                = MakeDefaultValue(vccInfo.RCS_licensed               , "0", false)
		vccInfo.SMS_licensed                = MakeDefaultValue(vccInfo.SMS_licensed               , "0", false)
		vccInfo.TWITTER_licensed            = MakeDefaultValue(vccInfo.TWITTER_licensed           , "0", false)
		vccInfo.WEBCONTACT_licensed         = MakeDefaultValue(vccInfo.WEBCONTACT_licensed        , "0", false)
		vccInfo.WHATSAPP_licensed           = MakeDefaultValue(vccInfo.WHATSAPP_licensed          , "0", false)
		vccInfo.YOUTUBE_licensed            = MakeDefaultValue(vccInfo.YOUTUBE_licensed           , "0", false)
	else
		SetLicensesDefaultValues() -- Ref #11602: Se invoca esta función para mantener el modelo de licencia antiguo por compatibilidad con Debian
	end
	
	-- VCC Session Timeouts
	local sessionTimeouts, encodedSessionTimeouts = __LoadVCCSessionTimeouts(vccInfo)
	vccInfo.SessionTimeouts = encodedSessionTimeouts

	-- VCC Security
	local security, encodedSecurity = __LoadVCCSecurity(vccInfo)
	vccInfo.Security = encodedSecurity
	local success, err = CheckExternalEntities(security)
	if not success then
		return false, err
	end

	-- Interactions Notifications
	local isEnabledNotifications = __LoadVCCInteractionNotifications(vccInfo)

	-- Widgets Integrations
	local widgetsIntegrations = {
		{Name = "sample" , Description = "Sample" , Config = {}},
		{Name = "example", Description = "Example"}
	}
	local widgetsDefinitions = m_integrations.widgets:LoadWidgetsDefinitions()

	-- Apps Integrations
	-- LogDebug("[PROVISIONING.VCCS:__ValidateAndMakeInfoToUpdate] 1 - Antes de '__LoadVCCAppsIntegrations' VCC %s\n vccInfo '%s'.", vccId, json.encode(vccInfo))
	__LoadVCCAppsIntegrations(vccInfo, isEnabledNotifications, widgetsDefinitions)

	-- CleanUp Values
	vccInfo.hasUsers = nil

	--LogDebug("[PROVISIONING.VCCS:__ValidateAndMakeInfoToUpdate] Maked VCC %s\n vccInfo '%s'\n adminUser: '%s'\n apiUser: '%s'.", vccId, json.encode(vccInfo), json.encode(firstAdminUser or {}), json.encode(firstApiUsers or {}))
	return true, vccInfo, firstAdminUser, firstApiUsers, firstSuperUser
end

local function auditFieldsByEntity(entity)

	--[[
		key = true|false| { alias = "key_alias"}
	
	]]--

	local entitiesFields = {
		Security = {
			captcha = true,
			mfa = true,
			ldap = true,
			authentication = true
		}
	}
	return entitiesFields[entity]
end

local function FilterAuditFields (entityData, entity)
	
	local auditFields = auditFieldsByEntity(entity)

	if not auditFields then
		return nil
	end

	local result = {}
	for key, value in pairs(entityData) do
		local auditField = auditFields[key]
		if auditField and type(auditField) == "table" then
			result[auditField.alias or key] = value
		elseif auditField then
			result[key] = value
		end
	end

	return result
end

local function getVccsAuditDataByEntity(entity, vccData, vcc)
	local entities = {
		Security   =  function(vccInfo, vccId) 
			if not vccInfo then
				local success = nil
				success, vccInfo = modelVCCs.Get(vccId)
				if not success or not vccInfo then
					return nil
				end
			end
			local securityData = modelVCCs.LoadVCCSecurity(vccInfo) or {}
			local filterData = FilterAuditFields(securityData, "Security")
			
			if filterData and filterData.captcha then
				filterData.captcha["secretKey"] = nil
				if filterData.captcha.type or filterData.captcha.enabled == "0" then
					filterData.captcha["type"] = nil
					filterData.captcha["publicKey"] = nil
				end
			end

			return filterData
		end
	}

	if entity then
		return entities[entity] and entities[entity](nil, vcc)
	end
end

local function UpdateVCC (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local authToken = web.vars and web.vars.auth and web.vars.auth.authToken

	LogInfo("[PROVISIONING.VCCS:UpdateVCC] Updating VCC....")
	LogDebug("[PROVISIONING.VCCS:UpdateVCC] Updating VCC. Posted data '%s'.", postedData)

	local result = {}

	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:UpdateVCC] Error Updating VCC. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	if not authToken then
		LogError("[PROVISIONING.VCCS:UpdateVCC] Error Updating VCC. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end
	local vccId = vccInfo.Id
	local success, vccInfo, firstAdminUser, firstApiUsers, firstSuperUser = __ValidateAndMakeInfoToUpdate(vccId, vccInfo, authToken)
	if not success then
		local errMsg = vccInfo
		local errInfo = firstAdminUser
		LogError("[PROVISIONING.VCCS:UpdateVCC] Error Validating VCC '%s'. Err: %s", vccId, errMsg)
		web.status = 500
		result.result = errMsg
		result.info = errInfo
		return result
	end

	local isNew = vccInfo.isNew
	--Old security data, for create audit
	local old_security_audit_data = getVccsAuditDataByEntity("Security", nil, vccId)
	local success, err = SaveVCC(vccInfo)
	if not success then
		--LogError("[PROVISIONING.VCCS:UpdateVCC] Error Updating VCC '%s'. Err: %s", vccId, errMsg)
		web.status = 500
		local info = utils.JSON_Decode(err) -- desde el provisioningmanager llega un JSON cuando es un error.
		result.result = type(info) == "table" and info.err.error or err-- Este código de error llega desde la Rocka "provisioningmanager"
		result.info = info
		return result
	end

	-- OK
	--
	LogInfo("[PROVISIONING.VCCS:UpdateVCC] '%s' saved successfully.", vccId)

	--Agrego a redis los totales actuales de config si es nuevo.
	if isNew then
		model_config_restrictions.AddToRedisEmptyVCCConfigurationTotals(vccId)
	end
	--Create Security Audit Data
	local new_security_audit_data = getVccsAuditDataByEntity("Security", nil, vccId)
	if new_security_audit_data then
		local ksHistorical = utils.GetDatasource():CassandraConnection("Historical")
		audits.Create({
			vcc = vccId,
			entity = "Security",
			user = inconcert.authUser.userName,
			fieldKey = (new_security_audit_data.authentication and new_security_audit_data.authentication.type or ""):lower(),
			oldFields = old_security_audit_data,
			newFields = new_security_audit_data,
			connection = ksHistorical
		})
		ksHistorical:Close()
	end

	result.code = "VCC WAS COMPLETELY UPDATED"
	result.warningInfo = {}

	if firstAdminUser then
		local success, err = CreateFirstUserInVCC(firstAdminUser)
		if not success then
			result.code = "PARTIAL UPDATED VCC"
			err = err or {}
			err.subCode = "UNABLE TO CREATE USER"
			table.insert(result.warningInfo, err)
		end
	end

	if firstSuperUser then
		local success, err = m_superUsers.CreateSuperUserInVCC(firstSuperUser)
		if not success then
			result.code = "PARTIAL UPDATED VCC"
			err = err or {}
			err.subCode = "UNABLE TO CREATE SUPER USER"
			table.insert(result.warningInfo, err)
		end
	end

	if firstApiUsers then
		local success, err = CreateFirstApiUsersInVCC(firstApiUsers)
		if not success then
			result.code = "PARTIAL UPDATED VCC"
			err = err or {}
			err.subCode = "UNABLE TO CREATE API USER"
			table.insert(result.warningInfo, err)
		end
	end

	web.status = 200
	result.result = "OK"
	return result
end

function DeleteVCC (vccId)
	assert(vccId)

	local ks_agent = utils.GetDatasource():CassandraConnection( "Agent" )
	local ks_prov = utils.GetDatasource():CassandraConnection( "Provisioning" )

	local result = modelVCCs.Delete(vccId, ks_prov, ks_agent)
	if not result.success then
		LogWarning("[PROVISIONING.VCCS:DeleteVCC] Error deleting VCC '%s'. Err: %s", vccId, result.errMsg)
	else
		LogInfo("[PROVISIONING.VCCS:DeleteVCC] '%s' deleted successfully.", vccId)
	end

	ks_agent:Close()
	ks_prov:Close()
	return result
end

local function EraseVCC (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:EraseVCC] Deleting VCC. User %s posted data '%s'.", userId, postedData)

	local result = {}

	-- POR AHORA QUEDA INHABILITADO EL BORRAR VCCs
	if true then
	 	LogError("[PROVISIONING.VCCS:EraseVCC] Unable to delete VCCs. Action Not IMPLEMENTED.")
		web.status = 403
		result.result = "FORBIDDEN"
		return result
	end

	local success, vccs = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:EraseVCC] Error Deleting VCC. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local vccId = vccs[1]
	local withCustomFields = true
	local success, vccInfo = modelVCCs.Get(vccId, nil, withCustomFields)
	if not success then
		LogError("[PROVISIONING.VCCS:EraseVCC] Error Deleting VCC %s. Not Found.", vccId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end

	if vccInfo.HasAllegroIntegration then
		LogError("[PROVISIONING.VCCS:EraseVCC] Error Deleting VCC %s. Unable to delete VCC with Allegro Integration.", vccId)
		web.status = 400
		result.result = "UNABLE TO DELETE VCC WITH ALLEGRO INTEGRATION"
		return result
	end

	local response = DeleteVCC(vccId)
	if not response.success then
		LogError("[PROVISIONING.VCCS:EraseVCC] Error Deleting VCC %s. Error: .", vccId, response.errMsg)
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		return result
	end

	web.status = 200
	result.result = "OK"
	return result
end

--[[
]]--
function CreateFirstUserInVCC (userInfo, ksAgent)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local encodedUserInfo = json.encode(userInfo)

	LogInfo("[PROVISIONING.VCCS:CreateFirstUserInVCC] Create first user in vcc. User %s. UserInfo: '%s'.", userId, encodedUserInfo)

	local moduleAgentUsers = require "apps.agent.controllers.users" -- un asco cambiar, TODO: realizar request!!
	local success, errType = moduleAgentUsers.UpdateUser(userInfo, ksAgent)

	if not success then
		if type(errType) == "string" then
			errType = {
				result = "VALUE NOT ALLOWED",
				msg = errType
			}
		end
		LogWarning("[PROVISIONING.VCCS:CreateFirstUserInVCC] Error creating first user. err: '%s'.", json.encode(errType or {}))
		return false, errType
	else
		LogInfo("[PROVISIONING.VCCS:CreateFirstUserInVCC] Created first user successfully.")
	end
	return true
end

--[[
]]--
function CreateFirstApiUsersInVCC (apiUsers)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local encodedApiUsers = json.encode(apiUsers)

	LogInfo("[PROVISIONING.VCCS:CreateFirstApiUsersInVCC] Create first API users in vcc. User %s. ApiUsers: '%s'.", userId, encodedApiUsers)

	local moduleAdmApiUsers = require "apps.administration.controllers.api_users"
	local success, errType = moduleAdmApiUsers.UpdateUsers(apiUsers)

	if not success then
		if type(errType) == "string" then
			errType = {
				result = "VALUE NOT ALLOWED",
				msg = errType
			}
		end
		LogWarning("[PROVISIONING.VCCS:CreateFirstApiUsersInVCC] Error creating first API users. err: '%s'.", json.encode(errType or {}))
		return false, errType
	else
		LogInfo("[PROVISIONING.VCCS:CreateFirstApiUsersInVCC] Created first API users successfully.")
	end
	return true
end

local function UpdateVCCTtsAsr (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:UpdateVCCTtsAsr] Saving VCC TTS/ASR Configuration.")
	--LogDebug("[PROVISIONING.VCCS:UpdateVCCTtsAsr] Saving VCC TTS/ASR Configuration. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:UpdateVCCTtsAsr] Error Updating VCC TTS/ASR Configuration. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local vccId = vccInfo.vcc
	
	--Los datos de TTS/ASR los guardamos en el model de VCCTtsAsrEngines
	if (vccInfo.ttsEngine and vccInfo.ttsProfile) or (vccInfo.asrEngine and vccInfo.asrProfile) or vccInfo.ttsasrEngines then
		LogInfo("[PROVISIONING.VCCS:UpdateVCCTtsAsr] Saving VCC TTS/ASR Engines.")
		local success, ttsasrEnginesOldData = modelVCCs.GetVCCTtsAsrEngines(vccId, nil)
		if not success then
			LogError("[PROVISIONING.VCCS:UpdateVCCTtsAsr] Error Retrieving VCC %s. Not Found.", vccId)
			web.status = 404
			result.result = "NOT FOUND"
			return result
		end
		ttsasrEnginesOldData = ttsasrEnginesOldData or {}
		
		--Hay que conservar los cambios que se hayan podido hacer en voices desde administracion en defaultVoice, prosodyVolume, prosodyRate
		local updateVoice = function(Id, customizedVoice)
			for _, engine in pairs(vccInfo.ttsasrEngines) do
				if engine.Id == Id then
					for voiceNumber, voice in pairs(engine.voices) do
						if voice.id == customizedVoice.id then
							engine.voices[voiceNumber] = customizedVoice
						end
					end
				end
			end
		end
		
		--Hay que conservar las custom grammars que puedan haber configurado desde administracion
		local insertCustomGrammar = function(Id, grammar)
			for _, engine in pairs(vccInfo.ttsasrEngines) do
				if engine.Id == Id then
					table.insert(engine.grammars, grammar)
				end
			end
		end
		
		if ttsasrEnginesOldData and ttsasrEnginesOldData.ttsasrEngines then
			for _, engine in pairs(json.decode(ttsasrEnginesOldData.ttsasrEngines)) do
				for _v, voice in pairs(engine.voices) do
					if voice.defaultVoice == 1 or voice.prosodyRate ~= "default" or voice.prosodyVolume ~= default then
						updateVoice(engine.Id, voice)
					end
				end
				
				for _gr, grammar in pairs(engine.grammars) do
					if grammar.type == "custom" then
						insertCustomGrammar(engine.Id, grammar)
					end
				end
			end
		end
		
		ttsasrEnginesOldData.ttsEngine = vccInfo.ttsEngine or ""
		ttsasrEnginesOldData.ttsProfile = vccInfo.ttsProfile or ""
		ttsasrEnginesOldData.asrEngine = vccInfo.asrEngine or ""
		ttsasrEnginesOldData.asrProfile = vccInfo.asrProfile or ""
		ttsasrEnginesOldData.ttsasrEngines = vccInfo.ttsasrEngines or {}
		
		local success, errMsg = modelVCCs.SaveVCCTtsAsrEngines(vccId, ttsasrEnginesOldData)
		if not success then
			LogError("[PROVISIONING.VCCS:UpdateVCCTtsAsr] Error Updating VCC '%s'. Err: %s", vccId, errMsg)
			web.status = 500
			result.result = errMsg -- Este código de error llega desde la Rocka "provisioningmanager"
			return result
		end
	end

	LogInfo("[PROVISIONING.VCCS:UpdateVCCTtsAsr] '%s' TTS/ASR Configuration saved successfully.", vccId)

	result.code = "VCC TTS/ASR CONFIGURATION WAS COMPLETELY UPDATED"
	web.status = 200
	result.result = "OK"
	return result
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.VCCS:Search] 1 - (START) Loading VCCs. Posted Info: '%s'", encodedInfo)

	local startCurrentProcessTime = startTime
	local success, vccs = LoadAllVCCs()
	LogDebug("[PROVISIONING.VCCS:SEARCH]    2 - Loaded All VCCs. ELAPSED Time '%s' seconds.", utils.GetElapsedTimeSec(startCurrentProcessTime))
	if not success then
		local err = vccs
		LogWarning("[PROVISIONING.VCCS:Search] Error loading vccs. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end
	--LogDebug("1 vccs: %s", vccs)

	startCurrentProcessTime = os.time()
	vccs = utils.FilterEntities(vccs, {IsSystemVCC = "0"})
	LogDebug("[PROVISIONING.VCCS:SEARCH]    3 - Filtered NOT System VCCs . ELAPSED Time '%s' seconds.", utils.GetElapsedTimeSec(startCurrentProcessTime))

	startCurrentProcessTime = os.time()
	local totalUnfilteredRows = utils.GetRowCount(vccs)
	LogDebug("[PROVISIONING.VCCS:SEARCH]    4 - Count VCCs - Total Unfiltered VCCs: %d. ELAPSED Time '%s' seconds.", totalUnfilteredRows, utils.GetElapsedTimeSec(startCurrentProcessTime))
	local totalAdded = 0
	local filteredRows = {}

	startCurrentProcessTime = os.time()
	local admUsers = require "apps.administration.controllers.users" -- un asco cambiar, TODO: realizar request!!
	local inProgress, encodedStatus = admUsers.IsAnyHelpDeskIntegrationInProgress()
	LogDebug("[PROVISIONING.VCCS:SEARCH]    5 - Validated Is Any HelpDesk Integration In Progress. ELAPSED Time '%s' seconds.", utils.GetElapsedTimeSec(startCurrentProcessTime))

	startCurrentProcessTime = os.time()
	local inProgressInfo = inProgress and json.decode(encodedStatus)
	LogDebug("[PROVISIONING.VCCS:SEARCH]    6 - Encoded status Is Any HelpDesk Integration In Progress. ELAPSED Time '%s' seconds.", utils.GetElapsedTimeSec(startCurrentProcessTime))

	startCurrentProcessTime = os.time()
	local hasLicenseSystem = m_helpers.utils.HasLicenseSystem()
	for id, vcc in pairs(vccs) do
		local vccId = utils.GetId(id)
		vcc.Id = vccId
		vcc.Name = vcc.Name or vccId
		if utils.MatchFilter(vcc, postData) then
			if inProgress then
				vcc.InProgressInfo = inProgressInfo
			end
			totalAdded = totalAdded + 1
			if hasLicenseSystem then
				vcc = modelVCCs.GetVccWithLicenseInfo(vcc)
			end
			if vcc then
				table.insert(filteredRows, vcc)
			end
		end
	end
	LogDebug("[PROVISIONING.VCCS:SEARCH]    7 - Filtered VCCs and Validating Licenses. ELAPSED Time '%s' seconds.", utils.GetElapsedTimeSec(startCurrentProcessTime))

	--include only rows that belong to logged user vcc
	startCurrentProcessTime = os.time()
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)
	LogDebug("[PROVISIONING.VCCS:SEARCH]    8 - Maked Response result. ELAPSED Time '%s' seconds.", utils.GetElapsedTimeSec(startCurrentProcessTime))

	local elapsedTime = utils.GetElapsedTimeSec(startTime)
	LogInfo("[PROVISIONING.VCCS:SEARCH] 9 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

local function UpdateVCCBotConfig (info)
	assert(info)
	assert(info.Id)
	assert(info.BotInfo)

	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, vccInfo = modelVCCs.Get(info.Id)
	if not success then
		LogError("[PROVISIONING.VCCS:UpdateVCCBotConfig] Error Updating VCC %s Bot Config. Not Found.", info.Id)
		return false
	end

	vccInfo.BotInfo = info.BotInfo
	SaveVCC(vccInfo)

	LogInfo("[PROVISIONING.VCCS:RASA:UpdateVCCBotConfig] Updated VCC BOT Config. User %s, BotInfo '%s'.", userId, json.encode(info.BotInfo))

	return true
end

local function UpdateVCCLlmConfig (info)
	assert(info)
	assert(info.Id)
	assert(info.LlmProfiles)

	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	local success, vccInfo = modelVCCs.Get(info.Id)
	if not success then
		LogError("[PROVISIONING.VCCS:UpdateVCCLlmConfig] Error Updating VCC %s LLM. Not Found.", info.Id)
		return false
	end

	vccInfo.LlmProfiles = info.LlmProfiles

	SaveVCC(vccInfo)

	LogInfo("[PROVISIONING.VCCS:RASA:UpdateVCCLlmConfig] Updated VCC LLM. User %s, LlmProfiles '%s'.", userId, info.LlmProfiles)

	return true
end

--[[
	COGNITIVE SERVICES (BOT RASA)
]]
local rasa = {}

--[[ RASA VCC ENABLED ---------------]]--

rasa.UpdateVCC = function(web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:RASA:UpdateVCC] Updating VCC RASA Information. User %s posted data '%s'.", userId, postedData)

	local result = {}
	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:RASA.UpdateVCC] Updating VCC RASA Information. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	-- Almaceno el Tipo de BOT RASA como activo/inactivo.
	UpdateVCCBotConfig({
		Id      = vccInfo.vcc,
		BotInfo = {Type = "RASA", Enabled = vccInfo.Enabled}
	})

	LogInfo("[PROVISIONING.VCCS:RASA:UpdateVCC] Successfully Updated VCC '%s' RASA Information.", vccInfo.vcc)

	web.status = 200
	result.result = "OK"
	return result
end

--[[
	RASA - IMPORT LUIS APP TO RASA 
]]--
rasa.ImportLuisApp = function(web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:RASA:ImportLuisApp] Importing LUIS App to RASA. User %s posted data '%s'.", userId, postedData)

	local result = {}
	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:RASA.ImportLuisApp] Importing LUIS App to RASA. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local success, err = modelVCCs.Rasa().ImportLuisApp(vccInfo)
	if not success then
		local stringErr = type(err) == "table" and json.encode(err) or err
		LogError("[PROVISIONING.VCCS:LUIS.ImportLuisApp] Error Importing LUIS App to RASA, VCC '%s'. Err: %s", vccInfo.vcc, stringErr)
		web.status = 500
		result.result = err -- Este código de error llega desde la Rocka "provisioningmanager"
		return result
	end

 	LogInfo("[PROVISIONING.VCCS:RASA:ImportLuisApp] Successfully Imported LUIS App to RASA. VCC '%s' .", vccInfo.vcc)

	web.status = 200
	result.result = "OK"
	return result
end

--[[
	RASA - IMPORT LUIS APP TO RASA 
]]--
rasa.LoadUnimportedLuisApps = function(web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:RASA:LoadUnimportedLuisApps] Retrieving Unimported LUIS Apps to RASA. User %s posted data '%s'.", userId, postedData)

	local result = {}
	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:RASA.LoadUnimportedLuisApps] Retrieving Unimported LUIS Apps to RASA. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local unimportedLuisApps, total = modelVCCs.Rasa().LoadUnimportedLuisApps(vccInfo)

 	LogInfo("[PROVISIONING.VCCS:RASA:LoadUnimportedLuisApps] Successfully Loaded Unimported LUIS Apps to RASA. VCC '%s'. Total %d", vccInfo.vcc, total)

	web.status = 200
	result.result = "OK"
	return unimportedLuisApps
end


--[[
	COGNITIVE SERVICES (BOT LUIS)
]]
local luis = {}

--[[ LUIS DEMO SUBSCRIPTIONS KEY ---------------]]--

luis.UpdateVCC = function(web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:LUIS.UpdateVCC] Updating VCC Luis Information. User %s posted data '%s'.", userId, postedData)

 	local result = {}

	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:LUIS.UpdateVCC] Updating VCC Luis Information. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	-- TENANT - Updating ---------------
	if vccInfo.tenantInfo then
		--LogDebug("1--> VCC: '%s' config Tenant.", vccInfo.vcc)
		local success, err = modelVCCs.Luis().StartAssignTenantAccountToVCC(vccInfo)
	else
		--LogDebug("1.1--> VCC: '%s' NO config Tenant.", vccInfo.vcc)
		local success, err = modelVCCs.Luis().RestoreDefaultTenantAzureAccountInVCC(vccInfo.vcc)
	end

	-- LanguageStudio - Updating --------------- 
	if vccInfo.languageStudioInfo then
		local success, err = modelVCCs.Luis().StartAssignLanguageStudioAccountToVCC(vccInfo)
	else
		local success, err = modelVCCs.Luis().RestoreDefaultLanguageStudioAccountInVCC(vccInfo.vcc)
	end

	-- AZURE ACCOUNTS & SUBSCRIPTIONS - Updating ---------------
	if vccInfo.account then
		--LogDebug("2.1--> VCC: '%s' ACTIVATING Azure Accounts & Subscriptions.", vccInfo.vcc)
		vccInfo.azure = vccInfo.account.azure
		vccInfo.subscriptions = vccInfo.account.subscriptions
		vccInfo.account = nil
		local success, err = modelVCCs.Luis().StartProcessAssignSubscriptionToVCC(vccInfo)
		if not success then
			local stringErr = type(err) == "table" and json.encode(err) or err
			LogError("[PROVISIONING.VCCS:LUIS.UpdateVCC] Error Assigning LUIS Subscription Key To VCC '%s'. Err: %s", vccInfo.vcc, stringErr)
			web.status = 500
			result.result = err -- Este código de error llega desde la Rocka "provisioningmanager"
			return result
		end
	else
		--LogDebug("2.2--> DEACTIVATING Azure Accounts & Subscriptions.", vccInfo.vcc)
		local success, err = modelVCCs.Luis().StartProcessUnAssignSubscriptionOfVCC(vccInfo)
	end

	-- DEMO - Updating ---------------
	local isDemoVCC, currentDemoInfo = modelVCCs.Luis().HasConfiguredDemoSubscriptionsInVcc(vccInfo.vcc)
	local hasActivateDemo = vccInfo.demo and vccInfo.demo.enabled or false

	--LogDebug("3--> VCC: '%s' is Demo: '%s' hasActivateDemo: '%s', '%s'.", vccInfo.vcc, isDemoVCC, hasActivateDemo, json.encode(currentDemoInfo or {}))

	if hasActivateDemo and not isDemoVCC then
	--	LogDebug("3.1--> ACTIVATING DEMO IN VCC '%s'.", vccInfo.vcc)
		local success, err = modelVCCs.Luis().StartAssignDemoSubscriptionKeyToVCC(vccInfo)
		-- if not success then
		-- 	LogError("[PROVISIONING.VCCS:LUIS.UpdateVCC] Error Updating VCC '%s' Luis Information. Err: %s", vccInfo.vcc, err)
		-- 	web.status = 500
		-- 	result.result = err -- Este código de error llega desde la Rocka "provisioningmanager"
		-- 	return result
		-- end
	elseif not hasActivateDemo and isDemoVCC then
	--	LogDebug("3.2--> DEACTIVATING DEMO IN VCC '%s'.", vccInfo.vcc)
		local success, err = modelVCCs.Luis().StartUnassignDemoSubscriptionKeyOfVCC(vccInfo)
		-- if not success then
		-- 	LogError("[PROVISIONING.VCCS:LUIS.UpdateVCC] Error Updating VCC '%s' Luis Information. Err: %s", vccInfo.vcc, err)
		-- 	web.status = 500
		-- 	result.result = err -- Este código de error llega desde la Rocka "provisioningmanager"
		-- 	return result
		-- end
	else
	--	LogDebug("3.3--> NO CAMBIA LA CONFIG DE DEMOSTRACION PARA EL VCC '%s'.", vccInfo.vcc)
		if currentDemoInfo then
			--LogDebug("3.3.1")
			local defaultDemoSubscriptionKey = modelVCCs.Luis().GetDefaultDemoSubscriptionKey()
			if vccInfo.demo and (vccInfo.demo.subscriptionKey ~= currentDemoInfo.subscriptionKey) then
				if vccInfo.demo.subscriptionKey == nil and defaultDemoSubscriptionKey == currentDemoInfo.subscriptionKey then
			--		LogDebug("3.3.1.1")
				else
			--		LogDebug("3.3.1.2")
					-- Si ha cambiado la DEMO tengo que quitar y volver a agregarla
					local success, err = modelVCCs.Luis().StartUnassignDemoSubscriptionKeyOfVCC(vccInfo)
					local success, err = modelVCCs.Luis().StartAssignDemoSubscriptionKeyToVCC(vccInfo)
				end
			else
			--	LogDebug("3.3.1.3")
			end
		else
		--	LogDebug("3.3.2")
		end
	end

	-- Almaceno el Tipo de BOT LUIS como activo/inactivo.
	UpdateVCCBotConfig({
		Id      = vccInfo.vcc,
		BotInfo = {Type = "LUIS", Enabled = vccInfo.Enabled}
	})

 	LogInfo("[PROVISIONING.VCCS:LUIS.UpdateVCC] Successfully Updated VCC '%s' Luis Information.", vccInfo.vcc)

	web.status = 200
	result.result = "OK"
	return result
end

--[[
	COGNITIVE SERVICES (QnA)
]]
local qna = {}

qna.UpdateVCC = function(web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:QNA.UpdateVCC] Updating VCC QNA Information. User %s posted data '%s'.", userId, postedData)

 	local result = {}

	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:QNA.UpdateVCC] Updating VCC QNA Information. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	-- Updating ---------------
	local _, total = modelVCCs.QnA().LoadVCCSubscriptions(vccInfo.vcc)
	if total > 0 then
	--	LogDebug("1.1--> DEACTIVATING QNA IN VCC '%s'. Total current subscriptions %s.", vccInfo.vcc, total)
		local success, err = modelVCCs.QnA().StartProcessUnAssignSubscriptionOfVCC(vccInfo)
		if not success then
			LogError("[PROVISIONING.VCCS:QNA.UpdateVCC] -Error Updating VCC '%s' QNA Information. Err: %s", vccInfo.vcc, err)
			web.status = 500
			result.result = err -- Este código de error llega desde la Rocka "provisioningmanager"
			return result
		end
	end
	if vccInfo.subscriptions then
	--	LogDebug("1.2--> ACTIVATING QNA IN VCC '%s'.", vccInfo.vcc)
		local success, err = modelVCCs.QnA().StartProcessAssignSubscriptionToVCC(vccInfo)
		if not success then
			LogError("[PROVISIONING.VCCS:QNA.UpdateVCC] +Error Updating VCC '%s' QNA Information. Err: %s", vccInfo.vcc, err)
			web.status = 500
			result.result = err -- Este código de error llega desde la Rocka "provisioningmanager"
			return result
		end
	end

	-- Updating --------------- fede
	local _, totalV2 = modelVCCs.QnA().LoadVCCSubscriptionsV2(vccInfo.vcc)
	if totalV2 > 0 then
	--	LogDebug("1.1--> DEACTIVATING QNA IN VCC '%s'. Total current subscriptions %s.", vccInfo.vcc, total)
		local success, err = modelVCCs.QnA().StartProcessUnAssignSubscriptionOfVCCV2(vccInfo)
		if not success then
			LogError("[PROVISIONING.VCCS:QNA.UpdateVCC] -Error Updating VCC '%s' QNA V2 Information. Err: %s", vccInfo.vcc, err)
			web.status = 500
			result.result = err -- Este código de error llega desde la Rocka "provisioningmanager"
			return result
		end
	end
	if vccInfo.subscriptionsV2 then -- fede
	--	LogDebug("1.2--> ACTIVATING QNA IN VCC '%s'.", vccInfo.vcc)
		local success, err = modelVCCs.QnA().StartProcessAssignSubscriptionToVCCV2(vccInfo)
		if not success then
			LogError("[PROVISIONING.VCCS:QNA.UpdateVCC] +Error Updating VCC '%s' QNA V2 Information. Err: %s", vccInfo.vcc, err)
			web.status = 500
			result.result = err -- Este código de error llega desde la Rocka "provisioningmanager"
			return result
		end
	end

 	LogInfo("[PROVISIONING.VCCS:QNA.UpdateVCC] Successfully Updated VCC '%s' QNA Information.", vccInfo.vcc)

	web.status = 200
	result.result = "OK"
	return result
end

--[[
	LLM
]]
local llm = {}

llm.UpdateVCC = function(web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:LLM.UpdateVCC] Updating VCC LLM Information. User %s posted data '%s'.", userId, postedData)

 	local result = {}

	local success, parsedData = pcall(json.decode, postedData)
	
	UpdateVCCLlmConfig({
		Id          = parsedData.vcc,
		LlmProfiles = parsedData.LlmProfiles
	})

 	LogInfo("[PROVISIONING.VCCS:LLM.UpdateVCC] Successfully Updated VCC '%s' LLM Information.", vcc)

	web.status = 200
	result.result = "OK"
	return result
end

--[[
	PEERS
]]
local peers = {}

peers.UpdateVCC = function(web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:PEERS.UpdateVCC] Updating VCC PEERS Information. User %s.", userId)
	--LogDebug("[PROVISIONING.VCCS:PEERS.UpdateVCC] --> Posted data '%s'.", postedData)

	local result = {}

	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:PEERS.UpdateVCC] Updating VCC PEERS Information. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local hasAllegroIntegration = utils.HasAllegroIntegration(vccInfo.vcc)
	if hasAllegroIntegration then
		LogError("[PROVISIONING.VCCS:PEERS.UpdateVCC] Updating VCC PEERS Information. Unable Assign Peers to VCC with Allegro Integration.")
		web.status = 400
		result.result = "UNABLE ASSIGN PEERS TO VCC WIT ALLEGRO INTEGRATION"
		result.err = err
		return result
	end

	-- Updating ---------------
	local success, err = modelVCCs.Peers().UpdatePeersInVCC(vccInfo.vcc, vccInfo.peers, vccInfo.assignationType)
	if not success then
		LogError("[PROVISIONING.VCCS:PEERS.UpdateVCC] Updating VCC PEERS Information. Error Updating Peers in VCC.")
		web.status = 400
		result.result = "ERROR UPDATING PEERS IN VCC"
		result.err = err
		return result
	end

	LogInfo("[PROVISIONING.VCCS:PEERS.UpdateVCC] Successfully Updated VCC '%s' PEERS Information.", vccInfo.vcc)
	result.warningList = err
	if not result.warningList or (#result.warningList.assign == 0 and #result.warningList.remove == 0) then
		result.code = "ALL PEERS UPDATED"
	else
		result.code = "PARTIAL UPDATED PEERS"
	end

	web.status = 200
	result.result = "OK"
	return result
end

peers.ResetPeersInRedis = function(web)

	local postedData = web.input["post_data"]
	local success, vccInfo = pcall(json.decode, postedData)
	
	LogInfo("[PROVISIONING.VCCS:PEERS.ResetPeersInRedis] Reseting VCC PEERS Information on Redis. VCC: %s",vccInfo.vcc)
	
	local result = {}

	local hasAllegroIntegration = utils.HasAllegroIntegration(vccInfo.vcc)
	if hasAllegroIntegration then
		LogError("[PROVISIONING.VCCS:PEERS.UpdateVCC] Updating VCC PEERS Information. Unable Assign Peers to VCC with Allegro Integration.")
		web.status = 400
		result.result = "UNABLE ASSIGN PEERS TO VCC WIT ALLEGRO INTEGRATION"
		result.err = err
		return result
	end

	-- Reset ---------------
	local success = modelVCCs.Peers().ResetPeersInRedis(vccInfo.vcc)
	if not success then
		LogError("[PROVISIONING.VCCS:PEERS.ResetPeersInRedis] Reseting VCC PEERS Information. Error Updating Peers in VCC.")
		web.status = 400
		result.result = "ERROR RESETING PEERS IN VCC"
		result.err = err
		return result
	end

	LogInfo("[PROVISIONING.VCCS:PEERS.ResetPeersInRedis] Successfully Reseting VCC '%s'.", vccInfo.vcc)
	web.status = 200
	result.result = "OK"
	return result
end
--[[
]]--
peers.LoadAvailablesVccPeers = function (web, currentVcc)
	local userVcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, userVcc)

	LogInfo("[PROVISIONING.VCCS.:LoadAvailablesVccPeers] Loading Availables Vcc '%s' Peers, user %s, vcc %s", currentVcc, userId, userVcc)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, freeVccPeers = modelPeers.GetFreeVCCPeers(keyspace)
	if not success then
		keyspace:Close()
		local err = freeVccPeers
		LogWarning("[PROVISIONING.VCCS:LoadAvailablesVccPeers] Error loading Free Vcc Peers. Err '%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	local success, vccPeers = modelPeers.GetVCCPeers(currentVcc, keyspace)
	keyspace:Close()
	if not success then
		local err = vccPeers
		LogWarning("[PROVISIONING.VCCS:LoadAvailablesVccPeers] Error loading Vcc '%s' Peers. Err '%s'.", currentVcc, err)
		web.status = 500
		return {result = "SERVER ERROR", err = vccPeers}
	end

	-- OK
	--
	-- Funcion que Separa el Id del VCC obteniendo el Tech y el Numero y los agrega en atributos discretos al objeto.
	local function SeparateAndUpdate (peer)
		if not peer then
			return
		end
		local tech, name = peers.SeparatePeer(peer.id)
		peer.tech = tech
		peer.name = name
	end
	-- Lista de Peers asignados al VCC y los libres.
	--
	local availablesVccPeers = {}
	-- Esta funcion cargar el array de peers libres y los utilizados por el current vcc.
	local function FillAvailableList (list, isAssigned)
		-- Se convierte el hash a un array para que luego pueda ser ordenado por numero de peer.
		local array = utils.HashToArray(list, "id")
		if #array > 1 then
			-- Ahora se ordena el array por numero de peer.
			array = utils.SortArray(array, function(peer)
				SeparateAndUpdate(peer)
				if not peer.name then
					return 0
				end
				return tonumber(peer.name) or 0
			end)
		else
			SeparateAndUpdate(array[1])
		end
		-- Ahora recorro el array ordenado y cargo el array de disponibles.
		for _, peer in ipairs(array) do
			table.insert(availablesVccPeers, {
				Id         = peer.id,
				Tech       = peer.tech,
				Name       = peer.name,
				HostId     = peer.hostId,
				Assigned   = isAssigned and peer.assignationType or nil
			})
		end
	end

	FillAvailableList(vccPeers, true)
	FillAvailableList(freeVccPeers, false)

	web.status = 200
	return availablesVccPeers
end

peers.SeparatePeer = function (peerId)
	return modelVCCs.Peers().SeparatePeer(peerId)
end

--[[
	PACKAGES
]]
local packages = {}

packages.UpdateVCC = function(web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:PACKAGES.UpdateVCC] Updating VCC PACKAGES Information. User %s.", userId)
	--LogDebug("[PROVISIONING.VCCS:PACKAGES.UpdateVCC] --> Posted data '%s'.", postedData)

	local result = {}

	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:PACKAGES.UpdateVCC] Updating VCC PACKAGES Information. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local hasAllegroIntegration = utils.HasAllegroIntegration(vccInfo.vcc)
	if hasAllegroIntegration then
		LogError("[PROVISIONING.VCCS:PACKAGES.UpdateVCC] Updating VCC PACKAGES Information. Unable Assign Packages to VCC with Allegro Integration.")
		web.status = 400
		result.result = "UNABLE ASSIGN PACKAGES TO VCC WIT ALLEGRO INTEGRATION"
		result.err = err
		return result
	end

	-- Updating ---------------
	local success, err = modelVCCs.Packages().UpdatePackagesInVCC(vccInfo.vcc, vccInfo.packages)
	if not success then
		LogError("[PROVISIONING.VCCS:PACKAGES.UpdateVCC] Updating VCC PACKAGES Information. Error Updating Packages in VCC.")
		web.status = 400
		result.result = "ERROR UPDATING PACKAGES IN VCC"
		result.err = err
		return result
	end

	LogInfo("[PROVISIONING.VCCS:PACKAGES.UpdateVCC] Successfully Updated VCC '%s' PACKAGES Information.", vccInfo.vcc)

	result.warningList = err
	if not result.warningList or (#result.warningList.assign == 0 and #result.warningList.remove == 0) then
		result.code = "ALL PACKAGES UPDATED"
	else
		result.code = "PARTIAL UPDATED PACKAGES"
	end

	web.status = 200
	result.result = "OK"
	return result
end

--[[
]]--
packages.LoadAvailablesVccPackages = function (web, currentVcc)
	local userVcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, userVcc)

	LogInfo("[PROVISIONING.VCCS:PACKAGES.LoadAvailablesVccPackages] Loading Availables Vcc '%s' Packages, user %s, vcc %s", currentVcc, userId, userVcc)
	--LogDebug("[PROVISIONING.VCCS:PACKAGES.LoadAvailablesVccPackages] 1 ----> web.vars.QUERY_STRING: '%s'", web.vars.QUERY_STRING)
	local webParameters = utils.ParseWebQueryString(web.vars.QUERY_STRING)
	--LogDebug("[PROVISIONING.VCCS:PACKAGES.LoadAvailablesVccPackages] 2 ----> parsed webParameters %s", utils.JSON_Encode(webParameters))

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, freeVccPackages = modelPackages.GetFreeVCCPackages(keyspace)
	if not success then
		keyspace:Close()
		local err = freeVccPackages
		LogWarning("[PROVISIONING.VCCS:PACKAGES.LoadAvailablesVccPackages] Error loading Free Vcc Packages. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	local success, vccPackages = modelPackages.GetVCCPackages(currentVcc, keyspace)
	if not success then
		keyspace:Close()
		local err = vccPackages
		LogWarning("[PROVISIONING.VCCS:PACKAGES.LoadAvailablesVccPackages] Error loading Vcc '%s' Packages. Err'%s'.", currentVcc, err)
		web.status = 500
		return {result = "SERVER ERROR", err = vccPackages}
	end

	-- OK
	--
	-- Lista de Packages asignados al VCC y los libres.
	local availablesVccPackages = {}
	local function FillAvailableList (list, isAssigned)
		for packageId in pairs(list) do
			local availablePackage = {Id = packageId, IsAssigned = isAssigned, data = nil}
			if webParameters and tostring(webParameters.full_data) == "1" then
				local success, packageInfo = modelPackages.Get(packageId, keyspace)
				if success then
					availablePackage.data = packageInfo
				end
			end
			table.insert(availablesVccPackages, availablePackage)
		end
	end

	FillAvailableList(vccPackages, true)
	FillAvailableList(freeVccPackages, false)
	keyspace:Close()

	web.status = 200
	return availablesVccPackages
end

--[[
	ACCOUNTS
]]
local accounts = {}

--[[
]]--
accounts.LoadAvailablesVccAccounts = function (web, currentVcc)
	local userVcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, userVcc)
	
	LogInfo("[PROVISIONING.VCCS:ACCOUNTS.LoadAvailablesVccAccounts] Loading Availables Vcc '%s' accounts, user %s, vcc %s", currentVcc, userId, userVcc)
	local vccAccounts = modelVCCs.Accounts().GetVCCAccounts(currentVcc) or {}
	local result = {}
	for accountType, accounts in pairs(vccAccounts) do
		for accountId, account in pairs(accounts) do
			result[accountType] = result[accountType] or {}
			table.insert(result[accountType], { id = accountId, name = account.name })
		end
	end
	web.status = 200
	return result
end

--[[
	INTEGRATIONS - HELPDESK
]]--

--[[
]]--
m_integrations.helpdesk.Synchronize = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:INTEGRATIONS.HELPDESK.Synchronize] Synchronize Users and Roles of VCC with HELPDESK. User %s.", userId)
	--LogDebug("[PROVISIONING.VCCS:INTEGRATIONS.HELPDESK.Synchronize] --> Posted data '%s'.", postedData)

	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:INTEGRATIONS.HELPDESK.Synchronize] Unable to Synchronizing VCC with HELPDESK. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end

	local admUsers = require "apps.administration.controllers.users" -- un asco cambiar, TODO: realizar request!!
	local success, result = admUsers.UpdateHelpDeskIntegrations(vccInfo)

	if not success then
		local err = result
		local stringErr = type(err) == "table" and json.encode(err) or err
		LogError("[PROVISIONING.VCCS:INTEGRATIONS.HELPDESK.Synchronize] Unable to Synchronizing VCC with HELPDESK VCC '%s'. Err: %s", vccInfo.vcc, stringErr)
		web.status = 500
		return result
	end

	web.status = 200
	result.result = "OK"
	return result
end

--[[
	widgets
]]--
m_integrations.widgets.api = require "apps.integrations.api" -- WIDGETS INTEGRATIONS API

m_integrations.widgets.cachedWidgetsDefinitions = nil
m_integrations.widgets.LoadWidgetsDefinitions = function()
	--LogDebug("[ROVISIONING.VCCS:INTEGRATIONS.WIDGETS.LoadWidgetsDefinitions] 1- CACHED Widgets Definitions: %s", m_integrations.widgets.cachedWidgetsDefinitions)
	if not m_integrations.widgets.cachedWidgetsDefinitions then
		local definitions = m_integrations.widgets.api.GetAvailablesWidgetsDefinitions()
		for _, def in ipairs(definitions) do
			local widgetDef = def
			widgetDef.Name = def.widgetId
			widgetDef.Description = def.displayName
		end
		m_integrations.widgets.cachedWidgetsDefinitions = definitions
	end
	--LogDebug("[ROVISIONING.VCCS:INTEGRATIONS.WIDGETS.LoadWidgetsDefinitions] 2- CACHED Widgets Definitions: %s", m_integrations.widgets.cachedWidgetsDefinitions)
	return m_integrations.widgets.cachedWidgetsDefinitions
end

--[[
	groups
]]--
local groups = {}

groups.LoadVccGroups = function (web, currentVcc)
	local userVcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, userVcc)
	LogInfo("[PROVISIONING.VCCS:GROUPS.LoadVccGroups] Loading Availables Vcc '%s' groups, user %s, vcc %s", currentVcc, userId, userVcc)
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local vccGroups = modelVCCs.Groups().GetVCCGroups(currentVcc, ksAgent) or {}
	ksAgent:Close()
	--No incluimos los grupos: "deployer", "full-system-access"
	local result = {}
	local excludeRoles = {
		["deployer"]           = true,
		["full-system-access"] = true
	}
	for vccGroupName, vccGroup in pairs(vccGroups) do
		vccGroupName = vccGroupName:lower()
		if not excludeRoles[vccGroupName] then
			result[vccGroupName] = vccGroup
		end
	end
	--Agregamos un grupo ficticio allusers, para que si no hay sincronizacion los usuarios se busquen por el filtro asociado a este grupo
	result["allusers"] = {
		Description = "All Users",
		Id = "allusers",
		IsDeployer = "0",
		IsSystemGroup = "1",
		Name = "allusers"
	}
	web.status = 200
	return result
end

--[[
	LDAP
]]--
local ldap = {}

--[[
]]--
ldap.GetLdapGroups = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	--LogDebug("[PROVISIONING.VCCS:LDAP.GetLdapGroups] --> Posted data '%s'.", postedData)

	local success, data = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:LDAP.GetLdapGroups] Unable to Load LDAP tree. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end

	if not data.vcc or not data.server or not data.baseDN or not data.adminUserDN or not data.password then
		LogError("[PROVISIONING.VCCS:LDAP.GetLdapGroups] Unable to Load LDAP tree. Invalid Request. Some required arguments missing")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end

	LogInfo("[PROVISIONING.VCCS:LDAP.GetLdapGroups] Loading LDAP tree for VCC %s. User %s.", data.vcc, userId)

	local authToken = web.vars.auth.authToken
	local jwtData, jwtErr = jwt.decode(data.password, md5.sumhexa(authToken), "HS256")
	if jwtErr then
		LogError("[PROVISIONING.VCCS:LDAP.GetLdapGroups] Error decoding jwtToken: '%s'", jwtErr)
		web.status = 400
		return {result = "INVALID REQUEST"}
	end
	data.password = jwtData.password

	local externalAuthFakePassword = modelVCCs.GetExternalAuthFakePassword()
	if data.password == externalAuthFakePassword then
		data.password = modelVCCs.GetCurrentExternalAuthPassword(data.vcc)
	end

	local success, result = modelVCCs.Ldap().GetLdapGroups(data.vcc, data.server, data.baseDN, data.adminUserDN, data.password, data.useTLS)
	if not success then
		local err = result
		local stringErr = type(err) == "table" and json.encode(err) or err
		LogError("[PROVISIONING.VCCS:LDAP.GetLdapGroups] Unable to Load LDAP tree. Err: %s", stringErr)
		web.status = 500
		return result
	end

	web.status = 200

	if type(result) == "table" and result.success == false then
		local _T = m_htmlRender.extension.makeTextTranslator()
		result.error = result.error and _T(result.error) or ""
	end
		
	return result
end	

--[[
	Azure AD
]]--
local azuread = {}

--[[
]]--
azuread.GetAzureADGroups = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	--LogDebug("[PROVISIONING.VCCS:LDAP.GetAzureADGroups] --> Posted data '%s'.", postedData)

	local success, data = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:AZUREAD.GetAzureADGroups] Unable to Load AZUREAD groups. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end

	if not data.vcc or not data.openid_configuration_url or not data.client_id or not data.client_secret then
		LogError("[PROVISIONING.VCCS:AZUREAD.GetAzureADGroups] Unable to Load AZUREAD groups. Invalid Request. Some required arguments missing")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end

	LogInfo("[PROVISIONING.VCCS:AZUREAD.GetAzureADGroups] Loading AZUREAD groups for VCC %s. User %s.", data.vcc, userId)

	local authToken = web.vars.auth.authToken
	local jwtData, jwtErr = jwt.decode(data.client_secret, md5.sumhexa(authToken), "HS256")
	if jwtErr then
		LogError("[PROVISIONING.VCCS:LDAP.GetLdapGroups] Error decoding jwtToken: '%s'", jwtErr)
		web.status = 400
		return {result = "INVALID REQUEST"}
	end
	data.client_secret = jwtData.password

	local externalAuthFakePassword = modelVCCs.GetExternalAuthFakePassword()
	if data.client_secret == externalAuthFakePassword then
		data.client_secret = modelVCCs.GetCurrentExternalAuthPassword(data.vcc)
	end

	local success, result = modelVCCs.AzureAD().GetAzureADGroups(data.openid_configuration_url, data.client_id, data.client_secret)
	if not success then
		local err = result
		local stringErr = type(err) == "table" and json.encode(err) or err
		LogError("[PROVISIONING.VCCS:AZUREAD.GetAzureADGroups] Unable to Load AZUREAD groups. Err: %s", stringErr)
		web.status = 500
		return result
	end

	web.status = 200

	if type(result) == "table" and result.success == false then
		local _T = m_htmlRender.extension.makeTextTranslator()
		result.error = result.error and _T(result.error) or ""
	end
		
	return result
end	

--[[
	CAMPAIGNS
]]
local campaigns = {}

--[[
]]--
campaigns.LoadVccCampaigns = function (web, currentVcc)
	local userVcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, userVcc)
	
	LogInfo("[PROVISIONING.VCCS:CAMPAIGNS.LoadVccCampaigns] Loading Vcc '%s' campaigns, user %s, vcc %s", currentVcc, userId, userVcc)
	local result = modelVCCs.Campaigns().GetVCCCampaigns(currentVcc) or {}

	web.status = 200
	return result
end

--[[
	Recording Configuration
]]
local recording_configuration = {}

--[[
]]--
recording_configuration.LoadDefaults = function (web)
	local userVcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, userVcc)
	LogInfo("[PROVISIONING.VCCS:RecordingConfiguration.LoadDefaults] Loading Default Recording Configuration, user %s, vcc %s", userId, userVcc)
	local result = modelVCCs.RecordingConfiguration().LoadDefaults() or {}
	web.status = 200
	return result
end

--[[
	MIGRATIONS - ALLEGRO VCC TO NATIVE
]]--

--[[

]]--
m_migrations.allegro.Require = function ()
	return require "apps.provisioning.controllers.migrations.init"
end

--[[

]]--
m_migrations.allegro.StartAsync = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogDebug("\n\n\n")
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.StartAsync] START ASYNC Migration Allegro VCC to Native VCC. Posted data '%s'.", postedData)
	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.StartAsync] Unable to Migrate Allegro VCC with Native VCC. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end
	local success, resp = m_migrations.allegro.Require():Start(vccInfo)
	if not success then
		local err = resp
		LogError("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.StartAsync] Unable to Migrate Allegro VCC with Native VCC %s. Err: %s", vccInfo.vcc, utils.JSON_Encode(err))
		web.status = 500
		return err
	end
	-- ok
	local result = resp
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.StartAsync] END ASYNC Migrated Allegro VCC with Native VCC %s. Result: %s", vccInfo.vcc, utils.JSON_Encode(result))
	LogDebug("\n\n\n")
	web.status = 200
	result.result = "OK"
	return result
end

--[[

]]--
m_migrations.allegro.Start = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Start] START Migration Allegro VCC to Native VCC. Posted data '%s'.", postedData)
	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Start] Unable to Migrate Allegro VCC with Native VCC. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local ksProv  = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, resp = m_migrations.allegro.Require():CanStartMigration({
		vccId             = vccInfo.vcc,
		ksAgent           = ksAgent,
		ksProv            = ksProv,
		step              = vccInfo.step,
		contactsMigration = vccInfo.contactsMigration
	})
	ksAgent:Close()
	ksProv:Close()
	if not success then
		local err = resp
		LogError("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Start] Unable to Migrate Allegro VCC with Native VCC %s. Err: %s", vccInfo.vcc, utils.JSON_Encode(err))
		web.status = 500
		return err
	end
	-- ejecuto el request asincronico
	utils.AsyncTasksWithAuthToken(
		"/inconcert/apps/provisioning/vccs/migrations/allegro/to/native/async/", -- url
		"MigrationAllegroVcc",                                                   -- taskId
		"POST",                                                                  -- method
		nil,                                                                     -- headers
		postedData,                                                              -- jsonData
		nil                                                                      -- cookies
	)
	-- ok
	local result = {}
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Start] END Migrated Allegro VCC with Native VCC %s. Result: %s", vccInfo.vcc, utils.JSON_Encode(result))
	web.status = 200
	result.result = "STARTED ASYNC MIGRATION - WAIT MOMENT"
	return result
end

--[[

]]--
m_migrations.allegro.RollbackAsync = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogDebug("\n\n\n")
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.RollbackAsync] Rollback ASYNC Migration Allegro VCC to Native VCC. Posted data '%s'.", postedData)
	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.RollbackAsync] Unable to Rollback Allegro VCC with Native VCC. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end
	local success, resp = m_migrations.allegro.Require():Rollback(vccInfo)
	if not success then
		local err = resp
		LogError("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.RollbackAsync] Unable to Rollback Allegro VCC with Native VCC %s. Err: %s", vccInfo.vcc, utils.JSON_Encode(err))
		web.status = 500
		return err
	end
	-- ok
	local result = {}
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.RollbackAsync] END Rollbacked ASYNC Allegro VCC with Native VCC %s. Result: %s", vccInfo.vcc, utils.JSON_Encode(result))
	LogDebug("\n\n\n")
	web.status = 200
	result.result = "OK"
	return result
end

--[[

]]--
m_migrations.allegro.Rollback = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Rollback] Rollback Migration Allegro VCC to Native VCC. Posted data '%s'.", postedData)
	local success, vccInfo = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Rollback] Unable to Rollback Allegro VCC with Native VCC. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, resp = m_migrations.allegro.Require():CanStartRollback(vccInfo.vcc, ksAgent, ksProv)
	ksAgent:Close()
	ksProv:Close()
	if not success then
		local err = resp
		LogError("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Rollback] Unable to Rollback Allegro VCC with Native VCC %s. Err: %s", vccInfo.vcc, utils.JSON_Encode(err))
		web.status = 500
		return err
	end
	-- ejecuto el request asincronico
	utils.AsyncTasksWithAuthToken(
		"/inconcert/apps/provisioning/vccs/migrations/allegro/rollback/async/", -- url
		"RollbackAllegroVcc",                                                   -- taskId
		"POST",                                                                 -- method
		nil,                                                                    -- headers
		postedData,                                                             -- jsonData
		nil                                                                     -- cookies
	)
	-- ok
	local result = {}
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Rollback] END Rollbacked Allegro VCC with Native VCC %s. Result: %s", vccInfo.vcc, utils.JSON_Encode(result))
	web.status = 200
	result.result = "STARTED ASYNC ROLLBACK - WAIT MOMENT"
	return result
end

--[[

]]--
m_migrations.allegro.IsRunning = function ()
	return m_migrations.allegro.Require():IsRunning()
end

--[[

]]--
m_migrations.allegro.Status = function (web)
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Status] Loading current migration or rollback status.")
	local response = {
		IsRunning        = nil,
		CurrentMigration = nil
	}
	local isRunning, currentMigration = m_migrations.allegro.IsRunning()
	if not isRunning then
		LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Status] No migration or rollback process is in progress.")
		response.IsRunning = false
		return response
	end
	response.IsRunning = true
	response.CurrentMigration = currentMigration
	LogInfo("[PROVISIONING.VCCS:MIGRATIONS.ALLEGRO.Status] There is a '%s' process in progress for vcc: '%s'.", true and "MIGRATION" or "ROLLBACK", "VCCeMMA")
	return response
end

--[[

]]--
m_migrations.allegro.GetAllegroInfo = function (vccId, ksAgent)
	return m_migrations.allegro.Require():GetAllegroInfo(vccId, ksAgent)
end

------------------------------
-- ALLEGRO - Cuando se invoca el Save desde otra App (Ej apps/allegro/controllers/provisioning_vccs.lua)
------------------------------
function ExternalSaveVCC (vccInfo)
	assert(vccInfo)
	LogInfo("[PROVISIONING.VCCS:ExternalSaveVCC] Updating VCC '%s'.", json.encode(vccInfo))

	local vccId = vccInfo.Id
	local success, vccInfo, firstAdminUser, firstApiUsers, firstSuperUser = __ValidateAndMakeInfoToUpdate(vccId, vccInfo)
	if not success then
		local errMsg = vccInfo
		LogError("[PROVISIONING.VCCS:ExternalSaveVCC] Error Validating VCC '%s'. Err: %s", vccId, errMsg)
		return false, errMsg
	end

	-- Ok
	local success, errMsg = SaveVCC(vccInfo)
	return success, errMsg, firstApiUsers, firstSuperUser
end

------------------------------
-- LICENSE
------------------------------
local function GetAvailableLicenseResources()
	return modelVCCs.GetAvailableLicenseResources()
end

local function CompanyIsEditable()
	return modelVCCs.CompanyIsEditable()
end

local function CompanyIsRequired()
	return modelVCCs.CompanyIsRequired()
end

local function ValidateCompany(web, company)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.VCCS:ValidateCompany] Validating company '%s'. User %s.", company, userId)	
	local companyInfo = modelVCCs.ValidateCompany(company)
	web.status = 200
	return companyInfo
end

------------------------------
-- REPOSITORY SERVICE
------------------------------
--[[
]]--
local function GetS3Regions (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	--LogDebug("[PROVISIONING.VCCS:GetS3Regions] --> Posted data '%s'.", postedData)

	LogInfo("[PROVISIONING.VCCS:GetS3Regions] Loading S3 Regions for VCC %s. User %s.", vcc, userId)

	local success, result = modelVCCs.GetS3Regions()
	if not success then
		local err = result
		local stringErr = type(err) == "table" and json.encode(err) or err
		LogError("[PROVISIONING.VCCS:GetS3Regions] Unable to Load S3 Regions. Err: %s", stringErr)
		web.status = 500
		return result
	end

	web.status = 200

	if type(result) == "table" and result.success == false then
		local _T = m_htmlRender.extension.makeTextTranslator()
		result.error = result.error and _T(result.error) or ""
	end
		
	return result
end

--[[
	SUPER USERS
]]--

--[[
]]--
m_superUsers.Save = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	LogInfo("[PROVISIONING.VCCS:SUPER_USERS.Save] Saving Super User: '%s'.", postedData)
	local success, info = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:SUPER_USERS.Save] Unable to Saving Super User VCC. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end
	--
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local currentSuperUser = modelUsers.GetSuperUser(info.vccId, info.Id, ksAgent, {"Id"})
	--LogDebug("[PROVISIONING.VCCS:SUPER_USERS.Save] ----> userId: '%s@%s' currentSuperUser: '%s'.", info.Id, info.vccId, json.encode(currentSuperUser or {}))
	info.newUser = not currentSuperUser and true -- SI, es un nuevo Usuario
										or false -- NO es un nuevo Usuario
	local superUser = __MakeTemplateSuperUser(info)
	local success, err = m_superUsers.CreateSuperUserInVCC(superUser, ksAgent)
	if not success then
		ksAgent:Close()
		LogError("[PROVISIONING.VCCS:SUPER_USERS.Save] Unable to Saving Super User VCC. Save User error: %s", encodedResp)
		web.status = 400
		return {
			result   = "API SAVING USER ERROR",
			rawError = err
		}
	end
	--ok
	local success, resp = m_superUsers.AssignSuperUserInAllCampaigns(superUser, info.vccId, ksAgent)

	ksAgent:Close()
	web.status = 200
	return {result = "OK"}
end

--[[

]]--
m_superUsers.CreateSuperUserInVCC = function(superUser, ksAgent)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("[PROVISIONING.VCCS:SUPER_USERS.CreateSuperUserInVCC] Create super user in vcc. User %s.", userId)
	return CreateFirstUserInVCC(superUser, ksAgent)
end

--[[

]]--
m_superUsers.AssignSuperUserInAllCampaigns = function(superUser, superUserVcc, ksAgent)
	assert(type(superUser) == "table")
	assert(type(superUserVcc) == "string")
	assert(ksAgent)
	if not superUser.newUser then
		local resp = {result = "UNABLE TO INCLUDE IN CAMPAIGNS - SUPER USER ALREADY EXIST", superUserId = superUser.Id}
		LogInfo("[PROVISIONING.VCCS:SUPER_USERS.AssignSuperUserInAllCampaigns] Unable to assing Super User '%s' in campaigns. Result: %s", superUser.Id, json.encode(resp))
		return true, resp
	end
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local vccCampaigns = utils.GetVCCCampaigns(superUserVcc, ksAgent)

	LogInfo("[PROVISIONING.VCCS:SUPER_USERS.AssignSuperUserInAllCampaigns] -> Assigning Super User '%s@%s' in '%s' campaigns. UserId: %s", superUser.Id, superUserVcc, utils.GetRowCount(vccCampaigns), userId)
	--[[
		Funcion que valida si puede agregar como Member de la campana al SuperUsuario.
		Se excluyen las campanas del sistema "internal_calls" y "sandbox"
	]]--
	local function MustUpdateCampaign(campaignId)
		local excludeCampaigns = {
			internal_calls = true,
			sandbox        = true
		}
		campaignId = campaignId:lower()
		if excludeCampaigns[campaignId] then
			LogDebug("[PROVISIONING.VCCS:SUPER_USERS.AssignSuperUserInAllCampaigns] MustUpdateCampaign-> Exclude campaignId '%s'", campaignId)
			return false
		end
		return true
	end
	--------------
	local warnings = {}
	for simpleCampId in pairs(vccCampaigns) do
		if MustUpdateCampaign(simpleCampId) then
			local usersCampaign = utils.GetCampaignMembers(superUserVcc, simpleCampId, ksAgent)
			usersCampaign = utils.Remap(usersCampaign, utils.GetId)
			if not usersCampaign[superUser.Id] then
				local fullCampaignId = utils.AppendVcc(simpleCampId, superUserVcc)
				local skipLimitsValidation = true
				local success, resp = m_superUsers.UpdateMembersCampaign(fullCampaignId, usersCampaign, ksAgent, nil, nil, skipLimitsValidation)
				if not success then
					table.insert(warnings, resp)
				end
			end
		end
	end
	LogDebug("[PROVISIONING.VCCS:SUPER_USERS.AssignSuperUserInAllCampaigns] -> Assigning Super User '%s@%s' in '%s' campaigns. warnings: %s", superUser.Id, superUserVcc, utils.GetRowCount(vccCampaigns), json.encode(warnings))
	return true, warnings
end

--[[
]]--
m_superUsers.UpdateMembersCampaign = function(campaignId, usersCampaign, ksAgent)
	LogInfo("[PROVISIONING.VCCS:SUPER_USERS.UpdateMembersCampaign] Invoke Administration Update members. Campaign: '%s'.", campaignId)
	local admCampController = utils.GeAdmController({controllerName = "campaigns", modelName = "Campaigns"})
	local success, resp = admCampController.UpdateMembersList(campaignId, usersCampaign, ksAgent)
	if not success then
		-- err
		local err = resp
		err.campaignId = campaignId
		LogWarning("[PROVISIONING.VCCS:SUPER_USERS.UpdateMembersCampaign] Unable to include super user in campaign: '%s'. Err: %s", campaignId, json.encode(err or {}))
		return false, err
	end
	-- ok
	return true
end

--[[
]]--
m_superUsers.Search = function (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	LogInfo("[PROVISIONING.VCCS:SUPER_USERS.Search] Searching Super Users: '%s'.", postedData)
	local success, info = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.VCCS:SUPER_USERS.Search] Unable to Search Super Users. Invalid Request.")
		web.status = 400
		return {result = "INVALID REQUEST"}
	end
	-- ok
	local withFullData = true
	local columns = {
		"areaCode",
		"countryCode",
		"FirstSurname",
		"Id",
		"LanguageId",
		"LastSurname",
		"Mail",
		"Name",
		"TimeZoneId"
	}
	local vccSuperUsers = modelUsers.GetVCCSuperUsers(info.vcc, withFullData, nil, columns)
	--LogDebug("[PROVISIONING.VCCS:SUPER_USERS.Search] Super Users List: '%s'.", json.encode(vccSuperUsers))
	-- convierto en un array y lo ordeno
	vccSuperUsers = utils.SortArray(
		utils.HashToArray(vccSuperUsers, "Id"),
		function(row) return row.Id end
	)
	web.status = 200
	return vccSuperUsers
end

------------------------------------
-- Mapas de las URLS
----------------------------------
function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn, validationDef)
		-- LogDebug("[[PROVISIONING.VCCs:MapUris] --> validationDef: %s", json.encode(validationDef or {}))
		responseFn = responseFn or replyFn
		local fnToExecute
		if validationDef then
			--LogDebug("[[PROVISIONING.VCCs:MapUris] 1 --> validationDef: %s", json.encode(validationDef or {}))
			fnToExecute = utils.WebJsonValidatorData(
				validationDef,
				fn
			)
		else
			--LogDebug("[[PROVISIONING.VCCs:MapUris] 2 -->")
			fnToExecute = fn
		end

		return responseFn
			.. LogRequest
			.. m_permissions:CheckPermission(
				"provisioning",
				actionId,
				fnToExecute
			)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix

	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- VCC i6
	module:dispatch_post(fn .. Search, p .. "/search/")
	module:dispatch_get (HasPermission(GetVCC, "view_vccs"), p .. "/get/(.+)")
	module:dispatch_post(
		HasPermission(
			UpdateVCC,
			"save_vcc",
			fn,
			m_validations.basic
		),
		p .. "/save/"
	)
	module:dispatch_post(HasPermission(EraseVCC, "delete_vcc", fn), p .. "/delete/(.*)")

	-- VCC i6 get GetMaxNumberConfigurationsAvailableByVcc
	module:dispatch_get(HasPermission(GetMaxNumberConfigurationsAvailableByVcc, "view_vccs"), p .. "/get_max_configs_available/")

	-- VCC i6 - cognitive services (BOT RASA UPDATE) (VCC BoTInfo Enabled)
	module:dispatch_post(HasPermission(rasa.UpdateVCC           , "save_vcc", fn), p .. "/cognitive_services/rasa/update/")

	-- VCC i6 - cognitive services (BOT RASA import LUIS App)
	module:dispatch_post(HasPermission(rasa.ImportLuisApp       , "save_vcc", fn), p .. "/cognitive_services/rasa/import/luis/app/")

	-- VCC i6 - cognitive services (BOT RASA - load unimported LUIS Apps to RASA)
	module:dispatch_post(HasPermission(rasa.LoadUnimportedLuisApps, "save_vcc", fn), p .. "/cognitive_services/rasa/import/luis/load_unimported_apps/")

	-- VCC i6 - cognitive services (BOT LUIS UPDATE) (VCC BoTInfo Enabled, demo subscription key + azure account & subscriptions + tenant accounts)
	module:dispatch_post(HasPermission(luis.UpdateVCC           , "save_vcc", fn), p .. "/cognitive_services/luis/update/")

	-- VCC i6 - cognitive services (LLM profiles)
	module:dispatch_post(HasPermission(llm.UpdateVCC           , "save_vcc", fn), p .. "/cognitive_services/llm/update/")

	-- VCC i6 - cognitive services (QnA) PAGA subscription key
	module:dispatch_post(HasPermission(qna.UpdateVCC            , "save_vcc", fn), p .. "/cognitive_services/qna/update/" )

	-- VCC i6 - packages
	module:dispatch_post(HasPermission(packages.UpdateVCC       , "save_vcc", fn), p .. "/packages/update/"               )
	module:dispatch_get (replyFn .. packages.LoadAvailablesVccPackages           , p .. "/packages/load/([^/]+)/"         )

	-- VCC i6 - peers
	module:dispatch_post(HasPermission(peers.UpdateVCC          , "save_vcc", fn), p .. "/peers/update/"                  )
	module:dispatch_post(replyFn .. peers.ResetPeersInRedis                      , p .. "/peers/reset/"                   )
	
	module:dispatch_get (replyFn .. peers.LoadAvailablesVccPeers                 , p .. "/peers/load/([^/]+)/"            )

	-- VCC i6 - TTS/ASR
	module:dispatch_post(HasPermission(UpdateVCCTtsAsr           , "save_vcc", fn), p .. "/ttsasr/update/"                  )

	-- VCC i6 - accounts
	module:dispatch_get (replyFn .. accounts.LoadAvailablesVccAccounts           , p .. "/accounts/load/([^/]+)/" )

	-- VCC i6 - integrations helpdesk
	module:dispatch_post(HasPermission(m_integrations.helpdesk.Synchronize, "save_vcc", fn), p .. "/integrations/helpdesk/synchronize/")

	-- VCC i6 - usergroups
	module:dispatch_get (replyFn .. groups.LoadVccGroups           , p .. "/groups/load/([^/]+)/" )

	-- VCC i6 - load ldap tree
	module:dispatch_post(HasPermission(ldap.GetLdapGroups, "save_vcc", fn), p .. "/ldap/getldapgroups/")

	-- VCC i6 - load azuread groups
	module:dispatch_post(HasPermission(azuread.GetAzureADGroups, "save_vcc", fn), p .. "/azuread/getazureadgroups/")

	-- VCC i6 - campaigns
	module:dispatch_get (replyFn .. campaigns.LoadVccCampaigns           , p .. "/campaigns/load/([^/]+)/" )

	-- VCC i6 - defaults recording configuration
	module:dispatch_get (replyFn .. recording_configuration.LoadDefaults           , p .. "/recording_configuration/load_defaults/" )

	-- License
	module:dispatch_get (HasPermission(GetAvailableLicenseResources, "view_vccs"), p .. "/get_available_license_resources/")
	module:dispatch_get (HasPermission(CompanyIsEditable, "view_vccs"), p .. "/company_is_editable/")
	module:dispatch_get (HasPermission(CompanyIsRequired, "view_vccs"), p .. "/company_is_required/")
	module:dispatch_get (HasPermission(ValidateCompany, "save_vcc"), p .. "/validate_company/(.+)")

	--repository service
	module:dispatch_get (replyFn .. GetS3Regions           , p .. "/load_s3_regions/" )

	-- VCC i6 - migration allegro vcc to native
	--
	-- (**) endpoints sincronicos (api) desde el cual se accede desde la web o consola
	module:dispatch_post(HasPermission(m_migrations.allegro.Start   , "save_vcc", fn)     , p .. "/migrations/allegro/to/native/"      )
	module:dispatch_post(HasPermission(m_migrations.allegro.Rollback, "save_vcc", fn)     , p .. "/migrations/allegro/rollback/"       )
	-- endpoints asincronicos desde el cual se accede desde los endpoints sincronicos (**)
	module:dispatch_post(HasPermission(m_migrations.allegro.StartAsync   , "save_vcc", fn), p .. "/migrations/allegro/to/native/async/")
	module:dispatch_post(HasPermission(m_migrations.allegro.RollbackAsync, "save_vcc", fn), p .. "/migrations/allegro/rollback/async/" )
	-- Status of Migration or Rollbacks
	module:dispatch_get (replyFn .. m_migrations.allegro.Status                           , p .. "/migrations/allegro/running/status/" )

	-- VCC i6 - Super Users
	module:dispatch_post(HasPermission(m_superUsers.Save   , "save_vcc", fn)              , p .. "/super_users/save/"                  )
	module:dispatch_post(fn .. m_superUsers.Search                                        , p .. "/super_users/search/"                )
	--
	m_htmlRender = htmlRender
	m_appLocation = appLocation
end
