module(..., package.seeall)

local utils = require "shared.apps.utils"
local PM = require("provisioningmanager")
local m_datasource

function GetName ()
	return "Routes"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

function ListAll (ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local success, routes = pcall(PM.ListRoutes, ksProv)
	--LogDebug("[provisioning:models.routes.ListAll]--> success: '%s', Routes: %s", success, json.encode(routes or {}))

	if not success then
		local errMsg = routes
		LogWarning("[provisioning:models.routes.ListAll] Error retrieving routes  -> '%s'", errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end
	routes = routes or {}

	for id, route in pairs(routes) do
		local routeId = utils.GetId(id)
		route.Id = routeId
	end

	return success, routes
end

function Get (routeId, ksProv)
	LogDebug("[provisioning:models.routes.Get] Retrieving ROUTE: '%s'.", routeId)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local success, currentRoute = pcall(PM.GetRoute, ksProv, routeId)
	--LogDebug("[provisioning:models.routes.Get] Retrieved ROUTE '%s' -> success: '%s', currentRoute: '%s'.", routeId, success, json.encode(currentRoute or {}))

	if releaseConn then
		ksProv:Close()
	end

	if not success or utils.TableIsEmpty(currentRoute) then
		local errMsg = "ROUTES DOES NOT EXIST"
		--LogWarning("[provisioning:models.routes.Get] Error retrieving ROUTES '%s'. ROUTES does not exist.'", routeId)
		return false, errMsg
	end

	currentRoute.Id = currentRoute.Id or routeId

	return success, currentRoute
end

function Save (routeId, routeInfo, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection( "Provisioning" )
		releaseConn = true
	end

	local success, err = PM.UpdateRoute(ksProv, routeId, routeInfo)
	if not success then
		LogWarning("[provisioning:models.routes.Save] Error saving routes '%s'. --> '%s'", routeId, json.encode(err))
	end

	if releaseConn then
		ksProv:Close()
	end

	return success, err
end

function SaveSchedulers (routeId, schedulers, ksProv)
	if not schedulers then
		return true
	end

	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection( "Provisioning" )
		releaseConn = true
	end

	-- utilizado para realizar rollback
	local currentsRoutesSchedulers = PM.GetRawRouteSchedulers(ksProv, routeId)
	--LogDebug("[provisioning:models.routes.SaveSchedulers] ---> routeId: '%s', currentsRoutesSchedulers: '%s'", routeId, json.encode(currentsRoutesSchedulers))

	-- primero los borro a todos antes de proceder a agregar. Si falla realizo rollback
	PM.DeleteRouteAllSchedules(ksProv, routeId)

	local success = true
	local err

	for _, schedulerSlotData in pairs(schedulers) do
		--LogDebug("[provisioning:models.routes.SaveSchedulers] routeId: '%s', key: '%s'. --> schedulerSlotData: '%s'", routeId, _, json.encode(schedulerSlotData))
		success, err = PM.AddRouteScheduleSlot(ksProv, routeId, schedulerSlotData)
		if not success then
			LogWarning("[provisioning:models.routes.SaveSchedulers] Error saving routes schedulers '%s'. --> '%s'", routeId, json.encode(err))
			break
		end
	end

	if not success then
		-- rollback
		LogInfo("[provisioning:models.routes.SaveSchedulers] RollBack Schedulers routeId: '%s' --> schedulers: '%s'", routeId, json.encode(currentsRoutesSchedulers))
		PM.InsertRawRouteSchedulers(ksProv, routeId, currentsRoutesSchedulers)
	end

	if releaseConn then
		ksProv:Close()
	end

	return success, err
end

function Delete (routeId, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection( "Provisioning" )
		releaseConn = true
	end
	local success, result = pcall(PM.DeleteRoute, ksProv, routeId)
	if not success or not result.success then
		LogWarning("[provisioning:models.routes.Delete] Error deleting routes '%s' --> '%s'", routeId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function LoadRoutesSchedulers (ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local success, routesSchedulers = pcall(PM.LoadRawRoutesSchedulers, ksProv)
	--LogDebug("[provisioning:models.routes.LoadRoutesSchedulers]--> success: '%s', routesSchedulers: %s", success, json.encode(routesSchedulers or {}))

	if not success then
		local errMsg = routesSchedulers
		LogWarning("[provisioning:models.routes.LoadRoutesSchedulers] Error retrieving routes  -> '%s'", errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end
	routesSchedulers = routesSchedulers or {}

	return success, routesSchedulers
end
