--[[
-- @file csv_exporter.lua
--
-- @brief Exporter plugin which can handle CSV files
--]]

-- UTF-8's BOM header
local UTF8_BOM = string.char(0xEF, 0xBB, 0xBF);

-- Plugin interfaces (defined later)
local exporter_interface;

-- class csv_exporter
local csv_exporter = {}
local __csv_exporter_mt = {}
__csv_exporter_mt.__index = csv_exporter

-- just opens a new handle
function csv_exporter:new()
	return setmetatable({
		interface = exporter_interface;
	}, __csv_exporter_mt);
end

-- writes the headers
function csv_exporter:write_headers(headers, callback)
	return self:write_row(headers, function(data)
		return callback( UTF8_BOM .. data);
	end);
end

-- writes a row
function csv_exporter:write_row(row, callback)

	local compliant_row = {};
	for _, cell in pairs(row) do
		if  type(cell) == "string" then
			cell = tostring(cell):gsub('"', '""'); -- Escape double quote replacing it with double double quote.
			cell = cell:gsub("[%c]", "");          -- Remove control characters from the string.
		end
		table.insert(compliant_row, '"'.. tostring(cell) ..'"');
	end

	return callback(table.concat(compliant_row, ',') .. '\r\n');
end

-- writes a lot of rows
function csv_exporter:bulk_write(rows, callback)
	local csv_rows = {};

	local on_row_written = function(row_string)
		table.insert(csv_rows, row_string);
	end

	for _, row in pairs(rows) do
		self:write_row(row, on_row_written);
	end

	return callback(table.concat(csv_rows));
end

-- does nothing, but it is required
function csv_exporter:close(callback)
	return callback(nil); -- nothing to write
end

--[[
-- Defines the interface required by the app
--]]
exporter_interface = {
	-- the type of file this exporter can handle
	type = "csv",

	practicable = true,

	content_type = "text/csv",

	-- indicates this file can be streamed on the network
	streamable = true,

	-- factory function used to create new exporters
	-- path is an interface requirement, we do not use it
    factory = function(report, _t, path)
	    --LogDebug("[REPORTS:csv_exporter:exporter_interface.factory]---> report: '%s', _t: '%s', path: '%s'", report, _t, path)
		return csv_exporter:new();
	end
};

return exporter_interface;
