--[[
-- @file pdf_exporter.lua
--
-- @brief Exporter plugin which can handle PDF files
--]]

require "hpdf"
local ormDate = require "LuaOrm.Date"
local utils = require "shared.apps.utils"

local m_T = nil
local function _T (s)
    if not m_T then
        return s
    end
    return m_T(s)
end
local function SetTranslatorFn (_t)
    m_T = _t
end

local function _TField (value)
    value = tostring(value)
    -- Ej: "Corte: "
    return string.format("%s%s ", _T(value), _T(":"))
end

local function clone(original)
    if type(original) ~= "table" then
        return original
    end

    local cloned = {}

    for key, value in pairs(original) do
        cloned[clone(key)] = clone(value)
    end
    
    return cloned
end

local function millimetersToDots(millimeters, dpi)
    dpi = dpi or 72
    local oneDotInInches = 1 / dpi
    local oneDotInMillimeters = oneDotInInches * 25.4
    return millimeters / oneDotInMillimeters
end

local function toScaleOfOne(num, maxNum)
    maxNum = maxNum or 255
    return num / maxNum
end

-- Plugin interfaces (defined later)
local exporter_interface;

-- class pdf_exporter
local pdf_exporter = {
    DEFAULT_MARGINS = {LEFT = millimetersToDots(7.0), TOP = millimetersToDots(7.0), RIGHT = millimetersToDots(7.0), BOTTOM = millimetersToDots(7.0)},
    DEFAULT_PAGE = {
        WIDTH = millimetersToDots(355.6),   -- US Legal in
        HEIGHT = millimetersToDots(215.9),  -- landscape.
        BACKGROUND_COLOR = {RED = toScaleOfOne(255), GREEN = toScaleOfOne(255), BLUE = toScaleOfOne(255)},  -- white
    },
    DEFAULT_FONT = {
        TRUE_TYPE_FILE = "/usr/share/fonts/truetype/ttf-dejavu/DejaVuSansMono.ttf",
		TRUE_TYPE_FILE_JESSIE = "/usr/share/fonts/truetype/dejavu/DejaVuSansMono.ttf",
        BUILTIN_BASE14 = "Courier", 
        TRUE_TYPE_ENCODING = "UTF-8",
        ENCODING = "StandardEncoding",
        SIZE = millimetersToDots(3.0),
        COLOR = {RED = toScaleOfOne(0), GREEN = toScaleOfOne(0), BLUE = toScaleOfOne(0)}  -- black
    },
    DEFAULT_LINE_SPACING = millimetersToDots(1.5),
    DEFAULT_COLUMN_SPACING = millimetersToDots(3.0),
    DEFAULT_HEADER_FONT_COLOR = {RED = toScaleOfOne(0), GREEN = toScaleOfOne(0), BLUE = toScaleOfOne(0)},  -- black
    DEFAULT_ODD_ROW = {
        BACKGROUND_COLOR = {RED = toScaleOfOne(235), GREEN = toScaleOfOne(235), BLUE = toScaleOfOne(235)},  -- lightgray
        HIGHLIGHT = false,
    },
}
local __pdf_exporter_mt = {}
__pdf_exporter_mt.__index = pdf_exporter

-- just opens a new handle
function pdf_exporter:new(report)
    local config = report.extraExportArgs and report.extraExportArgs.pdf or {}
    
    self = {}
    setmetatable(self, __pdf_exporter_mt);

    self.interface = exporter_interface
    
    self._report = report
    self._margins = {
        left = config.margins and config.margins.left and millimetersToDots(config.margins.left) or pdf_exporter.DEFAULT_MARGINS.LEFT, 
        top = config.margins and config.margins.top and millimetersToDots(config.margins.top) or pdf_exporter.DEFAULT_MARGINS.TOP, 
        right = config.margins and config.margins.right and millimetersToDots(config.margins.right) or pdf_exporter.DEFAULT_MARGINS.RIGHT, 
        bottom = config.margins and config.margins.bottom and millimetersToDots(config.margins.bottom) or pdf_exporter.DEFAULT_MARGINS.BOTTOM,
    }
    self._page = {
        width = config.page and config.page.width and millimetersToDots(config.page.width) or pdf_exporter.DEFAULT_PAGE.WIDTH,
        height = config.page and config.page.height and millimetersToDots(config.page.height) or pdf_exporter.DEFAULT_PAGE.HEIGHT,
        backgroundColor = {
            red = config.page and config.page.backgroundColor and config.page.backgroundColor.red and toScaleOfOne(config.page.backgroundColor.red) or 
                        pdf_exporter.DEFAULT_PAGE.BACKGROUND_COLOR.RED,
            green = config.page and config.page.backgroundColor and config.page.backgroundColor.green and toScaleOfOne(config.page.backgroundColor.green) or 
                        pdf_exporter.DEFAULT_PAGE.BACKGROUND_COLOR.GREEN,
            blue = config.page and config.page.backgroundColor and config.page.backgroundColor.blue and toScaleOfOne(config.page.backgroundColor.blue) or 
                        pdf_exporter.DEFAULT_PAGE.BACKGROUND_COLOR.BLUE,
        },
    }
    self._font = {
        -- TODO: accept font family and encoding
        size = config.font and config.font.size and millimetersToDots(config.font.size) or pdf_exporter.DEFAULT_FONT.SIZE,
        color = {
            red = config.font and config.font.color and config.font.color.red and toScaleOfOne(config.font.color.red) or pdf_exporter.DEFAULT_FONT.COLOR.RED,
            green = config.font and config.font.color and config.font.color.green and toScaleOfOne(config.font.color.green) or pdf_exporter.DEFAULT_FONT.COLOR.GREEN,
            blue = config.font and config.font.color and config.font.color.blue and toScaleOfOne(config.font.color.blue) or pdf_exporter.DEFAULT_FONT.COLOR.BLUE,
        },
    }
    self._lineSpacing = config.lineSpacing and millimetersToDots(config.lineSpacing) or pdf_exporter.DEFAULT_LINE_SPACING
    self._columnSpacing = config.columnSpacing and millimetersToDots(config.columnSpacing) or pdf_exporter.DEFAULT_COLUMN_SPACING
    
    --[[
            self._columnWidths = {
                -1,                                     -- Minimun width without overflow plus free space (e.g expanded).
                50,                                     -- Fixed width of 50.
                false,                              -- Remove this column.
                30,                                     -- Fixed width of 30.
                
                nil,                                    -- Not specified is equivalent to -1, that is
                nil,                                    -- minimun width without overflow plus free space (e.g expanded).
            }
    ]]
    -- TODO: allow the customization of which columns must be expanded to fill free space
    local columnWidthsInMillimiters = config.columnWidths or {}
    self._columnWidths = {}
    
    for index, width in pairs(columnWidthsInMillimiters) do
        if type(width) == "number" and width ~= -1 then  -- fixed width
            self._columnWidths[index] = millimetersToDots(width)
        else
            self._columnWidths[index] = width
        end
    end
    
    self._headerFontColor = {
        red = config.headerFontColor and config.headerFontColor.red and toScaleOfOne(config.headerFontColor.red) or pdf_exporter.DEFAULT_HEADER_FONT_COLOR.RED,
        green = config.headerFontColor and config.headerFontColor.green and toScaleOfOne(config.headerFontColor.green) or pdf_exporter.DEFAULT_HEADER_FONT_COLOR.GREEN,
        blue = config.headerFontColor and config.headerFontColor.blue and toScaleOfOne(config.headerFontColor.blue) or pdf_exporter.DEFAULT_HEADER_FONT_COLOR.BLUE,
    }
    -- TODO: this must be another configuration option
    self._headerBackgroundColor = {
        red = toScaleOfOne(192),
        green = toScaleOfOne(192),
        blue = toScaleOfOne(192),
    }
    self._oddRow = {
        backgroundColor = {
            red = config.oddRow and config.oddRow.backgroundColor and config.oddRow.backgroundColor.red and toScaleOfOne(config.oddRow.backgroundColor.red) or 
                        pdf_exporter.DEFAULT_ODD_ROW.BACKGROUND_COLOR.RED,
            green = config.oddRow and config.oddRow.backgroundColor and config.oddRow.backgroundColor.green and toScaleOfOne(config.oddRow.backgroundColor.green) or 
                            pdf_exporter.DEFAULT_ODD_ROW.BACKGROUND_COLOR.GREEN,
            blue = config.oddRow and config.oddRow.backgroundColor and config.oddRow.backgroundColor.blue and toScaleOfOne(config.oddRow.backgroundColor.blue) or 
                        pdf_exporter.DEFAULT_ODD_ROW.BACKGROUND_COLOR.BLUE,
        },
        -- highlight = <boolean>  setted below
    }
    
    if config.oddRow and config.oddRow.highlight ~= nil then
        self._oddRow.highlight = config.oddRow.highlight
    else
        self._oddRow.highlight = pdf_exporter.DEFAULT_ODD_ROW.HIGHLIGHT
    end
    
    self._pageHandlers = {}
    self._headers = nil
    -- TODO
    -- Hack until column types can be determined in standard way
    self._columnTypes = nil
    
    --[[
            self._reportTable will be populated with all data report, for example, 
            it could finally contain the following data:
            
            self._reportTable = {
                {type = "headers", value = {"Campaign", "Agent", "Start Date"}},
                {type = "row", value = {"Support", "Pepe", "2016-03-16 12:35:20"}},
                ...
                {type = "row", value = {"Sale", "Juan", "2016-03-18 09:40:23"}},
            }
    ]]
    self._reportTable = {}
    self._minColumnWidths = {}  -- computed minimum column widths that not cause overflow
    self._finalColumnWidths = {}  -- computed final column widths
    self._totalColumnsWidth = nil  -- total columns width plus spacings between columns (self._columnSpacing) after computed final column widths
    self._currentYCoord = nil
    self._bottomLineYCoord = self._margins.bottom + self._font.size + self._lineSpacing
    self._currentTimestamp = os.time()
    
    self._pdfHandler = hpdf.New()
    if not self._pdfHandler then
        error("Could not create a PDF object")
    end
    
    hpdf.UseUTFEncodings(self._pdfHandler)
    local success, fontNameOrErrMsg
    success, fontNameOrErrMsg = pcall(hpdf.LoadTTFontFromFile, self._pdfHandler, pdf_exporter.DEFAULT_FONT.TRUE_TYPE_FILE, true)
    if success then
        self._fontHandler = hpdf.GetFont(self._pdfHandler, fontNameOrErrMsg, pdf_exporter.DEFAULT_FONT.TRUE_TYPE_ENCODING)
        
        if not self._fontHandler then
            LogWarning("pdf_exporter:new - Failed to get font '%s' with enconding '%s'. Trying with '%s' and encoding '%s'...", 
                                    pdf_exporter.DEFAULT_FONT.TRUE_TYPE_FILE, pdf_exporter.DEFAULT_FONT.TRUE_TYPE_ENCODING, 
                                    pdf_exporter.DEFAULT_FONT.BUILTIN_BASE14, pdf_exporter.DEFAULT_FONT.ENCODING)
            self._fontHandler = hpdf.GetFont(self._pdfHandler, pdf_exporter.DEFAULT_FONT.BUILTIN_BASE14, pdf_exporter.DEFAULT_FONT.ENCODING)
        end
    else
		LogWarning("pdf_exporter:new - Failed to load true type font from file '%s'. Trying with '%s'...", 
                                pdf_exporter.DEFAULT_FONT.TRUE_TYPE_FILE, 
                                pdf_exporter.DEFAULT_FONT.TRUE_TYPE_FILE_JESSIE)
		
		hpdf.Free(self._pdfHandler)
		self._pdfHandler = nil
		self._pdfHandler = hpdf.New()
	    if not self._pdfHandler then
			error("Could not create a PDF object")
		end
		hpdf.UseUTFEncodings(self._pdfHandler)
		
		--Utilizo el path de Debian-Jessie
		success, fontNameOrErrMsg = pcall(hpdf.LoadTTFontFromFile, self._pdfHandler, pdf_exporter.DEFAULT_FONT.TRUE_TYPE_FILE_JESSIE, true)
		if success then
			 self._fontHandler = hpdf.GetFont(self._pdfHandler, fontNameOrErrMsg, pdf_exporter.DEFAULT_FONT.TRUE_TYPE_ENCODING)
        
			if not self._fontHandler then
				LogWarning("pdf_exporter:new - Failed to get font '%s' with enconding '%s'. Trying with '%s' and encoding '%s'...", 
										pdf_exporter.DEFAULT_FONT.TRUE_TYPE_FILE, pdf_exporter.DEFAULT_FONT.TRUE_TYPE_ENCODING, 
										pdf_exporter.DEFAULT_FONT.BUILTIN_BASE14, pdf_exporter.DEFAULT_FONT.ENCODING)
				self._fontHandler = hpdf.GetFont(self._pdfHandler, pdf_exporter.DEFAULT_FONT.BUILTIN_BASE14, pdf_exporter.DEFAULT_FONT.ENCODING)
			end
		else
			LogWarning("pdf_exporter:new - Failed to load true type font from file '%s'. Trying with '%s' and encoding '%s'...", 
									pdf_exporter.DEFAULT_FONT.TRUE_TYPE_FILE_JESSIE, 
									pdf_exporter.DEFAULT_FONT.BUILTIN_BASE14, pdf_exporter.DEFAULT_FONT.ENCODING)

			hpdf.Free(self._pdfHandler)
			self._pdfHandler = nil
			self._pdfHandler = hpdf.New()
		    if not self._pdfHandler then
				error("Could not create a PDF object")
			end
			hpdf.UseUTFEncodings(self._pdfHandler)
			self._fontHandler = hpdf.GetFont(self._pdfHandler, pdf_exporter.DEFAULT_FONT.BUILTIN_BASE14, pdf_exporter.DEFAULT_FONT.ENCODING)
		end
        
    end
    
    if not self._fontHandler then
        error(("Could not get font '%s' with enconding '%s'"):format(pdf_exporter.DEFAULT_FONT.BUILTIN_BASE14, pdf_exporter.DEFAULT_FONT.ENCODING))
    end
    
    self:_create_new_page()
    return self
end

-- writes the headers
function pdf_exporter:write_headers(headers, callback)
	--LogDebug("[REPORTS:pdf_exporter:write_headers]---> headers: '%s'", json.encode(headers or {}))
    table.insert(self._reportTable, {type = "headers", value = clone(headers)})
    callback(nil)
end

-- writes a row
function pdf_exporter:write_row(row, callback)
    table.insert(self._reportTable, {type = "row", value = clone(row)})
    
    -- TODO
    -- Hack until column types can be determined in standard way
    if not self._columnTypes then
        self._columnTypes = {}
        
        for cellIndex, cellValue in ipairs(row) do
            cellValue = tostring(cellValue)
            
            if cellValue:match("^[%d,.%s]*%d[%d,.%s]*%%?%s*$") then
                table.insert(self._columnTypes, "number")
            else
                table.insert(self._columnTypes, "string")
            end
        end
    end
    
    callback(nil)
end

-- writes a lot of rows
function pdf_exporter:bulk_write(rows, callback)
    for index, row in ipairs(rows) do
        self:write_row(row, function(data) end)
    end
    
    callback(nil)
end

-- close
function pdf_exporter:close(callback)
    callback(self:_export())
end

function pdf_exporter:_create_new_page()
	local _  -- ignored value
	local pageHandler = hpdf.AddPage(self._pdfHandler)  -- TODO: check that 'pageHandler' is valid
	table.insert(self._pageHandlers, pageHandler)

	-- reset Y coordinate
	self._currentYCoord = self._page.height - self._margins.top

	-- TODO: set page background color
	hpdf.Page_SetFontAndSize(pageHandler, self._fontHandler, self._font.size)
	hpdf.Page_SetWidth(pageHandler, self._page.width)
	hpdf.Page_SetHeight(pageHandler, self._page.height)
  
	-- report name

	local titleFontSize = self._font.size * 1.8
	self._currentYCoord = self._currentYCoord - titleFontSize

	_, self._currentYCoord = self:_print(self._report.label, {
		fontSize = titleFontSize,
		-- to avoid report description to go too down because of `titleFontSize'
		lineHeight = self._font.size,
	})

	-- report description

	_, self._currentYCoord = self:_print(self._report.description)
	-- put extra space below description
	self._currentYCoord = self._currentYCoord - (2 * self._lineSpacing)

	-- filters

	local filtersTable = {}
	local dateFrom
	local dateTo
	local period

	local tabulate = function (tableData)
		local columnWidths = {}

		for rowIndex, row in ipairs(tableData) do
			  for cellIndex, cell in ipairs(row) do
				  local cellWidth = hpdf.Font_TextWidth(self._fontHandler, cell, 
														  cell:len()).width * self._font.size / 1000

				  if not columnWidths[cellIndex] or columnWidths[cellIndex] < cellWidth then
					  columnWidths[cellIndex] = cellWidth
				  end
			  end
		end

		for rowIndex, row in ipairs(tableData) do
			local x = self._margins.left

				for cellIndex, cell in ipairs(row) do
					x, _ = self:_print(cell, {
						x = x,
						xInc = columnWidths[cellIndex],
						endLine = false,
					})
				end

				self._currentYCoord = self._currentYCoord - self._lineSpacing - self._font.size
		end
	end

	if self._report.type == "detail" or self._report.type == "livedetail" then
		for filterName, filterValue in pairs(self._report.queryFields) do
			if filterName ~= "scheduleId" and filterName ~= "sliceSize" and 
				filterName ~= "withContent" and filterName ~= "saveContentTo" then -- avoid this fields for when it is an schedule
				table.insert(filtersTable, {_TField(filterName), utils.ConvertValueToString(filterValue)})
			end
		end

		-- period

		dateFrom = self._report.startDate or ""
		dateTo = self._report.endDate or ""

		table.insert(filtersTable, {_TField("Period"), dateFrom.." - "..dateTo})
		tabulate(filtersTable)
	elseif self._report.type == "summary" then
		-- slice

		table.insert(filtersTable, {_TField("Slice"), self._report.sliceSize})

		-- other filters except `DateFrom' and `DateTo'

		for filterName, filterValue in pairs(self._report.filter) do
			if filterName ~= "startDate" and filterName ~= "endDate" and 
				filterName ~= "withContent" then  -- avoid this fields for when it is an schedule

				if filterName ~= "DateFrom" and filterName ~= "DateTo" then
					table.insert(filtersTable, {_TField(filterName), utils.ConvertValueToString(filterValue)})
				end
			end
		end

		-- period (`DateFrom' - `DateTo')

		dateFrom = self._report.filter.DateFrom or ""
		dateTo = self._report.filter.DateTo or ""

		table.insert(filtersTable, {_TField("Period"), dateFrom.." - "..dateTo})
		tabulate(filtersTable)
	elseif self._report.type == "realtime" then
		-- campaigns

		local campaignNames = ""
		local campaigns =  self._report.campaigns or {}
		for campaignIndex, campaignName in ipairs(campaigns) do
			campaignNames = campaignNames..utils.ConvertValueToString(campaignName)
			if campaignIndex < #campaigns then
				campaignNames = campaignNames..", "
			end
		end

		table.insert(filtersTable, {_TField("Campaigns"), campaignNames})

		-- period

		dateFrom = self._report.startDate or ""
		dateTo = self._report.endDate or ""

		table.insert(filtersTable, {_TField("Period"), dateFrom.." - "..dateTo})
		tabulate(filtersTable)
	end
  
	-- report generation date

	local generationDate = _TField("Report generation date") .. (ormDate(self._currentTimestamp):
		  toDifferentTimezoneString(self._report.timezone, "%Y-%m-%d %H:%M:%S"))
	local generationDateWidth = hpdf.Font_TextWidth(self._fontHandler,
		generationDate, generationDate:len()).width * self._font.size / 1000

	-- up one line to be in the same line as the last filter
	self._currentYCoord = self._currentYCoord + self._lineSpacing + self._font.size

	_, self._currentYCoord = self:_print(generationDate, {
		x = self._page.width - self._margins.right - generationDateWidth,
	})

	-- put extra space below filters
	self._currentYCoord = self._currentYCoord - self._lineSpacing

	-- headers
	if self._headers then
		self:_write_headers(self._headers, true, false --[[ to avoid posible infinite recursion due to self:_check_page_overflow() ]]
		)
	end

	-- page number and bottom line

	local pageNum = tostring(#self._pageHandlers)
	local pageLiteral = "Page "
	local pageNumWidth = hpdf.Font_TextWidth(self._fontHandler, pageNum, pageNum:len()).width * self._font.size / 1000
	local maxPageNumWidth = hpdf.Font_TextWidth(self._fontHandler, "9999", ("9999"):len()).width * self._font.size / 1000
	local pageLiteralWidth = hpdf.Font_TextWidth(self._fontHandler, pageLiteral, pageLiteral:len()).width * self._font.size / 1000

	self:_print(pageNum, {
		x = self._page.width - self._margins.right - pageNumWidth,
		y = self._margins.bottom,
	})

	self:_print(pageLiteral, {
		x = self._page.width - self._margins.right - maxPageNumWidth - pageLiteralWidth,
		y = self._margins.bottom,
	})

	self:_draw_line(self._margins.left, self._bottomLineYCoord, 
		self._page.width - self._margins.right, self._bottomLineYCoord)

	return pageHandler
end

function pdf_exporter:_check_page_overflow()
    if self._currentYCoord < self._bottomLineYCoord then
        self:_create_new_page()
    end
end

function pdf_exporter:_export()
    self:_compute_min_column_widths()
    self:_compute_final_column_widths()
    self:_make()
    local data = self:_get_data()
    self:_close()
    return data
end

function pdf_exporter:_compute_min_column_widths()
    for rowIndex, row in ipairs(self._reportTable) do
        for cellIndex, cellValue in ipairs(row.value) do
            cellValue = tostring(cellValue)
            local cellWidth
            
            if row.type == "headers" then
                local maxCellLineWidth = 0
                
                for cellLineIndex, cellLineText in ipairs(self:_get_word_wrapped_text(cellValue, 0)) do
                    local cellLineWidth = hpdf.Font_TextWidth(self._fontHandler, cellLineText, cellLineText:len()).width * self._font.size / 1000
                    
                    if cellLineWidth > maxCellLineWidth then
                        maxCellLineWidth = cellLineWidth
                    end
                end
                
                cellWidth = maxCellLineWidth
            else  -- row.type = "row"
                cellWidth = hpdf.Font_TextWidth(self._fontHandler, cellValue, cellValue:len()).width * self._font.size / 1000
            end
            
            if not self._minColumnWidths[cellIndex] or cellWidth > self._minColumnWidths[cellIndex] then
                self._minColumnWidths[cellIndex] = cellWidth
            end
        end
    end
end

function pdf_exporter:_compute_final_column_widths(notFillPage)
    if #self._reportTable == 0 then
        return
    end

    local firstRow = self._reportTable[1].value
    self._totalColumnsWidth = 0
    local totalColumnsWithDefaultWidth = 0
    
    for columnIndex = 1, #firstRow do
        if self._columnWidths[columnIndex] == nil or self._columnWidths[columnIndex] == -1 then
            self._finalColumnWidths[columnIndex] = self._minColumnWidths[columnIndex] or 0
            self._totalColumnsWidth = self._totalColumnsWidth + (self._minColumnWidths[columnIndex] or 0) + self._columnSpacing
            totalColumnsWithDefaultWidth = totalColumnsWithDefaultWidth + 1
        elseif self._columnWidths[columnIndex] ~= false then
            self._finalColumnWidths[columnIndex] = self._columnWidths[columnIndex]
            self._totalColumnsWidth = self._totalColumnsWidth + self._columnWidths[columnIndex] + self._columnSpacing
        end
    end
    
    if self._totalColumnsWidth > 0 then
        self._totalColumnsWidth = self._totalColumnsWidth - self._columnSpacing  -- remove column spacing after last column
    end
    
    if notFillPage then
        return
    end
    
    local freeSpace = self._page.width - self._margins.left - self._margins.right - self._totalColumnsWidth
    
    if totalColumnsWithDefaultWidth > 0 and freeSpace > 0 then
        local extraColumnWidthToFillPage = freeSpace / totalColumnsWithDefaultWidth
        
        for columnIndex = 1, #firstRow do
            if self._columnWidths[columnIndex] == nil or self._columnWidths[columnIndex] == -1 then
                self._finalColumnWidths[columnIndex] = self._finalColumnWidths[columnIndex] + extraColumnWidthToFillPage
            end
        end
        
        self._totalColumnsWidth = self._totalColumnsWidth + freeSpace
    end
end

function pdf_exporter:_make()
    local currentRowIsOdd = true
    
    for rowIndex, row in ipairs(self._reportTable) do
        if row.type == "headers" then
            self._headers = row.value  -- save headers for setting it at each page beginning
            self:_write_headers(row.value, true)
            currentRowIsOdd = true
        else  -- row.type = "row"
            if self._oddRow.highlight and currentRowIsOdd then
                self:_highlight_row(self._oddRow.backgroundColor)
            end
            currentRowIsOdd = not currentRowIsOdd
            
            self:_write_row(row.value, self._font.color, false)
        end
    end
end

function pdf_exporter:_write_headers(headers, wordWrap, checkPageOverflow)
    local linesNum = wordWrap and self:_get_row_lines_num(headers) or 1
    
    self:_highlight_row(self._headerBackgroundColor, linesNum)
    
    local availableContentWidth = self._page.width - self._margins.left - self._margins.right
    local x1 = self._margins.left
    local y1 = self._currentYCoord - self._lineSpacing - 
                        ((linesNum - 1) * (self._font.size + self._lineSpacing))
    local x2 = x1 + (self._totalColumnsWidth < availableContentWidth and self._totalColumnsWidth or 
                                    availableContentWidth)
    local y2 = y1
    
    self:_draw_line(x1, y1, x2, y2)
    
    self:_write_row(headers, self._headerFontColor, wordWrap, checkPageOverflow)
end

function pdf_exporter:_write_row(row, fontColor, wordWrap, checkPageOverflow)
    checkPageOverflow = checkPageOverflow == nil and true or checkPageOverflow
    
    local pageHandler = self._pageHandlers[#self._pageHandlers]
    local currentXCoord = self._margins.left
    -- TODO: simplify this method using pdf_exporter:_get_row_lines_num(row)
    local maxCellLinesNum = 1
    
    hpdf.Page_SetRGBFill(pageHandler, fontColor.red, fontColor.green, fontColor.blue)
    
    for cellIndex, cellValue in ipairs(row) do
        cellValue = tostring(cellValue)
        
        if self._columnWidths[cellIndex] ~= false and currentXCoord < self._page.width - self._margins.right then  -- only show non-removed columns that will be displayed completely or partialy
            hpdf.Page_GSave(pageHandler)
            
            -- determine cell lines
            local cellLines
            
            if wordWrap then
                cellLines = self:_get_word_wrapped_text(cellValue, self._finalColumnWidths[cellIndex])
            else
                cellLines = {cellValue}
            end
            
            if #cellLines > maxCellLinesNum then
                maxCellLinesNum = #cellLines
            end
            
            -- clip cell
            local clipWidth
            local clipHeight = #cellLines * (self._font.size + self._lineSpacing)
            
            if currentXCoord + self._finalColumnWidths[cellIndex] > self._page.width - self._margins.right then  -- if will occur overflow then clip to right margin start
                clipWidth = self._page.width - self._margins.right - currentXCoord
            else
                clipWidth = self._finalColumnWidths[cellIndex]
            end
            
            hpdf.Page_Rectangle(pageHandler,    currentXCoord, 
                                                                                self._currentYCoord - self._lineSpacing  -- one line
                                                                                - clipHeight + self._font.size + self._lineSpacing,  -- clipHeight - one line
                                                                                clipWidth, 
                                                                                clipHeight)
            hpdf.Page_Clip(pageHandler)
            hpdf.Page_EndPath(pageHandler)
            
            -- write cell lines
            hpdf.Page_BeginText(pageHandler)
            hpdf.Page_MoveTextPos(pageHandler, currentXCoord, self._currentYCoord)
            
            local rightAlignment = self._columnTypes ~= nil and 
                                                            self._columnTypes[cellIndex] == "number"
            
            for cellLineIndex, cellLineText in ipairs(cellLines) do
                if rightAlignment then
                    local cellLineTextWidth = hpdf.Font_TextWidth(
                        self._fontHandler, cellLineText, cellLineText:len()).width * 
                                                                                                    self._font.size / 1000
                    local offset = self._finalColumnWidths[cellIndex] - cellLineTextWidth
                    
                    hpdf.Page_MoveTextPos(pageHandler, offset, 0)
                    hpdf.Page_ShowText(pageHandler, cellLineText)
                    hpdf.Page_MoveTextPos(pageHandler, -offset, 0)  -- restore X coord
                else  -- left alignment
                    hpdf.Page_ShowText(pageHandler, cellLineText)
                end
                
                hpdf.Page_MoveTextPos(pageHandler, 0, -self._font.size - self._lineSpacing)
            end
            
            hpdf.Page_EndText(pageHandler)
            
            currentXCoord = currentXCoord + self._finalColumnWidths[cellIndex] + self._columnSpacing
            
            hpdf.Page_GRestore(pageHandler)
        end
    end
  
  self._currentYCoord = self._currentYCoord - maxCellLinesNum * (self._font.size + self._lineSpacing)
  
  if checkPageOverflow then
        self:_check_page_overflow()
    end
end

function pdf_exporter:_get_data()
    local TMP_FILENAME = os.tmpname()
    local saved = hpdf.SaveToFile(self._pdfHandler, TMP_FILENAME) == hpdf.OK
    
    local data
    local readed = false
    
    if saved then
        local tmpFile = io.open(TMP_FILENAME, "rb")
        
        if tmpFile then
            data = tmpFile:read("*all")
            
            if data then
                readed = true
            end
            
            tmpFile:close()
        end
    end
    
    local removed, errmsg, errno = os.remove(TMP_FILENAME)
    
    if not removed then
        LogWarning("pdf_exporter:_get_data - Failed to remove temporary file '%s'", TMP_FILENAME)
    end
    
    if not saved then
        error("Could not save PDF to temporary file")
    end
    
    if not readed then
        error("Could not read temporary PDF file")
    end
    
    return data
end

function pdf_exporter:_close()
    hpdf.Free(self._pdfHandler)
    self._pdfHandler = nil
    
    self._fontHandler = nil
    
    for pageIndex = 1, #self._pageHandlers do
        self._pageHandlers[pageIndex] = nil
    end
end

function pdf_exporter:_highlight_row(color, linesNum)
    linesNum = linesNum == nil and 1 or linesNum
    local pageHandler = self._pageHandlers[#self._pageHandlers]
    local availableContentWidth = self._page.width - self._margins.left - self._margins.right
    
    hpdf.Page_GSave(pageHandler)
    hpdf.Page_SetRGBFill(pageHandler, color.red, color.green, color.blue)
    hpdf.Page_Rectangle(pageHandler,    self._margins.left, 
                                                                        self._currentYCoord - self._lineSpacing - ((linesNum - 1) * (self._font.size + self._lineSpacing)),
                                                                        self._totalColumnsWidth < availableContentWidth and self._totalColumnsWidth or availableContentWidth, 
                                                                        linesNum * (self._font.size + self._lineSpacing))
    hpdf.Page_Fill(pageHandler)
    hpdf.Page_GRestore(pageHandler)
end

function pdf_exporter:_draw_line(x1, y1, x2, y2)
    local pageHandler = self._pageHandlers[#self._pageHandlers]
    
    hpdf.Page_GSave(pageHandler)
    
    hpdf.Page_MoveTo(pageHandler, x1, y1)
    hpdf.Page_LineTo(pageHandler, x2, y2)
    hpdf.Page_Stroke(pageHandler)
    
    hpdf.Page_GRestore(pageHandler)
end

function pdf_exporter:_get_word_wrapped_text(text, width)
    if text == "" then
        return {""}
    end
    
    local lines = {}
    
    while text:len() > 0 do
        local length = hpdf.Font_MeasureText(self._fontHandler, text, text:len(), width, self._font.size, 
                                                                                    0, 0,  -- TODO: handle char_space and word_space
                                                                                    true, nil)
        
        if length > 0 then
            table.insert(lines, text:sub(1, length))
            text = text:sub(length + 1)
        else
            local firstWord = text:match("^(%s*[^%s]*%s*)")
            table.insert(lines, firstWord)
            text = text:sub(firstWord:len() + 1)
        end
    end
    
    return lines
end

function pdf_exporter:_get_row_lines_num(row)
    local maxCellLinesNum = 1
    local currentXCoord = self._margins.left
    
    for cellIndex, cellValue in ipairs(row) do
        cellValue = tostring(cellValue)
        
        -- only analize non-removed columns that will be displayed completely or partialy
        if self._columnWidths[cellIndex] ~= false and currentXCoord < self._page.width - self._margins.right then
            local cellLines = self:_get_word_wrapped_text(cellValue, self._finalColumnWidths[cellIndex])
            
            if #cellLines > maxCellLinesNum then
                maxCellLinesNum = #cellLines
            end
            
            currentXCoord = currentXCoord + self._finalColumnWidths[cellIndex] + self._columnSpacing
        end
    end
    
    return maxCellLinesNum
end

--[[
--  args = {
--      x                           = ...,
--      y                           = ...,
--      endLine             = ...,
--      xInc                    = ...,
--      yDec                    = ...,
--      fontHandler     = ...,
--      fontColor           = ...,
--      fontSize            = ...,
--      lineHeight      = ...,
--      lineSpacing     = ...,
--  }
--
--  return args.x + args.xInc, args.y - args.yDec
--]]
function pdf_exporter:_print(line, args)
    local defaultArgs = {
        x = self._margins.left,
        y = self._currentYCoord,
        fontHandler = self._fontHandler,
        fontSize = self._font.size,
        lineSpacing = self._lineSpacing,
        fontColor = self._font.color,
        endLine = true,
    }
    
    args = args or {}
    args.x = args.x or defaultArgs.x
    args.y = args.y or defaultArgs.y
    args.fontHandler = args.fontHandler or defaultArgs.fontHandler
    args.fontSize = args.fontSize or defaultArgs.fontSize
    args.lineSpacing = args.lineSpacing or defaultArgs.lineSpacing
    args.fontColor = args.fontColor or defaultArgs.fontColor
    args.lineHeight = args.lineHeight or args.fontSize
    
    if args.endLine == nil then
        args.endLine = defaultArgs.endLine
    end
    
    args.xInc = args.xInc or args.endLine and 0 or 
        hpdf.Font_TextWidth(args.fontHandler, line, line:len()).width * 
                                                                                                                    args.fontSize / 1000
    args.yDec = args.yDec or args.endLine and args.lineSpacing + args.lineHeight or 0
    
    local pageHandler = self._pageHandlers[#self._pageHandlers]
    
    hpdf.Page_GSave(pageHandler)
    hpdf.Page_SetFontAndSize(pageHandler, args.fontHandler, args.fontSize)
  hpdf.Page_SetRGBFill(pageHandler, args.fontColor.red, 
                                                                    args.fontColor.green, 
                                                                    args.fontColor.blue)
  
  hpdf.Page_BeginText(pageHandler)
    hpdf.Page_MoveTextPos(pageHandler, args.x, args.y)
    hpdf.Page_ShowText(pageHandler, line)
    hpdf.Page_EndText(pageHandler)
    
    hpdf.Page_GRestore(pageHandler)
    
    return args.x + args.xInc, args.y - args.yDec
end

--[[
-- Defines the interface required by the app
--]]
exporter_interface = {
    -- the type of file this exporter can handle
    type = "pdf",

    practicable = true,

    content_type = "application/pdf",

    streamable = true,

    -- factory function used to create new exporters
    -- path is an interface requirement, we do not use it
    factory = function(report, _t, path)
	    --LogDebug("[REPORTS:pdf_exporter:exporter_interface.factory]---> report: '%s', _t: '%s', path: '%s'", report, _t, path)
        SetTranslatorFn(_t)
        return pdf_exporter:new(report)
    end
};

return exporter_interface;
