--[[
-- @file inconcert/apps/reports/init.lua
-- @brief Main file for application 'reports'
--]]
local orbit = require("orbit")
local inconcert_orbit_application = require("inconcert.orbit.application")
require("inconcert.translators")
require "lua_reporterclient"

local uuid           = require "uuid"
local json           = require "cjson"
local cassandra      = require "lua_cassandra"
local reporter       = require "reporter"
local manager        = require "reporter.manage"
local scheduler      = require "reporter.scheduler"
local wsapi_request  = require "wsapi.request"
local lconfigcontrol = require "LuaConfigurationControl"
local permissions    = require "shared.permissions.permissions"
local authFilter     = (require "authfilter")("agent")
local ir             = require "lua_iragentlink"
local fullcontact    = require "fullcontact"
local utils          = require "shared.apps.utils"
local exporter       = relativeRequire("exporter", ...)
local _              = relativeRequire("functional", ...)
local helpers        = relativeRequire("helpers", ...)
local model          = relativeRequire("model", ...)
local m_outboundEngineClient = nil
local model_processes = nil

local WrapRedis = require "wrapredis"
local redis = nil
local irLink = nil

local module_name = select(1, ...)
local module_dir  = select(2, ...)

module(module_name, package.seeall, orbit.new, inconcert_orbit_application.Decorate)
assert(module_dir, "should be running under inconcertWebHandler")

json.encode_empty_table_as_array("on")

local configControl
local IS_PRODUCTION_ENV = os.getenv("INCONCERT_ENV_MODE") == "PRODUCTION"

local customScopes = {
	cci				= true,
	webchat			= true,
	webcall			= true,
	contactform		= true,
	survey			= true,
	api				= true,
	batch_data		= true,
	process_flow	= true
}

local scopeLabels = {
	api				= "Contact Form API",
	batch_data		= "Batch Message",
	process_flow	= "Process Flow",
}

--Lista de categorias por default
local view_categories = {
	{ name = "CALL",			type = "CHANNEL",	channel = "CALL"		},
	{ name = "VOICEMAIL",		type = "CHANNEL",	channel = "CALL"		},
	{ name = "DIALER",			type = "CHANNEL",	channel = "CALL"		},
	{ name = "CHAT",			type = "CHANNEL",	channel = "CHAT"		},
	{ name = "MAIL",			type = "CHANNEL",	channel = "MAIL"		},
	{ name = "FACEBOOK",		type = "CHANNEL",	channel = "FACEBOOK"	},
	{ name = "INSTAGRAM",		type = "CHANNEL",	channel = "INSTAGRAM"	},
	{ name = "YOUTUBE",			type = "CHANNEL",	channel = "YOUTUBE"		},
	{ name = "APPSTORE",		type = "CHANNEL",	channel = "APPSTORE"	},
	{ name = "GOOGLEPLAY",		type = "CHANNEL",	channel = "GOOGLEPLAY"	},
	{ name = "RCS",				type = "CHANNEL",	channel = "RCS"			},
	{ name = "TWITTER",			type = "CHANNEL",	channel = "TWITTER"		},
	{ name = "WEBCONTACT",		type = "CHANNEL",	channel = "WEBCONTACT"	},
	{ name = "IM",				type = "CHANNEL",	channel = "IM"			},
	{ name = "SMS",				type = "CHANNEL",	channel = "SMS"			},
	{ name = "WHATSAPP",		type = "CHANNEL",	channel = "WHATSAPP"	},
	{ name = "TEAM",			type = "TEAM"								},
	{ name = "MANAGEMENT",		type = "MANAGEMENT"							},
	{ name = "CUSTOM",			type = "CUSTOM"								}
}

local m_allowedReportsTypes = {
	detail     = true,
	livedetail = true,
	summary    = true
}

local m_reporterClientInstance = nil
local function GetReporterClient ()
	if not m_reporterClientInstance then
		m_reporterClientInstance = lua_reporterclient.LuaReporterClient(
			inconcert.environment.apps.agent.cluster, "webhandler")
	end
	return m_reporterClientInstance
end

---
-- Some components expect a lualiblogger compatible logger instance.
-- The webhandler does not provide one, but we have the needed methods
-- to fake one.
--
local function MakeCompatibleLogger ()
	return {
		debug = LogDebug,
		info  = LogInfo,
		warn  = LogWarning,
		error = LogError,
		fatal = LogFatal
	}
end

local function putDiagnosticData (t)
	if not IS_PRODUCTION_ENV then
		t.raw_data = utils.DeepCopy(t)
	end
end

--[[
-- Helper function used to transform a null JSON to nil
--]]
local function getNullable(v)
	if v == json.null then
		return nil
	end
	return v
end

--[[
-- Helper function used in error responses
--]]
local function replyJson(web, status, response)
	if type(response) == "string" then
		response = { result = response }
	end
	web.status = status
	return json.encode(response)
end

--[[
	Obtiene los canales del VCC y si tienen licencias.
--]]
local function GetChannelsLicences (vcc, keyspace)
	local agent = require "apps.agent.agent"
	return agent.GetChannelsLicences(vcc, keyspace)
end

--[[
-- Local helper function. Returns the meta-data associated with a report
--]]
local function getReportHelper (report_id, _T)
	local ks, err = model.getCassandraConnection()
	if not ks then
		return nil, "Internal Server Error"
	end

	local ok, report, err = helpers.safeInvoke(manager.getReport, function(result, err)
			return true, result, err
		end, ks, inconcert.authUser.virtualContactCenter, report_id)

	if not ok then
		ks:Close()
		return nil, "Internal Server Error"
	end

	if not report then
		ks:Close()
		LogInfo("Report %s not found for VCC %s", report_id, inconcert.authUser.virtualContactCenter)
		return nil, nil -- So it becomes a 404
	end

	-- Get the columns and filters
	--LogDebug("[REPORTS:getReportHelper] report_id: %s', report.cut: '%s'", report_id, report.cut)
	if report.cut then
		local cutInfo, errMessage = manager.getCut(ks, inconcert.authUser.virtualContactCenter, report.cut)
		--LogDebug("[REPORTS:getReportHelper]  report_id: %s', report.cut: '%s', errMessage: '%s', cutInfo: '%s'", report_id, report.cut, errMessage, json.encode(cutInfo or {}))
		--assert(cutColumns, errMessage)
		if cutInfo then
			report.cutInfo = cutInfo
		else
			LogWarning("[REPORTS:getReportHelper] Cut '%s' definitions does not exist for report_id: '%s'.", report.cut, report_id)
		end
	end

	report.description = _T(report.description or "")
	report.label = _T(report.name)

	if report.columnDetails then
		for k, v in pairs(report.columnDetails) do
			if v.columnName then
				v.columnLabel = _T(v.columnName)
			end
		end
	end

	if report.lineCharts then
		for _, chart in pairs(report.lineCharts) do
			chart.name = _T(chart.name)
			for k, v in pairs(chart.indicators) do
				if v.name then
					v.name = _T(v.name)
				end
			end
		end
	end

	if report.totalCharts then
		for _, chart in pairs(report.totalCharts) do
			chart.name = _T(chart.name)
			for k, v in pairs(chart.indicators) do
				if v.name then
					v.name = _T(v.name)
				end
			end
		end
	end

	--LogDebug("[REPORTS:getReportHelper] Retrieving Report %s found for VCC %s. Report data: %s", report_id, report.vcc, json.encode(report))
	ks:Close()
	return report
end

--[[
-- Local helper function. Returns the meta-data associated with a dimension
--]]
local function getDimensionHelper(dimension_id, _T)
	local ks, err = model.getCassandraConnection()
	if not ks then
		return nil, "Internal Server Error"
	end

	local dimension, err
	_.some({ "system", inconcert.authUser.virtualContactCenter }, function(vcc)
		dimension, err = manager.getDimension(ks, vcc, dimension_id)
		return dimension
	end)
	ks:Close()

	if not dimension then
		LogInfo("Dimension %s not found for VCC %s", dimension_id, inconcert.authUser.virtualContactCenter)
		return nil, nil -- So it becomes a 404
	end

	dimension.description = _T(dimension.description or "")
	dimension.label = _T(dimension.displayName or dimension_id)

	if dimension.domain then
		local domainTable = {}
		for k, v in pairs(dimension.domain) do
			domainTable[v.value or v] = _T(v.label or v.value or v)
		end
		dimension.domain = domainTable
	end

	return dimension
end

--[[
-- Helper function used to obtain all elements of an specified type
--]]
local function listAllByEntity(name, listFunction, getFunction, iterator)
	iterator = iterator or function() end

	local ks, err = model.getCassandraConnection()
	if not ks then
		return false, "Internal Server Error"
	end

	local result = {}
	for _, vcc in ipairs({ "system", inconcert.authUser.virtualContactCenter }) do
		-- list entities first
		local ok = helpers.safeInvoke(listFunction, function(entities, err)
			if not entities then
				LogError("Failed to list %ss for VCC %s: %s", name, vcc, err)
				return false
			end
			-- and then load them one by one
			for _, entity_id in ipairs(entities) do
				local ok = helpers.safeInvoke(getFunction, function(entity, err)
					if not entity then
						LogError("Failed to load %s %s from VCC %s: %s", name, entity_id, vcc, err)
						return false
					end

					entity.id = entity_id

					if iterator(entity, entity_id) ~= false then
						result[entity_id] = entity
					end

					return true
				end, ks, vcc, entity_id)

				if not ok then
					return false
				end
			end
			-- all good
			return true
		end, ks, vcc)

		if not ok then
			ks:Close()
			return false, "Internal Server Error"
		end
	end

	ks:Close()
	return result
end


--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",

	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		LogInfo("[REPORTS.app_status:CheckStatus] ---> Checking status. hasChanged: '%s', newConfig: '%s' ...", hasChanged, newConfig)

		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end

		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"cassandra","redis", "irLinks", "elasticsearch"})

		app_status.current = newConfig

		--Configure Redis
		redis = WrapRedis(newConfig["redis"]["host"], tonumber(newConfig["redis"]["port"]), newConfig["redis"]["pwd"])

		local __, cachandra, state = _.firsty(newConfig.cassandra:childs())
		if cachandra and state ~= "deleted" then
			local reporterClient = GetReporterClient()
			LogInfo("Reports: Connecting to cassandra on %s:%d", cachandra.host, cachandra.port)
			cassandra.seed(cachandra.host, cachandra.port)
			cassandra.ip = cachandra.host
			cassandra.port = cachandra.port
			reporter.ConnectCassandra(cachandra.host, cachandra.port)
			reporter.SetGlobalTimer(inconcert.environment.apps.agent.cluster)
			reporterClient:ConfigureCassandra({
				{ host = cachandra.host, port = cachandra.port }
			})

			local elasticsearch = newConfig.elasticsearch and newConfig.elasticsearch[1] and {
				host = newConfig.elasticsearch[1].host,
				port = tonumber(newConfig.elasticsearch[1].port),
				-- elascitsearch user credentials
				credentials = {
					userId    = newConfig.elasticsearch[1].userId,   -- or "inconcert",
					userPwd   = newConfig.elasticsearch[1].userPwd   -- or "12345678"
				}
			}

			LogInfo("Reports: Connecting fullcontact to elasticsearch on %s", json.encode(elasticsearch or {}))

			fullcontact.init({
				owner = "REPORTS",
				cassandra = cachandra,
				elasticsearch = elasticsearch,
				logger = MakeCompatibleLogger()
			})

			model.init({
				cassandra = cassandra
			})

			m_outboundEngineClient = require "apps.outbound_engine.outboundengineclient"
			model_processes = require "apps.outbound_engine.models.processes"
		end

		if newConfig:DiffState("irLinks") == "updated" or newConfig:DiffState("irLinks") == "added" then
			local links = {}
			for key, value, state in newConfig["irLinks"]:childs() do
				if state ~= "deleted" then
					table.insert (links, value)
				end
			end
			LogInfo("Adding new agent links for cluster %s", inconcert.environment.apps.agent.cluster)
			irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
			irLink:DiscoverIRs(links)
		end

		LogInfo("Reports: Testing cassandra connection")
		local connection, err = model.getCassandraConnection()
		if connection then
			connection:Close()
			LogWarning("Reports: Cassandra connection validated successfully")
			app_status.actualStatus = "ready"
		else
			LogWarning("Reports: Could not connect to cassandra (%s)", err)
			app_status.actualStatus = "invalid"
		end

		return app_status.actualStatus
	end
}

--[[
-- Called on application start
--]]
function Initialize ()
	LogInfo("Initializing application 'reports'")

	exporter.load_exporters(module_dir .."/exporters")

	local base_config = inconcert.environment.apps.reports
	configControl = lconfigcontrol.Client({
			type         = "reports",
			address      = base_config.config_server.ip,
			port         = base_config.config_server.port,
			node_address = base_config.config_server.myIp,
			node_port    = base_config.config_server.myPort
		})

	configControl:AddServiceDependency("InteractionRouter")
	configControl:AddServiceDependency("Repository")
	configControl:AddServiceDependency("outbound_engine")

	configControl:TriggerNodeReady()
	return true
end

--[[
	Función que valida si el logged user tiene permisos sobre la app "administration", es decir Si Es o No un Supervisor.
	La misma se utiliza para saber la visibilidad de los datos de los usuarios de tipo supervisor.
]]--
local function HasPermissionToAdministrationApp (ksAgent)
	local vcc = inconcert.authUser.virtualContactCenter
	local fullUserId = inconcert.authUser.userName.."@"..vcc
	local userGroup = inconcert.authUser.userGroup
	local hasPermissionToAppAdministration = permissions.HasPermissionToApplication("administration", fullUserId, userGroup, nil, ksAgent)
	--LogDebug("[REPORTS:HasPermissionToAdministrationApp] User '%s' Role: '%s' --> %s HAS Permission in Administration APP.", fullUserId, userGroup, hasPermissionToAppAdministration and "" or "NOT")
	return hasPermissionToAppAdministration
end

local function getVccCampaignsWithAccounts (vcc, ksAgent)
	local campaignsWithAccounts = model.GetVccAccounts(vcc, ksAgent)
	--LogDebug("[REPORTS:getVccCampaignsWithAccounts] vcc: '%s', campaignsWithAccounts: '%s", vcc, json.encode(campaignsWithAccounts or {}))
	return campaignsWithAccounts
end

local function getVccCampaigns (ksAgent)
	return model.listVccCampaigns(inconcert.authUser.virtualContactCenter, ksAgent)
end

local function getCampaignDialBatches (campaignId, ksAgent)
	return model.listCampaignDialerBatches(campaignId, ksAgent)
end

local function getOutboundDialProcesses (vcc)
	return model_processes.LoadProcesses(vcc)
end

local function getCampaignProcess(campaignId, processes)
	for processId, process in pairs(processes) do
		if process.configuration.campaign == campaignId then
			process.id = processId
			return process
		end
	end
	return nil
end

local function getOutboundProcessDialBatches (campaignId, processes)
	local process = getCampaignProcess(campaignId, processes)
	if not process then
		return {}
	end

	local batches = model_processes.LoadBatchesOfProcessDefinition(process.id)
	local cleanBatches = {}
	for batchId, batch in pairs(batches) do
		batch = batch.configuration
		batch.id = batchId
		batch.processId = process.id
		cleanBatches[batch.name] = batch
	end
	return cleanBatches
end

local function GetUserCampaigns (vcc, userId, ksAgent)
	local fullUserId = utils.AppendVcc(utils.GetId(userId), vcc)
	return model.GetUserCampaigns(fullUserId, ksAgent)
end

local function GetCampaignMembers (campaignId, ksAgent)
	local vcc = inconcert.authUser.virtualContactCenter
	return model.GetCampaignMembers(vcc, campaignId, ksAgent)
end

local function IsMemberOfCampaign (userId, campaignId, ksAgent)
	local isMember = model.IsMemberOfCampaign(inconcert.authUser.virtualContactCenter, userId, campaignId, ksAgent)
	--LogDebug("[REPORTS:IsMemberOfCampaign] User: '%s' Is Member: '%s' of Campaign: '%s", userId, isMember, campaignId)
	return isMember
end

--[[
	Funcion encargada de filtrar las campanas que puede visualizar el logged user, dependiendo de
		@filterCampaignId (optional) = el valor viene si client side se ha especificado una campana.
		@interactionType (optional)  = el tipo de interaccion (que sale de la info de reporte)
										y sirve para filtrar las campanas de un channel especifico.
]]--
local function FilterCampaignsByLoggedUser (filterCampaignId, interactionType)
	--    -------------------------------------------------------
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local vcc = inconcert.authUser.virtualContactCenter
	local fullUserId = inconcert.authUser.userName.."@"..vcc
	-- Aca se valida si el usuario es de tipo administrador y si esta restringido.
	local hasPermissionToAppAdministration = HasPermissionToAdministrationApp(ksAgent)
	local isRestricted = permissions.IsPermissionEnabledForLoggedUser("administration", "admin_restricted", nil, ksAgent)

	if hasPermissionToAppAdministration and not isRestricted then
		ksAgent:Close()
		if filterCampaignId then
			return {filterCampaignId}
		end
		return ""
	end

	-- Funcion que valida que al menos exista una cuenta del channel en la campana
	local function ValidateProxyAccount (accounts)
		if not interactionType then
		-- Si no llega como parámetro el tipo de interaccion, la campana se debe incluir.
			--LogDebug("ValidateProxyAccount 1")
			return true
		end
		interactionType = interactionType:lower()
		for accountId, account in pairs(accounts) do
			if interactionType == (account.proxyType or ""):lower() then
				--LogDebug("ValidateProxyAccount 2")
				return true
			end
		end
		--LogDebug("ValidateProxyAccount 3")
		return false
	end

	filterCampaignId = filterCampaignId and filterCampaignId:lower()

	-- Aca se valida si el usuario es de tipo administrador.
	local userCampaigns

	-- Funcion que valida si la campana se tiene que incluir en la lista campanas.
	--
	local function MustIncludeCampaign (fullCampaignId, accounts)
		fullCampaignId = fullCampaignId:lower()
		local campaignId = utils.GetId(fullCampaignId)

		--LogDebug("MustIncludeCampaign 0.5 -> campaignId: %s, filterCampaignId: %s", campaignId, filterCampaignId)

		-- Si se ha proporciando una campana en el filtro, solo debe de validar la inclusión de la misma.
		if filterCampaignId and filterCampaignId ~= campaignId then
			--LogDebug("MustIncludeCampaign 1")
			return false
		end
		-- si no es un usuario administrador, tiene que valida contra las campanas a las cuales el logged user pertenezca.
		userCampaigns = userCampaigns or GetUserCampaigns(vcc, fullUserId, kAgent)
		if userCampaigns[fullCampaignId] then
			return ValidateProxyAccount(accounts)
		end
		--LogDebug("MustIncludeCampaign 2")
		return false
	end

	-- obtengo todas las campanas del vcc y sus cuentas.
	local campaignsWithAccounts = getVccCampaignsWithAccounts(vcc, ksAgent)
	ksAgent:Close()

	local filteredUserCampaigns = {}
	for campaignId, accounts in pairs(campaignsWithAccounts) do
		-- invoco a la funcion que realiza la validacion pertinente para saber si la campana es visible para el logged user.
		local mustIncludeCampaign = MustIncludeCampaign(campaignId, accounts)
		if mustIncludeCampaign then
			local campaignUid = utils.GetId(campaignId):lower()
			table.insert(filteredUserCampaigns, campaignUid)
		end
	end

	return filteredUserCampaigns
end

local function makeIdsToNamesMapping()
	local result = {
		Account = {},
		Page = {}
	}
	local campaignsWithAccounts = getVccCampaignsWithAccounts(inconcert.authUser.virtualContactCenter)
	for _, accounts in pairs(campaignsWithAccounts) do
		for _, account in pairs(accounts) do
			result.Account[account.uid] = account.name
			if account.pagesNames then
				for pageId, pageName in pairs(account.pagesNames) do
					result.Page[pageId] = pageName
				end
			end
		end
	end
	return result
end

local overrideStrategies = {

	--
	-- Given a filter that takes contact id, the report may choose to
	-- search by some contact data instead of contact id, in that case, an
	-- override must me performed
	GetContactByAddress = function(contactAddress)

		--
		-- This function will try to guess the address type based in some
		-- simple regular expressions, its not intended to validate it.
		local function guessAddressTypes(address2test)
			if string.match(address2test, '^([^@]+)@([^@]+)%.(.+)$') then
				return { "mail" }
			end

			-- test for phone (check for numbers)
			local maybeNumbers = address2test:gsub("[ +-]*", "")
			if string.match(maybeNumbers, "^%d+$") then
				return { "phone" }
			end

			if string.match(address2test, "^@[%w_]+") then
				return { "twitter" }
			end

			-- may be facebook, twitter
			return { "twitter", "facebook", "instagram", "youtube", "appstore", "googleplay" }
		end

		if not contactAddress or contactAddress == "" then
			return nil -- don't filter with this field
		end

		local contactId
		_.each(guessAddressTypes(contactAddress), function(addressType)
			LogInfo("Search contact by address %s/%s", addressType, contactAddress)
			local contact = fullcontact.GetByAddress(
					addressType,
					contactAddress,
					inconcert.authUser.virtualContactCenter,
					{ logical = true } -- accept deleted contacts
			)
			if contact then
				contactId = contact.Id
				LogInfo("Found contact %s for address %s/%s", contactId, addressType, contactAddress)
				return false -- break the loop
			end
		end)

		return contactId
	end,

	GetContactByExternalId = function(externalId)
		local contact = fullcontact.GetByExternalId(
			externalId,
			inconcert.authUser.virtualContactCenter,
			{ logical = true } -- accept deleted contacts
		)
		if contact then
			LogInfo("Found contact %s for external Id %s", contact.Id, externalId)
			return contact.Id
		end
	end
}

local function makeQueryRunner (report, queryParams, _T)
	LogInfo("[REPORTS:makeQueryRunner] Make Query (type:%s) (id:%s) (name:%s) (executionType: '%s').", report.type, report.id, report.name, queryParams.executionType)
	--LogDebug("[REPORTS:makeQueryRunner] Make Query (params:%s)", report.type, report.id, report.name, json.encode(queryParams))
	--LogDebug("---      makeQueryRunner       ---------> Query (%s) %s", report.type, json.encode(report))

	local vcc = inconcert.authUser.virtualContactCenter
	local has_allegro_integration = utils.HasAllegroIntegration(vcc) or false
	local idsToNames = makeIdsToNamesMapping()
	
	--LogDebug("1 ----> report.resultSize: %s", report.resultSize)
	--LogDebug("2 ----> queryParams.resultSize: %s", queryParams.resultSize)
	local resultSize = report.resultSize or queryParams.resultSize
	local exportType = queryParams.export_type or ""
	if not getNullable(resultSize) then
		resultSize = queryParams.export_type and 1000 or 40
	end
	--LogDebug("3 ----> resultSize: %s", resultSize)

	if queryParams and queryParams.executionType == "BY SCHEDULER" and queryParams.queryFields and queryParams.queryFields.scheduleId then
		local reportSchedule = model.getReportSchedule(queryParams.queryFields.scheduleId)
		if reportSchedule.RepeatPolicy == "EveryNHours" and reportSchedule.SendType ~= "database" then
			-- #14979: set isUTC flag to false, and startDate according to user timezone
			queryParams.filter.isUTC = false
			queryParams.queryFields.isUTC = false
			queryParams.filter.startDate = LuaOrm.Date(queryParams.filter.startDate):toDifferentTimezoneString(inconcert.authUser.timezone, "%Y-%m-%d %H:%M:%S")
			queryParams.queryFields.startDate = queryParams.filter.startDate
			queryParams.filter.DateFrom = queryParams.filter.startDate
			-- #14979. Extend to change to localtime the endDate & DateTo when schedule report is Hoursbefore.
			if queryParams.utz_endDate and queryParams.utz_endDate == true then
				queryParams.filter.endDate = LuaOrm.Date(queryParams.filter.endDate):toDifferentTimezoneString(inconcert.authUser.timezone, "%Y-%m-%d %H:%M:%S")
				queryParams.queryFields.endDate = queryParams.filter.endDate
				queryParams.filter.DateTo = queryParams.filter.endDate
			end
		end
	end	

	local function makeQuery (queryFunction, report, utils, redis, authFilter)
		return function()
			--LogDebug("------------> Query (%s) (%s)", report.type, json.encode(report))
			-- local ok, res, handle, err = pcall(queryFunction, report, utils, _T, idsToNames, redis, authFilter, GetReporterClient())
			-- if not ok then
			-- 	return nil, nil, res
			-- end
			-- return res, handle, err
			return queryFunction(report, utils, _T, idsToNames, redis, authFilter, GetReporterClient())
		end
	end

	local function GetExportContentType ()
		local defaultExportContentType = "HTML"
		if queryParams.exportContentType == nil then
			--LogDebug("[REPORTS:makeQueryRunner] 1 - queryParams.exportContentType: '%s'. Return nil.", queryParams.exportContentType)
			return defaultExportContentType
		end
		if type(queryParams.exportContentType) ~= "string" then
			--LogDebug("[REPORTS:makeQueryRunner] 2 - queryParams.exportContentType: '%s'. Return default :'%s'.", queryParams.exportContentType, defaultExportContentType)
			return defaultExportContentType
		end
		local supportedExporteContentType = {
			HTML          = true,
			JSON          = true,
			HTML_AND_JSON = true
		}
		local currentExportContentType = queryParams.exportContentType:upper()
		if supportedExporteContentType[currentExportContentType] then
			--LogDebug("[REPORTS:makeQueryRunner] 3 - queryParams.exportContentType: '%s'. Return supported :'%s'.", queryParams.exportContentType, currentExportContentType)
			return currentExportContentType
		end
		--LogDebug("[REPORTS:makeQueryRunner] 4 - queryParams.exportContentType: '%s' Unsupported. Return default :'%s'.", queryParams.exportContentType, defaultExportContentType)
		return defaultExportContentType
	end

	--[[
		Funcion que convierte en un array el filtro campana proporcionado client side.
		Dependiendo de la visibilidad y permisos del logged user.
	]]--
	local function FilterQueryCampaignsByLoggedUser (filter)
		if inconcert.authUser.api then return end

		local proxyType = report.interactionType or filter.Type
		--LogDebug("[REPORTS:makeQueryRunner.FilterQueryCampaignsByLoggedUser] -> START filter.Campaign: %s, proxyType: %s", filter.Campaign, proxyType)
		local campaigns
		if filter.Campaign ~= nil then
			campaigns = FilterCampaignsByLoggedUser(filter.Campaign, proxyType)
		else
			campaigns = FilterCampaignsByLoggedUser(nil, proxyType)
		end
		if campaigns then
			filter.Campaign = campaigns
		end
		--LogDebug("[REPORTS:makeQueryRunner.FilterQueryCampaignsByLoggedUser] -> END filter.Campaign: %s, proxyType: %s", json.encode(filter.Campaign), proxyType)
	end

	--[[
		Funcion que convierte en un array el filtro agentes proporcionado client side.
		Dependiendo de la visibilidad y permisos del logged user y dependiendo de las campanas retornadas por la funcion "FilterQueryCampaignsByLoggedUser"
	]]--
	local function FilterQueryAgentsByCampaigns (filter)
		--LogDebug("[REPORTS:makeQueryRunner.FilterQueryAgentsByCampaigns] -> START filter.Agent: '%s', filter.Campaign: '%s'", filter.Agent, filter.Campaign)
		-- Aca se valida si el usuario es de tipo administrador.
		local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
		local hasPermissionToAppAdministration = HasPermissionToAdministrationApp(ksAgent)
		ksAgent:Close()

		if hasPermissionToAppAdministration then
			if not filter.Agent then
				filter.Agent = ""
			end
			return
		end

		if not filter.Campaign then
			--LogDebug("[REPORTS:makeQueryRunner.FilterQueryAgentsByCampaigns] 1")
			filter.Agent = ""
			return
		end

		local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
		local agents = {}

		 if filter.Agent then
		-- Si en el filtro client side, SI se ha proporcionado un agente, tiene que validar que el agente sea member de al menos una de las campanas visibles del logged user.
			local agent = filter.Agent
			local isMember = false
			for _, campaignId in pairs(filter.Campaign) do
				isMember = IsMemberOfCampaign(agent, campaignId, ksAgent)
				if isMember then
					break
				end
			end
			if isMember then
				--LogDebug("[REPORTS:makeQueryRunner.FilterQueryAgentsByCampaigns] 2")
				table.insert(agents, agent)
			else
				--LogDebug("[REPORTS:makeQueryRunner.FilterQueryAgentsByCampaigns] 3")
			end
		else
		-- Si en el filtro client side, NO se ha proporciando un agente, tiene que retornar todos los agentes de todas las campanas visibles del logged user.

			--LogDebug("[REPORTS:makeQueryRunner.FilterQueryAgentsByCampaigns] 4")
			local hashMembersOfAllCampaigns = {}

			for _, campaignId in pairs(filter.Campaign) do
				local campaignMembers = GetCampaignMembers(campaignId, ksAgent)
				for userId in pairs(campaignMembers or {}) do
					userId = utils.GetId(userId:lower())
					if not hashMembersOfAllCampaigns[userId] then
						table.insert(agents, userId)
						hashMembersOfAllCampaigns[userId] = true
					end
				end
			end
		end

		filter.Agent = agents
		ksAgent:Close()

		--LogDebug("[REPORTS:makeQueryRunner.FilterQueryAgentsByCampaigns] -> END filter.Agent: '%s', filter.Campaign: '%s'", json.encode(filter.Agent), filter.Campaign)
	end

	local function FilterQueryCampaignsAndAgentsByLoggedUser (filter)
		FilterQueryCampaignsByLoggedUser(filter)
		FilterQueryAgentsByCampaigns(filter)
	end

	local function makeDetailQuery(report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeDetailQuery] -> Making Detail Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				if type(field) == "string" then
					fields[index] = {
						name = field,
						dimensionData = getDimensionHelper(field, _T)
					}
					if not fields[index].dimensionData then
						fields[index].label = _T(field)
					end
				else
					if field.dimension then
						field.dimensionData = getDimensionHelper(field.dimension, _T)
					end
					if field.columnName then
						field.label = _T(field.columnName)
					end
				end
			end
		end

		-- remove startDate and endDate from queryFields
		_.each({"startDate", "endDate"}, function(field)
			report[field] = _.remove(queryParams.queryFields, field)
		end)
		report = _.extend(report, {
			resultSize  = resultSize,
			lastHandle  = getNullable(queryParams.last_handle) or "",
			timezone    = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields, {
				VCC = inconcert.authUser.virtualContactCenter
			}),
			exportContentType = GetExportContentType()
		})
	
		-- Ejecuto el filtro del Query sobre los campos "Campaign" y "Agent"
		FilterQueryCampaignsByLoggedUser(queryParams.queryFields)

		processFields(report.projectFields)
		processFields(report.threadedFields)

		-- handle overrides
		_.each(report.filterOverrides or {}, function(data, fieldName)
			local finalValue
			local atLeastOneSet = _.some(data.fields, function(field)
				local fieldValue = report.queryFields[field.name]
				if fieldValue and fieldValue ~= "" then
					atLeastOneSet = true
					local overrideFn = overrideStrategies[field.strategy] or function(val)
						LogError("Unknown override strategy %s", field.strategy)
						return val -- identity
					end
					finalValue = overrideFn(fieldValue)
					return true -- break the loop
				end
			end)
			-- finalValue can be nil in the following cases
			--  1. None of the search fields were filled
			--  2. No contact matched the requested pattern
			-- In the second case I must make sure nothing matches, how do I do that? By providing a
			-- contactId that does not exist in the DB: uuid.new()
			report.queryFields[fieldName] = finalValue == nil and atLeastOneSet and uuid.new() or finalValue
		end)

		return makeQuery(reporter.QueryDetail, report, utils)
	end

	local function makeSummaryQuery(report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeSummaryQuery] -> Making Summary Query ...")

		report = _.extend(report, {
			resultSize = resultSize,
			lastHandle = getNullable(queryParams.last_handle)
		})
		if not report.lastHandle then
			-- set and remove fields that are not part of the filter
			_.each({"sliceSize"}, function(field)
				report[field] = _.remove(queryParams.filter, field)
			end)

			if (report.cut or ""):upper() == "AGENTSLICE" then
			-- Si es un reporte de tipo "equipo->estados por agente" (en si el CUT es igual a AgentSlice, tengo que filtrar por los agentes de las campanas que tenga asignadas el loggedUser).
			--
				--LogDebug("[REPORTS:makeQueryRunner.makeSummaryQuery] 1 report.cut: %s. Filter Campaigns And Agents By Logged User.", report.cut)
				FilterQueryCampaignsAndAgentsByLoggedUser(queryParams.filter)
			else
			-- Si NO es un reporte de tipo "equipo->estados por agente" (en si el CUT es distinto a AgentSlice, tengo que filtrar unicamente por las campanas que tenga asignadas el loggedUser).
			--
				--LogDebug("[REPORTS:makeQueryRunner.makeSummaryQuery] 2 report.cut: %s. Filter Campaigns By Logged User.", report.cut)
				FilterQueryCampaignsByLoggedUser(queryParams.filter)
			end

			local extendReportInfo = {
				timezone = inconcert.authUser.timezone,
				filter   = _.extend(queryParams.filter, {
					VCC = inconcert.authUser.virtualContactCenter
				}),
				queryFields = _.extend(queryParams.queryFields or {}, {
					VCC = inconcert.authUser.virtualContactCenter
				})
			}
			--LogDebug("[REPORTS:makeQueryRunner] makeSummaryQuery-> extendReportInfo: '%s'.", json.encode(extendReportInfo))
			report = _.extend(report, extendReportInfo)
		end

		return makeQuery(reporter.QuerySummary, report, utils)
	end

	local function makeRealtimeQuery (report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeRealtimeQuery] -> Making Realtime Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				field.label = _T(field.name)
				if field.description then
					field.description = _T(field.description)
				end
			end
		end

		local function getCampaigns (filter)
			-- Ejecuto el filtro del Query sobre el campo "Campaign"
			FilterQueryCampaignsByLoggedUser(filter)
			return filter.Campaign
		end
		local campaigns = getCampaigns(queryParams.filter)

		_.each({"startDate", "endDate"}, function(field)
			report[field] = _.remove(queryParams.filter, field)
		end)

		report = _.extend(report, {
			resultSize  = resultSize,
			lastHandle  = getNullable(queryParams.last_handle),
			campaigns   = campaigns,
			timezone    = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields or {}, {
				VCC = inconcert.authUser.virtualContactCenter
			})
		})
		processFields(report.projectFields)

		return makeQuery(reporter.QueryRealtime, report, utils)
	end

	local function makeDialerBatchQuery(report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeDialerBatchQuery] -> Making DialerBatch Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				field.label = _T(field.name)
				if field.description then
					field.description = _T(field.description)
				end
			end
		end

		local function getCampaigns(filter)
			-- Ejecuto el filtro del Query sobre el campo "Campaign"
			FilterQueryCampaignsByLoggedUser(filter)
			return filter.Campaign
		end

		local function getBatches (filter)
			local batches = {}
			local campaigns = getCampaigns(filter)
			--LogDebug("[getBatches] campaigns: '%s'", json.encode(campaigns or {}))

			local batchId = _.remove(filter, "Batch")
			for _, campaignId in pairs(campaigns or {}) do
				local campaignAtVcc = utils.AppendVcc(utils.GetId(campaignId), inconcert.authUser.virtualContactCenter)
				local campaignBatches = getCampaignDialBatches(campaignAtVcc)
				if batchId and batchId ~= "" then
					for campaignBatchId, batch in pairs(campaignBatches) do
						if string.upper(campaignBatchId) == string.upper(batchId) then
							batch.name = campaignBatchId
							batch.campaign = campaignId
							batch.campaignAtVcc = campaignAtVcc
							table.insert(batches, batch)
							break
						end
					end
				else
					for batchId, batch in pairs(campaignBatches) do
						batch.name = batchId
						batch.campaign = campaignId
						batch.campaignAtVcc = campaignAtVcc
						table.insert(batches, batch)
					end
				end
			end
			table.sort(batches, function(a, b)
				if a.campaign == b.campaign then
					return string.lower(a.name) < string.lower(b.name)
				end
				return string.lower(a.campaign) < string.lower(b.campaign)
			end)
			return batches
		end

		report = _.extend(report, {
			resultSize = resultSize,
			lastHandle = getNullable(queryParams.last_handle),
			batches    = getBatches(queryParams.filter),
			timezone   = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields or {}, {
				VCC = inconcert.authUser.virtualContactCenter
			})
		})
		processFields(report.projectFields)

		--add summary part
		if report.summary then
			report.summary = _.extend(report.summary, {
				vcc			= report.vcc,
				sliceSize	= "Total",
				resultSize	= 1,
				timezone	= inconcert.authUser.timezone,
				filter		= {
					VCC		= inconcert.authUser.virtualContactCenter
				},
				queryFields = _.extend(queryParams.queryFields or {}, {
					VCC = inconcert.authUser.virtualContactCenter
				})
			})
		end

		return makeQuery(reporter.QueryDialerBatch, report, utils)
	end

	local function makeDialerContactsQuery(report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeDialerContactsQuery] -> Making DialerContacts Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				field.label = _T(field.name)
				if field.dimension then
					field.dimensionData = getDimensionHelper(field.dimension, _T)
				end
				if field.description then
					field.description = _T(field.description)
				end
			end
		end
		--LogDebug("--> queryParams: %s", json.encode(queryParams))
		queryParams.queryFields = queryParams.queryFields or {}
		queryParams.queryFields.Campaign = queryParams.filter.Campaign

		local batchId = _.remove(queryParams.filter, "Batch")
		local contactId = _.remove(queryParams.filter, "ContactId")

		report = _.extend(report, {
			resultSize = resultSize,
			lastHandle = getNullable(queryParams.last_handle),
			batch	   = batchId,
			contactId  = contactId,
			timezone   = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields, {
				VCC = inconcert.authUser.virtualContactCenter
			})
		})

		-- Ejecuto el filtro del Query sobre el campo "Campaign"
		FilterQueryCampaignsByLoggedUser(queryParams.queryFields)

		processFields(report.projectFields)
		return makeQuery(reporter.QueryDialerContacts, report, utils)
	end

	local function makeLiveDetailQuery (report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeLiveDetailQuery] -> Making Live Detail Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				if type(field) == "string" then
					fields[index] = {
						name = field,
						dimensionData = getDimensionHelper(field, _T)
					}
					if not fields[index].dimensionData then
						fields[index].label = _T(field)
					end
				else
					if field.dimension then
						field.dimensionData = getDimensionHelper(field.dimension, _T)
					end
					if field.columnName then
						field.label = _T(field.columnName)
					end
				end
			end
		end

		-- remove startDate and endDate from queryFields
		_.each({"startDate", "endDate"}, function(field)
			report[field] = _.remove(queryParams.queryFields, field)
		end)
		report = _.extend(report, {
			resultSize  = resultSize,
			lastHandle  = getNullable(queryParams.last_handle) or "",
			timezone    = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields, {
				VCC = inconcert.authUser.virtualContactCenter
			}),
			exportType = exportType
		})

		-- Ejecuto el filtro del Query sobre el campo "Campaign"
		FilterQueryCampaignsByLoggedUser(queryParams.queryFields)

		processFields(report.projectFields)

		-- handle overrides
		_.each(report.filterOverrides or {}, function(data, fieldName)
			local finalValue
			local atLeastOneSet = _.some(data.fields, function(field)
				local fieldValue = report.queryFields[field.name]
				if fieldValue and fieldValue ~= "" then
					atLeastOneSet = true
					local overrideFn = overrideStrategies[field.strategy] or function(val)
						LogError("Unknown override strategy %s", field.strategy)
						return val -- identity
					end
					finalValue = overrideFn(fieldValue)
					return true -- break the loop
				end
			end)
			-- finalValue can be nil in the following cases
			--  1. None of the search fields were filled
			--  2. No contact matched the requested pattern
			-- In the second case I must make sure nothing matches, how do I do that? By providing a
			-- contactId that does not exist in the DB: uuid.new()
			report.queryFields[fieldName] = finalValue == nil and atLeastOneSet and uuid.new() or finalValue
		end)

		return makeQuery(reporter.QueryLiveDetail, report, utils, redis)
	end

	local function makeOutboundDialerBatchQuery(report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeDialerBatchQuery] -> Making DialerBatch Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				field.label = _T(field.name)
				if field.description then
					field.description = _T(field.description)
				end
			end
		end

		local function getCampaigns(filter)
			-- Ejecuto el filtro del Query sobre el campo "Campaign"
			FilterQueryCampaignsByLoggedUser(filter)
			return filter.Campaign
		end

		local function getBatches (filter)
			local batches = {}
			local campaigns = getCampaigns(filter)
			local processes = getOutboundDialProcesses(inconcert.authUser.virtualContactCenter)
			--LogDebug("[getBatches] campaigns: '%s'", json.encode(campaigns or {}))

			local batchId = _.remove(filter, "Batch")
			for _, campaignId in pairs(campaigns or {}) do
				local campaignAtVcc = utils.AppendVcc(utils.GetId(campaignId), inconcert.authUser.virtualContactCenter)
				local campaignBatches = getOutboundProcessDialBatches(campaignAtVcc, processes)
				if batchId and batchId ~= "" then
					for campaignBatchId, batch in pairs(campaignBatches) do
						if string.upper(campaignBatchId) == string.upper(batchId) then
							batch.campaign = campaignId
							batch.campaignAtVcc = campaignAtVcc
							table.insert(batches, batch)
							break
						end
					end
				else
					for batchId, batch in pairs(campaignBatches) do
						batch.campaign = campaignId
						batch.campaignAtVcc = campaignAtVcc
						table.insert(batches, batch)
					end
				end
			end
			table.sort(batches, function(a, b)
				if a.campaign == b.campaign then
					return string.lower(a.name) < string.lower(b.name)
				end
				return string.lower(a.campaign) < string.lower(b.campaign)
			end)
			return batches
		end

		report = _.extend(report, {
			resultSize = resultSize,
			lastHandle = getNullable(queryParams.last_handle),
			batches    = getBatches(queryParams.filter),
			timezone   = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields or {}, {
				VCC = inconcert.authUser.virtualContactCenter
			})
		})
		processFields(report.projectFields)

		--add summary part
		if report.summary then
			report.summary = _.extend(report.summary, {
				vcc			= report.vcc,
				sliceSize	= "Total",
				resultSize	= 1,
				timezone	= inconcert.authUser.timezone,
				filter		= {
					VCC		= inconcert.authUser.virtualContactCenter
				},
				queryFields = _.extend(queryParams.queryFields or {}, {
					VCC = inconcert.authUser.virtualContactCenter
				}),
			})
		end
		report.oeClient = m_outboundEngineClient

		return makeQuery(reporter.QueryOutboundDialerBatch, report, utils)
	end

	local function makeOutboundDialerContactsQuery(report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeOutboundDialerContactsQuery] -> Making DialerContacts Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				field.label = _T(field.name)
				if field.dimension then
					field.dimensionData = getDimensionHelper(field.dimension, _T)
				end
				if field.description then
					field.description = _T(field.description)
				end
			end
		end
		--LogDebug("--> queryParams: %s", json.encode(queryParams))
		queryParams.queryFields = queryParams.queryFields or {}
		queryParams.queryFields.Campaign = queryParams.filter.Campaign

		local batchId = _.remove(queryParams.filter, "Batch")
		local contactId = _.remove(queryParams.filter, "ContactId")
		if contactId and contactId ~= "" then
			contactId = utils.AppendVcc(utils.GetId(contactId), inconcert.authUser.virtualContactCenter)
		end
		local campaignId = queryParams.filter.Campaign .. '@' .. inconcert.authUser.virtualContactCenter
		local processes = getOutboundDialProcesses(inconcert.authUser.virtualContactCenter)
		local process = getCampaignProcess(campaignId, processes)

		report = _.extend(report, {
			resultSize = resultSize,
			lastHandle = getNullable(queryParams.last_handle),
			batch	   = batchId,
			contactId  = contactId,
			processId  = process and process.id,
			timezone   = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields, {
				VCC = inconcert.authUser.virtualContactCenter
			}),
			oeClient	= m_outboundEngineClient
		})

		-- Ejecuto el filtro del Query sobre el campo "Campaign"
		FilterQueryCampaignsByLoggedUser(queryParams.queryFields)

		processFields(report.projectFields)
		return makeQuery(reporter.QueryOutboundDialerContacts, report, utils)
	end

	local function makeCustomQuery (report, queryParams)
		--LogDebug("[REPORTS:makeQueryRunner.makeCustomQuery] -> Making Custom Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				if type(field) == "string" then
					fields[index] = {
						name = field,
						dimensionData = getDimensionHelper(field, _T)
					}
					if not fields[index].dimensionData then
						fields[index].label = _T(field)
					end
				else
					if field.dimension then
						field.dimensionData = getDimensionHelper(field.dimension, _T)
					end
					if field.columnName then
						field.label = _T(field.columnName)
					end
				end
			end
		end

		-- remove startDate and endDate from queryFields
		queryParams.queryFields = queryParams.queryFields or {}
		_.each({"startDate", "endDate"}, function(field)
			report[field] = _.remove(queryParams.queryFields, field)
		end)
		report = _.extend(report, {
			resultSize  = resultSize,
			lastHandle  = getNullable(queryParams.last_handle) or "",
			timezone    = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields, {
				VCC = inconcert.authUser.virtualContactCenter
			})
		})
		processFields(report.projectFields)

		-- handle overrides
		_.each(report.filterOverrides or {}, function(data, fieldName)
			local finalValue
			local atLeastOneSet = _.some(data.fields, function(field)
				local fieldValue = report.queryFields[field.name]
				if fieldValue and fieldValue ~= "" then
					atLeastOneSet = true
					local overrideFn = overrideStrategies[field.strategy] or function(val)
						LogError("Unknown override strategy %s", field.strategy)
						return val -- identity
					end
					finalValue = overrideFn(fieldValue)
					return true -- break the loop
				end
			end)
			-- finalValue can be nil in the following cases
			--  1. None of the search fields were filled
			--  2. No contact matched the requested pattern
			-- In the second case I must make sure nothing matches, how do I do that? By providing a
			-- contactId that does not exist in the DB: uuid.new()
			report.queryFields[fieldName] = finalValue == nil and atLeastOneSet and uuid.new() or finalValue
		end)

		return makeQuery(reporter.QueryRealtimeCustom, report, utils, redis, authFilter)
	end

	local function makeCompositeQuery(queryParams)
		LogDebug("[REPORTS:makeQueryRunner.makeCompositeQuery] -> Making Detail Query ...")

		local function processFields(fields)
			for index, field in pairs(fields or {}) do
				if type(field) == "string" then
					fields[index] = {
						name = field,
						dimensionData = getDimensionHelper(field, _T)
					}
					if not fields[index].dimensionData then
						fields[index].label = _T(field)
					end
				else
					if field.dimension then
						field.dimensionData = getDimensionHelper(field.dimension, _T)
					end
					if field.columnName then
						field.label = _T(field.columnName)
					end
				end
			end
		end

		report = _.extend(queryParams, {
			resultSize  = resultSize,
			lastHandle  = getNullable(queryParams.last_handle) or "",
			timezone    = inconcert.authUser.timezone,
			queryFields = _.extend(queryParams.queryFields, {
				VCC = inconcert.authUser.virtualContactCenter
			}),
			exportContentType = GetExportContentType()
		})
		processFields(report.projectFields)

		return makeQuery(reporter.QueryDetail, report, utils)
	end

	if report.type == "detail" then
		return makeDetailQuery(report, queryParams)
	end
	if report.type == "livedetail" then
		return makeLiveDetailQuery(report, queryParams)
	end 
	if report.type == "summary" then
		return makeSummaryQuery(report, queryParams)
	end
	if report.type == "realtime" then
		if report.subType == "custom" then
			-- custom
			return makeCustomQuery(report, queryParams)
		else
			-- normal
			return makeRealtimeQuery(report, queryParams)
		end
	end
	--allegro oe
	if report.type == "dialerbatch" then
		if has_allegro_integration then
			return makeDialerBatchQuery(report, queryParams)
		else
			return makeOutboundDialerBatchQuery(report, queryParams)
		end
	end
	if report.type == "dialercontacts" then
		if has_allegro_integration then
			return makeDialerContactsQuery(report, queryParams)
		else
			return makeOutboundDialerContactsQuery(report, queryParams)
		end
	end

	if report.type == "composite" then
		return makeCompositeQuery(queryParams)
	end
end

local function handleReportError()
	LogError("[REPORTS:handleReportError] error running report: %s", debug.traceback())
end

local function makeQueryExporter(consumerFn, report)
	report = _.extend(report, {
		resultSize = report.resultSize or (report.type == "summary" and 20 or 100),
		lastHandle = nil,
	})
	return function()
		local ok, result, lastHandle, err = xpcall(consumerFn, handleReportError)
		report.lastHandle = lastHandle
		return (ok and result), err
	end
end

local function makeTextTranslator()
	local translator = inconcert.translators.reports
	return function(...)
		return translator:Text(...)
	end, i18nlib.LCID[translator.language]
end

function GetApplicationTitle()
	local _T = makeTextTranslator()
	return _T("reports")
end

local htmlRender = require"inconcert.render".new("reports", {
	MainTitle = _M.GetApplicationTitle,
	Title = function() return _T("reports") end,
	model = nil,
	makeTextTranslator = makeTextTranslator
})

--[[
-- Returns the meta-data associated with a report
--]]
local function getReport(web, report_id)
	local _T = makeTextTranslator()
	web.headers["Content-Type"] = "application/json"

	local report, err = getReportHelper(report_id, _T)
	if not report then
		return replyJson(web, err and 500 or 404, _T(err or "Report not found"))
	end

	if not report.category or report.category == "Custom" or report.category:match('^Interaction') then
		-- hack until a migration is implemented
		report.category = "Interactions"
	else
		report.category = "States"
	end

	return replyJson(web, 200, report)
end

local function GetAcceptedInteractionTypes (from)
	local invokedFrom = from or "any"
	local vcc = inconcert.authUser.virtualContactCenter
	local keyspace = utils.GetDatasource():CassandraConnection("Agent")
	local channelsLicenses = GetChannelsLicences(vcc, ks_agent)

	local acceptedChannels = {}

	if invokedFrom == "viewer" or invokedFrom == "any" then
		_.each(getVccCampaigns(), function(campaignId)
			local accounts = model.listCampaignAccounts(campaignId, keyspace)
			_.each(accounts, function(account)
				if not acceptedChannels[account.proxyType] then
					--LogDebug("[REPORTS:GetAcceptedInteractionTypes] Accepting reports targeting interactions of type %s", account.proxyType)
					acceptedChannels[account.proxyType] = true
				end
			end)
		end)
		acceptedChannels["VOICEMAIL"] = acceptedChannels["CALL"] or false
	elseif invokedFrom == "designer" then
		acceptedChannels = channelsLicenses
	end
	keyspace:Close()

	--LogDebug("[REPORTS:GetAcceptedInteractionTypes] ----> invokedFrom: %s, acceptedChannels: %s'", invokedFrom, json.encode(acceptedChannels))

	return acceptedChannels
end

--[[
-- Helper function used by listReports
--]]
local function makeFilterFunction (acceptedChannels, from)
	local vcc = inconcert.authUser.virtualContactCenter
	local has_allegro_integration = utils.HasAllegroIntegration(vcc) or false

	acceptedChannels = acceptedChannels or GetAcceptedInteractionTypes(from)

	return (function(report)
		if report.allegro ~= nil and report.allegro ~= has_allegro_integration then
			return false
		end

		if report.showIn then
			if report.showIn.vccType then
				if report.showIn.vccType == "ONLY IN VCC WITHOUT ALLEGRO INTEGRATION" then
					if has_allegro_integration then
						return false
					end
				elseif report.showIn.vccType == "ONLY IN VCC WITH ALLEGRO INTEGRATION" then
					if not has_allegro_integration then
						return false
					end
				elseif report.showIn.vccType ~= "ALL VCCS" then
					assert("Unrecognized report showIn.vccType. " .. tostring(report.showIn.vccType))
				end
			end

			if report.showIn.gui == false then
				return false
			end
			if from == "viewer" and report.showIn.viewer == false then
				return false
			end
			if from == "designer" and report.showIn.designer == false then
				return false
			end
		end
		
		if report.vcc == vcc then
			--LogDebug("[REPORTS:makeFilterFunction] -1---> report.id: '%s', mustInclude: 'SI'.", report.id)
			return true
		end

		-- remove reports with shift (at least for now) in order to avoid duplicated data
		if report.id:match('.*WithShifts$') then
			return false
		end
		
		-- handle the special case of the category "InteractionsInternalChats"
		local categoryWasInteractionsInternalChats = report.category == "InteractionsInternalChats"

		-- hack until a migration is implemented
		if not report.category or report.category == "Custom" or report.category:match('^Interaction') then
			report.category = "Interactions"
		else
			report.category = "States"
		end

		-- handle the special case of the category "InteractionsInternalChats"
		if categoryWasInteractionsInternalChats then
			--LogDebug("[REPORTS:makeFilterFunction] -2---> report.id: '%s', mustInclude: 'SI'.", report.id)
			return true
		end

		if not (report.category and report.category == "Interactions" and report.interactionType) then
			--LogDebug("[REPORTS:makeFilterFunction] -3---> mustInclude: 'SI', report: '%s'.", json.encode(report))
			return true
		end

		local mustInclude = acceptedChannels[report.interactionType]
		--LogDebug("[REPORTS:makeFilterFunction] ----> mustInclude: '%s' report: '%s'.", mustInclude, json.encode(report))

		return mustInclude
	end)
end

--[[
-- Returns a list containing all report categories (for viewer)
--]]
local function getCategories (acceptedChannels)
	acceptedChannels = acceptedChannels or GetAcceptedInteractionTypes()

	local result = {}
	for _, category in pairs(view_categories) do
		if category.type == "CHANNEL" and not acceptedChannels[category.channel] then
			--LogDebug("[REPORTS:getCategories] ----> NO include category: %s", json.encode(category))
			--not available
		else
			table.insert(result, category)
		end
	end
	return result
end

local function listCategories(web)
	web.headers["Content-Type"] = "application/json"
	local result = getCategories()
	return replyJson(web, 200, result)
end

--[[
-- Returns a list containing a preview of all defined reports in the system
--]]
local function listReports (web, type, from)
	local _T = makeTextTranslator()
	web.headers["Content-Type"] = "application/json"

	local ks, err = model.getCassandraConnection()
	if not ks then
		return replyJson(web, 500, _T("Internal Server Error"))
	end

	local acceptedChannels = GetAcceptedInteractionTypes(from)
	local filter_function = makeFilterFunction(acceptedChannels, from)

	local result = {}
	for _, vcc in ipairs({ "system", inconcert.authUser.virtualContactCenter }) do
		local ok = helpers.safeInvoke(manager.listReports, function(reports, err)
			if not reports then
				LogError("[REPORTS:listReports] Failed to list reports for VCC %s: %s", vcc, err)
				return false
			end

			for _, report_id in ipairs(reports) do
				helpers.safeInvoke(manager.getReport, function(report, err)
					if not report then
						LogWarning("[REPORTS:listReports] Failed to get report %s from VCC %s: %s", report_id, vcc, err)
						return
					end
					if filter_function(report) then
						--LogDebug("[REPORTS:listReports] -1---> Si include report: %s", report_id)
						result[report_id] = {
							id            = report_id,
							vcc           = vcc,
							type          = report.type,
							name          = report.name,
							label         = _T(report.name),
							description   = _T(report.description or ""),
							category      = report.category,
							view_category = report.view_category or "CUSTOM",
							sub_type      = report.subType
						}
					--else
						--LogDebug("[REPORTS:listReports] -2---> NO include report: %s", report_id)
					end
				end, ks, vcc, report_id)
			end

			return true

		end, ks, vcc)

		if not ok then
			ks:Close()
			return replyJson(web, 500, _T("Internal Server Error"))
		end
	end

	ks:Close()
	return replyJson(web, 200, result)
end

local function listViewerReports (web, type)
	LogInfo("[REPORTS:listDesignerReports] Retrieving Reports for Viewer.")
	local from = "viewer"
	return listReports(web, type, from)
end

local function listDesignerReports (web, type)
	LogInfo("[REPORTS:listDesignerReports] Retrieving Reports for Designer.")
	local from = "designer"
	return listReports(web, type, from)
end

local function wrapErrorResponse (statusCode, response)
	if type(response) == "string" then
		response = { result = response }
	end
	response = json.encode(response)

	LogError("[REPORTS:wrapErrorResponse] Unable to execute report. Code '%s', err: %s", statusCode, response)

	local response_headers = {
		["Content-Type"] = "application/json",
		['Content-Length'] = #response
	}

	return statusCode, response_headers, coroutine.wrap(function()
		coroutine.yield(response)
	end)
end

local function executeReportCommon (wsapi_env, reportId, GetPermissionId, callback, overrideFields)
	local _T = makeTextTranslator()
	local request = wsapi_request.new(wsapi_env)

	LogInfo("[REPORTS:executeReportCommon] Executing common report %s. Method: '%s'", reportId, request.method)

	local input_readers = {
		POST = function()
			local post_data = wsapi_env.input:read(wsapi_env.input.length)
			--LogDebug("[REPORTS:executeReportCommon] 1 ---> Report %s POST data: '%s'", reportId, post_data)
			return pcall(json.decode, post_data)
		end,
		GET = function()
			--LogDebug("[REPORTS:executeReportCommon] 2 ---> Report %s GET data: '%s'", reportId, request.params.data)
			return pcall(json.decode, request.params.data)
		end
	}

	if not input_readers[request.method] then
		return wrapErrorResponse(405, _T("Invalid request method"))
	end

	local res, queryParams = input_readers[request.method]()
	if not res or not queryParams then
		return wrapErrorResponse(400, _T("Invalid request format"))
	end

	if overrideFields and queryParams.queryFields then
		for k, v in pairs(overrideFields) do
			queryParams.queryFields[k] = v
		end
	end

	local permissionId = GetPermissionId(queryParams)
	if permissionId then
		local hasPermission = permissions.IsPermissionEnabledForLoggedUser("reports", permissionId)
		if not hasPermission then
			return wrapErrorResponse(403, _T("You not have permissions."))
		end
	end

	local report, err = getReportHelper(reportId, _T)
	if not report then
		return wrapErrorResponse(err and 500 or 404, _T(err or "Report not found"))
	end

	-- and then invoke the template method
	return callback(
			makeQueryRunner(report, queryParams, _T),
			report,
			queryParams
		)
end

--[[
	Handler for reports/execute
		Esta funcion termina ejecutando las funciones de la rocka "reporter.QueryDetail" entre otras.
--]]
local function executeReport (wsapi_env, reportId)
	local _T = makeTextTranslator()

	LogInfo("[REPORTS:executeReport] Executing report %s", reportId)

	local function GetPermissionId(_)
		return "reports_viewer"
	end

	return executeReportCommon(wsapi_env, reportId, GetPermissionId, function(runQuery, report)
		local response_headers = {
			['Content-Type']                    = 'application/json',
			['Transfer-Encoding']               = 'chunked',  -- replaces Content-Length
			['x-inConcert-Buffering-Disable']   = 'yeah',     -- disable buffering
			['x-inConcert-Compression-Disable'] = 'yes'       -- make clear compression is disabled
		}

		--test runner
		--local result = runQuery()
		--LogDebug("PRUEBA %s", json.encode(result))

		return 200, response_headers, coroutine.wrap(function()
			-- transfer-encoding = chunked
			local function make_chunk(data)
				local formattedData = string.format("%x\r\n%s\r\n", #data, data)
				--LogInfo("[REPORTS:executeReport.make_chunk] Report '%s. coroutine.wrap .. formattedData: '%s'.", reportId, formattedData)
				return formattedData
			end

			local response
			local ok, result, last_handle, err_or_warn = xpcall(runQuery, handleReportError)

			if ok and result then
				response = {
					result = result,
					last_handle = last_handle
				}

				if err_or_warn and err_or_warn ~= "" then
					LogWarning("Executing report %s: %s", reportId, err_or_warn)
				end
			else
				local emsg = "Failed to execute report"

				if err_or_warn and type(err_or_warn) == "string" and err_or_warn ~= "" then
					emsg = err_or_warn
				end

				response = {
					result = _T(emsg)
				}

				LogError("[REPORTS:executeReport] Failed to execute %s report: %s", report.type, err_or_warn)
			end

			local encoded_response = json.encode(response)
			--LogDebug("--> encoded_response: %s", encoded_response)

			-- send chunk with complete body
			coroutine.yield(make_chunk(encoded_response))
			-- the final chunk
			coroutine.yield(make_chunk(""))
		end)
	end)
end

local function DeleteTemporalSourceContent (pathSourceContentDirectory)
	if not pathSourceContentDirectory then
		return
	end
	-- Idem al directorio temporal del archivo "detail_query.lua"
	local SOURCE_ROOT_PATH_TEMPORAL_FILES = "/tmp/"

	local sourcePath = SOURCE_ROOT_PATH_TEMPORAL_FILES .. pathSourceContentDirectory
	if string.sub(sourcePath, -1) == "/" then
		sourcePath=string.sub(sourcePath, 1, -2)
	end
	-- Se elimina el directorio temporal raiz, donde se encuentra el contexto temporal a subir al FTP
	LogInfo("[REPORTS:DeleteTemporalSourceContent] Remove temporal directory and files. Path: '%s'.", sourcePath)
	os.execute("rm -rf " .. sourcePath)
end

--[[
	Handler for reports/execute
		Esta funcion termina ejecutando las funciones de la rocka "reporter.QueryDetail" entre otras.
--]]
local function exportReport (wsapi_env, reportId)
	local _T = makeTextTranslator()

	LogInfo("[REPORTS:exportReport] Exporting report '%s'", reportId)

	 --[[
			Cuando la ejecucion de la exportacion del reporte se realiza mediante una accion del lado del cliente, se debe de validar el permiso.
			En caso de que a esta funcion la invoque el scheduler, no debe de validar el permiso.
	 ]]--
	local function GetPermissionId(queryParams)
		local permissionId
		if queryParams.executed_by_user then
			permissionId = "reports_viewer"
		end
		return permissionId
	end

	local start_time = socket.gettime()  -- Obtengo el tiempo de inicio
   
    local function hasExpired(timeout)
        return timeout and socket.gettime() - start_time >= timeout or nil
    end

	return executeReportCommon(wsapi_env, reportId, GetPermissionId, function(runQuery, report, queryParams)
		LogInfo("[REPORTS:exportReport:executeReportCommon] Exporting report '%s', queryParams: '%s'", reportId, json.encode(queryParams or {}))

		local exporter = exporter.open_exporter(queryParams.export_type, utils.DeepCopy(report), _T)
		if not exporter then
			LogInfo("[REPORTS:exportReport] Exporter %s not found", queryParams.export_type)
			return wrapErrorResponse(405, _T("Export type not supported"))
		end

		local filename = queryParams.filename or reportId
		runQuery = makeQueryExporter(runQuery, report)

		local content_disposition = "attachment;filename=".. filename
		if not filename:lower():match(".*%.".. exporter.interface.type:lower()) then
			content_disposition = content_disposition ..".".. exporter.interface.type
		end

		local response_headers = {
			["Content-Type"]                    = exporter.interface.content_type or "application/octet-stream",
			["Content-Disposition"]             = content_disposition,
			["Transfer-Encoding"]               = "chunked",  -- replaces Content-Length
			["x-inConcert-Buffering-Disable"]   = "yeah",     -- disable buffering
			["x-inConcert-Compression-Disable"] = "yes",      -- make clear compression is disabled
			["Set-Cookie"]                      = "downloadStarted=true; Path=/" -- used by the download plugin
		}

		local request_timeout = queryParams.timeout and tonumber(queryParams.timeout) or nil
		local sourceContentDirectory = queryParams.queryFields and queryParams.queryFields.saveContentTo

		--LogInfo("[REPORTS:exportReport] request_timeout %s sourceContentDirectory %s", queryParams.timeout, queryParams.queryFields.saveContentTo)
		return 200, response_headers, coroutine.wrap(function()
			local result, err
			local indexes = {} -- Used when check for invalid chars in string fields

			--LogDebug("[REPORTS:exportReport] 1 - coroutine.wrap")

			-- transfer-encoding = chunked
			local function make_chunk(data)
				return string.format("%x\r\n%s\r\n", #data, data)
			end

			local function get_more_data()
				result, err = runQuery()
				if not result then
					LogError("[REPORTS:exportReport] Something failed exporting report %s: %s", reportId, err)
				--else
					--LogDebug("[REPORTS:exportReport] 1 - coroutine.wrap 'get_more_data' result.rows: %s", #result.rows)
				end
				return (result ~= nil)
			end

			local function mark_row_header(row)
				table.insert(row, 1, "Header") -- first column
				return row
			end

			local function mark_nested_row(row)
				table.insert(row, 1, "Thread") -- second column
				table.insert(row, 1, "") -- first column
				return row
			end

			if not get_more_data() then
				return
			end

			local fnHasExpiredHttpRequest = function ()
				LogDebug("[REPORTS:exportReport] - Request timeout %s", reportId)
				DeleteTemporalSourceContent(sourceContentDirectory)
				coroutine.yield(make_chunk("REQUEST_HTTP_TIMEOUT"))
				-- the final chunk
				coroutine.yield(make_chunk(""))
			end

			-- Check request timeout 
			if sourceContentDirectory and hasExpired(request_timeout) then
				--LogDebug("[REPORTS:exportReport] 1 - fnHasExpiredHttpRequest")
				return fnHasExpiredHttpRequest()
			end

			if report.threaded then
				mark_row_header(result.headers)
			end

			--LogDebug("[REPORTS:exportReport] 2 - coroutine.wrap 'exporter:write_headers' starting")

			exporter:write_headers(result.headers, function(data)
				if report.threaded then
					mark_nested_row(result.threadingInfo.headers)
					exporter:write_row(result.threadingInfo.headers, function(newdata)
						if data and newdata then
							data = table.concat({ data, newdata })
						elseif newdata then
							data = newdata
						end
					end)
				end

				if data then
					coroutine.yield(make_chunk(data))
				end
			end)

			--LogDebug("[REPORTS:exportReport] 3 - coroutine.wrap 'loop rows' starting")

			while next(result.rows) do
				-- Check request timeout 
				if sourceContentDirectory and hasExpired(request_timeout) then
					--LogDebug("[REPORTS:exportReport] 2 - fnHasExpiredHttpRequest")
					return fnHasExpiredHttpRequest()
				end

				if report.threaded then
					local mixedrows = {}
					_.each(result.rows, function(row, idx)
						table.insert(mixedrows, mark_row_header(row))

						local rowThread = result.threadingInfo.records[idx]
						if rowThread and #rowThread > 0 then
							_.each(rowThread, function(nestedrow)
								table.insert(mixedrows, mark_nested_row(nestedrow))
							end)
						end
					end)
					result.rows = mixedrows
				end

				exporter:bulk_write(result.rows, function(data)
					if data then
						coroutine.yield(make_chunk(data))
					end
				end)

				if result.areAllRecords then
					--LogDebug("[REPORTS:exportReport] 3.1 - coroutine.wrap 'while rows' areAllRecords: %s", result.areAllRecords)
					break
				end
				
				local hasMoreData = get_more_data()
				--LogDebug("[REPORTS:exportReport] 3.2 - coroutine.wrap 'while rows' hasMoreData: %s", hasMoreData)
				if not hasMoreData then
					return
				end
			end

			--LogDebug("[REPORTS:exportReport] 4 - coroutine.wrap 'close' starting")

			exporter:close(function(close_data)
				if close_data then
					coroutine.yield(make_chunk(close_data))
				end
				-- Check request timeout 
				if sourceContentDirectory and hasExpired(request_timeout) then
					--LogDebug("[REPORTS:exportReport] 3 - fnHasExpiredHttpRequest")
					fnHasExpiredHttpRequest()
				else
					-- the final chunk
					coroutine.yield(make_chunk(""))
				end
			end)
		end)
	end, { withCache = true })
end

local function saveReport(web, request, ks)
	local _T = makeTextTranslator()
	local report = request

	if not report.name or not report.type or not report.description then
		return replyJson(web, 400, _T("Invalid Request"))
	end

	if not m_allowedReportsTypes[report.type] then
		return replyJson(web, 400, _T("Invalid report type"))
	end

	if report.type == "summary" then
		if not report.indicators then
			return replyJson(web, 400, _T("Cannot create summary report with no metrics"))
		end
		if not report.columnDetails then
			return replyJson(web, 400, _T("columnDetails is missing"))
		end
		if not report.cut then
			return replyJson(web, 400, _T("Summary filter is missing"))
		end
		-- TODO: Validate cut and indicators are both valid
		-- TODO: Validate columnDetails references all indicators
	else
		-- TODO: poor man's validations
	end

	local report_id = report.id or uuid.new()
	local vcc = inconcert.authUser.virtualContactCenter

	report.mutable = true
	LogInfo("Saving report %s@%s - %s", report_id, vcc, json.encode(report))

	report.__designer = true
	local success, err = manager.setReport(ks, vcc, report_id, report, force)
	if not success then
		LogError("Failed to save report %s@%s (%s): %s", report_id, vcc, report.name, err)
		return replyJson(web, 500, err)
	end

	return replyJson(web, 200, report_id)
end

local function deleteReport(web, request, ks)
	local vcc = inconcert.authUser.virtualContactCenter

	if not request.id then
		return replyJson(web, 400, _T("Invalid Request"))
	end

	LogInfo("Deleting report %s@%s", request.id, vcc)

	local success, err = manager.deleteReport(ks, vcc, request.id)
	if not success then
		LogError("Failed to delete report %s@%s: %s", request.id, vcc, err)
		return replyJson(web, 500, _T(err))
	end

	return replyJson(web, 200, "OK")
end

--------------------------------------------------[[
--------------------------------------------------
------------ METRICS
--------------------------------------------------
--------------------------------------------------]]

local function transformMetric(metric, composite, metrics, _T)
	putDiagnosticData(metric)

	metric.name = _T(metric.name or metric.id)
	metric.composite = metric.composite or (composite == true)
	metric.description = _T(metric.description or "")

	local metadata = metric.metadata or {}
	metadata.requiredFieldsHidden = metadata.requiredFieldsHidden or {}
	metadata.requiredFieldsMapping = metadata.requiredFieldsMapping or {}

	if not metric.composite then
		metric.requiredFields = _.map(metric.requiredFields or {}, function(v, k)
			if metadata.requiredFieldsHidden[k] then
				return nil
			end
			return {
				dimension  = metadata.requiredFieldsMapping[k] or k,
				comparator = "==",
				value      = v
			}
		end, false) --includeNil=false

		if metadata.requiredFieldsExtra then -- array of {dimension, comparator, value}
			_.concat(metric.requiredFields, metadata.requiredFieldsExtra, true) -- true = mutate metric.requiredFields
		end

		metadata.expression = metadata.expression or {
			type      = "complex",
			connector = "and",
			positive  = true,
			childs = _.map(metric.requiredFields, function(e)
				return {
					type       = "simple",
					field      = { id = e.dimension },
					comparator = e.comparator,
					value      = e.value
				}
			end)
		}

		metadata.requiredFieldsHidden = nil -- not needed
		metadata.requiredFieldsMapping = nil -- not needed
		metadata.requiredFieldsExtra = nil
		metric.script = nil
	else
		metric.formula = nil

		if not metric.format and metrics then
		-- format no set, resolve from the first indicator with format
			for _,v in pairs(metric.indicators) do
				local m = metrics[v]
				if m and m.format then
					metric.format = m.format
					break
				end
			end
		end
	end

	return metric
end

local function listMetrics(web)
	local _T = makeTextTranslator()
	web.headers["Content-Type"] = "application/json"

	local metrics, err = listAllByEntity(
			"metric", manager.listSummaryIndicators, manager.getSummaryIndicator,
			function(compositeMetric)
				return transformMetric(compositeMetric, false, nil, _T)
			end
		)
	if not metrics then
		return replyJson(web, 500, err)
	end

	local ok, err = listAllByEntity(
			"composite_metric", manager.listCompositeIndicators, manager.getCompositeIndicator,
			function(compositeMetric)
				if metrics[compositeMetric.id] then
					LogWarning("Metric %s will be overwritten", compositeMetric.id)
				end
				metrics[compositeMetric.id] = transformMetric(compositeMetric, true, metrics, _T)
			end
		)

	return ok and replyJson(web, 200, metrics) or replyJson(web, 500, err)
end


--[[
--
--]]
local function saveCompositeMetric(web, request, ks, metric)
	local _T = makeTextTranslator()
	metric.indicators = {}
	local operation = metric.metadata.operation

	if operation.name == "sum" then
		if type(operation.sum.values) ~= "table" or not next(operation.sum.values) then
			return replyJson(web, 400, _T("Invalid Operation Data"))
		end

		local addends = _.map(operation.sum.values, function(m)
			table.insert(metric.indicators, m.id) -- we're iterating already..
			return string.format('base["%s"]', m.id)
		end)
		metric.formula = string.format(
				"return %s", table.concat(addends, ' + ')
			)
	elseif operation.name == "average" then
		if type(operation.average.count) ~= "table" or #operation.average.count ~= 1 then
			return replyJson(web, 400, _T("Invalid Operation Data"))
		end
		if type(operation.average.sum) ~= "table" or #operation.average.sum ~= 1 then
			return replyJson(web, 400, _T("Invalid Operation Data"))
		end

		metric.indicators = { operation.average.count[1].id, operation.average.sum[1].id }
		metric.formula = string.format(
				'return base["%s"] / base["%s"]', operation.average.sum[1].id, operation.average.count[1].id
			)
	elseif operation.name == "percentage" then
		if type(operation.percentage.total) ~= "table" or #operation.percentage.total ~= 1 then
			return replyJson(web, 400, _T("Invalid Operation Data"))
		end
		if type(operation.percentage.value) ~= "table" or #operation.percentage.value ~= 1 then
			return replyJson(web, 400, _T("Invalid Operation Data"))
		end

		metric.format = { type = "Percentage" }
		metric.indicators = { operation.percentage.total[1].id, operation.percentage.value[1].id }
		metric.formula = string.format(
				'return base["%s"] * 100 / base["%s"]', operation.percentage.value[1].id, operation.percentage.total[1].id
			)
	elseif operation.name == "sub" then
		if type(operation.sub.minuend) ~= "table" or #operation.sub.minuend == 0 then
			return replyJson(web, 400, _T("Invalid Operation Data"))
		end
		if type(operation.sub.subtrahend) ~= "table" or #operation.sub.subtrahend == 0 then
			return replyJson(web, 400, _T("Invalid Operation Data"))
		end

		local addends = _.map(operation.sub.minuend, function(m)
			table.insert(metric.indicators, m.id) -- we're iterating already..
			return string.format('base["%s"]', m.id)
		end)

		local subtrahends = _.map(operation.sub.subtrahend, function(m)
			table.insert(metric.indicators, m.id) -- we're iterating already..
			return string.format('base["%s"]', m.id)
		end)

		metric.formula = string.format(
				"return %s - %s", table.concat(addends, ' + '), table.concat(subtrahends, ' - ')
			)
	end

	LogInfo("Saving composite metric %s", json.encode(metric))

	local success, err = manager.setCompositeIndicator(ks, metric.vcc, metric.id, metric)
	if not success then
		LogError("Error %s saving composite metric %s", err, json.encode(metric))
		return replyJson(web, 500, _T(err))
	end

	return replyJson(web, 200, transformMetric(metric, false, nil, _T))
end


--[[
--
--]]
local function saveSimpleMetric(web, request, ks, metric)
	local _T = makeTextTranslator()

	-- check custom metric licenses
	local limit = model.getCustomMetricsLimit(inconcert.authUser.virtualContactCenter)
	local customMetrics = manager.listSummaryIndicators(ks, inconcert.authUser.virtualContactCenter)

	if not _.contains(customMetrics, metric.id) and #customMetrics >= limit then
		LogWarning("Error saving metric %s. You ran out of licenses", json.encode(metric))
		return replyJson(web, 402 , _T("You ran out of licenses"))
	end

	-- ensure those fields are set when comparing
	local fieldsThatMustBeSet = {
		"StateRaw"
	}

	metric.scope = request.scope
	-- this is the list of fields the summarizer checks before executing the metric
	metric.requiredFields = { }
	metric.extras = request.extras and {
		MinMaxIndicator            = request.extras.MinMaxIndicator or nil,
		IncrementsPeriod           = request.extras.IncrementsPeriod or nil,
		IncrementsPeriodByDuration = request.extras.IncrementsPeriodByDuration or nil
	} or {}
	-- reference to dimensions in use
	metric.metadata.dimensionsInUse = {}
	-- there are some requiredFields that must be hidden, only used by the client
	metric.metadata.requiredFieldsHidden = {}
	-- requireFields needs a name, we map that name with the dimension ID here
	metric.metadata.requiredFieldsMapping = {}

	-- handy, looks up for the specified dimension in the database
	local function getDimension(dimensionId)
		local d, err
		_.some({ "system", inconcert.authUser.virtualContactCenter }, function(vcc)
			d, err = manager.getDimension(ks, vcc, dimensionId)
			return d
		end)
		return d, err
	end

	-- handy, looks up for the specified filter in the database
	local function getFilter(filterId)
		local d, err
		_.some({ "system", inconcert.authUser.virtualContactCenter }, function(vcc)
			d, err = manager.getCut(ks, vcc, filterId)
			return d
		end)
		return d, err
	end

	-- converts a dimension to something we can use in requiredFields
	local function getDimensionKey(dimension)
		-- webchat and cci are somewhat special
		if dimension.scope and customScopes[dimension.scope] then
			return dimension.scope ..".".. dimension.name
		end
		if dimension.compatibleWith and dimension.compatibleWith[metric.scope] then
			return dimension.compatibleWith[metric.scope]
		end
		-- anyway, this one must be builtin
		return dimension.id
	end

	-- casts a value to be used in required fields
	-- allows to do something like { Agent = "john" } or { IsTaken = false }
	local function processRequiredField(dimensionId, value)
		local dimension, err = getDimension(dimensionId)
		if not dimension then
			LogWarning("Unknown dimension '%s' in required fields (%s)", dimensionId, err)
			return dimensionId, value
		end

		if not _.contains(metric.metadata.dimensionsInUse, dimension.id) then
			table.insert(metric.metadata.dimensionsInUse, dimension.id)
		end

		if dimension.type == "number" or dimension.type == "Time" then
			value = tonumber(value)
		elseif dimension.type == "boolean" then
			if type(value) ~= "boolean" then
				value = (value == "true")
			end
		else
			value = tostring(value) -- just in case (value should be string already)
		end

		return getDimensionKey(dimension), value, dimension.hidden, dimension.byCampaign
	end

	local function serializeExpression(formula, campaignScope)
		local function negate(expr)
			return "not(" .. expr .. ")"
		end
		local function key2detail(key, scope)
			scope = scope or "detail"
			return string.format(scope.."['%s']", key)
		end
		local function quote(value)
			return type(value) == "string" and ('"' .. value .. '"') or tostring(value)
		end

		if formula.type == "simple" then
			local key, value, hidden, byCampaign = processRequiredField(formula.field.id, formula.value)

			local scope = byCampaign and campaignScope or nil

			local comparator = formula.comparator == "<>" and "~=" or formula.comparator
			if type(value) == "boolean" and comparator == "~=" then
				value = not value
				comparator = "=="
			end

			local expr = table.concat({ key2detail(key, scope), comparator, quote(value) }, ' ')

			if (comparator ~= "==" and _.contains(fieldsThatMustBeSet, key)) then
				expr = "(" .. key2detail(key, scope)  .." and ".. expr ..")"
			end

			return expr

		elseif formula.type == "complex" then
			local expr = table.concat( _.map(formula.childs, function(e)
				local ef = serializeExpression(e, campaignScope)
				if not formula.positive then
					ef = negate(ef)
				end
				return ef
			end), " ".. formula.connector .." ")
			return #expr > 0 and ( "( " .. expr .. " )" ) or ""
		end
	end

	_.each(request.requiredFields or { }, function(field)
		local key, value, hidden = processRequiredField(field.dimension, field.value)

		metric.requiredFields[key] = value
		if key ~= field.dimension then
			metric.metadata.requiredFieldsMapping[key] = field.dimension
		end
		if hidden then
			metric.metadata.requiredFieldsHidden[key] = true
		end
	end)

	local chunks = {
		nesting = 0,
		fbegin  = { },
		fend    = { }
	}
	local nested = function(f)
		chunks.nesting = chunks.nesting + 1
		f()
		chunks.nesting = chunks.nesting - 1
	end
	local append = function(code, target)
		table.insert(target or chunks.fbegin, {
			nesting = chunks.nesting,
			code = code
		})
	end

	-- here is where the if ... then is built
	local function addConditon(campaignScope)
		if request.metadata.expression then
			local cond = serializeExpression(request.metadata.expression, campaignScope)
			if #cond > 0 then
				append("if ".. cond .." then")
				append("end", chunks.fend)
				chunks.nesting = chunks.nesting + 1
			end
		end
	end

	local function GetFilterFields()
		if string.find(metric.metadata.filter, "custom_") then
			local filterId = metric.metadata.filter:gsub("custom_", "")
			local filter, err = getFilter(filterId)
			if not filter then
				return nil, "Error getting custom filter ".. filterId..": "..err
			end
			if filter.scope ~= metric.scope then
				return nil, "Filter scope is not the same that metric scope"
			end
			local fields = {}
			for _, field in pairs(filter.columns) do
				if field ~= "Slice" and field ~= "Shift" then
					fields[field] = true
				end
			end
			return fields
		else
			if metric.scope == "Interaction" then
				return {
					Type = true,
					Campaign = true,
					Account = true,
					DispositionCode = true,
					Agent = metric.metadata.filter == "CampaignAgent" or nil,
					Page = metric.metadata.filter == "CampaignFacebookPost" or nil,
					Post = metric.metadata.filter == "CampaignFacebookPost" or nil,
					Tweet = metric.metadata.filter == "CampaignTweet" or nil,
					BatchId = metric.metadata.filter == "DialerCampaign" or nil,
					PhoneType = metric.metadata.filter == "DialerCampaign" or nil,
				}
			elseif metric.scope == "InteractionAgent" then
				return {
					Type = true,
					Campaign = true,
					Account = true,
					DispositionCode = true,
					Agent = true,
				}
			elseif metric.scope == "InteractionAttentionLevel" then
				return {
					Type = true,
					Campaign = true,
					Account = true,
					DispositionCode = true,
					AttentionLevel = true,
				}
			elseif metric.scope == "InteractionCount" then
				return {
					Type = true,
					Campaign = true,
				}
			elseif metric.scope == "InteractionRule" then
				return {
					Type = true,
					Campaign = true,
					Account = true,
					Direction = true,
					Rule = true,
				}
			elseif metric.scope == "InteractionVoiceBlaster" then
				return {
					Campaign = true,
					BatchId = true,
					VoiceBlasterId = true,
					MenuId = true,
				}
			elseif metric.scope == "Agent" then
				return {
					Type = true,
					Agent = true,
					Campaign = metric.metadata.filter == "AgentCampaign",
				}
			end
		end
		return fields
	end

	local filterFields, err = GetFilterFields()
	if not filterFields then
		LogError("Error (%s) getting fields to save metric %s", err, json.encode(metric))
		return replyJson(web, 500, _T(err))
	end

	--in this case the condition is inside 
	local isAgentScopeForCampaign = metric.scope == "Agent" and filterFields.Campaign

	if not isAgentScopeForCampaign then
		addConditon()
	end

	local increment
	if metric.metadata.increment == "default" then
		increment = "1"
	elseif metric.metadata.increment == "time_lapse" then
		increment = "extras.ToDouble - extras.FromDouble"
	else
		local incrementScope = isAgentScopeForCampaign and "campaignData" or "detail"
		increment = incrementScope.."['".. metric.metadata.increment .. "']"
	end

	-- build fields table
	local fields = {
		VCC = "detail.VCC"
	}

	if metric.scope == "InteractionCount" then
		increment       = [[detail.amount]]
		fields.Date     = [[detail.timestamp]]
		for field, _ in pairs(filterFields) do
			fields[field] = "detail."..field:lower()..' or "N/A"'
		end

	elseif metric.scope:match('^Interaction.*') then
		metric.sub_scope = metric.metadata.interactionType
		fields.Date            = [[extras.EndDateDouble]]
		fields.Shift           = [[helpers.GetShift(extras.EndDateDouble)]]

		for field, _ in pairs(filterFields) do
			if field == "Agent" and (metric.scope == "Interaction" or metric.scope == "InteractionAttentionLevel") then
				fields.Agent = 'agentId'
				append("for _, agentId in ipairs(detail.Agent) do")
				append("end", chunks.fend)
				chunks.nesting = chunks.nesting + 1
			else
				fields[field] = "detail."..field..' or "N/A"'
			end
		end

		local cciProject   = metric.cciProject or ""
		local batch_dataProject   = metric.batch_dataProject or ""
		local process_flowProject   = metric.process_flowProject or ""
		local integrationType = metric.integrationType or ""

		if cciProject ~= "" or batch_dataProject ~= "" or process_flowProject ~= "" or integrationType ~= "" then
			-- allow to group by all dimensions on each project
			_.each(metric.metadata.dimensions, function(d)
				local key = d.name
				fields[key] = "detail['".. key .. "']"
				if not metric.requiredFields[key] then
					fields[key] = string.format("%s ~= nil and tostring(%s) or nil", fields[key], fields[key])
				end
			end)

			if cciProject ~= "" then
				fields["cci.__SID"] = "detail['cci.__SID']"
			end

			if batch_dataProject ~= "" then
				fields["batch_data.__SID"] = "detail['batch_data.__SID']"
			end

			if process_flowProject ~= "" then
				fields["process_flow.__SID"] = "detail['process_flow.__SID']"
			end

			if integrationType ~= "" then
				fields[integrationType..".__SID"] = "detail['"..integrationType..".__SID']"
			end

			-- not needed anymore
			metric.metadata.dimensions = nil
		end

	elseif metric.scope == "Agent" then
		if metric.extras.IncrementsPeriod or metric.extras.IncrementsPeriodByDuration then
			append("local shifts = helpers.GetShifts(extras.LapseFromDouble, extras.DateDouble)")

			append("for _, shift in ipairs(shifts) do")
			append("end", chunks.fend)
			chunks.nesting = chunks.nesting + 1

			fields.DateFrom  = [[shift.start]]
			fields.DateTo    = [[shift.finish]]
			fields.Shift     = [[shift.shift]]

		else
			fields.Date  = "extras.DateDouble"
			fields.Shift = "helpers.GetShift(extras.FromDouble)"
		end

		for field, _ in pairs(filterFields) do
			if field == "Campaign" then
				fields.Campaign = "campaign"

				append("for campaign, campaignData in ipairs(extras.Campaigns) do")

				if isAgentScopeForCampaign then
					addConditon("campaignData")
				end

				append("end", chunks.fend)
			else
				fields[field] = "detail."..field..' or "N/A"' 
			end
		end
	else
		LogError("Metric scope %s not supported: %s", metric.scope, json.encode(metric))
		return replyJson(web, 500, _T("Metric scope not supported"))
	end

	metric.metadata.groupFields = {}
	append("table.insert(results, {")
	nested(function()
		append("fields = {")
		nested(function()
			_.each(fields, function(v, k)
				append(string.format("['%s'] = %s,", k, v))
				table.insert(metric.metadata.groupFields, k)
			end)
		end)
		append(" },")
		append(string.format("increment = math.floor(%s)", increment))
	end)
	append("})")

	local formula = table.concat(_.map(_.concat(chunks.fbegin, chunks.fend), function(v)
		return ( string.rep("  ", v.nesting) .. v.code )
	end), "\n")

	LogInfo("Saving metric %s with formula:\n[[\n%s\n]]\n", json.encode(metric), formula)

	metric.script = formula

	local success, err = manager.setSummaryIndicator(ks, metric.vcc, metric.id, metric)
	if not success then
		LogError("Error %s saving metric %s", err, json.encode(metric))
		return replyJson(web, 500, _T(err))
	end

	return replyJson(web, 200, transformMetric(metric, false, nil, _T))
end

--[[
-- Creates a new metric
--]]
local function saveMetric(web, request, ks)
	local _T = makeTextTranslator()
	-- force metric has at leats name and format
	if not request.name or not request.format then
		return replyJson(web, 400, _T("Invalid Request"))
	end

	if type(request.defaultMaxLength) ~= "boolean" then
		return replyJson(web, 400, _T("Default max length must be boolean"))
	end

	local metric = utils.DeepCopy(request)
	metric.__designer = true
	metric.id         = metric.id or uuid.new() -- id is not set when creatig a new metric
	metric.vcc        = inconcert.authUser.virtualContactCenter

	if metric.composite then
		return saveCompositeMetric(web, request, ks, metric)
	else
		return saveSimpleMetric(web, request, ks, metric)
	end -- not composite
end

local function deleteMetric(web, request, ks)
	local _T = makeTextTranslator()
	local vcc = inconcert.authUser.virtualContactCenter

	if not request.id then
		return replyJson(web, 400, _T("Invalid Request"))
	end

	local function checkReports()
		return not _.some(manager.listReports(ks, vcc, "summary"), function(id)
			local report = manager.getReport(ks, vcc, id)
			return report and _.contains(report.indicators, request.id)
		end), _T("Metric is being used by one or more reports")
	end

	LogInfo("Deleting metric %s@%s.", request.id, vcc)

	local status = 400

	local success, err = checkReports()
	if success then
		status = nil
		if request.composite then
			success, err = manager.deleteCompositeIndicator(ks, vcc, request.id)
		else
			success, err = manager.deleteSummaryIndicator(ks, vcc, request.id)
		end
	end

	if not success then
		LogError("Error %s deleting metric (request: %s)", err, json.encode(request))
		return replyJson(web, status or 500, _T(err))
	end

	return replyJson(web, 200, "OK")
end

--------------------------------------------------[[
--------------------------------------------------
------------ FILTERS
--------------------------------------------------
--------------------------------------------------]]

local function makeFilterTransformFunction(filterType, vccIsCustom)
	local transformationTable = {
		detail     = { "mandatoryFields", "optionalFields", "sortFields" },
		livedetail = { "mandatoryFields", "optionalFields", "sortFields" },
		summary    = { "filter", "columns" }
	}

	local propertyIterator = vccIsCustom and
		function(filter, p)
			filter[p] = filter.metadata[p]
		end
		or
		function(filter, p)
			if filter.vcc == "system" then
				filter[p] = _.map(filter[p], function(e) return { id = e } end)
			else
				filter[p] = filter.metadata[p]
			end
		end

	return function(filter, id, _T)
		putDiagnosticData(filter)

		filter.type  = filterType
		filter.id    = filter.id or id
		filter.name  = filter.name or filter.id
		filter.label = _T(filter.name)

		_.each(transformationTable[filterType], function(p)
			propertyIterator(filter, p)
		end)

		filter.metadata = nil -- not needed client-side (yet)
		return filter
	end
end

local function saveFilter(web, request, ks)
	local _T = makeTextTranslator()
	local vcc = inconcert.authUser.virtualContactCenter

	local function validateRequest()
		if not m_allowedReportsTypes[request.type] then
			return false
		end
		if not request.name then
			return false
		end
		return true
	end

	if not validateRequest() then
		return replyJson(web, 400, _T("Invalid Request"))
	end

	local filter = utils.DeepCopy(request)
	filter.__designer     = true
	filter.id             = filter.id or uuid.new()
	filter.requiredFields = filter.requiredFields or { }
	filter.metadata       = {
		-- detail or livedetail
		mandatoryFields = filter.mandatoryFields,
		optionalFields  = filter.optionalFields,
		sortFields      = filter.sortFields,
		-- summary
		filter          = filter.filter,
		columns         = filter.columns,
	}

	local customFilters = manager.listFilters(ks, vcc)
	local customCuts = manager.listCuts(ks, vcc)
	local totalFilters = #customFilters + #customCuts

	-- check filter licenses
	local limit = model.getCustomFiltersLimit(vcc)

	if totalFilters >= limit and not (_.contains(customFilters, filter.id) or _.contains(customCuts, filter.id)) then
		LogWarning("Error saving filter %s. You ran out of licenses", json.encode(filter))
		return replyJson(web, 402 , _T("You ran out of licenses"))
	end

	if vcc ~= "system" then
		filter.requiredFields["VCC"] = vcc
	end

	local function getDimensionField(idc)
		local d, err
		for __, searchVcc in ipairs({ "system", vcc }) do
			d, err = manager.getDimension(ks, searchVcc, idc.id)
			if d then
				if d.compatibleWith and d.compatibleWith[filter.scope] then
					return d.compatibleWith[filter.scope]
				end
				if d.scope and customScopes[d.scope] then
					return d.scope ..".".. d.name
				end
				-- anyway, this one must be builtin
				return d.id
			end
		end
		return nil, err
	end

	if filter.type == "detail" or filter.type == "livedetail" then
		if not filter.dateField or not filter.sortFields or #filter.sortFields == 0 then
			return replyJson(web, 400, _T("Invalid Request"))
		end

		filter.mandatoryFields = _.map(filter.metadata.mandatoryFields, getDimensionField)
		filter.optionalFields  = _.map(filter.metadata.optionalFields, getDimensionField)
		filter.sortFields      = _.map(filter.metadata.sortFields, getDimensionField)
		table.insert(filter.mandatoryFields, "VCC")

		LogInfo("Saving %s filter %s", filter.type, json.encode(filter))

		local success, err = manager.setFilter(ks, vcc, filter.id, filter)
		if not success then
			LogError("Error %s saving %s filter %s", err, filter.type, json.encode(filter))
			return replyJson(web, 500, _T(err))
		end
	else -- summary
		if not filter.columns or #filter.columns == 0 then
			return replyJson(web, 400, _T("Invalid Request"))
		end

		filter.filter  = _.map(filter.metadata.filter, getDimensionField)
		filter.columns = _.map(filter.metadata.columns, getDimensionField)
		table.insert(filter.filter, "VCC")

		LogInfo("Saving summary filter %s", json.encode(filter))

		local success, err = manager.setCut(ks, vcc, filter.id, filter)
		if not success then
			LogError("Error %s saving summary filter %s", err, json.encode(filter))
			return replyJson(web, 500, _T(err))
		end
	end

	return replyJson(web, 200, makeFilterTransformFunction(filter.type, filter.id)(filter, filter.id, _T))
end

local function deleteFilter(web, request, ks)
	local _T = makeTextTranslator()
	local vcc = inconcert.authUser.virtualContactCenter

	if not request.id then
		return replyJson(web, 400, _T("Invalid Request"))
	end

	local function checkReports(reportType)
		local attr
		if reportType == "summary" then
			attr = "cut"
		elseif reportType == "detail" then
			attr = "filterName"
		elseif reportType == "livedetail" then
			attr = "filterName"
		else
			assert(false, "deleteFilter - Unrecognized report reportType " .. tostring(reportType))
		end
		return not _.some(manager.listReports(ks, vcc), function(id)
			local report = manager.getReport(ks, vcc, id)
			return report and report[attr] == request.id
		end), _T("Filter is being used by one or more reports")
	end

	LogInfo("Deleting %s filter %s@%s.", request.type, request.id, vcc)

	local status = 400

	local success, err = checkReports(request.type)
	if success then
		status = nil
		if request.type == "summary" then
			success, err = manager.deleteCut(ks, vcc, request.id)
		elseif request.type == "detail" then
			success, err = manager.deleteFilter(ks, vcc, request.id)
		elseif request.type == "livedetail" then
			success, err = manager.deleteFilter(ks, vcc, request.id)
		end
	end

	if not success then
		LogError("Error %s deleting %s filter (request: %s)", request.type, err, json.encode(request))
		return replyJson(web, status or 500, _T(err))
	end

	return replyJson(web, 200, "OK")
end


--------------------------------------------------
------------ DIMENSIONS
--------------------------------------------------

local function transformDimension(d, id, _T)
	d.id            = d.id or id
	d.name          = d.name or d.id
	d.label         = _T(d.displayName or d.name or "")
	d.description   = _T(d.description or "")
	d.format        = d.format or ""
	d.multiplier    = d.multiplier or 1
	d.isProjectable = type(d.isProjectable) == "nil" or d.isProjectable
	d.isCondition   = type(d.isCondition) == "nil" or d.isCondition
	d.isGrouper     = type(d.isGrouper) == "nil" or d.isGrouper
	if d.scope and customScopes[d.scope] then
		d.name = d.scope ..".".. d.name
	end
	if d.scope and scopeLabels[d.scope] then
		d.scopeLabel = _T(scopeLabels[d.scope])
	end
	if (d.domain) then
		for k, v in pairs(d.domain) do
			if (type(v) == "table") then
				v.label = _T(v.label or "")
			end
		end
	end
	return d
end

local function saveDimension(web, request, ks)
	local _T = makeTextTranslator()
	local vcc = inconcert.authUser.virtualContactCenter

	if not request.name or not request.type then
		return replyJson(web, 400, _T("Invalid Request"))
	end

	if request.type == "enum" and not (type(request.domain) == "table" and #request.domain > 0) then
		return replyJson(web, 400, _T("Invalid Enumeration Domain"))
	end

  if type(request.defaultMaxLength) ~= "boolean" then
	  return replyJson(web, 400, _T("Default max length must be boolean"))
	end

	local dimension = {
		__designer  = true,
		id          = request.id or uuid.new(),
		scope       = request.scope or "api",
		name        = request.name,
		displayName = request.displayName,
		description = request.description,
		type        = request.type,
		domain      = request.domain,
		projectId   = request.projectId,
		maxLength = request.maxLength,
		defaultMaxLength = request.defaultMaxLength,
	}

	local success, err = manager.setDimension(ks, vcc, dimension.id, dimension)
	if not success then
		LogError("Error %s saving dimension %s", err, json.encode(filter))
		return replyJson(web, 500, _T(err))
	end

	return replyJson(web, 200, transformDimension(dimension, nil, _T))
end

local function deleteDimension(web, request, ks)
	local _T = makeTextTranslator()
	local vcc = inconcert.authUser.virtualContactCenter

	if not request.id then
		return replyJson(web, 400, _T("Invalid Request"))
	end

	local function checkMetrics()
		return not _.some(manager.listSummaryIndicators(ks, vcc), function(id)
			local metric = manager.getSummaryIndicator(ks, vcc, id)
			return metric.metadata and _.contains(metric.metadata.dimensionsInUse or {}, request.id)
		end), _T("Field is being used by one or more metrics")
	end

	local function checkReports()
		return not _.some(manager.listReports(ks, vcc, "detail"), function(id)
			local report = manager.getReport(ks, vcc, id)
			return _.some(report.projectFields, {
				dimension = request.id
			})
		end), _T("Field is being used by one or more reports")
	end

	local function checkFilters()
		return not _.some(manager.listFilters(ks, vcc), function(id)
			local filter = manager.getFilter(ks, vcc, id)
			-- TODO
		end) and not _.some(manager.listCuts(ks, vcc), function(id)
			local cut = manager.getCut(ks, vcc, id)
			-- TODO
		end), _T("Field is being used by one or more filters")
	end

	local status = 400

	local success, err = checkMetrics()
	if success then
		success, err = checkReports()
		if success then
			success, err = checkFilters()
			if success then
				status = nil
				success, err = manager.deleteDimension(ks, vcc, request.id)
			end
		end
	end

	if not success then
		LogError("Error %s deleting custom field (request: %s)", err, json.encode(request))
		return replyJson(web, status or 500, _T(err))
	end

	return replyJson(web, 200, "OK")
end

local function IncludeReportInfoToSchedule (schedule, currentVcc, ks, _T)
	assert(schedule)
	assert(currentVcc)
	assert(ks)
	_T = _T or makeTextTranslator()

	--LogDebug("[REPORTS:IncludeReportInfoToSchedule] Including Rpt Info to schedule id %s, vcc %s", schedule.Id, currentVcc)

	local reportId = schedule.Report or ""
	for _, vcc in pairs({"system", currentVcc}) do
		local column_family = "Reports"
		local key = vcc
		local res, report_info = pcall(ks.GetColumnValue, ks, key, column_family, reportId)

		if res and report_info then
			report_info = json.decode(report_info)
			report_info.description = _T(report_info.description)
			report_info.label = _T(report_info.name)
			report_info.typeLabel = _T(report_info.type)
			report_info.isCustomReport = vcc == currentVcc
			schedule.__report_info = report_info
			schedule.RepeatPolicy = schedule.RepeatPolicy or ""

			if schedule.__history then
				local history = {}
				for utc_date, info in pairs(schedule.__history) do
					local timezone_date = utils.ConvertUTCDateToDate( utc_date, nil, nil, true --[[ is full date --]] )
					local string_date = timezone_date:toString("%Y/%m/%d %H:%M:%S")
					info.__timestamp = timezone_date:time()
					history[string_date] = info
				end
				schedule.__history = history
			end
			break
		end
	end
end

local function listSchedules (web, post_data, ks)
	local _T = makeTextTranslator()
	local current_vcc = inconcert.authUser.virtualContactCenter
	local params = {
		vcc = current_vcc,
		history = {
			-- Retorna 40 registros en la subgrilla (historico) por cada row principal
			maxRows = 40,
			order   = "desc"
		}
	}

	local startTime = os.time()

	LogInfo("[REPORTS:listSchedules] 1 -(START) Loading Report Schedulers in VCC '%s'. Info: %s", current_vcc, json.encode(post_data))

	local schedules = scheduler.get_schedules_with_history(utils, params)
	local response = {
		vcc = current_vcc,
		rows = {},
		last = "",
		total_rows = utils.GetRowCount(schedules)
	}

	--LogDebug("[REPORTS:listSchedules] 1.1 - After retrieve 'schedulers with history'. Info: %s", json.encode(response))

	for schedule_id, schedule in pairs(schedules) do
		if utils.MatchFilter(schedule, post_data) then
			IncludeReportInfoToSchedule(schedule, current_vcc, ks, _T)
			table.insert(response.rows, schedule)
		end
	end

	--LogDebug("[REPORTS:listSchedules] 1.2 - After loop retrieve 'schedulers with history'.")

	post_data.limit = nil -- El limite de rows por pagina, no aplica para el caso de los schedulers, ya que el paginado se realiza clientside.
	post_data.__order_attr_name = "Report"
	local schedules_response = utils.MakeSearchResponse(response, post_data)

	--LogDebug("[REPORTS:listSchedules] 1.3 - After make search response 'schedulers with history'.")
	local jsonResult = replyJson(web, 200, schedules_response)
	--LogDebug("[REPORTS:listSchedules] 1.4 - Response %s.", jsonResult)

	local endTime = os.time()
	local elapsedTime = endTime - startTime

	LogInfo("[REPORTS.listSchedules] 2 - (END) Loaded Report Schedulers in VCC '%s'. Total Schedulers Rows: %d. ELAPSED Time '%s' seconds.", current_vcc, #schedules_response.rows, elapsedTime)

	return true and jsonResult
end

-- unicamente comapara la fecha si ha cambiado de dia (no la hora)
local function compare_repeat_date( repeat_date_hour )
	if not repeat_date_hour then
		return "equal"
	end
	local current = LuaOrm.Date( { year = LuaOrm.Date():year(), month = LuaOrm.Date():month(), day = LuaOrm.Date():day() } )
	local repeat_date = LuaOrm.Date( { year = repeat_date_hour:year(), month = repeat_date_hour:month(), day = repeat_date_hour:day() } )
	if repeat_date:time() < current:time() then
		return "minor"
	elseif repeat_date:time() > current:time() then
		return "major"
	end
	return "equal"
end

-- Calcula el valor a asignar en los campos "RepeatWeekday" "RepeatDay" del schedule.
-- Para el campo "RepeatWeekday" los valores posibles son del 1 al 7,
--    si el calculo es menor decrementa 1
--    si el calculo es mayor incrementa 1
--    en el caso que al decrementar llegue al valor 0, comienza en valor 7
--    en el caso que al incrementar llegue al valor 8, comienza en valor 0
---
--    La logica para el campo "RepeatDay" es igual a la anterior, pero con los valor maximo en 31
--
local function calculateOverflowRepeatInfo( repeat_date_hour, value, minimum, maximum )
	local value = tonumber( value )
	if not repeat_date_hour then
		return value
	end
	local changed = compare_repeat_date( repeat_date_hour ) -- invoca a funcion que compara si la fecha ha cambiado (depende del timezone de usario)
	if changed == "minor" then
		value = value - 1
		if value < minimum then
			value = maximum
		end
	elseif changed == "major" then
		value = value + 1
		if value > maximum then
			value = minimum
		end
	end
	return value
end

local FIRST_WEEK_DAY  = 1
local LAST_WEEK_DAY   = 7
local FIRST_MONTH_DAY = 1
local LAST_MONTH_DAY  = 31

local function convertScheduleDateToUTCDate( sch )
	if not sch then
		return
	end
	local format_hour = "%H:%M:%S"
	local format_date = "%Y/%m/%d"

	local formatted_date = utils.ConvertDateToUTCDate( sch.OneTimeDate, sch.OneTimeHour )
	if formatted_date then
		sch.OneTimeDate  = formatted_date:toString( format_date )
		sch.OneTimeHour  = formatted_date:toString( format_hour )
	end

	local repeat_date_hour = utils.ConvertHourToUTCHour( sch.RepeatHour )
	if repeat_date_hour then
		sch.RepeatHour = repeat_date_hour:toString( format_hour )
	end

	formatted_date = utils.ConvertDateToUTCDate( sch.StartsOnDate, sch.StartsOnHour or "00:00:00" )
	if formatted_date then
		sch.StartsOnDate = formatted_date:toString( format_date )
		sch.StartsOnHour = formatted_date:toString( format_hour )
	end

	formatted_date = utils.ConvertDateToUTCDate( sch.EndsOnDate, sch.EndsOnHour or "23:59:00" )
	if formatted_date then
		sch.EndsOnDate = formatted_date:toString( format_date )
		sch.EndsOnHour = formatted_date:toString( format_hour )
	end

	if sch.RepeatWeekday or sch.RepeatDay then
		if sch.RepeatWeekday then
			sch.RepeatWeekday = calculateOverflowRepeatInfo( repeat_date_hour, sch.RepeatWeekday, FIRST_WEEK_DAY, LAST_WEEK_DAY )
		end
		if sch.RepeatDay then
			sch.RepeatDay = calculateOverflowRepeatInfo( repeat_date_hour, sch.RepeatDay, FIRST_MONTH_DAY, LAST_MONTH_DAY )
		end
	end
end

local function convertScheduleUTCDateToDate( sch )
	if not sch then
		return
	end
	local format_hour = "%H:%M:%S"
	local format_date = "%Y/%m/%d"

	local formatted_date = utils.ConvertUTCDateToDate( sch.OneTimeDate, sch.OneTimeHour )
	if formatted_date then
		sch.OneTimeDate  = formatted_date:toString( format_date )
		sch.OneTimeHour  = formatted_date:toString( format_hour )
	end

	local repeat_date_hour = utils.ConvertUTCHourToHour( sch.RepeatHour )
	if repeat_date_hour then
		sch.RepeatHour   = repeat_date_hour:toString( format_hour )
	end

	formatted_date = utils.ConvertUTCDateToDate( sch.StartsOnDate, sch.StartsOnHour )
	if formatted_date then
		sch.StartsOnDate  = formatted_date:toString( format_date )
		sch.StartsOnHour  = formatted_date:toString( format_hour )
	end

	formatted_date = utils.ConvertUTCDateToDate( sch.EndsOnDate, sch.EndsOnHour )
	if formatted_date then
		sch.EndsOnDate  = formatted_date:toString( format_date )
		sch.EndsOnHour  = formatted_date:toString( format_hour )
	end

	if sch.RepeatWeekday or sch.RepeatDay then
		if sch.RepeatWeekday then
			sch.RepeatWeekday = calculateOverflowRepeatInfo( repeat_date_hour, sch.RepeatWeekday, FIRST_WEEK_DAY, LAST_WEEK_DAY )
		end
		if sch.RepeatDay then
			sch.RepeatDay = calculateOverflowRepeatInfo( repeat_date_hour, sch.RepeatDay, FIRST_MONTH_DAY, LAST_MONTH_DAY )
		end
	end
end
-------------------------------------------------

local function ApiExecute (self, reportId, queryParams)
	LogInfo("[REPORTS:ApiExecute] Api Executing Report '%s', queryParams: '%s'.", reportId, json.encode(queryParams or {}))
	queryParams.queryFields = queryParams.queryFields or {}
	queryParams.queryFields.withCache = true
	local _T = makeTextTranslator()
	local report, err
	if type(reportId) =="table" and queryParams.composite then
		report = reportId
	else
		report, err = getReportHelper(reportId, _T)
		if not report then
			LogError("[REPORTS:ApiExecute] Api Executing Report '%s' not found.", reportId)
			return false
		end
	end
	--LogDebug("[REPORTS:ApiExecute] Api Executing Report '%s', reportInfo: '%s'.", reportId, json.encode(report or {}))

	local runQuery = makeQueryRunner(report, queryParams, _T)
	--makeQuery
	LogInfo("[REPORTS:ApiExecute] Api Executing Report '%s', runQuery: '%s'.", reportId, runQuery)

	local fn = coroutine.wrap(function()
		---- transfer-encoding = chunked
		-- local function make_chunk(data)
		-- 	local formattedData = string.format("%x\r\n%s\r\n", #data, data)
		-- 	LogInfo("[REPORTS:executeReport.make_chunk] Report '%s. coroutine.wrap .. formattedData: '%s'.", reportId, formattedData)
		-- 	return formattedData
		-- end

		local response
		local result, last_handle, err_or_warn = runQuery()
	
		if result then
			response = {
				result = result,
				last_handle = last_handle
			}
			if err_or_warn and err_or_warn ~= "" then
				LogWarning("[REPORTS:ApiExecute] Executing report %s: %s", reportId, err_or_warn)
			end
		elseif err_or_warn == "other_node" then
			response = "other_node"
		else
			response = {
				result = _T("Failed to execute report")
			}
			LogError("[REPORTS:ApiExecute] Failed to execute %s report: %s", report and report.type, err_or_warn)
		end

		--local encodedResponse = json.encode(response)
		--LogDebug("[REPORTS:ApiExecute] --> encodedResponse: %s", encodedResponse)

		-- send chunk with complete body
		--coroutine.yield(make_chunk(encodedResponse))
		-- the final chunk
		--coroutine.yield(make_chunk(""))

		coroutine.yield(response)
	end)


	local resp = fn()

	--LogDebug("[REPORTS:ApiExecute] --> resp: %s", resp)

	return type(resp) == "table", resp
end
	-- return true, coroutine.wrap(function()
	-- 	-- transfer-encoding = chunked
	-- 	local function make_chunk(data)
	-- 		return string.format("%x\r\n%s\r\n", #data, data)
	-- 	end

	-- 	local response
	-- 	local result, last_handle, err_or_warn = runQuery()

	-- 	if result then
	-- 		response = {
	-- 			result = result,
	-- 			last_handle = last_handle
	-- 		}

	-- 		if err_or_warn and err_or_warn ~= "" then
	-- 			LogWarning("Executing report %s: %s", reportId, err_or_warn)
	-- 		end
	-- 	else
	-- 		response = {
	-- 			result = _T("Failed to execute report")
	-- 		}
	-- 		LogError("[REPORTS:executeReport] Failed to execute %s report: %s", report.type, err_or_warn)
	-- 	end

	-- 	local encoded_response = json.encode(response)
	-- 	--LogDebug("--> encoded_response: %s", encoded_response)

	-- 	-- send chunk with complete body
	-- 	coroutine.yield(make_chunk(encoded_response))
	-- 	-- the final chunk
	-- 	coroutine.yield(make_chunk(""))
	-- end)

--end

local api = {}

function GetApi(cassandra)
	Initialize()

	reporter.ConnectCassandra(cassandra.ip, cassandra.port)
	reporter.SetGlobalTimer(inconcert.environment.apps.agent.cluster)

	local reporterClient = GetReporterClient()
	reporterClient:ConfigureCassandra({
		{ host = cassandra.ip, port = cassandra.port }
	})

	-- local elasticsearch = newConfig.elasticsearch and newConfig.elasticsearch[1] and {
	-- 	host = newConfig.elasticsearch[1].host,
	-- 	port = tonumber(newConfig.elasticsearch[1].port)
	-- }
	-- LogInfo("Reports: Connecting fullcontact to elasticsearch on %s", json.encode(elasticsearch or {}))

	-- fullcontact.init({
	-- 	owner = "REPORTS",
	-- 	cassandra = cachandra,
	-- 	elasticsearch = elasticsearch,
	-- 	logger = MakeCompatibleLogger()
	-- })
	model.init({
		cassandra = cassandra
	})
	return api
end

api.Execute = ApiExecute
api.executeReport = executeReport
-------------------------------------------------





--[[
-- Save schedule
--]]
local function saveSchedule( web, request, ks )
	local _T = makeTextTranslator()
	local user = inconcert.authUser.userName
	local vcc = inconcert.authUser.virtualContactCenter
	local schedules = request or {}

-- TODO: Realizar validaciones
	--LogInfo("[REPORTS:saveSchedule] Saving schedule. User '%s', Vcc %s, schedule: '%s'.", user, vcc, json.encode( schedules ))

	local validate = function()
		local response = { success = true }
		for sch_id, sch in pairs( schedules ) do
			if not sch.isNew then
				local success, schedule = scheduler.get_schedule( utils, sch_id )
				if schedule.State == "running" then
					response.success = false
					response.result = "CURRENT_SCHEDULE_IS_RUNNING"
					response.message = _T("Current schedule is running.")
					response.code = 400
					break
				end
			end
		end
		return response
	end

	local is_valid = validate()
	if not is_valid.success then
		LogError("[REPORTS:saveSchedule] Error saving schedule. %s", is_valid.message)
		return replyJson( web, is_valid.code, is_valid )
	end

	for sch_id, sch in pairs( schedules ) do
		if sch.isNew then
			sch.Id = uuid.new()
		else
			sch.Id = sch_id
		end
		sch.isNew = nil
		sch.__report_info = nil -- si esta data llega se debe de eliminar, es solo utilizada del lado del cliente.
		sch.State = nil -- si esta data llega se debe de eliminar, es solo utilizada del lado del cliente.

		-- asingo valores por default
		sch.User = user
		sch.Vcc  = vcc
		sch.Params = sch.Params or {}

		convertScheduleDateToUTCDate( sch )
		local success, err = scheduler.add_schedule(utils, redis, sch)
		if not success then
			LogError("[REPORTS:saveSchedule] Unable to update schedule. %s", json.encode(err))
			local response = {}
			response.success = false
			response.result = err.result
			response.message = err.message or _T("Error updating scheduler.")
			response.code = 400
			return replyJson(web, response.code, response)
		end
	end

	LogInfo("[REPORTS:saveSchedule] Successfully schedule saved.")

	return replyJson( web, 200, {} )
end

local function getSchedule (web, sch_id)
	local _T = makeTextTranslator()
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("[REPORTS:getSchedule] Retrieving schedule id %s, user %s, vcc %s", sch_id, userId, vcc)

	local success, schedule = scheduler.get_schedule( utils, sch_id )

	if success then
		local ks, err = model.getCassandraConnection()
		IncludeReportInfoToSchedule(schedule, vcc, ks, _T)
		ks:Close()
		convertScheduleUTCDateToDate( schedule )
		return replyJson( web, 200, schedule )
	else
		web.status = 404
		local err = schedule
		LogError("[REPORTS:getSchedule] Error retrieving schedule id '%s'. Err: '%s'.", sch_id, json.encode(err))
		local msgError = "Error retriving schedule"
		if err.result == "SCHEDULE_DOES_NOT_EXIST" then
			msgError = "Schedule does not exist"
		end
		return replyJson(web, 404, _T(msgError))
	end
end

local function deleteSchedule( web, request, ks )
	local _T = makeTextTranslator()

	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	LogInfo("[REPORTS:deleteSchedule] Delete schedules. user %s, vcc %s", userId, vcc)

	local validate = function()
		local response = { success = true }
		for _, sch_id in pairs( request ) do
			local success, schedule = scheduler.get_schedule( utils, sch_id )
			if schedule.State == "running" then
				response.success = false
				response.result = "CURRENT_SCHEDULE_IS_RUNNING"
				response.message = _T("Current schedule is running.")
				response.code = 400
				break
			end
		end
		return response
	end

	local is_valid = validate()
	if not is_valid.success then
		LogError("[REPORTS:deleteSchedule] Error deleting schedule. %s", is_valid.message)
		return replyJson( web, is_valid.code, is_valid )
	end

	for _, sch_id in ipairs( request ) do
		local sch_info = { Id = sch_id }
		local success, err = scheduler.delete_schedule( utils, redis, sch_info )

		if not success then
			LogError("[REPORTS:deleteSchedule] Error %s deleting schedule (request: %s)", json.encode(err or {}), json.encode(request))
			return replyJson( web, 400, err )
		end
	end

	return replyJson( web, 200, "OK" )
end

local function activateSchedule( web, request, ks )
	local _T = makeTextTranslator()

	local user = inconcert.authUser.userName
	local vcc = inconcert.authUser.virtualContactCenter
	local schedule = request or {}

	schedule.User = user
	schedule.Vcc  = vcc

	LogInfo("Activating schedule. User '%s', Vcc %s, schedule: '%s'.", user, vcc, json.encode( schedule ))

	local validate = function( sch_id )
		local response = { success = true }
		local success, schedule = scheduler.get_schedule( utils, sch_id )
		if schedule.State == "running" then
			response.success = false
			response.result = "CURRENT_SCHEDULE_IS_RUNNING"
			response.message = _T("Current schedule is running.")
			response.code = 400
		end
		return response
	end

	local is_valid = validate( schedule.Id )
	if not is_valid.success then
		LogError("Error activating schedule. %s", is_valid.message)
		return replyJson( web, is_valid.code, is_valid )
	end

	scheduler.activate_schedule( utils, redis, schedule )

	LogInfo("Successfully schedule activated.")

	return replyJson( web, 200, {} )
end

local function deactivateSchedule( web, request, ks )
	local _T = makeTextTranslator()

	local user = inconcert.authUser.userName
	local vcc = inconcert.authUser.virtualContactCenter
	local schedule = request or {}

	schedule.User = user
	schedule.Vcc  = vcc

	LogInfo("Deactivating schedule. User '%s', Vcc %s, schedule: '%s'.", user, vcc, json.encode( schedule ))

	local validate = function( sch_id )
		local response = { success = true }
		local success, schedule = scheduler.get_schedule( utils, sch_id )
		if schedule.State == "running" then
			response.success = false
			response.result = "CURRENT_SCHEDULE_IS_RUNNING"
			response.message = _T("Current schedule is running.")
			response.code = 400
		end
		return response
	end

	local is_valid = validate( schedule.Id )
	if not is_valid.success then
		LogError("Error deactivating schedule. %s", is_valid.message)
		return replyJson( web, is_valid.code, is_valid )
	end

	scheduler.deactivate_schedule( utils, redis, schedule )

	LogInfo("Successfully schedule deactivated.")

	return replyJson( web, 200, {} )
end

--[[
-- Execute schedule
--]]
local function executeSchedule (web, request, ks)
	local _T = makeTextTranslator()
	local _T = makeTextTranslator()

	local user = inconcert.authUser.userName
	local vcc = inconcert.authUser.virtualContactCenter
	local schedule = request or {}

	schedule.User = user
	schedule.Vcc  = vcc

	LogInfo("[REPORTS:executeSchedule] Executing scheduler. User '%s', Vcc %s, schedule: '%s'.", user, vcc, json.encode(schedule))

	local validate = function( sch_id )
		local response = { success = true }
		local success, schedule = scheduler.get_schedule( utils, sch_id )
		if schedule.State == "running" then
			response.success = false
			response.result = "CURRENT_SCHEDULE_IS_RUNNING"
			response.message = _T("Current schedule is running.")
			response.code = 400
		end
		return response
	end

	local is_valid = validate(schedule.Id)
	if not is_valid.success then
		LogError("[REPORTS:executeSchedule] Error Executing schedule. %s", is_valid.message)
		return replyJson(web, is_valid.code, is_valid)
	end


	local base_config = inconcert.environment.apps.reports
	local host = base_config.config_server.myIp..":"..base_config.config_server.myPort
	local url = host.."/inconcert/apps/reports"

	local success, err = scheduler.exec(
		irLink,
		authFilter,
		model.getDbUtils(),
		redis,
		schedule.Id,
		url
	)

	if not success then
		LogError("[REPORTS:executeSchedule] Error executing scheduler. %s", schedule.Id)
		local response = {}
		response.success = false
		response.result = err
		response.message = _T("Schedule does not exist.")
		response.code = 400
		return replyJson(web, response.code, response)
	end

	LogInfo("[REPORTS:executeSchedule] Successfully executed scheduler.")

	return replyJson(web, 200, {})
end

local function exportRecordNotFilesUploadedToFtp(wsapi_env)
	local userId = inconcert.authUser.userName
	local _T = makeTextTranslator()
	local post_data = (wsapi_request.new(wsapi_env)).params["data"]
	LogInfo("[REPORTS:executeSchedule] Export files record by logged UserId %s post_data %s", userId, post_data)
	local post_data_dec = json.decode(post_data)
	if not post_data_dec.filesNames then
		wsapi_env.status = 404
		return wrapErrorResponse(404, "Error exporting, no files")
	end

	local fileNames = post_data_dec.filesNames
	

	local csv = utils.Table2Csv(fileNames, nil, nil, "|")
	if not csv then
		wsapi_env.status = 500
		return wrapErrorResponse(500, "Error exporting")
	end

	csv = string.gsub('"'..csv..'"', "|", '","' )
	local LN = "\n"
	csv = string.gsub(csv, LN, '"'..LN..'"')
	local updateDate = utils.ConvertUTCDateToDate( LuaOrm.Date():toString("%Y-%m-%d %H:%M:%S"), nil, "%Y-%m-%d %H:%M:%S", true, inconcert.authUser.timezone or nil ) 
	local suggestedFileName = string.format("(%s) - %s.CSV", updateDate, "Record of files not uploaded to FTP")
	local headers = {}
	headers["Response"]            = "defaultError"
	headers["Content-Type"]        = "application/octet-stream"
	headers["Content-Length"]      = #csv
	headers["Response"]            = "ok"
	headers["Content-Disposition"] = [[attachment; filename="]] .. suggestedFileName .. '"'
	headers["Set-Cookie"]          = "download_started=true; Path=/"

	LogInfo("[REPORTS:executeSchedule] Exporting to Csv File Name '%s'.", suggestedFileName)

	local coroutineIter = 0
	return 200, headers, coroutine.wrap(function()
		coroutineIter = coroutineIter + 1
		coroutine.yield(csv)
	end)
end

local function getDefaultValues (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = inconcert.authUser.userName.."@"..vcc

	--LogDebug("[REPORTS:getDefaultValues] Retrieving default values with user `%s' and vcc `%s'", userId, vcc)
	
	local default_values = {
		DEFAULT_FIELD_MAX_LENGTH = manager.DEFAULT_FIELD_MAX_LENGTH
	}

	return replyJson(web, 200, default_values)
end

local function getCampaignOptions (web)
	web.headers["Content-Type"] = "application/json"

	local vcc = inconcert.authUser.virtualContactCenter
	local fullUserId = inconcert.authUser.userName.."@"..vcc
	local userGroup = inconcert.authUser.userGroup
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local hasPermissionToAppAdministration = HasPermissionToAdministrationApp(ksAgent)
	local isRestricted = permissions.IsPermissionEnabledForLoggedUser("administration", "admin_restricted", nil, ksAgent)

	LogInfo("[REPORTS:getCampaignOptions] Retrieving campaign options with user '%s' and vcc '%s', roleId: '%s, with permission in App Administration = '%s'.", fullUserId, vcc, userGroup, hasPermissionToAppAdministration)
	
	local result = {}
	local campaignsWithAccounts = getVccCampaignsWithAccounts(vcc, ksAgent)
	local userCampaigns

	--[[
		Funcion encargada de validar si la campana queda visible para el logged User.
	]]--
	local function MustIncludeCampaign (campaignId)
		-- La campana interna no se incluye en los filtros ni la sandbox
		local campaign = model.GetCampaign(campaignId, ksAgent, { "Id", "IsInternal", "SystemType"})

		if campaign.IsInternal == "1" or (campaign.SystemType and campaign.SystemType == "sandbox") then
			return false
		end
		-- Is admin user and is not restricted
		if hasPermissionToAppAdministration and not isRestricted then
		-- Si es un usuario administrador, NO realiza ninguna validacion e incluye la campana.
			return true
		end
		-- Si NO es un usuario administrador, tiene que validar que el logged user pertenezca a la campana.
		userCampaigns = userCampaigns or GetUserCampaigns(vcc, fullUserId, kAgent)
		if userCampaigns[campaignId:lower()] then
			--LogDebug("[REPORTS:getCampaignOptions] SI Incluyo la Campana. El Usuario NO tiene permisos sobre la app 'Administration' y SI pertenece a la campana: '%s", campaignId)
			return true
		end
		--LogDebug("[REPORTS:getCampaignOptions] NO Incluyo la Campana. El Usuario NO tiene permisos sobre la app 'Administration' y NO pertenece a la campana: '%s", campaignId)
		return false
	end

	for campaignId, accounts in pairs(campaignsWithAccounts) do
		local mustIncludeCampaign = MustIncludeCampaign(campaignId)

		if mustIncludeCampaign then
			local campaignUid = utils.SeparateVCC(campaignId)
			local campaignAttLevels = model.GetAttentionLevels(campaignId, ksAgent)
			local campaignResult = {}
			for accountId, account in pairs(accounts) do
				local accountResult = {
					name       = account.name,
					proxyType  = account.proxyType,
					campaignId = campaignUid,
					attLevels  = {}
				}
				for attLevelId, _ in pairs(account.AttentionLevels or {}) do
					if campaignAttLevels[attLevelId] then
						accountResult.attLevels[campaignAttLevels[attLevelId].name] = {
							campaignId = campaignUid
						}
					end
				end
				local accountUid = account.uid
				if account.proxyType == "FACEBOOK" or account.proxyType == "INSTAGRAM" then
					accountResult.pages = {}
					for pageId, pageName in pairs(account.pagesNames or {}) do
						accountResult.pages[pageId] = {
							name       = pageName,
							accountId  = accountUid,
							campaignId = campaignUid
						}
					end
				end
				if account.proxyType == "WHATSAPP" then
					local _, whatsappAccountId = utils.Separate(accountId, "_")
					accountUid = whatsappAccountId or accountId
				end
				campaignResult[accountUid] = accountResult
			end
			result[campaignUid] = campaignResult
		end
	end
	ksAgent:Close()

	return replyJson(web, 200, result)
end

--------------------------------------------------
------------ URL HANDLERS
--------------------------------------------------

local function wrapList(name, listFunction, getFunction, iterator)
	iterator = iterator or _.Identity
	return (function(web)
		local _T = makeTextTranslator()
		web.headers["Content-Type"] = "application/json"
		local res, err = listAllByEntity(name, listFunction, getFunction, function(elem)
			if elem.hidden then
				return false
			end
			return iterator(elem, nil, _T)
		end)
		return res and replyJson(web, 200, res) or replyJson(web, 500, err)
	end)
end

local function wrapListFilters()
	return (function(web, reportType)
		LogInfo("[REPORTS:wrapListFilters] Loading filters type: '%s'", reportType)
		if reportType == "detail" then
			local listDetail = wrapList("filter", manager.listDetailFilters, manager.getFilter, makeFilterTransformFunction("detail"))
			return listDetail(web)
		elseif reportType == "livedetail" then
			local listLiveDetail = wrapList("filter", manager.listLiveDetailFilters, manager.getFilter, makeFilterTransformFunction("livedetail"))
			return listLiveDetail(web)
		elseif reportType == "summary" then
			local listSummary = wrapList("cut", manager.listCuts, manager.getCut, makeFilterTransformFunction("summary"))
			return listSummary(web)
		else
			return replyJson(web, 400, "Invalid Request")
		end
	end)
end

local function wrapListDimensions()
	--wrapList("dimension", manager.listAllDimensions, manager.getDimension, transformDimension)
	return (function(web, dimensionsType)
		LogInfo("[REPORTS:wrapListDimensions] Loading dimensions type: '%s'", dimensionsType)
		if dimensionsType == "archive" then
			local listArchiveDimensions = wrapList("dimension", manager.listArchiveDimensions, manager.getDimension, transformDimension)
			return listArchiveDimensions(web)
		elseif dimensionsType == "live" then
			local listLiveDimensions = wrapList("dimension", manager.listLiveDimensions, manager.getDimension, transformDimension)
			return listLiveDimensions(web)
		elseif dimensionsType == "all" then
			local listAllDimensions = wrapList("dimension", manager.listAllDimensions, manager.getDimension, transformDimension)
			return listAllDimensions(web)
		else
			return replyJson(web, 400, "Invalid Request")
		end
	end)
end

function _post(url, requestHandler)
	_M:dispatch_post(function(web)
		web.headers["Content-Type"] = "application/json"

		local web_input = web.input["post_data"]
		local post_data = json.decode(web_input)
		if not post_data then
			LogInfo("[%s] Invalid post data (%s)", url, web_input)
			return replyJson(web, 400, "Invalid Request")
		end

		--LogDebug("[%s]: Handling POST request (%s)", url, web_input)

		local ks, err = model.getCassandraConnection()
		if not ks then
			return replyJson(web, 500, "Internal Server Error")
		end

		local res = helpers.safeInvoke(requestHandler, function(res)
			return res
		end, web, post_data, ks)

		ks:Close()

		return res or replyJson(web, 500, "Internal Server Error")
	end, url)
end

function _get(url, handler)
	_M:dispatch_get(handler, url)
end

function _wsapi(url, handler)
	_M:dispatch_wsapi(handler, url)
end

local function has_permission( action, func ) return permissions:CheckPermission( "reports", action, func ) end

-- reports viewer
_get("/viewer/list_all/", listViewerReports)
_get("/list_categories/", listCategories)
_get("/get/(.+)"        , getReport)
_wsapi("/execute/(.+)"  , executeReport)
_wsapi("/export/(.+)"   , exportReport) -- A esta url/funcion las invoca la rocka "reporter"

-- options
_get ("/options/campaigns/(.*)", getCampaignOptions)

_get("/export_formats/", function(web)
	web.headers["Content-Type"] = "application/json"
	return replyJson(web, 200, exporter.get_types())
end)

_get("/schedule_job/", function(web)
	web.headers["Content-Type"] = "application/json"

	scheduler.cleanup_schedules( utils, redis )

	local base_config = inconcert.environment.apps.reports
	local host = base_config.config_server.myIp..":"..base_config.config_server.myPort
	local url = host.."/inconcert/apps/reports"

--[[
	Dentro de estas funciones de "run" del "schedule" entre tantas cosas que hace termina invocado
	 a la URL "/export/" funcion "exportReport" de este mismo archivo, para obtener el detalle del reporte.
--]]
	return replyJson(web, 200, scheduler.run(
			irLink,
			authFilter,
			model.getDbUtils(),
			redis,
			url
		)
	)
end)

---------
-- reports designer
---------
--
-- reports
_get("/designer/list_all/"      , listDesignerReports                                             )

-- custom reports
_post("/save/"                  , has_permission( "save_custom_reports"   , saveReport           ))
_post("/delete/"                , has_permission( "delete_custom_reports" , deleteReport         ))

-- custom metrics
_get("/metrics/get_all/"        , listMetrics)
_post("/metrics/save/"          , has_permission( "save_custom_metrics"   , saveMetric           ))
_post("/metrics/delete/"        , has_permission( "delete_custom_metrics" , deleteMetric         ))

-- custom filters
_get("/filters/get_all/(.+)"    , wrapListFilters())
_post("/filters/save/"          , has_permission( "save_custom_filters"   , saveFilter           ))
_post("/filters/delete/"        , has_permission( "delete_custom_filters" , deleteFilter         ))

-- dimensions
_get("/dimensions/get_all/(.+)" , wrapListDimensions())
_post("/dimensions/save/"       , has_permission( "save_custom_fields"    , saveDimension        ))
_post("/dimensions/delete/"     , has_permission( "delete_custom_fields"  , deleteDimension      ))

-- schedules
_post("/schedules/get_all/"     , listSchedules )
_get ("/schedules/get/(.+)"     , getSchedule   )
_post("/schedules/save/"        , has_permission( "save_schedule"         , saveSchedule         ))
_post("/schedules/delete/"      , has_permission( "delete_schedule"       , deleteSchedule       ))
_post("/schedules/activate/"    , has_permission( "activate_schedule"     , activateSchedule     ))
_post("/schedules/deactivate/"  , has_permission( "activate_schedule"     , deactivateSchedule   ))
_post("/schedules/execute/"     , has_permission( "execute_schedule"      , executeSchedule      ))
_wsapi("/schedules/export_not_uploaded_files_record/"   , exportRecordNotFilesUploadedToFtp)
-- default values
_get("/default_values/get_all/", getDefaultValues)

--scripts and pages

--paginas
local function render_pages(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:RenderPageLP("/apps/reports/pages/" .. name .. ".html", web, env)
end

--scripts
function render_script(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:ScriptLp("/apps/reports/scripts/" .. name .. ".js", web, env)
end

_M:dispatch_get(render_script, "/scripts/([^.]+)%.([^.]+)%.?([^.]*)%.js")
_M:dispatch_get(render_pages, "/pages/([^.]+)%.([^.]+)%.?([^.]*)%.html")

--
--  (scripts and styles) urls/apis
local function serve_file(app_module, cache)
	return function (web)
		local filename = string.sub(web.path_info, 2, #web.path_info)
		if not cache then
			app_module.response.DontCache(web)
		end
		return _M:serve_static(web, module_dir.."/"..filename)
	end
end
_M:dispatch_get(serve_file(_M), "/images/.+",".+%.png", ".+%.gif", "/styles/.+", "/css/.+")
