/*jshint bitwise:true, browser:true, curly:true, eqeqeq:true, noarg:true, noempty:true, strict:true, undef:true, unused:true, esversion: 9 */
/*jshint quotmark:false*/
/*global $, _, EventEmitter, inConcert*/

this.inConcert = this.inConcert || {};
this.inConcert.reports = this.inConcert.reports || {};

(function () {
	"use strict";

	const format = inConcert.Utils.format;
	const helpers = inConcert.reports.helpers;
	const CACHE_TIME = 300 * 1000; // 5 minutes cache
	const CAMPAIGNS_CACHE_TIME = 60 * 1000; // 1 minuto

	// Handy
	var isCacheValid = function(obj) {
		if (obj && ((new Date()).getTime() - CACHE_TIME <= obj.__timestamp)) {
			//console.log("loading with cache");
			return true;
		}
		else {
			//console.log("loading without cache");
			return false;
		}
	};

	/*!
	 * Assuming A and B are object of the same type, or at least having the attribute \p attr in
	 * common, this method creates a comparision function that will compare A and B using the
	 * attribute \p attr. So \code A < B \endcode if and only if \code a[attr] < b[attr] \endcode.
	 *
	 * \see \c compareById and \c compareByName
	 */
	var makeAttributeComparator = function(attr) {
		return (function(a, b) {
			return a[attr].localeCompare(b[attr], undefined, {
				sensitivity: "base"
			});
		});
	};

	/*!
	 * Sorted insertion using binary search
	 */
	var insertOrdered = function(list, elem, compareFunction) {
		var insertOrderedImpl = function(list, elem, startIndex, endIndex, compareFunction) {
			if (startIndex === endIndex) {
				return list.splice(startIndex, 0, elem);
			}

			var pivot = Math.floor((startIndex + endIndex) / 2);

			var compareResult = compareFunction(elem, list[pivot]);
			if (compareResult < 0) {
				insertOrderedImpl(list, elem, startIndex, pivot, compareFunction);
			} else if (compareResult > 0) {
				insertOrderedImpl(list, elem, pivot + 1, endIndex, compareFunction);
			} else {
				list.splice(pivot, 0, elem);
			}
		};

		insertOrderedImpl(list, elem, 0, list.length, compareFunction);
	};

	/*!
	 * Deletes the items with id \p id from \p array
	*/
	var deleteFromArray = function(array, id) {
		_.find(array, function(elem, index) {
			if (elem.id === id) {
				array.splice(index, 1);
				return true;
			}
		});
	};

	/*!
	 * Generates the filter's description base on its fields
	 */
	var buildFilterDescription = function(filter) {
		var rejectedFields = function(field) {
			return field.id === "VCC";
		};

		var reduce = function(params) {
			if (params.array.length === 0) {
				return params.defaultValue;
			}
			params.separator = params.separator || ",";
			return _.reduce(params.array, function(memo, field, index, list)  {
				if (index === 0) {
					return memo + " " + field.label;
				}
				return memo + (index === list.length - 1 && params.last ? params.last : params.separator) + " " + field.label;
			}, params.memo);
		};

		var buildSummaryFilterDescription = function(summaryFilter) {
			var filters = reduce({
				array       : _.reject(summaryFilter.filter, rejectedFields),
				memo        : "<%=_T('Filter by')%>",
				defaultValue: "<%=_T('No filters')%>",
				last        : " " + "<%=_T(',')%>"
			});
			var columns = reduce({
				array       : summaryFilter.columns,
				memo        : "<%=_T('Group by')%>",
				defaultValue: "<%=_T('No grouping')%>",
				separator   : " " + "<%=_T(',')%>"
			});
			return format("{0}. {1}.", columns, filters);
		};

		var buildDetailFilterDescription = function(detailFilter) {
			var optionalFields = reduce({
				array       : _.reject(detailFilter.optionalFields, rejectedFields),
				memo        : "<%=_T('optionally by')%>",
				defaultValue: "",
				last        : " " + "<%=_T(',')%>"
			});
			var mandatoryFields = reduce({
				array       : _.reject(detailFilter.mandatoryFields, rejectedFields),
				memo        : "<%=_T('by')%>",
				defaultValue: "",
				last        : optionalFields === "" ? (" " + "<%=_T(',')%>") : undefined
			});
			var sortFields = reduce({
				array       : _.reject(detailFilter.sortFields, rejectedFields),
				memo        : "<%=_T('Sorted by')%>",
				defaultValue: "<%=_T('Default sorting')%>",
				last        : " " + "<%=_T(',')%>"
			});

			if (mandatoryFields === "" && optionalFields === "") {
				return format("{0}. {1}.", "<%=_T('No filters')%>", sortFields);
			}

			return format("<%=_T('Filter {0}{1}{2}. {3}.')%>",
				mandatoryFields,
				(mandatoryFields !== "" && optionalFields !== "") ? "<%=_T(' , ')%>" : "",
				optionalFields,
				sortFields
			);
		};

		var buildLiveDetailFilterDescription = buildDetailFilterDescription;

		try {
			switch (filter.type) {
				case "summary":
					return buildSummaryFilterDescription(filter);
				case "detail" :
					return buildDetailFilterDescription(filter);
				case "livedetail" :
					return buildLiveDetailFilterDescription(filter);
				default:
					throw("buildFilterDescription - Unrecognized filter.type: " + filter.type);
			}
		} catch(e) {
			return "ERROR: Filter uses invalid fields";
		}
	};

	/*! A common comparison function */
	//var compareById = makeAttributeComparator("id");

	/*! Another common comparison function */
	var compareByName = makeAttributeComparator("name");
	var compareByLabel = makeAttributeComparator("label");

	/*! Base class for reports, it only defines the interface */
	function ReportBase(id, type, controller)
	{
		this.id = id;
		this.type = type;
		this.name = "";
		this.description = "";
		this.controller = controller;
		this.exec_data = {};
		this.bak_exec_data = {};
	}

	ReportBase.prototype.get_id = function() {
		return this.id;
	};
	ReportBase.prototype.get_type = function() {
		return this.type;
	};
	ReportBase.prototype.get_name = function() {
		return this.label || this.name;
	};
	ReportBase.prototype.get_description = function() {
		return this.description;
	};
	ReportBase.prototype.backup_state = function() {
		this.bak_exec_data = $.extend({}, this.exec_data);
	};
	ReportBase.prototype.restore_state = function() {
		this.exec_data = $.extend({}, this.bak_exec_data);
	};
	/*! Should return a copy of the filters for this report */
	ReportBase.prototype.getFilters = function() {
		throw "ReportBase.getFilters is abstract";
	};
	/*! Sets a filter on this report. \note prepare() must be called first */
	ReportBase.prototype.setFilterValue = function(/*k, v*/) {
		throw "ReportBase.setFilterValue is abstract";
	};
	/*! Prepares this report for execution */
	ReportBase.prototype.prepare = function() {
		throw "ReportBase.prepare is abstract";
	};
	/*! Get filters */
	ReportBase.prototype.get_filters = function() {
		return {};
	};
	/*! Checks if is single day filter */
	ReportBase.prototype.is_single_day_filter = function() {
		return false;
	};
	/*! Get filter date */
	ReportBase.prototype.get_single_day_date_filter = function() {
		return null;
	};
	/*! Checks if is single year filter */
	ReportBase.prototype.is_single_year_filter = function() {
		return false;
	};
	/*! Get filter date */
	ReportBase.prototype.get_single_year_date_filter = function() {
		return null;
	};
	/*! Get slice date */
	ReportBase.prototype.get_slice_date_column = function() {
		return null;
	};
	/*! Get proxy type */
	ReportBase.prototype.get_proxy_type = function() {
		return null;
	};
	/*! Get filters key */
	ReportBase.prototype.get_filters_key = function() {
		return "filter";
	};
	
	ReportBase.prototype.__makeStateFilterId = function(stateId) {
		return stateId;
	};

	ReportBase.prototype.__updateStateFilter = function(filter, callback) {
		var self = this;
		filter.type = "enum";

		var states = [];
		let states_added = {};
		$.each(filter.domain, function(_, state) {
			if (!state.custom) {
				const value = self.__makeStateFilterId(state.value);
				states_added[value] = true;
				states.push({
					label : state.label,
					value : value
				});
			}
		});

		var availableStates = inConcert.app.getAvailableStates();
		$.each(availableStates, function(_, state) {
			let value = self.__makeStateFilterId("State: " + state.Description);
			if (state.vcc !== "system" && !states_added[value]) {
				states.push({
					label	: state.Description,
					value	: value
				});
			}
		});
		filter.domain = states;
		
		callback();
	};

	ReportBase.prototype.__updateCampaignFilter = function(filter, callback) {
		this.controller.getCampaignOptions(this.get_proxy_type(), function(options) {
			filter.options = options;
			callback();
		});
	};

	ReportBase.prototype.__updateAccountFilter = function(filter, callback) {
		this.controller.getAccountOptions(this.get_proxy_type(), null, function(options) {
			filter.options = options;
			callback();
		});
	};

	ReportBase.prototype.__updateAttLevelFilter = function(filter, callback) {
		this.controller.getAttLevelOptions(this.get_proxy_type(), null, null, function(options) {
			filter.options = options;
			callback();
		});
	};

	ReportBase.prototype.__updatePageFilter = function(filter, callback) {
		this.controller.getPageOptions(this.get_proxy_type(), null, null, function(options) {
			filter.options = options;
			callback();
		});
	};
	
	ReportBase.prototype.__updateIntegrationFilter = function(filter, callback) {
		var scope = filter.name.replace(".__SID", "");
		if (this[scope + "Project"]) {
			filter.preset = true;
			filter.value = this[scope + "Project"];
		}
		else {
			filter.preset = false;
			filter.value = null;
		}
		callback();
	};

	/*! Updates report filters */
	ReportBase.prototype.update_filters = function(callback) {
		var self = this;
		var filters_key = self.get_filters_key();
		
		function updateFilter(filter_id, filter, callback) {
			switch(filter_id) {
				case "State":
					return self.__updateStateFilter(filter, callback);
				case "Campaign":
				case "DialerCampaign":
					return self.__updateCampaignFilter(filter, callback);
				case "Account":
					return self.__updateAccountFilter(filter, callback);
				case "AttentionLevel":
					return self.__updateAttLevelFilter(filter, callback);
				case "Page":
					return self.__updatePageFilter(filter, callback);
			}
			if (filter_id.indexOf(".__SID") > 0) {
				return self.__updateIntegrationFilter(filter, callback);
			}
			callback();
		}

		var fieldsToProcess = [];
		$.each(self.definition[filters_key], function(filter_id) {
			fieldsToProcess.push(filter_id);
		});

		(function processFields(index) {
			if (index === fieldsToProcess.length) {
				return callback();
			}
			var filter_id = fieldsToProcess[index];
			updateFilter(filter_id, self.definition[filters_key][filter_id], function() {
				processFields(index + 1);
			});
		})(0);
	};

	ReportBase.prototype.processFilterValue = function(name, value) {
		switch(name) {
			case "Account":
				var proxyType = this.get_proxy_type();
				if (proxyType !== "TWITTER") {
					//Si es Twitter debo enviar el Name
					value = this.controller.getAccountIdByName(value) || value;
				}
				break;
			case "Page":
				value = this.controller.getPageIdByName(value) || value;
				break;
		}
		return value;
	};

	ReportBase.prototype.processFilterLabel = function(name, value) {
		switch(name) {
			case "Account":
				value = this.controller.getAccountName(value) || value;
				break;
			case "Page":
				value = this.controller.getPageName(value) || value;
				break;
		}
		return value;
	};

	/*! Has line charts */
	ReportBase.prototype.has_line_charts = function() {
		return !!this.lineCharts;
	};
	/*! Has total charts */
	ReportBase.prototype.has_total_charts = function() {
		return !!this.totalCharts;
	};
	/*! Has charts */
	ReportBase.prototype.has_charts = function() {
		return this.has_line_charts() || this.has_total_charts();
	};
	/*! Can show charts */
	ReportBase.prototype.can_show_charts = function() {
		return this.has_charts() && this.has_all_filters_set();
	};

	ReportBase.prototype.has_all_filters_set = function() {
		var self = this;
		var filters_key = self.get_filters_key();
		var has_all = true;
		$.each(self.definition[filters_key], function(filterId) {
			var filterValue = self.exec_data[filters_key][filterId];
			if (!filterValue && filterValue !== false) {
				has_all = false;
				return false;
			}
		});
		return has_all;
	};

	/*! Make base lines charts data */
	ReportBase.prototype.make_line_charts_data = function(columnsMetadata) {
		if (!this.has_line_charts()) {
			return null;
		}
		var columnsMetadata_d = _.reduce(columnsMetadata, function(memo, metadata) {
			memo[metadata.name] = metadata;
			return memo;
		}, {});
		var data = {
			chartsRows : {},
			chartsIndicators : {},
			chartsByIndicator : {},
			rowsCount : 0,
			xIndicator : this.get_line_charts_x_indicator()
		};
		$.each(this.lineCharts, function(_, chart) {
			var chartRows = [];
			data.chartsRows[chart.name] = chartRows;
			var chartIndicators = [];
			data.chartsIndicators[chart.name] = chartIndicators;
			if (data.xIndicator) {
				data.chartsByIndicator[data.xIndicator] = data.chartsByIndicator[data.xIndicator] || {};
				data.chartsByIndicator[data.xIndicator][chart.name] = chartRows;
			}
			$.each(chart.indicators, function(_, indicator) {
				var metadata = columnsMetadata_d[indicator.id] || {};
				data.chartsByIndicator[indicator.id] = data.chartsByIndicator[indicator.id] || {};
				data.chartsByIndicator[indicator.id][chart.name] = chartRows;
				chartIndicators.push({
					id : indicator.id,
					name : indicator.name || metadata.label || "",
					color : indicator.color || null
				});
			});
		});
		return data;
	};

	ReportBase.prototype.add_line_charts_row = function(data) {
		if (!data) {
			return;
		}
		var x = {
			value : data.rowsCount,
			label : data.rowsCount
		};
		$.each(data.chartsRows, function(_, rows) {
			rows.push({ x : x });
		});	
		data.rowsCount++;
	};
	
	ReportBase.prototype.add_line_charts_value = function(data, name, type, value, text) {
		if (!data || !data.chartsByIndicator[name]) {
			return;
		}
		var dot = this.make_line_charts_dot(value, text);
		var index = name;
		if (name === data.xIndicator) {
			dot.value = data.rowsCount - 1;
			index = "x";
		}
		else {
			dot.value = this.get_charts_value(type, value);
		}
		$.each(data.chartsByIndicator[name], function(_, chartRows) {
			chartRows[data.rowsCount - 1][index] = dot;
		});
	};
	
	ReportBase.prototype.make_line_charts_dot = function(value, text) {
		return {
			value : value,
			label : text
		};
	};
	
	ReportBase.prototype.get_line_charts_x_indicator = function() {
		return null;
	};
	
	/*! Make base total charts data */
	ReportBase.prototype.make_total_charts_data = function(columnsMetadata, totals) {
		var self = this;
		if (!self.has_total_charts()) {
			return null;
		}
		var columnsMetadata_d = _.reduce(columnsMetadata, function(memo, metadata) {
			memo[metadata.name] = metadata;
			return memo;
		}, {});
		var totals_d = _.reduce(totals, function(memo, value, index) {
			var metadata = columnsMetadata[index] || {};
			memo[metadata.name] = self.get_charts_value(metadata.type, value);
			return memo;
		}, {});
		var totals_label_d = _.reduce(totals, function(memo, value, index) {
			var metadata = columnsMetadata[index] || {};
			memo[metadata.name] = self.get_charts_label(metadata.type, value);
			return memo;
		}, {});
		var data = {};
		$.each(this.totalCharts, function(_, chart) {
			var indicators = [];
			$.each(chart.indicators, function(_, indicator) {
				var metadata = columnsMetadata_d[indicator.id] || {};
				indicators.push({
					id : indicator.id,
					name : indicator.name || metadata.label || "",
					color : indicator.color || null,
					total : indicator.total || false,
					value : totals_d[indicator.id],
					valueLabel : totals_label_d[indicator.id]
				});
			});
			data[chart.name] = indicators;
		});
		return data;
	};
	
	ReportBase.prototype.get_charts_value = function(type, value) {
		if (typeof(value) === "number") {
			return value;
		}
		switch(type) {
			case "Time":
				var parts = value.split(":");
				if (parts.length !== 3) {
					return 0;
				}
				return parseInt(parts[0]) * 360 + parseInt(parts[1]) * 60 + parseInt(parts[2]);
			case "percentage":
			case "Percentage":
				return parseFloat(value.replace("%", "")) || 0;
			default:
				return 0;
		}
	};

	ReportBase.prototype.get_charts_label = function(type, value) {
		if ( typeof(value) === "string" ) {
			return inConcert.Utils.escapeAttrib(value);
		}
		if ( typeof(value) === "number" && Math.floor(value) !== value ) {
			value = value.toFixed(2);
		}
		return value;
	};

	/*!
	 * A specialization for \c ReportBase targeting reports of type 'detail'
	 */
	function DetailReport(id, controller, raw_metadata, report_detail_type)
	{
		this.report_detail_type = report_detail_type || "detail";
		DetailReport.uber.constructor.call(this, id, this.report_detail_type, controller);
		$.extend(this, {
			filterOverrides: {}
		}, raw_metadata);
		this.raw_data = raw_metadata;
		this.initialize(raw_metadata);
	}

	inConcert.Utils.inherit(DetailReport, ReportBase);

	/*!
	 * Initializes this instance using the given meta-data
	 * \scope private
	 */
	DetailReport.prototype.initialize = function(raw_metadata)
	{
		var self = this;

		// definition is immutable
		this.definition = {
			queryFields: {
				startDate: {
					name       : "Start Date",
					label      : "<%=_T('Start Date')%>",
					type       : "date",
					time       : "00:00:00",
					before_than: "endDate",
					required   : true
				},
				endDate: {
					name      : "End Date",
					label     : "<%=_T('End Date')%>",
					type      : "date",
					time      : "23:59:59",
					after_than: "startDate",
					required  : true
				}
			},
			pending_fields: [],
			filter: {
				name: raw_metadata.filterName,
				scope: ""
			}
		};

		this.projectFields = _.map(this.projectFields, function(field)
		{
			field = typeof(field) === "string" && {
				dimension: field
			} || field;

			var dimension = self.controller.loadDimension(field.dimension, {sync: true}) ||
				{
					id         : field.dimension,
					name       : field.name || field.dimension,
					displayName: field.name || field.dimension,
					label      : field.label || field.name || field.dimension,
					type       : "string",
					format     : "",
					multiplier : 1
				};

			dimension.alias = field.alias || dimension.label || field.displayName;

			return dimension;
		});

		var filter = this.controller.getFilter(this.definition.filter.name, this.report_detail_type);
		if (filter) {
			var append_field_from_array = function(field_list, base_props) {
				field_list.forEach(function(field) {
					var fieldName = field.name;

					if (!self.controller.canConfigureFilter(fieldName)) {
						self.definition.pending_fields.push(fieldName);
						return;
					}

					var override = self.filterOverrides[fieldName];
					if ( override ) {
						_.each(override.fields, function(ovField) {
							self.definition.queryFields[ovField.name] = _.extend({
								name       : ovField.name,
								label      : ovField.label || ovField.name,
								description: ovField.description,
								type       : ovField.type || "string",
								domain     : ovField.domain
							}, base_props);
						});
						return;
					}

					self.definition.queryFields[fieldName] = $.extend({
						name       : fieldName,
						label      : field.label || field.name || fieldName,
						description: field.description,
						type       : !field.customizable && field.type || "string",
						domain     : field.domain,
						can_sort   : (filter.sortFields.indexOf(fieldName) !== -1)
					}, base_props);
				});
			};

			this.definition.filter.scope = filter.scope;
			append_field_from_array(filter.mandatoryFields, { required: true  });
			append_field_from_array(filter.optionalFields,  { required: false });

			var filter_values = raw_metadata.filters || {};
			$.each(filter_values, function(field_name, field_value) {
				var target_field = self.definition.queryFields[field_name];
				if ( target_field ) {
					target_field.preset = true;
					target_field.value  = field_value;
				} else {
					inConcert.logger.error("Field [" + field_name + "] not found in definition");
				}
			});
		}

		Object.freeze(this.definition);
	};

	DetailReport.prototype.getFilters = function() {
		return $.extend({}, this.definition.queryFields);
	};

	DetailReport.prototype.setFilterValue = function(field_name, field_value) {
		this.exec_data.queryFields[field_name] = this.processFilterValue(field_name, field_value);
	};

	DetailReport.prototype.prepare = function() {
		this.exec_data = {
			firstKey       : "",
			queryFields    : {},
			last_handle    : null,
			pending_fields : this.definition.pending_fields
		};
	};

	DetailReport.prototype.get_filters = function() {
		return this.exec_data && this.exec_data.queryFields || {};
	};

	DetailReport.prototype.get_proxy_type = function() {
		return this.filters && this.filters.Type || null;
	};

	DetailReport.prototype.get_filters_key = function() {
		return "queryFields";
	};

	/******************************************************************************
	*
	*******************************************************************************/
	function LiveDetailReport(id, controller, raw_metadata)
	{
		const report_detail_type = "livedetail";
		LiveDetailReport.uber.constructor.call(this, id, controller, raw_metadata, report_detail_type);
	}

	inConcert.Utils.inherit(LiveDetailReport, DetailReport);

	/*!
	 * A specialization for \c ReportBase targeting reports of type 'summary'
	 */
	function SummaryReport(id, controller, raw_metadata)
	{
		SummaryReport.uber.constructor.call(this, id, "summary", controller);
		$.extend(this, raw_metadata);
		this.raw_data = raw_metadata;
		this.initialize(raw_metadata);
	}

	inConcert.Utils.inherit(SummaryReport, ReportBase);

	/*!
	 * Initializes this instance using the given meta-data
	 * \scope private
	 */
	SummaryReport.prototype.initialize = function(raw_metadata) {
		var self = this;

		// definition is immutable
		self.definition = {
			filter : {
				DateFrom: {
					name       : "Start Date",
					label      : "<%=_T('Start Date')%>",
					type       : "date",
					time       : "00:00:00",
					before_than: "DateTo",
					required   : true
				},
				DateTo: {
					name      : "End Date",
					label     : "<%=_T('End Date')%>",
					type      : "date",
					time      : "23:59:59.999",
					after_than: "DateFrom",
					required  : true
				},
				sliceSize: {
					name    : "Slice Size",
					label   : "<%=_T('Slice Size')%>",
					type    : "enum",
					required: true,
					domain: [
						{ value: "15Min", label: "<%=_T('15 minutes')%>", selected: true},
						{ value: "30Min", label: "<%=_T('30 minutes')%>" },
						{ value: "Hour",  label: "<%=_T('Hour')%>" },
						{ value: "Day",   label: "<%=_T('Day')%>" },
						{ value: "Month", label: "<%=_T('Month')%>" }
					]
				}
			},
			pending_fields: [],
			cut: raw_metadata.cut
		};

		self.indicators = _.compact(_.map(self.indicators, function(indicator) {
			var metric = self.controller.loadMetric(indicator.match(/[^|]+/), {sync:true});
			if (metric) { // there may be some unsupported metrics around there
				var metricParams = _.pick(metric, 'id', 'name', 'description', 'scope', 'sub_scope');
				var columnDetails = raw_metadata.columnDetails[indicator];
				if (columnDetails) { // some indicators may not be reflected in the columns
					metricParams.alias = columnDetails.columnName || metricParams.name;
					if (metric.extras && metric.extras.MinMaxIndicator) {
						metricParams.isMinMax = true;
						metricParams.min = columnDetails.min;
						metricParams.name = (metricParams.min && "(Min) " || "(Max) ") + (metric.name || metric.id);
					}
					metricParams.name_with_scope = helpers.addScopeToMetricName(metricParams);
				}
				return metricParams;
			}
		}));

		var cut = self.controller.getCut(self.definition.cut);
		if (cut && cut.filter) {
			cut.filter.forEach(function(field) {
				var fieldName = field.name;
				if (self.controller.canConfigureFilter(fieldName)) {
					self.definition.filter[fieldName] = ({
						name    : fieldName,
						type    : field.type,
						domain	: field.domain,
						required: false,
						label	: field.label || field.name || fieldName
					});
				} else {
					self.definition.pending_fields.push(fieldName);
				}
			});
		}

		Object.freeze(self.definition);
	};

	SummaryReport.prototype.getFilters = function() {
		return $.extend({}, this.definition.filter);
	};

	SummaryReport.prototype.setFilterValue = function(filter_id, filter_value) {
		this.exec_data.filter[filter_id] = this.processFilterValue(filter_id, filter_value);
	};

	SummaryReport.prototype.prepare = function() {
		this.exec_data = {
			filter         : {},
			last_handle    : null,
			pending_fields : this.definition.pending_fields
		};
	};
	
	SummaryReport.prototype.get_filters = function() {
		return this.exec_data && this.exec_data.filter || {};
	};
	
	SummaryReport.prototype.get_proxy_type = function() {
		return this.interactionType;
	};
	
	SummaryReport.prototype.is_single_day_filter = function() {
		var filters = this.get_filters();
		if (!filters.DateFrom || !filters.DateTo) {
			return false;
		}
		return filters.DateFrom.split(" ")[0] === filters.DateTo.split(" ")[0];
	};
	
	SummaryReport.prototype.get_single_day_date_filter = function() {
		var filters = this.get_filters();
		return filters.DateFrom.split(" ")[0] + " ";
	};
	
	var singleYearSlices = {
		"15Min"		: true,
		"30Min"		: true,
		"Hour"		: true
 	};
	SummaryReport.prototype.is_single_year_filter = function() {
		var filters = this.get_filters();
		if (!singleYearSlices[filters.sliceSize] || !filters.DateFrom || !filters.DateTo) {
			return false;
		}
		return filters.DateFrom.substring(0, 5) === filters.DateTo.substring(0, 5);
	};
	
	SummaryReport.prototype.get_single_year_date_filter = function() {
		var filters = this.get_filters();
		return filters.DateFrom.substring(0, 5);
	};
	
	SummaryReport.prototype.get_slice_date_column = function() {
		return "Slice";
	};

	SummaryReport.prototype.get_line_charts_x_indicator = function() {
		return "Slice";
	};
	
	SummaryReport.prototype.__makeStateFilterId = function(stateId) {
		return stateId.replace("State: ", "");
	};

	function RealtimeReport(id, controller, raw_metadata)
	{
		RealtimeReport.uber.constructor.call(this, id, "realtime", controller);
		$.extend(this, raw_metadata);
		this.raw_data = raw_metadata;
		this.initialize(raw_metadata);
	}

	inConcert.Utils.inherit(RealtimeReport, ReportBase);

	RealtimeReport.prototype.initialize = function(/*raw_metadata*/)
	{
		this.definition = {
			filter : {
				startDate: {
					name       : "Start Date",
					label      : "<%=_T('Start Date')%>",
					type       : "date",
					time       : "00:00:00",
					before_than: "DateTo",
					required   : true
				},
				endDate: {
					name      : "End Date",
					label     : "<%=_T('End Date')%>",
					type      : "date",
					time      : "23:59:59.999",
					after_than: "DateFrom",
					required  : true
				},
				Campaign: {
					name      : "Campaign",
					label     : "<%=_T('Campaign')%>",
					type      : "string",
					required  : false
				}
			}
		};
	};

	RealtimeReport.prototype.getFilters = function() {
		return $.extend({}, this.definition.filter);
	};

	RealtimeReport.prototype.setFilterValue = function(filter_id, filter_value) {
		this.exec_data.filter[filter_id] = this.processFilterValue(filter_id, filter_value);
	};

	RealtimeReport.prototype.prepare = function() {
		this.exec_data = {
			filter     : {},
			last_handle: null
		};
	};
	
	RealtimeReport.prototype.get_filters = function() {
		return this.exec_data && this.exec_data.filter || {};
	};
	
	RealtimeReport.prototype.get_proxy_type = function() {
		return this.interactionType;
	};

	function DialerBatchReport(id, controller, raw_metadata)
	{
		DialerBatchReport.uber.constructor.call(this, id, "dialerbatch", controller);
		$.extend(this, raw_metadata);
		this.raw_data = raw_metadata;
		this.initialize(raw_metadata);
	}

	inConcert.Utils.inherit(DialerBatchReport, ReportBase);

	DialerBatchReport.prototype.initialize = function(/*raw_metadata*/)
	{
		this.definition = {
			filter : {
				Campaign: {
					name      : "Campaign",
					label     : "<%=_T('Campaign')%>",
					type      : "string",
					required  : true
				},
				Batch: {
					name      : "Batch",
					label     : "<%=_T('Batch')%>",
					type      : "string",
					required  : false
				}
			}
		};
	};

	DialerBatchReport.prototype.getFilters = function() {
		return $.extend({}, this.definition.filter);
	};

	DialerBatchReport.prototype.setFilterValue = function(filter_id, filter_value) {
		this.exec_data.filter[filter_id] = this.processFilterValue(filter_id, filter_value);
	};

	DialerBatchReport.prototype.prepare = function() {
		this.exec_data = {
			filter     : {},
			last_handle: null
		};
	};
	
	DialerBatchReport.prototype.get_filters = function() {
		return this.exec_data && this.exec_data.filter || {};
	};
	
	DialerBatchReport.prototype.get_proxy_type = function() {
		return this.filters && this.filters.Type || null;
	};

	function DialerContactsReport(id, controller, raw_metadata)
	{
		DialerContactsReport.uber.constructor.call(this, id, "dialercontacts", controller);
		$.extend(this, raw_metadata);
		this.raw_data = raw_metadata;
		this.initialize(raw_metadata);
	}

	inConcert.Utils.inherit(DialerContactsReport, ReportBase);

	DialerContactsReport.prototype.initialize = function(/*raw_metadata*/)
	{
		this.definition = {
			filter : {
				Campaign: {
					name      : "Campaign",
					label     : "<%=_T('Campaign')%>",
					type      : "string",
					required  : true
				},
				Batch: {
					name      : "Batch",
					label     : "<%=_T('Batch')%>",
					type      : "string",
					required  : inConcert.app.has_allegro_integration()
				},
				ContactId: {
					name      : "ContactId",
					label     : "<%=_T('Contact Id')%>",
					type      : "string",
					required  : false
				}
			}
		};
	};

	DialerContactsReport.prototype.getFilters = function() {
		return $.extend({}, this.definition.filter);
	};

	DialerContactsReport.prototype.setFilterValue = function(filter_id, filter_value) {
		this.exec_data.filter[filter_id] = this.processFilterValue(filter_id, filter_value);
	};

	DialerContactsReport.prototype.prepare = function() {
		this.exec_data = {
			filter     : {},
			last_handle: null
		};
	};
	
	DialerContactsReport.prototype.get_filters = function() {
		return this.exec_data && this.exec_data.filter || {};
	};
	
	DialerContactsReport.prototype.get_proxy_type = function() {
		return this.interactionType;
	};

	/*!
	 * Abstraction layer for reports management
	 *
	 * Events:
	 * \li reportAdded  : function(report)
	 * \li reportUpdated: function(report)
	 * \li reportDeleted: function(report)
	 * \li metricAdded  : function(metric)
	 * \li metricUpdated: function(metric)
	 * \li metricDeleted: function(metric)
	 */
	function ReportsController(base_domain)
	{
		this.m_baseDomain = base_domain;
		this.m_currentRequest = null;
		this.m_currentRequestParams = null;

		// holds the list of reports classified by type
		this.m_reportsByType  = {};
		// holds cache metadata for reports classified by type and report id
		this.m_loadedReports = { detail: {}, livedetail: {}, summary: {}, realtime: {}, dialerbatch: {}, dialercontacts: {} };
		// cache default values of server side
		this.m_defaultValues = undefined;

		const self = this;
		this.report_factories = {
			detail: {
				create: function(report_id, raw_metadata) {
					return new DetailReport(report_id, self, raw_metadata);
				}
			},
			livedetail: {
				create: function(report_id, raw_metadata) {
					return new LiveDetailReport(report_id, self, raw_metadata);
				}
			},
			summary: {
				create: function(report_id, raw_metadata){
					return new SummaryReport(report_id, self, raw_metadata);
				}
			},
			realtime: {
				create: function(report_id, raw_metadata){
					return new RealtimeReport(report_id, self, raw_metadata);
				}
			},
			dialerbatch: {
				create: function(report_id, raw_metadata){
					return new DialerBatchReport(report_id, self, raw_metadata);
				}
			},
			dialercontacts: {
				create: function(report_id, raw_metadata){
					return new DialerContactsReport(report_id, self, raw_metadata);
				}
			}
		};
	}

	inConcert.Utils.inherit(ReportsController, EventEmitter);

	/*! Given a method and its arguments, returns the full URL */
	ReportsController.prototype.buildUrl = function(method, args) {
		return this.m_baseDomain + "/inconcert/apps/reports/" + method + "/" + (args || "");
	};

	/*! \return The filter with name \p filterName */
	ReportsController.prototype.getFilter = function(filterName, filter_type) {
		if (filter_type === "detail") {
			return this.detailFilters.__by_id[filterName];
		}
		else if (filter_type === "livedetail") {
			return this.liveDetailFilters.__by_id[filterName];
		}
		else {
			throw("ReportsController.prototype.getFilter - Unrecognized filter_type: " + filter_type + " for filter name: " + filterName);
		}
	};

	/*! \return The cut with name \p cutName */
	ReportsController.prototype.getCut = function(cutName) {
		return this.summaryFilters.__by_id[cutName];
	};

	/*! \return Export types supported by the server */
	ReportsController.prototype.loadExportFormats = function() {
		var self = this;
		var deferred = new $.Deferred();

		inConcert.invoke({
			url        : self.buildUrl("export_formats"),
			type       : "GET",
			dataType   : "json",
			success: function(response) {
				deferred.resolve(response);
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

  /**
	 * \return Default values of the server side.
	 *
	 * NOTE: This method look first for cached values.
	 */
	ReportsController.prototype.loadDefaultValues = function() {
		var self = this;
		var deferred = new $.Deferred();

		if (self.m_defaultValues !== undefined) {
			_.defer(_.bind(deferred.resolve, deferred, self.m_defaultValues));
				return deferred;
		}

		inConcert.invoke({
			url        : self.buildUrl("default_values/get_all"),
			type       : "GET",
			dataType   : "json",
			success: function(response) {
				deferred.resolve(response);
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	/*! \return \c true if the filter with name \p filter_name can be configured client-side */
	ReportsController.prototype.canConfigureFilter = function(filter_name) {
		return (filter_name !== "VCC");
	};

	/*!
	 * Wraps a call to inConcert.invoke in order to limit API calls
	 */
	ReportsController.prototype.wrapInvoke = function(params) {
		const self = this;

		if (this.m_currentRequest) {
			console.log("ReportsController.prototype.wrapInvoke - Aborting current request: '" + this.m_currentRequest + "' URL: '" + params.url + "'");
			this.m_currentRequest.abort();
			this.m_currentRequest = null;
			this.m_currentRequestParams = null;
		}

		let params_on_success = params.success;
		params.success = function(response) {
			self.m_currentRequest = null;
			if ( params_on_success ) {
				params_on_success.call(this, response);
			}
			self.m_currentRequestParams = null;
		};

		let params_on_error = params.error;
		params.error = function(xhr, error) {
			self.m_currentRequest = null;
			if ( params_on_error ) {
				params_on_error.call(this, xhr, error);
			}
			self.m_currentRequestParams = null;
		};

		this.m_currentRequestParams = params;
		this.m_currentRequest = inConcert.invoke(params);
	};

	/*!
	 * Lists all entities of an specific type
	 * \scope private
	 */
	ReportsController.prototype.__commonListing = function(params)
	{
		const self = this;

		let deferred = new $.Deferred();

		params = $.extend({
			accessor  : undefined,
			url       : undefined,
			overlap   : false,
			type      : "all",
			iterator  : _.identity,
			comparator: compareByName,
			beforeGet : function(callback) {
				callback();
			}
		}, params);

		if ( isCacheValid(self[params.accessor]) ) {
			_.defer(function() {
				deferred.resolve(self[params.accessor][params.type] || []);
			});
			return deferred;
		}

		params.beforeGet(function(errorThrown) {
			if (errorThrown) {
				deferred.reject(errorThrown);
				return;
			}

			let invokeFunction, invokeName;
			if (params.overlap) {
				invokeFunction = inConcert.invoke;
				invokeName = "inConcert.invoke";
			}
			else {
				invokeFunction = self.wrapInvoke;
				invokeName = "ReportsController.wrapInvoke";
			}
			let url = self.buildUrl(params.url, params.args);

			// const uid = inConcert.Utils.generateGuid();
			// let startDate = new Date();
			// console.log("A) __commonListing - START request --> UID: '" + uid + "' - Start date: '" + startDate + "' - URL: '" + url + "' - INVOKE TYPE: '" + invokeName + "'");

			invokeFunction.call(self, {
				url     : url,
				type    : "GET",
				dataType: "json",
				success : function(response)
				{
					// let endDate = new Date();
					// console.log("B) __commonListing (SUCCES) - END request --> UID: '" + uid + "' - End date " + endDate + "' - URL: '" + url + "'");

					let cache = {
						__by_id    : response,
						__timestamp: (new Date()).getTime(),
						all        : [],
						builtin    : [],
						custom     : []
					};

					// startDate = endDate;
					// console.log("C) __commonListing - START itering items --> UID: '" + uid + "' - Start date: '" + startDate + "'");

					//let itemCount = 0;

					/*
						TODO: Este codigo queda comentado, aca se puede realizar el recorrido del loop (*) de forma asincronica,
								por si llegan a venir muchos datos y el loop demore.
						Habia unos problemas de sincronizacion con las dimensiones, request que dependen request, por esto lo dejo comentado
						y si es necesario habra que miralo.
						Ref #7101
					*/
					// inConcert.Utils.foreach_timeout(
					// {
						// rows : response,
						// max_append_per_loop: 200,
						// pause_between_loop_ms: 50,
						// waiting_message: "Please wait, adding rows for --> UID: '" + uid + "' URL '" + url + "', {0} of {1}.",
						// // por cada registro que procesamos se invoca esta funcion
						// append_fn: function( value, added, key )
						// {
							// itemCount++;
							// value = params.iterator(value, key, cache);
							// insertOrdered(cache.all, value, params.comparator);
							// if (value.vcc === "system") {
								// insertOrdered(cache.builtin, value, params.comparator);
							// } else {
								// insertOrdered(cache.custom, value, params.comparator);
							// }
						// },
						// // la funcion, para que muestre en pantalla un waiting durante la carga o un log.
						// waiting_fn : function( message ) {
							// console.log(message);
						// },
						// // Esta funcion es invocada cuando finaliza el recorrido de la lista y la carga del DOM.
						// complete_fn: function()
						// {
							// //cache.__timestamp = (new Date()).getTime();
							// self[params.accessor] = cache;
							// deferred.resolve(cache[params.type] || []);

							// endDate = new Date();
							// const elapsedTimeSec = (endDate.getTime() - startDate.getTime()) / 1000;
							// console.log("D) __commonListing - END itering items --> UID: '" + uid + "' - End date " + endDate + "' - Elapsed time (sec): '" + elapsedTimeSec + "' --> TOTAL ITEMS: " + itemCount);
						// }
					// });

					// (*) loop
					_.each(response, function(value, key) {
						//itemCount++;
						value = params.iterator(value, key, cache);
						insertOrdered(cache.all, value, params.comparator);
						if (value.vcc === "system") {
							insertOrdered(cache.builtin, value, params.comparator);
						} else {
							insertOrdered(cache.custom, value, params.comparator);
						}
					});

					self[params.accessor] = cache;
					deferred.resolve(cache[params.type] || []);

					// endDate = new Date();
					// const elapsedTimeSec = (endDate.getTime() - startDate.getTime()) / 1000;
					// console.log("D) __commonListing - END itering items --> UID: '" + uid + "' - End date " + endDate + "' - Elapsed time (sec): '" + elapsedTimeSec + "' --> TOTAL ITEMS: " + itemCount);
				},
				error: function(xhr, error) {
					// const endDate = new Date();
					// console.log("B) __commonListing (ERROR) - END request --> UID: '" + uid + "' - End date " + endDate + " - URL: '" + url + "'");
					deferred.reject(error.result);
				}
			});
		});

		return deferred;
	};

	/*!
	 * Loads and returns an entity, \p id must be unique
	 * \scope private
	 */
	ReportsController.prototype.__commonLoad = function(id, params) {
		var self = this;

		params = $.extend({
			accessor    : undefined,
			listFunction: undefined,
			sync        : false
		}, params);

		if (params && params.sync) {
			if (!self[params.accessor]) {
				throw new Error(params.accessor + " not loaded");
			}
			return self[params.accessor].__by_id[id];
		}

		var deferred = new $.Deferred();

		params.listFunction.call(self).done(function() {
			var obj = self[params.accessor].__by_id[id];
			if (!obj) {
				deferred.reject("Not Found");
			} else {
				deferred.resolve(obj);
			}
		}).fail(function(errorThrown) {
			deferred.reject(errorThrown);
		});

		return deferred;
	};

	/*!
	 * Helper
	 * \scope private
	 */
	ReportsController.prototype.__listFilters = function(filterType, options)
	{
		var self = this;

		var customIterator = options.iterator || _.identity;
		delete options.iterator;
		var iterator = _.compose(function(f) {
				f.type = filterType;
				f.description = buildFilterDescription(f);
				return f;
			},
			customIterator
		);

		return this.__commonListing($.extend({
			url     : "filters/get_all",
			args    : filterType,
			iterator: iterator,
			beforeGet: function(callback) {
				self.listDimensions().done(function() {
					callback();
				}).fail(callback);
			}
		}, options));
	};

	/*!
	 * Helper
	 * \scope private
	 */
	ReportsController.prototype.__makeFilterTransform = function(type) {
		const self = this;
		function loadDimension(idc) {
			return self.loadDimension(idc.id, {sync: true});
		}
		switch(type.toLowerCase()) {
			case "detail":
			case "livedetail":
				return function(f) {
					f.mandatoryFields = _.map(f.mandatoryFields, loadDimension);
					f.optionalFields  = _.map(f.optionalFields, loadDimension);
					f.sortFields      = _.map(f.sortFields, loadDimension);
					return f;
				};
			case "summary":
				return function(f) {
					f.filter  = _.map(f.filter, loadDimension);
					f.columns = _.map(f.columns, loadDimension);
					return f;
				};
			default:
				throw("ReportsController.prototype.__makeFilterTransform - Unrecognized filter type: " + type);
		}
	};

	/*!
	 * \brief Gets the list of reports from the server if it's not in the cache
	 *
	 * \param report_type Type of reports the user wants to load (all, summary, detail, custom)
	 * \param options An object with properties
	 * \param callback Function to execute when done: \code callback(result, error) \endcode
	 *
	 * \note Set \p options.overlap to \c false if you don't want requests to be caceled by another action
	 *
	 * \scope public
	 */
	ReportsController.prototype.listReports = function(report_type, options) {
		var self = this;
		options = options || { };
		return this.__commonListing({
			accessor  : "reports",
			url       : "designer/list_all",
			type      : report_type,
			overlap   : options.overlap,
			iterator  : function(report, id, cache) {
				report.name = report.name || report.id;
				report.full_name = self.getReportFullLabel(report);
				cache[report.type] = cache[report.type] || [ ];
				insertOrdered(cache[report.type], report, compareByLabel);
				return report;
			}
		});
	};
	
	//Es como la de arriba pero tira el request y nada mas
	ReportsController.prototype.getAllReports = function(callback) {
		callback = callback || $.noop;

		return inConcert.invoke({
			url: this.buildUrl("viewer/list_all"),
			type: "GET",
			timeout: 60000,
			success: function( response )
			{
				callback( true, response );
			},
			error:function( err /*, text_status*/) {
				callback( false, err );
			}
		});
	};
	
	//Get report categories
	ReportsController.prototype.getCategories = function(callback) {
		callback = callback || $.noop;

		return inConcert.invoke({
			url: this.buildUrl("list_categories"),
			type: "GET",
			timeout: 60000,
			success: function( response )
			{
				callback( true, response );
			},
			error:function( err /*, text_status*/) {
				callback( false, err );
			}
		});
	};

	/*!
	 * Loads all dimensions of the specific type
	 * \param type string Specifies which dimensions to list (all|builtin|custom), default: all
	 * \return A promise providing .done(dimensions) and .fail(errorThrown)
	 */
	ReportsController.prototype.listDimensions = function(type) {
		/*
			Se deja en (overlap = true), ya que en la funcion "loadCustomFieldGroups" del "reports_helpers", invoca de forma simultanea para obtener las dimensiones customs (de tipo "api" y "batch_data"),
				y cuando se hace esto, el primer request se aborta.
		*/
		return this.__commonListing({
			accessor: "dimensions",
			url : "dimensions/get_all",
			type: type,
			args: "all", // archive & live
			overlap: true
		});
	};

	ReportsController.prototype.listArchiveDimensions = function(type) {
		/*
			Usa (overlap = true "inConcert.invoke")  en lugar de (overlap = false "ReportsController.wrapInvoke"),
				ya que esta ultima aborta el request en curso, y como la carga de dimensiones de tipo "archive" y de tipo "live",
				se ejecutan de forma simultanea (puede existir mas de un request de este tipo al mimso tiempo).
		*/
		return this.__commonListing({
			accessor: "archiveDimensions",
			url : "dimensions/get_all",
			type: type,
			args: "archive",
			overlap: true
		});
	};

	ReportsController.prototype.listLiveDimensions = function(type) {
		/*
			Usa (overlap = true "inConcert.invoke")  en lugar de (overlap = false "ReportsController.wrapInvoke"),
				ya que esta ultima aborta el request en curso, y como la carga de dimensiones de tipo "archive" y de tipo "live",
				se ejecutan de forma simultanea (puede existir mas de un request de este tipo al mimso tiempo).
		*/
		return this.__commonListing({
			accessor: "liveDimensions",
			url : "dimensions/get_all",
			type: type,
			args: "live",
			overlap: true
		});
	};

	/*!
	};
	 * Loads filters from the server
	 * \param type string Specifies which filters to list (all|builtin|custom), default: all
	 */
	ReportsController.prototype.listLiveDetailFilters = function(type) {
		return this.__listFilters("livedetail", {
			accessor: "liveDetailFilters",
			type    : type,
			iterator: this.__makeFilterTransform("livedetail")
		});
	};

	/*!
	 * Loads filters from the server
	 * \param type string Specifies which filters to list (all|builtin|custom), default: all
	 */
	ReportsController.prototype.listDetailFilters = function(type) {
		return this.__listFilters("detail", {
			accessor: "detailFilters",
			type    : type,
			iterator: this.__makeFilterTransform("detail")
		});
	};

	/*!
	 * \brief Loads all defined summary filters form the server if they're in the cache.
	 * \param type string Specifies which filters to list (all|builtin|custom), default: all
	 * \return A promise providing .done(summaryFilters) and .fail(errorThrown)
	 */
	ReportsController.prototype.listSummaryFilters = function(type) {
		return this.__listFilters("summary", {
			accessor: "summaryFilters",
			type    : type,
			iterator: this.__makeFilterTransform("summary")
		});
	};

	/*!
	 * Loads all metrics, calls \p callback when done
	 * \param type string Specifies which metrics to list (all|builtin|custom), default: all
	 */
	ReportsController.prototype.listMetrics = function(type) {
		var self = this;

		//console.log("START ReportsController.prototype.listMetrics");

		return this.__commonListing({
			accessor : "metrics",
			url      : "metrics/get_all",
			type     : type,
			iterator : function(metric, id)
			{
				metric.id   = metric.id || id;
				metric.name = metric.name || metric.id;
				(metric.requiredFields || []).forEach(function(field) {
					field.dimension = self.dimensions.__by_id[field.dimension] || {
						id  : field.dimension,
						name: field.dimension
					};
				});
				return metric;
			},
			beforeGet: function(callback) {
				self.listDimensions().done(function() {
					callback();
				}).fail(callback);
			}
		});
	};

	/*!
	 * \brief Gets a report meta-data from the server if it's not in the cache
	 * \scope public
	 */
	ReportsController.prototype.loadReport = function(report_type, report_id) {
		var self = this;
		var deferred = new $.Deferred();

		// check if it is already in cache
		var loaded_report = self.m_loadedReports[report_type][report_id];
		if (loaded_report) {
			_.defer(_.bind(deferred.resolve, deferred, loaded_report));
			return deferred;
		}

		var loadReportImpl = function() {
			self.wrapInvoke({
				url        : self.buildUrl("get", report_id),
				type       : "GET",
				dataType   : "json",
				success: function(response) {
					var factory = self.report_factories[report_type];

					var report = factory.create(report_id, response);
					if (response.vcc === "system") {
						self.m_loadedReports[report_type][report_id] = report;
					}

					deferred.resolve(report);
				},
				error: function(xhr, error) {
					deferred.reject(error.result);
				}
			});
		};

		var onError = _.bind(deferred.reject, deferred);

		// make sure everything is loaded
		if (report_type === "summary" || report_type === "dialerbatch" || report_type === "dialercontacts") {
			self.listMetrics().done(function() {
				self.listSummaryFilters()
					.done(loadReportImpl)
					.fail(onError);
			}).fail(onError);

		}
		else if (report_type === "detail" || report_type === "realtime") {
			self.listDetailFilters()
				.done(loadReportImpl)
				.fail(onError);
		}
		else if (report_type === "livedetail") {
			self.listLiveDetailFilters()
				.done(loadReportImpl)
				.fail(onError);
		}
		else {
			throw("ReportsController.prototype.listMetrics - Unrecognized report_type: " + report_type);
		}
		return deferred;
	};

	/*
	 * Loads and returns a dimension
	 *
	 * \return The dimension if \p options.sync is set. If \p options.sync is not
	 * set this function returns a promise which provides \c .done(dimension) and
	 * \c .fail(errorThrown) events
	 *
	 * \remarks This function throws when called with an invalid type, or
	 * when \p options.sync is set but dimensions have not yet been loaded.
	 */
	ReportsController.prototype.loadDimension = function(id, options)
	{
		var extendedOptions = _.extend({
			accessor    : "dimensions",
			listFunction: this.listDimensions
		}, options);

		return this.__commonLoad(id, extendedOptions);
	};

	/*!
	 * Loads and returns a summary or detail filter
	 *
	 * \return The filter if \p options.sync is set. If \p options.sync is not
	 * set this function returns a promise which provides \c .done(filter) and
	 * \c .fail(errorThrown) events
	 *
	 * \remarks This function throws when called with an invalid type, or
	 * when \p options.sync is set but filters have not yet been loaded.
	 */
	ReportsController.prototype.loadFilter = function(id, type, options)
	{
		let info;
		if (type === "detail") {
			info = {
				accessor    : "detailFilters",
				listFunction: this.listDetailFilters
			};
		}
		else if (type === "livedetail") {
			info = {
				accessor    : "liveDetailFilters",
				listFunction: this.listLiveDetailFilters
			};
		}
		else if (type === "summary" || type === "dialerbatch") {
			info = {
				accessor    : "summaryFilters",
				listFunction: this.listSummaryFilters
			};
		}
		else {
			throw("ReportsController.prototype.loadFilter - Unrecognized type: " + type);
		}

		return this.__commonLoad(id, _.extend(info, options));
	};

	/*
	 * Loads and returns a metric
	 *
	 * \return The metric if \p options.sync is set. If \p options.sync is not
	 * set this function returns a promise which provides \c .done(metric) and
	 * \c .fail(errorThrown) events
	 *
	 * \remarks This function throws when called with an invalid type, or
	 * when \p options.sync is set but metrics have not yet been loaded.
	 */
	ReportsController.prototype.loadMetric = function(id, options) {
		return this.__commonLoad(id, _.extend({
			accessor    : "metrics",
			listFunction: this.listMetrics
		}, options));
	};

	ReportsController.prototype.saveReport = function(data)
	{
		var self = this;
		var deferred = new $.Deferred();

		this.listReports("custom").done(function()
		{
			var existing = _.find(self.reports.all, helpers.makeFilter({ name: data.name }));
			if (existing && existing.id !== data.id) {
				_.defer(_.bind(deferred.reject, deferred, "<%=_T('A report with the same name already exists')%>"));
				return deferred;
			}

			var report = data.id && _.find(self.reports.custom, function(e) { return e.id === data.id; });
			self.wrapInvoke({
				url        : self.buildUrl("save"),
				type       : "POST",
				data       : JSON.stringify(data),
				dataType   : "json",
				contentType: "application/json",
				success: function(response) {
					if (report) {
						var old_name = report.name;
						report.name = data.name;
						report.description = data.description;
						if (old_name !== report.name) {
							self.reports.all.sort(compareByName);
							self.reports.custom.sort(compareByName);
							self.reports[report.type].sort(compareByName);
						}
						self.emit("reportUpdated", report);
					} else {
						report = self.reports.__by_id[response.result] = {
							id         : response.result,
							name       : data.name,
							description: data.description,
							type       : data.type
						};

						insertOrdered(self.reports.all, report, compareByName);
						insertOrdered(self.reports.custom, report, compareByName);
						insertOrdered(self.reports[report.type], report, compareByName);
						self.emit("reportAdded", report);
					}

					deferred.resolve(response);
				},
				error: function(xhr, error) {
					deferred.reject(error.result);
				}
			});
		}).fail(_.bind(deferred.reject, deferred));

		return deferred;
	};

	ReportsController.prototype.saveMetric = function(data)
	{
		var self = this;
		var deferred = new $.Deferred();

		var existing = _.find(this.metrics.all, helpers.makeFilter({ name: data.name }));
		if (existing && existing.id !== data.id) {
			_.defer(_.bind(deferred.reject, deferred, "A metric with the same name already exists"));
			return deferred;
		}

		data.requiredFields = [];
		if (!data.composite && (data.integrationType || data.cciProject || data.batch_dataProject || 
			data.process_flowProject) 
		) {
			var handleCustomProject = function(scope, projectId) {
				if (!(scope && projectId)) {
					return [];
				}
				var dimensions = _.select(self.dimensions.custom, helpers.makeFilter({
					scope    : scope,
					projectId: projectId
				}));
				if (dimensions.length > 0) {
					var firstDimension = _.first(dimensions);
					if (firstDimension.requiredFieldName && firstDimension.requiredFieldValue) {
						data.requiredFields.push({
							dimension : firstDimension.requiredFieldName,
							comparator: '=',
							value     : firstDimension.requiredFieldValue
						});
					}
				}
				return dimensions;
			};

			data.metadata.dimensions = _.map( _.union(
				handleCustomProject(data.integrationType, data[data.integrationType + "Project"]),
				handleCustomProject("cci", data.cciProject),
				handleCustomProject("batch_data", data.batch_dataProject),
				handleCustomProject("process_flow", data.process_flowProject)
			), function(d) {
				return _.pick(d, "id", "name", "scope", "sourceName");
			});

			if (data.metadata.dimensions.length === 0) {
				_.defer(_.bind(deferred.reject, deferred, "Selected projects generate no useful data."));
				return deferred;
			}
		}

		var metric = data.id && this.metrics.__by_id[data.id];
		this.wrapInvoke({
			url        : this.buildUrl("metrics/save"),
			type       : "POST",
			data       : JSON.stringify(data),
			dataType   : "json",
			contentType: "application/json",
			success: function(response) {
				if (metric) {
					var old_name = metric.name;
					metric = $.extend(metric, response);
					if (old_name !== metric.name) {
						self.metrics.all.sort(compareByName);
						self.metrics.custom.sort(compareByName);
					}
					self.emit("metricUpdated", metric);
				} else {
					metric = self.metrics.__by_id[response.id] = response;
					insertOrdered(self.metrics.all, metric, compareByName);
					insertOrdered(self.metrics.custom, metric, compareByName);
					self.emit("metricAdded", metric);
				}

				(metric.requiredFields || []).forEach(function(field) {
					field.dimension = self.dimensions.__by_id[field.dimension] || {
						id  : field.dimension,
						name: field.dimension
					};
				});

				deferred.resolve(metric.id);
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	ReportsController.prototype.saveFilter = function(data) {
		var self = this;
		var deferred = new $.Deferred();
		var filtersContainer;

		if (data.type === "detail") {
			filtersContainer = self.detailFilters;
		}
		else if (data.type === "livedetail") {
			filtersContainer = self.liveDetailFilters;
		}
		else if (data.type === "summary") {
			filtersContainer = self.summaryFilters;
		}
		else {
			throw("ReportsController.prototype.saveFilter - Unrecognized data.type " + data.type);
		}

		var existing = _.find(filtersContainer.all, helpers.makeFilter({ name: data.name }));
		if (existing && existing.id !== data.id) {
			_.defer(_.bind(deferred.reject, deferred, "A filter with the same name already exists"));
			return deferred;
		}

		function idsEqual(e1, e2) {
			return e1 && e2 && e1.id === e2.id;
		}

		function setsEqual(set1, set2, strict) {
			return set1.length === set2.length && _.every(set1, strict ? function(e1, idx) {
				return idsEqual(e1, set2[idx]);
			} : function(e1) {
				return _.find(set2, _.partial(idsEqual, e1));
			});
		}

		var isVcc = _.partial(idsEqual, {id: "VCC"});
		var searchIterators = {
			summary: function(f) {
				let isEqualScope = (f.scope || "").toLowerCase() === (data.scope || "").toLowerCase();
				return isEqualScope &&
					setsEqual(f.columns, data.columns, true) &&
					setsEqual(_.reject(f.filter, isVcc), data.filter);
			},
			detail: function(f) {
				let isEqualScope = (f.scope || "").toLowerCase() === (data.scope || "").toLowerCase();
				return isEqualScope &&
					setsEqual(f.sortFields, data.sortFields, true) &&
					setsEqual(_.reject(f.mandatoryFields, isVcc), data.mandatoryFields) &&
					setsEqual(f.optionalFields, data.optionalFields);
			}
		};

		existing = _.find(filtersContainer.all, searchIterators[data.type]);
		if (existing && existing.id !== data.id) {
			_.defer(function() {
				let html_filter_in_conflict = format('<span style="font-weight:bold;font-style:italic;">"{0}"</span>', existing.name || existing.id);
				let message = format("<%=_T('The filter {0} appears to have the same configuration as this one.')%>", html_filter_in_conflict);
				deferred.reject(message);
			});
			return deferred;
		}

		data.requiredFields = { };
		if (data.integrationType || data.cciProject || data.batch_dataProject || 
			data.process_flowProject
		) {
			var handleCustomProject = function(scope, projectId) {
				if (!(scope && projectId)) {
					return;
				}
				var d = _.find(self.dimensions.custom, helpers.makeFilter({
					scope    : scope,
					projectId: projectId
				}));
				if (d && d.requiredFieldName) {
					data.requiredFields[ d.requiredFieldName ] = d.requiredFieldValue;
				}
			};

			handleCustomProject(data.integrationType, data[data.integrationType + "Project"]);
			handleCustomProject("cci", data.cciProject);
			handleCustomProject("batch_data", data.batch_dataProject);
			handleCustomProject("process_flow", data.process_flowProject);
		}

		var filter = data.id && filtersContainer.__by_id[data.id];
		self.wrapInvoke({
			url        : self.buildUrl("filters/save"),
			type       : "POST",
			data       : JSON.stringify(data),
			dataType   : "json",
			contentType: "application/json",
			success: function(response) {
				response = self.__makeFilterTransform(data.type)(response);
				if (filter) {
					var old_name = filter.name;
					filter =  $.extend(filter, response);
					filter.description = buildFilterDescription(filter);
					if (old_name !== filter.name) {
						filtersContainer.all.sort(compareByName);
						filtersContainer.custom.sort(compareByName);
					}
					self.emit("filterUpdated", filter);
				} else {
					filter = filtersContainer.__by_id[response.id] = $.extend({
						type: data.type
					}, response);
					filter.description = buildFilterDescription(filter);
					insertOrdered(filtersContainer.all, filter, compareByName);
					insertOrdered(filtersContainer.custom, filter, compareByName);
					self.emit("filterAdded", filter);
				}

				deferred.resolve(filter.id);
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	ReportsController.prototype.saveDimension = function(data) {
		var self = this;
		var deferred = new $.Deferred();

		var existing = _.find(self.dimensions.all, helpers.makeFilter({ name: data.name }));
		if (existing && existing.id !== data.id) {
			_.defer(_.bind(deferred.reject, deferred, "A field with the same name already exists"));
			return deferred;
		}

		var dimension = data.id && self.dimensions.__by_id[data.id];
		self.wrapInvoke({
			url        : self.buildUrl("dimensions/save"),
			type       : "POST",
			data       : JSON.stringify(data),
			dataType   : "json",
			contentType: "application/json",
			success: function(response) {
				if (dimension) {
					var old_name = dimension.name;
					dimension =  $.extend(dimension, response);
					if (old_name !== dimension.name) {
						self.dimensions.all.sort(compareByName);
						self.dimensions.custom.sort(compareByName);
					}
					self.emit("dimensionUpdated", dimension);
				} else {
					dimension = self.dimensions.__by_id[response.id] = response;
					insertOrdered(self.dimensions.all, dimension, compareByName);
					insertOrdered(self.dimensions.custom, dimension, compareByName);
					self.emit("dimensionAdded", dimension);
				}

				deferred.resolve(dimension.id);
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	ReportsController.prototype.deleteReport = function(reportId) {
		var self = this;
		var deferred = new $.Deferred();

		var report = self.reports.__by_id[reportId];
		if (report === undefined) {
			_.defer(_.bind(deferred.reject, deferred, "<%=_T('Report could not be found')%>"));
			return deferred;
		}

		self.wrapInvoke({
			url        : self.buildUrl("delete"),
			type       : "POST",
			data       : JSON.stringify({ id: report.id }),
			dataType   : "json",
			contentType: "application/json",
			success: function() {
				delete self.reports.__by_id[report.id];
				//delete self.m_loadedReports[report.type][report.id];
				deleteFromArray(self.reports.all,          report.id);
				deleteFromArray(self.reports[report.type], report.id);
				deleteFromArray(self.reports.custom,       report.id);

				self.emit("reportDeleted", report);
				deferred.resolve();
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	ReportsController.prototype.deleteMetric = function(metricId) {
		var self = this;
		var deferred = new $.Deferred();

		var metric = self.metrics.__by_id[metricId];
		if (metric === undefined) {
			_.defer(_.bind(deferred.reject, deferred, "<%=_T('Metric could not be found')%>"));
			return deferred;
		}

		self.wrapInvoke({
			url        : self.buildUrl("metrics/delete"),
			type       : "POST",
			data       : JSON.stringify({ id: metric.id, composite: metric.composite }),
			dataType   : "json",
			contentType: "application/json",
			success: function() {
				delete self.metrics.__by_id[metric.id];
				deleteFromArray(self.metrics.all,    metric.id);
				deleteFromArray(self.metrics.custom, metric.id);

				self.emit("metricDeleted", metric);
				deferred.resolve();
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	ReportsController.prototype.deleteFilter = function(id, type) {
		const self = this;
		let deferred = new $.Deferred();

		let filtersContainer;

		if (type === "detail") {
			filtersContainer = this.detailFilters;
		}
		else if (type === "livedetail") {
			filtersContainer = this.liveDetailFilters;
		}
		else if (type === "summary") {
			filtersContainer = this.summaryFilters;
		}
		else {
			throw("ReportsController.prototype.deleteFilter - Unrecognized data.type " + type);
		}

		var filter = filtersContainer.__by_id[id];
		if (filter === undefined) {
			_.defer(_.bind(deferred.reject, deferred, "<%=_T('Filter could not be found')%>"));
			return;
		}

		this.wrapInvoke({
			url        : this.buildUrl("filters/delete"),
			type       : "POST",
			data       : JSON.stringify({ id: id, type: type }),
			dataType   : "json",
			contentType: "application/json",
			success: function() {
				delete filtersContainer.__by_id[filter.id];
				deleteFromArray(filtersContainer.all,    filter.id);
				deleteFromArray(filtersContainer.custom, filter.id);

				self.emit("filterDeleted", filter);
				deferred.resolve();
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	ReportsController.prototype.deleteDimension = function(id) {
		var self = this;
		var deferred = new $.Deferred();

		var dimension = self.dimensions.__by_id[id];
		if (dimension === undefined) {
			_.defer(_.bind(deferred.reject, deferred, "<%=_T('Custom Field could not be found')%>"));
			return deferred;
		}

		self.wrapInvoke({
			url        : self.buildUrl("dimensions/delete"),
			type       : "POST",
			data       : JSON.stringify({ id: dimension.id }),
			dataType   : "json",
			contentType: "application/json",
			success: function() {
				delete self.dimensions.__by_id[dimension.id];
				deleteFromArray(self.dimensions.all,    dimension.id);
				deleteFromArray(self.dimensions.custom, dimension.id);

				self.emit("dimensionDeleted", dimension);
				deferred.resolve();
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	ReportsController.prototype.execute = function(report, result_size, totals) {
		var self = this;
		var deferred = new $.Deferred();

		var post_data = $.extend({
			utc_offset: new Date().getTimezoneOffset(),
			executed_by_user: true,
			resultSize: result_size || 20
		}, report.exec_data);

		if (totals) {
			post_data = JSON.parse(JSON.stringify(post_data));
			post_data.filter.sliceSize = "Total";
			post_data.last_handle = null;
		}

		self.wrapInvoke({
			url        : self.buildUrl("execute", report.get_id()),
			type       : "POST",
			dataType   : "json",
			contentType: "application/json",
			data       : JSON.stringify(post_data),
			timeout    : 600000,
			success: function(response) {
				if ( typeof(response.result) === "object" ) {
					report.exec_data.last_handle = response.last_handle;
					deferred.resolve(response.result);
				} else { // the server sends result as string to indicate error
					deferred.reject(response.result || "Unknown server error");
				}
			},
			error: function(xhr, error) {
				deferred.reject(error.result);
			}
		});

		return deferred;
	};

	ReportsController.prototype.update = function(report, result_size) {
		// there's nothing special here...
		return this.execute(report, result_size);
	};

	ReportsController.prototype.totals = function(report) {
		return this.execute(report, 1, true);
	};


	/*!
	 * Starts exporting a report, invokes \c callback when download is started or something fails
	 */
	ReportsController.prototype.exportReport = function(report, export_type, callback)
	{
		var post_data = $.extend({
			export_type: export_type,
			utc_offset : new Date().getTimezoneOffset(),
			executed_by_user: true
		}, report.exec_data);

		$.fileDownload(this.buildUrl("export", report.get_id()), {
			httpMethod : "GET",
			cookieName : "downloadStarted",
			data       : {
				data: JSON.stringify(post_data)
			}
		}).done(function() {
			callback(true);
		}).fail(function(response) {
			var errMsg = "";
			try {
				response = JSON.parse(response);
				if (response.relistReportssult) {
					errMsg = response.relistReportssult;
				}
				else if (response.result) {
					errMsg = response.result;
				}
			} catch (err) {}
			callback(false, errMsg);
		});
	};

	ReportsController.prototype.getReportFullLabel = function(report) {
		return format("{0} - {1}",
				inConcert.reports.view_categories[report.view_category || "CUSTOM"].label,
				report.label || report.name
			);
	};

	ReportsController.prototype.__makeCampaigns = function(campaigns) {
		const self = this;
		this.campaigns = {
			timestamp : (new Date()).getTime()
		};
		this.accounts = {};
		this.attLevels = {};
		this.pages = {};
		$.each(campaigns, function(campaignId, campaignAccounts) {
			$.each(campaignAccounts, function(accountId, account) {
				//campaigns
				////all
				self.campaigns.ALL = self.campaigns.ALL || {};
				self.campaigns.ALL[campaignId] = self.campaigns.ALL[campaignId] || {
					dialer : !!account.dialer,
					accounts : {}
				};
				self.campaigns.ALL[campaignId].accounts[accountId] = account;
				////type
				self.campaigns[account.proxyType] = self.campaigns[account.proxyType] || {};
				self.campaigns[account.proxyType][campaignId] = self.campaigns[account.proxyType][campaignId] || {
					dialer : !!account.dialer,
					accounts : {}
				};
				self.campaigns[account.proxyType][campaignId].accounts[accountId] = account;
				//accounts
				////all
				self.accounts.ALL = self.accounts.ALL || {};
				self.accounts.ALL[accountId] = account;
				////type
				self.accounts[account.proxyType] = self.accounts[account.proxyType] || {};
				self.accounts[account.proxyType][accountId] = account;
				//att levels
				$.each(account.attLevels, function(attLevelId/*, _*/) {
					////campaign all
					self.campaigns.ALL[campaignId].attLevels = 
						self.campaigns.ALL[campaignId].attLevels || {};
					self.campaigns.ALL[campaignId].attLevels[attLevelId] = 
						self.campaigns.ALL[campaignId].attLevels[attLevelId] || {};
					////campaign type
					self.campaigns[account.proxyType][campaignId].attLevels = 
						self.campaigns[account.proxyType][campaignId].attLevels || {};
					self.campaigns[account.proxyType][campaignId].attLevels[attLevelId] = 
						self.campaigns[account.proxyType][campaignId].attLevels[attLevelId] || {};
					////att level
					self.attLevels[attLevelId] = self.attLevels[attLevelId] || {};
				});
				//pages
				if (account.pages) {
					$.each(account.pages, function(pageId, page) {
						////campaign all
						self.campaigns.ALL[campaignId].pages = 
							self.campaigns.ALL[campaignId].pages || {};
						self.campaigns.ALL[campaignId].pages[pageId] = page;
						////campaign type
						self.campaigns[account.proxyType][campaignId].pages = 
							self.campaigns[account.proxyType][campaignId].pages || {};
						self.campaigns[account.proxyType][campaignId].pages[pageId] = page;
						////page
						self.pages[pageId] = page;
					});
				}
			});
		});
	};

	ReportsController.prototype.__loadCampaigns = function(callback = $.noop) {
		const self = this;
		if (this.campaigns && this.campaigns.timestamp + CAMPAIGNS_CACHE_TIME > (new Date()).getTime()) {
			return callback(this.campaigns);
		}
		inConcert.invoke({
			url: this.buildUrl("options/campaigns"),
			type: "GET",
			timeout: 60000,
			success: function(campaigns) {
				self.__makeCampaigns(campaigns);
				callback();
			},
			error:function(/*err , text_status*/) {
				inConcert.app.getMainWindow().showBannerMessage({
					message : "<%=_T('Error loading campaigns.')%>",
					messageType : "Error",
					timeout : 5000
				});
				self.__makeCampaigns({});
				callback();
			}
		});
	};

	ReportsController.prototype.getCampaignOptions = function(proxyType, callback = $.noop) {
		proxyType = proxyType || "ALL";
		if (proxyType === "VOICEMAIL") {
			proxyType = "CALL";
		}
		const self = this;
		this.__loadCampaigns(function() {
			var campaigns = self.campaigns[proxyType] || {};
			var ids = _.map(campaigns, function(_, id) {
				return id;
			});
			callback(ids);
		});
	};

	ReportsController.prototype.getAccountOptions = function(proxyType, campaignId, callback = $.noop) {
		proxyType = proxyType || "ALL";
		const self = this;
		this.__loadCampaigns(function() {
			var accounts = campaignId ?
				(self.campaigns[proxyType] && self.campaigns[proxyType][campaignId] &&
					self.campaigns[proxyType][campaignId].accounts || {}) :
				(self.accounts[proxyType] || {});
			var ids = _.map(accounts, function(account) {
				return account.name;
			});
			callback(ids);
		});
	};

	ReportsController.prototype.getAttLevelOptions = function(proxyType, campaignId, accountId, callback) {
		const self = this;
		this.__loadCampaigns(function() {
			var attLevels;
			if (campaignId) {
				if (accountId) {
					attLevels = self.campaigns[proxyType] && self.campaigns[proxyType][campaignId] && 
						self.campaigns[proxyType][campaignId][accountId] && 
						self.campaigns[proxyType][campaignId][accountId].attLevels || {};
				}
				else {
					attLevels = self.campaigns[proxyType] && self.campaigns[proxyType][campaignId] && 
							self.campaigns[proxyType][campaignId].attLevels || {};
				}
			}
			else if (accountId) {
				attLevels = self.accounts[proxyType] && self.accounts[proxyType][accountId] && 
						self.accounts[proxyType][accountId].attLevels || {};
			}
			else {
				attLevels = self.attLevels;
			}
			var ids = _.map(attLevels, function(_, id) {
				return id;
			});
			callback(ids);
		});
	};

	ReportsController.prototype.getPageOptions = function(proxyType, campaignId, accountId, callback) {
		const self = this;
		this.__loadCampaigns(function() {
			var pages;
			if (campaignId) {
				if (accountId) {
					pages = self.campaigns[proxyType] && self.campaigns[proxyType][campaignId] && 
						self.campaigns[proxyType][campaignId][accountId] && 
						self.campaigns[proxyType][campaignId][accountId].pages || {};
				}
				else {
					pages = self.campaigns[proxyType] && self.campaigns[proxyType][campaignId] && 
							self.campaigns[proxyType][campaignId].pages || {};
				}
			}
			else if (accountId) {
				pages = self.accounts[proxyType] && self.accounts[proxyType][accountId] && 
						self.accounts[proxyType][accountId].pages || {};
			}
			else {
				pages = self.pages;
			}
			var ids = _.map(pages, function(account) {
				return account.name;
			});
			callback(ids);
		});
	};

	ReportsController.prototype.getAccounts = function() {
		return this.accounts && this.accounts.ALL || {};
	};

	ReportsController.prototype.getAttLevels = function() {
		return this.attLevels && this.attLevels.ALL || {};
	};

	ReportsController.prototype.getFacebookPages = function() {
		return this.pages || {};
	};

	ReportsController.prototype.getAccountName = function(accountId) {
		var accounts = this.getAccounts();
		return accounts[accountId] && accounts[accountId].name || null;
	};

	ReportsController.prototype.getPageName = function(pageId) {
		var pages = this.getFacebookPages();
		return pages[pageId] && pages[pageId].name || null;
	};

	ReportsController.prototype.getAccountIdByName = function(accountName) {
		accountName = (accountName || "").toLowerCase();
		var id = null;
		$.each(this.getAccounts(), function(accountId, account) {
			if (account.name.toLowerCase() === accountName) {
				id = accountId;
				return false;
			}
		});
		return id;
	};

	ReportsController.prototype.getPageIdByName = function(pageName) {
		pageName = (pageName || "").toLowerCase();
		var id = null;
		$.each(this.getFacebookPages(), function(pageId, page) {
			if (page.name.toLowerCase() === pageName) {
				id = pageId;
				return false;
			}
		});
		return id;
	};

	if (typeof inConcert.reports.ReportsController !== 'function') {
		inConcert.reports.ReportsController = ReportsController;
	}
}());
