/*jshint bitwise:true, browser:true, curly:true, eqeqeq:true, noarg:true, noempty:true, strict:true, undef:true, unused:true, esversion: 9 */
/*jshint quotmark:false */
/*global $, _, inConcert*/

this.inConcert = this.inConcert || {};
this.inConcert.reports = this.inConcert.reports || {};

(function () {
	"use strict";

	var format = inConcert.Utils.format;

	var MessageTypes =
	{
		ERROR: "messageError",
		INFO : "messageInfo"
	};

	function ReportsDashboard(params)
	{
		$.extend(this, {
			m_appList    : [],
			m_previousApp: null,
			m_currentApp : null,
			m_controller : null,
			m_defaultIds : {
				message      : "reports_message_container",
				close_message: "reports_close_message",
				waiting      : "waiting_for_right_panel"
			},
			m_externalIds: params && params.ids || {}
		});
	}

	ReportsDashboard.prototype.getDOMElement = function( id ) {
		return $( "#" + (this.m_externalIds[id] || this.m_defaultIds[id]) || "");
	};

	ReportsDashboard.prototype.registerApp = function(app) {
		var self = this;
		self.m_appList.push(app);
	};

	ReportsDashboard.prototype.currentApp = function() {
		return this.m_currentApp ? this.m_currentApp.name : "";
	};

	ReportsDashboard.prototype.switchToApp = function(app_name) {
		var self = this;

		if (self.m_currentApp && self.m_currentApp.name === app_name) {
			return;
		}

		var target_apps = _.select(self.m_appList, function(app) {
			return (app.name === app_name);
		});
		if (target_apps.length === 0) {
			return;
		}

		if (self.m_currentApp) {
			self.m_currentApp.instance.leave();
			self.m_previousApp = self.m_currentApp;
		}

		self.m_currentApp = target_apps[0];

		var $targets = self.$dashboard.find('[rd-app]');
		$targets.filter('[rd-app!='+ self.m_currentApp.name +']').hide();
		$targets.filter('[rd-app=' + self.m_currentApp.name +']').show();

		$('#navigation_buttons').children('[rd-app!='+self.m_currentApp.name+']').show();
		$('#navigation_buttons').children('[rd-app='+self.m_currentApp.name+']').hide();

		self.m_currentApp.instance.enter();
	};

	ReportsDashboard.prototype.leaveApp = function(app_name) {
		var self = this;

		if (!self.m_previousApp || self.m_currentApp.name !== app_name) {
			return;
		}

		self.switchToApp(self.m_previousApp.name);
	};

	ReportsDashboard.prototype.get_controller = function() {
		return this.m_controller;
	};

	/*!
	 * Call it when you want to block the right panel, it's awesome to make the user wait
	 * \param options object An object with parameters indicating how to block the panel
	 * \li options.with_animation boolean  Set to \c true if you want the animation
	 * \li options.with_message string The text being displayed (of any)
	 */
	ReportsDashboard.prototype.block_right_panel = function(options) {
		options.with_message = options.with_message || "";
		var $waiting_for_right_panel = this.getDOMElement( 'waiting' );
		if (options.with_animation) {
			$waiting_for_right_panel.children('.internalWaitingAnimation').show();
		} else {
			$waiting_for_right_panel.children('.internalWaitingAnimation').hide();
		}
		$waiting_for_right_panel.children('[waiting_message]').text(options.with_message);
		$waiting_for_right_panel.show();
	};

	/*!
	 * The oposite to \c block_right_panel
	 * \returns {undefined}
	 */
	ReportsDashboard.prototype.unblock_right_panel = function() {
		this.getDOMElement( 'waiting' ).hide();
	};

	/*!
	 * Hides or schedules hiding the container
	 * \param options object An object with parameters indicating when to hide the container:
	 * \li \c options.async Indicates whether to schedule hiding the container or not
	 * \li \c options.timeout Indicates when to hide the container, defaults to 5 seconds
	 * if and only if \p async is \c true
	 */
	ReportsDashboard.prototype.hide_message = function(options) {
		options = options || {};

		var self = this;
		var hide_container_impl = function() {
			const $container = self.getDOMElement( 'message' );
			$container.hide();
			if (self.$dashboard) {
			// $dashboard is null when instantiated in administration 'cause render() is never called
				self.$dashboard.removeClass('reportsWithMessagesContainer');
			}
			const remove_message_wrap = function ($message_container) {
				if (!$message_container) {
					return;
				}
				const $parent = $message_container.parent();
				if ($parent.attr(`rpts_container_message_wrap`)) {
					$message_container.unwrap();
				}
			};

			remove_message_wrap($container);
		};

		if (self._message_timer) {
			clearTimeout(self._message_timer);
		}

		if (!options.async) {
			hide_container_impl();
			return;
		}

		self._message_timer = setTimeout(function() {
			self._message_timer = null;
			hide_container_impl();
		}, options.timeout || 5000);
	};

	/*!
	 * Shows a message
	 * \param type \c MessageTypes Specifies the level of the message (info, error)
	 * \param content \c string|object Specifies the message to show, it can be a plain
	 * string or a \c jQuery object
	 * \param timeout \c integer Specifies how many seconds to wait before hiding the message
	 * \return undefined
	 */
	ReportsDashboard.prototype.show_message = function(messageType, content, timeout, wrap_classes) {
		var self = this;
		timeout = timeout || 5000;

		var $container = this.getDOMElement( 'message' );
		$.each(MessageTypes, function(k, v) {
			$container.removeClass(v);
		});

		var $message_content = $container.children('.messageContent');

		if ( typeof(content) === "string" ) {
			$message_content.html('<label class="messageText">' + content + '</label>');
		} else {
			$message_content.html('').append(content);
		}

		$container.addClass(messageType).show();

		if (wrap_classes) {
			const $wrap = $(`<div rpts_container_message_wrap="true" class="${wrap_classes}">`);
			$container.wrap($wrap);
		}

		if (self.$dashboard) {
		// $dashboard is null when instantiated in administration 'cause render() is never called
			self.$dashboard.addClass('reportsWithMessagesContainer')
				.scrollTo($container);
		}
		else {
			$("#conf_grid_scroll").scrollTo(0, 200);
		}

		if (timeout > 0) {
			self.hide_message({
				async  : true,
				timeout: timeout
			});
		}
	};

	ReportsDashboard.prototype.load_inner_page = function(page_url, callback) {
		inConcert.invoke({
			url        : page_url,
			type       : "GET",
			contentType: "text/html",
			success: function(response) {
				callback(response);
			},
			error: function(xhr, error) {
				callback(null, error.result);
			}
		});
	};

	ReportsDashboard.prototype.can_show = function() {
		return inConcert.permissions.to_category( "reports", "viewer" );
	};

	ReportsDashboard.prototype.render = function(callback = $.noop) {
		const self = this;

		function show() {
			self.$dashboard.show();
			callback();
		}

		if (this.$dashboard) {
			return show();
		}

		this.load_inner_page("/inconcert/apps/reports/pages/reports_main.<%=_L%>.html#grunt-cache-bust", function(content, error) {
			if (!content) {
				inConcert.app.getMainWindow().showBannerMessage({
					message : format("<%=_T('Rendering reports dashboard failed with error {0}')%>", error),
					messageType : "Error",
					timeout : 5000
				});
				callback();
				return;
			}

			self.$dashboard = $('#reports_dashboard').html(content);
			self.m_controller = inConcert.reports.m_ReportsController || new inConcert.reports.ReportsController(inConcert.app.getBaseDomain());
			inConcert.reports.m_ReportsController = self.m_controller;

			// static stuff
			self.getDOMElement( 'close_message' )
				.text("<%=_T('close')%>")
				.click(function() {
					self.hide_message();
				});

			// initialize all registers apps
			// some function must be exposed through a proxy in order to avoid undesired behavior
			self.m_appList.forEach(function(app) {
				app.instance = app.factory($.extend({}, self, {
					block_right_panel: function(options) {
						if ( app.name === self.currentApp() ) {
							self.block_right_panel(options);
						}
					},
					unblock_right_panel: function() {
						if ( app.name === self.currentApp() ) {
							self.unblock_right_panel();
						}
					}
				}));
			});

			_.defer(function() {
				self.switchToApp("viewer");
			});

			// finally. show the dashboard
			show();
		});
	};

	inConcert.reports.external = inConcert.reports.external || {};
	if ( typeof inConcert.reports.external.dashboard !== "function" ) {
		inConcert.reports.external.dashboard = ReportsDashboard;
	}

	if (typeof inConcert.reports.dashboard !== 'object') {
		inConcert.reports.dashboard = new ReportsDashboard();
	}
}());
