local json = require "cjson"
local utils = require "shared.apps.utils"

module(..., package.seeall)

json.encode_empty_table_as_array("on")

local m_datasource

function GetName ()
	return "Posts"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

function GetTypes ()
	return { "FACEBOOK", "INSTAGRAM", "YOUTUBE", "TWITTER" }
end

local function IsClosed (post)
	return post.state == "CLOSED" or post.state == "DELETED"
end

---
-- Helper to perform the actual search on ES.
--
function search_elastic (vcc, query)
	local requestInfo = utils.GetElasticSearchInstance():MakeRequestInfo({
		query = query
	})

	local http = require "socket.http"
	local ltn12 = require "ltn12"
	local response_body = {}
	local index_name = ("vcc_posts-%s"):format(vcc)
	local url = ("/%s/_search"):format(index_name)

	LogInfo("[POSTS.models.search_elastic] --> Executing elastic search. '%s:%s - url: %s'\nPOST_DATA: %s",
		inconcert.environment.apps.agent.elasticsearch.host,
		inconcert.environment.apps.agent.elasticsearch.port,
		url,
		requestInfo.encodedData
	)

	local ok, response_code, response_headers, response_status_line = http.request({
		host = inconcert.environment.apps.agent.elasticsearch.host,
		port = inconcert.environment.apps.agent.elasticsearch.port,
		url = url,
		headers = requestInfo.headers,
		method = "POST",
		source = requestInfo.source,
		sink = ltn12.sink.table(response_body),
		redirect = false
	})
	if not ok then
		LogError("[POSTS.models.search_elastic] %s\n%s\n%s\n%s", response_code, json.encode(response_headers), response_status_line, json.encode(response_body))
		return false, { error = response_status_line, status = response_code }
	end

	local raw_response = table.concat(response_body)
	if not response_headers["content-type"]:match("^application/json") then
		LogError("[POSTS.models.search_elastic] Unknown response from elasticsearch. headers: %s\nresponse: %s", json.encode(response_headers), json.encode(raw_response))
		return false, { error = "inconcert unknown", status = 500 }
	end
	local response = json.decode(raw_response)
	if response.error then
		LogError("[POSTS.models.search_elastic] Request failed with the following error: '%s'", json.encode(response.error))
		return false, response
	end

	--LogDebug("[POSTS.models.search_elastic] -> Ok search. response: %s", json.encode(response))
	return true, response
end

function ExistsInElastic (vcc, id)
	local query = {
		query = {
			bool = {
				filter = { term = { _id = id } } 
			}
		}
	}
	local status, result = search_elastic(vcc, query)
	return status and result and result.hits and result.hits.total.value == 1
end

function GetAccountStatistics (id, ks)
	local release = false
	if not ks then
		release = true
		ks = m_datasource.cassandra:CassandraConnection("Interactions")
	end
	local data = ks:GetColumnValue(id, "SocialNetworkAccounts", "statistics")
	local statistics = data and json.decode(data) or nil
	if release then
		ks:Close()
	end
	return statistics
end

function GetPostStatistics (id, ks)
	local release = false
	if not ks then
		release = true
		ks = m_datasource.cassandra:CassandraConnection("Interactions")
	end
	local data = ks:GetColumnValue(id, "SocialNetworkPosts", "statistics")
	local statistics = data and json.decode(data) or nil
	if release then
		ks:Close()
	end
	return statistics
end

function GetPost (id, includeStatistics, ks)
	local post = GetPostFromRedis(id) or GetPostFromArchive(id, ks)
	if not post or type(post) ~= "table" then
		return nil
	end
	if includeStatistics then
		post.statistics = GetPostStatistics(id, ks)
	end
	return post
end

--retrieves a full post JSON from redis
function GetPostFromRedis (id)
	local data = m_datasource.redis:get("SocialNetworkPost:"..id)
	if not data or data == "" then
		return
	end
	local res, decoded = pcall(json.decode, data)
	if res and decoded then
		return decoded
	else
		LogError("Post '%s' detail is undecodable: '%s'", postId, data)
	end
end

function GetPostFromArchive (id, ks)
	local release = false
	if not ks then
		release = true
		ks = m_datasource.cassandra:CassandraConnection("Interactions")
	end

	local data = ks:GetColumnValue(id, "SocialNetworkPosts", "original")
	local post = data and json.decode(data) or nil

	if release then
		ks:Close()
	end
	return post
end

function GetPostRequest (id)
	local data = m_datasource.redis:get("SocialNetworkPostRequest:"..id)
	if not data or data == "" then
		return
	end
	local res, decoded = pcall(json.decode, data)
	if res and decoded then
		return decoded
	else
		LogError("Post request '%s' is undecodable: '%s'", id, data)
	end
end

function SetTemporalFile (id, data)	
	m_datasource.redis:set("SocialNetworkTemporalFile:" .. id, data)
	m_datasource.redis:expire("SocialNetworkTemporalFile:" .. id, 180)
end

function GetTemporalUploadedFile (file_info)
	return m_datasource.redis:get("SocialNetworkTemporalFile:" .. id)
end

function DeleteTemporalUploadedFile (file_info)
	m_datasource.redis:del("SocialNetworkTemporalFile:" .. id)
end
