/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:true, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/*global $, inConcert */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}

/**
 * @namespace
 * @name SupervisorDashboardConfig
 */
 
(function () {

	'use strict';

	const DEFAULT_INSTANCE_TYPE = "CONFIGURATION-GUI";
	const ONLY_DATA_INSTANCE_TYPE = "CONFIGURATION-ONLY-DATA";

	/**
	 * Creates a new archive dashboard instance.
	 *
	 * @author Ellio de Mello (cany@inconcertcc.com)
	 * @constructor
	 * @memberOf SupervisorDashboardConfig
	 * @param dashboard_instance
	 */
	function DashboardConfig(dashboard_instance)
	{
		this.m_type = 'configuration';
		this.m_id = 'dashboard_config';
		this.m_dashboard = dashboard_instance;
		this.m_alarms_config_instances = {};
		this.m_system_indicators = null;
		this.m_ready = false;
		this.initialize();
	}

	if (typeof inConcert.sup_dashboard_config !== 'function') {
		inConcert.sup_dashboard_config = DashboardConfig;
	}

	DashboardConfig.prototype.init = $.noop;
	DashboardConfig.prototype.on_change_config_selector = $.noop;
	DashboardConfig.prototype.refresh = $.noop;

	DashboardConfig.prototype.initialize = function()
	{
		var self = this;
		var model = this.m_dashboard.get_model_manager();

		model.get_system_indicators(function(response)
		{
			if (!response) {
				var message =  "<%=_T('Error retrieving system indicators.')%>";
				self.m_dashboard.show_error_message(message);
				return;
			}
			self.m_system_indicators = response;
			self.m_ready = true;
		});
	};

	DashboardConfig.prototype.is_ready = function()
	{
		return this.m_ready;
	};

	DashboardConfig.prototype.setup = function(campaign_id, instance_type)
	{
		var self = this;

		if (!this.m_ready) {
			// Si aun no ha termiando de procesar los request del server (this.initialize), ejecuta este timeout hacia la misma funcion, esperando que el mismo termine.
			window.setTimeout(function() {
				self.setup(campaign_id, instance_type);
			}, 100);
			return;
		}

		this.m_dashboard.get_config(function(config, error)
		{
			if (error){
				// TODO: agregar mensaje de error
				return;
			}
			config = config || {};
			var campaign_config = config[campaign_id] || {};
			var campaign_alarms = campaign_config.alarms || {};

			var alarms_config = new inConcert.sup_alarms_config({
				parent              : self,
				campaign_id         : campaign_id,
				prefix_container_id : 'sup_dash_camp_alarms_config_',
				currents_alarms     : campaign_alarms,
				instance_type       : instance_type
			});
			/*
			El id de la instancia para el hash se genera concatenando el tipo de instancia.
				ONLY_DATA_INSTANCE_TYPE: Es la instancia utilizada para obtener la configuracion de alarmas para la campana,
											cuando la GUI de la app supervisor aun no se ha iniciado (es decir, no ha entrado a la app supervisor desde que inicio sesion).
				DEFAULT_INSTANCE_TYPE: Es la opcion por defecto y la que utiliza el supervisor para configurar alarmas, etc, etc.
			*/
			var instance_id = campaign_id + "_" + (instance_type === ONLY_DATA_INSTANCE_TYPE ? ONLY_DATA_INSTANCE_TYPE : DEFAULT_INSTANCE_TYPE);
			self.m_alarms_config_instances[instance_id] = alarms_config;
		});
	};

	DashboardConfig.prototype.__get_alarms_instance = function(campaign_id)
	{
		// Primero busco en el hash de instancias la instancia por defecto.
		var instance_id = campaign_id + "_" + DEFAULT_INSTANCE_TYPE;
		var instance = this.m_alarms_config_instances[instance_id];
		if (instance) {
			return instance;
		}
		// Luego si la instancia por defecto aun no se ha instanciado, tengo que obtener la instancia de la campana utilizada para las notificaciones que se visualizan
		//  en el acceso directo a la app supervisor.
		instance_id = campaign_id + "_" + ONLY_DATA_INSTANCE_TYPE;
		instance = this.m_alarms_config_instances[instance_id];
		return instance;
	};

	DashboardConfig.prototype.is_ready_alarms_campaign_config = function(campaign_id)
	{
		var instance = this.__get_alarms_instance(campaign_id);
		return instance ? true : false;
	};

	DashboardConfig.prototype.get_operations_viewer_node = function(campaign_dom_id)
	{
		return $('#sup_dash_operations_viewer_' + campaign_dom_id);
	};

	DashboardConfig.prototype.get_interactions_viewer_node = function(campaign_dom_id)
	{
		return $('#sup_dash_interaction_viewer_' + campaign_dom_id);
	};

	DashboardConfig.prototype.$get_campaign_node = function(campaign_id)
	{
		return this.m_dashboard.get_campaign_node(campaign_id);
	};

	DashboardConfig.prototype.get_campaign_normalized_id = function(campaign_id)
	{
		return this.m_dashboard.get_normalized_id(campaign_id);
	};

	DashboardConfig.prototype.show = function(campaign_id)
	{
		let has_permission = this.m_dashboard.has_permission_to('sup_view_alarms_config', campaign_id);
		if (!has_permission) {
			inConcert.app.getMainWindow().showBannerMessage({ 
				message : "<%=_T('NOT HAS PERMISSIONS TO SHOW ALARMS')%>", 
				timeout : 5000
			});
			return false;
		}

		let $campaign = this.$get_campaign_node(campaign_id);
		$campaign
			.removeClass('alarmsCondition') // si ya existia (no deberia eh!!) la quita primero, para que la misma no quede 2 veces
			.addClass('alarmsCondition');

		let campaign_dom_id = this.get_campaign_normalized_id(campaign_id);

		// hide dashboard and menu
		$campaign.find('[viewer_type="operations"]').hide();
		$campaign.find('[viewer_type="interactions"]').hide();

		$campaign.find('[viewer_type="alarms_config"]').show();
		let $back = this.m_dashboard.$back_button(campaign_id);
		let $full_screen_back_button = this.m_dashboard.$full_screen_back_button(campaign_id);
		$back.show();
		$full_screen_back_button.show();
		this.m_dashboard.on_dashboard_closed(campaign_id, this.m_id);
		
		if ( this.m_dashboard.get_campaign_id_in_full_screen() === campaign_id) {
			const $sup_full_screen_header_container = $('#sup_full_screen_header_container');
			$sup_full_screen_header_container.find('[action_id="campaign_menu"]').show();
			$sup_full_screen_header_container.find('[action_id="campaign_menu_config_view"]').hide();
		}
		else
		{
			$('#sup_dash_panel_menu_' + campaign_dom_id).show();
			$('#sup_dash_panel_menu_config_view_' + campaign_dom_id).hide();
		}

		return true;
	};

	DashboardConfig.prototype.has_configured_alarms = function(campaign_id)
	{
		var alarms_config = this.__get_alarms_instance(campaign_id);
		if (!alarms_config) {
			// TODO: no se deberia dar
			return;
		}
		return alarms_config.has_configured_alarms();
	};

	DashboardConfig.prototype.get_alarms_config = function(campaign_id)
	{
		var alarms_config = this.__get_alarms_instance(campaign_id);
		if (!alarms_config) {
			// TODO: no se deberia dar
			return;
		}
		return alarms_config.get_config();
	};

	DashboardConfig.prototype.get_system_indicators = function()
	{
		return this.m_system_indicators;
	};
}());
