/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global inConcert, _ */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}


/**
 * @namespace
 * @name SupervisorDashboardMonitor
 */
(function () {

	'use strict';

	const _T = inConcert.i18n.getString;
	var format = inConcert.Utils.format;

	/**
	 * Creates a new monitor dashboard instance.
	 *
	 * @author Sebastián Moreno (smoreno@inconcertcc.com)
	 * @constructor
	 * @memberOf SupervisorDashboardMonitor
	 * @param dashboard_instance
	 */
	function DashboardMonitor(dashboard_instance) {

		DashboardMonitor.uber.constructor.call(this, 'monitoring', "<%=_T('Monitoring')%>", dashboard_instance);

		this.permission_id = 'sup_view_monitoring';
		this.navigation_folder_id = 'active';
		this.m_view_classes_name = 'monitoringViewer';
		this.m_model = dashboard_instance.get_model_manager();
		this.m_dashboard = dashboard_instance ? dashboard_instance : null;

		// Search
		this.m_page_by_campaign = {};
		this.m_request_search_monitor = null;

		// current config
		this.m_config = dashboard_instance.get_config_manager();

		// finite state machine
		// possible states by campaign: idle, refreshing, processing
		this.__state = {};
	}

	inConcert.Utils.inherit(DashboardMonitor, inConcert.sup_dashboard_interactions_view);

	if (typeof inConcert.sup_dashboard_monitor !== 'function') {
		inConcert.sup_dashboard_monitor = DashboardMonitor;
	}

	DashboardMonitor.prototype.__allowed_bulk_actions = function(interaction)
	{
		var logged_user = inConcert.app.getLoggedUserId();
		var owner = interaction.getConversationOwner();
		var allowed_bulk_actions = logged_user === owner ? false : true;
		return allowed_bulk_actions;
	};

	DashboardMonitor.prototype.__must_show_alarm = function (campaign_id, interaction)
	{
		var category = "conversation_time_on_state";
		return this.m_dashboard.eval_notifications_alarms(campaign_id, category, interaction);
	};

	DashboardMonitor.prototype.__get_interaction_gui_config = function()
	{
		var self = this;

		var config = DashboardMonitor.uber.__get_interaction_gui_config.call(this);

		config.allowed_bulk_actions = function(interaction) {
			return self.__allowed_bulk_actions(interaction);
		};

		config.columns = $.extend(true, config.columns || {}, {
			interaction_state : { show: true }
		});

		config.get_specifics_interaction_classes = function( interaction, $interaction ) {
			return self.get_specifics_interaction_classes( interaction, $interaction );
		};

		config.callbacks = {
			can_open_client_window: function( interaction, callback )
			{
				callback = callback || $.noop;
				if ( interaction.belongsToMe() ) {
					callback(false);
					return;
				}
				if ( !self.has_permission_to_monitoring( interaction ) ) {
					callback(false);
					return;
				}
				callback(true);
			},
			transfer:{
				on_before_transfer_execute : function(transfer_to, interactions) {
					$.each(interactions, function(i, interaction){
						const view = interaction && inConcert.app.getInteractionManager().getInteractionClient(interaction);
						if (view && view.on_before_transfer_execute) {
							view.on_before_transfer_execute(transfer_to);
						}
					});
				}
			}
		};

		return config;
	};

	DashboardMonitor.prototype.on_change_config_selector = function( campaign_id )
	{
		DashboardMonitor.uber.on_change_config_selector.call(this, campaign_id);
		this.cleanup_search_text(campaign_id);
	};

	DashboardMonitor.prototype.has_permission_to_monitoring = function( interaction )
	{
		var app_id = 'supervisor';
		var action_id = 'sup_monitoring';
		var campaign_id = interaction.campaign;

		if ( !inConcert.permissions.to_action( app_id, action_id, undefined /* category */, campaign_id, true /* no log */ ) ) {
			inConcert.app.getMainWindow().showBannerMessage({
				message : format( "<%=_T('You have not permissions to monitoring conversations on {0} campaign.')%>", campaign_id ), 
				timeout : 5000
			});
			return false;
		}
		return true;
	};

	DashboardMonitor.prototype.__get_prefix = function( /*interaction*/ )
	{
		return 'supervisor_int_monitor_';
	};

	DashboardMonitor.prototype.__get_dom_interaction_params = function( interaction )
	{
		return {
			prefix: this.__get_prefix( interaction ),
			allowed_bulk_actions: this.__allowed_bulk_actions(interaction)
		};
	};

	DashboardMonitor.prototype.process_interactions = function(campaign_id, interactions)
	{
		var self = this;

		if (this.view_has_changed(campaign_id)){
			return;
		}
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		var interactions_container = this.get_interactions_container(campaign_dom_id);
		var scroll_container = interactions_container.parent();
		var scroll_top = scroll_container.scrollTop();
		var detached_container = interactions_container.detach();

		// build markup for every conversation.
		interactions.forEach(function(interaction)
		{
			// create and append node.
			var parmams_interaction = self.__get_dom_interaction_params(interaction);
			var $interaction_node = $( self.m_interaction_gui.$build_interaction_node( interaction, self.m_templates.interaction_node_tpl, parmams_interaction ) );
			self.insert_interaction_in_panel_list( interaction, $interaction_node, detached_container );
			// update interaction on panel
			self.m_interaction_gui.update_interaction_data( interaction, $interaction_node );
		});

		scroll_container.append(detached_container);

		detached_container.find("[int_labels_disposition]").each(function(_, element) {
			self.m_interaction_gui.toggleInteractionLabels($(element));
		});
		scroll_container.scrollTop(scroll_top);
	};

	/**
	* TODO: Documentar. Que es init_in_page. Que nodo es current_node.
	*/
	DashboardMonitor.prototype.search_monitor = function ($current_node, init_in_page, callback = $.noop)
	{
		const self = this;

		console.info("[DashboardMonitor.search_monitor] START ");

		if (this.m_request_search_monitor) {
			console.info("[DashboardMonitor.search_monitor] --> Unable to start search monitor. Already exist another request: " + this.m_request_search_monitor);
			callback(false);
			return;
		}

		let campaign_id = $current_node.attr('campaign_id');
		var search_query = this.m_dashboard.getQueryBuilderFilter(campaign_id);
		if(!search_query) {
			return;
		}
		// Dejo disabled la navegacion, hasta que culmine de procesar el request
		let navigation = this.get_navigation(campaign_id);
		if (navigation) {
			navigation.disable_navigation(true);
		}

		var onSearchFinished = function(filtered_interactions) {
			callback(filtered_interactions);
			self.hide_show_select_all_batch_actions(campaign_id);
			if (navigation) {
				// Dejo enabled la navegacion, ya que ha terminado de procesar el Request.
				navigation.disable_navigation(false);
			}
			self.m_request_search_monitor = null;
		};

		// configuracion seleccionada de la campana en el tree selector

		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		var $dynamic = $('#sup_dash_interaction_viewer_dynamic_' + campaign_dom_id);

		//Pills
		var $pillContainer = $dynamic.find(".form-group-pill-card-outlined");
		var $noPill = $pillContainer.find(".inlineMessagesContainer");
		var $pillGroup = $pillContainer.find(".form-group-pill-container");
		$pillGroup.empty();
		if(search_query && search_query.rules && $.isArray(search_query.rules) && search_query.rules.length > 0)
		{
			var addRemoveButtonClick;
			addRemoveButtonClick = function(ruleId) {
				$("#" + ruleId + "_remove").click(function() {
					var ruleIdx = parseInt(this.id.replace("sup_monitor" + campaign_dom_id + "_pill_", ""));
					//Eliminar del queryBuilder
					if(self.m_dashboard.queryBuilderObj[campaign_dom_id].deleteRule(ruleIdx))
					{
						//Eliminar del contenedor de pills
						$(this).parent().remove();
					}
				});
			};			
			for (var i = 0; i < search_query.rules.length; i++) {
				var ruleId = "sup_monitor" + campaign_dom_id + "_pill_" + i;
				var rulePillText = _T(search_query.rules[i].label || search_query.rules[i].id) + " | " + _T(search_query.rules[i].operator) + " | " + "<strong>" + _T(search_query.rules[i].value) + "</strong>";
				const $lbl = "" +
					"<label id='" + ruleId + "_label' class='form-group-pill'>" +
						"<span>" + rulePillText + "</span>" +
						"<button id='" + ruleId + "_remove' class='pillButtonDelete'></button>" +
					"</label>";
				$pillGroup.append($lbl);
				$pillGroup.show();
				$noPill.hide();

				addRemoveButtonClick(ruleId);
			}
		}
		else
		{
			$pillGroup.hide();
			$noPill.show();
		}

		//labels
		let selected_labels = navigation.get_selected_labels();
		if (selected_labels && $.isArray(selected_labels) && selected_labels.length > 0) {
			var labelsRule = {
				id : "labels",
				operator : "in",
				value : selected_labels
			};
			search_query.rules.push(labelsRule);
		}	

		if(!search_query.rules.find(rule => rule.id === "campaign")) {
			//campaign
			var campaignRule = {
				id : "campaign",
				operator : "equal",
				value : campaign_id
			};
			search_query.rules.push(campaignRule);
		}

		search_query.rules = this.m_dashboard.getCommonConfigRules(search_query.rules, campaign_id);

		var hasCustomFilter = function() {
			return (search_query.rules.find(rule => rule.id === "agent") || search_query.rules.find(rule => rule.id === "labels") || search_query.rules.find(rule => rule.id === "type") !== undefined);
		};

		// if custom filter is set, get all pages.
		var default_page = hasCustomFilter ? 0 : 1;
		var current_page = init_in_page || ( this.m_page_by_campaign[campaign_id] ? this.m_page_by_campaign[campaign_id].filter.page : default_page );

		var filter = {
			//campaign_id        : campaign_id,
			//account_id         : campaign_config.account_id,
			//attention_level_id : campaign_config.al_id,
			page               : current_page,
			rows_per_page      : this.INTERACTIONS_PER_PAGE,
			query              : search_query,
			sort               : [{state_timestamp: { order: "asc" }, priority: {order: "desc"}}]
		};

		this.m_page_by_campaign[campaign_id] = { paginating : false, filter : filter };

		this.m_request_search_monitor = this.m_model.search_monitor_interactions(filter, function(status, result)
		{
			if (status) {
				$dynamic.find("[query_builder_clean]").prop("disabled", !self.m_dashboard.queryBuilderObj[campaign_dom_id].hasRules());
				// Update pagination info.
				self.m_page_by_campaign[campaign_id].totalRows = result.total || 0;
				self.m_page_by_campaign[campaign_id].last_page = Math.ceil(result.total / self.INTERACTIONS_PER_PAGE);
				self.update_current_pagination(campaign_id, result.interactions.length || 0);
				onSearchFinished(result.interactions);
			}
			else{
				inConcert.app.getMainWindow().showBannerMessage({ message : format("<%=_T('Unable to get campaign conversations for monitor view on campaign: {0}.')%>", campaign_id), timeout : 5000});
				onSearchFinished([]);
			}
		});
	};

	DashboardMonitor.prototype.get_view_paged_info = function()
	{
		return this.m_page_by_campaign;
	};

	DashboardMonitor.prototype.execute_new_search = function(campaign_id, callback, cleanup_search_text, filter_params)
	{
		callback = callback || $.noop;
		var self = this;

		if (cleanup_search_text) {
			this.cleanup_search_text(campaign_id);
		}
		if (filter_params) {
			this.set_values_search_text(campaign_id, filter_params);
		}

		this.__state[campaign_id] = 'processing';
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		this.show_waiting(campaign_dom_id, "<%=_T('Searching conversations')%>");
		var $current_node = this.m_dashboard.get_campaign_node(campaign_id);

		this.search_monitor(
			$current_node,
			undefined,
			function(interactions) {
				if (interactions) {
					self.clear_list(campaign_dom_id);
					if (interactions.length === 0){
						self.toggle_no_conversations_msg(campaign_dom_id, "<%=_T('No monitorable conversations to show. Try another search.')%>");
					}
					else{
						self.process_interactions(campaign_id, interactions);
					}
				}
				self.disable_batch_actions(campaign_id);
				self.hide_waiting(campaign_dom_id);
				self.__state[campaign_id] = 'idle';
				callback();
			});
	};

	DashboardMonitor.prototype.init = function($current_node, filter_params, on_init = $.noop)
	{
		DashboardMonitor.uber.init.call(this, $current_node, { has_batch_actions : true });

		var self = this;
		var campaign_id = $current_node.attr('campaign_id');
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);

		//var $search = $current_node.find('[id="sup_dash_interaction_viewer_dynamic_' + campaign_dom_id + '"]');
		//var $search_box = $search.find('[filter_search]');
		//var $search_button = $search.find('[filter_search_execute]');

		// Set agent if it was specified.
		let init_on_page;
		if (filter_params) {
			init_on_page = 1;
		}

		// setup custom monitor pagination.
		this.setup_view_pagination($current_node, campaign_id);

		// initial search
		this.__state[campaign_id] = 'processing';
		this.show_waiting(campaign_dom_id, "<%=_T('Searching conversations')%>"); // Show waiting

		this.search_monitor(
			$current_node,
			init_on_page,
			function(interactions) {
				if (interactions) {
					self.clear_list(campaign_dom_id); // Clear conversations list and setup menu actions.
					if ( interactions.length === 0 ) {
						self.toggle_no_conversations_msg(campaign_dom_id, "<%=_T('No monitorable conversations to show.')%>");
					}
					else {
						self.process_interactions(campaign_id, interactions);
					}
				}
				self.hide_waiting(campaign_dom_id);
				self.on_view_load_finished(campaign_id);
				self.__state[campaign_id] = 'idle';
				on_init();
			}
		);
	};

	DashboardMonitor.prototype.update_current_pagination = function(campaignId, pageInteractions)
	{
		let pageInfo = this.m_page_by_campaign[campaignId];
		if (!pageInfo) {
			return;
		}
		let currentPage = pageInfo.filter.page;
		let lastPage = pageInfo.last_page;
		let totalRows = pageInfo.totalRows;

		DashboardMonitor.uber.update_current_pagination.call(this, campaignId, currentPage, lastPage, pageInteractions, totalRows);
	};

	DashboardMonitor.prototype.refresh = function ($current_node)
	{
		const self = this;

		let campaign_id = $current_node.attr('campaign_id');
		if (this.view_has_changed(campaign_id)){
			return;
		}
		let campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		// go and search for conversations
		this.search_monitor(
			$current_node,
			undefined, 
			function(interactions)
			{
				if (!self.allowed_to_refresh_view(campaign_id)) {
					self.hide_waiting(campaign_dom_id);
					return;
				}
				if (interactions) {
					// if no conversations, show message to the user and return.
					if ( interactions.length === 0 ){
						self.clear_list(campaign_dom_id);
						self.toggle_no_conversations_msg(campaign_dom_id, "<%=_T('No monitorable conversations to show.')%>");
					}
					else {
						// we got conversations, remove no conversations message (if exists.)
						self.remove_no_conversations_msg(campaign_dom_id);
					}
					self.process_off_dom_interaction_update(
						campaign_dom_id,
						interactions,
						self.m_templates.interaction_node_tpl,
						function( interaction, $interaction_node /*, previous_interaction_data*/ ) {
							// update interaction on panel
							self.m_interaction_gui.update_interaction_data( interaction, $interaction_node );
						}
					);
				}
				self.hide_waiting(campaign_dom_id);
			}
		);
	};
}());
