/* jshint smarttabs:true, quotmark:false, esversion: 9 */
/* global $, inConcert */


if (!this.inConcert) {
	this.inConcert = {};
}

(function () {

	"use strict";
	const FORMAT = inConcert.Utils.format;
	const REQUEST_TIMEOUT_MS = 300000; // 5 minutes
	const _T = inConcert.i18n.getString;

	function LuisAPI(params)
	{
		if(typeof(params) !== "object") {
			throw("LuisAPI.Constructor - expected an object with configuration");
		}
		if(typeof(params.uri ) !== "string") {
			throw("LuisAPI.Constructor - base uri is required");
		}

		this.uri = params.uri + "/cognitive_services";
		this.cached_cultures = null;
	}

	if (typeof inConcert.LuisAPI !== "function") {
		inConcert.LuisAPI = LuisAPI;
	}

/********************************************************************************************************
	FUNCIONES GENERICAS
********************************************************************************************************/

	/*
	
	*/
	LuisAPI.prototype.__MakePaginatedParameters = function(params)
	{
		if (typeof(params) !== "object") {
			return "without_pagination";
		}
		if (typeof(params.page) !== "number") {
			return "without_pagination";
		}
		const page_number = params.page < 1 ? 1: params.page;
		const paginated_parameters = {page: page_number, rows_per_page: params.rows_per_page};
		return encodeURIComponent(JSON.stringify(paginated_parameters));
	};

	/*
	
	*/
	LuisAPI.prototype.__Get = function(uri, callback = $.noop)
	{
		if (typeof(uri) !== "string") {
			throw("LuisAPI.prototype.__Get: uri variable must be string type.");
		}
		if (typeof(callback) !== "function") {
			throw("LuisAPI.prototype.__Get: callback variable must be function type.");
		}

		const request_handler = inConcert.invoke({
			url    : uri,
			type   : "GET",
			timeout: REQUEST_TIMEOUT_MS,
			success: function (list, result, server_response)
			{
				if (typeof(server_response) === "object") {
					server_response = JSON.stringify(server_response);
				}
				//console.log(uri + "\nlist: " + list + "\nresult: " + result + "\nserver_response: " + server_response);
				if (result.toLowerCase() === "success") {
					callback(true, list);
				}
				else {
					// Ejemplo: code 204 (nocontent) para el caso de getTrain #7167
					callback(false, server_response);
				}
			},
			error  : function (code, error) {callback(false, error);}
		});
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.__Delete = function(uri, callback, data)
	{
		if (typeof(uri) !== "string") {
			throw("LuisAPI.prototype.__Delete: uri variable must be string type.");
		}
		if (typeof(callback) !== "function") {
			throw("LuisAPI.prototype.__Delete: callback variable must be function type.");
		}

		const request_handler = inConcert.invoke({
			url         : uri,
			type        : "DELETE",
			data        : JSON.stringify(data || {}),
			contentType : "application/json",
			timeout     : REQUEST_TIMEOUT_MS,
			success     : function (response) { callback(true, response); },
			error       : function (code, error) {callback(false, error);}
		});
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.__Post = function(uri, callback, data)
	{
		if (typeof(uri) !== "string") {
			throw("LuisAPI.prototype.__Post: uri variable must be string type.");
		}
		if (typeof(callback) !== "function") {
			throw("LuisAPI.prototype.__Post: callback variable must be function type.");
		}
		if (typeof(data) !== "object") {
			throw("LuisAPI.prototype.__Post: data variable must be object type.");
		}

		const request_handler = inConcert.invoke({
			url         : uri,
			type        : "POST",
			data        : JSON.stringify(data),
			dataType    : "json",
			contentType : "application/json",
			timeout     : REQUEST_TIMEOUT_MS,
			success     : function (response)    { callback(true, response); },
			error       : function (code, error) { callback(false, error);   }
		});
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.__Put = function(uri, callback, data)
	{
		if (typeof(uri) !== "string") {
			throw("LuisAPI.prototype.__Put: uri variable must be string type.");
		}
		if (typeof(callback) !== "function") {
			throw("LuisAPI.prototype.__Put: callback variable must be function type.");
		}
		if (typeof(data) !== "object") {
			throw("LuisAPI.prototype.__Put: data variable must be object type.");
		}

		const request_handler = inConcert.invoke({
			url         : uri,
			type        : "PUT",
			data        : JSON.stringify(data),
			dataType    : "json",
			contentType : "application/json",
			timeout     : REQUEST_TIMEOUT_MS,
			success     : function (response)    { callback(true, response); },
			error       : function (code, error) { callback(false, error  ); }
		});
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.getCognitiveErrorMsg = function(err)
	{
		if (typeof err !== "object") {
			return "";
		}
		if (typeof err.cognitiveServiceError !== "object") {
			return "";
		}
		if (typeof err.cognitiveServiceError.response !== "object") {
			if (typeof err.cognitiveServiceError.reason !== "string") {
				return "";
			}
			const msg_err = err.cognitiveServiceError.reason;
			if (typeof msg_err !== "string") {
				return "";
			}
			return " " + _T(msg_err);
		}
		if (typeof err.cognitiveServiceError.response.error !== "object") {
			return "";
		}
		const msg_err = err.cognitiveServiceError.response.error.message;
		if (typeof msg_err !== "string") {
			return "";
		}
		return " " + _T(msg_err);
	};

/********************************************************************************************************
	SUBSCRIPTIONS
********************************************************************************************************/
	/*
	
	*/
	LuisAPI.prototype.getVCCSubscriptionsKeys = function(callback = $.noop)
	{

		const vcc_info = inConcert.app.get_vcc_info();
		if (!vcc_info) {
			callback(false);
			return;
		}
		if (!vcc_info.MsCognitiveServices) {
			callback(false);
			return;
		}
		if (!vcc_info.MsCognitiveServices.SubscriptionsKeys) {
			callback(false);
			return;
		}
		callback(true, vcc_info.MsCognitiveServices.SubscriptionsKeys);
	};

/********************************************************************************************************
	APPS
********************************************************************************************************/
	/*
	
	*/
	LuisAPI.prototype.loadCultures = function(api_version, callback = $.noop) 
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/load_cultures/
		 */

		
		// desactivo el caché para las cultures dado que son diferentes según la versión del api que se esté usando
		if (this.cached_cultures && this.cached_cultures[api_version]) {
			callback(true, this.cached_cultures[api_version]);
			return; 
		} else {
			this.cached_cultures = {};
		}

		const self = this;
		const uri = FORMAT("{0}/apps/load_cultures/", this.uri);

		const request_handler = this.__Get(uri + api_version, function(success, list) {
			if (success) {
				self.cached_cultures[api_version] = list;

				callback(true, list);
				return;
			}
			// error
			callback(false, list);
		});
		return request_handler;
	};

	LuisAPI.prototype.loadCulturesForIntents = function(api_version, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/load_cultures/
		 */

		
		// desactivo el caché para las cultures dado que son diferentes según la versión del api que se esté usando
		if (this.cached_cultures_for_intents && this.cached_cultures_for_intents[api_version]) {
			callback(true, this.cached_cultures_for_intents[api_version]);
			return; 
		} else {
			this.cached_cultures_for_intents = {};
		}

		const self = this;
		const uri = FORMAT("{0}/apps/load_cultures_for_intents/", this.uri);

		const request_handler = this.__Get(uri + api_version, function(success, list) {
			if (success) {
				self.cached_cultures_for_intents[api_version] = list;
				callback(true, list);
				return;
			}
			// error
			callback(false, list);
		});
		return request_handler;
	};

	
	LuisAPI.prototype.loadAvailableApiVersions = function(callback = $.noop) 
	{

		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/load_available_api_versions/
		 */

		const uri = FORMAT("{0}/apps/load_available_api_versions/", this.uri);

		const request_handler = this.__Get(uri, function(success, list) {
			if (success) {
				callback(true, list);
				return;
			}
			// error
			callback(false, list);
		});
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.loadApps = function(callback, publishedOnly)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/load_all/
		 */

		const uri = FORMAT("{0}/apps/load_all/", this.uri);

		const request_handler = this.__Post(uri, function(success, list) {
			if (success) {
				let appList = list.rows || [];
				if (publishedOnly) {
					appList = [];
					$.each(list.rows || [], function(_, app) {
						if (app.inconcertInfo && app.inconcertInfo.endpointUrl) {
							appList.push(app);
						}
					});
				}
				callback(true, appList);
				return;
			}
			// error
			callback(false, list);
		}, {});
		return request_handler;
	};

	LuisAPI.prototype.addApp = function(data, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/add/
		 */

		const uri = FORMAT("{0}/apps/add/", this.uri);

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	LuisAPI.prototype.importApp = function(data, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/import/
		 */

		const uri = FORMAT("{0}/apps/import", this.uri);

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	LuisAPI.prototype.renameApp = function(data, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/rename/
		 */

		const uri = FORMAT("{0}/apps/rename/", this.uri);

		const request_handler = this.__Put(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteApps = function(data, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/delete/
		 */

		const uri = FORMAT("{0}/apps/delete/", this.uri);

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	LuisAPI.prototype.trainApp = function(data, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/train/
		 */

		const uri = FORMAT("{0}/apps/train/", this.uri);

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
		ATENCION!!!
		Esta funcion NO debe invocarse, sin que previamente se haya realizado un TRAIN (POST) de la APP.
		Ya que sino los scores de los utterances NO cambian.
		Ref #8900
	
		https://westus.dev.cognitive.microsoft.com/docs/services/5890b47c39e2bb17b84a55ff/operations/5890b47c39e2bb052c5b9c46
			train - Get version training status
			Gets the training status of all models (intents and entities) for the specified LUIS app. You must call the train API to train the LUIS app before you call this API to get training status.

			Select the testing console in the region where you created your resource:
	*/
	LuisAPI.prototype.getTrainApp = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/get_train/d469528f-a521-4035-a9ea-36a346b83df8/0.1
		 */

		const uri = FORMAT("{0}/apps/get_train/{1}/{2}",
			this.uri,
			encodeURIComponent(params.appId   || ""),
			encodeURIComponent(params.version || "")
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};

	LuisAPI.prototype.getAppVersion = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/get_version/d469528f-a521-4035-a9ea-36a346b83df8/0.1
		 */

		const uri = FORMAT("{0}/apps/get_version/{1}/{2}",
			this.uri,
			encodeURIComponent(params.appId   || ""),
			encodeURIComponent(params.version || "")
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.loadAppVersions = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/load_versions/d469528f-a521-4035-a9ea-36a346b83df8/0/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/load_versions/{1}/{2}",
			this.uri,
			encodeURIComponent(params.appId   || ""),
			this.__MakePaginatedParameters(params.paginated)
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};

	LuisAPI.prototype.publishApp = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/publish/
		 */

		const uri = FORMAT("{0}/apps/publish/", this.uri);

		const data = {
			appId          : params.appId,
			subscriptionKey: params.subscriptionKey,
			timezone       : params.timezone,
			verbose        : params.verbose,
			language       : params.language,
			// opcionales
			versionId      : params.version
		};

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppSuggestedUtterance = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/suggested_utterances/delete/f23db312-1ff6-47d5-be25-f6a739daf5ee/el utterance/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/suggested_utterances/delete/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId         || ""),
			encodeURIComponent(params.utteranceName || ""),
			encodeURIComponent(params.version       || "")
		);

		const request_handler = this.__Delete(uri, callback);
		return request_handler;
	};

/********************************************************************************************************
	INTENTS
********************************************************************************************************/
	/*
	
	*/
	LuisAPI.prototype.loadAllIntents = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/intents/load_all/d469528f-a521-4035-a9ea-36a346b83df8/0/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/intents/load_all/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId || ""),
			this.__MakePaginatedParameters(params.paginated),
			encodeURIComponent(params.version || "")
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.createAppIntent = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/intents/create/
		 */

		const uri = FORMAT("{0}/apps/intents/create/", this.uri);

		const data = {
			appId: params.appId,
			name : params.name
		};
		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	LuisAPI.prototype.renameAppIntent = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/intents/rename/d469528f-a521-4035-a9ea-36a346b83df8/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */

		const uri = FORMAT("{0}/apps/intents/rename/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.intentId || ""),
			encodeURIComponent(params.version  || "")
		);

		const data = {
			name: params.name
		};
		const request_handler = this.__Put(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppIntent = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/intents/delete/d469528f-a521-4035-a9ea-36a346b83df8/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/intents/delete/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.intentId || ""),
			encodeURIComponent(params.version  || "")
		);

		const request_handler = this.__Delete(uri, callback);
		return request_handler;
	};



/********************************************************************************************************
	UTTERANCES
********************************************************************************************************/

	/*
	
	*/
	LuisAPI.prototype.loadUtterancesByAppIntent = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/intents/utterances/load_all/d469528f-a521-4035-a9ea-36a346b83df8/INTENT_NAME/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/intents/utterances/load_all/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId      || ""),
			encodeURIComponent(params.intentName || ""),
			encodeURIComponent(params.version    || "")
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.addUtteranceAppIntent = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/intents/utterances/add/
		 */

		const uri = FORMAT("{0}/apps/intents/utterances/add/", this.uri);

		// requerido
		let data = {
			appId     : params.appId,
			intentName: params.intentName,
			utterance : params.utterance,
			language  : params.language,  
			dataset      : params.dataset 
		};

		// opcional
		if (params.entityLabels) {
			data.entityLabels = [];
			$.each(params.entityLabels, function(_, info) {
				data.entityLabels.push({
					entityName    : info.entityName,
					startCharIndex: info.startCharIndex,
					endCharIndex  : info.endCharIndex,
					children      : info.children
				});
			});
		}

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppUtterance = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/intents/utterances/delete/f23db312-1ff6-47d5-be25-f6a739daf5ee/-8010780/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/intents/utterances/delete/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId       || ""),
			encodeURIComponent(params.utteranceId || ""),
			encodeURIComponent(params.version     || "")
		);

		const request_handler = this.__Delete(uri, callback);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.loadSuggestedUtterancesByAppIntent = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/intents/utterances/suggested/load_all/d469528f-a521-4035-a9ea-36a346b83df8/INTENT_ID/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/intents/utterances/suggested/load_all/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.intentId || ""),
			encodeURIComponent(params.version  || "")
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};


/********************************************************************************************************
	ENTITIES MACHINE LEARNED, COMPOSITE, HIERARCHICAL
********************************************************************************************************/

	/*
	
	*/
	LuisAPI.prototype.loadAllEntities = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/load_all/d469528f-a521-4035-a9ea-36a346b83df8/0/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/entities/load_all/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId   || ""),
			this.__MakePaginatedParameters(params.paginated),
			encodeURIComponent(params.version || "")
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};

/********************************************************************************************************
	ENTITIES MACHINE LEARNED
********************************************************************************************************/

	/*
	
	*/
	LuisAPI.prototype.createAppEntityMachineLearned = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/machine_learned/create/
		 */

		const uri = FORMAT("{0}/apps/entities/machine_learned/create/", this.uri);

		const data = {
			appId: params.appId,
			name : params.name,
			is_new : params.is_new
		};

		if (params.children) {
			data.children = [];
			params.children.forEach(child => {
				data.children.push(child);
			});
		}

		if (params.regex) {
			data.regex = [];
			params.regex.forEach(child => {
				data.regex.push(child);
			});
		}

		if (params.prebuilts) {
			data.prebuilts = [];
			params.prebuilts.forEach(child => {
				data.prebuilts.push(child);
			});
		}

		if (params.list) {
			data.list = [];
			params.list.forEach(child => {
				data.list.push(child);
			});
		}

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.renameAppEntityMachineLearned = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/machine_learned/rename/d469528f-a521-4035-a9ea-36a346b83df8/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */

		const uri = FORMAT("{0}/apps/entities/machine_learned/rename/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.entityId || ""),
			encodeURIComponent(params.version  || "")
		);

		const data = {
			name: params.name
		};

		const request_handler = this.__Put(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppEntityMachineLearned = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/machine_learned/delete/f23db312-1ff6-47d5-be25-f6a739daf5ee/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/entities/machine_learned/delete/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.entityId || ""),
			encodeURIComponent(params.version  || "")
		);

		const request_handler = this.__Delete(uri, callback);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.createAppEntityChildMachineLearned = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/machine_learned/create_child/
		 */

		const uri = FORMAT("{0}/apps/entities/machine_learned/create_child/", this.uri);

		const data = {
			appId   : params.appId,
			entityId: params.entityId,
			name    : params.name
		};

		if (params.children) {
			data.children = [];
			params.children.forEach(child => {
				data.children.push(child);
			});
		}

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppEntityChildMachineLearned = function(params, callback = $.noop)
	{
		return this.deleteAppEntityMachineLearned(params, callback);
	};

	/*
	
	*/
	LuisAPI.prototype.renameAppEntityChildMachineLearned = function(params, callback = $.noop)
	{
		return this.renameAppEntityMachineLearned(params, callback);
	};

/********************************************************************************************************
	ENTITIES COMPOSITE
********************************************************************************************************/

	/*
	
	*/
	LuisAPI.prototype.createAppEntityComposite = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/composite/create/
		 */

		const uri = FORMAT("{0}/apps/entities/composite/create/", this.uri);

		const data = {
			appId   : params.appId,
			name    : params.name,
			children: []
		};

		params.children.forEach(child => {
			data.children.push(child.name);
		});
		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.renameAppEntityComposite = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/composite/rename/d469528f-a521-4035-a9ea-36a346b83df8/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */

		const uri = FORMAT("{0}/apps/entities/composite/rename/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.entityId || ""),
			encodeURIComponent(params.version  || "")
		);

		const data = {
			name    : params.name,
			children: []
		};

		params.children.forEach(child => {
			data.children.push(child.name);
		});
		const request_handler = this.__Put(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppEntityComposite = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/composite/delete/f23db312-1ff6-47d5-be25-f6a739daf5ee/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/entities/composite/delete/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.entityId || ""),
			encodeURIComponent(params.version  || "")
		);

		const request_handler = this.__Delete(uri, callback);
		return request_handler;
	};

/********************************************************************************************************
	ENTITIES HIERARCHICAL
********************************************************************************************************/

	/*
	
	*/
	LuisAPI.prototype.createAppEntityHierarchical = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/hierarchical/create/
		 */
		const uri = FORMAT("{0}/apps/entities/hierarchical/create/", this.uri);

		const data = {
			appId   : params.appId,
			name    : params.name,
			children: []
		};

		params.children.forEach(child => {
			data.children.push(child.name);
		});

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.renameAppEntityHierarchical = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/hierarchical/rename/d469528f-a521-4035-a9ea-36a346b83df8/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */

		const uri = FORMAT("{0}/apps/entities/hierarchical/rename/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.entityId || ""),
			encodeURIComponent(params.version  || "")
		);

		const data = {
			name    : params.name,
			children: []
		};

		params.children.forEach(child => {
			data.children.push(child.name);
		});
		const request_handler = this.__Put(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppEntityHierarchical = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/hierarchical/delete/f23db312-1ff6-47d5-be25-f6a739daf5ee/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/entities/hierarchical/delete/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.entityId || ""),
			encodeURIComponent(params.version  || "")
		);

		const request_handler = this.__Delete(uri, callback);
		return request_handler;
	};


/********************************************************************************************************
	ENTITIES PREBUILT
********************************************************************************************************/

	/*
	
	*/
	LuisAPI.prototype.addAppEntityPrebuilt = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/prebuilt/add/
		 */
		const uri = FORMAT("{0}/apps/entities/prebuilt/add/", this.uri);

		const data = {
			appId: params.appId,
			list : []
		};

		params.children.forEach(child => {
			data.list.push(child.name);
		});

		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppEntityPrebuilt = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/prebuilt/delete/f23db312-1ff6-47d5-be25-f6a739daf5ee/7b1d9a5f-2804-42a1-9b58-03b07788477a/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/entities/prebuilt/delete/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.entityId || ""),
			encodeURIComponent(params.version  || "")
		);

		const request_handler = this.__Delete(uri, callback);
		return request_handler;
	};

	/*
	*/
	LuisAPI.prototype.loadAvailablePrebuiltEntityList = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/entities/prebuilt/load_available/d469528f-a521-4035-a9ea-36a346b83df8/0.1
		 */
		const uri = FORMAT("{0}/apps/entities/prebuilt/load_available/{1}/{2}",
			this.uri,
			encodeURIComponent(params.appId   || ""),
			encodeURIComponent(params.version || "")
		);

		const request_handler = this.__Get(uri, function(success, list) {
			if (success) {
				callback(true, list);
				return;
			}
			// error
			callback(false, list);
		});
		return request_handler;
	};


/********************************************************************************************************
	FEATURES
********************************************************************************************************/

	/*
	
	*/
	LuisAPI.prototype.loadAllFeatures = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/features/load_all/d469528f-a521-4035-a9ea-36a346b83df8/0/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/features/load_all/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId   || ""),
			this.__MakePaginatedParameters(params.paginated),
			encodeURIComponent(params.version || "")
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.loadAllFeaturesPhrases = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/features/phraselist/load_all/d469528f-a521-4035-a9ea-36a346b83df8/0/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/features/phraselist/load_all/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId   || ""),
			this.__MakePaginatedParameters(params.paginated),
			encodeURIComponent(params.version || "")
		);

		const request_handler = this.__Get(uri, callback);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.createAppFeaturesPhrase = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/features/phraselist/create/
		 */

		const uri = FORMAT("{0}/apps/features/phraselist/create/", this.uri);

		const data = {
			appId         : params.appId,
			name          : params.name,
			phrases       : params.phrases,
			isExchangeable: params.isExchangeable,
			isActive      : params.isActive
		};
		const request_handler = this.__Post(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.updateAppFeaturesPhrase = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *  http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/features/phraselist/update/d469528f-a521-4035-a9ea-36a346b83df8/113023/0.1
		 */

		const uri = FORMAT("{0}/apps/features/phraselist/update/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.phraseId || ""),
			encodeURIComponent(params.version  || "")
		);

		const data = {
			appId         : params.appId,
			name          : params.name,
			phrases       : params.phrases,
			isExchangeable: params.isExchangeable,
			isActive      : params.isActive
		};
		const request_handler = this.__Put(uri, callback, data);
		return request_handler;
	};

	/*
	
	*/
	LuisAPI.prototype.deleteAppFeaturesPhrase = function(params, callback = $.noop)
	{
		/*
		 * sample URI:
		 *   http://i6.inconcert.cc/inconcert/apps/cognitive_services/cognitive_services/apps/features/phraselist/delete/d469528f-a521-4035-a9ea-36a346b83df8/113023/0.1
		 */
		params = params || {};

		const uri = FORMAT("{0}/apps/features/phraselist/delete/{1}/{2}/{3}",
			this.uri,
			encodeURIComponent(params.appId    || ""),
			encodeURIComponent(params.phraseId || ""),
			encodeURIComponent(params.version  || "")
		);

		const request_handler = this.__Delete(uri, callback);
		return request_handler;
	};

}());
