local xml = nil --on demand
local socketUrl = nil --on demand
local http = nil--on demand
local https = nil
local ltn12 = nil--on demand

module(..., package.seeall)

function makeUrl(domain, port, path, isHttps)
	return (isHttps and "https://" or "http://") .. domain .. (port and (":" .. port) or "") .. (path or "")
end

local function isHttps(url)
	return url:match("https://") and true or false
end

function call(method, url, urlParams, body, contentType, headers)
	ltn12 = ltn12 or require "ltn12"
	
	local urlParams_s = encodeUrlEncodedRequest(urlParams)
	url = url .. (urlParams_s ~= "" and ("?" .. urlParams_s) or "")
	body = body or ""
	headers = headers or {}
	headers["Content-Length"] = tostring(#body)
	if contentType then
		headers["Content-Type"] = contentType
	end
	
	local source = ltn12.source.string(body)
	local ok, response_code, response_headers, response_status_line
	local response_body = {}

	if isHttps(url) then
		if not https then
			local Synchronize = require "Synchronize"
			https = Synchronize("luasec_mutex", function()
				return require ("ssl.https")
			end)
		end
		ok, response_code, response_headers, response_status_line = https.request({
			url = url,
			method = method,
			headers = headers,
			source = source,
			sink = ltn12.sink.table(response_body)
		})
	else
		http = http or require "socket.http"
		ok, response_code, response_headers, response_status_line = http.request({
			url = url,
			method = method,
			headers = headers,
			source = source,
			sink = ltn12.sink.table(response_body)
		})
	end

	response_body = table.concat(response_body)
	if not ok and type(response_code) ~= "number" then
		response_code = -1
	end
	return response_code, response_body, response_headers
end

function get(url, urlParams, headers)
	return call("GET", url, urlParams, nil, nil, headers)
end

function post(url, urlParams, body, contentType, headers)
	return call("POST", url, urlParams, body, contentType, headers)
end

function put(url, urlParams, body, contentType, headers)
	return call("PUT", url, urlParams, body, contentType, headers)
end

function delete(url, urlParams, body, contentType, headers)
	return call("DELETE", url, urlParams, body, contentType, headers)
end

function callObject(method, url, urlParams, requestType, requestData, headers, responseType, errorResponseType)
	requestType = requestType or "text"
	requestData = requestData or ""
	responseType = responseType or "text"
	
	local body = encodeRequest(requestType, requestData)
	local contentType = getContentType(requestType)

	local response_code, response_body, response_headers = call(method, url, urlParams, body, contentType, headers)

	local isError = response_code ~= 200
	responseType = isError and errorResponseType or responseType
	local status, responseData = pcall(decodeResponse, responseType, response_body)
	if status then
		response_body = responseData
	elseif not isError then
		error("error parsing response: " .. response_body)
	end
	return response_code, response_body, response_headers
end

function getObject(url, urlParams, headers, responseType, errorResponseType)
	return callObject("GET", url, urlParams, "text", "", headers, responseType, errorResponseType)
end

function postObject(url, urlParams, requestType, requestData, headers, responseType, errorResponseType)
	return callObject("POST", url, urlParams, requestType, requestData, headers, responseType, errorResponseType)
end

function putObject(url, urlParams, requestType, requestData, headers, responseType, errorResponseType)
	return callObject("PUT", url, urlParams, requestType, requestData, headers, responseType, errorResponseType)
end

function deleteObject(url, urlParams, requestType, requestData, headers, responseType, errorResponseType)
	return callObject("DELETE", url, urlParams, requestType, requestData, headers, responseType, errorResponseType)
end

function getContentType(dataType)
	if dataType == "text" then
		return "text/plain"
	elseif dataType == "json" then
		return "application/json"
	elseif dataType == "xml" then
		return "application/xml"
	elseif dataType == "url" then
		return "application/x-www-form-urlencoded"
	else
		error("unsuported dataType " .. dataType)
	end
end

function encodeRequest(dataType, data)
	if dataType == "text" then
		return tostring(data)
	elseif dataType == "json" then
		return encodeJsonRequest(data)
	elseif dataType == "xml" then
		return encodeXmlRequest(data)
	elseif dataType == "url" then
		return encodeUrlEncodedRequest(data)
	else
		error("unsuported dataType " .. dataType)
	end
end

function decodeResponse(dataType, response)
	if dataType == "text" then
		return tostring(response)
	elseif dataType == "json" then
		return decodeJsonResponse(response)
	elseif dataType == "xml" then
		return decodeXmlResponse(response)
	elseif dataType == "url" then
		return decodeUrlEncodedResponse(response)
	else
		error("unsuported dataType " .. dataType)
	end
end

function encodeJsonRequest(data)
	return json.encode(data)
end

function decodeJsonResponse(response)
	return json.decode(response)
end

function encodeUrlEncodedRequest(data)
	socketUrl = socketUrl or require "socket.url"
	data = data or {}
	local request = ""
	for k, v in pairs(data) do
		request = request .. (request == "" and "" or "&") .. k .. "=" .. socketUrl.escape(v)
	end
	return request
end

function decodeUrlEncodedResponse(response)
	socketUrl = socketUrl or require "socket.url"
	local data = {}
	string.gsub(response or "", "([^&]+)", function(value) 
		local k, v = string.match(value, "([^=]+)=([^=]+)")
		data[k] = socketUrl.unescape(v)
	end)
	return data
end

function encodeXmlRequest(data)
	xml = xml or require "pl.xml"
	
	local function getObject(name, data, childrenNames)
		local xmlData = xml.new(name)
		local dataType = type(data)
		if dataType == "string" then
			xmlData:text(data)
		elseif dataType == "number" then
			xmlData:text(tostring(data))
		elseif dataType == "boolean" then
			xmlData:text(data and "true" or "false")
		elseif dataType == "table" then
			for key, value in pairs(data) do
				local childName
				if type(key) == "number" then
					childName = childrenNames[name] or (name .. "_child")
				else
					childName = key
				end
				local childXml = getObject(childName, value)
				xmlData:add_direct_child(childXml)
			end
		end
		return xmlData
	end

	local xmlObject = getObject(data.rootName or "root", data.data, data.childrenNames or {})
	return xml.tostring(xmlObject)
end

function decodeXmlResponse(response)
	xml = xml or require "pl.xml"
	
	local function loadData(root, xmlData)
		local key = xmlData.tag
		local text = xmlData:get_text()
		local data = ""
		if text ~= "" then
			data = text
		elseif xmlData:first_childtag() then
			data = {}
			for xmlChild in xmlData:children() do
				data = loadData(data, xmlChild)
			end
		end
		if #root > 0 then
			--is an array
			table.insert(root, data)
		elseif root[key] then
			--should be an array
			local firstElement = root[key]
			root = {}
			table.insert(root, firstElement)
			table.insert(root, data)
		else
			--is an object
			root[key] = data
		end
		return root
	end

	local xmlObject = xml.parse(response)
	local data = loadData({}, xmlObject)
	return data
end
