/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 9 */

/* global inConcert */

/*
Documentacion: https://developer.mozilla.org/en-US/docs/Web/API/notification

Sample Usage:
	let desktop_notifications = inConcert.app.getDesktopNotifications();
	desktop_notifications.notify({
		message: "Notification GUID " + inConcert.Utils.generateGuid()
	});
*/

if (!this.inConcert) {
	this.inConcert = {};
}

(function () {
	"use strict";

	const _T = inConcert.i18n.getString;
	const FORMAT = inConcert.Utils.format;
	const CHANNEL_IMG_PATH = "/inconcert/shared/notifications/images/";

	//---------------------------------------------------------------------
	// DesktopNotifications
	//
	function DesktopNotifications ()
	{
		this.m_browser_supports = false;
		this.m_listening = false;
		this.m_has_permission = null;
		this.m_enabled_sound_notifications = false;
		this.m_enabled_visual_notifications = false;
		this.m_away_sound = null;
		this.m_play_promise = null; // sound 
		this.m_notifications_by_interactions = {};
		this.m_interactions_by_notifications = {};
		this.__initialize();
	}

	if (typeof inConcert.DesktopNotifications !== "function") {
		inConcert.DesktopNotifications = DesktopNotifications;
	}

	DesktopNotifications.prototype.__initialize = function()
	{
		const self = this;

		if (!window.Notification) {
			let browser_name = inConcert.Utils.browserDetect().name;
			let message = FORMAT("Your '{0}' browser does not support desktop notifications.", browser_name);
			inConcert.logger.warn({
				action_name: "inConcert.DesktopNotifications",
				action_type: "initialize",
				message    : message
			});
			return;
		}

		inConcert.app.messagebus.subscribe("/loaded_user_permissions/updated", function(_) {
			self.update_permissions();
		});

		this.m_browser_supports = true;
	};

	/*
		Funcion publica.
		Que es invocada cuando se realiza el logut del agente.
	*/
	DesktopNotifications.prototype.dispose = function()
	{
		this.__close_all_notifications();
		this.__stop_sound();
	};

	DesktopNotifications.prototype.browser_supports_desktop_notifications = function()
	{
		return this.m_browser_supports;
	};

	/*
		Funcion publica que se invoca cuando el usuario ha terminado de iniciar sesion y la app se encuentra ready (clase MainWindow)
		$(document).ready
			..
			inConcert.app.bootstrap(...)
	*/
	DesktopNotifications.prototype.start_listening = function(callback = $.noop)
	{
		const self = this;

		if (!this.browser_supports_desktop_notifications()) {
			let message = "Your browser does not support desktop notifications.";
			console.warn({
				action_name: "DesktopNotifications",
				action_type: "set_visual_enabled",
				message    : message
			});
			this.m_listening = false;
			callback(false, message);
			return;
		}
		if (Notification.permission === "granted") {
			this.m_listening = true;
			callback(true);
			return;
		}
		Notification.requestPermission(function (permission) {
			if (permission === "granted") {
				self.m_listening = true;
				callback(true);
			}
			else {
				/*
					Aca, puede que el usuario haya denegado el permiso, o tambien puede ser un error del navegador, Ej:
					Puede dar este Error en los iFrame en CHROME:

					Permission for the Notification API may no longer be requested from a cross-origin iframe. You should consider requesting permission from a top-level frame or opening a new window instead. See https://www.chromestatus.com/feature/6451284559265792 for more details.
				*/
				let message = _T("Notification permission is denied.");
				console.warn({
					action_name: "DesktopNotifications",
					action_type: "set_visual_enabled",
					message    : message
				});
				self.m_listening = false;
				callback(false, message);
			}
		});
	};

	DesktopNotifications.prototype.stop_listening = function()
	{
		this.m_listening = false;
	};

	DesktopNotifications.prototype.update_permissions = function()
	{
		this.m_has_permission = inConcert.permissions.to_action("settings", "change_personal_settings");
	};

	/*
		Funcion Privada.
		Inicia la reproduccion de un sonido en la notificacion.
	*/
	DesktopNotifications.prototype.__play_sound = function()
	{
		const self = this;
	
		var play_sound = function() {
			if (!self.m_enabled_sound_notifications) {
				return;
			}
			let url = "inconcert/apps/agent/sounds/notif_away.mp3";
	
			self.m_away_sound = new Audio(url);
			if (!self.m_away_sound) {
				return;
			}
			let times = 1;
			self.m_away_sound.addEventListener("ended", function() {
				if (times > 0) {
					this.currentTime = 0;
					times = times -1;
					this.play();
				}
			}, false);
	
			times = times -1;
			self.m_play_promise = self.m_away_sound.play();
		};
		if (self.m_away_sound) {
			self.__stop_sound(function() {
				play_sound();
			});
			return;
		}
		play_sound();
	};

	/*
		Funcion Privada.
			Detiene el sonido que se encuentra reproduciendo en la notificacion

			El promise del sound (no de la notificacion visual), genera un error en consola de Chrome, por lo que hay que quedarse con el promise en el play para luego poder detenerlo.
			El error:
				Uncaught (in promise) DOMException: The play() request was interrupted by a call to pause().
			Referncia a la solucion:
				Ref: https://developers.google.com/web/updates/2017/06/play-request-was-interrupted
	*/
	DesktopNotifications.prototype.__stop_sound = function(callback = $.noop)
	{
		if (!this.m_away_sound) {
			callback();
			return;
		}
		if (this.m_play_promise) {
			const self = this;
			this.m_play_promise.then(_ => {
				// Automatic playback started!
				// Show playing UI.
				// We can now safely pause video...
				self.m_away_sound.pause();
				self.m_away_sound.currentTime = 0;
				self.m_away_sound = null;
				self.m_play_promise = null;
				callback();
			  })
			  .catch(error => {
				callback();
				// Auto-play was prevented
				// Show paused UI.
			  });
		}
		else {
			callback();
		}
	};

	DesktopNotifications.prototype.__get_campaign_name = function(interaction)
	{
		if (!interaction) {
			return;
		}
		let campaign_id = interaction.campaign;
		let campaign = inConcert.app.getCampaign(campaign_id);
		let campaign_name = campaign ? campaign.getDescription() : campaign_id;
		return campaign_name;
	};

	DesktopNotifications.prototype.__get_channel_info = function(type = "")
	{
		type = type.toUpperCase();

		let channels_info = {
		// Los canales ENABLED ---
			CHAT: {
				enabled: true,
				name: _T("Web chat"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertWebchat.png";}
			},
			FACEBOOK: {
				enabled: true,
				name: _T("Facebook"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertFacebook.png";}
			},
			INSTAGRAM: {
				enabled: true,
				name: _T("Instagram"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertInstagram.png";}
			},
			YOUTUBE: {
				enabled: true,
				name: _T("Youtube"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertYoutube.png";}
			},
			APPSTORE: {
				enabled: true,
				name: _T("Appstore"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertAppstore.png";}
			},
			GOOGLEPLAY: {
				enabled: true,
				name: _T("Googleplay"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertGoogleplay.png";}
			},
			RCS: {
				enabled: true,
				name: _T("Rcs"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertRcs.png";}
			},
			INTERNAL_CHAT: {
				enabled: true,
				name: _T("Internal Chat"),
				img : function(interaction) {
				//	return CHANNEL_IMG_PATH + "inconcertInternalChat.png";
					return inConcert.app.getBaseUrl() + interaction.getAvatarURI();
				}
			},
			MAIL: {
				enabled: true,
				name: _T("E-mail"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertMail.png";}
			},
			SMS: {
				enabled: true,
				name: _T("SMS"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertSMS.png";}
			},
			TWITTER: {
				enabled: true,
				name: _T("Twitter"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertTwitter.png";}
			},
			WEBCONTACT: {
				enabled: true,
				name: _T("Web contact"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertWebContact.png";}
			},
			WHATSAPP: {
				enabled: true,
				name: _T("Whatsapp"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertWhatsApp.png";}
			},
			CALL: {
				enabled: true,
				name: _T("Phone Call"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertPhone.png";}
			},
		// Los canales DISABLED ---
			IM: {
				enabled: false,
				name: _T("Chat"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertIM.png";}
			},
			VOICEMAIL: {
				enabled: false,
				name: _T("Voice mail"),
				img : function(/*interaction*/) {return CHANNEL_IMG_PATH + "inconcertVoiceMail.png";}
			}
		};
		let channel_info = channels_info[type];
		if (!channel_info) {
			throw("DesktopNotifications.prototype.__get_channel_info. Unrecognized channel " + type);
		}

		return channel_info;
	};

	/*
		Funcion que realiza las validaciones para saber si la notificacion visual puede ser ejecutada.
	*/
	DesktopNotifications.prototype.__can_notify = function(message_info, callback = $.noop)
	{
		// Si la clase no ha sido instanciada.
		if (!this.m_browser_supports) {
			callback(false);
			return;
		}
		// Si aun no esta escuchando
		if (!this.m_listening) {
			callback(false);
			return;
		}
		if (this.m_has_permission === null) {
		// Si aun no se han validado permisos, valido si tiene permisos para activar las notificaciones visuales y sonoras.
			this.m_has_permission = this.update_permissions();
		}
		// Si no tiene permisos
		if (this.m_has_permission === false) {
			callback(false);
			return;
		}

		this.m_enabled_sound_notifications = inConcert.app.getUserSoundNotifications();
		this.m_enabled_visual_notifications = inConcert.app.getUserVisualNotifications();
		// Si no ha configurado ni notificaciones visuales ni sonoras
		if (!this.m_enabled_visual_notifications && !this.m_enabled_sound_notifications) {
			callback(false);
			return;
		}
		let has_focus = document.hasFocus();
		if (has_focus) {
			// Si la ventana del navegador tiene el Foco con la App Omnichannel, no debo realizar ninguna notificacion
			callback(false);
			return;
		}
		if (typeof(message_info) !== "object") {
			throw("DesktopNotifications.prototype.__can_notify - message_info is not a object.");
		}
		// Si viene dado el tipo de interaccion, valido si el mismo se encuentra enabled.
		if (message_info.interaction_type) {
			let channel_info = this.__get_channel_info(message_info.interaction_type);
			if (!channel_info.enabled) {
				callback(false);
				return;
			}
		}
		// Si tengo un callback para validar si se debe de visualizar lo ejecuto.
		if (typeof(message_info.can_show) === "function") {
			message_info.can_show(function(can_show) {
				// aca viene la respuesta del callback en true/false
				callback(can_show);
			});
			return;
		}

		// SI SI, puedo visualizar la notificacion
		callback(true);
	};

	DesktopNotifications.prototype.__add_notification_to_list = function(message_info, n)
	{
		let interaction_id = message_info.interaction_id;
		this.m_notifications_by_interactions[interaction_id] = this.m_notifications_by_interactions[interaction_id] || {};
		this.m_notifications_by_interactions[interaction_id][n.tag] = n;
		this.m_interactions_by_notifications[n.tag] = interaction_id;
	};

	DesktopNotifications.prototype.__remove_notification_of_list = function(n)
	{
		const self = this;
		window.setTimeout(function() {
			let interaction_id = self.m_interactions_by_notifications[n.tag];
			delete self.m_interactions_by_notifications[n.tag];
			delete self.m_notifications_by_interactions[interaction_id][n.tag];
			let total = Object.keys(self.m_notifications_by_interactions[interaction_id]).length;
			if (total === 0) {
				delete self.m_notifications_by_interactions[interaction_id];
			}
		});
	};

	DesktopNotifications.prototype.__remove_notifications_of_interaction = function(n)
	{
		let interaction_id = this.m_interactions_by_notifications[n.tag];
		$.each(this.m_notifications_by_interactions[interaction_id], function(_, notification) {
			notification.close();
		});
	};

	/*
		Funcion que se encarga de cerrar todas las notificaciones existentes.
	*/
	DesktopNotifications.prototype.__close_all_notifications = function()
	{
		const self = this;
		$.each(this.m_interactions_by_notifications, function(_, interaction_id) {
			self.remove_interaction_notifications(interaction_id);
		});
	};

	/*
		Funcion publica.
		Encargada de cerrar todas las notificaciones existentes para una interaccion determinada.
		La misma se invoca desde la misma clase o de forma externa.
		Externa se ejecuta, cuando una interaccion deja de estar en la bandeja de Asignadas (puede ser cuando la misma se finaliza)
		Secuencia externa:
				Application.onDataReceived
					|-> MainWindow.prototype.updateInteractions
						|-> InteractionPanel.updateInteractions
							|-> InteractionPanel.removeInteraction
								|-> AssignedInteractionsPanel.onInteractionRemoved 
									|-> AssignedInteractionsPanel.remove_interaction_desktop_notifications
										|-> DesktopNotifications.remove_interaction_notifications
	*/
	DesktopNotifications.prototype.remove_interaction_notifications = function(interaction_id)
	{
		$.each(this.m_notifications_by_interactions[interaction_id] || {}, function(_, notification) {
			notification.close();
		});
	};

	/*
		Funcion que se encarga de invocar al API de notificaciones.
	*/
	DesktopNotifications.prototype.__notify = function(message, message_info)
	{
		const self = this;

		var notify = function() {
		// A la notificacion la realiza en un timeout, asi libera el navegador y nos evitamos futuros problemas.
			window.setTimeout(function() {
				// invocamos el callback con las opciones del mensaje
				message_info.get_options(function(options) {
					// Creamos una notificacion con el mensaje y las opciones

					options.silent = true; // Silenciamos el sonido automatico de la notificacion del navegador, ya que el sonido del mismo lo realizamos en ejecutando nuestro audio (new Audio...)

					let n = new Notification(message, options);
					// Bindeamos el evento click
					n.onclick = function() {
						var onclick = message_info.onclick || $.noop;
						onclick(); // ejecutamos el evento click definido en el callback de la notificacion (opcional)
						window.focus(); // le damos foco a la ventana
						n.close(); // cerramos la notificacion
						self.__remove_notifications_of_interaction(n); // cerramos todos las notificaciones que se encuentren visibles para la interaccion.
					};
					n.onclose = function() {
						// elimina la notificacion actual de la lista interna de notificaciones.
						self.__remove_notification_of_list(n);
					};
					self.__play_sound(); // ejecuto notificacion sonora.
					self.__add_notification_to_list(message_info, n); //agrego el objeto notificacion a la lista interna de notificaciones.
					if (options.close_ms) {
						// agrego en un timeout de N ms, el close automatico de la notificacion.
						// esto es para el caso de chrome que los mismo no se cierran de forma automatica.
						window.setTimeout(n.close.bind(n), options.close_ms);
					}
				});
			});
		};

		if (this.m_enabled_visual_notifications) {
		// Si la notificacion visual se encuentra enabled, procedo a visualizarla y tambien a escuchar la notificacion sonora.
			switch (Notification.permission) {
				case "denied":
				// Si no tenemos permisos en el navegador para la notificacion visual, solo escuchamos la notificacion sonora.
					console.warn("DesktopNotifications.prototype.__notify - Notification is " + Notification.permission);
					this.__play_sound();
				break;
				case "granted":
				// Si tenemos permisos en el navegador para la notificacion visual, procedemos a visualizarla.
					notify();
				break;
				case "default":
				// Si tenemos permisos por default en el navegador para las notificaciones visuales, realizamos un request para habilitar los mismos en el navegador y proseguir.
					Notification.requestPermission(function (permission) {
						if (permission === "granted") {
							self.m_listening = true;
							notify();
						}
						else {
							self.m_listening = false;
							this.__play_sound();
						}
					});
				break;
				default:
					console.warn("DesktopNotifications.prototype.__notify - Notification is " + Notification.permission);
			}
		}
		else {
		// Si la notificacion visual se encuentra disabled, solo escuchamos la notificacion sonora.
			this.__play_sound();
		}
	};

	/*
		Funcion que invoca a la funcion que valida si la interaccion se puede notificar y en el caso que pueda realiza la visualizacion de la notificacion.
	*/
	DesktopNotifications.prototype.notify = function(message_info)
	{
		const self = this;
		this.__can_notify(message_info, function(can_show) {
			if (!can_show) {
				return;
			}
			if (typeof(message_info.message) === "string") {
			// Si el mensaje ya viene como string invoco a la funcion para visualizarlo
				self.__notify(message_info.message, message_info);
				return;
			}
			// Si viene como funcion, invoco al callback para obtener el mensaje y luego lo visualizo
			message_info.message(function(message) {
				self.__notify(message, message_info);
			});
		});
	};

	/*
		Funcion que es invocada unicamente desde la misma clase, pero si fuese necesario invocarla de forma externa, se podria invocar.
	*/
	DesktopNotifications.prototype.notify_interaction = function(interaction, options = {})
	{
		const self = this;
		if (!interaction) {
			return;
		}

		var notify = function ()
		{
			self.notify({
				interaction_id: interaction.id,
				interaction_type: interaction.type,
				can_show: function(callback = $.noop)
				{
				// Callback que es invocado para determinar si la notificacion se debe de visualizar
					var can_show_fn = options.can_show || function() {
						callback(true);
					};
					can_show_fn(callback);
				},
				message: function(callback = $.noop)
				{
				// Callback que crear el mensaje a mostrar
					let message;
					if (options.message) {
						if (typeof(options.message) === "string") {
							message = options.message;
						}
						else if (typeof(options.message) === "function") {
							message = options.message();
						}
						else {
							throw("DesktopNotifications.prototype.notify_interaction - Invalid type options.message " + options.message);
						}
					}
					else {
						let campaign_name = self.__get_campaign_name(interaction);
						
						if (campaign_name) {
							if (interaction.type.toUpperCase() === "CALL") {
								message = FORMAT(_T("New call in campaign {0}"), campaign_name);
							}
							else{
								message = FORMAT(_T("New conversation in campaign {0}"), campaign_name);
							}
						}
						else {
							if (interaction.type.toUpperCase() === "CALL") {
								message = _T("Has received a new call");
							}
							else{
								message = _T("Has received a new call");
							}
						}
					}
					callback(message);
				},
				get_options: function(callback = $.noop)
				{
				// Callback que crear el body del mensaje a mostrar
					let body;
					let close_ms = options.close_ms || 10000;

					if (options.body) {
					// Si el body ya viene dado o viene una funcion lo muestro o invoco a la funcion para obtenerlo
						if (typeof(options.body) === "string") {
							body = options.body;
						}
						else if (typeof(options.body) === "function") {
							body = options.body();
						}
						else {
							throw("DesktopNotifications.prototype.notify_interaction - Invalid type options.body " + options.body);
						}
					}
					else {
					// Si el body no viene dado, obtengo el nombre del contacto o del member
						let contact_name = interaction.getClientName();
						if (contact_name) {
							if (interaction.isInternal()) {
								body = FORMAT(_T("of user {0}"), contact_name);
							}
							else {
								body = FORMAT(_T("contact {0}"), contact_name);
							}
						}
					}

					let channel_info = self.__get_channel_info(interaction.type);
					callback({
						tag     : inConcert.Utils.generateGuid(),
						body    : body,
						icon    : channel_info.img(interaction),
						close_ms: close_ms
					});
				},
				onclick: function()
				{
					let panel = inConcert.app.getMainWindow().getPanel("AssignedInteractions");
					let $interaction = panel.$findInteraction(interaction);
					panel.openInteractionClient(
						interaction,
						$interaction
					);
				}
			});
		};

		notify();

		// if (options.no_inflate_interaction) {
		// // Si la interaccion fue inflada antes de invocar a esta funcion, no la inflo nuevamente (caso de notificaciones de nuevos eventos).
		// 	notify();
		// 	return;
		// }
		// this.__inflate_interaction(interaction, function(success) {
		// 	if (!success) {
		// 		return;
		// 	}
		// 	notify();
		// });
	};

	/*
		Funcion Publica.
		Interacciones sincronicas (Live Panel)
		La misma se encarga de notificar que un nuevo mensaje de Chat ha llegado al agente, para los canales sincronicos.
		Estos canales son: (WEB CHAT, INTERNAL MESSAGE)
		La misma es invocada desde la clase que maneja la sesion de Chat.
			Secuencia:
				ChatSession.onMessageReceived
					|-> ChatSession.notifyDesktopNewMessage 
						|-> DesktopNotifications.notify_chat_message
	*/
	DesktopNotifications.prototype.notify_chat_message = function(interaction, message)
	{
		const self = this;
		if (!interaction || !message) {
			return;
		}
		this.notify_interaction(interaction, {
			can_show: function(callback = $.noop) {
				if (message.owned) {
					callback(false);
					return;
				}
				callback(true);
			},
			message: function() {
				let channel_info = self.__get_channel_info(interaction.type);
				let msg = FORMAT(_T("New message"), channel_info.name);

				let contact_name = interaction.getClientName();
				if (contact_name) {
					if (interaction.isInternal()) {
						msg = FORMAT(_T("{0} of user {1}"), msg, contact_name);
					}
					else {
						msg = FORMAT(_T("{0} of contact {1}"), msg, contact_name);
					}
				}
				let campaign_name = self.__get_campaign_name(interaction);
				if (campaign_name) {
					msg = FORMAT(_T("{0} in campaign {1}"), msg, campaign_name);
				}
				return msg;
			},
			body: function() {
				return message.text || "";
			}
		});
	};

	DesktopNotifications.prototype.__inflate_interaction = function(interaction, callback = $.noop)
	{
		if (!interaction) {
			callback(false);
			return;
		}
		interaction.inflate(function() {
			callback(true);
		});
	};

	/*
		Funcion Publica.
		Interacciones asincronicas (Active Panel)
		Encargada de notificar los nuevos eventos que han llegado a una intearccion existente.
		La misma es utilizada para los canales que soportan eventos (interaction.supportsEvents).
		Estos canales son: (MAIL, FACEBOOK, TWITTER, SMS, WHATSAPP)
		La misma es invocada desde el DataPoller de mensajes.
			Secuencia:
				Application.onDataReceived
					|-> MainWindow.prototype.updateInteractions
						|-> InteractionPanel.prototype.updateInteractions
							|-> InteractionPanel.prototype.addInteraction
								|-> AssignedInteractionsPanel.prototype.updateInteractionData
									|-> InteractionPanel.updateInteractionData
										|-> InteractionGUI.update_interaction_data
											|-> InteractionGUI.update_monitoring_info
												|-> AssignedInteractionsPanel.onInteractionUpdated
													|-> AssignedInteractionsPanel.notify_desktop_new_events_interactions
														|-> DesktopNotifications.notify_events_interaction
	*/
	DesktopNotifications.prototype.notify_events_interaction = function(interaction, wasInflated = false)
	{
		const self = this;
		if (!interaction) {
			return;
		}
		// 1 - Valida si la interaccion soporta eventos
		if (!interaction.supportsEvents()) {
			return;
		}
		// 2 - Valida si existen nuevos eventos
		let total_new_events = interaction.getTotalNewEvents();
		if (total_new_events <= 0) {
			return;
		}

		const notify = function()
		{
			// Se obtienn solo los nuevos mensajes del contacto (los eventos entrantes)
			let new_contact_events = interaction.getNewContactEvents();
			$.each(new_contact_events, function(_, new_contact_event)
			{
				// Invoco a la funcion que realiza la notificacion, y paso un objeto con callbacks para visualizar y validar el mensaje correcto.
				let sorted_index = interaction.get_sorted_index_by_original_index(new_contact_event.__index);
				self.notify_interaction(interaction, {
					can_show: function(callback = $.noop) {
					// Callback por compatibilidad y si se desea realizar alguna validacion especifica
						callback(true);
					},
					message: function() {
					// Callback encargado de crear el mensaje
						let campaign_name = self.__get_campaign_name(interaction);
						let source = interaction.getSource(sorted_index);
						if (source) {
							return FORMAT(_T("New message of contact {0} in campaign {1}"), source, campaign_name);
						}
						return FORMAT(_T("New message in campaign {0}"), campaign_name);
					},
					body: function() {
						// Callback encargado de crear el body del mensaje
						return interaction.getEventShortDescription(sorted_index);
					},
					no_inflate_interaction: true // La interaccion ya fue inflada, entonces flagueo para que la funcion generica encargada de realizar la notificacion, no la infle nuevamente.
				});
			});
		};

		// 3 - Tiene que inflar la interaccion para poder validar cuales son estos nuevos mensajes. Si se inflo previamente, ya no se debe volver a inflar y solo se llamara el notify
		if (wasInflated === false)
		{
			this.__inflate_interaction(interaction, function(success) {
				if (!success) {
					return;
				}
				notify();
			});
		}
		else {
			notify();
		}
	};

	/*
		Funcion Publica.
		Interacciones asincronicas (Active Panel)
		Encargada de recorrer el array de notificaciones e ir notificando de a una.
		La misma es invocada desde el Panel de Interacciones asignadas (AssignedInteractionsPanel) cuando han llegado nuevas interacciones en el DataPoller de mensajes.
			Secuencia:
				Application.onDataReceived
					|-> MainWindow.updateInteractions
						|-> AssignedInteractionsPanel.notify_desktop_new_interactions
							|-> DesktopNotifications.notify_batch_interactions
	*/
	DesktopNotifications.prototype.notify_batch_interactions = function(interactions)
	{
		const self = this;

		if (!interactions || interactions.length === 0) {
			return;
		}

		// Funcion encargada de realizar un loop sobre las interacciones e ir notificandolas.
		var notify_batch_interactions = function()
		{
			let interaction;
			let interactions_by_channels = {};
			// 1 - Loopeo el array de interacciones y las categorizo por channel, para que las misma se notifiquen por canal
			for (let i = 0; i < interactions.length; i++) {
				interaction = interactions[i];
				interactions_by_channels[interaction.type] = interactions_by_channels[interaction.type] || [];
				interactions_by_channels[interaction.type].push(interaction);
			}
			// 2 - Recorro la lista de interacciones por channel y comienzo a notificar.
			$.each(interactions_by_channels, function(_, interactions_by_channel)
			{
				let total_channel = interactions_by_channel.length;
				if (total_channel === 1) {
					// Si en el array de interacciones por channel, encuentro 1 interaccion para el canal, envio derecho la notificacion.
					interaction = interactions_by_channel[0];
					self.notify_interaction(interaction);
				}
				else {
					// Si existe mas de 1 en el canal, notifico pasando como parametros funciones para obtener el mensaje y el body correspondiente segun un batch de interacciones por canal.
					self.notify_interaction(interaction, {
						message: function() {
							if (interaction.type.toUpperCase() === "CALL") {
								return FORMAT(_T("Has received a {0} new calls"), total_channel);
							}

							return FORMAT(_T("Has received a {0} new conversations"), total_channel);
						},
						body: function() {
							let total_campaigns = 0;
							let campaigns = {};
							$.each(interactions_by_channel, function(_, interaction_by_channel) {
								if (!campaigns[interaction_by_channel.campaign]) {
									total_campaigns++;
									campaigns[interaction_by_channel.campaign] = (campaigns[interaction_by_channel.campaign] || 0) + 1;
								}
							});
							if (total_campaigns === 1) {
								let campaign_name = self.__get_campaign_name(interactions_by_channel[0]);
								return FORMAT(_T("of campaign {0}"), campaign_name);
							}
							return FORMAT(_T("of {0} campaigns"), total_campaigns);
						}
					});
				}
			});
		};

		// Funcion can_show retorna siempre true, queda compatibilidad y si se llega a necesitar implementar en un futuro.
		let info = {
			can_show: function(callback = $.noop) {
				callback(true);
			}
		};

		// Aca se valida si es posible realizar las notificaciones.
		this.__can_notify(info, function(can_show) {
			if (!can_show) {
				return;
			}
			// La validacion paso correctamente, ahora comenzamos a realizar el bacth de notificaciones.
			notify_batch_interactions();
		});
	};

	/*
		Funcion que es invocada desde la App Supervisor "SupAlarmsNotifications" (supervisor_alarms_notifications.js).
	*/
	DesktopNotifications.prototype.notify_supervisor_alarm = function(campaign_id, total_new_alarms)
	{
		if (total_new_alarms === 0) {
			return;
		}
		let campaign_id_without_vcc = inConcert.Utils.getIdWithoutVCC(campaign_id);
		this.notify({
			message: _T("Supervisor - Alarms"),
			get_options: function(callback = $.noop) {
				let body;
				if (total_new_alarms === 1) {
					body = FORMAT(_T("Has received {0} new alarm in the {1} campaign."), total_new_alarms, campaign_id_without_vcc);
				}
				else {
					body = FORMAT(_T("Has received {0} new alarms in the {1} campaign."), total_new_alarms, campaign_id_without_vcc);
				}

				let img = CHANNEL_IMG_PATH + "supervisorAlarm.png";
				callback({
					tag : inConcert.Utils.generateGuid(),
					body: body,
					icon: img//,
					//requireInteraction: true, // atributo para la notificacion NO la cierre el navegador de forma automatica.
					// close_ms: 10000 // atributo nuestro interno para que cierre de forma automatica después de N MS (este atributo debe ir en conjunto con el atributo anterior "requireInteraction = true")
				});
			},
			onclick: function() {
				const mw = inConcert.app.getMainWindow();
				mw.jumpToApp("supervisor");
			}
		});
	};

}());
