module(..., package.seeall)

local utils = require "shared.apps.utils"
local m_cachedSystemApplicationActions -- estas se pueden cachear en memoria (en la misma VM) ya que las system NO cambian
local cjson = require "cjson"

local function CanValidatePermission ()
	if not inconcert.authUser or inconcert.authUser.api then
		LogDebug("[PERMISSIONS:CanValidatePermission] No validate Permission for API User: '%s'", json.encode(inconcert.authUser))
		return false
	end
	return true
end

function GetName ()
	return "Permissions"
end

local function GetRedisClient ()
	local agentModel = require "apps.agent.model"
	local ok, redis = pcall(agentModel.GetRedisClient)
	if not ok then
		LogWarning("[PERMISSIONS:GetRedisClient] Unable to retrieve Redis client.")
		return
	end
	return redis
end

--[[
]]--
local function GetAppVccOverride (appId, vccInfo)
	if not vccInfo then
		return false
	end
	local vccPermissions = vccInfo.vccPermissions
	--LogDebug("[PERMISSIONS:GetAppVccOverride] ---> vcc: '%s' vccPermissions: '%s'", vccInfo.Id, json.encode(vccPermissions or {}))
	if not vccPermissions then
		return false
	end
	return vccPermissions.override and vccPermissions.override[appId] or nil
end

--[[
]]--
local function IsEnabledAppVccOverride (appId, vccInfo)
	--LogDebug("[PERMISSIONS:IsEnabledAppVccOverride] --> vccInfo.vccPermissions: %s", json.encode(vccInfo.vccPermissions or {}))
	local resp = {
		hasPermission  = true,                             -- por defecto siempre tiene permiso en el vcc
		vccAppOverride = GetAppVccOverride(appId, vccInfo) -- valida si a nivel VCC se ha configurado
	}
	if not vccInfo then
		return resp
	end
	--LogDebug("[PERMISSIONS:IsEnabledAppVccOverride] ---> vcc: '%s' appId: '%s' ->  resp: '%s'", vccInfo.Id, appId, json.encode(resp or {}))
	if not resp.vccAppOverride then 
		return resp
	end
	if type(resp.vccAppOverride.hasPermission) == "boolean"
		and resp.vccAppOverride.hasPermission == false
	then
		LogInfo("[PERMISSIONS:IsEnabledAppVccOverride] ---> vcc: '%s' has appId: '%s' disabled", vccInfo.Id, appId)
		resp.hasPermission = false
	end
	return resp
end

--[[
]]--
function IsEnabledActionAppVccOverride (appId, actionId, vccInfo)
	--LogInfo("[PERMISSIONS:IsEnabledActionAppVccOverride] VALIDATING---> action '%s' in appId: '%s'.", actionId, appId)
	--LogDebug("[PERMISSIONS:IsEnabledActionAppVccOverride] --> vccInfo.vccPermissions: %s", json.encode(vccInfo.vccPermissions or {}))
	-- valida si a nivel VCC se ha configurado.
	local vccPermissionsConfig = IsEnabledAppVccOverride(appId, vccInfo)
	if not vccPermissionsConfig.hasPermission then
		return false -- se ha inhabilitado la app a nivel vcc
	end
	-- valida las acciones a nivel vcc
	local actions = vccPermissionsConfig.vccAppOverride and vccPermissionsConfig.vccAppOverride.actions or nil
	if actions then
		if type(actions[actionId]) == "boolean" then
			local isEnabledAction = actions[actionId]
			LogInfo("[PERMISSIONS:IsEnabledActionAppVccOverride] ---> vcc: '%s' has disabled action '%s' in appId: '%s'.", vccInfo.Id, actionId, appId)
			return isEnabledAction -- true or false
		end
	end
	-- por defecto
	return true
end

--[[
	Se cachean por 60 segundos los roles, ya que sino en la validación de permisos en batch, debería ir contra cassandra todas las veces.
]]--
local function MakeVccRolesCachedId (vcc)
	assert(vcc)
	local key = string.format(
		"Permissions:VCC:%s:Cached:Roles",
		vcc
	)
	--LogDebug("[PERMISSIONS:MakeVccRolesCachedId]------------> key: '%s'", key)
	return key
end

local function EraseVccCacheRoles (vcc)
	assert(vcc)
	--LogDebug("[PERMISSIONS:EraseVccCacheRoles] Erasing cached roles vcc: '%s'", vcc)
	local key = MakeVccRolesCachedId(vcc)
	local redis = GetRedisClient()
	if not redis then
		return
	end	
	return redis:del(key)
end

local function SetVccCacheRoles (vcc, roles)
	assert(vcc)
	assert(roles)
	--LogDebug("[PERMISSIONS:SetVccCacheRoles] Saving cache roles vcc: '%s'", vcc)
	local key = MakeVccRolesCachedId(vcc)
	local redis = GetRedisClient()
	if not redis then
		return
	end	
	local encodedRoles = cjson.encode(roles)
	redis:set(key, encodedRoles)
	-- La key expira al minuto
	local MINUTE_IN_SEC = 60
	redis:expire(key, MINUTE_IN_SEC)
end

local function GetVccCachedRoles (vcc)
	assert(vcc)
	--LogDebug("[PERMISSIONS:GetVccCachedRoles] Retrieving cached roles vcc: '%s'", vcc)
	local key = MakeVccRolesCachedId(vcc)
	local redis = GetRedisClient()
	if not redis then
		return
	end		
	local encodedRoles = redis:get(key)
	local ok, roles = pcall(cjson.decode, encodedRoles)
	if not ok then
		return
	end
	return roles
end

--[[
	TODO: Entiendo que NO habria problemas como está, pero quizás haya que revisar si tiene que obtener el Rol "deployer" !!
]]
function GetGroups (vcc, keyspace, cleanUpCache)
	assert(vcc)
	cleanUpCache = cleanUpCache or false

	local vccRoles
	if cleanUpCache then
		--LogDebug("[PERMISSIONS:GetGroups] ---> 1 %s", vcc)
		EraseVccCacheRoles(vcc) -- los elimino de redis
	end

	-- los obtengo de redis
	vccRoles = GetVccCachedRoles(vcc)
	if vccRoles then
		--LogDebug("[PERMISSIONS:GetGroups] ---> 2 %s", vcc)
		return vccRoles
	end

	local function GetVccDbRoles ()
		--LogDebug("[PERMISSIONS:GetGroups.GetVccDbRoles] load from cassandra vcc: %s", vcc)
--		LogDebug("permissions - cassandra 1 - GetVccDbRoles %s", vcc)
		return utils.GetAllColumns {
			keyspace = keyspace or "Agent",
			columnFamily = "Group",
			superColumns = true,
			key = vcc,
			columns = {} -- todos los grupos del vcc
		} or {}
	end

	-- los obtengo de cassandra
	vccRoles = GetVccDbRoles()
	local totalVccRoles = utils.GetRowCount(vccRoles)
	if totalVccRoles == 0 then
		--LogDebug("[PERMISSIONS:GetGroups] ---> 3 %s", vcc)
		return vccRoles
	end

	-- los cacheo en redis
	SetVccCacheRoles(vcc, vccRoles)
	--LogDebug("[PERMISSIONS:GetGroups] ---> 4 %s", vcc)
	return vccRoles
end

function GetGroup (fullGroupId, keyspace, vccGroups)
	--LogDebug("[PERMISSIONS:GetGroup] fullGroupId: '%s'.", fullGroupId)
	local simpleGroupId, vcc = utils.SeparateVCC(fullGroupId)
	vccGroups = vccGroups or GetGroups(vcc, keyspace)
	--LogDebug("[PERMISSIONS:GetGroup] vccGroups: '%s'.", cjson.encode(vccGroups))
	local group
	for groupId, vccGroup in pairs(vccGroups) do
		if simpleGroupId:lower() == groupId:lower() then
			group = vccGroup
			break
		end
	end
	--LogDebug("[PERMISSIONS:GetGroup] retreive group: '%s'.", cjson.encode(group or {}))
	return group, vcc
end

function SaveGroup (vcc, groups, keyspace)
	--LogDebug("[PERMISSIONS:SaveGroup] vcc: '%s' '%s'.", vcc, cjson.encode(groups))
	utils.SaveRows {
		keyspace = keyspace or "Agent",
		columnFamily = "Group",
		superColumns = true,
		rows = {
			[vcc] = groups
		}
	}
end

function RemoveGroups (vcc, groups, keyspace)
	--LogDebug("[PERMISSIONS:RemoveGroups] Removing roles in vcc: '%s' roles: '%s'.", vcc, cjson.encode(groups))
	utils.RemoveColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "Group",
		superColumns = true,
		rows = {
			[vcc] = groups
		}
	}
	EraseVccCacheRoles(vcc) -- los elimino de redis
	return true
end

--[[
	Esta funcion se encarga de obtener el GroupId tal cual se ha almacenado en la CF "Group", es decir con Mayúsculas y Minúsculas.
	Para que la comparación contra el parámetro recibido "fullGroupId" sea exacta, ya en este parámetro muchas veces llegan todos en minúsuclas o mayúsculas.
]]--
local function SanitizeGroupId (fullGroupId, keyspace)
	local group, vcc = GetGroup(fullGroupId, keyspace)
	if not group then
		return fullGroupId
	end
	sanitizedGroupId = utils.AppendVcc(group.Id, vcc)
	--LogDebug("[PERMISSIONS:SanitizeGroupId] OrignialGroupId: '%s', SanitizedGroupId: '%s'.", fullGroupId, sanitizedGroupId)
	return sanitizedGroupId
end

--[[
Group permissions are stored in the GroupPermission ColumnFamily
Each Row is the groupId, and each columnName is the application
Each columnValue is a json of the form {viewagent: true, viewcampaignchat: true, ...} 
]]
function GetGroupPermissions (fullGroupId, keyspace)
	local sanitizedGroupId = SanitizeGroupId(fullGroupId, keyspace)
	local permissions = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "GroupPermission",
		key = sanitizedGroupId
	}
	--LogDebug("[PERMISSIONS:GetGroupPermissions] --> OrignialGroupId: '%s' sanitizedGroupId: '%s' permissions: '%s'", fullGroupId, sanitizedGroupId, cjson.encode(permissions))-- and cjson.encode(permissions) or permissions)
	return permissions
end

local function GetAllGroupsSystemAndVCC (vcc, keyspace, cleanUpCache)
	cleanUpCache = cleanUpCache or false
	local system_roles = GetGroups("system", keyspace, cleanUpCache)
	local vcc_roles = GetGroups(vcc, keyspace, cleanUpCache)
	local all_roles = {}
	local function union ( roles_list, current_vcc )
		for role_id, role in pairs( roles_list ) do
			local full_role_id = utils.AppendVcc(role_id, current_vcc)
			all_roles[full_role_id] = role
		end
	end
	union( system_roles, "system" )
	union( vcc_roles, vcc )
	return all_roles
end

function LoadAllGroupPermissions (vcc, keyspace, cleanUpCache)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	cleanUpCache = cleanUpCache or false
	local roles = GetAllGroupsSystemAndVCC(vcc, keyspace, cleanUpCache)
	for roleId, role in pairs( roles ) do
		role.Permissions = GetGroupPermissions(roleId, keyspace)
	end
	return roles
end

function SaveGroupPermissions (groupId, permissions, keyspace)
	--LogDebug("[PERMISSIONS:SaveGroupPermissions] groupId: '%s' '%s'.", groupId, cjson.encode(permissions))
	utils.SaveRows {
		keyspace = keyspace or "Agent",
		columnFamily = "GroupPermission",
		rows = {
			[groupId] = permissions
		}
	}
end

function DeleteGroupPermissions (groupId, keyspace)
	utils.RemoveColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "GroupPermission",
		rows = {groupId}
	}
end

function UpdateAppRolePermissions (applicationId, fullRoleId, actionsInfo, keyspace, rolePermissions, vcc)
	assert(applicationId)
	assert(fullRoleId)
	assert(actionsInfo)
	assert(keyspace)

	LogInfo("[PERMISSIONS:UpdateAppRolePermissions] Updating App '%s' Role '%s'.", applicationId, fullRoleId)

	rolePermissions = rolePermissions or GetGroupPermissions(fullRoleId, keyspace)
	--LogDebug("[PERMISSIONS:UpdateAppRolePermissions] Role '%s' Current Permissions: '%s'.", fullRoleId, cjson.encode(rolePermissions or {}))

	if not rolePermissions then
		LogWarning("[PERMISSIONS:UpdateAppRolePermissions] Unable to Update App '%s' Role '%s' Permissions. Role Permissions does not exist.", applicationId, fullRoleId)
		return
	end

	local appPermissions = rolePermissions[applicationId] or {}
	local totalCurrentPermissions = utils.GetRowCount(appPermissions)

	LogInfo("[PERMISSIONS:UpdateAppRolePermissions] App: '%s' Role '%s' total: %d.", applicationId, fullRoleId, totalCurrentPermissions)
	--LogDebug("[PERMISSIONS:UpdateAppRolePermissions] Current Permissions: '%s'.", cjson.encode(appPermissions))

	for _, actionInfo in ipairs(actionsInfo) do
		--LogDebug("[PERMISSIONS:UpdateAppRolePermissions] 1 - (antes) permisos.%s: %s", actionInfo.Id, appPermissions[actionInfo.Id])
		if actionInfo.Enabled == "REMOVE" then
			appPermissions[actionInfo.Id] = nil
		else
			appPermissions[actionInfo.Id] = actionInfo.Enabled or nil
		end
		--LogDebug("[PERMISSIONS:UpdateAppRolePermissions] 2 - (despues) permisos.%s: %s", actionInfo.Id, appPermissions[actionInfo.Id])
	end

	-- Elimina de la CF "GroupPermissions"
	local function ErasePermissions (simpleRoleId)
	-- delete role permissions
		local vccRoleId = simpleRoleId .. "@" .. vcc
		DeleteGroupPermissions(vccRoleId, keyspace)
		LogInfo("[PERMISSIONS:UpdateAppRolePermissions] - Modified deleted of the '%s' Role in vcc '%s.", simpleRoleId, vcc )
	end

	-- Elimina de la CF "Groups"
	local function EraseRole (simpleRoleId)
		-- remove role
		local success, err = RemoveGroups(vcc, {simpleRoleId}, keyspace)
		if not success then
			LogWarning("[PERMISSIONS:UpdateAppRolePermissions] Removing roles error: %s.", cjson.encode(err))
		end
	-- delete role permissions
		ErasePermissions(simpleRoleId)
	end

	local hasAnyAction = next(appPermissions)
	if not hasAnyAction then
		rolePermissions[applicationId] = nil
		local simpleRoleId = utils.GetId(fullRoleId)
		local defaultSystemRolePermissions = GetGroupPermissions(simpleRoleId .. "@system", keyspace)
		local hasChangedSystemRole = not utils.StructuralCompare(defaultSystemRolePermissions, rolePermissions) --, CanValidateFn)

		if hasChangedSystemRole then
			keyspace:InsertColumn(fullRoleId, "GroupPermission", applicationId, cjson.encode(appPermissions))
			LogInfo("[PERMISSIONS:UpdateAppRolePermissions] -Updated App: '%s' Role '%s' Current Permissions: '%s'.", applicationId, fullRoleId, cjson.encode(appPermissions))
		else
			LogInfo("[PERMISSIONS:UpdateAppRolePermissions] Removing Custom Rol '%s'.", fullRoleId)
			EraseRole(simpleRoleId)
		end
	else
		keyspace:InsertColumn(fullRoleId, "GroupPermission", applicationId, cjson.encode(appPermissions))
		LogInfo("[PERMISSIONS:UpdateAppRolePermissions] --Updated App: '%s' Role '%s' Current Permissions: '%s'.", applicationId, fullRoleId, cjson.encode(appPermissions))
	end
end

function GetRolesHaveTheActionAssigned (applicationId, actionId, vcc, keyspace, retrieveTrueAndFalse)
	assert(applicationId)
	assert(actionId)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	retrieveTrueAndFalse = retrieveTrueAndFalse == "TRUE_AND_FALSE" and retrieveTrueAndFalse or "TRUE"

	LogInfo("[PERMISSIONS:GetRolesHaveTheActionAssigned] Retrieve roles have the app '%s' action '%s' assigned. VCC: '%s', Retrieve Type: '%s'.", applicationId, actionId, vcc, retrieveTrueAndFalse)

	local cleanUpCache = true
	local allRoles = LoadAllGroupPermissions(vcc, keyspace, cleanUpCache)
	--LogDebug("--> allRoles: '%s'", cjson.encode(allRoles))

	local rolesWhithAssignedAction = {}
	for roleId, roleInfo in pairs(allRoles) do
		for appId, permissions in pairs(roleInfo.Permissions or {}) do
			if appId == applicationId then
				if (
					retrieveTrueAndFalse == "TRUE_AND_FALSE"
					and (
							permissions[actionId] == true
							or permissions[actionId] == false
						)
					)
					or (
						retrieveTrueAndFalse == "TRUE"
						and permissions[actionId] == true
					)
				then
					table.insert(rolesWhithAssignedAction, {RoleId = roleId, Permissions = roleInfo.Permissions})
				end
			end
		end
	end

	--LogDebug("[PERMISSIONS:GetRolesHaveTheActionAssigned] Retrieving roles have the app '%s' action '%s' assigned. VCC: '%s', roles: '%s'.", applicationId, actionId, vcc, cjson.encode(rolesWhithAssignedAction))

	return rolesWhithAssignedAction
end

function GetVCCCampaigns (vcc, keyspace)
--		LogDebug("permissions - cassandra 3 - GetVCCCampaigns %s", vcc)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCCampaigns",
		key = vcc
	} or {}
end

function GetAllVCCCampaignsPermissions (vcc, keyspace)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local all_campaigns_permissions = {}
	local vcc_campaigns = GetVCCCampaigns(vcc, keyspace)
	for simple_id in pairs(vcc_campaigns) do
		local full_id = utils.AppendVcc( simple_id, vcc )
		all_campaigns_permissions[full_id] = GetCampaignPermissions( full_id, keyspace )
	end
	return all_campaigns_permissions
end

function GetUserPermissions (userId, keyspace, vcc)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local full_user_id = utils.AppendVcc( utils.GetId( userId ), vcc )
--		LogDebug("permissions - cassandra 4 - GetUserPermissions %s", userId)
	local permissions = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "UserPermission",
		key = full_user_id
	}
--	LogDebug("--> %s permissions %s", full_user_id, cjson.encode(permissions))
	return permissions
end

function GetUsersOverridePermissions (vcc, keyspace)
	assert(vcc)
--		LogDebug("permissions - cassandra 9 - GetUsersOverridePermissions %s", vcc)
	local list = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "UserPermission",
		startKey = "",
		endKey = ""
	} or {}
	local usersOverridePermissions = {}
	for fullUserId, permission in pairs(list) do
		--LogDebug("%s - %s", fullUserId, permission)
		local userId, userVcc = utils.SeparateVCC(fullUserId)
		if (userVcc or ""):lower() == vcc:lower() then
			if not utils.TableIsEmpty(permission) then
				usersOverridePermissions[fullUserId] = permission
			end
		end
	end
	return usersOverridePermissions
end

local function GetVCCUsers (vcc, keyspace)
--		LogDebug("permissions - cassandra 5 - GetVCCUsers %s", vcc)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCUsers",
		key = vcc,
		columnCount = 50000
	} or {}
end

function LoadAllVccUsersPermissions (vcc, keyspace)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local users = GetVCCUsers(vcc, keyspace)
	for user_id in pairs( users ) do
		local full_user_id = utils.AppendVcc( user_id, vcc )
		GetUserPermissions(full_user_id, keyspace, vcc)
	end
end

function GetUser (vcc, simpleId, keyspace, columns)
	simpleId = simpleId:lower()
	local user       = utils.GetAllColumns({
		keyspace     = keyspace or "Agent",
		columnFamily = "Users",
		key          = utils.AppendVcc(simpleId, vcc),
		columns      = columns
	}) or {}
	user.Avatar = nil

	return user
end

function SaveUser (vcc, user, keyspace)
	utils.SaveRows {
		keyspace = keyspace or "Agent",
		columnFamily = "Users",
		rows = {
			[utils.AppendVcc(user.Id, vcc)] = user
		}
	}
end

function SaveUserPermissions (userId, permissions, keyspace)
	--perform a remove to clean previous data
	DeleteUserPermissions(userId, keyspace)

	local vcc = inconcert.authUser.virtualContactCenter
	local simple_id = utils.GetId( userId )

	local user = GetUser(vcc, simple_id, keyspace)
	user.Id = simple_id
	user.HasRoleOverride = permissions and permissions.has_override == "1" and "1" or "0"

	SaveUser(vcc, user, keyspace)
	if keyspace then
		keyspace:InsertColumn(vcc, "VCCUsers", simple_id, json.encode(user))
	end

	utils.SaveRows {
		keyspace = keyspace or "Agent",
		columnFamily = "UserPermission",
		rows = {
			[userId] = permissions
		}
	}
end

function DeleteUserPermissions (userId, keyspace)
	local vcc = inconcert.authUser.virtualContactCenter
	local simple_id = utils.GetId( userId )

	local user = GetUser(vcc, simple_id, keyspace)
	user.Id = simple_id
	user.HasRoleOverride = "0"

	SaveUser(vcc, user, keyspace)
	if keyspace then
		keyspace:InsertColumn(vcc, "VCCUsers", simple_id, json.encode(user))
	end

	utils.RemoveColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "UserPermission",
		rows = {userId}
	}
end

function GetUserCampaigns (userId, keyspace, vcc)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local full_user_id = utils.AppendVcc(userId, vcc)
--		LogDebug("permissions - cassandra 6 - GetUserCampaigns %s", userId)
	local userCampaigns = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "UserCampaigns",
		key = full_user_id
	}
--LogDebug("GetUserCampaigns %s -> %s", userId, cjson.encode(userCampaigns or {}))
	return userCampaigns or {}
end

function GetCampaignPermissions (campaignId, keyspace)
--		LogDebug("permissions - cassandra 7 - GetCampaignPermissions %s", campaignId)
	local encoded_permissions = utils.GetColumnValue({
		keyspace     = keyspace or "Agent",
		columnFamily = "Campaigns",
		key          = campaignId,
		column       = "permissions"
	})
	if encoded_permissions then
		return cjson.decode( encoded_permissions )
	end
end

function IsModifiedSystemGroup (group_id, keyspace, vcc)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local full_group_id = utils.AppendVcc( group_id, vcc )
	--LogDebug("[PERMISSIONS:IsModifiedSystemGroup] full_group_id: '%s'", full_group_id)
	local groupPermissions = GetGroupPermissions( full_group_id, keyspace )

	local response = { modified = false }
	if not utils.TableIsEmpty(groupPermissions) then
		--LogDebug("[PERMISSIONS:IsModifiedSystemGroup] The %s system role, is modified!!", group_id)
		response.modified = true
		response.group_permissions = groupPermissions
	end

	response.default_system_group_permissions = GetGroupPermissions( group_id .. "@system", keyspace )
	return response
end

-- recibe los permisos como se levantan de la BD o cuando llegan en los post
-- y los agrupa por campanas
--
-- Recibe en "get_campaign_permissions":
--   {"members":{"cany":{"apps":{"administration":true},"actions":{"administration":{"deletecampaign":true,"save_campaign_roles_permissions":true}}}},"groups":[]}
--
-- Retorna:
--  {"campro@inconcert":{"member":{"apps":{"administration":true},"actions":{"administration":{"deletecampaign":true,"save_campaign_roles_permissions":true}}}},"campro200@inconcert":[],"camp1@inconcert":[],"camp2@inconcert":[]}
--
--
function MakeGroupedPermissionsByCampaigns (params)
	local GetRolePermissions = function( campaignId, groupsPermissions, userGroup, vcc )
		local userGroupWithVCC = utils.AppendVcc( userGroup, vcc )
		local groupPermissions = ( groupsPermissions and groupsPermissions[ userGroupWithVCC ] ) or nil
		if not groupPermissions then
			return false
		end
		return groupPermissions
	end

	--LogDebug("1 MakeGroupedPermissionsByCampaigns -> params: %s", cjson.encode(params or {}))

	local current_vcc = params.vcc or inconcert.authUser.virtualContactCenter
	local permissions = {}
	local specific_user_group = params.specific_user_group
	local user_id = params.user_id

	local campaignsPermissions = {}
	for campaignId in pairs( params.user_campaigns ) do

		local simple_id = utils.GetId( campaignId )
		local campaignId = utils.AppendVcc( simple_id, current_vcc )
		campaignsPermissions[campaignId] = campaignsPermissions[campaignId] or params.get_campaign_permissions( campaignId )
		local campaignPermissions = campaignsPermissions[campaignId]

		--LogDebug("2 MakeGroupedPermissionsByCampaigns -> campaignPermissions: %s", cjson.encode(campaignPermissions or {}))

		permissions[ campaignId ] = permissions[ campaignId ] or {}

		if campaignPermissions then
			if specific_user_group then
				-- los permisos del grupo dentro de la campana
				-- primero me fijo si el grupo es del vcc del usuario
				local group_permissions = GetRolePermissions( campaignId, campaignPermissions.groups, specific_user_group, current_vcc )
				if not group_permissions then
				-- si no es busco si el grupo pertence a un vcc del sistema
					group_permissions = GetRolePermissions( campaignId, campaignPermissions.groups, specific_user_group, "system" )
					if group_permissions then
						permissions[ campaignId ].group = group_permissions
					end
				else
					permissions[ campaignId ].group = group_permissions
				end
			else
			-- los permisos de todos los grupos
				permissions[ campaignId ].groups = campaignPermissions.groups
			end

			-- los permisos del grupo dentro de la campana
			local memberPermissions = ( campaignPermissions.members and campaignPermissions.members[ user_id ] ) or nil
			if memberPermissions then
				permissions[ campaignId ].member = memberPermissions
			end
		end
	end

	--LogDebug("3 - MakeGroupedPermissionsByCampaigns -> role: '%s', user: '%s', permissions: %s", specific_user_group, user_id, cjson.encode(permissions or {}))

	return permissions
end

--[[
User permissions are stored in the UserPermission ColumnFamily
Each Row is the userId, and each columnName is the application
Each columnValue is a json of the form:
{viewagent: true, viewcampaignchat: true, ...},
]]
function GetUserPermissionsOnCampaigns (user_id, specific_user_group, userCampaigns, keyspace, campaignsPermissions, vcc)
--	LogDebug("[PERMISSIONS:GetUserPermissionsOnCampaigns] user_id: '%s', specific_user_group: '%s'", user_id, specific_user_group)
	if not userCampaigns then
--		LogDebug("a-desde cassandra----")
		userCampaigns = GetUserCampaigns(user_id, keyspace, vcc)
	else
--		LogDebug("b-desde cache----")
	end
	return MakeGroupedPermissionsByCampaigns(
		{
			get_campaign_permissions = function(campaign_id)
				if campaignsPermissions and campaignsPermissions[campaign_id] then
				--LogDebug("1-desde cache----%s", campaign_id)
					return campaignsPermissions[campaign_id]
				end
				--LogDebug("2-desde cassandra----%s", campaign_id)
				return GetCampaignPermissions( campaign_id, keyspace )
			end,
			specific_user_group = specific_user_group,
			user_campaigns = userCampaigns,
			user_id = user_id,
			vcc = vcc
		}
	)
end

--[[
	La clave en la cual se van almacenar en REDIS las acciones custom pertenecientes al VCC.
]]--
local function MakeKeyVccApplicationsActions (vcc)
	if not vcc then
		return
	end
	local key = string.format(
		"Permissions:VCC:%s:Applications:Actions",
		vcc
	)
	--LogDebug("[PERMISSIONS:MakeKeyVccApplicationsActions]------------> key: '%s'", key)
	return key
end

--[[
	Las acciones custom pertenecientes al VCC se recuperan desde REDIS, y tienen la siguiente estructura:
	{
		apps_launcher = {
			Actions = {
				apps_launcher_cuatro = {
					ActionId          = "apps_launcher_cuatro",
					Entity            = entity,
					Domain            = { domain_id= "entities", override_type = "vcc", is_overrideable = '1' },
					Category          = "view",
					Description       = "App Cuatro",
					License = nil,
					Data = {
						webdesigner = {
							projectId = "integraction"
						}
					}
				},
				...
			}
		}
	}
]]--
local function GetVccApplicationsActions (vcc)
	vcc = vcc or inconcert.authUser and inconcert.authUser.virtualContactCenter
	if not vcc then
		return
	end
	local redis = GetRedisClient()
	if not redis then
		return
	end

	local key = MakeKeyVccApplicationsActions(vcc)
	local encodedCurrentVccAppsActions = redis:get(key)
	--LogDebug("[PERMISSIONS:GetVccApplicationsActions] -- > encodedCurrentVccAppsActions: '%s'", encodedCurrentVccAppsActions)

	local ok, currentVccAppActions = pcall(cjson.decode, encodedCurrentVccAppsActions)
	if not ok then
		return false
	end
	return currentVccAppActions
end

--[[
Application actions are stored in the Application ColumnFamily
Each Row is the applicationId, an actions are stored in the column Actions
Actions are stored as json in the following form:
Actions: {
	campaignhistory: { //Id
		Domain: "Campaign",
		Category: "Visualization",
		Description: "View campaign history",
		License: "CC_ENABLED",
		Data: null // in lua it is not stored
	},
	configchat: {
		...
	},
	...
}
]]
function GetApplicationsActions (keyspace, vcc)
	if not m_cachedSystemApplicationActions then
		--LogDebug("permissions - cassandra 8 - GetApplicationsActions SYSTEM")
		m_cachedSystemApplicationActions = utils.GetAllColumns {
			keyspace = keyspace or "Agent",
			columnFamily = "Application",
			startKey = "",
			endKey = "",
			columns = {"Actions", "Name", "IsOverrideable", "IsVisibleFor"}
		}
	end
	-- Función encargada de limpiar las Applicaciones que se hayan eliminado.
	-- Esto no debería suceder, pero .... si se eliminan, cassandra NO elimina el index de la app, entonces nos llega una app vacía, y se rompe todo.
	local function PurgeEmptyApps ()
		local purgeList = {}
		for appId, appInfo in pairs(m_cachedSystemApplicationActions) do
			--LogDebug("[PERMISSIONS.GetApplicationsActions] --> appId: '%s', appInfo: '%s'.\n", appId, cjson.encode(appInfo))
			if not appInfo or not appInfo.Actions or not appInfo.Name then
				table.insert(purgeList, appId)
			end
		end
		for _, appId in ipairs(purgeList) do
			--LogInfo("[PERMISSIONS:GetApplicationsActions] Purge appId: '%s'.", appId)
			m_cachedSystemApplicationActions[appId] = nil
		end
	end

	PurgeEmptyApps()

	local applicationActions = utils.DeepCopy(m_cachedSystemApplicationActions)

	local vccApplicationsActions = GetVccApplicationsActions(vcc)
	for appId, vccAppInfo in pairs(vccApplicationsActions or {}) do
		if applicationActions[appId] and applicationActions[appId].Actions then
			for vccActionId, vccActionInfo in pairs(vccAppInfo.Actions or {}) do
				if not applicationActions[vccActionId] then
					--LogDebug("[PERMISSIONS.GetApplicationsActions] Add Vcc Action: '%s' to App: '%s'.", vccActionId, appId)
					applicationActions[appId].Actions[vccActionId] = vccActionInfo
				end
			end
		end
	end

	--[[
		Función que filtra las applications a retornar.
			NO_INCLUDE_DEPLOYER_ROLE = retorna todos las applications menos el rol deployer.
			ONLY_INCLUDE_DEPLOYER_ROLE = retorna unicamente las applications para el rol deployer.
	]]--
	local function FilterApplications (apps)
		local role, fullRoleId
		if inconcert.authUser and inconcert.authUser.userGroup and inconcert.authUser.groupVCC then
			fullRoleId = utils.AppendVcc(inconcert.authUser.userGroup, inconcert.authUser.groupVCC)
			role = GetGroup(fullRoleId, keyspace)
			--LogDebug("[PERMISSIONS.GetApplicationsActions] 0.1 Filter Applications fullRoleId: '%s' role.IsDeployer: '%s', role.IsFullSystemAccess '%s'.", fullRoleId, (role and role.IsDeployer), (role and role.IsFullSystemAccess))
		end

		local includeType = "NO_INCLUDE_DEPLOYER_ROLE" -- Valor por defecto
		if role
			and tostring(role.IsDeployer) == "1"
		then
			includeType = "ONLY_INCLUDE_DEPLOYER_ROLE"
		end
		--LogDebug("[PERMISSIONS.GetApplicationsActions] 0.4 Filter Applications includeType: '%s'.", includeType)

		local filteredApps = {}
		for appId, app in pairs(apps) do
			--LogDebug("[PERMISSIONS.GetApplicationsActions] 0.5 Filter Applications App: '%s' IsVisibleFor: '%s'.", appId, app.IsVisibleFor)
			if app.IsVisibleFor then
				if app.IsVisibleFor.Role == includeType then
					--LogDebug("[PERMISSIONS.GetApplicationsActions] 1 Filter Applications App: '%s' includeType: '%s' SI LA INCLUYE !!.", appId, includeType)
					filteredApps[appId] = app
				else
					--LogDebug("[PERMISSIONS.GetApplicationsActions] Filter Applications includeType: '%s'. EXCLUDE App '%s' for ROLE: '%s' !!!!!!!!!!", includeType, appId, fullRoleId)
				end
			else
				--LogDebug("[PERMISSIONS.GetApplicationsActions] 3 Filter Applications App: '%s' includeType: '%s' SI LA INCLUYE !!.", appId, includeType)
				filteredApps[appId] = app
			end
		end

		return filteredApps
	end

	local filteredApplicationsActions = FilterApplications(applicationActions)
	return filteredApplicationsActions
end

function ReloadApplicationsActions (keyspace)
	LogInfo("[PERMISSIONS.ReloadApplicationsActions] Reloading cached Application Actions.")
	m_cachedSystemApplicationActions = nil
	return GetApplicationsActions(keyspace)
end

--[[
	Las acciones custom pertenecientes al VCC se almacenan en REDIS con la siguiente estructura:
	{
		apps_launcher = {
			Actions = {
				apps_launcher_cuatro = {
					ActionId          = "apps_launcher_cuatro",
					Entity            = entity,
					Domain            = { domain_id= "entities", override_type = "vcc", is_overrideable = '1' },
					Category          = "view",
					Description       = "App Cuatro",
					License = nil,
					Data = {
						webdesigner = {
							projectId = "integraction"
						}
					}
				},
				...
			}
		}
	}

	Como se puede ver se almacena en la app las acciones que se agregan de forma dinámica.
	Las mismas se recuperan cuando se obtienen las ApplicationActions (las system desde cassandra) en la función "GetApplicationsActions"
]]--
function SaveVccApplicationActions (vcc, appInfo, keyspace)
	assert(vcc)
	assert(appInfo)
	assert(appInfo.ApplicationId)
	assert(appInfo.Actions)
	assert(keyspace)

	LogInfo("[PERMISSIONS.SaveVccApplicationActions] Saving Vcc '%s' Application Actions '%s'.", vcc, cjson.encode(appInfo))

	local appId = appInfo.ApplicationId
	local appName = keyspace:GetColumnValue(appId, "Application", "Name")

	if not appName then
		LogWarning("[PERMISSIONS.SaveVccApplicationActions] Unable to Saving Vcc '%s' Application '%s'. Application does not exist.", vcc, appId)
		return false
	end

	local redis = GetRedisClient()
	if not redis then
		LogWarning("[PERMISSIONS.SaveVccApplicationActions] Unable to Saving Vcc '%s' Application '%s'. Redis Error.", vcc, appId)
		return
	end

	local key = MakeKeyVccApplicationsActions(vcc)
	local encodedCurrentVccAppsActions = redis:get(key)
	--LogDebug("[PERMISSIONS:SaveVccApplicationActions]-1-----------> encodedCurrentVccAppsActions: '%s'", encodedCurrentVccAppsActions)

	local ok, currentVccAppActions = pcall(cjson.decode, encodedCurrentVccAppsActions)
	if not ok then
		currentVccAppActions = {}
	end
	currentVccAppActions[appId] = currentVccAppActions[appId] or {Actions={}}

	local entity = {id = appId, name = appName}

	for _, actionInfo in ipairs(appInfo.Actions) do
		local actionId = actionInfo.Id
		local description = actionInfo.Description or actionId
		local domain = actionInfo.Domain or {domain_id= "entities", override_type = "vcc", is_overrideable = '1'}
		local category = actionInfo.Category or "view"
		local license = actionInfo.License or nil
		local data = actionInfo.Data or nil

		currentVccAppActions[appId].Actions[actionId] = {
			ActionId         = actionId,
			Description      = description,
			Entity           = entity,
			Domain           = domain,
			Category         = category,
			License          = license,
			Data             = data
		}
	end

	encodedCurrentVccAppsActions = cjson.encode(currentVccAppActions)
	--LogDebug("[PERMISSIONS:SaveVccApplicationActions]-2-----------> encodedCurrentVccAppsActions: '%s'", encodedCurrentVccAppsActions)
	redis:set(key, encodedCurrentVccAppsActions)

	-- Recargo las aplicacions cacheadas
	ReloadApplicationsActions(keyspace)

	return true
end

--[[
	Elimina las acciones custom del VCC, las que quedan en REDIS.
]]--
function DeleteVccApplicationActions (vcc, appInfo, keyspace)
	assert(vcc)
	assert(appInfo)
	assert(appInfo.ApplicationId)
	assert(appInfo.Actions)
	assert(keyspace)

	LogInfo("[PERMISSIONS.DeleteVccApplicationActions] Deleting Vcc '%s' Application Actions '%s'.", vcc, cjson.encode(appInfo))

	local appId = appInfo.ApplicationId
	local appName = keyspace:GetColumnValue(appId, "Application", "Name")

	if not appName then
		LogWarning("[PERMISSIONS.DeleteVccApplicationActions] Unable to Deleting Vcc '%s' Application '%s'. Application does not exist.", vcc, appId)
		return false
	end

	local redis = GetRedisClient()
	if not redis then
		LogWarning("[PERMISSIONS.DeleteVccApplicationActions] Unable to Deleting Vcc '%s' Application '%s'. Redis Error.", vcc, appId)
		return
	end

	local key = MakeKeyVccApplicationsActions(vcc)
	local encodedCurrentVccAppsActions = redis:get(key)
	--LogDebug("[PERMISSIONS:DeleteVccApplicationActions]-1-----------> encodedCurrentVccAppsActions: '%s'", encodedCurrentVccAppsActions)

	local ok, currentVccAppsActions = pcall(cjson.decode, encodedCurrentVccAppsActions)
	if not ok then
		LogWarning("[PERMISSIONS.DeleteVccApplicationActions] -Unable to Deleting Vcc '%s' Application '%s'. VCC Application Actions does not exist.", vcc, appId)
		return
	end
	if not currentVccAppsActions[appId] then
		LogWarning("[PERMISSIONS.DeleteVccApplicationActions] --Unable to Deleting Vcc '%s' Application '%s'. VCC Application Actions does not exist.", vcc, appId)
		return
	end

	for _, actionInfo in ipairs(appInfo.Actions) do
		local actionId = actionInfo.Id
		if currentVccAppsActions[appId].Actions and currentVccAppsActions[appId].Actions[actionId] then
			currentVccAppsActions[appId].Actions[actionId] = nil
		end
	end

	encodedCurrentVccAppsActions = cjson.encode(currentVccAppsActions)
	--LogDebug("[PERMISSIONS:DeleteVccApplicationActions]-2-----------> encodedCurrentVccAppsActions: '%s'", encodedCurrentVccAppsActions)
	redis:set(key, encodedCurrentVccAppsActions)

	-- Recargo las aplicacions cacheadas
	ReloadApplicationsActions(keyspace)

	return true
end

function GetApplicationActions (applicationId, keyspace)
	local applicationsActions = GetApplicationsActions( keyspace )
	return applicationsActions[ applicationId ].Actions
end

function GetUserGroupPermissions (group, keyspace, vcc)
	if not group then
		--LogDebug("[PERMISSIONS:GetUserGroupPermissions]-1")
		return {}
	end
	-- primero obtengo los rols permissions de un grupo custom
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local userGroup = utils.AppendVcc(group, vcc)
	local groupPermissions = GetGroupPermissions( userGroup, keyspace )
	if not utils.TableIsEmpty(groupPermissions) then
		---- si es un role custom, retorno sus permisos
		--LogDebug("[PERMISSIONS:GetUserGroupPermissions]-2 userGroup: '%s', groupPermissions: %s", userGroup, cjson.encode(groupPermissions))
		return groupPermissions
	end
	-- si no es custom voy a buscar los override del rol del sistema dentro del vcc
	local response = IsModifiedSystemGroup(group, keyspace, vcc)
	if response.modified then
		groupPermissions = response.group_permissions
		if groupPermissions then
			---- si el role del sistema tiene override ( fue modificado en el vcc ) retorno sus valores.
			--LogDebug("[PERMISSIONS:GetUserGroupPermissions]-3 userGroup: '%s', groupPermissions: %s", userGroup, cjson.encode(groupPermissions))
			return groupPermissions
		end
	end
	-- si el rol del sistema no tiene override, retorno sus permisos ( los por default )
	local systemGroup = utils.AppendVcc(group, "system")
	groupPermissions = GetGroupPermissions( systemGroup, keyspace ) or {}
	--LogDebug("[PERMISSIONS:GetUserGroupPermissions]-4 userGroup: '%s@system' groupPermissions: %s", group, cjson.encode(groupPermissions))
	return groupPermissions
end

local function ActionBelongsTo (app_name, action_id, override_type, keyspace)
	local application_actions = GetApplicationsActions( keyspace )
	assert( application_actions, "Permissions - Application Actions is required." )
	local app = application_actions[ app_name ]
	if not app then
		local err = LogWarning("[PERMISSIONS:ActionBelongsTo] ===> App '%s' does not exist!!!. Action '%s'.", app_name, action_id)
		--LogDebug("[PERMISSIONS:ActionBelongsTo] ===> - app '%s' Current Actions '%s'.", app_name, json.encode(application_actions))
		--assert(app, err)
		return false
	end
	local action = app.Actions[ action_id ]
	if not action then
		LogWarning("[PERMISSIONS:ActionBelongsTo] ---> Action '%s' does not exist, in app '%s' !!!", action_id, app_name)
		return false
	end
	return ( action.Domain.override_type or "" ):lower() == override_type:lower()
end

--[[
	Funcion Privada
]]--
local function _StartCategorize (permissions, keyspace, applicationsActions)
	----Get application's actions to group actions by category
	--LogDebug("[PERMISSIONS._StartCategorize] 1--> applicationsActions: %s", applicationsActions)
	local currentAppsActions
	if applicationsActions then
		currentAppsActions = applicationsActions
	else
		--LogDebug("[PERMISSIONS:_StartCategorize] 2")
		currentAppsActions = GetApplicationsActions(keyspace)
	end
	--Obtain category for each action
	local categoryByAction = {}
	for appName, app in pairs(currentAppsActions) do
		--LogDebug("[PERMISSIONS._StartCategorize] 3--> appName: '%s', app.Actions: '%s'.", appName, app.Actions)
		for actionName, action in pairs(app.Actions) do
			--LogDebug("[PERMISSIONS._StartCategorize]   3.1--> appName: '%s', actionName: '%s', app.Category: '%s'.", appName, actionName, action.Category)
			categoryByAction[actionName] = action.Category
		end
	end
	--group permissions by category
	local function GroupByCategory ( app, permissions )
		local groups = {}
		for action, enabled in pairs (permissions) do
			if not categoryByAction[action] then
				--local errMsg = LogWarning("5---> Permissions - _StartCategorize: App: '%s' Unknown action -> '%s' : categoryByAction: '%s'", app, action, categoryByAction[action])
				--assert(categoryByAction[action], errMsg)
			end
			local category = categoryByAction[action] or "Unknown"
			groups[category] = groups[category] or {}
			groups[category][action] = enabled
		end
		return groups
	end
	-- mapping callback
	local Map = function( app, permissions )
		--LogDebug("[PERMISSIONS._StartCategorize] 4--> app: '%s', permissions: '%s'.", app, permissions)
		return GroupByCategory( app, permissions )
	end
	-- start mapping
	return utils.Map( permissions or {}, Map )
end

--[[
	Funcion Publica
]]--
function Categorize (permissions, keyspace, applicationsActions)
	----Get application's actions to group actions by category
	--LogDebug("[PERMISSIONS.Categorize] --> Public funcion start.")
	return _StartCategorize(permissions, keyspace, applicationsActions)
end

local function CategorizeCampaignPermissions (permissions, keyspace, applicationsActions)
	if type( permissions ) ~= "table"  then
		return
	end

	local function CategorizeAllCampRoles()
		local categorized_roles = {}
		for role_id, role in pairs( permissions.groups or {} ) do
			categorized_roles[role_id] = _StartCategorize(role.actions or {}, keyspace, applicationsActions)
		end
		return categorized_roles
	end

	return {
		groups = CategorizeAllCampRoles(),
		group  = _StartCategorize(permissions.group and permissions.group.actions, keyspace, applicationsActions),
		member = _StartCategorize(permissions.member and permissions.member.actions, keyspace, applicationsActions)
	}
end

function CategorizeCampaignsPermissions (campaignsPermissions, keyspace, applicationsActions)
	local categorized = {}
	for campaignId, permissions in pairs( campaignsPermissions or {} ) do
		categorized[ campaignId ] = CategorizeCampaignPermissions(permissions, keyspace, applicationsActions)
	end
	return categorized
end

function CategorizeRolePermmissions (groupPermissions, keyspace, applicationsActions)
	return _StartCategorize(groupPermissions, keyspace, applicationsActions)
end

-- categoriza todos los permisos: los permisos del rol, los override a nivel vcc, los override a nivel campana/rol y campana/member
function CategorizeAllPermissions (groupPermissions, userCampaignsPermissions, userVccPermissions, keyspace, applicationsActions, custom)
	local rolePermissions
	if custom and custom.rolePermissions then
		--LogDebug("1.1 CategorizeAllPermissions roles desde cache")
		rolePermissions = custom.rolePermissions
	else
		--LogDebug("1.2 CategorizeAllPermissions roles calcuando")
		rolePermissions = CategorizeRolePermmissions(groupPermissions, keyspace, applicationsActions)
	end

	local campaignPermissions
	if custom and custom.campaignPermissions then
		--LogDebug("2.1 CategorizeAllPermissions campaigns desde cache")
		campaignPermissions = custom.campaignPermissions
	else
		--LogDebug("2.2 CategorizeAllPermissions campaigns calculando")
		campaignPermissions = CategorizeCampaignsPermissions(userCampaignsPermissions, keyspace, applicationsActions)
	end

	local vccPermissions
	if custom and custom.vccPermissions then
		--LogDebug("3.1 CategorizeAllPermissions vcc permissions desde cache")
		vccPermissions = custom.vccPermissions
	else
		--LogDebug("3.2 CategorizeAllPermissions vcc permissions calculando")
		vccPermissions = _StartCategorize(userVccPermissions and userVccPermissions.actions, keyspace, applicationsActions)
	end

	local categorizedPermissions = {
		groupPermissions = rolePermissions,
		userPermissions  = {
			campaigns = campaignPermissions,
			vcc = vccPermissions
		}
	}

	return categorizedPermissions
end

-- userPermissions and groupPermissions are optionals
local function GetSpecificUserPermissions (userId, userGroup, userCampaignsPermissions, groupPermissions, userVccPermissions, keyspace, campaignsPermissions, userCampaigns, vcc, applicationsActions)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	if not groupPermissions then
--LogDebug("1 permissions GetSpecificUserPermissions")
		groupPermissions = GetUserGroupPermissions(userGroup, keyspace, vcc)
--LogDebug("1.1 --> permissions GetSpecificUserPermissions - group: '%s@%s'groupPermissions: %s", userGroup, vcc, json.encode(groupPermissions or {}))
	end
	--now get campaign user permission for userId
	if not userCampaignsPermissions then
--LogDebug("2 permissions GetSpecificUserPermissions")
		userCampaignsPermissions = GetUserPermissionsOnCampaigns(userId, userGroup, userCampaigns, keyspace, campaignsPermissions, vcc)
		--LogDebug("--> %s userCampaignsPermissions %s", userId, cjson.encode(userCampaignsPermissions))
	end
	--now get vcc user permission for userId
	if not userVccPermissions then
--LogDebug("3 permissions GetSpecificUserPermissions")
		userVccPermissions = GetUserPermissions(userId, keyspace, vcc) or {}
		--LogDebug("--> %s userVccPermissions %s", userId, cjson.encode(userVccPermissions))
	end
--LogDebug("4 permissions GetSpecificUserPermissions applicationsActions: %s", cjson.encode(applicationsActions))
	local categorizedPermissions = CategorizeAllPermissions(groupPermissions, userCampaignsPermissions, userVccPermissions, keyspace, applicationsActions)
--LogDebug("5 permissions GetSpecificUserPermissions")
	return categorizedPermissions
end

function GetLoggedUserPermissions (keyspace, applicationsActions)
	local vcc = inconcert.authUser.virtualContactCenter
	applicationsActions = applicationsActions or GetApplicationsActions(keyspace, vcc)
	local fullUserId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	--LogInfo("[PERMISSIONS.GetLoggedUserPermissions] Get Logged User '%s' Permissions. Role: '%s'.", fullUserId, inconcert.authUser.userGroup)
	local loggedUserPermissions = GetSpecificUserPermissions(inconcert.authUser.userName, inconcert.authUser.userGroup, nil, nil, nil, keyspace, nil, nil, vcc, applicationsActions)
	--LogDebug("[PERMISSIONS.GetLoggedUserPermissions] Permissions for user %s, group %s.\nPermissions: '%s'.", fullUserId, inconcert.authUser.userGroup, cjson.encode(loggedUserPermissions))
	return loggedUserPermissions
end

--[[
	--Permissions has the following structure
	--where userPermissions are custom permissions defined by campaign
	--and group permissions are permissions defined by group
	{
		"groupPermissions" : {
			"administration" : {
				"view" : {
					"viewentitygroup" : true
				},
				"delete" : {
					"deletegroup" : true
				},
				"Groups" : {
					"viewgroups" : true
				},
				"save" : {
					"savegroup" : true
				},
				"Settings" : {
					"personalsettings" : true
				}
			}
		},
		"userPermissions" : {
			"administration" : {
				"view" : {
					"viewentitygroup" : false
				},
				"Groups" : {
					"viewgroups" : false
				},
				"Settings" : {
					"personalsettings" : true
				}
			}
		}
	}
]]

local function SearchPermission (permissions, application, action, userId, campaignId --[[ opcional --]])
	--
	local searchCategories = function ( categories )
		for categoryName, category in pairs (categories or {} ) do
			if category[ action ] ~= nil then
				return category[ action ]
			end
		end
		return nil
	end

	local findOnOverrideCampaigns = function ()
		local campaignsPermissions = permissions.userPermissions.campaigns

		local validateCampaignPermission = function( campaign_id )
			local has_permission
			local campaignPermissions = campaignsPermissions[ campaign_id ]
			if not campaignPermissions then
				return has_permission
			end
			has_permission = searchCategories( campaignPermissions.member[ application ] )
			if has_permission ~= nil then
				LogInfo("User %s has member permission %s for action %s in campaign %s", userId, has_permission and "enabled" or "disabled", action, campaign_id)
				return has_permission
			end
			has_permission = searchCategories( campaignPermissions.group[ application ] )
			if has_permission ~= nil then
				LogInfo("User %s has role permission %s for action %s in campaign %s", userId, has_permission and "enabled" or "disabled", action, campaign_id)
				return has_permission
			end

			return has_permission
		end

		--Search user's custom permission ( campaign/rol or campaign/member )
		local hasPermission
		if campaignId then
		-- busca en una campana especifica
			hasPermission = validateCampaignPermission( campaignId )
		else
		-- busca en todas las campanas
		-- si al menos una campana tiene permisos, salgo con el permiso en true
			local total_campaigns = 0
			local enable_configured = 0
			local disable_configured = 0
			local not_configured = 0

			for campaign_id, permissions in pairs( campaignsPermissions ) do
				total_campaigns = total_campaigns + 1
				local result = validateCampaignPermission( campaign_id )
				if result == true then
					enable_configured = enable_configured + 1
				elseif result == false then
					disable_configured = disable_configured + 1
				else -- result == nil
					not_configured = not_configured + 1
				end
			end
			if total_campaigns > 0 then
				if disable_configured == total_campaigns then
					hasPermission = false
				end
			end
		end
		return hasPermission
	end

	local vccPermissions = permissions.userPermissions.vcc
	local findOnOverrideVCC = function ()
		--Search user's custom permission
		local hasPermission = searchCategories( vccPermissions[ application ] )
		if hasPermission ~= nil then
			LogInfo("User %s has permission %s for action %s in vcc", userId, hasPermission and "enabled" or "disabled", action)
		end
		return hasPermission
	end

	local hasPermission
	if ActionBelongsTo( application, action, "vcc" ) then
		--Search user's custom vcc permission
		hasPermission = findOnOverrideVCC()
	else
		--Search user's custom campaign permission ( campaign/rol or campaign/member )
		hasPermission = findOnOverrideCampaigns()
	end

	if hasPermission ~= nil then
		return hasPermission
	end

	--Search user's group permissions
	hasPermission = searchCategories( permissions.groupPermissions[application] )
	if hasPermission ~= nil then
		if not hasPermission then
			LogWarning("User %s has permission DISABLED for action %s in group %s", userId, action, inconcert.authUser.userGroup)
		end
		return hasPermission
	end

	--LogDebug( "User %s has no configured permission for action %s", userId, action )
	return false
end

--[[
	application,
	action,
	campaignId  -- opcional
--]]
function IsPermissionEnabledForUser (application, action, fullCampaignId, keyspace, applicationsActions)
	if not CanValidatePermission() then
		return true
	end
	-- valida contra la configuracion del VCC.vccPermissions
	local vccInfo = inconcert and inconcert.authUser or nil
	if not IsEnabledActionAppVccOverride(application, action, vccInfo) then
		return false
	end
	local userId = inconcert.authUser.userName .. "@" .. inconcert.authUser.virtualContactCenter
	local permissions = GetLoggedUserPermissions(keyspace, applicationsActions)
	--LogDebug("[SHARED.MODEL.PERMISSIONS] IsPermissionEnabledForUser --- application: '%s', action: '%s', fullCampaignId: '%s'.", application, action, fullCampaignId)
	--LogDebug("[SHARED.MODEL.PERMISSIONS] IsPermissionEnabledForUser --- permissions: '%s'.", json.encode(permissions or {}))
	return SearchPermission(permissions, application, action, userId, fullCampaignId)
end

function IsPermissionEnabledForSpecificUser (application, action, fullUserId, userGroup, userVccPermissions, groupPermissions, userCampaignsPermissions, keyspace, applicationsActions, skipCanValidationPermission)
	if not skipCanValidationPermission and not CanValidatePermission() then
		return true
	end
	assert(application)
	assert(action)
	assert(fullUserId)
	assert(userGroup)
	-- valida contra la configuracion del VCC.vccPermissions
	local vccInfo = inconcert and inconcert.authUser or nil
	if not IsEnabledActionAppVccOverride(application, action, vccInfo) then
		return false
	end
	local userId, vcc = utils.SeparateVCC(fullUserId)
	local permissions = GetSpecificUserPermissions(userId, userGroup, userCampaignsPermissions, groupPermissions, userVccPermissions, keyspace, nil, nil, vcc, applicationsActions)
	return SearchPermission(permissions, application, action, userId)
end

--[[
	requeridos: app_id, permissions
	opcionales: campaign_id
--]]
local function HasPermissionToApplicationOverrideCampaign (app_id, permissions, campaign_id, skipCanValidationPermission)
	if not skipCanValidationPermission and not CanValidatePermission() then
		return true
	end
	assert( app_id, "Permissions - Application id is required." )
	assert( permissions, "Permissions - permissions group is required." )

	--LogDebug( "Validate permissions to '%s' application, for user '%s', campaign: '%s' ( OVERRIDE CAMPAIGN ).", app_id, full_user_id, campaign_id )

	local has_permission = false
	local group_app_permissions = (permissions.groupPermissions or {})[app_id]
	local campaigns_permissions = (permissions.userPermissions or {}).campaigns

	-- LogDebug( "- app '%s' overrride campaign_id: %s----> group_app_permissions '%s'.", app_id, campaign_id, cjson.encode(group_app_permissions) )
	-- LogDebug( "- app '%s' overrride campaign_id: %s----> campaigns_permissions '%s'.", app_id, campaign_id, cjson.encode(campaigns_permissions) )

	if not utils.TableIsEmpty(group_app_permissions) or not utils.TableIsEmpty(campaigns_permissions) then
		local template_id = "%s_%s"

		local allowed_actions = {}
		-- el peso del grupo al cual pertence la accion.
		local perm_type_weight = { group = 0, campaign_group = 1, campaign_member = 2 }

		local add_allowed_action = function( action_type, action_id, perm_type )
			local id = string.format( template_id, action_type, action_id )
			allowed_actions[ id ] = { enabled = true, weight = perm_type_weight[ perm_type ] }
		end

		local del_allowed_action = function( action_type, action_id, perm_type )
			local id = string.format( template_id, action_type, action_id )
			if allowed_actions[ id ] then
				local delete_action_weight = perm_type_weight[ perm_type ]
				if delete_action_weight > allowed_actions[ id ].weight then
				-- solo la elimina a la accion previamente agregada, si tiene un peso mayor.
					allowed_actions[ id ] = nil
				end
			end
		end

		local fill_allowed_actions = function( list, perm_type )
			list = list or {}
			for action_type, action_list in pairs( list ) do
				for action_id, enabled in pairs( action_list ) do
					if enabled then
						add_allowed_action( action_type, action_id, perm_type )
					else
						del_allowed_action( action_type, action_id, perm_type )
					end
				end
			end
		end

		-- 1 - cargo las acciones permitidas para el grupo (rol)
		fill_allowed_actions( group_app_permissions, "group" )

		-- 2 - cargo las acciones permitidas para el grupo (rol) dentro de cada campana ( realizo override del anterior )
		for cmp_id, perm in pairs( campaigns_permissions ) do
			if campaign_id == nil or ( campaign_id:lower() == cmp_id:lower() ) then
				if perm.group then
					fill_allowed_actions( perm.group[ app_id ], "campaign_group" )
				end
			end
		end

		-- 3 - cargo las acciones permitidas para el member dentro de cada campana ( realizo override de los 2 anteriores )
		for cmp_id, perm in pairs( campaigns_permissions ) do
			if campaign_id == nil or ( campaign_id:lower() == cmp_id:lower() ) then
				if perm.member then
					fill_allowed_actions( perm.member[ app_id ], "campaign_member" )
				end
			end
		end

		-- Si al menos tiene una accion permitida, quiere decir que SI tiene permiso
		for _, v in pairs( allowed_actions ) do
			has_permission = true
			break
		end
	end

	--if has_permission then
		--LogDebug("User %s have permission to application %s. ( OVERRIDE CAMPAIGN )", full_user_id, app_id )
	--end

	return has_permission
end

local function HasPermissionToApplicationOverrideVCC (app_id, full_user_id, user_group, permissions, skipCanValidationPermission)
	if not skipCanValidationPermission and not CanValidatePermission() then
		return true
	end
	assert(app_id, "Permissions - Application id is required.")

	--LogDebug("Validate permissions to '%s' application, for user '%s'  ( OVERRIDE VCC ).", app_id, full_user_id)

	local has_permission = false
	local vcc_permissions = (permissions.userPermissions or {}).vcc
	local group_permissions = (permissions.groupPermissions or {})

	if group_permissions[ app_id ] or vcc_permissions then
		local template_id = "%s_%s"
		local allowed_actions = {}

		local function add_allowed_action(action_type, action_id)
			local id = string.format(template_id, action_type, action_id)
			allowed_actions[id] = true
		end
		local function del_allowed_action(action_type, action_id)
			local id = string.format(template_id, action_type, action_id)
			allowed_actions[id] = nil
		end

		local function fill_allowed_actions(list, listType)
			list = list or {}
			local continue = true
			for action_type, action_list in pairs(list) do
				for action_id, enabled in pairs(action_list) do
					if enabled then
						add_allowed_action(action_type, action_id)
						-- Si estamos validando el Override a nivel VCC "USER_PERMISSIONS", con que exista una acción en ENABLED, ya quiere decir que tenemos ENABLED el permiso sobre la APP.
						--   Entonces salimos del bucle y no validamos más nada.
						if "USER_PERMISSIONS" == listType then
							continue = false
							break
						end
					else
						del_allowed_action(action_type, action_id)
					end
				end
				if not continue then
					break
				end
			end
		end

		fill_allowed_actions(group_permissions[app_id], "ROLE_PERMISSIONS")
		fill_allowed_actions(vcc_permissions[app_id], "USER_PERMISSIONS")

		for id, enabled in pairs( allowed_actions ) do
			if enabled then
				has_permission = true
				break
			end
		end
	end

	-- if has_permission then
		-- LogDebug("User %s have permission to application %s ( OVERRIDE VCC ).", userId, app_id)
	-- end

	return has_permission
end

function HasPermissionToApplication (app_id, user_id, user_group, userCampaignPermissions, groupPermissions, campaign_id, keyspace, userVccPermissions, campaignsPermissions, userCampaigns, applicationsActions, customPermissions, skipCanValidationPermission, vccInfo)
	if not skipCanValidationPermission and not CanValidatePermission() then
		return true
	end

	assert( app_id, "Permissions - Application id is required." )
	assert( user_id, "Permissions - User id is required." )
	assert( user_group, "Permissions - User group is required." )

	-- valida si a nivel VCC se ha configurado.
	vccInfo = vccInfo or inconcert and inconcert.authUser or nil
	local vccPermissionsConfig = IsEnabledAppVccOverride(app_id, vccInfo)
	if not vccPermissionsConfig.hasPermission then
		return false -- se ha inhabilitado la app a nivel vcc
	end

	local user_id_without_vcc, vcc = utils.SeparateVCC(user_id)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local full_user_id = utils.AppendVcc(user_id_without_vcc, vcc)
--LogDebug("1 userCampaigns %s", userCampaigns)
	userCampaigns = userCampaigns or nil
--LogDebug("2 userCampaigns %s", userCampaigns)

	local permissions
	if customPermissions then
		permissions = customPermissions
	else
		permissions = GetSpecificUserPermissions(user_id_without_vcc, user_group, userCampaignPermissions, groupPermissions, userVccPermissions, keyspace, campaignsPermissions, userCampaigns, vcc, applicationsActions)
	end

	--LogDebug("-> custom: '%s' -> %s %s %s", customPermissions ~= nil, user_id_without_vcc, user_group, json.encode(permissions))

	local has_permission
	if campaign_id then
		--LogDebug("1")
		has_permission = HasPermissionToApplicationOverrideCampaign( app_id, permissions, campaign_id, skipCanValidationPermission)
	else
		--LogDebug("2")
		has_permission = HasPermissionToApplicationOverrideVCC( app_id, full_user_id, user_group, permissions, skipCanValidationPermission)
		if not has_permission then
		--LogDebug("3")
			has_permission = HasPermissionToApplicationOverrideCampaign( app_id, permissions, nil, skipCanValidationPermission)
		end
	end

	--if not has_permission then
		--LogWarning("User %s not have permission to application %s.", full_user_id, app_id)
	--end

	return has_permission
end

function RoleHasPermissionToApplication (app_id, role_id, role_permissions, keyspace)
	if not CanValidatePermission() then
		return true
	end

	assert( app_id, "Permissions - Application id is required." )
	assert( role_id, "Permissions - Role id is required." )

	role_permissions = role_permissions or GetGroupPermissions(role_id, keyspace)

	local has_permission = false

	if role_permissions and role_permissions[ app_id ] then
		for id, enabled in pairs( role_permissions[ app_id ] ) do
			if enabled then
				has_permission = true
				break
			end
		end
	end

	LogInfo("[PERMISSIONS.RoleHasPermissionToApplication] Role %s %s have permission to application %s.", role_id, (has_permission and "" or "not "), app_id)

	return has_permission
end

function RoleHasPermissionToApplicationAction (appId, actionId, roleId, rolePermissions, keyspace)
	if not CanValidatePermission() then
		return true
	end

	assert( appId , "Permissions - Application id is required." )
	assert( roleId, "Permissions - Acttion id is required." )
	assert( roleId, "Permissions - Role id is required." )

	rolePermissions = rolePermissions or GetGroupPermissions(roleId, keyspace)

	local hasPermission = false
	local actionStatus
	if rolePermissions and rolePermissions[appId] then
		actionStatus = rolePermissions[appId][actionId]
		if actionStatus then
			hasPermission = true
		end
	end

	LogInfo("[PERMISSIONS.RoleHasPermissionToApplicationAction] Role '%s' %s have permission to application '%s' action: '%s'. Action Status: '%s'.", roleId, (hasPermission and "" or "not "), appId, actionId, (actionStatus and "ENABLED" or "DISABLED"))

	return hasPermission
end
