
local odbc = require "odbc"
local cjson = require "cjson"
local Charsets = require "shared.relational_databases.charsets"
local SQLServer = require "shared.relational_databases.sql_sqlserver"
local MySQL = require "shared.relational_databases.sql_mysql"
module(..., package.seeall)

local sql = sql or {}

local function logDs(datasource)
	return json.encode(datasource):gsub(datasource.Password, "***")
end

local function logConnection(connection)
	return logDs(connection.datasource)
end

local function getDriver(datasource)
	if datasource.Type == "sqlserver" then
		return SQLServer.Init()
	elseif datasource.Type == "mysql" then
		return MySQL.Init()
	else
		error("Unsuported DBMS " .. datasource.Type)
	end
end

sql.connect = function(datasource)
	LogInfo("[RELATIONAL_DB.sql:connect] %s", logDs(datasource))
	local driver = getDriver(datasource)
	local connectionData = driver.makeConnectionData(datasource)
	local cnn, env = odbc.driverconnect(connectionData)
	if not cnn then
		LogError("[RELATIONAL_DB.sql:connect] %s - %s", logDs(datasource), tostring(env))
		return false, env
	end
	return { datasource = datasource, cnn = cnn, env = env }
end

sql.disconnect = function(connection)
	
	if not connection then
		return
	end
	LogDebug("[RELATIONAL_DB.sql:disconnect] %s", logConnection(connection))
	if connection.cnn then
		connection.cnn:destroy()
		connection.cnn = nil
	end
	if connection.env then
		connection.env:destroy()
		connection.env = nil
	end
end

local function run(connection, query)
	query = Charsets.Init().fromutf8(query)
	LogDebug("[RELATIONAL_DB.sql:run] %s query %s", logConnection(connection), query)
	local stmt, err = connection.cnn:statement()
	if not stmt then
		LogError("[RELATIONAL_DB.sql:run] %s error opening statement for query %s - %s", 
			logConnection(connection), query, tostring(err))
		return false, err
	end
	local cur, err = stmt:execute(query)
	LogDebug("[RELATIONAL_DB.sql:run] cur %s, err %s", cur, err)
	if not cur then
		stmt:destroy()
		LogError("[RELATIONAL_DB.sql:run] %s error running query %s - %s", 
			logConnection(connection), query, tostring(err))
		return false, err
	end
	return cur, stmt
end

sql.executeWithTransaction = function(connection, query)
	connection.cnn:setautocommit(false)
	local result, err = sql.query(connection, query)
	if not result then
		connection.cnn:rollback()
	end
	connection.cnn:setautocommit(true)
	return result, err
end


sql.query = function(connection, query, options)
	options = options or {}
	local filter = options.filter
	--run
	local cur, stmt = run(connection, query)
	if not cur then
		return false, stmt
	end
	--process operation result
	if type(cur) == "number" then
		stmt:destroy()
		return { count = cur }
	end
	--process query result
	local result = {
		columns = {},
		rows = {}
	}
	--columns
	local colNames = cur:colnames()
	local colTypes = cur:coltypes()
	for index, colName in ipairs(colNames) do
		local colType = colTypes[index]
		table.insert(result.columns, { name = colName, type = colType })
	end
	--rows
	stmt:foreach("", nil, function(row_db)
		local row_a = {}
		for _, value in pairs(row_db) do
			if type(value) == "userdata" then
				value = json.null
			end
			if type(value) =="string"  then
				value = Charsets.Init().toutf8(value)
			end
			table.insert(row_a, value)
		end
		
		local row = {}
		for index, colName in ipairs(colNames) do
			row[colName] = row_a[index]
		end
		if not filter or filter(row) then
			table.insert(result.rows, row)
		end
	end)
	cur:close()
	stmt:destroy()
	return result
end

sql.connectAndQuery = function(datasource, query)
	local connection, err = sql.connect(datasource)
	if not connection then
		return false, err
	end
	local result, err = sql.query(connection, query)
	sql.disconnect(connection)
	return result, err
end

sql.escapeValue = function(datasource, value)
	local driver = getDriver(datasource)
	return driver.escapeValue(value)
end

local RELATIONAL_DB = {
	Connect         = sql.connect,
	ConnectAndQuery = sql.connectAndQuery,
	Disconnect      = sql.disconnect,
	EscapeValue     = sql.escapeValue,
	Query           = sql.query,
	ExecuteSchema	= sql.executeSchema,
}

function Init()
	return RELATIONAL_DB
end